﻿/** @file angleclipper.cpp  Angle Clipper.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "render/angleclipper.h"
#include <QVector>
#include <de/Error>
#include <de/Log>
#include <doomsday/console/var.h>
#include "dd_def.h"
#include "render/rend_main.h"
/** @file api_render.cpp  Public API of the renderer.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#define DENG_NO_API_MACROS_RENDER
#ifndef __CLIENT__
#  error "api_render.cpp is for the client only"
#endif
#include "de_platform.h"
#include "api_render.h"
#include <de/Log>
#include <doomsday/console/exec.h>
#include <doomsday/defs/sprite.h>
#include <doomsday/res/Sprites>
#include <doomsday/world/Materials>
#include "dd_main.h"  // App_Resources
#include "def_main.h"
#include "sys_system.h"  // novideo
#include "gl/sys_opengl.h"
#include "render/r_main.h"
#include "render/billboard.h"  // Rend_SpriteMaterialSpec
#include "render/rend_model.h"
#include "resource/clientresources.h"
#include "MaterialVariantSpec"
/** @file billboard.cpp  Rendering billboard "sprites".
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "de_base.h"
#include "render/billboard.h"
#include "clientapp.h"
#include "ClientTexture"
#include "gl/gl_main.h"
#include "MaterialVariantSpec"
#include "misc/r_util.h"
#include "render/rend_main.h"
#include "render/rendersystem.h"
#include "render/vissprite.h"
#include "world/p_players.h"  // viewPlayer, ddPlayers
#include <de/vector1.h>
#include <de/concurrency.h>
#include <de/GLInfo>
#include <doomsday/console/var.h>
#include <doomsday/world/Materials>
/** @file blockmapvisual.cpp  Graphical Blockmap Visual.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "de_base.h"
#include "render/blockmapvisual.h"
#include <de/aabox.h>
#include <de/concurrency.h>
#include <de/GLInfo>
#include <de/Vector>
#include "gl/gl_main.h"
#include "gl/gl_texmanager.h"
#include "world/blockmap.h"
#include "world/lineblockmap.h"
#include "world/map.h"
#include "world/p_object.h"
#include "world/p_players.h"
#include "ConvexSubspace"
#include "Face"
#include "HEdge"
#include "api_fontrender.h"
#include "render/rend_font.h"
#include "ui/ui_main.h"
/** @file cameralensfx.cpp  Camera lens effects.
 *
 * Renders camera lens effects, i.e., special effects applied to a "raw" world
 * frame. ConsoleEffect-derived isntances are put onto a stack; each console
 * has its own effect stack.
 *
 * Given the following stack of effects:
 * - A
 * - B
 * - C
 *
 * The following sequence of methods is called during the rendering of a frame:
 *  1. A.beginFrame
 *  2. B.beginFrame
 *  3. C.beginFrame
 *  4. A.draw
 *  5. B.draw
 *  6. C.draw
 *  7. C.endFrame   <-- reverse order
 *  8. B.endFrame
 *  9. A.endFrame
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de_platform.h"
#include "render/cameralensfx.h"
#include "render/rend_main.h"
#include "render/viewports.h"
#include "render/fx/bloom.h"
#include "render/fx/colorfilter.h"
#include "render/fx/lensflares.h"
#include "render/fx/ramp.h"
#include "render/fx/vignette.h"
#include "world/p_players.h"
#include "clientapp.h"
#include "ui/clientwindow.h"
#include "ui/viewcompositor.h"
#include <doomsday/console/cmd.h>
#include <de/libcore.h>
#include <de/Rectangle>
#include <de/Drawable>
#include <de/GLFramebuffer>
#include <QList>
/** @file consoleeffect.cpp
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "render/consoleeffect.h"
#include "render/viewports.h"
#include "render/rendersystem.h"
#include "world/p_players.h"
#include "clientapp.h"
/** @file decoration.cpp  World surface decoration.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2014 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 2006-2007 Jamie Jones <jamie_jones_au@yahoo.com.au>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "render/decoration.h"
#include "world/surface.h"
#include "world/clientserverworld.h"
#include <doomsday/world/MaterialManifest>
/** @file drawlist.cpp  Drawable primitive list.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "de_base.h"
#include "render/drawlist.h"
#include <de/concurrency.h>
#include <de/memoryzone.h>
#include <de/GLInfo>
#include "clientapp.h"
#include "gl/gl_main.h"
#include "render/rend_main.h"
#include "render/store.h"
/** @file drawlists.cpp  Drawable primitive list collection/management.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "de_platform.h"
#include "render/drawlists.h"
#include <de/Log>
#include <de/memoryzone.h>
#include <QMultiHash>
#include <QtAlgorithms>
/** @file environ.cpp  Environment rendering.
 *
 * @authors Copyright (c) 2015-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "render/environ.h"
#include "world/map.h"
#include "client/clientsubsector.h"
#include "clientapp.h"
#include <de/FileSystem>
#include <de/ImageFile>
#include <de/filesys/AssetObserver>
#include <doomsday/world/world.h>
/** @file lightdecoration.cpp  World surface light decoration.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2014 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 2006-2007 Jamie Jones <jamie_jones_au@yahoo.com.au>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "render/lightdecoration.h"
#include "world/convexsubspace.h"
#include "world/map.h"
#include "Surface"
#include "client/clientsubsector.h"
#include "render/rend_main.h" // Rend_ApplyLightAdaptation
#include "def_main.h"
#include <doomsday/console/var.h>
/** @file lumobj.cpp  Luminous object.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "de_platform.h"
#include "render/lumobj.h"
#include <de/vector1.h>
#include <doomsday/console/var.h>
#include "render/rend_halo.h"
#include "render/vissprite.h"
#include "world/map.h"
/** @file render/model.cpp  Drawable model.
 *
 * @authors Copyright (c) 2015-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "render/model.h"
/** @file modelloader.cpp  Model loader.
 *
 * @authors Copyright (c) 2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "render/modelloader.h"
#include "render/model.h"
#include "render/rendersystem.h"
#include "clientapp.h"
#include "busyrunner.h"
#include <de/filesys/AssetObserver>
#include <de/MultiAtlas>
#include <de/ScriptedInfo>
#include <QHash>
#include <QSet>
/** @file modelrenderer.cpp  Model renderer.
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "render/modelrenderer.h"
#include "render/modelloader.h"
#include "render/stateanimator.h"
#include "render/rend_main.h"
#include "render/vissprite.h"
#include "render/environ.h"
#include "render/rendersystem.h"
#include "gl/gl_main.h"
#include "world/p_players.h"
#include "world/p_object.h"
#include "world/bspleaf.h"
#include "world/clientmobjthinkerdata.h"
#include "world/convexsubspace.h"
#include "clientapp.h"
#include "world/clientserverworld.h"
#include <de/App>
#include <de/ModelBank>
#include <de/NativePointerValue>
#include <de/TextValue>
/** @file playerweaponanimator.cpp  Player weapon animator.
 *
 * @authors Copyright (c) 2015-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "render/playerweaponanimator.h"
#include "render/stateanimator.h"
#include "render/vissprite.h"
#include "clientplayer.h"
#include "clientapp.h"
#include "client/cl_def.h"
#include "def_main.h"
#include "render/r_main.h"
#include "render/rendersystem.h"
#include <de/timer.h>
#include <de/Garbage>
#include <de/AnimationVector>
/** @file projectionlist.cpp  Projection list.
 *
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "render/projectionlist.h"
#include <de/memoryzone.h>
/** @file r_draw.cpp  Misc Drawing Routines.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include <de/concurrency.h>
#include <doomsday/res/Textures>
#include <doomsday/world/Materials>
#include <de/GLInfo>
#include "clientapp.h"
#include "sys_system.h"
#include "render/r_main.h"
#include "render/r_draw.h"
#include "resource/clientresources.h"
#include "gl/gl_draw.h"
#include "gl/gl_main.h"
#include "gl/sys_opengl.h"
#include "api_resource.h"
#include "MaterialAnimator"
#include "world/p_players.h" // displayPlayer
/** @file r_main.cpp
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "de_platform.h"
#include <de/vector1.h>
#include <de/GLInfo>
#include <de/GLState>
#include <de/GLFramebuffer>
#include <doomsday/defs/sprite.h>
#include <doomsday/world/Materials>
#include <doomsday/res/Sprites>
#include "dd_def.h"  // finesine
#include "clientapp.h"
#include "gl/gl_main.h"
#include "render/billboard.h"
#include "render/modelrenderer.h"
#include "render/r_main.h"
#include "render/rend_main.h"
#include "render/rend_model.h"
#include "render/rendersystem.h"
#include "render/vissprite.h"
#include "resource/clientresources.h"
#include "world/map.h"
#include "world/p_players.h"
#include "BspLeaf"
#include "ConvexSubspace"
#include "client/clientsubsector.h"
/** @file r_things.cpp  Map Object => Vissprite Projection.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 2006 Jamie Jones <jamie_jones_au@yahoo.com.au>
 * @authors Copyright © 1993-1996 by id Software, Inc.
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "de_platform.h"
#include "render/r_things.h"
#include <de/vector1.h>
#include <de/ModelDrawable>
#include <doomsday/defs/sprite.h>
#include <doomsday/world/Materials>
#include "clientapp.h"
#include "dd_main.h"  // App_World()
#include "dd_loop.h"  // frameTimePos
#include "def_main.h"  // states
#include "misc/r_util.h"
#include "gl/gl_main.h"
#include "gl/gl_tex.h"
#include "gl/gl_texmanager.h"  // GL_PrepareFlaremap
#include "network/net_main.h"  // clients[]
#include "render/rendersystem.h"
#include "render/r_main.h"
#include "render/angleclipper.h"
#include "render/stateanimator.h"
#include "render/rend_halo.h"
#include "render/vissprite.h"
#include "world/map.h"
#include "world/p_object.h"
#include "world/p_players.h"
#include "world/clientmobjthinkerdata.h"
#include "BspLeaf"
#include "ConvexSubspace"
#include "client/clientsubsector.h"
/** @file rend_fakeradio.cpp  Geometry generation for faked, radiosity lighting.
 *
 * @authors Copyright © 2004-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "render/rend_fakeradio.h"
#include <de/Vector>
#include <doomsday/console/var.h>
#include "clientapp.h"
#include "gl/gl_texmanager.h"
#include "MaterialAnimator"
#include "MaterialVariantSpec"
#include "world/map.h"
#include "ConvexSubspace"
#include "Line"
#include "Surface"
#include "client/clientsubsector.h"
#include "Face"
#include "HEdge"
#include "WallEdge"
#include "render/rend_main.h"
#include "render/rendersystem.h"
#include "render/r_main.h"  // levelFullBright
#include "render/drawlist.h"
#include "render/shadowedge.h"
#include "render/viewports.h"  // R_FrameCount()
#include "render/store.h"
/** @file rend_font.cpp  Font Renderer.
 *
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#define DENG_NO_API_MACROS_FONT_RENDER
#include "de_base.h"
#include "render/rend_font.h"
#include "sys_system.h" // novideo
#include <cctype>
#include <cstdio>
#include <cstdlib>
#include <de/concurrency.h>
#include <de/GLState>
#include <de/GLInfo>
#include "BitmapFont"
#include "CompositeBitmapFont"
#include "gl/gl_main.h"
#include "gl/gl_draw.h"
#include "gl/gl_texmanager.h"
#include "api_fontrender.h"
#include "render/rend_main.h"
/** @file rend_halo.cpp Halos and Lens Flares.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de_platform.h"
#include "dd_main.h"
#include "render/rend_halo.h"
#include "render/rend_main.h"
#include "render/fx/bloom.h"
#include "gl/gl_main.h"
#include "gl/gl_texmanager.h"
#include "world/p_players.h"
#include <de/concurrency.h>
#include <de/GLState>
#include <de/GLInfo>
#include <doomsday/console/var.h>
#include <doomsday/console/cmd.h>
#include <doomsday/doomsdayapp.h>
#include <cmath>
#define NUM_FLARES          5
/** @file rend_main.cpp  World Map Renderer.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 2006 Jamie Jones <jamie_jones_au@yahoo.com.au>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "de_platform.h"
#include "render/rend_main.h"
#include "MaterialVariantSpec"
#include "ClientTexture"
#include "Face"
#include "world/map.h"
#include "world/blockmap.h"
#include "world/lineowner.h"
#include "world/p_object.h"
#include "world/p_players.h"
#include "world/sky.h"
#include "world/thinkers.h"
#include "world/surface.h"
#include "BspLeaf"
#include "Contact"
#include "ConvexSubspace"
#include "client/clientsubsector.h"
#include "client/clskyplane.h"
#include "LightDecoration"
#include "Lumobj"
#include "SkyFixEdge"
#include "TriangleStripBuilder"
#include "WallEdge"
#include "gl/gl_main.h"
#include "gl/gl_tex.h"  // pointlight_analysis_t
#include "gl/gl_texmanager.h"
#include "gl/sys_opengl.h"
#include "api_fontrender.h"
#include "misc/r_util.h"
#include "render/fx/bloom.h"
#include "render/fx/vignette.h"
#include "render/fx/lensflares.h"
#include "render/angleclipper.h"
#include "render/blockmapvisual.h"
#include "render/billboard.h"
#include "render/cameralensfx.h"
#include "render/modelrenderer.h"
#include "render/r_main.h"
#include "render/r_things.h"
#include "render/rendersystem.h"
#include "render/rend_fakeradio.h"
#include "render/rend_halo.h"
#include "render/rend_particle.h"
#include "render/rendpoly.h"
#include "render/skydrawable.h"
#include "render/store.h"
#include "render/viewports.h"
#include "render/vissprite.h"
#include "render/vr.h"
#include "ui/editors/rendererappearanceeditor.h"
#include "ui/editors/modelasseteditor.h"
#include "ui/ui_main.h"
#include "ui/postprocessing.h"
#include "sys_system.h"
#include "dd_main.h"
#include "clientapp.h"
#include "network/net_main.h"
#include <doomsday/console/cmd.h>
#include <doomsday/console/var.h>
#include <doomsday/defs/sprite.h>
#include <doomsday/res/TextureManifest>
#include <doomsday/world/Materials>
#include <doomsday/BspNode>
#include <de/concurrency.h>
#include <de/timer.h>
#include <de/texgamma.h>
#include <de/vector1.h>
#include <de/GLInfo>
#include <de/GLState>
#include <QtAlgorithms>
#include <QBitArray>
#include <cmath>
#include <cstdio>
#include <cstdlib>
/** @file rend_model.cpp  3D Model Rendering (frame models).
 *
 * @note Light vectors and triangle normals are in an entirely independent,
 *       right-handed coordinate system.
 *
 * There is some more confusion with Y and Z axes as the game uses Z as the
 * vertical axis and the rendering code and model definitions use the Y axis.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "clientapp.h"
#include "dd_def.h"
#include "dd_main.h" // App_World()
#include "world/p_players.h"
#include "world/clientmobjthinkerdata.h"
#include "render/rend_model.h"
#include "render/rend_main.h"
#include "render/rendersystem.h"
#include "render/vissprite.h"
#include "render/vectorlightdata.h"
#include "render/modelrenderer.h"
#include "gl/gl_main.h"
#include "gl/gl_texmanager.h"
#include "MaterialVariantSpec"
#include "ClientTexture"
#include "ClientMaterial"
#include <doomsday/console/var.h>
#include <doomsday/world/Materials>
#include <de/Log>
#include <de/ArrayValue>
#include <de/GLInfo>
#include <de/binangle.h>
#include <de/memory.h>
#include <de/concurrency.h>
#include <cstdlib>
#include <cmath>
#include <cstring>
/** @file rend_particle.cpp  Particle effect rendering.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "de_base.h"
#include "render/rend_particle.h"
#include "gl/gl_main.h"
#include "gl/gl_texmanager.h"
#include "gl/texturecontent.h"
#include "world/map.h"
#include "world/p_players.h"
#include "BspLeaf"
#include "ConvexSubspace"
#include "Line"
#include "Plane"
#include "client/clientsubsector.h"
#include "resource/image.h"
#include "render/r_main.h"
#include "render/viewports.h"
#include "render/rend_main.h"
#include "render/rend_model.h"
#include "render/vissprite.h"
#include "clientapp.h"
#include "misc/r_util.h"
#include "sys_system.h"  // novideo
#include <doomsday/console/var.h>
#include <doomsday/filesys/fs_main.h>
#include <de/concurrency.h>
#include <de/vector1.h>
#include <de/Folder>
#include <de/GLInfo>
#include <de/ImageFile>
#include <cstdlib>
/** @file rendersystem.cpp  Renderer subsystem.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de_platform.h"
#include "render/rendersystem.h"
#include <de/memory.h>
#include <de/memoryzone.h>
#include <de/Function>
#include <de/GLUniform>
#include <de/PackageLoader>
#include <de/ScriptSystem>
#include <de/ScriptedInfo>
#include <de/TextValue>
#include "clientapp.h"
#include "render/environ.h"
#include "render/rend_main.h"
#include "render/rend_halo.h"
#include "render/angleclipper.h"
#include "render/modelrenderer.h"
#include "render/skydrawable.h"
#include "render/store.h"
#include "world/clientserverworld.h"
#include "gl/gl_main.h"
#include "gl/gl_texmanager.h"
#include "ConvexSubspace"
#include "Subsector"
#include "Surface"
#include "Contact"
#include "misc/r_util.h"
/** @file rendpoly.cpp  RendPoly data buffers.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "de_base.h"
#include "render/rendpoly.h"
#include "misc/color.h"
#include "WallEdge"
/** @file render/shadervar.cpp  Shader variable.
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "render/shadervar.h"
#include <de/ScriptedInfo>
#include <de/ScriptSystem>
/** @file shadowedge.cpp  FakeRadio Shadow Edge Geometry
 *
 * @authors Copyright © 2004-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "render/shadowedge.h"
#include "Face"
#include "HEdge"
#include "ConvexSubspace"
#include "Plane"
#include "Sector"
#include "Surface"
#include "world/lineowner.h"
#include "client/clientsubsector.h"
#include "render/rend_main.h"
#include "MaterialAnimator"
#include "WallEdge"
/** @file skydrawable.cpp  Drawable specialized for the sky.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "render/skydrawable.h"
#include "clientapp.h"
#include "client/cl_def.h" // clientPaused
#include "gl/gl_main.h"
#include "gl/gl_tex.h"
#include "MaterialVariantSpec"
#include "resource/clientresources.h"
#include "resource/framemodeldef.h"
#include "render/rend_main.h"
#include "render/rend_model.h"
#include "render/vissprite.h"
#include "world/sky.h"
#include "ClientTexture"
#include <doomsday/console/var.h>
#include <doomsday/console/exec.h>
#include <doomsday/world/Materials>
#include <de/concurrency.h>
#include <de/timer.h>
#include <de/Log>
#include <de/GLInfo>
#include <cmath>
#define MAX_LAYERS  2
#define MAX_MODELS  32
/** @file skyfixedge.cpp  Sky Fix Edge Geometry.
 *
 * @authors Copyright © 2011-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "render/skyfixedge.h"
#include "Face"
#include "world/map.h"
#include "world/maputil.h" // R_SideBackClosed, remove me
#include "world/p_players.h"
#include "ConvexSubspace"
#include "Plane"
#include "Sector"
#include "Surface"
#include "MaterialAnimator"
#include "render/rend_main.h"
#include "client/clientsubsector.h"
#include "client/clskyplane.h"
/** @file stateanimator.cpp
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "render/stateanimator.h"
#include "render/rendersystem.h"
#include "render/modelloader.h"
#include "render/shadervar.h"
#include "clientapp.h"
#include "dd_loop.h"
#include <doomsday/world/thinkerdata.h>
#include <de/ConditionalTrigger>
#include <de/ScriptedInfo>
#include <de/ScriptSystem>
#include <de/RecordValue>
#include <de/NoneValue>
#include <de/NativePointerValue>
#include <de/GLUniform>
#include <QSet>
/** @file store.cpp  Store.
 *
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "render/store.h"
#include <de/libcore.h>
#include <de/memory.h>
/** @file render/trianglestripbuilder.cpp Triangle Strip Geometry Builder.
 *
 * @authors Copyright © 2011-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include <QScopedPointer>
#include "render/trianglestripbuilder.h"
/** @file vectorlightlist.cpp Vector light list.
 *
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "render/vectorlightlist.h"
#include <de/memoryzone.h>
/** @file viewports.cpp  Player viewports and related low-level rendering.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de_platform.h"
#include "render/viewports.h"
#include <QBitArray>
#include <de/concurrency.h>
#include <de/timer.h>
#include <de/vector1.h>
#include <de/GLInfo>
#include <de/GLState>
#include <doomsday/filesys/fs_util.h>
#include "clientapp.h"
#include "api_console.h"
#include "dd_main.h"
#include "dd_loop.h"
#include "gl/gl_main.h"
#include "api_render.h"
#include "render/angleclipper.h"
#include "render/cameralensfx.h"
#include "render/fx/bloom.h"
#include "render/playerweaponanimator.h"
#include "render/r_draw.h"
#include "render/r_main.h"
#include "render/rendersystem.h"
#include "render/rendpoly.h"
#include "render/skydrawable.h"
#include "render/vissprite.h"
#include "render/vr.h"
#include "network/net_demo.h"
#include "world/linesighttest.h"
#include "world/thinkers.h"
#include "world/p_object.h"
#include "world/p_players.h"
#include "world/sky.h"
#include "BspLeaf"
#include "ConvexSubspace"
#include "Surface"
#include "Contact"
#include "client/clientsubsector.h"
#include "ui/ui_main.h"
#include "ui/clientwindow.h"
/** @file vissprite.cpp  Projected visible sprite ("vissprite") management.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2016 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 2006 Jamie Jones <jamie_jones_au@yahoo.com.au>
 * @authors Copyright © 1993-1996 by id Software, Inc.
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "render/vissprite.h"
#include "clientapp.h"
#include "world/map.h"
#include "world/bspleaf.h"
#include "world/convexsubspace.h"
#include "client/clientsubsector.h"
/** @file vr.cpp  Stereoscopic rendering and Oculus Rift support.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2015 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 2013 Christopher Bruns <cmbruns@rotatingpenguin.com>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de_base.h"
#include "render/vr.h"
#include <de/timer.h>
#include <de/BaseGuiApp>
#include <doomsday/console/cmd.h>
#include <doomsday/console/var.h>
#include "ui/clientwindow.h"
/** @file walledge.cpp  Wall Edge Geometry.
 *
 * @authors Copyright © 2011-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "de_base.h"
#include "render/walledge.h"
#include "BspLeaf"
#include "ConvexSubspace"
#include "Sector"
#include "world/lineowner.h"
#include "world/p_players.h"
#include "world/maputil.h"
#include "world/surface.h"
#include "client/clientsubsector.h"
#include "MaterialAnimator"
#include "render/rend_main.h"  /// devRendSkyMode @todo remove me
#include "Face"
#include <QtAlgorithms>
/** @file render/wallspec.cpp Wall Geometry Specification.
 *
 * @authors Copyright © 2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "de_base.h"
#include "MaterialAnimator"
#include "Sector"
#include "Surface"
#include "world/p_players.h" // viewPlayer
#include "render/rend_main.h"
#include "render/walledge.h"
using namespace de;
using namespace world;
namespace internal {
    /**
     * @param point  @em View-relative point in map-space.
     */
    static inline binangle_t pointToAngle(Vector2d const &point) {
        return bamsAtan2(dint(point.y * 100), dint(point.x * 100));
    }
    /**
     * Simple data structure for pooling POD elements. Note that unlike a traditional
     * ObjectPool (pattern), the pooled elements are @em not owned by the pool!
     */
    class ElementPool {
    public:
        /**
         * Base for POD elements.
         */
        struct Element {
        private:
            Element *_prev, *_next;
            friend class ElementPool;
        };
        /**
         * Begin reusing elements in the pool.
         */
        void rewind() {
            _rover = _first;
        }
        /**
         * Add a new @em unused object to the pool.
         *
         * @param elem  Element to be linked in the pool. Ownership is unaffected.
         */
        void add(Element *elem) {
            if(!_last) _last = elem;
            if(_first) _first->_prev = elem;
            elem->_next = _first;
            elem->_prev = nullptr;
            _first = elem;
        }
        /**
         * Returns a pointer to the next unused element in the pool; otherwise @c nullptr.
         */
        void *get() {
            if(!_rover) return nullptr;
            Element *next = _rover;
            _rover = _rover->_next;
            return next;
        }
        /**
         * Release the element @a elem (@important which is assumed to have been added
         * previously!), moving it to the list of used elements, for later reuse.
         */
        void release(Element *elem) {
            DENG2_ASSERT(_last);
            if(elem == _last) {
                DENG2_ASSERT(!_rover);
                _rover = elem;
                return;
            }
            DENG2_ASSERT(elem->_next);
            elem->_next->_prev = elem->_prev;
            if(elem->_prev) {
                elem->_prev->_next = elem->_next;
            }
            else {
                _first = _first->_next;
                _first->_prev = nullptr;
            }
            _last->_next = elem;
            elem->_prev = _last;
            elem->_next = nullptr;
            _last = elem;
            if(!_rover) {
                _rover = _last;
            }
        }
    private:
        Element *_first = nullptr;
        Element *_last  = nullptr;
        Element *_rover = nullptr;
    };
}  // namespace internal
using namespace ::internal;
DENG2_PIMPL_NOREF(AngleClipper) {
    struct Clipper : public ElementPool::Element, AngleRange {
        Clipper *prev;
        Clipper *next;
    };
    ElementPool clipNodes;         ///< The list of clipnodes.
    Clipper *clipHead = nullptr;   ///< Head of the clipped-range list.
    struct Occluder : public ElementPool::Element, AngleRange {
        Occluder *prev;
        Occluder *next;
        bool topHalf;              ///< @c true= top, rather than bottom, half.
        Vector3f normal;           ///< Of the occlusion plane.
    };
    ElementPool occNodes;          ///< The list of occlusion nodes.
    Occluder *occHead = nullptr;   ///< Head of the occlusion-range list.
    QVector<binangle_t> angleBuf;  ///< Scratch buffer for sorting angles.
    ~Impl() {
        clearRangeList(&clipHead);
        clearRangeList(&occHead);
    }
    template <typename NodeType>
    void clearRangeList(NodeType **head) {
        DENG2_ASSERT(head);
        while(*head) {
            auto *next = static_cast<NodeType *>((*head)->next);
            delete *head;
            *head = next;
        }
    }
    /**
     * The specified range must be safe!
     */
    dint isRangeVisible(binangle_t from, binangle_t to) const {
        for(Clipper *i = clipHead; i; i = i->next) {
            if(from >= i->from && to <= i->to)
                return false;
        }
        return true;
    }
    /**
     * @return  Non-zero iff the range is not entirely clipped; otherwise @c 0.
     */
    dint safeCheckRange(binangle_t from, binangle_t to) const {
        if(from > to) {
            return (isRangeVisible(from, BANG_MAX) || isRangeVisible(0, to));
        }
        return isRangeVisible(from, to);
    }
    void removeRange(Clipper *crange) {
        if(clipHead == crange)
            clipHead = crange->next;
        if(crange->prev)
            crange->prev->next = crange->next;
        if(crange->next)
            crange->next->prev = crange->prev;
        clipNodes.release(crange);
    }
    Clipper *newClipNode(binangle_t from, binangle_t to) {
        auto *crange = reinterpret_cast<Clipper *>(clipNodes.get());
        if(!crange) {
            clipNodes.add(crange = new Clipper);
        }
        crange->from = from;
        crange->to   = to;
        crange->prev = nullptr;
        crange->next = nullptr;
        return crange;
    }
    void addRange(binangle_t from, binangle_t to) {
        cutOcclusionRange(from, to);
        if(!clipHead) {
            clipHead = newClipNode(from, to);
            /*
            LOG_AS("AngleClipper::addRange");
            LOG_DEBUG(String("New head added: %1 => %2")
                        .arg(clipHead->from, 0, 16)
                        .arg(clipHead->to,   0, 16));
            */
            return;
        }
        for(Clipper *i = clipHead; i; i = i->next) {
            /*
            LOG_AS("AngleClipper::addRange");
            LOG_DEBUG(String("0x%1: %2 => %3")
                        .arg((quintptr)i, QT_POINTER_SIZE * 2, 16, QChar('0'))
                        .arg(i->from, 0, 16)
                        .arg(i->to,   0, 16));
            */
            if(from >= i->from && to <= i->to) {
                /*
                LOG_AS("AngleClipper::addRange");
                LOG_DEBUG("Range already exists");
                */
                return;  // The new range already exists.
            }
#ifdef DENG2_DEBUG
            if(i == i->next)
                throw Error("AngleClipper::addRange", String("loop1 0x%1 linked to itself: %2 => %3")
                                                        .arg((quintptr)i, QT_POINTER_SIZE * 2, 16, QChar('0'))
                                                        .arg(i->from, 0, 16)
                                                        .arg(i->to,   0, 16));
#endif
        }
        for(Clipper *i = clipHead; i;) {
            if(i->from >= from && i->to <= to) {
                Clipper *contained = i;
                /*
                LOG_AS("AngleClipper::addRange");
                LOG_DEBUG(String("Removing contained range %1 => %2")
                            .arg(contained->from, 0, 16)
                            .arg(contained->to,   0, 16));
                */
                i = i->next;
                removeRange(contained);
                continue;
            }
            i = i->next;
        }
        Clipper *crange = nullptr;
        for(Clipper *i = clipHead; i; i = i->next) {
            if(i->from < to) {
                crange = i;
            }
            if(i->from >= from && i->from <= to) {
                /*
                LOG_AS("AngleClipper::addRange");
                LOG_DEBUG(String("Overlapping start: %1 => %2 - adjusting to %3 => %4")
                            .arg(i->from, 0, 16)
                            .arg(i->to,   0, 16)
                            .arg(from,    0, 16)
                            .arg(i->to,   0, 16));
                */
                i->from = from;
                return;
            }
            if(i->to >= from && i->to <= to) {
                /*
                LOG_AS("AngleClipper::addRange");
                LOG_DEBUG(String("Overlapping end: %1 => %2")
                            .arg(i->from, 0, 16)
                            .arg(i->to,   0, 16));
                */
                crange = i->next;
                if(!crange) {
                    i->to = to;
                    /*
                    LOG_AS("AngleClipper::addRange");
                    LOG_DEBUG(String("No next, adjusting end (now %1 => %2)")
                                .arg(i->from, 0, 16)
                                .arg(i->to,   0, 16));
                    */
                }
                else {
                    if(crange->from <= to) {
                        i->to = crange->to;
                        /*
                        LOG_AS("AngleClipper::addRange");
                        LOG_DEBUG(String("merging with the next (%1 => %2)")
                                    .arg(crange->from, 0, 16)
                                    .arg(crange->to,   0, 16));
                        */
                        removeRange(crange);
                    }
                    else {
                        i->to = to;
                        /*
                        LOG_AS("AngleClipper::addRange");
                        LOG_DEBUG(String("Not merger w/next (now %1 => %2)")
                                    .arg(i->from, 0, 16)
                                    .arg(i->to,   0, 16));
                        */
                    }
                }
                return;
            }
        }
        if(!crange) {
            crange = clipHead;
            clipHead = newClipNode(from, to);
            clipHead->next = crange;
            if(crange)
                crange->prev = clipHead;
        }
        else {
            Clipper *added = newClipNode(from, to);
            added->next = crange->next;
            if(added->next)
                added->next->prev = added;
            added->prev = crange;
            crange->next = added;
        }
    }
    void removeOcclusionRange(Occluder *orange) {
        if(occHead == orange)
            occHead = orange->next;
        if(orange->prev)
            orange->prev->next = orange->next;
        if(orange->next)
            orange->next->prev = orange->prev;
        occNodes.release(orange);
    }
    Occluder *newOcclusionRange(binangle_t from, binangle_t to, Vector3f const &normal,
        bool topHalf) {
        auto *orange = reinterpret_cast<Occluder *>(occNodes.get());
        if(!orange) {
            occNodes.add(orange = new Occluder);
        }
        orange->from    = from;
        orange->to      = to;
        orange->prev    = nullptr;
        orange->next    = nullptr;
        orange->topHalf = topHalf;
        orange->normal  = normal;
        return orange;
    }
    /**
     * @pre The given range is "safe".
     */
    void addOcclusionRange(binangle_t from, binangle_t to, Vector3f const &normal,
        bool topHalf) {
        if(from > to) return;
        Occluder *newor = newOcclusionRange(from, to, normal, topHalf);
        if(!occHead) {
            occHead = newor;
            occHead->next = occHead->prev = nullptr;
            return;
        }
        Occluder *after = nullptr;
        for(Occluder *orange = occHead; orange; orange = orange->next) {
            if(orange->from > from) {
                newor->next = orange;
                newor->prev = orange->prev;
                orange->prev = newor;
                if(newor->prev)
                    newor->prev->next = newor;
                else
                    occHead = newor;  // We have a new head.
                return;
            }
            after = orange;
        }
        after->next = newor;
        newor->prev = after;
        newor->next = nullptr;
    }
    /**
     * If necessary, cut the given range in two.
     */
    void safeAddOcclusionRange(binangle_t startAngle, binangle_t endAngle,
                               Vector3f const &normal, bool tophalf) {
        if(!safeCheckRange(startAngle, endAngle)) return;
        if(startAngle > endAngle) {
            addOcclusionRange(startAngle, BANG_MAX, normal, tophalf);
            DENG2_DEBUG_ONLY(occlusionRanger(3));
            addOcclusionRange(0, endAngle, normal, tophalf);
            DENG2_DEBUG_ONLY(occlusionRanger(4));
        }
        else {
            addOcclusionRange(startAngle, endAngle, normal, tophalf);
            DENG2_DEBUG_ONLY(occlusionRanger(5));
        }
    }
    /**
     * Attempts to merge the two given occnodes.
     *
     * @return @c 0= Could not be merged.
     *         @c 1= orange was merged into other.
     *         @c 2= other was merged into orange.
     */
    dint tryMergeOccludes(Occluder *orange, Occluder *other) {
        if(!orange->normal.z) return 0;
        Vector3f cross = orange->normal.cross(other->normal);
        if(!cross.x && !cross.y && !cross.z) {
            removeOcclusionRange(orange);
            return 1;
        }
        binangle_t crossAngle = bamsAtan2(dint(cross.y), dint(cross.x));
        if(crossAngle >= orange->from && crossAngle <= orange->to)
            return 0;  // Inside the range, can't do a thing.
        crossAngle += BANG_180;
        if(crossAngle >= orange->from && crossAngle <= orange->to)
            return 0;  // Inside the range, can't do a thing.
        crossAngle = (orange->from + orange->to) >> (1 + BAMS_BITS - 13);
        cross.x = 100 * FIX2FLT(fineCosine[crossAngle]);
        cross.y = 100 * FIX2FLT(finesine  [crossAngle]);
        cross.z = -(orange->normal.x * cross.x +
                    orange->normal.y * cross.y) / orange->normal.z;
        if(cross.dot(other->normal) < 0) {
            removeOcclusionRange(other);
            return 2;
        }
        else {
            removeOcclusionRange(orange);
            return 1;
        }
    }
    /**
     * Try to merge oranges with matching ranges. (Quite a number may be produced
     * as a result of the cuts.)
     */
    void mergeOccludes() {
        for(Occluder *orange = occHead; orange && orange->next; ) {
            auto *next = orange->next;
            for(Occluder *other = next; other && orange->from == other->from; other = other->next) {
                if(other->topHalf != orange->topHalf) continue;
                if(orange->to != other->to) continue;
                dint result = tryMergeOccludes(orange, other);
                if(result == 2) {
                    next = next->next;
                }
                break;
            }
            orange = next;
        }
    }
    /**
     * Everything in the given range is removed from the occlusion nodes.
     */
    void cutOcclusionRange(binangle_t from, binangle_t to) {
        DENG2_DEBUG_ONLY(occlusionRanger(1));
        Occluder *after  = nullptr;
        for(Occluder *orange = occHead; orange && orange->from < to; orange = orange->next) {
            after = orange;
        }
        for(Occluder *orange = occHead; orange; ) {
            auto *next = orange->next;
            if(from <= orange->to) {
                if(orange->from >= to) break;
                switch(orange->relationship(AngleRange(from, to))) {
                case 0:  // The cut range completely includes this orange.
                    removeOcclusionRange(orange);
                    break;
                case 1:  // The cut range contains the beginning of the orange.
                    orange->from = to;
                    break;
                case 2:  // The cut range contains the end of the orange.
                    orange->to = from;
                    break;
                case 3: {  // The orange contains the whole cut range.
                    Occluder *part = newOcclusionRange(to, orange->to, orange->normal,
                                                       orange->topHalf);
                    part->prev = after;
                    if(after) {
                        part->next = after->next;
                        after->next = part;
                    }
                    else {
                        part->next = occHead;
                        occHead = part;
                    }
                    if(part->next)
                        part->next->prev = part;
                    orange->to = from;
                    break; }
                default:  // No meaningful relationship (in this context).
                    break;
                }
            }
            orange = next;
        }
        DENG2_DEBUG_ONLY(occlusionRanger(2));
        mergeOccludes();
        DENG2_DEBUG_ONLY(occlusionRanger(6));
    }
#ifdef DENG2_DEBUG
    void occlusionLister() {
        for(Occluder *orange = occHead; orange; orange = orange->next) {
            LOG_MSG(String("from: %1 to: %2 topHalf: %3")
                        .arg(orange->from, 0, 16)
                        .arg(orange->to,   0, 16)
                        .arg(DENG2_BOOL_YESNO(orange->topHalf)));
        }
    }
    void occlusionRanger(int mark) {
        for(Occluder *orange = occHead; orange; orange = orange->next) {
            if(orange->prev && orange->prev->from > orange->from) {
                occlusionLister();
                throw Error("AngleClipper::occlusionRanger", String("Order %1 has failed").arg(mark));
            }
        }
    }
#endif
};
AngleClipper::AngleClipper() : d(new Impl)
{}
dint AngleClipper::isFull() const {
    if(::devNoCulling) return false;
    return d->clipHead && d->clipHead->from == 0 && d->clipHead->to == BANG_MAX;
}
dint AngleClipper::isAngleVisible(binangle_t bang) const {
    if(::devNoCulling) return true;
    for(Impl::Clipper const *crange = d->clipHead; crange; crange = crange->next) {
        if(bang > crange->from && bang < crange->to)
            return false;
    }
    return true;  // Not occluded.
}
dint AngleClipper::isPointVisible(Vector3d const &point) const {
    if(::devNoCulling) return true;
    Vector3d const viewRelPoint = point - Rend_EyeOrigin().xzy();
    binangle_t const angle      = pointToAngle(viewRelPoint);
    if(!isAngleVisible(angle)) return false;
    for(Impl::Occluder const *orange = d->occHead; orange; orange = orange->next) {
        if(angle >= orange->from && angle <= orange->to) {
            if(orange->from > angle)
                return true;  // No more possibilities.
            if(viewRelPoint.dot(orange->normal) > 0)
                return false;
        }
    }
    return true;  // Not occluded.
}
dint AngleClipper::isPolyVisible(Face const &poly) const {
    DENG2_ASSERT(poly.isConvex());
    if(::devNoCulling) return true;
    if(poly.hedgeCount() > d->angleBuf.count()) {
        d->angleBuf.resize(poly.hedgeCount());
    }
    Vector2d const eyeOrigin = Rend_EyeOrigin().xz();
    dint n = 0;
    HEdge const *hedge = poly.hedge();
    do {
        d->angleBuf[n++] = pointToAngle(hedge->origin() - eyeOrigin);
    } while((hedge = &hedge->next()) != poly.hedge());
    for(dint i = 0; i < poly.hedgeCount() - 1; ++i) {
        binangle_t angLen = d->angleBuf.at(i + 1) - d->angleBuf.at(i);
        if(angLen == BANG_180) return true;
        if(angLen < BANG_180) {
            if(d->safeCheckRange(d->angleBuf.at(i), d->angleBuf.at(i + 1)))
                return true;
        }
        else {
            if(d->safeCheckRange(d->angleBuf.at(i + 1), d->angleBuf.at(i)))
                return true;
        }
    }
    return false;  // Completely occluded.
}
void AngleClipper::clearRanges() {
    d->clipHead = nullptr;
    d->clipNodes.rewind();  // Start reusing ranges.
    d->occHead = nullptr;
    d->occNodes.rewind();   // Start reusing ranges.
}
dint AngleClipper::safeAddRange(binangle_t from, binangle_t to) {
    if(from > to) {
        d->addRange(from, BANG_MAX);
        d->addRange(0, to);
    }
    else {
        d->addRange(from, to);
    }
    return true;
}
void AngleClipper::addRangeFromViewRelPoints(Vector2d const &from, Vector2d const &to) {
    Vector2d const eyeOrigin = Rend_EyeOrigin().xz();
    safeAddRange(pointToAngle(to   - eyeOrigin),
                 pointToAngle(from - eyeOrigin));
}
void AngleClipper::addViewRelOcclusion(Vector2d const &from, Vector2d const &to,
    coord_t height, bool topHalf) {
    Vector3d const eyeOrigin    = Rend_EyeOrigin().xzy();
    auto const eyeToV1          = Vector3d(from, height) - eyeOrigin;
    auto const eyeToV2          = Vector3d(to,   height) - eyeOrigin;
    binangle_t const startAngle = pointToAngle(eyeToV2);
    binangle_t const endAngle   = pointToAngle(eyeToV1);
    if(startAngle == endAngle) return;
    Vector3f const normal = (topHalf? eyeToV2 : eyeToV1).cross(topHalf? eyeToV1 : eyeToV2);
#ifdef DENG2_DEBUG
    if(Vector3f(0, 0, (topHalf ? 1000 : -1000)).dot(normal) < 0) {
        LOG_AS("AngleClipper::addViewRelOcclusion");
        LOGDEV_GL_WARNING("Wrong side v1:%s v2:%s eyeOrigin:%s!")
                << from.asText() << to.asText()
                << Vector2d(eyeOrigin).asText();
        DENG2_ASSERT(!"Failed AngleClipper::addViewRelOcclusion: Side test");
    }
#endif
    d->safeAddOcclusionRange(startAngle, endAngle, normal, topHalf);
}
dint AngleClipper::checkRangeFromViewRelPoints(Vector2d const &from, Vector2d const &to) {
    if(::devNoCulling) return true;
    Vector2d const eyeOrigin = Rend_EyeOrigin().xz();
    return d->safeCheckRange(pointToAngle(to   - eyeOrigin) - BANG_45/90,
                             pointToAngle(from - eyeOrigin) + BANG_45/90);
}
#ifdef DENG2_DEBUG
void AngleClipper::validate() {
    for(Impl::Clipper *i = d->clipHead; i; i = i->next) {
        if(i == d->clipHead) {
            if(i->prev)
                throw Error("AngleClipper::validate", "Cliphead->prev != NULL");
        }
        if(i->prev) {
            if(i->prev->next != i)
                throw Error("AngleClipper::validate", "Prev->next != this");
        }
        else if(i != d->clipHead) {
            throw Error("AngleClipper::validate", "prev == NULL, this isn't clipHead");
        }
        if(i->next) {
            if(i->next->prev != i)
                throw Error("AngleClipper::validate", "Next->prev != this");
        }
    }
}
#endif
#undef M_ScreenShot
DENG_EXTERN_C dint M_ScreenShot(char const *name, dint flags);
#undef Models_CacheForState
DENG_EXTERN_C void Models_CacheForState(dint stateIndex) {
    if (FrameModelDef *modelDef = App_Resources().modelDefForState(stateIndex)) {
        App_Resources().cache(modelDef);
    }
}
#undef R_SetBorderGfx
DENG_EXTERN_C void R_SetBorderGfx(struct uri_s const *const *paths);
#undef Rend_CacheForMobjType
DENG_EXTERN_C void Rend_CacheForMobjType(dint num) {
    LOG_AS("Rend.CacheForMobjType");
    if (::novideo) return;
    if (!((::useModels && ::precacheSkins) || ::precacheSprites)) return;
    if (num < 0 || num >= ::runtimeDefs.mobjInfo.size()) return;
    de::MaterialVariantSpec const &spec = Rend_SpriteMaterialSpec();
    for (dint i = 0; i < DED_Definitions()->states.size(); ++i) {
        if (::runtimeDefs.stateInfo[i].owner != &::runtimeDefs.mobjInfo[num])
            continue;
        Models_CacheForState(i);
        if (::precacheSprites) {
            if (state_t *state = Def_GetState(i)) {
                App_Resources().cache(state->sprite, spec);
            }
        }
    }
}
#undef R_RenderPlayerView
#undef R_SetViewOrigin
#undef R_SetViewAngle
#undef R_SetViewPitch
#undef R_ViewWindowGeometry
#undef R_ViewWindowOrigin
#undef R_ViewWindowSize
#undef R_SetViewWindowGeometry
#undef R_ViewPortGeometry
#undef R_ViewPortOrigin
#undef R_ViewPortSize
#undef R_SetViewPortPlayer
DENG_EXTERN_C void R_RenderPlayerView(dint num);
DENG_EXTERN_C void R_SetViewOrigin(dint consoleNum, coord_t const origin[3]);
DENG_EXTERN_C void R_SetViewAngle(dint consoleNum, angle_t angle);
DENG_EXTERN_C void R_SetViewPitch(dint consoleNum, dfloat pitch);
DENG_EXTERN_C dint R_ViewWindowGeometry(dint consoleNum, RectRaw *geometry);
DENG_EXTERN_C dint R_ViewWindowOrigin(dint consoleNum, Point2Raw *origin);
DENG_EXTERN_C dint R_ViewWindowSize(dint consoleNum, Size2Raw *size);
DENG_EXTERN_C void R_SetViewWindowGeometry(dint consoleNum, RectRaw const *geometry, dd_bool interpolate);
DENG_EXTERN_C dint R_ViewPortGeometry(dint consoleNum, RectRaw *geometry);
DENG_EXTERN_C dint R_ViewPortOrigin(dint consoleNum, Point2Raw *origin);
DENG_EXTERN_C dint R_ViewPortSize(dint consoleNum, Size2Raw *size);
DENG_EXTERN_C void R_SetViewPortPlayer(dint consoleNum, dint viewPlayer);
#undef R_SkyParams
DENG_EXTERN_C void R_SkyParams(dint layer, dint param, void *data);
static inline MaterialVariantSpec const &pspriteMaterialSpec_GetSpriteInfo() {
    return App_Resources().materialSpec(PSpriteContext, 0, 1, 0, 0, GL_CLAMP_TO_EDGE, GL_CLAMP_TO_EDGE,
                                             0, -2, -1, false, true, true, false);
}
#undef R_GetSpriteInfo
DENG_EXTERN_C dd_bool R_GetSpriteInfo(dint id, dint frame, spriteinfo_t *info) {
    if (!info) return false;
    de::zapPtr(info);
    auto const *sprDef = res::Sprites::get().spritePtr(id, frame);
    if (!sprDef) {
        LOG_RES_WARNING("Invalid sprite id:%i and/or frame:%i") << id << frame;
        return false;
    }
    defn::Sprite sprite(*sprDef);
    if (!sprite.hasView(0)) {
        LOG_RES_WARNING("Sprite id:%i frame:%i has no front view")
            << id << frame;
        return false;
    }
    defn::Sprite::View const spriteView = sprite.view(0);
    info->material = world::Materials::get().materialPtr(*spriteView.material);
    info->flip     = spriteView.mirrorX;
    if (::novideo) return true;  // We can't prepare the material.
    MaterialAnimator &matAnimator = reinterpret_cast<world::Material *>(info->material)->as<ClientMaterial>()
            .getAnimator(pspriteMaterialSpec_GetSpriteInfo());
    matAnimator.prepare();  // Ensure we have up-to-date info.
    Vector2ui const &matDimensions = matAnimator.dimensions();
    TextureVariant *tex            = matAnimator.texUnit(MaterialAnimator::TU_LAYER0).texture;
    Vector2i const &texDimensions  = tex->base().origin();
    dint const texBorder           = tex->spec().variant.border;
    info->geometry.origin.x    = -texDimensions.x + -texBorder;
    info->geometry.origin.y    = -texDimensions.y +  texBorder;
    info->geometry.size.width  = matDimensions.x + texBorder * 2;
    info->geometry.size.height = matDimensions.y + texBorder * 2;
    tex->glCoords(&info->texCoord[0], &info->texCoord[1]);
    return true;
}
#undef R_ChooseAlignModeAndScaleFactor
#undef R_ChooseScaleMode2
#undef R_ChooseScaleMode
DENG_EXTERN_C dd_bool R_ChooseAlignModeAndScaleFactor(dfloat *scale, dint width, dint height, dint availWidth, dint availHeight, scalemode_t scaleMode);
DENG_EXTERN_C scalemode_t R_ChooseScaleMode2(dint width, dint height, dint availWidth, dint availHeight, scalemode_t overrideMode, dfloat stretchEpsilon);
DENG_EXTERN_C scalemode_t R_ChooseScaleMode(dint width, dint height, dint availWidth, dint availHeight, scalemode_t overrideMode);
#undef R_SetupFog
DENG_EXTERN_C void R_SetupFog(dfloat start, dfloat end, dfloat density, dfloat const *rgb) {
    DENG2_ASSERT(rgb);
    Con_Execute(CMDS_DDAY, "fog on", true, false);
    Con_Executef(CMDS_DDAY, true, "fog start %f", start);
    Con_Executef(CMDS_DDAY, true, "fog end %f", end);
    Con_Executef(CMDS_DDAY, true, "fog density %f", density);
    Con_Executef(CMDS_DDAY, true, "fog color %.0f %.0f %.0f",
                 rgb[0] * 255, rgb[1] * 255, rgb[2] * 255);
}
#undef R_SetupFogDefaults
DENG_EXTERN_C void R_SetupFogDefaults() {
    Con_Execute(CMDS_DDAY, "fog off", true, false);
}
DENG_DECLARE_API(Rend) = {
    { DE_API_RENDER },
    R_SetupFogDefaults,
    R_SetupFog,
    Rend_CacheForMobjType,
    Models_CacheForState,
    R_RenderPlayerView,
    R_SetViewOrigin,
    R_SetViewAngle,
    R_SetViewPitch,
    R_ViewWindowGeometry,
    R_ViewWindowOrigin,
    R_ViewWindowSize,
    R_SetViewWindowGeometry,
    R_SetBorderGfx,
    R_ViewPortGeometry,
    R_ViewPortOrigin,
    R_ViewPortSize,
    R_SetViewPortPlayer,
    R_ChooseAlignModeAndScaleFactor,
    R_ChooseScaleMode2,
    R_ChooseScaleMode,
    R_GetSpriteInfo,
    R_SkyParams,
    M_ScreenShot
};
dint spriteLight = 4;
dfloat maxSpriteAngle = 60;
dbyte noSpriteTrans;
dint useSpriteAlpha = 1;
dint useSpriteBlend = 1;
dint alwaysAlign;
dint noSpriteZWrite;
dbyte devNoSprites;
static inline void drawQuad(dgl_vertex_t *v, dgl_color_t *c, dgl_texcoord_t *tc) {
    DGL_Begin(DGL_QUADS);
        DGL_Color4ubv(c[0].rgba);
        DGL_TexCoord2fv(0, tc[0].st);
        DGL_Vertex3fv(v[0].xyz);
        DGL_Color4ubv(c[1].rgba);
        DGL_TexCoord2fv(0, tc[1].st);
        DGL_Vertex3fv(v[1].xyz);
        DGL_Color4ubv(c[2].rgba);
        DGL_TexCoord2fv(0, tc[2].st);
        DGL_Vertex3fv(v[2].xyz);
        DGL_Color4ubv(c[3].rgba);
        DGL_TexCoord2fv(0, tc[3].st);
        DGL_Vertex3fv(v[3].xyz);
    DGL_End();
}
void Rend_DrawMaskedWall(drawmaskedwallparams_t const &parms) {
    DENG2_ASSERT_IN_RENDER_THREAD();
    DENG_ASSERT_GL_CONTEXT_ACTIVE();
    TextureVariant *tex = nullptr;
    if(::renderTextures) {
        MaterialAnimator *matAnimator = parms.animator;
        DENG2_ASSERT(matAnimator);
        matAnimator->prepare();
        tex = matAnimator->texUnit(MaterialAnimator::TU_LAYER0).texture;
    }
    bool withDyn = false;
    dint normal = 0, dyn = 1;
    if(parms.modTex) {
        if(IS_MUL) {
            normal = 1;
            dyn    = 0;
        }
        else {
            normal = 0;
            dyn    = 1;
        }
        GL_SelectTexUnits(2);
        DGL_ModulateTexture(IS_MUL? 4 : 5);
        DGL_SetInteger(DGL_ACTIVE_TEXTURE, IS_MUL? 0 : 1);
        GL_BindTextureUnmanaged(renderTextures ? parms.modTex : 0,
                                gl::ClampToEdge, gl::ClampToEdge);
        DGL_SetModulationColor(parms.modColor);
        DGL_SetInteger(DGL_ACTIVE_TEXTURE, IS_MUL? 1 : 0);
        GL_BindTexture(tex);
        withDyn = true;
    }
    else {
        DGL_ModulateTexture(1);
        DGL_Enable(DGL_TEXTURE_2D);
        GL_BindTexture(tex);
        normal = 0;
    }
    GL_BlendMode(parms.blendMode);
    byte normalTarget = normal? 1 : 0;
    byte dynTarget    =    dyn? 1 : 0;
    if(withDyn) {
        DGL_Begin(DGL_QUADS);
            DGL_Color4fv(parms.vertices[0].color);
            DGL_TexCoord2f(normalTarget, parms.texCoord[0][0], parms.texCoord[1][1]);
            DGL_TexCoord2f(dynTarget, parms.modTexCoord[0][0], parms.modTexCoord[1][1]);
            DGL_Vertex3f(parms.vertices[0].pos[0],
                         parms.vertices[0].pos[2],
                         parms.vertices[0].pos[1]);
            DGL_Color4fv(parms.vertices[1].color);
            DGL_TexCoord2f(normalTarget, parms.texCoord[0][0], parms.texCoord[0][1]);
            DGL_TexCoord2f(dynTarget, parms.modTexCoord[0][0], parms.modTexCoord[0][1]);
            DGL_Vertex3f(parms.vertices[1].pos[0],
                         parms.vertices[1].pos[2],
                         parms.vertices[1].pos[1]);
            DGL_Color4fv(parms.vertices[3].color);
            DGL_TexCoord2f(normalTarget, parms.texCoord[1][0], parms.texCoord[0][1]);
            DGL_TexCoord2f(dynTarget, parms.modTexCoord[1][0], parms.modTexCoord[0][1]);
            DGL_Vertex3f(parms.vertices[3].pos[0],
                         parms.vertices[3].pos[2],
                         parms.vertices[3].pos[1]);
            DGL_Color4fv(parms.vertices[2].color);
            DGL_TexCoord2f(normalTarget, parms.texCoord[1][0], parms.texCoord[1][1]);
            DGL_TexCoord2f(dynTarget, parms.modTexCoord[1][0], parms.modTexCoord[1][1]);
            DGL_Vertex3f(parms.vertices[2].pos[0],
                         parms.vertices[2].pos[2],
                         parms.vertices[2].pos[1]);
        DGL_End();
        GL_SelectTexUnits(1);
        DGL_ModulateTexture(1);
    }
    else {
        DGL_Begin(DGL_QUADS);
            DGL_Color4fv(parms.vertices[0].color);
            DGL_TexCoord2f(0, parms.texCoord[0][0], parms.texCoord[1][1]);
            DGL_Vertex3f(parms.vertices[0].pos[0],
                       parms.vertices[0].pos[2],
                       parms.vertices[0].pos[1]);
            DGL_Color4fv(parms.vertices[1].color);
            DGL_TexCoord2f(0, parms.texCoord[0][0], parms.texCoord[0][1]);
            DGL_Vertex3f(parms.vertices[1].pos[0],
                       parms.vertices[1].pos[2],
                       parms.vertices[1].pos[1]);
            DGL_Color4fv(parms.vertices[3].color);
            DGL_TexCoord2f(0, parms.texCoord[1][0], parms.texCoord[0][1]);
            DGL_Vertex3f(parms.vertices[3].pos[0],
                       parms.vertices[3].pos[2],
                       parms.vertices[3].pos[1]);
            DGL_Color4fv(parms.vertices[2].color);
            DGL_TexCoord2f(0, parms.texCoord[1][0], parms.texCoord[1][1]);
            DGL_Vertex3f(parms.vertices[2].pos[0],
                       parms.vertices[2].pos[2],
                       parms.vertices[2].pos[1]);
        DGL_End();
    }
    DGL_Disable(DGL_TEXTURE_2D);
    GL_BlendMode(BM_NORMAL);
}
/**
 * Set all the colors in the array to that specified.
 */
static void applyUniformColor(dint count, dgl_color_t *colors, dfloat const *rgba) {
    for(; count-- > 0; colors++) {
        colors->rgba[0] = dbyte(255 * rgba[0]);
        colors->rgba[1] = dbyte(255 * rgba[1]);
        colors->rgba[2] = dbyte(255 * rgba[2]);
        colors->rgba[3] = dbyte(255 * rgba[3]);
    }
}
/**
 * Calculate vertex lighting.
 */
static void Spr_VertexColors(dint count, dgl_color_t *out, dgl_vertex_t *normals,
    duint lightListIdx, dint maxLights, dfloat const *_ambient) {
    DENG2_ASSERT(out && normals && _ambient);
    dbyte const opacity = 255 * _ambient[3];
    Vector3f const ambient(_ambient);
    Vector3f const saturated(1, 1, 1);
    Vector3ub colorClamped;
    for(dint i = 0; i < count; ++i) {
        if(maxLights > 0) {
            Vector3f const normal(normals[i].xyz);
            Vector3f accum[2];  // Begin with total darkness [color, extra].
            dint numProcessed = 0;
            ClientApp::renderSystem().forAllVectorLights(lightListIdx, [&maxLights, &normal
                                         , &accum, &numProcessed](VectorLightData const &vlight) {
                numProcessed += 1;
                dfloat strength = vlight.direction.dot(normal) + vlight.offset;  // Shift toward the light a little.
                if(strength > 0) strength *= vlight.lightSide;
                else             strength *= vlight.darkSide;
                accum[vlight.affectedByAmbient ? 0 : 1]
                    += vlight.color * de::clamp(-1.f, strength, 1.f);
                return (maxLights && numProcessed == maxLights);
            });
            colorClamped = ((accum[0].max(ambient) + accum[1]).min(saturated) * 255).toVector3ub();
        }
        else if(i == 0) {
            colorClamped = (ambient.min(saturated) * 255).toVector3ub();
        }
        out[i].rgba[0] = colorClamped.x;
        out[i].rgba[1] = colorClamped.y;
        out[i].rgba[2] = colorClamped.z;
        out[i].rgba[3] = opacity;
    }
}
MaterialVariantSpec const &PSprite_MaterialSpec() {
    return App_Resources().materialSpec(SpriteContext, 0, 0, 0, 0, GL_CLAMP_TO_EDGE, GL_CLAMP_TO_EDGE,
                                             1, -2, 0, false, true, true, false);
}
void Rend_DrawPSprite(rendpspriteparams_t const &parms) {
    DENG2_ASSERT_IN_RENDER_THREAD();
    DENG_ASSERT_GL_CONTEXT_ACTIVE();
    if(::renderTextures == 1) {
        GL_SetPSprite(parms.mat, 0, 0);
        DGL_Enable(DGL_TEXTURE_2D);
    }
    else if(::renderTextures == 2) {
        MaterialAnimator &matAnimator = ClientMaterial::find(de::Uri("System", Path("gray")))
                .getAnimator(PSprite_MaterialSpec());
        matAnimator.prepare();
        GL_BindTexture(matAnimator.texUnit(MaterialAnimator::TU_LAYER0).texture);
        DGL_Enable(DGL_TEXTURE_2D);
    }
    dfloat v1[] = { parms.pos[0],               parms.pos[1] };
    dfloat v2[] = { parms.pos[0] + parms.width, parms.pos[1] };
    dfloat v3[] = { parms.pos[0] + parms.width, parms.pos[1] + parms.height };
    dfloat v4[] = { parms.pos[0],               parms.pos[1] + parms.height };
    Vector3f const &frontVec = viewPlayer->viewport().frontVec;
    dgl_vertex_t quadNormals[4];
    for(dint i = 0; i < 4; ++i) {
        quadNormals[i].xyz[0] = frontVec.x;
        quadNormals[i].xyz[1] = frontVec.z;
        quadNormals[i].xyz[2] = frontVec.y;
    }
    dgl_color_t quadColors[4];
    if(!parms.vLightListIdx) {
        applyUniformColor(4, quadColors, parms.ambientColor);
    }
    else {
        Spr_VertexColors(4, quadColors, quadNormals, parms.vLightListIdx, spriteLight + 1, parms.ambientColor);
    }
    dgl_texcoord_t tcs[4], *tc = tcs;
    dgl_color_t *c = quadColors;
    tc[0].st[0] = parms.texOffset[0] *  (parms.texFlip[0]? 1:0);
    tc[0].st[1] = parms.texOffset[1] *  (parms.texFlip[1]? 1:0);
    tc[1].st[0] = parms.texOffset[0] * (!parms.texFlip[0]? 1:0);
    tc[1].st[1] = parms.texOffset[1] *  (parms.texFlip[1]? 1:0);
    tc[2].st[0] = parms.texOffset[0] * (!parms.texFlip[0]? 1:0);
    tc[2].st[1] = parms.texOffset[1] * (!parms.texFlip[1]? 1:0);
    tc[3].st[0] = parms.texOffset[0] *  (parms.texFlip[0]? 1:0);
    tc[3].st[1] = parms.texOffset[1] * (!parms.texFlip[1]? 1:0);
    DGL_Begin(DGL_QUADS);
        DGL_Color4ubv(c[0].rgba);
        DGL_TexCoord2fv(0, tc[0].st);
        DGL_Vertex2fv(v1);
        DGL_Color4ubv(c[1].rgba);
        DGL_TexCoord2fv(0, tc[1].st);
        DGL_Vertex2fv(v2);
        DGL_Color4ubv(c[2].rgba);
        DGL_TexCoord2fv(0, tc[2].st);
        DGL_Vertex2fv(v3);
        DGL_Color4ubv(c[3].rgba);
        DGL_TexCoord2fv(0, tc[3].st);
        DGL_Vertex2fv(v4);
    DGL_End();
    if(renderTextures) {
        DGL_Disable(DGL_TEXTURE_2D);
    }
}
MaterialVariantSpec const &Rend_SpriteMaterialSpec(dint tclass, dint tmap) {
    return App_Resources().materialSpec(SpriteContext, 0, 1, tclass, tmap, GL_CLAMP_TO_EDGE, GL_CLAMP_TO_EDGE,
                                        1, -2, -1, true, true, true, false);
}
void Rend_DrawSprite(vissprite_t const &spr) {
    drawspriteparams_t const &parm = *VS_SPRITE(&spr);
    DENG2_ASSERT_IN_RENDER_THREAD();
    DENG_ASSERT_GL_CONTEXT_ACTIVE();
    TextureVariant *tex = nullptr;
    Vector2f size;
    dfloat viewOffsetX = 0;  ///< View-aligned offset to center point.
    dfloat s = 1, t = 1;     ///< Bottom right coords.
    if(MaterialAnimator *matAnimator = parm.matAnimator) {
        matAnimator->prepare();
        tex = matAnimator->texUnit(MaterialAnimator::TU_LAYER0).texture;
        dint const texBorder = tex->spec().variant.border;
        size        = matAnimator->dimensions() + Vector2ui(texBorder * 2, texBorder * 2);
        viewOffsetX = -size.x / 2 + -tex->base().origin().x;
        tex->glCoords(&s, &t);
    }
    if(renderTextures == 2) {
        ClientMaterial &debugMaterial = ClientMaterial::find(de::Uri("System", Path("gray")));
        MaterialAnimator &matAnimator = debugMaterial.getAnimator(Rend_SpriteMaterialSpec());
        matAnimator.prepare();
        tex = matAnimator.texUnit(MaterialAnimator::TU_LAYER0).texture;
    }
    if(renderTextures) {
        GL_BindTexture(tex);
        DGL_Enable(DGL_TEXTURE_2D);
    }
    else {
        GL_SetNoTexture();
    }
    coord_t spriteCenter[3] = { spr.pose.origin[0] + spr.pose.srvo[0],
                                spr.pose.origin[1] + spr.pose.srvo[1],
                                spr.pose.origin[2] + spr.pose.srvo[2] };
    coord_t v1[3], v2[3], v3[3], v4[3];
    R_ProjectViewRelativeLine2D(spriteCenter, spr.pose.viewAligned,
                                size.x, viewOffsetX, v1, v4);
    v2[0] = v1[0];
    v2[1] = v1[1];
    v3[0] = v4[0];
    v3[1] = v4[1];
    v1[2] = v4[2] = spriteCenter[2] - size.y / 2;
    v2[2] = v3[2] = spriteCenter[2] + size.y / 2;
    coord_t surfaceNormal[3];
    V3d_PointCrossProduct(surfaceNormal, v2, v1, v3);
    V3d_Normalize(surfaceNormal);
/*#if _DEBUG
    DGL_Begin(DGL_LINES);
    DGL_Color4f(1, 0, 0, 1);
    DGL_Vertex3f(spriteCenter[0], spriteCenter[2], spriteCenter[1]);
    DGL_Color4f(1, 0, 0, 0);
    DGL_Vertex3f(spriteCenter[0] + surfaceNormal[0] * 10,
               spriteCenter[2] + surfaceNormal[2] * 10,
               spriteCenter[1] + surfaceNormal[1] * 10);
    DGL_End();
#endif*/
    dgl_color_t quadColors[4];
    dgl_vertex_t quadNormals[4];
    for(dint i = 0; i < 4; ++i) {
        V3f_Copyd(quadNormals[i].xyz, surfaceNormal);
    }
    if(!spr.light.vLightListIdx) {
        applyUniformColor(4, quadColors, &spr.light.ambientColor[0]);
    }
    else {
        Spr_VertexColors(4, quadColors, quadNormals, spr.light.vLightListIdx,
                         spriteLight + 1, &spr.light.ambientColor[0]);
    }
    bool restoreMatrix = false;
    bool restoreZ      = false;
    if(spr.pose.viewAligned || alwaysAlign >= 2) {
        restoreMatrix = true;
        DGL_MatrixMode(DGL_MODELVIEW);
        DGL_PushMatrix();
        DGL_Translatef(spriteCenter[0], spriteCenter[2], spriteCenter[1]);
        if(!spr.pose.viewAligned) {
            dfloat s_dx = v1[0] - v2[0];
            dfloat s_dy = v1[1] - v2[1];
            if(alwaysAlign == 2) {
                dfloat dx = spriteCenter[0] - vOrigin.x;
                dfloat dy = spriteCenter[1] - vOrigin.z;
                dfloat spriteAngle = BANG2DEG(
                    bamsAtan2(spriteCenter[2] - vOrigin.y, std::sqrt(dx * dx + dy * dy)));
                if(spriteAngle > 180)
                    spriteAngle -= 360;
                if(fabs(spriteAngle) > maxSpriteAngle) {
                    dfloat turnAngle = (spriteAngle > 0? spriteAngle - maxSpriteAngle :
                                                         spriteAngle + maxSpriteAngle);
                    DGL_Rotatef(turnAngle, s_dx, 0, s_dy);
                }
            }
            else {
                DGL_Rotatef(vpitch * .5f, s_dx, 0, s_dy);
            }
        }
        else {
            DGL_Rotatef(vpitch, viewsidex, 0, viewsidey);
        }
        DGL_Translatef(-spriteCenter[0], -spriteCenter[2], -spriteCenter[1]);
    }
    if(parm.blendMode != BM_NORMAL) {
        GL_BlendMode(parm.blendMode);
    }
    if(parm.noZWrite || spr.light.ambientColor[3] < .98f ||
       !(parm.blendMode == BM_NORMAL || parm.blendMode == BM_ZEROALPHA)) {
        restoreZ = true;
        DGL_Disable(DGL_DEPTH_WRITE);
    }
    dgl_vertex_t vs[4], *v = vs;
    dgl_texcoord_t tcs[4], *tc = tcs;
    v[0].xyz[0] = v1[0];
    v[0].xyz[1] = v1[2];
    v[0].xyz[2] = v1[1];
    v[1].xyz[0] = v2[0];
    v[1].xyz[1] = v2[2];
    v[1].xyz[2] = v2[1];
    v[2].xyz[0] = v3[0];
    v[2].xyz[1] = v3[2];
    v[2].xyz[2] = v3[1];
    v[3].xyz[0] = v4[0];
    v[3].xyz[1] = v4[2];
    v[3].xyz[2] = v4[1];
    tc[0].st[0] = s *  (parm.matFlip[0]? 1:0);
    tc[0].st[1] = t * (!parm.matFlip[1]? 1:0);
    tc[1].st[0] = s *  (parm.matFlip[0]? 1:0);
    tc[1].st[1] = t *  (parm.matFlip[1]? 1:0);
    tc[2].st[0] = s * (!parm.matFlip[0]? 1:0);
    tc[2].st[1] = t *  (parm.matFlip[1]? 1:0);
    tc[3].st[0] = s * (!parm.matFlip[0]? 1:0);
    tc[3].st[1] = t * (!parm.matFlip[1]? 1:0);
    drawQuad(v, quadColors, tc);
    if(renderTextures) {
        DGL_Disable(DGL_TEXTURE_2D);
    }
    if(devMobjVLights && spr.light.vLightListIdx) {
        DGL_PushState();
        DGL_Disable(DGL_DEPTH_TEST);
        DGL_CullFace(DGL_NONE);
        DGL_MatrixMode(DGL_MODELVIEW);
        DGL_PushMatrix();
        DGL_Translatef(spr.pose.origin[0], spr.pose.origin[2], spr.pose.origin[1]);
        coord_t const distFromViewer = de::abs(spr.pose.distance);
        ClientApp::renderSystem().forAllVectorLights(spr.light.vLightListIdx, [&distFromViewer] (VectorLightData const &vlight) {
            if(distFromViewer < 1600 - 8) {
                Rend_DrawVectorLight(vlight, 1 - distFromViewer / 1600);
            }
            return LoopContinue;
        });
        DGL_MatrixMode(DGL_MODELVIEW);
        DGL_PopMatrix();
        DGL_PopState();
    }
    if(restoreMatrix) {
        DGL_PopMatrix();
    }
    if(parm.blendMode != BM_NORMAL) {
        GL_BlendMode(BM_NORMAL);
    }
    if(restoreZ) {
        DGL_Enable(DGL_DEPTH_WRITE);
    }
}
void Rend_SpriteRegister() {
    C_VAR_INT   ("rend-sprite-align",       &alwaysAlign,       0, 0, 3);
    C_VAR_FLOAT ("rend-sprite-align-angle", &maxSpriteAngle,    0, 0, 90);
    C_VAR_INT   ("rend-sprite-alpha",       &useSpriteAlpha,    0, 0, 1);
    C_VAR_INT   ("rend-sprite-blend",       &useSpriteBlend,    0, 0, 1);
    C_VAR_INT   ("rend-sprite-lights",      &spriteLight,       0, 0, 10);
    C_VAR_BYTE  ("rend-sprite-mode",        &noSpriteTrans,     0, 0, 1);
    C_VAR_INT   ("rend-sprite-noz",         &noSpriteZWrite,    0, 0, 1);
    C_VAR_BYTE  ("rend-sprite-precache",    &precacheSprites,   0, 0, 1);
    C_VAR_BYTE  ("rend-dev-nosprite",       &devNoSprites,      CVF_NO_ARCHIVE, 0, 1);
}
byte bmapShowDebug; // 1 = mobjs, 2 = lines, 3 = BSP leafs, 4 = polyobjs. cvar
dfloat bmapDebugSize = 1.5f; // cvar
static void drawMobj(mobj_t const &mob) {
    AABoxd const bounds = Mobj_Bounds(mob);
    DGL_Vertex2f(bounds.minX, bounds.minY);
    DGL_Vertex2f(bounds.maxX, bounds.minY);
    DGL_Vertex2f(bounds.maxX, bounds.maxY);
    DGL_Vertex2f(bounds.minX, bounds.maxY);
}
static void drawLine(Line const &line) {
    DGL_Vertex2f(line.from().x(), line.from().y());
    DGL_Vertex2f(line.to  ().x(), line.to  ().y());
}
static void drawSubspace(ConvexSubspace const &subspace) {
    dfloat const scale = de::max(bmapDebugSize, 1.f);
    dfloat const width = (DENG_GAMEVIEW_WIDTH / 16) / scale;
    Face const &poly = subspace.poly();
    HEdge *base = poly.hedge();
    HEdge *hedge = base;
    do {
        Vector2d start = hedge->origin();
        Vector2d end   = hedge->twin().origin();
        DGL_Begin(DGL_LINES);
            DGL_Vertex2f(start.x, start.y);
            DGL_Vertex2f(end.x, end.y);
        DGL_End();
        ddouble length = (end - start).length();
        if (length > 0) {
            Vector2d const unit = (end - start) / length;
            Vector2d const normal(-unit.y, unit.x);
            GL_BindTextureUnmanaged(GL_PrepareLSTexture(LST_DYNAMIC));
            DGL_Enable(DGL_TEXTURE_2D);
            GL_BlendMode(BM_ADD);
            DGL_Begin(DGL_QUADS);
                DGL_TexCoord2f(0, 0.75f, 0.5f);
                DGL_Vertex2f(start.x, start.y);
                DGL_TexCoord2f(0, 0.75f, 0.5f);
                DGL_Vertex2f(end.x, end.y);
                DGL_TexCoord2f(0, 0.75f, 1);
                DGL_Vertex2f(end.x - normal.x * width, end.y - normal.y * width);
                DGL_TexCoord2f(0, 0.75f, 1);
                DGL_Vertex2f(start.x - normal.x * width, start.y - normal.y * width);
            DGL_End();
            DGL_Disable(DGL_TEXTURE_2D);
            GL_BlendMode(BM_NORMAL);
        }
        start = Vector2d(poly.bounds().minX, poly.bounds().minY);
        end   = Vector2d(poly.bounds().maxX, poly.bounds().maxY);
        DGL_Begin(DGL_LINES);
            DGL_Vertex2f(start.x, start.y);
            DGL_Vertex2f(  end.x, start.y);
            DGL_Vertex2f(  end.x, start.y);
            DGL_Vertex2f(  end.x,   end.y);
            DGL_Vertex2f(  end.x,   end.y);
            DGL_Vertex2f(start.x,   end.y);
            DGL_Vertex2f(start.x,   end.y);
            DGL_Vertex2f(start.x, start.y);
        DGL_End();
    } while ((hedge = &hedge->next()) != base);
}
static int drawCellLines(Blockmap const &bmap, BlockmapCell const &cell, void *) {
    DGL_Begin(DGL_LINES);
        bmap.forAllInCell(cell, [] (void *object) {
            Line &line = *(Line *)object;
            if (line.validCount() != validCount) {
                line.setValidCount(validCount);
                drawLine(line);
            }
            return LoopContinue;
        });
    DGL_End();
    return false; // Continue iteration.
}
static int drawCellPolyobjs(Blockmap const &bmap, BlockmapCell const &cell, void *context) {
    DGL_Begin(DGL_LINES);
        bmap.forAllInCell(cell, [&context] (void *object) {
            Polyobj &pob = *(Polyobj *)object;
            for (Line *line : pob.lines()) {
                if (line->validCount() != validCount) {
                    line->setValidCount(validCount);
                    drawLine(*line);
                }
            }
            return LoopContinue;
        });
    DGL_End();
    return false; // Continue iteration.
}
static int drawCellMobjs(Blockmap const &bmap, BlockmapCell const &cell, void *) {
    DGL_Begin(DGL_QUADS);
        bmap.forAllInCell(cell, [] (void *object) {
            mobj_t &mob = *(mobj_t *)object;
            if (mob.validCount != validCount) {
                mob.validCount = validCount;
                drawMobj(mob);
            }
            return LoopContinue;
        });
    DGL_End();
    return false; // Continue iteration.
}
static int drawCellSubspaces(Blockmap const &bmap, BlockmapCell const &cell, void *) {
    bmap.forAllInCell(cell, [] (void *object) {
        ConvexSubspace *sub = (ConvexSubspace *)object;
        if (sub->validCount() != validCount) {
            sub->setValidCount(validCount);
            drawSubspace(*sub);
        }
        return LoopContinue;
    });
    return false; // Continue iteration.
}
static void drawBackground(Blockmap const &bmap) {
    BlockmapCell const &dimensions = bmap.dimensions();
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    DGL_Scalef(bmap.cellSize(), bmap.cellSize(), 1);
    /*
     * Draw the translucent quad which represents the "used" cells.
     */
    DGL_Color4f(.25f, .25f, .25f, .66f);
    DGL_Begin(DGL_QUADS);
        DGL_Vertex2f(0,            0);
        DGL_Vertex2f(dimensions.x, 0);
        DGL_Vertex2f(dimensions.x, dimensions.y);
        DGL_Vertex2f(0,            dimensions.y);
    DGL_End();
    /*
     * Draw the "null cells" over the top.
     */
    DGL_Color4f(0, 0, 0, .95f);
    BlockmapCell cell;
    for (cell.y = 0; cell.y < dimensions.y; ++cell.y)
    for (cell.x = 0; cell.x < dimensions.x; ++cell.x) {
        if (bmap.cellElementCount(cell)) continue;
        DGL_Begin(DGL_QUADS);
            DGL_Vertex2f(cell.x,     cell.y);
            DGL_Vertex2f(cell.x + 1, cell.y);
            DGL_Vertex2f(cell.x + 1, cell.y + 1);
            DGL_Vertex2f(cell.x,     cell.y + 1);
        DGL_End();
    }
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
}
static void drawCellInfo(Vector2d const &origin_, char const *info) {
    DGL_Enable(DGL_TEXTURE_2D);
    FR_SetFont(fontFixed);
    FR_LoadDefaultAttrib();
    FR_SetShadowOffset(UI_SHADOW_OFFSET, UI_SHADOW_OFFSET);
    FR_SetShadowStrength(UI_SHADOW_STRENGTH);
    Size2Raw size = {{{16 + FR_TextWidth(info),
                       16 + FR_SingleLineHeight(info)}}};
    Point2Raw origin = {{{int(origin_.x), int(origin_.y)}}};
    origin.x -= size.width / 2;
    origin.x += 8;
    origin.y += size.height / 2;
    UI_SetColor(UI_Color(UIC_TEXT));
    UI_TextOutEx2(info, &origin, UI_Color(UIC_TITLE), 1, ALIGN_LEFT, DTF_ONLY_SHADOW);
    DGL_Disable(DGL_TEXTURE_2D);
}
static void drawBlockmapInfo(Vector2d const &origin_, Blockmap const &blockmap) {
    DGL_Enable(DGL_TEXTURE_2D);
    Point2Raw origin = {{{int(origin_.x), int(origin_.y)}}};
    FR_SetFont(fontFixed);
    FR_LoadDefaultAttrib();
    FR_SetShadowOffset(UI_SHADOW_OFFSET, UI_SHADOW_OFFSET);
    FR_SetShadowStrength(UI_SHADOW_STRENGTH);
    Size2Raw size;
    size.width = 16 + FR_TextWidth("(+000.0, +000.0) (+000.0, +000.0)");
    int th = FR_SingleLineHeight("Info");
    size.height = th * 4 + 16;
    origin.x -= size.width;
    origin.y -= size.height;
    origin.x += 8;
    origin.y += 8 + th/2;
    UI_TextOutEx2("Blockmap", &origin, UI_Color(UIC_TITLE), 1, ALIGN_LEFT, DTF_ONLY_SHADOW);
    origin.y += th;
    Vector2ui const &bmapDimensions = blockmap.dimensions();
    char buf[80];
    dd_snprintf(buf, 80, "Dimensions:(%u, %u) #%li", bmapDimensions.x, bmapDimensions.y,
                                                    (long) bmapDimensions.y * bmapDimensions.x);
    UI_TextOutEx2(buf, &origin, UI_Color(UIC_TEXT), 1, ALIGN_LEFT, DTF_ONLY_SHADOW);
    origin.y += th;
    dd_snprintf(buf, 80, "Cell dimensions:(%.3f, %.3f)", blockmap.cellSize(), blockmap.cellSize());
    UI_TextOutEx2(buf, &origin, UI_Color(UIC_TEXT), 1, ALIGN_LEFT, DTF_ONLY_SHADOW);
    origin.y += th;
    dd_snprintf(buf, 80, "(%+06.0f, %+06.0f) (%+06.0f, %+06.0f)",
                         blockmap.bounds().minX, blockmap.bounds().minY,
                         blockmap.bounds().maxX, blockmap.bounds().maxY);
    UI_TextOutEx2(buf, &origin, UI_Color(UIC_TEXT), 1, ALIGN_LEFT, DTF_ONLY_SHADOW);
    DGL_Disable(DGL_TEXTURE_2D);
}
static void drawCellInfoBox(Vector2d const &origin, Blockmap const &blockmap,
    char const *objectTypeName, BlockmapCell const &cell) {
    uint count = blockmap.cellElementCount(cell);
    char info[160]; dd_snprintf(info, 160, "Cell:(%u, %u) %s:#%u", cell.x, cell.y, objectTypeName, count);
    drawCellInfo(origin, info);
}
/**
 * @param bmap        Blockmap to be rendered.
 * @param followMobj  Mobj to center/focus the visual on. Can be @c 0.
 * @param cellDrawer  Blockmap cell content drawing callback. Can be @c 0.
 */
static void drawBlockmap(Blockmap const &bmap, mobj_t *followMobj,
    int (*cellDrawer) (Blockmap const &, BlockmapCell const &, void *)) {
    BlockmapCellBlock vCellBlock;
    BlockmapCell vCell;
    BlockmapCell const &dimensions = bmap.dimensions();
    Vector2d const cellDimensions = bmap.cellDimensions();
    if (followMobj) {
        bool didClip;
        vCell = bmap.toCell(followMobj->origin, &didClip);
        if (didClip)
            followMobj = 0; // Outside the blockmap.
        if (followMobj) {
            AABoxd mobBox = Mobj_Bounds(*followMobj);
            mobBox.minX -= DDMOBJ_RADIUS_MAX * 2;
            mobBox.minY -= DDMOBJ_RADIUS_MAX * 2;
            mobBox.maxX += DDMOBJ_RADIUS_MAX * 2;
            mobBox.maxY += DDMOBJ_RADIUS_MAX * 2;
            vCellBlock = bmap.toCellBlock(mobBox);
        }
    }
    if (followMobj) {
        DGL_Translatef(-(vCell.x * cellDimensions.x), -(vCell.y * cellDimensions.y), 0);
    }
    else {
        DGL_Translatef(-(cellDimensions.x * dimensions.x)/2,
                       -(cellDimensions.y * dimensions.y)/2, 0);
    }
    drawBackground(bmap);
    if (followMobj) {
        DGL_Begin(DGL_QUADS);
        BlockmapCell cell;
        for (cell.y = vCellBlock.min.y; cell.y < vCellBlock.max.y; ++cell.y)
        for (cell.x = vCellBlock.min.x; cell.x < vCellBlock.max.x; ++cell.x) {
            if (cell == vCell) {
                DGL_Color4f(.66f, .66f, 1, .66f);
            }
            else {
                DGL_Color4f(.33f, .33f, .66f, .33f);
            }
            Vector2d const start = cellDimensions * cell;
            Vector2d const end   = start + cellDimensions;
            DGL_Vertex2f(start.x, start.y);
            DGL_Vertex2f(  end.x, start.y);
            DGL_Vertex2f(  end.x,   end.y);
            DGL_Vertex2f(start.x,   end.y);
        }
        DGL_End();
    }
    /**
     * Draw the Gridmap visual.
     * @note Gridmap uses a cell unit size of [width:1,height:1], so we need to
     *       scale it up so it aligns correctly.
     */
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    DGL_Scalef(cellDimensions.x, cellDimensions.y, 1);
    bmap.drawDebugVisual();
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
    /*
     * Draw the blockmap-linked data.
     * Translate the modelview matrix so that objects can be drawn using
     * the map space coordinates directly.
     */
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    DGL_Translatef(-bmap.origin().x, -bmap.origin().y, 0);
    if (cellDrawer) {
        if (followMobj) {
            /*
             * Draw cell contents color coded according to their range from the
             * followed Mobj.
             */
            validCount++;
            DGL_Color4f(.33f, 0, 0, .75f);
            BlockmapCell cell;
            for (cell.y = 0; cell.y < dimensions.y; ++cell.y)
            for (cell.x = 0; cell.x < dimensions.x; ++cell.x) {
                if (   cell.x >= vCellBlock.min.x && cell.x < vCellBlock.max.x
                    && cell.y >= vCellBlock.min.y && cell.y < vCellBlock.max.y) {
                    continue;
                }
                if (!bmap.cellElementCount(cell)) continue;
                cellDrawer(bmap, cell, 0/*no params*/);
            }
            validCount++;
            DGL_Color3f(1, .5f, 0);
            for (cell.y = vCellBlock.min.y; cell.y < vCellBlock.max.y; ++cell.y)
            for (cell.x = vCellBlock.min.x; cell.x < vCellBlock.max.x; ++cell.x) {
                if (cell == vCell) continue;
                if (!bmap.cellElementCount(cell)) continue;
                cellDrawer(bmap, cell, 0/*no params*/);
            }
            validCount++;
            DGL_Color3f(1, 1, 0);
            if (bmap.cellElementCount(vCell)) {
                cellDrawer(bmap, vCell, 0/*no params*/);
            }
        }
        else {
            validCount++;
            DGL_Color4f(.33f, 0, 0, .75f);
            BlockmapCell cell;
            for (cell.y = 0; cell.y < dimensions.y; ++cell.y)
            for (cell.x = 0; cell.x < dimensions.x; ++cell.x) {
                if (!bmap.cellElementCount(cell)) continue;
                cellDrawer(bmap, cell, 0/*no params*/);
            }
        }
    }
    /*
     * Draw the followed mobj, if any.
     */
    if (followMobj) {
        validCount++;
        DGL_Color3f(0, 1, 0);
        DGL_Begin(DGL_QUADS);
            drawMobj(*followMobj);
        DGL_End();
    }
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
}
void Rend_BlockmapDebug() {
    if (!bmapShowDebug || bmapShowDebug > 4) return;
    if (!App_World().hasMap()) return;
    Map &map = App_World().map();
    Blockmap const *blockmap;
    int (*cellDrawer) (Blockmap const &, BlockmapCell const &, void *);
    char const *objectTypeName;
    switch (bmapShowDebug) {
    default: // Mobjs.
        blockmap = &map.mobjBlockmap();
        cellDrawer = drawCellMobjs;
        objectTypeName = "Mobjs";
        break;
    case 2: // Lines.
        blockmap = &map.lineBlockmap();
        cellDrawer = drawCellLines;
        objectTypeName = "Lines";
        break;
    case 3: // BSP leafs.
        blockmap = &map.subspaceBlockmap();
        cellDrawer = drawCellSubspaces;
        objectTypeName = "Subspaces";
        break;
    case 4: // Polyobjs.
        blockmap = &map.polyobjBlockmap();
        cellDrawer = drawCellPolyobjs;
        objectTypeName = "Polyobjs";
        break;
    }
    DENG_ASSERT_IN_MAIN_THREAD();
    DENG_ASSERT_GL_CONTEXT_ACTIVE();
    /*
     * Draw the blockmap.
     */
    DGL_MatrixMode(DGL_PROJECTION);
    DGL_PushMatrix();
    DGL_LoadIdentity();
    DGL_Ortho(0, 0, DENG_GAMEVIEW_WIDTH, DENG_GAMEVIEW_HEIGHT, -1, 1);
    DGL_Translatef((DENG_GAMEVIEW_WIDTH / 2), (DENG_GAMEVIEW_HEIGHT / 2), 0);
    float scale = bmapDebugSize / de::max(DENG_GAMEVIEW_HEIGHT / 100, 1);
    DGL_Scalef(scale, -scale, 1);
    mobj_t *followMobj = 0;
    if (viewPlayer && viewPlayer->publicData().mo) {
        followMobj = viewPlayer->publicData().mo;
    }
    drawBlockmap(*blockmap, followMobj, cellDrawer);
    DGL_MatrixMode(DGL_PROJECTION);
    DGL_PopMatrix();
    /*
     * Draw HUD info.
     */
    DGL_MatrixMode(DGL_PROJECTION);
    DGL_PushMatrix();
    DGL_LoadIdentity();
    DGL_Ortho(0, 0, DENG_GAMEVIEW_WIDTH, DENG_GAMEVIEW_HEIGHT, -1, 1);
    if (followMobj) {
        bool didClip;
        BlockmapCell cell = blockmap->toCell(followMobj->origin, &didClip);
        if (!didClip) {
            drawCellInfoBox(Vector2d(DENG_GAMEVIEW_WIDTH / 2, 30), *blockmap,
                            objectTypeName, cell);
        }
    }
    drawBlockmapInfo(Vector2d(DENG_GAMEVIEW_WIDTH - 10, DENG_GAMEVIEW_HEIGHT - 10),
                     *blockmap);
    DGL_MatrixMode(DGL_PROJECTION);
    DGL_PopMatrix();
}
static int fxFramePlayerNum; ///< Player view currently being drawn.
#define IDX_LENS_FLARES         2
void LensFx_Init() {
    for (int i = 0; i < DDMAXPLAYERS; ++i) {
        ConsoleEffectStack &stack = DD_Player(i)->fxStack();
        stack.effects
                << new fx::Bloom(i)
                << new fx::Vignette(i)
                << new fx::LensFlares(i)        // IDX_LENS_FLARES
                << new fx::ColorFilter(i)
                << new fx::Ramp(i);
    }
}
void LensFx_Shutdown() {
    LensFx_GLRelease();
    for (int i = 0; i < DDMAXPLAYERS; ++i) {
        DD_Player(i)->fxStack().clear();
    }
}
void LensFx_GLRelease() {
    for (int i = 0; i < DDMAXPLAYERS; ++i) {
        foreach (ConsoleEffect *effect, DD_Player(i)->fxStack().effects) {
            if (effect->isInited()) {
                effect->glDeinit();
            }
        }
    }
}
void LensFx_Draw(int playerNum) {
    ClientPlayer &player = ClientApp::player(playerNum);
    fxFramePlayerNum = playerNum;
    player.viewCompositor().gameView().resolveSamples();
    GLState::push()
            .setTarget(player.viewCompositor().gameView().resolvedFramebuffer());
    auto const &effects = player.fxStack().effects;
    foreach (ConsoleEffect *effect, effects) {
        if (!effect->isInited()) {
            effect->glInit();
        }
    }
    foreach (ConsoleEffect *effect, effects) {
        effect->beginFrame();
    }
    foreach (ConsoleEffect *effect, effects) {
        effect->draw();
    }
    for (int i = effects.size() - 1; i >= 0; --i) {
        effects.at(i)->endFrame();
    }
    GLState::pop();
}
void LensFx_MarkLightVisibleInFrame(IPointLightSource const &lightSource) {
    auto const &effects = DD_Player(fxFramePlayerNum)->fxStack().effects;
    static_cast<fx::LensFlares *>(effects.at(IDX_LENS_FLARES))->
            markLightPotentiallyVisibleForCurrentFrame(&lightSource);
}
DENG2_PIMPL_NOREF(ConsoleEffect) {
    int console;
    bool inited;
    Impl() : console(0), inited(false)
    {}
};
ConsoleEffect::ConsoleEffect(int console) : d(new Impl) {
    d->console = console;
}
ConsoleEffect::~ConsoleEffect()
{}
int ConsoleEffect::console() const {
    return d->console;
}
Rectanglei ConsoleEffect::viewRect() const {
    return Rectanglei::fromSize(R_Console3DViewRect(d->console).size());
}
bool ConsoleEffect::isInited() const {
    return d->inited;
}
GLShaderBank &ConsoleEffect::shaders() const {
    return ClientApp::renderSystem().shaders();
}
void ConsoleEffect::glInit() {
    d->inited = true;
}
void ConsoleEffect::glDeinit() {
    d->inited = false;
}
void ConsoleEffect::beginFrame()
{}
void ConsoleEffect::draw()
{}
void ConsoleEffect::endFrame()
{}
DENG2_PIMPL_NOREF(Decoration) {
    MaterialAnimator::Decoration const *source = nullptr;
    Surface *surface = nullptr;
};
Decoration::Decoration(MaterialAnimator::Decoration const &source, Vector3d const &origin)
    : MapObject(origin)
    , d(new Impl) {
    d->source = &source;
}
Decoration::~Decoration()
{}
String Decoration::description() const {
    auto desc = String(    _E(l) "Origin: "   _E(.)_E(i) "%1" _E(.)
                       " " _E(l) "Material: " _E(.)_E(i) "%2" _E(.)
                       " " _E(l) "Surface: "  _E(.)_E(i) "%3" _E(.))
                  .arg(origin().asText())
                  .arg(source().decor().material().manifest().composeUri().asText())
                  .arg(String("[0x%1]").arg(de::dintptr(&surface()), 0, 16));
#ifdef DENG2_DEBUG
    desc.prepend(String(_E(b) "Decoration " _E(.) "[0x%1]\n").arg(de::dintptr(this), 0, 16));
#endif
    return desc;
}
MaterialAnimator::Decoration const &Decoration::source() const {
    DENG2_ASSERT(d->source);
    return *d->source;
}
bool Decoration::hasSurface() const {
    return d->surface != nullptr;
}
Surface &Decoration::surface() {
    if(hasSurface()) return *d->surface;
    throw MissingSurfaceError("Decoration::surface", "No surface is attributed");
}
Surface const &Decoration::surface() const {
    if(hasSurface()) return *d->surface;
    throw MissingSurfaceError("Decoration::surface", "No surface is attributed");
}
void Decoration::setSurface(Surface *newSurface) {
    d->surface = newSurface;
}
/**
 * Drawing condition flags.
 *
 * @todo Most of these are actually list specification parameters. Rather than
 * set them each time an identified list is drawn it would be better to record
 * in the list itself. -ds
 */
enum DrawCondition {
    NoBlend            = 0x00000001,
    Blend              = 0x00000002,
    SetLightEnv0       = 0x00000004,
    SetLightEnv1       = 0x00000008,
    JustOneLight       = 0x00000010,
    ManyLights         = 0x00000020,
    SetBlendMode       = 0x00000040, // Primitive-specific blending.
    SetMatrixDTexture0 = 0x00000080,
    SetMatrixDTexture1 = 0x00000100,
    SetMatrixTexture0  = 0x00000200,
    SetMatrixTexture1  = 0x00000400,
    NoColor            = 0x00000800,
    Skip               = 0x80000000,
    SetLightEnv        = SetLightEnv0 | SetLightEnv1,
    SetMatrixDTexture  = SetMatrixDTexture0 | SetMatrixDTexture1,
    SetMatrixTexture   = SetMatrixTexture0 | SetMatrixTexture1
};
static duint32 const BLEND_MODE_MASK = 0xf;
Q_DECLARE_FLAGS(DrawConditions, DrawCondition)
Q_DECLARE_OPERATORS_FOR_FLAGS(DrawConditions)
DrawList::PrimitiveParams::PrimitiveParams(de::gl::Primitive type,
                                           de::Vector2f texScale,
                                           de::Vector2f texOffset,
                                           de::Vector2f detailTexScale,
                                           de::Vector2f detailTexOffset,
                                           Flags        flags,
                                           blendmode_t  blendMode,
                                           DGLuint      modTexture,
                                           de::Vector3f modColor)
    : type           (type)
    , flags_blendMode(uint(blendMode) | uint(flags))
    , texScale       (texScale)
    , texOffset      (texOffset)
    , detailTexScale (detailTexScale)
    , detailTexOffset(detailTexOffset)
    , modTexture     (modTexture)
    , modColor       (modColor)
{}
DENG2_PIMPL(DrawList) {
    /**
     * Each Element begins a block of GL commands/geometry to apply/transfer.
     */
    struct Element {
        duint size;  ///< Size of this element (zero = n/a).
        Element *next() {
            if (!size) return nullptr;
            auto *elem = (Element *) ((duint8 *) (this) + size);
            if (!elem->size) return nullptr;
            return elem;
        }
        struct Data {
            Store const *buffer;
            duint numIndices;
            duint *indices;
            PrimitiveParams primitive;
            /**
             * Draw the geometry for this element.
             */
            void draw(DrawConditions const &conditions, TexUnitMap const &texUnitMap) {
                DENG2_ASSERT(buffer);
                if (conditions & SetLightEnv) {
                    DGL_SetInteger(DGL_ACTIVE_TEXTURE, (conditions & SetLightEnv0)? 0 : 1);
                    GL_BindTextureUnmanaged(!renderTextures? 0 : primitive.modTexture,
                                            gl::ClampToEdge, gl::ClampToEdge);
                    DGL_SetModulationColor(Vector4f(primitive.modColor, 0.f));
                }
                if (conditions & SetMatrixTexture) {
                    if (conditions & SetMatrixTexture0) {
                        DGL_SetInteger(DGL_ACTIVE_TEXTURE, 0);
                        DGL_MatrixMode(DGL_TEXTURE);
                        DGL_PushMatrix();
                        DGL_LoadIdentity();
                        DGL_Translatef(primitive.texOffset.x * primitive.texScale.x,
                                       primitive.texOffset.y * primitive.texScale.y, 1);
                        DGL_Scalef(primitive.texScale.x, primitive.texScale.y, 1);
                    }
                    if (conditions & SetMatrixTexture1) {
                        DGL_SetInteger(DGL_ACTIVE_TEXTURE, 1);
                        DGL_MatrixMode(DGL_TEXTURE);
                        DGL_PushMatrix();
                        DGL_LoadIdentity();
                        DGL_Translatef(primitive.texOffset.x * primitive.texScale.x,
                                       primitive.texOffset.y * primitive.texScale.y, 1);
                        DGL_Scalef(primitive.texScale.x, primitive.texScale.y, 1);
                    }
                }
                if (conditions & SetMatrixDTexture) {
                    if (conditions & SetMatrixDTexture0) {
                        DGL_SetInteger(DGL_ACTIVE_TEXTURE, 0);
                        DGL_MatrixMode(DGL_TEXTURE);
                        DGL_PushMatrix();
                        DGL_LoadIdentity();
                        DGL_Translatef(primitive.detailTexOffset.x * primitive.detailTexScale.x,
                                       primitive.detailTexOffset.y * primitive.detailTexScale.y, 1);
                        DGL_Scalef(primitive.detailTexScale.x, primitive.detailTexScale.y, 1);
                    }
                    if (conditions & SetMatrixDTexture1) {
                        DGL_SetInteger(DGL_ACTIVE_TEXTURE, 1);
                        DGL_MatrixMode(DGL_TEXTURE);
                        DGL_PushMatrix();
                        DGL_LoadIdentity();
                        DGL_Translatef(primitive.detailTexOffset.x * primitive.detailTexScale.x,
                                       primitive.detailTexOffset.y * primitive.detailTexScale.y, 1);
                        DGL_Scalef(primitive.detailTexScale.x, primitive.detailTexScale.y, 1);
                    }
                }
                if (conditions & SetBlendMode) {
                    GL_BlendMode(blendmode_t(primitive.flags_blendMode & BLEND_MODE_MASK));
                }
                DGL_Begin(primitive.type == gl::TriangleStrip? DGL_TRIANGLE_STRIP : DGL_TRIANGLE_FAN);
                for (duint i = 0; i < numIndices; ++i) {
                    duint const index = indices[i];
                    for (dint k = 0; k < MAX_TEX_UNITS; ++k) {
                        Vector2f const *tc = nullptr;  // No mapping.
                        switch (texUnitMap[k]) {
                        case AttributeSpec::TexCoord0:   tc = buffer->texCoords[0]; break;
                        case AttributeSpec::TexCoord1:   tc = buffer->texCoords[1]; break;
                        case AttributeSpec::ModTexCoord: tc = buffer->modCoords;    break;
                        default: break;
                        }
                        if (tc) {
                            DGL_TexCoord2f(k, tc[index].x, tc[index].y);
                        }
                    }
                    if (!(conditions & NoColor)) {
                        Vector4ub const &color = buffer->colorCoords[index];
                        DGL_Color4ub(color.x, color.y, color.z, color.w);
                    }
                    Vector3f const &pos = buffer->posCoords[index];
                    DGL_Vertex3f(pos.x, pos.z, pos.y);
                }
                DGL_End();
                if (conditions & SetMatrixDTexture) {
                    if (conditions & SetMatrixDTexture0) {
                        DGL_SetInteger(DGL_ACTIVE_TEXTURE, 0);
                        DGL_MatrixMode(DGL_TEXTURE);
                        DGL_PopMatrix();
                    }
                    if (conditions & SetMatrixDTexture1) {
                        DGL_SetInteger(DGL_ACTIVE_TEXTURE, 1);
                        DGL_MatrixMode(DGL_TEXTURE);
                        DGL_PopMatrix();
                    }
                }
                if (conditions & SetMatrixTexture) {
                    if (conditions & SetMatrixTexture0) {
                        DGL_SetInteger(DGL_ACTIVE_TEXTURE, 0);
                        DGL_MatrixMode(DGL_TEXTURE);
                        DGL_PopMatrix();
                    }
                    if (conditions & SetMatrixTexture1) {
                        DGL_SetInteger(DGL_ACTIVE_TEXTURE, 1);
                        DGL_MatrixMode(DGL_TEXTURE);
                        DGL_PopMatrix();
                    }
                }
            }
        } data;
    };
    Spec spec;                 ///< List specification.
    dsize dataSize = 0;        ///< Number of bytes allocated for the data.
    duint8 *data   = nullptr;  ///< Data for a number of polygons (The List).
    duint8 *cursor = nullptr;  ///< Data pointer for reading/writing.
    Element *last  = nullptr;  ///< Last element (if any).
    Impl(Public *i, Spec const &spec) : Base(i), spec(spec) {}
    ~Impl() { clearAllData(); }
    void clearAllData() {
        if (data) {
            Z_Free(data); data = nullptr;
#ifdef DENG_DEBUG
            Z_CheckHeap();
#endif
        }
        cursor   = nullptr;
        last     = nullptr;
        dataSize = 0;
    }
    /**
     * @return  Start of the allocated data.
     */
    void *allocateData(dsize bytes) {
        dint const PADDING = 16;
        if (!bytes) return nullptr;
        dint const startOffset = cursor - data;
        dsize const required   = startOffset + bytes + PADDING;
        if (required > dataSize) {
            duint8 const *oldData   = data;
            dint const cursorOffset = (cursor? cursor - oldData : -1);
            dint const lastOffset   = (last? (duint8 *) last - oldData : -1);
            if (dataSize == 0) {
                dataSize = 1024;
            }
            while (dataSize < required) {
                dataSize *= 2;
            }
            data = (duint8 *) Z_Realloc(data, dataSize, PU_APPSTATIC);
            cursor = (cursorOffset >= 0? data + cursorOffset : data);
            last   = (lastOffset >= 0? (Element *) (data + lastOffset) : 0);
            if (oldData) {
                for (Element *elem = first(); elem && elem <= last; elem = elem->next()) {
                    if (elem->data.indices) {
                        elem->data.indices = (duint *) (data + ((duint8 *) elem->data.indices - oldData));
                    }
                }
            }
        }
        cursor += bytes;
        return data + startOffset;
    }
    Element *first() const {
        auto *elem = (Element *)data;
        if (!elem->size) return nullptr;
        return elem;
    }
    /**
     * Configure GL state for drawing in this @a mode.
     * @return  The conditions to select primitives.
     */
    DrawConditions pushGLState(DrawMode mode) {
        switch (mode) {
        case DM_SKYMASK:
            DENG2_ASSERT(spec.group == SkyMaskGeom);
            return NoColor;
        case DM_ALL:  // All surfaces.
            DENG2_ASSERT(spec.group == UnlitGeom || spec.group == LitGeom);
            if (spec.unit(TU_INTER).hasTexture()) {
                GL_SelectTexUnits(2);
                GL_BindTo(spec.unit(TU_PRIMARY), 0);
                GL_BindTo(spec.unit(TU_INTER  ), 1);
                DGL_ModulateTexture(2);
                dfloat color[4] = { 0, 0, 0, spec.unit(TU_INTER).opacity };
                DGL_SetModulationColor(color);
            }
            else if (!spec.unit(TU_PRIMARY).hasTexture()) {
                return 0;
            }
            else {
                GL_SelectTexUnits(1);
                GL_Bind(spec.unit(TU_PRIMARY));
                DGL_ModulateTexture(1);
            }
            if (spec.unit(TU_INTER).hasTexture()) {
                return SetMatrixTexture0 | SetMatrixTexture1;
            }
            return SetMatrixTexture0;
        case DM_LIGHT_MOD_TEXTURE:
            DENG2_ASSERT(spec.group == LitGeom);
            GL_BindTo(spec.unit(TU_PRIMARY), 1);
            return SetMatrixTexture1 | SetLightEnv0 | JustOneLight | NoBlend;
        case DM_TEXTURE_PLUS_LIGHT:
            DENG2_ASSERT(spec.group == LitGeom);
            GL_BindTo(spec.unit(TU_PRIMARY), 0);
            return SetMatrixTexture0 | SetLightEnv1 | NoBlend;
        case DM_FIRST_LIGHT:
            DENG2_ASSERT(spec.group == LitGeom);
            return SetLightEnv0 | ManyLights | Blend;
        case DM_BLENDED: {
            DENG2_ASSERT(spec.group == UnlitGeom || spec.group == LitGeom);
            if (!spec.unit(TU_INTER).hasTexture()) {
                return Skip;
            }
            GL_SelectTexUnits(2);
            GL_BindTo(spec.unit(TU_PRIMARY), 0);
            GL_BindTo(spec.unit(TU_INTER  ), 1);
            DGL_ModulateTexture(2);
            dfloat color[4] = { 0, 0, 0, spec.unit(TU_INTER).opacity };
            DGL_SetModulationColor(color);
            return SetMatrixTexture0 | SetMatrixTexture1; }
        case DM_BLENDED_FIRST_LIGHT:
            DENG2_ASSERT(spec.group == LitGeom);
            if (!spec.unit(TU_INTER).hasTexture()) {
                return Skip;
            }
            return SetMatrixTexture1 | SetLightEnv0;
        case DM_WITHOUT_TEXTURE:
            DENG2_ASSERT(spec.group == LitGeom);
            return 0;
        case DM_LIGHTS:
            DENG2_ASSERT(spec.group == LightGeom);
            GL_Bind(spec.unit(TU_PRIMARY));
            return 0;
        case DM_BLENDED_MOD_TEXTURE:
            DENG2_ASSERT(spec.group == LitGeom);
            if (!spec.unit(TU_INTER).hasTexture()) {
                break;
            }
        case DM_MOD_TEXTURE:
        case DM_MOD_TEXTURE_MANY_LIGHTS:
            DENG2_ASSERT(spec.group == LitGeom);
            if (spec.unit(TU_INTER).hasTexture()) {
                GL_SelectTexUnits(2);
                GL_BindTo(spec.unit(TU_PRIMARY), 0);
                GL_BindTo(spec.unit(TU_INTER  ), 1);
                DGL_ModulateTexture(3);
                dfloat color[4] = { 0, 0, 0, spec.unit(TU_INTER).opacity };
                DGL_SetModulationColor(color);
                return SetMatrixTexture0 | SetMatrixTexture1;
            }
            GL_SelectTexUnits(1);
            GL_Bind(spec.unit(TU_PRIMARY));
            DGL_ModulateTexture(0);
            if (mode == DM_MOD_TEXTURE_MANY_LIGHTS) {
                return SetMatrixTexture0 | ManyLights;
            }
            return SetMatrixTexture0;
        case DM_UNBLENDED_MOD_TEXTURE_AND_DETAIL:
            DENG2_ASSERT(spec.group == LitGeom);
            if (spec.unit(TU_INTER).hasTexture()) {
                break;
            }
            if (spec.unit(TU_PRIMARY_DETAIL).hasTexture()) {
                GL_SelectTexUnits(2);
                DGL_ModulateTexture(9); // Tex+Detail, no color.
                GL_BindTo(spec.unit(TU_PRIMARY       ), 0);
                GL_BindTo(spec.unit(TU_PRIMARY_DETAIL), 1);
                return SetMatrixTexture0 | SetMatrixDTexture1;
            }
            else {
                GL_SelectTexUnits(1);
                DGL_ModulateTexture(0);
                GL_Bind(spec.unit(TU_PRIMARY));
                return SetMatrixTexture0;
            }
            break;
        case DM_ALL_DETAILS:
            DENG2_ASSERT(spec.group == UnlitGeom || spec.group == LitGeom);
            if (spec.unit(TU_PRIMARY_DETAIL).hasTexture()) {
                GL_Bind(spec.unit(TU_PRIMARY_DETAIL));
                return SetMatrixDTexture0;
            }
            break;
        case DM_UNBLENDED_TEXTURE_AND_DETAIL:
            DENG2_ASSERT(spec.group == UnlitGeom || spec.group == LitGeom);
            if (spec.unit(TU_INTER).hasTexture()) {
                break;
            }
            if (spec.unit(TU_PRIMARY_DETAIL).hasTexture()) {
                GL_SelectTexUnits(2);
                DGL_ModulateTexture(8);
                GL_BindTo(spec.unit(TU_PRIMARY       ), 0);
                GL_BindTo(spec.unit(TU_PRIMARY_DETAIL), 1);
                return SetMatrixTexture0 | SetMatrixDTexture1;
            }
            else {
                GL_SelectTexUnits(1);
                DGL_ModulateTexture(1);
                GL_Bind(spec.unit(TU_PRIMARY));
                return SetMatrixTexture0;
            }
            break;
        case DM_BLENDED_DETAILS: {
            DENG2_ASSERT(spec.group == UnlitGeom || spec.group == LitGeom);
            if (!spec.unit(TU_INTER).hasTexture()) {
                break;
            }
            if (!spec.unit(TU_PRIMARY_DETAIL).hasTexture() ||
                !spec.unit(TU_INTER_DETAIL  ).hasTexture()) {
                break;
            }
            GL_BindTo(spec.unit(TU_PRIMARY_DETAIL), 0);
            GL_BindTo(spec.unit(TU_INTER_DETAIL  ), 1);
            dfloat color[4] = { 0, 0, 0, spec.unit(TU_INTER_DETAIL).opacity };
            DGL_SetModulationColor(color);
            return SetMatrixDTexture0 | SetMatrixDTexture1; }
        case DM_SHADOW:
            DENG2_ASSERT(spec.group == ShadowGeom);
            if (spec.unit(TU_PRIMARY).hasTexture()) {
                GL_Bind(spec.unit(TU_PRIMARY));
            }
            else {
                GL_BindTextureUnmanaged(0);
            }
            if (!spec.unit(TU_PRIMARY).hasTexture()) {
                DGL_MatrixMode(DGL_MODELVIEW);
                DGL_PushMatrix();
                DGL_Translatef(vOrigin.x, vOrigin.y, vOrigin.z);
                DGL_Scalef(.99f, .99f, .99f);
                DGL_Translatef(-vOrigin.x, -vOrigin.y, -vOrigin.z);
            }
            return 0;
        case DM_MASKED_SHINY:
            DENG2_ASSERT(spec.group == ShineGeom);
            if (spec.unit(TU_INTER).hasTexture()) {
                GL_SelectTexUnits(2);
                GL_BindTo(spec.unit(TU_INTER), 1);
                dfloat color[4] = { 0, 0, 0, 1 };
                DGL_SetModulationColor(color);
            }
        case DM_ALL_SHINY:
        case DM_SHINY:
            DENG2_ASSERT(spec.group == ShineGeom);
            GL_BindTo(spec.unit(TU_PRIMARY), 0);
            if (!spec.unit(TU_INTER).hasTexture()) {
                GL_SelectTexUnits(1);
            }
            if (mode == DM_ALL_SHINY) {
                return SetBlendMode;
            }
            if (mode == DM_MASKED_SHINY) {
                return SetBlendMode | SetMatrixTexture1;
            }
            return SetBlendMode | NoBlend;
        default: break;
        }
        return Skip;
    }
    /**
     * Restore GL state after drawing in the specified @a mode.
     */
    void popGLState(DrawMode mode) {
        switch (mode) {
        default: break;
        case DM_ALL:
            if (spec.unit(TU_INTER).hasTexture()) {
                GL_SelectTexUnits(1);
                DGL_ModulateTexture(1);
            }
            break;
        case DM_BLENDED:
            if (spec.unit(TU_INTER).hasTexture()) {
                GL_SelectTexUnits(1);
                DGL_ModulateTexture(1);
            }
            break;
        case DM_BLENDED_MOD_TEXTURE:
        case DM_MOD_TEXTURE:
        case DM_MOD_TEXTURE_MANY_LIGHTS:
            if (spec.unit(TU_INTER).hasTexture()) {
                GL_SelectTexUnits(1);
                DGL_ModulateTexture(1);
            }
            else if (mode != DM_BLENDED_MOD_TEXTURE) {
                DGL_ModulateTexture(1);
            }
            break;
        case DM_UNBLENDED_MOD_TEXTURE_AND_DETAIL:
            if (!spec.unit(TU_INTER).hasTexture()) {
                if (spec.unit(TU_PRIMARY_DETAIL).hasTexture()) {
                    GL_SelectTexUnits(1);
                    DGL_ModulateTexture(1);
                }
                else {
                    DGL_ModulateTexture(1);
                }
            }
            break;
        case DM_UNBLENDED_TEXTURE_AND_DETAIL:
            if (!spec.unit(TU_INTER).hasTexture() &&
               spec.unit(TU_PRIMARY_DETAIL).hasTexture()) {
                GL_SelectTexUnits(1);
                DGL_ModulateTexture(1);
            }
            break;
        case DM_SHADOW:
            if (!spec.unit(TU_PRIMARY).hasTexture()) {
                DGL_MatrixMode(DGL_MODELVIEW);
                DGL_PopMatrix();
            }
            break;
        case DM_SHINY:
        case DM_ALL_SHINY:
        case DM_MASKED_SHINY:
            GL_BlendMode(BM_NORMAL);
            if (mode == DM_MASKED_SHINY && spec.unit(TU_INTER).hasTexture()) {
                GL_SelectTexUnits(1);
            }
            break;
        }
    }
};
DrawList::DrawList(Spec const &spec) : d(new Impl(this, spec))
{}
bool DrawList::isEmpty() const {
    return d->last == nullptr;
}
DrawList &DrawList::write(Store const &buffer,
                          DrawList::Indices const &indices,
                          PrimitiveParams const &params) {
    if (indices.isEmpty()) return *this;
    return write(buffer, indices.constData(), indices.size(), params);
}
DrawList &DrawList::write(Store const &buffer, Indices const &indices, gl::Primitive primitiveType) {
    return write(buffer, indices.constData(), indices.size(), primitiveType);
}
DrawList &DrawList::write(Store const &buffer, duint const *indices, int indexCount, gl::Primitive primitiveType) {
    static PrimitiveParams defaultParams(gl::TriangleFan); // NOTE: rendering is single-threaded atm
    defaultParams.type = primitiveType;
    return write(buffer, indices, indexCount, defaultParams);
}
DrawList &DrawList::write(Store const &buffer, duint const *indices, int indexCount,
                          PrimitiveParams const &params) {
#ifdef DENG2_DEBUG
    using Parm = PrimitiveParams;
    DENG2_ASSERT(!(spec().group == LightGeom  && ((params.flags_blendMode & Parm::OneLight) ||
                                                  (params.flags_blendMode & Parm::ManyLights) ||
                                                  params.modTexture)));
    DENG2_ASSERT(!(spec().group == LitGeom    && !Rend_IsMTexLights() && ((params.flags_blendMode & Parm::OneLight) ||
                                                                          params.modTexture)));
    DENG2_ASSERT(!(spec().group == ShadowGeom && ((params.flags_blendMode & Parm::OneLight) ||
                                                  (params.flags_blendMode & Parm::ManyLights) ||
                                                  params.modTexture)));
#endif
    if (!indexCount) return *this;
    d->last = (Impl::Element *) d->allocateData(sizeof(Impl::Element));
    d->last->size = 0;
    d->last->data.buffer     = &buffer;
    d->last->data.primitive  = params;
    d->last->data.numIndices = indexCount;
    auto *lti = (duint *) d->allocateData(sizeof(duint) * d->last->data.numIndices);
    DENG_ASSERT(d->last != nullptr);
    d->last->data.indices = lti;
    for (duint i = 0; i < d->last->data.numIndices; ++i) {
        d->last->data.indices[i] = indices[i];
    }
    d->last->size = d->cursor - (duint8 *) d->last;
    *(duint *) d->cursor = 0;
    return *this;
}
void DrawList::draw(DrawMode mode, TexUnitMap const &texUnitMap) const {
    using Parm = PrimitiveParams;
    DENG2_ASSERT_IN_RENDER_THREAD();
    DENG_ASSERT_GL_CONTEXT_ACTIVE();
    DrawConditions const conditions = d->pushGLState(mode);
    if (conditions & Skip) return; // Assume no state changes were made.
    bool bypass = false;
    if (d->spec.unit(TU_INTER).hasTexture()) {
        if (conditions.testFlag(NoBlend)) {
            return;
        }
        if (conditions.testFlag(Blend)) {
            bypass = true;
        }
    }
    if (!bypass) {
        if (!conditions.testFlag(JustOneLight) &&
            !conditions.testFlag(ManyLights)) {
            bypass = true;
        }
    }
    bool skip = false;
    for (Impl::Element *elem = d->first(); elem; elem = elem->next()) {
        if (!bypass) {
            skip = false;
            if (conditions.testFlag(JustOneLight)
                    && (elem->data.primitive.flags_blendMode & Parm::ManyLights)) {
                skip = true;
            }
            else if (conditions.testFlag(ManyLights)
                     && (elem->data.primitive.flags_blendMode & Parm::OneLight)) {
                skip = true;
            }
        }
        if (!skip) {
            elem->data.draw(conditions, texUnitMap);
            DENG2_ASSERT(!Sys_GLCheckError());
        }
    }
    d->popGLState(mode);
}
DrawList::Spec &DrawList::spec() {
    return d->spec;
}
DrawList::Spec const &DrawList::spec() const {
    return d->spec;
}
void DrawList::clear() {
    d->clearAllData();
}
void DrawList::rewind() {
    d->cursor = d->data;
    d->last   = nullptr;
}
void DrawList::reserveSpace(DrawList::Indices &indices, uint count) // static
{
    if (indices.size() < int(count)) indices.resize(int(count));
}
typedef QMultiHash<GLuint, DrawList *> DrawListHash;
DENG2_PIMPL(DrawLists) {
    QScopedPointer<DrawList> skyMaskList;
    DrawListHash unlitHash;
    DrawListHash litHash;
    DrawListHash dynHash;
    DrawListHash shinyHash;
    DrawListHash shadowHash;
    Impl(Public *i) : Base(i) {
        DrawListSpec newSpec;
        newSpec.group = SkyMaskGeom;
        skyMaskList.reset(new DrawList(newSpec));
    }
    DrawListHash &listHash(GeomGroup group) {
        switch(group) {
        case UnlitGeom:  return unlitHash;
        case LitGeom:    return litHash;
        case LightGeom:  return dynHash;
        case ShadowGeom: return shadowHash;
        case ShineGeom:  return shinyHash;
        case SkyMaskGeom: break; // n/a?
        }
        DENG2_ASSERT(false);
        return unlitHash;
    }
};
DrawLists::DrawLists() : d(new Impl(this))
{}
static void clearAllLists(DrawListHash &hash) {
    foreach(DrawList *list, hash) {
        list->clear();
    }
    qDeleteAll(hash);
    hash.clear();
}
void DrawLists::clear() {
    clearAllLists(d->unlitHash);
    clearAllLists(d->litHash);
    clearAllLists(d->dynHash);
    clearAllLists(d->shadowHash);
    clearAllLists(d->shinyHash);
    d->skyMaskList->clear();
}
static void resetList(DrawList &list) {
    list.rewind();
    DrawListSpec &listSpec = list.spec();
    listSpec.unit(TU_INTER).unmanaged.glName = 0;
    listSpec.unit(TU_INTER).texture          = 0;
    listSpec.unit(TU_INTER).opacity          = 0;
    listSpec.unit(TU_INTER_DETAIL).unmanaged.glName = 0;
    listSpec.unit(TU_INTER_DETAIL).texture          = 0;
    listSpec.unit(TU_INTER_DETAIL).opacity          = 0;
}
static void resetAllLists(DrawListHash &hash) {
    foreach(DrawList *list, hash) {
        resetList(*list);
    }
}
void DrawLists::reset() {
    resetAllLists(d->unlitHash);
    resetAllLists(d->litHash);
    resetAllLists(d->dynHash);
    resetAllLists(d->shadowHash);
    resetAllLists(d->shinyHash);
    resetList(*d->skyMaskList);
}
/**
 * Specialized texture unit comparision function that ignores properties which
 * are applied per-primitive and which should not result in list separation.
 *
 * (These properties are written to the primitive header and applied dynamically
 * when reading back the draw list.)
 */
static bool compareTexUnit(GLTextureUnit const &lhs, GLTextureUnit const &rhs) {
    if(lhs.texture) {
        if(lhs.texture != rhs.texture) return false;
    }
    else {
        if(lhs.unmanaged != rhs.unmanaged) return false;
    }
    if(!de::fequal(lhs.opacity, rhs.opacity)) return false;
    return true;
}
DrawList &DrawLists::find(DrawListSpec const &spec) {
    if(spec.group == SkyMaskGeom) {
        return *d->skyMaskList;
    }
    DrawList *convertable = 0;
    GLuint const key  = spec.unit(TU_PRIMARY).getTextureGLName();
    DrawListHash &hash = d->listHash(spec.group);
    for(DrawListHash::const_iterator it = hash.find(key);
        it != hash.end() && it.key() == key; ++it) {
        DrawList *list = it.value();
        DrawListSpec const &listSpec = list->spec();
        if((spec.group == ShineGeom &&
            compareTexUnit(listSpec.unit(TU_PRIMARY), spec.unit(TU_PRIMARY))) ||
           (spec.group != ShineGeom &&
            compareTexUnit(listSpec.unit(TU_PRIMARY), spec.unit(TU_PRIMARY)) &&
            compareTexUnit(listSpec.unit(TU_PRIMARY_DETAIL), spec.unit(TU_PRIMARY_DETAIL)))) {
            if(!listSpec.unit(TU_INTER).hasTexture() &&
               !spec.unit(TU_INTER).hasTexture()) {
                return *list;
            }
            if(list->isEmpty() && !convertable && spec.unit(TU_INTER).hasTexture()) {
                convertable = list;
            }
            if((spec.group == ShineGeom &&
                compareTexUnit(listSpec.unit(TU_INTER), spec.unit(TU_INTER))) ||
               (spec.group != ShineGeom &&
                compareTexUnit(listSpec.unit(TU_INTER), spec.unit(TU_INTER)) &&
                compareTexUnit(listSpec.unit(TU_INTER_DETAIL), spec.unit(TU_INTER_DETAIL)))) {
                return *list;
            }
        }
    }
    if(convertable) {
        if(spec.group == ShineGeom) {
            convertable->spec().unit(TU_INTER) = spec.unit(TU_INTER);
        }
        else {
            convertable->spec().unit(TU_INTER) = spec.unit(TU_INTER);
            convertable->spec().unit(TU_INTER_DETAIL) = spec.unit(TU_INTER_DETAIL);
        }
        return *convertable;
    }
    return *hash.insert(key, new DrawList(spec)).value();
}
int DrawLists::findAll(GeomGroup group, FoundLists &found) {
    found.clear();
    if(group == SkyMaskGeom) {
        if(!d->skyMaskList->isEmpty()) {
            found.append(d->skyMaskList.data());
        }
    }
    else {
        DrawListHash const &hash = d->listHash(group);
        foreach(DrawList *list, hash) {
            if(!list->isEmpty()) {
                found.append(list);
            }
        }
    }
    return found.count();
}
static String const ID_DEFAULT       ("default");
static String const DEF_PATH         ("path");
static String const DEF_INTERIOR_PATH("interior.path");
static String const DEF_EXTERIOR_PATH("exterior.path");
namespace render {
DENG2_PIMPL(Environment)
, DENG2_OBSERVES(filesys::AssetObserver, Availability)
, DENG2_OBSERVES(World, MapChange) {
    enum { Interior, Exterior };
    filesys::AssetObserver observer { "texture\\.reflect\\..*" };
    LoopCallback mainCall;
    struct EnvMaps {
        Path interior;
        Path exterior;
    };
    QHash<String, EnvMaps> maps;
    GLTexture reflectionTextures[2];
    Impl(Public *i) : Base(i) {
        observer.audienceForAvailability() += this;
        World::get().audienceForMapChange() += this;
        for (auto &tex : reflectionTextures) {
            tex.setMinFilter(gl::Linear, gl::MipLinear);
            tex.setWrap(gl::ClampToEdge, gl::ClampToEdge);
        }
    }
    ~Impl() {
        release();
    }
    void assetAvailabilityChanged(String const &identifier,
                                  filesys::AssetObserver::Event event) {
        LOG_RES_MSG("Texture asset \"%s\" is now %s")
            << identifier
            << (event == filesys::AssetObserver::Added ? "available" : "unavailable");
        String const mapId = identifier.substr(16).toLower();
        switch (event) {
        case filesys::AssetObserver::Added:
            addMapsFromAsset(mapId, App::asset(identifier));
            break;
        case filesys::AssetObserver::Removed:
            maps.remove(mapId);
            break;
        }
    }
    void addMapsFromAsset(String const &mapId, Package::Asset const &asset) {
        EnvMaps env;
        if (asset.has(DEF_PATH)) {
            env.interior = env.exterior = asset.absolutePath(DEF_PATH);
        }
        if (asset.has(DEF_INTERIOR_PATH)) {
            env.interior = asset.absolutePath(DEF_INTERIOR_PATH);
        }
        if (asset.has(DEF_EXTERIOR_PATH)) {
            env.exterior = asset.absolutePath(DEF_EXTERIOR_PATH);
        }
        maps.insert(mapId, env);
    }
    void release() {
        for (auto &tex : reflectionTextures) {
            tex.clear();
        }
    }
    void loadCubeMap(GLTexture &tex, String const &path) {
        try {
            ImageFile const &imgFile = App::rootFolder().locate<ImageFile const>(path);
            LOG_GL_MSG("Loading reflection cube map %s") << imgFile.description();
            Image img = imgFile.image();
            Image::Size const size(img.width() / 6, img.height());
            tex.setImage(gl::NegativeX, img.subImage(Rectanglei(0*size.x, 0, size.x, size.y)));
            tex.setImage(gl::PositiveZ, img.subImage(Rectanglei(1*size.x, 0, size.x, size.y)));
            tex.setImage(gl::PositiveX, img.subImage(Rectanglei(2*size.x, 0, size.x, size.y)));
            tex.setImage(gl::NegativeZ, img.subImage(Rectanglei(3*size.x, 0, size.x, size.y)));
            tex.setImage(gl::NegativeY, img.subImage(Rectanglei(4*size.x, 0, size.x, size.y)));
            tex.setImage(gl::PositiveY, img.subImage(Rectanglei(5*size.x, 0, size.x, size.y)));
            tex.generateMipmap();
        }
        catch (Error const &er) {
            LOG_GL_WARNING("Failed to load reflection cube map from path \"%s\": %s")
                << path << er.asText();
        }
    }
    void worldMapChanged() {
        mainCall.enqueue([this] () { loadTexturesForCurrentMap(); });
    }
    void loadTexturesForCurrentMap() {
        DENG2_ASSERT_IN_MAIN_THREAD();
        release();
        String const mapId = ClientApp::world().map().id().toLower();
        auto found = maps.constFind(mapId);
        if (found != maps.constEnd()) {
            loadEnvMaps(found.value());
        }
        else {
            found = maps.constFind(ID_DEFAULT);
            if (found != maps.constEnd()) {
                loadEnvMaps(found.value());
            }
        }
    }
    void loadEnvMaps(EnvMaps const &env) {
        DENG2_ASSERT(!(env.interior.isEmpty() && env.exterior.isEmpty()));
        if (!env.exterior.isEmpty()) {
            loadCubeMap(reflectionTextures[Exterior], env.exterior);
        }
        loadCubeMap(reflectionTextures[Interior], env.interior);
    }
};
Environment::Environment() : d(new Impl(this))
{}
void Environment::glDeinit() {
    d->release();
}
GLTexture const &Environment::defaultReflection() const {
    return d->reflectionTextures[Impl::Interior];
}
GLTexture const &Environment::reflectionInSubsector(world::Subsector const *subsec) const {
    if (!subsec) {
        return defaultReflection();
    }
    if (subsec->as<world::ClientSubsector>().hasSkyPlane()) {
        return d->reflectionTextures[Impl::Exterior];
    }
    return d->reflectionTextures[Impl::Interior];
}
}  // namespace render
static dfloat angleFadeFactor = .1f; ///< cvar
static dfloat brightFactor    = 1;   ///< cvar
LightDecoration::LightDecoration(MaterialAnimator::Decoration const &source, Vector3d const &origin)
    : Decoration(source, origin)
    , Source()
{}
String LightDecoration::description() const {
    String desc;
#ifdef DENG2_DEBUG
    desc.prepend(String(_E(b) "LightDecoration " _E(.) "[0x%1]\n").arg(de::dintptr(this), 0, 16));
#endif
    return Decoration::description() + "\n" + desc;
}
dfloat LightDecoration::occlusion(Vector3d const &eye) const {
    if (source().elevation() < 2 && ::angleFadeFactor > 0) // Close the surface?
    {
        Vector3d const vecFromOriginToEye = (origin() - eye).normalize();
        auto dot = dfloat( -surface().normal().dot(vecFromOriginToEye) );
        if (dot < ::angleFadeFactor / 2) {
            return 0;
        }
        else if (dot < 3 * ::angleFadeFactor) {
            return (dot - ::angleFadeFactor / 2) / (2.5f * ::angleFadeFactor);
        }
    }
    return 1;
}
/**
 * @return  @c > 0 if @a lightlevel passes the min max limit condition.
 */
static dfloat checkLightLevel(dfloat lightlevel, dfloat min, dfloat max) {
    if (de::fequal(min, max)) return 1;
    return de::clamp(0.f, (lightlevel - min) / dfloat(max - min), 1.f);
}
Lumobj *LightDecoration::generateLumobj() const {
    if (source().color() == Vector3f(0, 0, 0))
        return nullptr;
    ConvexSubspace *subspace = bspLeafAtOrigin().subspacePtr();
    if (!subspace) return nullptr;
    dfloat intensity = subspace->subsector().as<ClientSubsector>().lightSourceIntensity();
    Rend_ApplyLightAdaptation(intensity);
    dfloat lightLevels[2];
    source().lightLevels(lightLevels[0], lightLevels[1]);
    intensity = checkLightLevel(intensity, lightLevels[0], lightLevels[1]);
    if (intensity < .0001f)
        return nullptr;
    dfloat fadeMul = intensity * ::brightFactor;
    if (fadeMul <= 0)
        return nullptr;
    Lumobj *lum = new Lumobj(origin(), source().radius(), source().color() * fadeMul);
    lum->setSource(this);
    lum->setMaxDistance (MAX_DECOR_DISTANCE)
        .setLightmap    (Lumobj::Side, source().tex())
        .setLightmap    (Lumobj::Down, source().floorTex())
        .setLightmap    (Lumobj::Up,   source().ceilTex())
        .setFlareSize   (source().flareSize())
        .setFlareTexture(source().flareTex());
    return lum;
}
void LightDecoration::consoleRegister() // static
{
    C_VAR_FLOAT("rend-light-decor-angle",  &::angleFadeFactor, 0, 0, 1);
    C_VAR_FLOAT("rend-light-decor-bright", &::brightFactor,    0, 0, 10);
}
dfloat LightDecoration::angleFadeFactor() // static
{
    return ::angleFadeFactor;
}
dfloat LightDecoration::brightFactor() // static
{
    return ::brightFactor;
}
static dint radiusMax     = 320;    ///< Absolute maximum lumobj radius (cvar).
static dfloat radiusScale = 5.2f;  ///< Radius scale factor (cvar).
dfloat Lumobj::Source::occlusion(Vector3d const & /*eye*/) const
{
    return 1;  // Fully visible.
}
DENG2_PIMPL_NOREF(Lumobj) {
    Source const *source = nullptr;      ///< Source of the lumobj (if any, not owned).
    mobj_t const *sourceMobj = nullptr;  ///< Mobj associated with the lumobj (if any).
    ddouble maxDistance = 0;             ///< Used when rendering to limit the number drawn lumobjs.
    Vector3f color = Vector3f(1, 1, 1);  ///< Light color/intensity.
    ddouble radius = 256;                ///< Radius in map space units.
    ddouble zOffset = 0;                 ///< Z-axis origin offset in map space units.
    dfloat flareSize = 0;                ///< Scale factor.
    DGLuint flareTex = 0;                ///< Custom flare texture (@todo should be Texture ptr).
    ClientTexture *sideTex = nullptr;
    ClientTexture *downTex = nullptr;
    ClientTexture *upTex   = nullptr;
    Impl() {}
    Impl(Impl const &other)
        : source     (other.source)
        , sourceMobj (other.sourceMobj)
        , maxDistance(other.maxDistance)
        , color      (other.color)
        , radius     (other.radius)
        , zOffset    (other.zOffset)
        , flareSize  (other.flareSize)
        , flareTex   (other.flareTex)
        , sideTex    (other.sideTex)
        , downTex    (other.downTex)
        , upTex      (other.upTex)
    {}
};
Lumobj::Lumobj(Vector3d const &origin, ddouble radius, Vector3f const &color)
    : MapObject(origin), d(new Impl()) {
    setRadius(radius);
    setColor(color);
}
Lumobj::Lumobj(Lumobj const &other)
    : MapObject(other.origin()), d(new Impl(*other.d))
{}
void Lumobj::setSource(Source const *newSource) {
    d->source = newSource;
}
void Lumobj::setSourceMobj(mobj_t const *mo) {
    d->sourceMobj = mo;
}
mobj_t const *Lumobj::sourceMobj() const {
    return d->sourceMobj;
}
Vector3f const &Lumobj::color() const {
    return d->color;
}
Lumobj &Lumobj::setColor(Vector3f const &newColor) {
    if(d->color != newColor) {
        d->color = newColor;
    }
    return *this;
}
ddouble Lumobj::radius() const {
    return d->radius;
}
Lumobj &Lumobj::setRadius(ddouble newRadius) {
    newRadius *= 40 * ::radiusScale;
    newRadius = de::clamp<ddouble>(.0001, de::abs(newRadius), ::radiusMax);
    if(d->radius != newRadius) {
        d->radius = newRadius;
    }
    return *this;
}
AABoxd Lumobj::bounds() const {
    return AABoxd(origin().x - d->radius, origin().y - d->radius,
                  origin().x + d->radius, origin().y + d->radius);
}
ddouble Lumobj::zOffset() const {
    return d->zOffset;
}
Lumobj &Lumobj::setZOffset(ddouble newZOffset) {
    d->zOffset = newZOffset;
    return *this;
}
ddouble Lumobj::maxDistance() const {
    return d->maxDistance;
}
Lumobj &Lumobj::setMaxDistance(ddouble newMaxDistance) {
    d->maxDistance = newMaxDistance;
    return *this;
}
ClientTexture *Lumobj::lightmap(LightmapSemantic semantic) const {
    switch(semantic) {
    case Side: return d->sideTex;
    case Down: return d->downTex;
    case Up:   return d->upTex;
    };
    DENG2_ASSERT(false);
    return d->sideTex;
}
Lumobj &Lumobj::setLightmap(LightmapSemantic semantic, ClientTexture *newTexture) {
    switch(semantic) {
    case Side: d->sideTex = newTexture; break;
    case Down: d->downTex = newTexture; break;
    case Up:   d->upTex   = newTexture; break;
    };
    return *this;
}
dfloat Lumobj::flareSize() const {
    return d->flareSize;
}
Lumobj &Lumobj::setFlareSize(dfloat newFlareSize) {
    d->flareSize = de::max(0.f, newFlareSize);
    return *this;
}
DGLuint Lumobj::flareTexture() const {
    return d->flareTex;
}
Lumobj &Lumobj::setFlareTexture(DGLuint newTexture) {
    d->flareTex = newTexture;
    return *this;
}
dfloat Lumobj::attenuation(ddouble distFromEye) const {
    if(distFromEye > 0 && d->maxDistance > 0) {
        if(distFromEye > d->maxDistance) return 0;
        if(distFromEye > .67 * d->maxDistance)
            return (d->maxDistance - distFromEye) / (.33 * d->maxDistance);
    }
    return 1;
}
void Lumobj::generateFlare(Vector3d const &eye, ddouble distFromEye) {
    if(d->maxDistance > 0 && distFromEye > d->maxDistance)
        return;
    if(!d->source) return;
    vissprite_t *vis = R_NewVisSprite(VSPR_FLARE);
    vis->pose.origin             = origin();
    vis->pose.distance           = distFromEye;
    V3f_Set(vis->data.flare.color, d->color.x, d->color.y, d->color.z);
    vis->data.flare.mul          = d->source->occlusion(eye) * attenuation(distFromEye);
    vis->data.flare.size         = d->flareSize > 0? de::max(1.f, d->flareSize * 60 * (50 + haloSize) / 100.0f) : 0;
    vis->data.flare.tex          = d->flareTex;
    vis->data.flare.lumIdx       = indexInMap();
    vis->data.flare.isDecoration = true;
}
dint Lumobj::radiusMax()  // static
{
    return ::radiusMax;
}
dfloat Lumobj::radiusFactor()  // static
{
    return ::radiusScale;
}
void Lumobj::consoleRegister()  // static
{
    C_VAR_INT  ("rend-light-radius-max",   &::radiusMax,   0, 64,  512);
    C_VAR_FLOAT("rend-light-radius-scale", &::radiusScale, 0, .1f, 10);
}
namespace render {
String const Model::DEF_TIMELINE("timeline");
Model::AnimSequence::AnimSequence(String const &name, Record const &def)
    : name(name)
    , def(&def) {
    if (def.hasSubrecord(DEF_TIMELINE)) {
        timeline = new Timeline;
        timeline->addFromInfo(def.subrecord(DEF_TIMELINE));
    }
    else if (def.hasMember(DEF_TIMELINE)) {
        sharedTimeline = def.gets(DEF_TIMELINE);
    }
}
Model::~Model() {
    unsetAtlas();
    qDeleteAll(timelines.values());
}
} // namespace render
namespace render {
using namespace de;
String const ModelLoader::DEF_ANIMATION ("animation");
String const ModelLoader::DEF_MATERIAL  ("material");
String const ModelLoader::DEF_PASS      ("pass");
String const ModelLoader::DEF_RENDER    ("render");
static String const DEF_ALIGNMENT_PITCH ("alignment.pitch");
static String const DEF_ALIGNMENT_YAW   ("alignment.yaw");
static String const DEF_AUTOSCALE       ("autoscale");
static String const DEF_BLENDFUNC       ("blendFunc");
static String const DEF_BLENDOP         ("blendOp");
static String const DEF_DEPTHFUNC       ("depthFunc");
static String const DEF_DEPTHWRITE      ("depthWrite");
static String const DEF_FOV             ("fov");
static String const DEF_FRONT_VECTOR    ("front");
static String const DEF_MESHES          ("meshes");
static String const DEF_MIRROR          ("mirror");
static String const DEF_OFFSET          ("offset");
static String const DEF_SEQUENCE        ("sequence");
static String const DEF_SHADER          ("shader");
static String const DEF_STATE           ("state");
static String const DEF_TEXTURE_MAPPING ("textureMapping");
static String const DEF_TIMELINE        ("timeline");
static String const DEF_UP_VECTOR       ("up");
static String const DEF_WEAPON_OPACITY   ("opacityFromWeapon");
static String const DEF_WEAPON_FULLBRIGHT("fullbrightFromWeapon");
static String const DEF_VARIANT         ("variant");
static String const SHADER_DEFAULT      ("model.skeletal.generic");
static String const MATERIAL_DEFAULT    ("default");
static String const VAR_U_MAP_TIME          ("uMapTime");
static String const VAR_U_PROJECTION_MATRIX ("uProjectionMatrix");
static String const VAR_U_VIEW_MATRIX       ("uViewMatrix");
static Atlas::Size const MAX_ATLAS_SIZE (8192, 8192);
DENG2_PIMPL(ModelLoader)
, DENG2_OBSERVES(filesys::AssetObserver, Availability)
, DENG2_OBSERVES(Bank, Load)
, DENG2_OBSERVES(ModelDrawable, AboutToGLInit)
, DENG2_OBSERVES(BusyRunner, DeferredGLTask)
, public MultiAtlas::IAtlasFactory {
    MultiAtlas atlasPool { *this };
    filesys::AssetObserver observer { "model\\..*" };
    ModelBank bank {
        []() -> ModelDrawable * { return new render::Model; }
    };
    LockableT<QSet<String>> pendingModels;
    struct Program : public GLProgram {
        String shaderName;
        Record const *def = nullptr;
        int useCount = 1; ///< Number of models using the program.
    };
    /**
     * All shader programs needed for loaded models. All programs are ready for drawing,
     * and the common uniforms are bound.
     */
    struct Programs : public QHash<String, Program *> {
        ~Programs() {
#ifdef DENG2_DEBUG
            for (auto i = constBegin(); i != constEnd(); ++i) {
                qDebug() << i.key() << i.value();
                DENG2_ASSERT(i.value());
                qWarning() << "ModelRenderer: Program" << i.key()
                           << "still has" << i.value()->useCount << "users";
            }
#endif
            DENG2_ASSERT(empty());
        }
    };
    Programs programs;
    Impl(Public *i) : Base(i) {
        observer.audienceForAvailability() += this;
        bank.audienceForLoad() += this;
    }
    void init() {
        loadProgram(SHADER_DEFAULT);
#if defined (DENG_HAVE_BUSYRUNNER)
        ClientApp::busyRunner().audienceForDeferredGLTask() += this;
#endif
    }
    void deinit() {
#if defined (DENG_HAVE_BUSYRUNNER)
        ClientApp::busyRunner().audienceForDeferredGLTask() -= this;
#endif
        bank.unloadAll(Bank::ImmediatelyInCurrentThread);
        pendingModels.value.clear();
        atlasPool.clear();
        unloadProgram(*programs[SHADER_DEFAULT]);
    }
    Atlas *makeAtlas(MultiAtlas &) override {
        Atlas *atlas = AtlasTexture::newWithKdTreeAllocator(
                    Atlas::DeferredAllocations,
                    GLTexture::maximumSize().min(MAX_ATLAS_SIZE));
        atlas->setBorderSize(1);
        atlas->setMarginSize(0);
        LOG_GL_MSG("New %ix%i model texture atlas allocated")
                << atlas->totalSize().x << atlas->totalSize().y;
        return atlas;
    }
    void assetAvailabilityChanged(String const &identifier, filesys::AssetObserver::Event event) override {
        LOG_RES_MSG("Model asset \"%s\" is now %s")
                << identifier
                << (event == filesys::AssetObserver::Added? "available" :
                                                            "unavailable");
        if (event == filesys::AssetObserver::Added) {
            bank.add(identifier, App::asset(identifier).absolutePath("path"));
            bank.load(identifier);
        }
        else {
            auto const &model = bank.model<render::Model const>(identifier);
            for (auto const &pass : model.passes) {
                DENG2_ASSERT(pass.program);
                unloadProgram(*static_cast<Program *>(pass.program));
            }
            if (model.passes.isEmpty()) {
                if (model.program()) {
                    unloadProgram(*static_cast<Program *>(model.program()));
                }
            }
            else {
                DENG2_ASSERT(!model.program());
            }
            bank.remove(identifier); {
                DENG2_GUARD(pendingModels);
                pendingModels.value.remove(identifier);
            }
        }
    }
    /**
     * Initializes one or more uninitialized models for rendering.
     * Must be called from the main thread.
     *
     * @param maxCount  Maximum number of models to initialize.
     */
    void initPendingModels(int maxCount) {
        DENG2_GUARD(pendingModels);
        while (!pendingModels.value.isEmpty() && maxCount > 0) {
            String const identifier = *pendingModels.value.begin();
            pendingModels.value.remove(identifier);
            if (!bank.has(identifier)) {
                continue;
            }
            LOG_GL_MSG("Initializing \"%s\"") << identifier;
            auto &model = bank.model<render::Model>(identifier);
            model.glInit();
            --maxCount;
        }
    }
    BusyRunner::DeferredResult performDeferredGLTask() override {
        initPendingModels(1);
        DENG2_GUARD(pendingModels);
        return pendingModels.value.isEmpty()? BusyRunner::AllTasksCompleted
                                            : BusyRunner::TasksPending;
    }
    /**
     * Initializes a program and stores it in the programs hash. If the program already
     * has been loaded before, the program use count is incremented.
     *
     * @param name  Name of the shader program.
     *
     * @return Shader program.
     */
    Program *loadProgram(String const &name) {
        if (programs.contains(name)) {
            programs[name]->useCount++;
            return programs[name];
        }
        std::unique_ptr<Program> prog(new Program);
        prog->shaderName = name;
        prog->def = &ClientApp::shaders()[name].valueAsRecord(); // for lookups later
        LOG_RES_VERBOSE("Loading model shader \"%s\"") << name;
        ClientApp::shaders().build(*prog, name);
        DENG2_FOR_PUBLIC_AUDIENCE2(NewProgram, i) {
            i->newProgramCreated(*prog);
        }
        auto &render = ClientApp::renderSystem();
        if (prog->def->hasMember(VAR_U_MAP_TIME)) {
            *prog << render.uMapTime();
        }
        if (prog->def->hasMember(VAR_U_PROJECTION_MATRIX)) {
            *prog << render.uProjectionMatrix();
        }
        if (prog->def->hasMember(VAR_U_VIEW_MATRIX)) {
            *prog << render.uViewMatrix();
        }
        programs[name] = prog.get();
        return prog.release();
    }
    /**
     * Decrements a program's use count, and deletes the program altogether if it has
     * no more users.
     *
     * @param name  Name of the shader program.
     */
    void unloadProgram(Program &program) {
        if (--program.useCount == 0) {
            String name = program.shaderName;
            LOG_RES_VERBOSE("Model shader \"%s\" unloaded (no more users)") << name;
            delete &program;
            DENG2_ASSERT(programs.contains(name));
            programs.remove(name);
        }
    }
    void modelAboutToGLInit(ModelDrawable &drawable) override {
        auto &model = static_cast<render::Model &>(drawable); {
            DENG2_GUARD(pendingModels);
            pendingModels.value.remove(model.identifier);
        }
        model.setAtlas(*model.textures);
    }
    static gl::Comparison textToComparison(String const &text) {
        static struct { char const *txt; gl::Comparison comp; } const cs[] = {
            { "Never",          gl::Never },
            { "Always",         gl::Always },
            { "Equal",          gl::Equal },
            { "NotEqual",       gl::NotEqual },
            { "Less",           gl::Less },
            { "Greater",        gl::Greater },
            { "LessOrEqual",    gl::LessOrEqual },
            { "GreaterOrEqual", gl::GreaterOrEqual }
        };
        for (auto const &p : cs) {
            if (text == p.txt) {
                return p.comp;
            }
        }
        throw DefinitionError("ModelRenderer::textToComparison",
                              QString("Invalid comparison function \"%1\"").arg(text));
    }
    static gl::Blend textToBlendFunc(String const &text) {
        static struct { char const *txt; gl::Blend blend; } const bs[] = {
            { "Zero",              gl::Zero },
            { "One",               gl::One },
            { "SrcColor",          gl::SrcColor },
            { "OneMinusSrcColor",  gl::OneMinusSrcColor },
            { "SrcAlpha",          gl::SrcAlpha },
            { "OneMinusSrcAlpha",  gl::OneMinusSrcAlpha },
            { "DestColor",         gl::DestColor },
            { "OneMinusDestColor", gl::OneMinusDestColor },
            { "DestAlpha",         gl::DestAlpha },
            { "OneMinusDestAlpha", gl::OneMinusDestAlpha }
        };
        for (auto const &p : bs) {
            if (text == p.txt) {
                return p.blend;
            }
        }
        throw DefinitionError("ModelRenderer::textToBlendFunc",
                              QString("Invalid blending function \"%1\"").arg(text));
    }
    static gl::BlendOp textToBlendOp(String const &text) {
        if (text == "Add") return gl::Add;
        if (text == "Subtract") return gl::Subtract;
        if (text == "ReverseSubtract") return gl::ReverseSubtract;
        throw DefinitionError("ModelRenderer::textToBlendOp",
                              QString("Invalid blending operation \"%1\"").arg(text));
    }
    /**
     * Compose texture mappings of one or more shaders.
     *
     * All shaders used on one model must have the same mappings. Otherwise
     * the vertex data in the static model VBO is not compatible with all of
     * them (redoing the VBOs during drawing is inadvisable).
     *
     * @param mapping    Model texture mapping.
     * @param shaderDef  Shader definition.
     */
    void composeTextureMappings(ModelDrawable::Mapping &mapping,
                                Record const &shaderDef) {
        if (shaderDef.has(DEF_TEXTURE_MAPPING)) {
            ArrayValue const &array = shaderDef.geta(DEF_TEXTURE_MAPPING);
            for (int i = 0; i < int(array.size()); ++i) {
                ModelDrawable::TextureMap map = ModelDrawable::textToTextureMap(array.element(i).asText());
                if (i == mapping.size()) {
                    mapping << map;
                }
                else if (mapping.at(i) != map) {
                    QStringList list;
                    for (auto map : mapping) list << ModelDrawable::textureMapToText(map);
                    throw TextureMappingError("ModelRenderer::composeTextureMappings",
                                              QString("Texture mapping <%1> is incompatible with shader %2")
                                                  .arg(list.join(", "))
                                                  .arg(ScriptedInfo::sourceLocation(shaderDef)));
                }
            }
        }
    }
    /**
     * When model assets have been loaded, we can parse their metadata to see if there
     * are any animation sequences defined. If so, we'll set up a shared lookup table
     * that determines which sequences to start in which mobj states.
     *
     * @param path  Model asset id.
     */
    void bankLoaded(DotPath const &path) override {
        render::Model &model = bank.model<render::Model>(path);
        model.audienceForAboutToGLInit() += this;
        auto const asset = App::asset(path);
        Vector3f front(0, 0, 1);
        Vector3f up   (0, 1, 0);
        if (asset.has(DEF_FRONT_VECTOR)) {
            front = Vector3f(asset.geta(DEF_FRONT_VECTOR));
        }
        if (asset.has(DEF_UP_VECTOR)) {
            up = Vector3f(asset.geta(DEF_UP_VECTOR));
        }
        bool mirror = ScriptedInfo::isTrue(asset, DEF_MIRROR);
        model.cull = mirror? gl::Back : gl::Front;
        model.transformation = Matrix4f::frame(front, up, !mirror);
        if (asset.has(DEF_OFFSET)) {
            model.offset = vectorFromValue<Vector3f>(asset.get(DEF_OFFSET));
        }
        applyFlagOperation(model.flags, render::Model::AutoscaleToThingHeight,
                           ScriptedInfo::isTrue(asset, DEF_AUTOSCALE)? SetFlags : UnsetFlags);
        applyFlagOperation(model.flags, render::Model::ThingOpacityAsAmbientLightAlpha,
                           ScriptedInfo::isTrue(asset, DEF_WEAPON_OPACITY, true)? SetFlags : UnsetFlags);
        applyFlagOperation(model.flags, render::Model::ThingFullBrightAsAmbientLight,
                           ScriptedInfo::isTrue(asset, DEF_WEAPON_FULLBRIGHT)? SetFlags : UnsetFlags);
        model.alignYaw   = parseAlignment(asset, DEF_ALIGNMENT_YAW);
        model.alignPitch = parseAlignment(asset, DEF_ALIGNMENT_PITCH);
        model.pspriteFOV = asset.getf(DEF_FOV, 0.0f);
        model.materialIndexForName.insert(MATERIAL_DEFAULT, 0);
        if (asset.has(DEF_MATERIAL)) {
            asset.subrecord(DEF_MATERIAL)
                .forSubrecords([this, &model](String const &blockName, Record const &block) {
                    try {
                        if (ScriptedInfo::blockType(block) == DEF_VARIANT) {
                            String const materialName = blockName;
                            if (!model.materialIndexForName.contains(materialName)) {
                                model.materialIndexForName.insert(materialName,
                                                                  model.addMaterial());
                            }
                            block.forSubrecords([this, &model, &materialName](
                                                    String const &matName, Record const &matDef) {
                                setupMaterial(model,
                                              matName,
                                              model.materialIndexForName[materialName],
                                              matDef);
                                return LoopContinue;
                            });
                        }
                        else {
                            setupMaterial(model, blockName, 0, block);
                        }
                    }
                    catch (const Error &er) {
                        LOG_GL_ERROR("Material variant \"%s\" is invalid: %s")
                            << blockName << er.asText();
                    }
                    return LoopContinue;
                });
        }
        if (asset.has(DEF_ANIMATION)) {
            auto states = ScriptedInfo::subrecordsOfType(DEF_STATE, asset.subrecord(DEF_ANIMATION));
            DENG2_FOR_EACH_CONST(Record::Subrecords, state, states) {
                auto const seqs = ScriptedInfo::subrecordsOfType(DEF_SEQUENCE, *state.value());
                if (!seqs.isEmpty()) {
                    if (model.animationCount() > 0) {
                        for (String key : ScriptedInfo::sortRecordsBySource(seqs)) {
                            model.animations[state.key()] << render::Model::AnimSequence(key, *seqs[key]);
                        }
                    }
                    else {
                        LOG_GL_WARNING("Model \"%s\" has no animation sequences in the model file "
                                       "but animation sequence definition found at %s")
                                << path
                                << ScriptedInfo::sourceLocation
                                   (*seqs[ScriptedInfo::sortRecordsBySource(seqs).first()]);
                    }
                }
            }
            auto timelines = ScriptedInfo::subrecordsOfType(DEF_TIMELINE, asset.subrecord(DEF_ANIMATION));
            DENG2_FOR_EACH_CONST(Record::Subrecords, timeline, timelines) {
                Timeline *scheduler = new Timeline;
                scheduler->addFromInfo(*timeline.value());
                model.timelines[timeline.key()] = scheduler;
            }
        }
        ModelDrawable::Mapping textureMapping;
        String modelShader = SHADER_DEFAULT;
        if (asset.has(DEF_RENDER)) {
            Record const &renderBlock = asset.subrecord(DEF_RENDER);
            modelShader = renderBlock.gets(DEF_SHADER, modelShader);
            auto passes = ScriptedInfo::subrecordsOfType(DEF_PASS, renderBlock);
            for (String key : ScriptedInfo::sortRecordsBySource(passes)) {
                try {
                    auto const &def = *passes[key];
                    ModelDrawable::Pass pass;
                    pass.name = key;
                    pass.meshes.resize(model.meshCount());
                    for (Value const *value : def.geta(DEF_MESHES).elements()) {
                        int meshId = identifierFromText(value->asText(), [&model] (String const &text) {
                            return model.meshId(text);
                        });
                        if (meshId < 0 || meshId >= model.meshCount()) {
                            throw DefinitionError("ModelLoader::bankLoaded",
                                                  "Unknown mesh \"" + value->asText() + "\" in " +
                                                      ScriptedInfo::sourceLocation(def));
                        }
                        pass.meshes.setBit(meshId, true);
                    }
                    if (def.has(DEF_BLENDFUNC)) {
                        ArrayValue const &blendDef = def.geta(DEF_BLENDFUNC);
                        pass.blendFunc.first  = textToBlendFunc(blendDef.at(0).asText());
                        pass.blendFunc.second = textToBlendFunc(blendDef.at(1).asText());
                    }
                    pass.blendOp = textToBlendOp(def.gets(DEF_BLENDOP, "Add"));
                    pass.depthFunc = textToComparison(def.gets(DEF_DEPTHFUNC, "Less"));
                    pass.depthWrite = ScriptedInfo::isTrue(def, DEF_DEPTHWRITE, true);
                    String const passShader = def.gets(DEF_SHADER, modelShader);
                    pass.program = loadProgram(passShader);
                    composeTextureMappings(textureMapping,
                                           ClientApp::shaders()[passShader]);
                    model.passes.append(pass);
                }
                catch (Error const &er) {
                    LOG_RES_ERROR("Rendering pass \"%s\" in asset \"%s\" is invalid: %s")
                            << key << path << er.asText();
                }
            }
        }
        if (model.passes.isEmpty()) {
            try {
                model.setProgram(loadProgram(modelShader));
                composeTextureMappings(textureMapping,
                                       ClientApp::shaders()[modelShader]);
            }
            catch (Error const &er) {
                LOG_RES_ERROR("Asset \"%s\" cannot use shader \"%s\": %s")
                        << path << modelShader << er.asText();
            }
        }
        model.setTextureMapping(textureMapping);
        model.textures.reset(new MultiAtlas::AllocGroup(atlasPool)); {
            DENG2_GUARD(pendingModels);
            pendingModels.value.insert(path);
        }
    }
    render::Model::Alignment parseAlignment(RecordAccessor const &def, String const &key) {
        if (!def.has(key)) return render::Model::NotAligned;
        String const value = def.gets(key);
        if (!value.compareWithoutCase("movement")) {
            return render::Model::AlignToMomentum;
        }
        else if (!value.compareWithoutCase("view")) {
            return render::Model::AlignToView;
        }
        else if (!value.compareWithoutCase("random")) {
            return render::Model::AlignRandomly;
        }
        else if (ScriptedInfo::isFalse(def, key, false)) {
            return render::Model::NotAligned;
        }
        else {
            throw DefinitionError("ModelRenderer::parseAlignment",
                                  String("Unknown alignment value \"%1\" for %2 in %3")
                                  .arg(value)
                                  .arg(key)
                                  .arg(ScriptedInfo::sourceLocation(def)));
        }
    }
    void setupMaterial(ModelDrawable &model,
                       String const &meshName,
                       duint materialIndex,
                       Record const &matDef) {
        int mid = identifierFromText(meshName,
                                     [&model](const String &text) { return model.meshId(text); });
        if (mid < 0 || mid >= model.meshCount()) {
            throw DefinitionError("ModelLoader::setupMaterial",
                                  "Mesh \"" + meshName + "\" not found in " +
                                      ScriptedInfo::sourceLocation(matDef));
        }
        const ModelDrawable::MeshId mesh{duint(mid), materialIndex};
        setupMaterialTexture(model, mesh, matDef, QStringLiteral("diffuseMap"),  ModelDrawable::Diffuse);
        setupMaterialTexture(model, mesh, matDef, QStringLiteral("normalMap"),   ModelDrawable::Normals);
        setupMaterialTexture(model, mesh, matDef, QStringLiteral("heightMap"),   ModelDrawable::Height);
        setupMaterialTexture(model, mesh, matDef, QStringLiteral("specularMap"), ModelDrawable::Specular);
        setupMaterialTexture(model, mesh, matDef, QStringLiteral("emissiveMap"), ModelDrawable::Emissive);
    }
    void setupMaterialTexture(ModelDrawable &model,
                              ModelDrawable::MeshId const &mesh,
                              Record const &matDef,
                              String const &textureName,
                              ModelDrawable::TextureMap map) {
        if (matDef.has(textureName)) {
            String const path = ScriptedInfo::absolutePathInContext(matDef, matDef.gets(textureName));
            model.setTexturePath(mesh, map, path);
        }
    }
    DENG2_PIMPL_AUDIENCE(NewProgram)
};
DENG2_AUDIENCE_METHOD(ModelLoader, NewProgram)
ModelLoader::ModelLoader()
    : d(new Impl(this))
{}
ModelBank &ModelLoader::bank() {
    return d->bank;
}
ModelBank const &ModelLoader::bank() const {
    return d->bank;
}
void ModelLoader::glInit() {
    d->init();
}
void ModelLoader::glDeinit() {
    d->deinit();
}
String ModelLoader::shaderName(GLProgram const &program) const {
    return static_cast<Impl::Program const &>(program).shaderName;
}
Record const &ModelLoader::shaderDefinition(GLProgram const &program) const {
    return *static_cast<Impl::Program const &>(program).def;
}
int ModelLoader::identifierFromText(String const &text,
                                    std::function<int (String const &)> resolver) // static
{
    int id = 0;
    if (text.beginsWith('@')) {
        id = text.mid(1).toInt();
    }
    else {
        id = resolver(text);
    }
    return id;
}
} // namespace render
static int constexpr MAX_LIGHTS = 4;
float weaponFixedFOV = 95.f;
DENG2_PIMPL(ModelRenderer)
, DENG2_OBSERVES(render::ModelLoader, NewProgram) {
    render::ModelLoader loader;
    GLUniform uMvpMatrix        { "uMvpMatrix",        GLUniform::Mat4 };
    GLUniform uWorldMatrix      { "uWorldMatrix",      GLUniform::Mat4 }; // included in uMvpMatrix
    GLUniform uReflectionMatrix { "uReflectionMatrix", GLUniform::Mat4 };
    GLUniform uTex              { "uTex",              GLUniform::Sampler2D };
    GLUniform uReflectionTex    { "uReflectionTex",    GLUniform::SamplerCube };
    GLUniform uEyePos           { "uEyePos",           GLUniform::Vec3 };
    GLUniform uAmbientLight     { "uAmbientLight",     GLUniform::Vec4 };
    GLUniform uLightDirs        { "uLightDirs",        GLUniform::Vec3Array, MAX_LIGHTS };
    GLUniform uLightIntensities { "uLightIntensities", GLUniform::Vec4Array, MAX_LIGHTS };
    GLUniform uFogRange         { "uFogRange",         GLUniform::Vec4 };
    GLUniform uFogColor         { "uFogColor",         GLUniform::Vec4 };
    Matrix4f inverseLocal; ///< Translation ignored, this is used for light vectors.
    int lightCount = 0;
    Variable lightIntensityFactor;
    Impl(Public *i) : Base(i) {
        loader.audienceForNewProgram() += this;
        lightIntensityFactor.set(NumberValue(1.75));
    }
    void init() {
        loader.glInit();
    }
    void deinit() {
        loader.glDeinit();
    }
    void newProgramCreated(GLProgram &program) {
        program << uMvpMatrix
                << uReflectionMatrix
                << uWorldMatrix
                << uTex
                << uReflectionTex
                << uEyePos
                << uAmbientLight
                << uLightDirs
                << uLightIntensities
                << uFogRange
                << uFogColor;
    }
    void setupLighting(const VisEntityLighting &lighting,
                       const mobj_t *excludeSourceMobj,
                       float alpha = 1.0f) {
        setAmbientLight(lighting.ambientColor * .6f, alpha);
        clearLights();
        ClientApp::renderSystem().forAllVectorLights(lighting.vLightListIdx,
                                                     [this, excludeSourceMobj]
                                                     (VectorLightData const &vlight) {
            if (excludeSourceMobj && vlight.sourceMobj == excludeSourceMobj) {
                return LoopContinue;
            }
            addLight(vlight.direction.xzy(), vlight.color);
            return LoopContinue;
        });
        /*
        {
            Vector4f ambient = uAmbientLight;
            if (model.flags & render::Model::ThingAlphaAsAmbientLightAlpha) {
                ambient.w = p.alpha;
            }
            if (model.flags & render::Model::ThingFullBrightAsAmbientLight) {
                ambient = {1.0f, 1.0f, 1.0f, ambient.w};
            }
            uAmbientLight = ambient;
        }
*/
    }
    void setAmbientLight(Vector3f const &ambientIntensity, float alpha) {
        uAmbientLight = Vector4f(ambientIntensity, alpha);
    }
    void clearLights() {
        lightCount = 0;
        for (int i = 0; i < MAX_LIGHTS; ++i) {
            uLightDirs       .set(i, Vector3f());
            uLightIntensities.set(i, Vector4f());
        }
    }
    void addLight(Vector3f const &direction, Vector3f const &intensity) {
        if (lightCount == MAX_LIGHTS) return;
        int idx = lightCount;
        uLightDirs       .set(idx, (inverseLocal * direction).normalize());
        uLightIntensities.set(idx, Vector4f(intensity, intensity.max()) * lightIntensityFactor);
        lightCount++;
    }
    void setupPose(const Vec3d &modelWorldOrigin,
                   const Vec3f &modelOffset,
                   float yawAngle,
                   float pitchAngle,
                   float fixedFOVAngle, /* zero to use default FOV */
                   bool usePSpriteClipPlane,
                   const Mat4f *preModelToLocal = nullptr) {
        Vector3f const aspectCorrect(1.0f, 1.0f/1.2f, 1.0f);
        Vector3d origin = modelWorldOrigin + modelOffset * aspectCorrect;
        Matrix4f modelToLocal =
                Matrix4f::rotate(-90 + yawAngle, Vector3f(0, 1, 0) /* vertical axis for yaw */) *
                Matrix4f::rotate(pitchAngle,     Vector3f(1, 0, 0));
        Vector3f relativePos = Rend_EyeOrigin() - origin;
        uReflectionMatrix = Matrix4f::rotate(-yawAngle,  Vector3f(0, 1, 0)) *
                            Matrix4f::rotate(pitchAngle, Vector3f(0, 0, 1));
        if (preModelToLocal) {
            modelToLocal = modelToLocal * (*preModelToLocal);
        }
        const Matrix4f localToWorld = Matrix4f::translate(origin) *
                                      Matrix4f::scale(aspectCorrect); // Inverse aspect correction.
        const Matrix4f viewProj =
            Rend_GetProjectionMatrix(
                fixedFOVAngle,
                usePSpriteClipPlane ? 0.1f /* near plane distance: IssueID #2373 */ : 1.0f) *
            ClientApp::renderSystem().uViewMatrix().toMatrix4f();
        const Matrix4f localToScreen = viewProj * localToWorld;
        uWorldMatrix = localToWorld * modelToLocal;
        setTransformation(relativePos, modelToLocal, localToScreen);
    }
    /**
     * Sets up the transformation matrices.
     *
     * "Local space" is the same as world space but relative to the object's origin.
     * That is, (0,0,0) being the object's origin.
     *
     * @param relativeEyePos  Position of the eye in relation to object (in world space).
     * @param modelToLocal    Transformation from model space to the object's local space
     *                        (object's local frame in world space).
     * @param localToScreen   Transformation from local space to screen (projected 2D) space.
     */
    void setTransformation(Vector3f const &relativeEyePos,
                           Matrix4f const &modelToLocal,
                           Matrix4f const &localToScreen) {
        uMvpMatrix   = localToScreen * modelToLocal;
        inverseLocal = modelToLocal.inverse();
        uEyePos      = inverseLocal * relativeEyePos;
    }
    void setReflectionForSubsector(world::Subsector const *subsec) {
        uReflectionTex = ClientApp::renderSystem()
                            .environment().reflectionInSubsector(subsec);
    }
    void setReflectionForObject(mobj_t const *object) {
        if (object && Mobj_HasSubsector(*object)) {
            setReflectionForSubsector(&Mobj_Subsector(*object));
        }
        else {
            uReflectionTex = ClientApp::renderSystem()
                                 .environment().defaultReflection();
        }
    }
    template <typename Params> // generic to accommodate psprites and vispsprites
    void draw(Params const &p) {
        DGL_FogParams(uFogRange, uFogColor);
        uTex = static_cast<AtlasTexture const *>(p.model->textures->atlas());
        p.model->draw(&p.animator->appearance(), p.animator);
    }
};
ModelRenderer::ModelRenderer() : d(new Impl(this))
{}
void ModelRenderer::glInit() {
    d->init();
}
void ModelRenderer::glDeinit() {
    d->deinit();
}
render::ModelLoader &ModelRenderer::loader() {
    return d->loader;
}
render::ModelLoader const &ModelRenderer::loader() const {
    return d->loader;
}
ModelBank &ModelRenderer::bank() {
    return d->loader.bank();
}
render::Model::StateAnims const *ModelRenderer::animations(DotPath const &modelId) const {
    auto const &model = d->loader.bank().model<render::Model const>(modelId);
    if (!model.animations.isEmpty()) {
        return &model.animations;
    }
    return nullptr;
}
void ModelRenderer::render(vissprite_t const &spr) {
    /*
     * Work in progress:
     *
     * Here is the contact point between the old renderer and the new GL2 model renderer.
     * In the future, vissprites should form a class hierarchy, and the entire drawing
     * operation should be encapsulated within. This will allow drawing a model (or a
     * sprite, etc.) by creating a VisSprite instance and telling it to draw itself.
     */
    drawmodel2params_t const &p = spr.data.model2;
    auto const *mobjData = (p.object? &THINKER_DATA(p.object->thinker, ClientMobjThinkerData) :
                                      nullptr);
    d->setReflectionForObject(p.object);
    d->setupPose((spr.pose.origin + spr.pose.srvo).xzy(),
                 p.model->offset,
                 spr.pose.yaw + spr.pose.yawAngleOffset,
                 spr.pose.pitch + spr.pose.pitchAngleOffset,
                 0.0f, /* regular FOV */
                 false, /* regular clip planes */
                 mobjData ? &mobjData->modelTransformation() : nullptr);
    d->setupLighting(spr.light, nullptr);
    GLState::push().setCull(p.model->cull);
    d->draw(p);
    GLState::pop();
}
void ModelRenderer::render(vispsprite_t const &pspr, mobj_t const *playerMobj) {
    auto const &p = pspr.data.model2;
    world::ConvexSubspace const *sub = pspr.bspLeaf ? pspr.bspLeaf->subspacePtr() : nullptr;
    d->setReflectionForSubsector(sub ? sub->subsectorPtr() : nullptr);
    dfloat yaw   = vang   + p.yawAngleOffset;
    dfloat pitch = vpitch + p.pitchAngleOffset;
    Matrix4f eyeSpace = Matrix4f::rotate(180 - yaw, Vector3f(0, 1, 0))
                      * Matrix4f::rotate(pitch    , Vector3f(1, 0, 0));
    Matrix4f xform = p.model->transformation;
    d->setupPose(Rend_EyeOrigin(),
                 eyeSpace * p.model->offset,
                 -90 - yaw,
                 pitch,
                 p.model->pspriteFOV > 0 ? p.model->pspriteFOV : weaponFixedFOV,
                 true, /* nearby clip planes */
                 &xform);
    float opacity = 1.0f;
    if (p.model->flags & render::Model::ThingOpacityAsAmbientLightAlpha) {
        opacity = pspr.alpha;
    }
    d->setupLighting(pspr.light, playerMobj /* player holding weapon is excluded */, opacity);
    if (p.model->flags & render::Model::ThingFullBrightAsAmbientLight) {
        d->setAmbientLight({1.0f, 1.0f, 1.0f}, opacity);
    }
    GLState::push().setCull(p.model->cull);
    d->draw(p);
    GLState::pop();
}
static render::StateAnimator &animatorInstance(Context &ctx) {
    if (auto *self = maybeAs<NativePointerValue>(ctx.selfInstance().get(Record::VAR_NATIVE_SELF))) {
        if (auto *obj = self->nativeObject<render::StateAnimator>()) {
            return *obj;
        }
    }
    throw Value::IllegalError("ModelRenderer::animatorInstance",
                              "Not a StateAnimator instance");
}
static Value *Function_StateAnimator_Thing(Context &ctx, Function::ArgumentValues const &) {
    render::StateAnimator &anim = animatorInstance(ctx);
    if (anim.ownerNamespaceName() == QStringLiteral("__thing__")) {
        return anim[anim.ownerNamespaceName()].value().duplicate();
    }
    return nullptr;
}
static Value *Function_StateAnimator_PlayingSequences(Context &ctx, Function::ArgumentValues const &) {
    render::StateAnimator &anim = animatorInstance(ctx);
    std::unique_ptr<ArrayValue> playing(new ArrayValue);
    for (int i = 0; i < anim.count(); ++i) {
        playing->add(new NumberValue(anim.at(i).animId));
    }
    return playing.release();
}
static Value *Function_StateAnimator_StartSequence(Context &ctx, Function::ArgumentValues const &args) {
    render::StateAnimator &anim = animatorInstance(ctx);
    int animId = anim.animationId(args.at(0)->asText());
    if (animId >= 0) {
        int priority = args.at(1)->asInt();
        bool looping = args.at(2)->isTrue();
        String node  = args.at(3)->asText();
        anim.startAnimation(animId, priority, looping, node);
    }
    else {
        LOG_SCR_ERROR("%s has no animation \"%s\"")
                << anim.objectNamespace().gets("ID")
                << args.at(0)->asText();
    }
    return nullptr;
}
static Value *Function_StateAnimator_StartTimeline(Context &ctx, Function::ArgumentValues const &args) {
    render::StateAnimator &anim = animatorInstance(ctx);
    render::Model const &model = anim.model();
    String const timelineName = args.first()->asText();
    if (model.timelines.contains(timelineName)) {
        anim.scheduler().start(*model.timelines[timelineName],
                               &anim.objectNamespace(),
                               timelineName);
        return new TextValue(timelineName);
    }
    else {
        LOG_SCR_ERROR("%s has no timeline \"%s\"")
            << anim.objectNamespace().gets("ID") << timelineName;
    }
    return nullptr;
}
static Value *Function_StateAnimator_StopTimeline(Context &ctx, Function::ArgumentValues const &args) {
    render::StateAnimator &anim = animatorInstance(ctx);
    anim.scheduler().stop(args.first()->asText());
    return nullptr;
}
void ModelRenderer::initBindings(Binder &binder, Record &module) // static
{ {
        Record &anim = module.addSubrecord("StateAnimator");
        binder.init(anim)
                << DENG2_FUNC_NOARG(StateAnimator_Thing,            "thing")
                << DENG2_FUNC_NOARG(StateAnimator_PlayingSequences, "playingSequences")
                << DENG2_FUNC_DEFS (StateAnimator_StartSequence,    "startSequence",
                                    "sequence" << "priority" << "looping" << "node",
                                    Function::Defaults({ std::make_pair("priority", new NumberValue(0)),
                                                         std::make_pair("looping",  new NumberValue(0)),
                                                         std::make_pair("node",     new TextValue) }))
                << DENG2_FUNC      (StateAnimator_StartTimeline,    "startTimeline", "name")
                << DENG2_FUNC      (StateAnimator_StopTimeline,     "stopTimeline", "name");
    }
}
namespace render {
DENG2_PIMPL_NOREF(PlayerWeaponAnimator)
, DENG2_OBSERVES(Asset, Deletion) {
    ClientPlayer *player;
    String identifier;
    std::unique_ptr<StateAnimator> animator;
    AnimationVector2 angleOffset { Animation::Linear };
    Impl(ClientPlayer *plr)
        : player(plr)
    {}
    void setupAsset(String const &identifier) {
        this->identifier = identifier;
        angleOffset = Vector2f();
        if (animator) {
            animator->model().audienceForDeletion() -= this;
        }
        if (modelBank().has(identifier)) {
            auto &model = modelBank().model<Model>(identifier);
            model.audienceForDeletion() += this;
            animator.reset(new StateAnimator(identifier, model));
            animator->setOwnerNamespace(player->info(), QStringLiteral("__player__"));
        }
        else {
            animator.reset();
            this->identifier.clear();
        }
    }
    void assetBeingDeleted(Asset &) {
        de::trash(animator.release());
    }
    static ModelBank &modelBank() {
        return ClientApp::renderSystem().modelRenderer().bank();
    }
};
PlayerWeaponAnimator::PlayerWeaponAnimator(ClientPlayer *plr)
    : d(new Impl(plr))
{}
void PlayerWeaponAnimator::setAsset(String const &identifier) {
    d->setupAsset(identifier);
}
String PlayerWeaponAnimator::assetId() const {
    return d->identifier;
}
void PlayerWeaponAnimator::stateChanged(state_s const *state) {
    if (d->animator) {
        d->animator->triggerByState(Def_GetStateName(state));
    }
}
StateAnimator &PlayerWeaponAnimator::animator() {
    DENG2_ASSERT(hasModel());
    return *d->animator;
}
void PlayerWeaponAnimator::setupVisPSprite(vispsprite_t &spr) const {
    DENG2_ASSERT(hasModel());
    spr.type = VPSPR_MODEL2;
    spr.data.model2.model = model();
    spr.data.model2.animator = d->animator.get();
    float bob[2] = { *(float *) gx.GetPointer(DD_PSPRITE_BOB_X),
                     *(float *) gx.GetPointer(DD_PSPRITE_BOB_Y) };
    Vector2f angles(
    /* yaw: */   bob[0] * weaponOffsetScale,
    /* pitch: */ (32 - bob[1]) * weaponOffsetScale * weaponOffsetScaleY / 1000.0f);
    TimeSpan const span = 1.0 / Timer_TicksPerSecond();
    d->angleOffset.setValueIfDifferentTarget(angles, span);
    spr.data.model2.yawAngleOffset   = d->angleOffset.x;
    spr.data.model2.pitchAngleOffset = d->angleOffset.y;
}
void PlayerWeaponAnimator::advanceTime(TimeSpan const &elapsed) {
    if (clientPaused) return;
    if (d->animator) {
        d->animator->advanceTime(elapsed);
    }
}
bool PlayerWeaponAnimator::hasModel() const {
    return bool(d->animator);
}
Model const *PlayerWeaponAnimator::model() const {
    if (!hasModel()) return nullptr;
    return &d->animator->model();
}
} // namespace render
ProjectionList::Node *ProjectionList::firstNode = nullptr;
ProjectionList::Node *ProjectionList::cursorNode = nullptr;
void ProjectionList::init()  // static
{
    static bool firstTime = true;
    if(firstTime) {
        firstNode  = 0;
        cursorNode = 0;
        firstTime  = false;
    }
}
void ProjectionList::rewind()  // static
{
    cursorNode = firstNode;
}
static dfloat luminosity(ProjectedTextureData const &pt)  // static
{
    return (pt.color.x + pt.color.y + pt.color.z) / 3 * pt.color.w;
}
ProjectionList &ProjectionList::add(ProjectedTextureData &texp) {
    Node *node = newNode();
    node->projection = texp;
    if(head && sortByLuma) {
        dfloat luma = luminosity(node->projection);
        Node *iter = head;
        Node *last = iter;
        do {
            if(luminosity(iter->projection) > luma) {
                last = iter;
                iter = iter->next;
            }
            else {
                node->next = last->next;
                last->next = node;
                return *this;
            }
        } while(iter);
    }
    node->next = head;
    head = node;
    return *this;
}
ProjectionList::Node *ProjectionList::newNode()  // static
{
    Node *node;
    if(!cursorNode) {
        node = (Node *) Z_Malloc(sizeof(*node), PU_APPSTATIC, nullptr);
        node->nextUsed = firstNode;
        firstNode = node;
    }
    else {
        node = cursorNode;
        cursorNode = cursorNode->nextUsed;
    }
    node->next = nullptr;
    return node;
}
enum {
    BG_BACKGROUND,
    BG_TOP,
    BG_RIGHT,
    BG_BOTTOM,
    BG_LEFT,
    BG_TOPLEFT,
    BG_TOPRIGHT,
    BG_BOTTOMRIGHT,
    BG_BOTTOMLEFT
};
static bool initedDraw = false;
static int borderSize;
static de::Uri *borderGraphicsNames[9];
static patchid_t borderPatches[9];
static void loadViewBorderPatches() {
    borderPatches[0] = 0;
    for(uint i = 1; i < 9; ++i) {
        borderPatches[i] = res::Textures::get().declarePatch(borderGraphicsNames[i]->path());
    }
    borderSize = 0;
    patchinfo_t info;
    if(!R_GetPatchInfo(borderPatches[BG_TOP], &info)) {
        return;
    }
    borderSize = info.geometry.size.height;
}
static ClientTexture &borderTexture(int borderComp) {
    res::TextureScheme &patches = res::Textures::get().textureScheme("Patches");
    DENG2_ASSERT(borderComp >= 0 && borderComp < 9);
    return static_cast<ClientTexture &>(patches.findByUniqueId(borderPatches[borderComp]).texture());
}
#undef R_SetBorderGfx
DENG_EXTERN_C void R_SetBorderGfx(struct uri_s const *const *paths) {
    DENG2_ASSERT(initedDraw);
    if(!paths) return;
    for(uint i = 0; i < 9; ++i) {
        if(paths[i]) {
            if(!borderGraphicsNames[i]) {
                borderGraphicsNames[i] = new de::Uri;
            }
            *(borderGraphicsNames[i]) = *reinterpret_cast<de::Uri const *>(paths[i]);
        }
        else {
            if(borderGraphicsNames[i]) {
                delete borderGraphicsNames[i];
            }
            borderGraphicsNames[i] = 0;
        }
    }
    loadViewBorderPatches();
}
void R_InitViewWindow() {
    if(Sys_IsShuttingDown()) return;
    for(int i = 0; i < DDMAXPLAYERS; ++i) {
        R_SetupDefaultViewWindow(i);
    }
    if(initedDraw) {
        for(int i = 0; i < 9; ++i) {
            if(borderGraphicsNames[i]) {
                delete borderGraphicsNames[i];
            }
        }
    }
    de::zap(borderGraphicsNames);
    de::zap(borderPatches);
    borderSize = 0;
    initedDraw = true;
}
void R_ShutdownViewWindow() {
    if(!initedDraw) return;
    for(int i = 0; i < 9; ++i) {
        if(borderGraphicsNames[i]) {
            delete borderGraphicsNames[i];
        }
    }
    de::zap(borderGraphicsNames);
    initedDraw = false;
}
TextureVariantSpec const &Rend_PatchTextureSpec(int flags, gl::Wrapping wrapS,
    gl::Wrapping wrapT) {
    return ClientApp::resources().textureSpec(TC_UI, flags, 0, 0, 0,
        GL_Wrap(wrapS), GL_Wrap(wrapT), 0, -3, 0, false, false, false, false);
}
void R_DrawPatch(ClientTexture &texture, int x, int y, int w, int h, bool useOffsets) {
    if(texture.manifest().schemeName().compareWithoutCase("Patches")) {
        LOG_AS("R_DrawPatch3");
        LOGDEV_GL_WARNING("Cannot draw a non-patch [%p]") << dintptr(&texture);
        return;
    }
    TextureVariantSpec const &texSpec =
        Rend_PatchTextureSpec(0 | (texture.isFlagged(res::Texture::Monochrome)        ? TSF_MONOCHROME : 0)
                                | (texture.isFlagged(res::Texture::UpscaleAndSharpen) ? TSF_UPSCALE_AND_SHARPEN : 0));
    GL_BindTexture(texture.prepareVariant(texSpec));
    if(useOffsets) {
        x += texture.origin().x;
        y += texture.origin().y;
    }
    GL_DrawRectf2Color(x, y, w, h, 1, 1, 1, 1);
}
void R_DrawPatch(ClientTexture &tex, int x, int y) {
    R_DrawPatch(tex, x, y, tex.width(), tex.height());
}
void R_DrawPatchTiled(ClientTexture &texture, int x, int y, int w, int h,
    gl::Wrapping wrapS, gl::Wrapping wrapT) {
    TextureVariantSpec const &spec =
        Rend_PatchTextureSpec(0 | (texture.isFlagged(res::Texture::Monochrome)        ? TSF_MONOCHROME : 0)
                                | (texture.isFlagged(res::Texture::UpscaleAndSharpen) ? TSF_UPSCALE_AND_SHARPEN : 0),
                              wrapS, wrapT);
    GL_BindTexture(texture.prepareVariant(spec));
    GL_DrawRectf2Tiled(x, y, w, h, texture.width(), texture.height());
}
static MaterialVariantSpec const &bgMaterialSpec() {
    return ClientApp::resources().materialSpec(UiContext, 0, 0, 0, 0,
                                                    GL_REPEAT, GL_REPEAT, 0, -3,
                                                    0, false, false, false, false);
}
void R_DrawViewBorder() {
    DENG2_ASSERT(initedDraw);
    viewport_t const *port = R_CurrentViewPort();
    viewdata_t const *vd = &DD_Player(displayPlayer)->viewport();
    DENG2_ASSERT(port != 0 && vd != 0);
    if (!borderGraphicsNames[BG_BACKGROUND]) return;
    if (vd->window.isNull()) return;
    if (vd->window.size() >= port->geometry.size()) return;
    Vector2i const origin = port->geometry.topLeft;
    DENG2_ASSERT_IN_RENDER_THREAD();
    DENG_ASSERT_GL_CONTEXT_ACTIVE();
    DGL_Enable(DGL_TEXTURE_2D);
    DGL_MatrixMode(DGL_TEXTURE);
    DGL_PushMatrix();
    int border;
    if(port->geometry.width() >= port->geometry.height()) {
        DGL_Scalef(float(SCREENHEIGHT) / port->geometry.height(),
                   float(SCREENHEIGHT) / port->geometry.height(), 1);
        border = float(borderSize) / SCREENHEIGHT * port->geometry.height();
    }
    else {
        DGL_Scalef(float(SCREENWIDTH) / port->geometry.width(),
                   float(SCREENWIDTH) / port->geometry.width(), 1);
        border = float(borderSize) / SCREENWIDTH * port->geometry.width();
    }
    DGL_Color4f(1, 1, 1, 1);
    try {
        MaterialAnimator &matAnimator = ClientMaterial::find(*borderGraphicsNames[BG_BACKGROUND])
                .getAnimator(bgMaterialSpec());
        matAnimator.prepare();
        GL_BindTexture(matAnimator.texUnit(MaterialAnimator::TU_LAYER0).texture);
        Vector2ui const &matDimensions = matAnimator.dimensions();
        GL_DrawCutRectf2Tiled(origin.x, origin.y, port->geometry.width(), port->geometry.height(),
                              matDimensions.x, matDimensions.y, 0, 0,
                              origin.x + vd->window.topLeft.x - border,
                              origin.y + vd->window.topLeft.y - border,
                              vd->window.width() + 2 * border, vd->window.height() + 2 * border);
    }
    catch(world::MaterialManifest::MissingMaterialError const &)
    {} // Ignore this error.
    if(border) {
        R_DrawPatchTiled(borderTexture(BG_TOP),    origin.x + vd->window.topLeft.x,          origin.y + vd->window.topLeft.y - border, vd->window.width(), border, gl::Repeat, gl::ClampToEdge);
        R_DrawPatchTiled(borderTexture(BG_BOTTOM), origin.x + vd->window.topLeft.x,          origin.y + vd->window.bottomRight.y, vd->window.width(), border, gl::Repeat, gl::ClampToEdge);
        R_DrawPatchTiled(borderTexture(BG_LEFT),   origin.x + vd->window.topLeft.x - border, origin.y + vd->window.topLeft.y, border, vd->window.height(), gl::ClampToEdge, gl::Repeat);
        R_DrawPatchTiled(borderTexture(BG_RIGHT),  origin.x + vd->window.topRight().x,       origin.y + vd->window.topRight().y, border, vd->window.height(), gl::ClampToEdge, gl::Repeat);
    }
    DGL_MatrixMode(DGL_TEXTURE);
    DGL_PopMatrix();
    if(border) {
        R_DrawPatch(borderTexture(BG_TOPLEFT),     origin.x + vd->window.topLeft.x - border,      origin.y + vd->window.topLeft.y - border, border, border, false);
        R_DrawPatch(borderTexture(BG_TOPRIGHT),    origin.x + vd->window.topRight().x,            origin.y + vd->window.topLeft.y - border, border, border, false);
        R_DrawPatch(borderTexture(BG_BOTTOMRIGHT), origin.x + vd->window.bottomRight.x,           origin.y + vd->window.bottomRight.y, border, border, false);
        R_DrawPatch(borderTexture(BG_BOTTOMLEFT),  origin.x + vd->window.bottomLeft().x - border, origin.y + vd->window.bottomRight.y, border, border, false);
    }
    DGL_Disable(DGL_TEXTURE_2D);
}
dint levelFullBright;
dint weaponOffsetScaleY = 1000;
dint psp3d;
dfloat pspLightLevelMultiplier = 1;
dfloat pspOffset[2];
/*
 * Console variables:
 */
dfloat weaponFOVShift    = 45;
dfloat weaponOffsetScale = 0.3183f;  // 1/Pi
dbyte weaponScaleMode    = SCALEMODE_SMART_STRETCH;
static MaterialVariantSpec const &pspriteMaterialSpec() {
    return ClientApp::resources().materialSpec(PSpriteContext, 0, 1, 0, 0, GL_CLAMP_TO_EDGE, GL_CLAMP_TO_EDGE,
                                 0, -2, 0, false, true, true, false);
}
static void setupPSpriteParams(rendpspriteparams_t &parm, vispsprite_t const &vs) {
    static dint const WEAPONTOP = 32;  /// @todo Currently hardcoded here and in the plugins.
    dfloat const offScaleY = ::weaponOffsetScaleY / 1000.0f;
    DENG2_ASSERT(vs.psp);
    ddpsprite_t const &psp = *vs.psp;
    DENG2_ASSERT(psp.statePtr);
    state_t const &state = *psp.statePtr;
    defn::Sprite::View const spriteView = defn::Sprite(res::Sprites::get().sprite(state.sprite, state.frame)).view(0);
    MaterialAnimator &matAnimator = ClientMaterial::find(*spriteView.material)
            .getAnimator(pspriteMaterialSpec());
    matAnimator.prepare();
    TextureVariant const &tex             = *matAnimator.texUnit(MaterialAnimator::TU_LAYER0).texture;
    Vector2i const &texOrigin             = tex.base().origin();
    variantspecification_t const &texSpec = tex.spec().variant;
    parm.pos[0] = psp.pos[0] + texOrigin.x + pspOffset[0] - texSpec.border;
    parm.pos[1] = WEAPONTOP + offScaleY * (psp.pos[1] - WEAPONTOP) + texOrigin.y
                + pspOffset[1] - texSpec.border;
    Vector2ui const dimensions = matAnimator.dimensions() + Vector2ui(texSpec.border, texSpec.border) * 2;
    parm.width  = dimensions.x;
    parm.height = dimensions.y;
    tex.glCoords(&parm.texOffset[0], &parm.texOffset[1]);
    parm.texFlip[0] = spriteView.mirrorX;
    parm.texFlip[1] = false;
    parm.mat        = &matAnimator.material();
    parm.ambientColor[3] = vs.alpha;
    if (vs.light.isFullBright) {
        parm.ambientColor[0] = parm.ambientColor[1] = parm.ambientColor[2] = 1;
        parm.vLightListIdx = 0;
    }
    else {
        DENG2_ASSERT(vs.bspLeaf);
#if 0
        world::Map const &map = ClientApp::world().map();
        if (useBias && map.hasLightGrid()) {
            Vector4f color = map.lightGrid().evaluate(vs.origin);
            for (dint i = 0; i < 3; ++i) {
                color[i] += Rend_LightAdaptationDelta(color[i]);
            }
            V3f_Set(parm.ambientColor, color.x, color.y, color.z);
        }
        else
#endif
        {
            auto const &subsec   = vs.bspLeaf->subspace().subsector().as<world::ClientSubsector>();
            Vector4f const color = subsec.lightSourceColorfIntensity();
            dfloat lightLevel = color.w;
            lightLevel += Rend_ExtraLightDelta();
            lightLevel *= pspLightLevelMultiplier;
            Rend_ApplyLightAdaptation(lightLevel);
            for (dint i = 0; i < 3; ++i) {
                parm.ambientColor[i] = lightLevel * color[i];
            }
        }
        Rend_ApplyTorchLight(parm.ambientColor, 0);
        parm.vLightListIdx =
                Rend_CollectAffectingLights(vs.origin, Vector3f(parm.ambientColor),
                                            vs.bspLeaf->subspacePtr());
    }
}
void Rend_Draw2DPlayerSprites() {
    if (!viewPlayer) return;
    ddplayer_t const &ddpl = viewPlayer->publicData();
    if (ddpl.flags & DDPF_CAMERA  ) return;
    if (ddpl.flags & DDPF_CHASECAM) return;
    if (fogParams.usingFog) {
        DGL_Enable(DGL_FOG);
    }
    for (vispsprite_t const &vs : visPSprites) {
        if (vs.type != VPSPR_SPRITE) continue;  // No...
        if (!vs.psp || !vs.psp->statePtr) continue;
        try {
            rendpspriteparams_t parm; setupPSpriteParams(parm, vs);
            Rend_DrawPSprite(parm);
        }
        catch (Resources::MissingResourceManifestError const &er) {
            state_t const &state = *vs.psp->statePtr;
            LOG_GL_WARNING("Drawing psprite '%i' frame '%i': %s")
                    << state.sprite << state.frame << er.asText();
        }
    }
    if (fogParams.usingFog) {
        DGL_Disable(DGL_FOG);
    }
}
static void setupModelParamsForVisPSprite(vissprite_t &vis, vispsprite_t const &spr) {
    drawmodelparams_t *params = VS_MODEL(&vis);
    params->mf = spr.data.model.mf;
    params->nextMF = spr.data.model.nextMF;
    params->inter = spr.data.model.inter;
    params->alwaysInterpolate = false;
    params->id = spr.data.model.id;
    params->selector = spr.data.model.selector;
    params->flags = spr.data.model.flags;
    vis.pose.origin = spr.origin;
    vis.pose.srvo[0] = spr.data.model.visOff[0];
    vis.pose.srvo[1] = spr.data.model.visOff[1];
    vis.pose.srvo[2] = spr.data.model.visOff[2] - spr.data.model.floorClip;
    vis.pose.topZ = spr.data.model.topZ;
    vis.pose.distance = -10;
    vis.pose.yaw = spr.data.model.yaw;
    vis.pose.extraYawAngle = 0;
    vis.pose.yawAngleOffset = spr.data.model.yawAngleOffset;
    vis.pose.pitch = spr.data.model.pitch;
    vis.pose.extraPitchAngle = 0;
    vis.pose.pitchAngleOffset = spr.data.model.pitchAngleOffset;
    vis.pose.extraScale = 0;
    vis.pose.viewAligned = spr.data.model.viewAligned;
    vis.pose.mirrored = (mirrorHudModels? true : false);
    params->shineYawOffset = -vang;
    params->shinePitchOffset = vpitch + 90;
    params->shineTranslateWithViewerPos = false;
    params->shinepspriteCoordSpace = true;
    vis.light.ambientColor[3] = spr.alpha;
    if ((levelFullBright || spr.light.isFullBright) &&
       !spr.data.model.mf->testSubFlag(0, MFF_DIM)) {
        vis.light.ambientColor[0] = vis.light.ambientColor[1] = vis.light.ambientColor[2] = 1;
        vis.light.vLightListIdx = 0;
    }
    else {
        vis.light.setupLighting(vis.pose.origin, vis.pose.distance, *spr.bspLeaf);
    }
}
void Rend_Draw3DPlayerSprites() {
    GL_ProjectionMatrix(true /* fixed FOV for psprites */);
    Rend_ModelViewMatrix(false /* don't apply view angle rotation */);
    bool first = true;
    for (vispsprite_t const &spr : visPSprites) {
        if (spr.type != VPSPR_MODEL &&
            spr.type != VPSPR_MODEL2) continue;
        if (first) {
            first = false;
            GLState::current().target().clear(GLFramebuffer::Depth);
        }
        if (spr.type == VPSPR_MODEL) {
            vissprite_t vs;
            setupModelParamsForVisPSprite(vs, spr);
            Rend_DrawModel(vs);
        }
        else {
            vispsprite_t lit = spr;
            lit.light.setupLighting(spr.origin + Vector3d(0, 0, -10), -10, *spr.bspLeaf);
            ClientApp::renderSystem().modelRenderer()
                    .render(lit, viewPlayer->publicData().mo);
        }
    }
    GL_ProjectionMatrix(false);
}
static void evaluateLighting(Vector3d const &origin, ConvexSubspace &subspaceAtOrigin,
    coord_t distToEye, bool fullbright, Vector4f &ambientColor, duint *vLightListIdx) {
    if(fullbright) {
        ambientColor = Vector3f(1, 1, 1);
        *vLightListIdx = 0;
    }
    else {
        auto &subsec = subspaceAtOrigin.subsector().as<world::ClientSubsector>();
#if 0
        Map &map = subsec.sector().map();
        if(useBias && map.hasLightGrid()) {
            Vector4f color = map.lightGrid().evaluate(origin);
            for(dint i = 0; i < 3; ++i) {
                color[i] += Rend_LightAdaptationDelta(color[i]);
            }
            ambientColor = color;
        }
        else
#endif
        {
            Vector4f const color = subsec.lightSourceColorfIntensity();
            dfloat lightLevel = color.w;
            /* if(spr->type == VSPR_DECORATION)
            {
                lightLevel += R_WallAngleLightLevelDelta(line, side);
            } */
            lightLevel = Rend_AttenuateLightLevel(distToEye, lightLevel);
            lightLevel = de::clamp(0.f, lightLevel + Rend_ExtraLightDelta(), 1.f);
            Rend_ApplyLightAdaptation(lightLevel);
            ambientColor = color * lightLevel;
        }
        Rend_ApplyTorchLight(ambientColor, distToEye);
        *vLightListIdx = Rend_CollectAffectingLights(origin, ambientColor, &subspaceAtOrigin);
    }
}
static Vector3d mobjOriginSmoothed(mobj_t *mob) {
    DENG2_ASSERT(mob);
    coord_t origin[] = { mob->origin[0], mob->origin[1], mob->origin[2] };
    if(isClient && mob->dPlayer && P_GetDDPlayerIdx(mob->dPlayer) != consolePlayer) {
        Smoother_Evaluate(DD_Player(P_GetDDPlayerIdx(mob->dPlayer))->smoother(), origin);
    }
    return origin;
}
/**
 * Determine the correct Z coordinate for the mobj. The visible Z coordinate
 * may be slightly different than the actual Z coordinate due to smoothed
 * plane movement.
 *
 * @todo fixme: Should use the visual plane heights of subsectors.
 */
static void findMobjZOrigin(mobj_t &mob, bool floorAdjust, vissprite_t &vis) {
    validCount++;
    Mobj_Map(mob).forAllSectorsTouchingMobj(mob, [&mob, &floorAdjust, &vis](Sector &sector) {
        if (floorAdjust && fequal(mob.origin[2], sector.floor().height())) {
            vis.pose.origin.z = sector.floor().heightSmoothed();
        }
        if (fequal(mob.origin[2] + mob.height, sector.ceiling().height())) {
            vis.pose.origin.z = sector.ceiling().heightSmoothed() - mob.height;
        }
        return LoopContinue;
    });
}
void R_ProjectSprite(mobj_t &mob) {
    if((mob.ddFlags & DDMF_DONTDRAW)) return;
    if(!Mobj_HasSubsector(mob)) return;
    if(!mob.state || !runtimeDefs.states.indexOf(mob.state)) return;
    Record const *spriteRec = Mobj_SpritePtr(mob);
    if(!spriteRec) return;
    dfloat const alpha = Mobj_Alpha(mob);
    if(alpha <= 0) return;
    ConvexSubspace &subspace = Mobj_BspLeafAtOrigin(mob).subspace();
    auto &subsec = subspace.subsector().as<ClientSubsector>();
    if(!subsec.hasWorldVolume()) return;
    ClientMobjThinkerData const *mobjData = THINKER_DATA_MAYBE(mob.thinker, ClientMobjThinkerData);
    Vector3d const moPos = mobjOriginSmoothed(&mob);
    coord_t const distFromEye = Rend_PointDist2D(moPos);
    FrameModelDef *mf = nullptr, *nextmf = nullptr;
    dfloat interp = 0;
    render::StateAnimator const *animator = nullptr; // GL2 model present?
    if(useModels) {
        mf = Mobj_ModelDef(mob, &nextmf, &interp);
        if(mf) {
            DENG2_ASSERT(mf->select == (mob.selector & DDMOBJ_SELECTOR_MASK))
            if(maxModelDistance && !(mf->flags & MFF_NO_DISTANCE_CHECK)
               && distFromEye > maxModelDistance) {
                mf = nextmf = nullptr;
                interp = -1;
            }
        }
        if(mobjData) {
            animator = mobjData->animator();
        }
    }
    bool const hasModel = (mf || animator);
    ClientMaterial *mat = nullptr;
    bool matFlipS = false;
    bool matFlipT = false;
    defn::Sprite const sprite(*spriteRec);
    defn::Sprite::View const spriteView = sprite.nearestView(mob.angle, R_ViewPointToAngle(mob.origin), !!hasModel); {
        if (auto *sprMat = world::Materials::get().materialPtr(*spriteView.material)) {
            mat = &sprMat->as<ClientMaterial>();
        }
        matFlipS = spriteView.mirrorX;
    }
    /*catch(defn::Sprite::MissingViewError const &er)
    {
        LOG_GL_WARNING("Projecting sprite '%i' frame '%i': %s")
                << mob.sprite << mob.frame << er.asText();
    }*/
    if(!mat) return;
    MaterialAnimator &matAnimator = mat->getAnimator(Rend_SpriteMaterialSpec(mob.tclass, mob.tmap));
    matAnimator.prepare();
    Vector2ui const &matDimensions = matAnimator.dimensions();
    TextureVariant *tex            = matAnimator.texUnit(MaterialAnimator::TU_LAYER0).texture;
    if(!tex || tex->base().manifest().schemeName().compareWithoutCase(QStringLiteral("Sprites"))) {
        return;
    }
    bool const fullbright = ((mob.state->flags & STF_FULLBRIGHT) != 0 || levelFullBright);
    bool viewAlign  = (!hasModel && ((mob.ddFlags & DDMF_VIEWALIGN) || alwaysAlign == 1))
                       || alwaysAlign == 3;
    coord_t const visWidth = Mobj_VisualRadius(mob) * 2;  /// @todo ignorant of rotation...
    Vector2d v1, v2;
    R_ProjectViewRelativeLine2D(moPos, hasModel || viewAlign, visWidth,
                                (hasModel? 0 : coord_t(-tex->base().origin().x) - (visWidth / 2.0f)),
                                v1, v2);
    if(!ClientApp::renderSystem().angleClipper().checkRangeFromViewRelPoints(v1, v2)) {
        coord_t const MAX_OBJECT_RADIUS = 128;
        if(!hasModel) return;
        viewdata_t const *viewData = &viewPlayer->viewport();
        Vector2d delta(distFromEye, moPos.z + (mob.height / 2) - viewData->current.origin.z);
        if(M_ApproxDistance(delta.x, delta.y) > MAX_OBJECT_RADIUS)
            return;
    }
    vissprite_t *vis = R_NewVisSprite(animator ? VSPR_MODELDRAWABLE :
                                            mf ? VSPR_MODEL :
                                                 VSPR_SPRITE);
    vis->pose.origin   = moPos;
    vis->pose.distance = distFromEye;
    Plane &floor     = subsec.visFloor();
    Plane &ceiling   = subsec.visCeiling();
    bool floorAdjust = false;
    if(!Mobj_OriginBehindVisPlane(&mob)) {
        floorAdjust = de::abs(floor.heightSmoothed() - floor.height()) < 8;
        findMobjZOrigin(mob, floorAdjust, *vis);
    }
    coord_t topZ = vis->pose.origin.z + -tex->base().origin().y;  // global z top
    coord_t floorClip = mob.floorClip;
    if(mob.ddFlags & DDMF_BOB) {
        floorClip += Mobj_BobOffset(mob);
    }
    dfloat yaw = 0, pitch = 0;
    if((mf && mf->testSubFlag(0, MFF_ALIGN_YAW)) ||
       (animator && animator->model().alignYaw == render::Model::AlignToView)) {
        viewdata_t const *viewData = &viewPlayer->viewport();
        Vector2d delta(moPos.y - viewData->current.origin.y,
                       moPos.x - viewData->current.origin.x);
        yaw = 90 - (BANG2RAD(bamsAtan2(delta.x * 10, delta.y * 10)) - PI / 2) / PI * 180;
    }
    else if(mf && mf->testSubFlag(0, MFF_SPIN)) {
        yaw = modelSpinSpeed * 70 * App_World().time() + MOBJ_TO_ID(&mob) % 360;
    }
    else if((mf && mf->testSubFlag(0, MFF_MOVEMENT_YAW)) ||
            (animator && animator->model().alignYaw == render::Model::AlignToMomentum)) {
        yaw = R_MovementXYYaw(mob.mom[0], mob.mom[1]);
    }
    else {
        yaw = Mobj_AngleSmoothed(&mob) / dfloat(ANGLE_MAX) * -360.f;
    }
    if((mf && mf->testSubFlag(0, MFF_IDANGLE)) ||
       (animator && animator->model().alignYaw == render::Model::AlignRandomly)) {
        yaw += MOBJ_TO_ID(&mob) % 360;  // arbitrary
    }
    if((mf && mf->testSubFlag(0, MFF_ALIGN_PITCH)) ||
       (animator && animator->model().alignPitch == render::Model::AlignToView)) {
        viewdata_t const *viewData = &viewPlayer->viewport();
        Vector2d delta(vis->pose.midZ() - viewData->current.origin.z, distFromEye);
        pitch = -BANG2DEG(bamsAtan2(delta.x * 10, delta.y * 10));
    }
    else if((mf && mf->testSubFlag(0, MFF_MOVEMENT_PITCH)) ||
            (animator && animator->model().alignPitch == render::Model::AlignToMomentum)) {
        pitch = R_MovementXYZPitch(mob.mom[0], mob.mom[1], mob.mom[2]);
    }
    Vector3d visOff;
    if((hasModel && useSRVO > 0) || (!hasModel && useSRVO > 1)) {
        if(mob.tics >= 0) {
            visOff = Vector3d(mob.srvo) * (mob.tics - frameTimePos) / (float) mob.state->tics;
        }
        if(!INRANGE_OF(mob.mom[0], 0, NOMOMENTUM_THRESHOLD) ||
           !INRANGE_OF(mob.mom[1], 0, NOMOMENTUM_THRESHOLD) ||
           !INRANGE_OF(mob.mom[2], 0, NOMOMENTUM_THRESHOLD)) {
            visOff += Vector3d(~mob.ddFlags & DDMF_MOVEBLOCKEDX ? mob.mom[VX] : 0.0,
                               ~mob.ddFlags & DDMF_MOVEBLOCKEDY ? mob.mom[VY] : 0.0,
                               ~mob.ddFlags & DDMF_MOVEBLOCKEDZ ? mob.mom[VZ] : 0.0) *
                      frameTimePos;
        }
    }
    if(!hasModel) {
        bool const brightShadow = (mob.ddFlags & DDMF_BRIGHTSHADOW) != 0;
        bool const fitTop       = (mob.ddFlags & DDMF_FITTOP)       != 0;
        bool const fitBottom    = (mob.ddFlags & DDMF_NOFITBOTTOM)  == 0;
        blendmode_t blendMode;
        if(brightShadow) {
            blendMode = BM_ADD;
        }
        else if(noSpriteTrans && alpha >= .98f) {
            blendMode = BM_ZEROALPHA;
        }
        else {
            blendMode = BM_NORMAL;
        }
        if(matDimensions.y < ceiling.heightSmoothed() - floor.heightSmoothed()) {
            if(fitTop && topZ > ceiling.heightSmoothed())
                topZ = ceiling.heightSmoothed();
            if(floorAdjust && fitBottom && topZ - matDimensions.y < floor.heightSmoothed())
                topZ = floor.heightSmoothed() + matDimensions.y;
        }
        topZ -= floorClip;
        Vector3d const origin(vis->pose.origin.x, vis->pose.origin.y, topZ - matDimensions.y / 2.0f);
        Vector4f ambientColor;
        duint vLightListIdx = 0;
        evaluateLighting(origin, subspace, vis->pose.distance, fullbright, ambientColor, &vLightListIdx);
        ambientColor.w = alpha;
        VisSprite_SetupSprite(vis,
                              VisEntityPose(origin, visOff, viewAlign),
                              VisEntityLighting(ambientColor, vLightListIdx),
                              floor.heightSmoothed(), ceiling.heightSmoothed(),
                              floorClip, topZ, *mat, matFlipS, matFlipT, blendMode,
                              mob.tclass, mob.tmap,
                              &Mobj_BspLeafAtOrigin(mob),
                              floorAdjust, fitTop, fitBottom);
    }
    else // It will be drawn as a 3D model.
    {
        Vector4f ambientColor;
        duint vLightListIdx = 0;
        evaluateLighting(vis->pose.origin, subspace, vis->pose.distance,
                         fullbright && !animator, // GL2 models lit with more granularity
                         ambientColor, &vLightListIdx);
        ambientColor.w = alpha;
        if(animator) {
            vis->pose = VisEntityPose(vis->pose.origin,
                                      Vector3d(visOff.x, visOff.y, visOff.z - floorClip),
                                      /*viewAlign*/ false, topZ, yaw, 0, pitch, 0);
            vis->light = VisEntityLighting(ambientColor, vLightListIdx);
            vis->data.model2.object   = &mob;
            vis->data.model2.animator = animator;
            vis->data.model2.model    = &animator->model();
        }
        else {
            DENG2_ASSERT(mf);
            VisSprite_SetupModel(vis,
                                 VisEntityPose(vis->pose.origin,
                                               Vector3d(visOff.x, visOff.y, visOff.z - floorClip),
                                               viewAlign, topZ, yaw, 0, pitch, 0),
                                 VisEntityLighting(ambientColor, vLightListIdx),
                                 mf, nextmf, interp,
                                 mob.thinker.id, mob.selector,
                                 &Mobj_BspLeafAtOrigin(mob),
                                 mob.ddFlags, mob.tmap,
                                 fullbright && !(mf && mf->testSubFlag(0, MFF_DIM)), false);
        }
    }
    if(mob.lumIdx != Lumobj::NoIndex && haloMode > 0) {
        try {
            MaterialAnimator &matAnimator = ClientMaterial::find(*spriteView.material)
                    .getAnimator(Rend_SpriteMaterialSpec(mob.tclass, mob.tmap));
            matAnimator.prepare();
            Vector2ui const &matDimensions = matAnimator.dimensions();
            TextureVariant *tex            = matAnimator.texUnit(MaterialAnimator::TU_LAYER0).texture;
            if(!tex || tex->base().manifest().schemeName().compareWithoutCase(QStringLiteral("Sprites"))) {
                return;
            }
            auto const *pl = (pointlight_analysis_t const *) tex->base().analysisDataPointer(res::Texture::BrightPointAnalysis);
            DENG2_ASSERT(pl);
            Lumobj const &lob = subsec.sector().map().lumobj(mob.lumIdx);
            vissprite_t *vis  = R_NewVisSprite(VSPR_FLARE);
            vis->pose.distance = distFromEye;
            vis->pose.origin = moPos + visOff;
            vis->pose.origin.z += lob.zOffset();
            dfloat flareSize = pl->brightMul;
            dfloat xOffset = matDimensions.x * pl->originX - -tex->base().origin().x;
            ded_light_t const *def = (mob.state? runtimeDefs.stateInfo[runtimeDefs.states.indexOf(mob.state)].light : 0);
            if(def) {
                if(def->size)
                    flareSize = def->size;
                if(def->haloRadius)
                    flareSize = def->haloRadius;
                if(def->offset[0])
                    xOffset = def->offset[0];
                vis->data.flare.flags = def->flags;
            }
            vis->data.flare.size = flareSize * 60 * (50 + haloSize) / 100.0f;
            if(vis->data.flare.size < 8)
                vis->data.flare.size = 8;
            V3f_Set(vis->data.flare.color, lob.color().x, lob.color().y, lob.color().z);
            vis->data.flare.factor = mob.haloFactors[DoomsdayApp::players().indexOf(viewPlayer)];
            vis->data.flare.xOff = xOffset;
            vis->data.flare.mul = 1;
            vis->data.flare.tex = 0;
            if(def && def->flare) {
                de::Uri const &flaremapResourceUri = *def->flare;
                if(flaremapResourceUri.path().toStringRef().compareWithoutCase("-")) {
                    vis->data.flare.tex = GL_PrepareFlaremap(flaremapResourceUri);
                }
            }
        }
        catch(defn::Sprite::MissingViewError const &er) {
            LOG_GL_WARNING("Projecting flare source for sprite '%i' frame '%i': %s")
                    << mob.sprite << mob.frame << er.asText();
        }
        catch(Resources::MissingResourceManifestError const &er) {
            LOG_GL_WARNING("Projecting flare source for sprite '%i' frame '%i': %s")
                    << mob.sprite << mob.frame << er.asText();
        }
    }
}
enum WallShadow {
    TopShadow,
    BottomShadow,
    LeftShadow,
    RightShadow
};
static dfloat const MIN_OPEN            = 0.1f;
static ddouble const MINDIFF            = 8;       ///< Min plane height difference (world units).
static ddouble const INDIFF             = 8;       ///< Max plane height for indifference offset.
static dfloat const MIN_SHADOW_DARKNESS = .0001f;  ///< Minimum to qualify.
static ddouble const MIN_SHADOW_SIZE    = 1;       ///< In map units.
dint rendFakeRadio              = true;  ///< cvar
static dfloat fakeRadioDarkness = 1.2f;  ///< cvar
byte devFakeRadioUpdate         = true;  ///< cvar
/**
 * Returns the "shadow darkness" (factor) for the given @a ambientLight (level), derived
 * from values in Config.
 *
 * @param ambientLight  Ambient light level to process. It is assumed that adaptation has
 *                      @em NOT yet been applied (it will be).
 */
static inline dfloat calcShadowDarkness(dfloat ambientLight) {
    ambientLight += Rend_LightAdaptationDelta(ambientLight);
    return (0.6f - ambientLight * 0.4f) * 0.65f * ::fakeRadioDarkness;
}
/**
 * Returns the "shadow size" in map units for the given @a ambientLight (level).
 *
 * @param ambientLight  Ambient light level to process. It is assumed that adaptation has
 *                      @em NOT yet been applied (it will be).
 */
static inline dfloat calcShadowSize(dfloat ambientLight) {
    return 2 * (8 + 16 - ambientLight * 16);  /// @todo Make cvars out of constants.
}
/**
 * Returns the "wall height" (i.e., distance in map units) of the wall described by
 * @a leftEdge and @a rightEdge.
 *
 * @param leftEdge    WallEdge describing the logical left-edge of the wall section.
 * @param rightEdge   WallEdge describing the logical right-edge of the wall section.
 *
 * @see wallWidth()
 * @see wallDimensions()
 */
static inline ddouble wallHeight(WallEdge const &leftEdge, WallEdge const &rightEdge) {
    return rightEdge.top().origin().z - leftEdge.bottom().origin().z;
}
/**
 * Returns the "wall width" (i.e., distance in map units) of the wall described by
 * @a leftEdge and @a rightEdge.
 *
 * @param leftEdge    WallEdge describing the logical left-edge of the wall section.
 * @param rightEdge   WallEdge describing the logical right-edge of the wall section.
 *
 * @see wallHeight()
 * @see wallDimensions()
 */
static inline ddouble wallWidth(WallEdge const &leftEdge, WallEdge const &rightEdge) {
    return de::abs((rightEdge.origin() - leftEdge.origin()).length());
}
/**
 * Returns the "wall dimensions" (i.e., distance in map units) of the wall described
 * by @a leftEdge and @a rightEdge.
 *
 * @param leftEdge    WallEdge describing the logical left-edge of the wall section.
 * @param rightEdge   WallEdge describing the logical right-edge of the wall section.
 *
 * @see wallWidth()
 * @see wallHeight()
 */
#if 0
static Vector2d wallDimensions(WallEdge const &leftEdge, WallEdge const &rightEdge) {
    return Vector2d(wallWidth(leftEdge, rightEdge), wallHeight(leftEdge, rightEdge));
}
#endif
/**
 * Returns the "wall offset" (i.e., distance in map units from the LineSide's vertex) of
 * the wall described by @a leftEdge and @a rightEdge.
 *
 * @param leftEdge    WallEdge describing the logical left-edge of the wall section.
 * @param rightEdge   WallEdge describing the logical right-edge of the wall section.
 */
static ddouble wallOffset(WallEdge const &leftEdge, WallEdge const &/*rightEdge*/)
{
    return leftEdge.lineSideOffset();
}
/**
 * Return the "wall side-openness" (factor) of specified side of the wall described by
 * @a leftEdge and @a rightEdge.
 *
 * @param leftEdge    WallEdge describing the logical left-edge of the wall section.
 * @param rightEdge   WallEdge describing the logical right-edge of the wall section.
 * @param rightSide   Use the right edge if @c true; otherwise the left edge.
 */
static dfloat wallSideOpenness(WallEdge const &leftEdge, WallEdge const &/*rightEdge*/, bool rightSide)
{
    return leftEdge.lineSide().radioCornerSide(rightSide).corner;
}
/**
 * Returns @c true if the wall described by @a leftEdge and @a rightEdge should receive
 * the specified @a shadow.
 *
 * @param leftEdge    WallEdge describing the logical left-edge of the wall section.
 * @param rightEdge   WallEdge describing the logical right-edge of the wall section.
 * @param shadow      WallShadow identifier.
 * @param shadowSize  Shadow size in map units.
 */
static bool wallReceivesShadow(WallEdge const &leftEdge, WallEdge const &rightEdge,
    WallShadow shadow, dfloat shadowSize) {
    if(shadowSize <= 0) return false;
    LineSide const &side = leftEdge.lineSide();
    DENG2_ASSERT(side.leftHEdge());
    auto const &subsec      = side.leftHEdge()->face().mapElementAs<ConvexSubspace>().subsector().as<world::ClientSubsector>();
    Plane const &visFloor   = subsec.visFloor  ();
    Plane const &visCeiling = subsec.visCeiling();
    switch(shadow) {
    case TopShadow:
        return visCeiling.castsShadow()
               && rightEdge.top   ().z() > visCeiling.heightSmoothed() - shadowSize
               && leftEdge .bottom().z() < visCeiling.heightSmoothed();
    case BottomShadow:
        return visFloor.castsShadow()
               && leftEdge .bottom().z() < visFloor  .heightSmoothed() + shadowSize
               && rightEdge.top   ().z() > visFloor  .heightSmoothed();
    case LeftShadow:
        return (visFloor.castsShadow() || visCeiling.castsShadow())
               && wallSideOpenness(leftEdge, rightEdge, false/*left side*/) > 0
               && leftEdge.lineSideOffset() < shadowSize;
    case RightShadow:
        return (visFloor.castsShadow() || visCeiling.castsShadow())
               && wallSideOpenness(leftEdge, rightEdge, true/*right side*/) > 0
               && leftEdge.lineSideOffset() + wallWidth(leftEdge, rightEdge) > side.line().length() - shadowSize;
    }
    DENG2_ASSERT(!"Unknown WallShadow");
    return false;
}
/**
 * Determine the horizontal offset for a FakeRadio wall, shadow geometry.
 *
 * @param lineLength  If negative; implies that the texture is flipped horizontally.
 * @param segOffset   Offset to the start of the segment.
 */
static inline dfloat calcTexCoordX(dfloat lineLength, dfloat segOffset) {
    return (lineLength > 0 ? segOffset : lineLength + segOffset);
}
/**
 * Determine the vertical offset for a FakeRadio wall, shadow geometry.
 *
 * @param z          Z height of the vertex.
 * @param bottom     Z height of the bottom of the wall section.
 * @param top        Z height of the top of the wall section.
 * @param texHeight  If negative; implies that the texture is flipped vertically.
 */
static inline dfloat calcTexCoordY(dfloat z, dfloat bottom, dfloat top, dfloat texHeight) {
    return (texHeight > 0 ? top - z : bottom - z);
}
struct ProjectedShadowData {
    lightingtexid_t texture;
    Vector2f texOrigin;
    Vector2f texDimensions;
    Vector2f texCoords[4];  ///< { bl, tl, br, tr }
};
static void setTopShadowParams(WallEdge const &leftEdge, WallEdge const &rightEdge, ddouble shadowSize,
    ProjectedShadowData &projected) {
    LineSide /*const*/ &side = leftEdge.lineSide();
    DENG2_ASSERT(side.leftHEdge());
    auto const &space       = side.leftHEdge()->face().mapElementAs<ConvexSubspace>();
    auto const &subsec      = space.subsector().as<world::ClientSubsector>();
    Plane const &visFloor   = subsec.visFloor  ();
    Plane const &visCeiling = subsec.visCeiling();
    projected = {};
    projected.texDimensions = Vector2f(0, shadowSize);
    projected.texOrigin     = Vector2f(0, calcTexCoordY(leftEdge.top().z(), subsec.visFloor().heightSmoothed()
                                                        , subsec.visCeiling().heightSmoothed(), shadowSize));
    projected.texture       = LST_RADIO_OO;
    edgespan_t const &edgeSpan = side.radioEdgeSpan(true/*top*/);
    if(side.radioCornerSide(0/*left*/ ).corner == -1 ||
       side.radioCornerSide(1/*right*/).corner == -1)
    {
        projected.texture         = LST_RADIO_OO;
        projected.texDimensions.x = edgeSpan.length;
        projected.texOrigin.x     = calcTexCoordX(edgeSpan.length, edgeSpan.shift + wallOffset(leftEdge, rightEdge));
        if(   (side.radioCornerSide(0/*left*/).corner == -1 && side.radioCornerSide(1/*right*/).corner == -1)
           || (side.radioCornerTop (0/*left*/).corner == -1 && side.radioCornerTop (1/*right*/).corner == -1))
        {
            projected.texture = LST_RADIO_OO;//CC;
        }
        else if(side.radioCornerSide(1/*right*/).corner == -1)
        {
            if(-side.radioCornerTop(0/*left*/).pOffset < 0 && side.radioCornerBottom(0/*left*/).pHeight < visCeiling.heightSmoothed())
            {
                projected.texture         = LST_RADIO_OE;
                projected.texDimensions.x = -edgeSpan.length;
                projected.texOrigin.x     = calcTexCoordX(-edgeSpan.length, edgeSpan.shift + wallOffset(leftEdge, rightEdge));
            }
        }
        else  // Left corner faces outwards.
        {
            if(-side.radioCornerTop(1/*right*/).pOffset < 0 && side.radioCornerBottom(1/*right*/).pHeight < visCeiling.heightSmoothed())
            {
                projected.texture = LST_RADIO_OE;
            }
        }
    }
    else {
        projected.texDimensions.x = edgeSpan.length;
        projected.texOrigin.x     = calcTexCoordX(edgeSpan.length, edgeSpan.shift + wallOffset(leftEdge, rightEdge));
        if(side.radioCornerTop(0/*left*/).corner == -1 && side.radioCornerTop(1/*right*/).corner == -1)
        {
            projected.texture = LST_RADIO_OO;//CC;
        }
        else if(side.radioCornerTop(1/*right*/).corner == -1 && side.radioCornerTop(0/*left*/).corner > MIN_OPEN)
        {
            projected.texture = LST_RADIO_OO;
        }
        else if(side.radioCornerTop(0/*left*/).corner == -1 && side.radioCornerTop(1/*right*/).corner > MIN_OPEN)
        {
            projected.texture = LST_RADIO_OO;
        }
        else if(side.radioCornerTop(0/*left*/).corner <= MIN_OPEN && side.radioCornerTop(1/*right*/).corner <= MIN_OPEN)
        {
            projected.texture = LST_RADIO_OO;
            if(side.radioCornerTop(0/*left*/).proximity && side.radioCornerTop(1/*right*/).proximity)
            {
                if(-side.radioCornerTop(0/*left*/).pOffset >= 0 && -side.radioCornerTop(1/*right*/).pOffset < 0)
                {
                    projected.texture = LST_RADIO_CO;
                    if(shadowSize > -side.radioCornerTop(0/*left*/).pOffset)
                    {
                        if(-side.radioCornerTop(0/*left*/).pOffset < INDIFF)
                        {
                            projected.texture = LST_RADIO_OE;
                        }
                        else {
                            projected.texDimensions.y = -side.radioCornerTop(0/*left*/).pOffset;
                            projected.texOrigin.y     = calcTexCoordY(leftEdge.top().z(), visFloor.heightSmoothed(), visCeiling.heightSmoothed(), projected.texDimensions.y);
                        }
                    }
                }
                else if(-side.radioCornerTop(0/*left*/).pOffset < 0 && -side.radioCornerTop(1/*right*/).pOffset >= 0)
                {
                    projected.texture         = LST_RADIO_CO;
                    projected.texDimensions.x = -edgeSpan.length;
                    projected.texOrigin.x     = calcTexCoordX(-edgeSpan.length, edgeSpan.shift + wallOffset(leftEdge, rightEdge));
                    if(shadowSize > -side.radioCornerTop(1/*right*/).pOffset)
                    {
                        if(-side.radioCornerTop(1/*right*/).pOffset < INDIFF)
                        {
                            projected.texture = LST_RADIO_OE;
                        }
                        else {
                            projected.texDimensions.y = -side.radioCornerTop(1/*right*/).pOffset;
                            projected.texOrigin.y     = calcTexCoordY(leftEdge.top().z(), visFloor.heightSmoothed(), visCeiling.heightSmoothed(), projected.texDimensions.y);
                        }
                    }
                }
            }
            else {
                if(-side.radioCornerTop(0/*left*/).pOffset < -MINDIFF)
                {
                    projected.texture         = LST_RADIO_OE;
                    projected.texDimensions.x = -edgeSpan.length;
                    projected.texOrigin.x     = calcTexCoordX(-edgeSpan.length, edgeSpan.shift + wallOffset(leftEdge, rightEdge));
                }
                else if(-side.radioCornerTop(1/*right*/).pOffset < -MINDIFF)
                {
                    projected.texture = LST_RADIO_OE;
                }
            }
        }
        else if(side.radioCornerTop(0/*left*/).corner <= MIN_OPEN)
        {
            if(-side.radioCornerTop(0/*left*/).pOffset < 0)
                projected.texture = LST_RADIO_CO;
            else
                projected.texture = LST_RADIO_OO;
            projected.texDimensions.x = -edgeSpan.length;
            projected.texOrigin.x     = calcTexCoordX(-edgeSpan.length, edgeSpan.shift + wallOffset(leftEdge, rightEdge));
        }
        else if(side.radioCornerTop(1/*right*/).corner <= MIN_OPEN)
        {
            if(-side.radioCornerTop(1/*right*/).pOffset < 0)
                projected.texture = LST_RADIO_CO;
            else
                projected.texture = LST_RADIO_OO;
        }
        else  // C/C ???
        {
            projected.texture = LST_RADIO_OO;
        }
    }
}
static void setBottomShadowParams(WallEdge const &leftEdge, WallEdge const &rightEdge, ddouble shadowSize,
    ProjectedShadowData &projected) {
    LineSide /*const*/ &side = leftEdge.lineSide();
    DENG2_ASSERT(side.leftHEdge());
    auto const &subsec      = side.leftHEdge()->face().mapElementAs<ConvexSubspace>().subsector().as<world::ClientSubsector>();
    Plane const &visFloor   = subsec.visFloor  ();
    Plane const &visCeiling = subsec.visCeiling();
    projected = {};
    projected.texDimensions.y = -shadowSize;
    projected.texOrigin.y     = calcTexCoordY(leftEdge.top().z(), visFloor.heightSmoothed(), visCeiling.heightSmoothed(), -shadowSize);
    projected.texture         = LST_RADIO_OO;
    edgespan_t const &edgeSpan = side.radioEdgeSpan(false/*bottom*/);
    if(side.radioCornerSide(0/*left*/).corner == -1 || side.radioCornerSide(1/*right*/).corner == -1)
    {
        projected.texture         = LST_RADIO_OO;
        projected.texDimensions.x = edgeSpan.length;
        projected.texOrigin.x     = calcTexCoordX(edgeSpan.length, edgeSpan.shift + wallOffset(leftEdge, rightEdge));
        if(   (side.radioCornerSide  (0/*left*/).corner == -1 && side.radioCornerSide  (1/*right*/).corner == -1)
           || (side.radioCornerBottom(0/*left*/).corner == -1 && side.radioCornerBottom(1/*right*/).corner == -1))
        {
            projected.texture = LST_RADIO_OO;//CC;
        }
        else if(side.radioCornerSide(1/*right*/).corner == -1)
        {
            if(side.radioCornerBottom(0/*left*/).pOffset < 0 && side.radioCornerTop(0/*left*/).pHeight > visFloor.heightSmoothed())
            {
                projected.texture         = LST_RADIO_OE;
                projected.texDimensions.x = -edgeSpan.length;
                projected.texOrigin.x     = calcTexCoordX(-edgeSpan.length, edgeSpan.shift + wallOffset(leftEdge, rightEdge));
            }
        }
        else  // Left corner faces outwards.
        {
            if(side.radioCornerBottom(1/*right*/).pOffset < 0 && side.radioCornerTop(1/*right*/).pHeight > visFloor.heightSmoothed())
            {
                projected.texture = LST_RADIO_OE;
            }
        }
    }
    else  // Corners WITH a neighbor back sector.
    {
        projected.texDimensions.x = edgeSpan.length;
        projected.texOrigin.x     = calcTexCoordX(edgeSpan.length, edgeSpan.shift + wallOffset(leftEdge, rightEdge));
        if(side.radioCornerBottom(0/*left*/).corner == -1 && side.radioCornerBottom(1/*right*/).corner == -1)
        {
            projected.texture = LST_RADIO_OO;//CC;
        }
        else if(side.radioCornerBottom(1/*right*/).corner == -1 && side.radioCornerBottom(0/*right*/).corner > MIN_OPEN)
        {
            projected.texture = LST_RADIO_OO;
        }
        else if(side.radioCornerBottom(0/*left*/).corner == -1 && side.radioCornerBottom(1/*right*/).corner > MIN_OPEN)
        {
            projected.texture = LST_RADIO_OO;
        }
        else if(side.radioCornerBottom(0/*left*/).corner <= MIN_OPEN && side.radioCornerBottom(1/*right*/).corner <= MIN_OPEN)
        {
            projected.texture = LST_RADIO_OO;
            if(side.radioCornerBottom(0/*left*/).proximity && side.radioCornerBottom(1/*right*/).proximity)
            {
                if(side.radioCornerBottom(0/*left*/).pOffset >= 0 && side.radioCornerBottom(1/*right*/).pOffset < 0)
                {
                    projected.texture = LST_RADIO_CO;
                    if(shadowSize > side.radioCornerBottom(0/*left*/).pOffset)
                    {
                        if(side.radioCornerBottom(0/*left*/).pOffset < INDIFF)
                        {
                            projected.texture = LST_RADIO_OE;
                        }
                        else {
                            projected.texDimensions.y = -side.radioCornerBottom(0/*left*/).pOffset;
                            projected.texOrigin.y     = calcTexCoordY(leftEdge.top().z(), visFloor.heightSmoothed(), visCeiling.heightSmoothed(), projected.texDimensions.y);
                        }
                    }
                }
                else if(side.radioCornerBottom(0/*left*/).pOffset < 0 && side.radioCornerBottom(1/*right*/).pOffset >= 0)
                {
                    projected.texture         = LST_RADIO_CO;
                    projected.texDimensions.x = -edgeSpan.length;
                    projected.texOrigin.x     = calcTexCoordX(-edgeSpan.length, edgeSpan.shift + wallOffset(leftEdge, rightEdge));
                    if(shadowSize > side.radioCornerBottom(1/*right*/).pOffset)
                    {
                        if(side.radioCornerBottom(1/*right*/).pOffset < INDIFF)
                        {
                            projected.texture = LST_RADIO_OE;
                        }
                        else {
                            projected.texDimensions.y = -side.radioCornerBottom(1/*right*/).pOffset;
                            projected.texOrigin.y     = calcTexCoordY(leftEdge.top().z(), visFloor.heightSmoothed(), visCeiling.heightSmoothed(), projected.texDimensions.y);
                        }
                    }
                }
            }
            else {
                if(side.radioCornerBottom(0/*left*/).pOffset < -MINDIFF)
                {
                    projected.texture         = LST_RADIO_OE;
                    projected.texDimensions.x = -edgeSpan.length;
                    projected.texOrigin.x     = calcTexCoordX(-edgeSpan.length, edgeSpan.shift + wallOffset(leftEdge, rightEdge));
                }
                else if(side.radioCornerBottom(1/*right*/).pOffset < -MINDIFF)
                {
                    projected.texture = LST_RADIO_OE;
                }
            }
        }
        else if(side.radioCornerBottom(0/*left*/).corner <= MIN_OPEN)
        {
            if(side.radioCornerBottom(0/*left*/).pOffset < 0)
                projected.texture = LST_RADIO_CO;
            else
                projected.texture = LST_RADIO_OO;
            projected.texDimensions.x = -edgeSpan.length;
            projected.texOrigin.x     = calcTexCoordX(-edgeSpan.length, edgeSpan.shift + wallOffset(leftEdge, rightEdge));
        }
        else if(side.radioCornerBottom(1/*right*/).corner <= MIN_OPEN)
        {
            if(side.radioCornerBottom(1/*right*/).pOffset < 0)
                projected.texture = LST_RADIO_CO;
            else
                projected.texture = LST_RADIO_OO;
        }
        else  // C/C ???
        {
            projected.texture = LST_RADIO_OO;
        }
    }
}
static void setSideShadowParams(WallEdge const &leftEdge, WallEdge const &rightEdge, bool rightSide,
    ddouble shadowSize, ProjectedShadowData &projected) {
    LineSide /*const*/ &side = leftEdge.lineSide();
    HEdge const *hedge = side.leftHEdge();
    DENG2_ASSERT(hedge);
    auto const &subsec      = hedge->face().mapElementAs<ConvexSubspace>().subsector().as<world::ClientSubsector>();
    Plane const &visFloor   = subsec.visFloor  ();
    Plane const &visCeiling = subsec.visCeiling();
    DENG2_ASSERT(visFloor.castsShadow() || visCeiling.castsShadow());  // sanity check.
    projected = {};
    projected.texOrigin     = Vector2f(0, leftEdge.bottom().z() - visFloor.heightSmoothed());
    projected.texDimensions = Vector2f(0, visCeiling.heightSmoothed() - visFloor.heightSmoothed());
    if(rightSide) {
        projected.texOrigin.x = -side.line().length() + wallOffset(leftEdge, rightEdge);
        if(shadowSize > side.line().length()) {
            projected.texDimensions.x = -side.line().length();
            if(side.radioCornerSide(0/*left*/).corner > MIN_OPEN)
                projected.texDimensions.x /= 2;
        }
        else {
            projected.texDimensions.x = -shadowSize;
        }
    }
    else {
        projected.texOrigin.x = wallOffset(leftEdge, rightEdge);
        if(shadowSize > side.line().length()) {
            projected.texDimensions.x = side.line().length();
            if(side.radioCornerSide(1/*right*/).corner > MIN_OPEN)
                projected.texDimensions.x /= 2;
        }
        else {
            projected.texDimensions.x = shadowSize;
        }
    }
    if(!hedge->twin().hasFace() || leftEdge.spec().section == LineSide::Middle) {
        if(!visFloor.castsShadow()) {
            projected.texDimensions.y = -(visCeiling.heightSmoothed() - visFloor.heightSmoothed());
            projected.texOrigin.y     = calcTexCoordY(leftEdge.top().z(), visFloor.heightSmoothed(), visCeiling.heightSmoothed(), projected.texDimensions.y);
            projected.texture         = LST_RADIO_CO;
        }
        else if(!visCeiling.castsShadow()) {
            projected.texture = LST_RADIO_CO;
        }
        else {
            projected.texture = LST_RADIO_CC;
        }
    }
    else {
        auto const &bSpace = hedge->twin().face().mapElementAs<ConvexSubspace>();
        if (bSpace.hasSubsector()) {
            auto const &bSubsec  = bSpace.subsector().as<world::ClientSubsector>();
            ddouble const bFloor = bSubsec.visFloor  ().heightSmoothed();
            ddouble const bCeil  = bSubsec.visCeiling().heightSmoothed();
            if (bFloor > visFloor.heightSmoothed() && bCeil < visCeiling.heightSmoothed()) {
                if (visFloor.castsShadow() && visCeiling.castsShadow()) {
                    projected.texture = LST_RADIO_CC;
                }
                else if (!visFloor.castsShadow()) {
                    projected.texOrigin.y     = leftEdge.bottom().z() - visCeiling.heightSmoothed();
                    projected.texDimensions.y = -(visCeiling.heightSmoothed() - visFloor.heightSmoothed());
                    projected.texture         = LST_RADIO_CO;
                }
                else {
                    projected.texture = LST_RADIO_CO;
                }
            }
            else if (bFloor > visFloor.heightSmoothed()) {
                if (visFloor.castsShadow() && visCeiling.castsShadow()) {
                    projected.texture = LST_RADIO_CC;
                }
                else if (!visFloor.castsShadow()) {
                    projected.texOrigin.y     = leftEdge.bottom().z() - visCeiling.heightSmoothed();
                    projected.texDimensions.y = -(visCeiling.heightSmoothed() - visFloor.heightSmoothed());
                    projected.texture         = LST_RADIO_CO;
                }
                else {
                    projected.texture = LST_RADIO_CO;
                }
            }
            else if (bCeil < visCeiling.heightSmoothed()) {
                if (visFloor.castsShadow() && visCeiling.castsShadow()) {
                    projected.texture = LST_RADIO_CC;
                }
                else if (!visFloor.castsShadow()) {
                    projected.texOrigin.y     = leftEdge.bottom().z() - visCeiling.heightSmoothed();
                    projected.texDimensions.y = -(visCeiling.heightSmoothed() - visFloor.heightSmoothed());
                    projected.texture         = LST_RADIO_CO;
                }
                else {
                    projected.texture = LST_RADIO_CO;
                }
            }
            }
    }
}
static void quadTexCoords(Vector2f *tc, WallEdge const &leftEdge, WallEdge const &rightEdge,
    Vector2f const &texOrigin, Vector2f const &texDimensions, bool horizontal) {
    DENG2_ASSERT(tc);
    if(horizontal) {
        tc[0] = (texOrigin / texDimensions).yx();
        tc[2] = tc[0] + Vector2f(0                              , wallWidth(leftEdge, rightEdge)) / texDimensions.yx();
        tc[3] = tc[0] + Vector2f(wallHeight(leftEdge, rightEdge), wallWidth(leftEdge, rightEdge)) / texDimensions.yx();
        tc[1] = Vector2f(tc[3].x, tc[0].y);
    }
    else  // Vertical.
    {
        tc[1] = texOrigin / texDimensions;
        tc[0] = tc[1] + Vector2f(0                             , wallHeight(leftEdge, rightEdge)) / texDimensions;
        tc[2] = tc[1] + Vector2f(wallWidth(leftEdge, rightEdge), wallHeight(leftEdge, rightEdge)) / texDimensions;
        tc[3] = Vector2f(tc[2].x, tc[1].y);
    }
}
static bool projectWallShadow(WallEdge const &leftEdge, WallEdge const &rightEdge,
    WallShadow shadow, ddouble shadowSize, ProjectedShadowData &projected) {
    if(!wallReceivesShadow(leftEdge, rightEdge, shadow, shadowSize))
        return false;
    switch(shadow) {
    case TopShadow:
        setTopShadowParams(leftEdge, rightEdge, shadowSize, projected);
        quadTexCoords(projected.texCoords, leftEdge, rightEdge,
                      projected.texOrigin, projected.texDimensions, false/*vertical*/);
        return true;
    case BottomShadow:
        setBottomShadowParams(leftEdge, rightEdge, shadowSize, projected);
        quadTexCoords(projected.texCoords, leftEdge, rightEdge,
                      projected.texOrigin, projected.texDimensions, false/*vertical*/);
        return true;
    case LeftShadow:
        setSideShadowParams(leftEdge, rightEdge, false/*left side*/, shadowSize, projected);
        quadTexCoords(projected.texCoords, leftEdge, rightEdge,
                      projected.texOrigin, projected.texDimensions, true/*horizontal*/);
        return true;
    case RightShadow:
        setSideShadowParams(leftEdge, rightEdge, true/*right side*/, shadowSize, projected);
        quadTexCoords(projected.texCoords, leftEdge, rightEdge,
                      projected.texOrigin, projected.texDimensions, true/*horizontal*/);
        return true;
    }
    DENG2_ASSERT(!"Unknown WallShadow");
    return false;
}
static void drawWallShadow(Vector3f const *posCoords, WallEdge const &leftEdge, WallEdge const &rightEdge,
    dfloat shadowDark, ProjectedShadowData const &tp) {
    DENG2_ASSERT(posCoords);
    Vector4ub const shadowColor(0, 0, 0, 255 * de::clamp(0.f, shadowDark, 1.0f));
    DrawListSpec listSpec;
    listSpec.group = ShadowGeom;
    listSpec.texunits[TU_PRIMARY] = GLTextureUnit(GL_PrepareLSTexture(tp.texture), gl::ClampToEdge, gl::ClampToEdge);
    DrawList &shadowList = ClientApp::renderSystem().drawLists().find(listSpec);
    static DrawList::Indices indices;
    if (indices.size() < 64) indices.resize(64);
    if(leftEdge.divisionCount() || rightEdge.divisionCount()) {
        Store &buffer = ClientApp::renderSystem().buffer(); {
            duint const numVerts = 3 + rightEdge.divisionCount();
            duint const base     = buffer.allocateVertices(numVerts);
            if (indices.size() < int(numVerts)) indices.resize(numVerts);
            for(duint i = 0; i < numVerts; ++i) {
                indices[i] = base + i;
            }
            buffer.posCoords   [indices[0           ]] = posCoords[0];
            buffer.colorCoords [indices[0           ]] = shadowColor;
            buffer.texCoords[0][indices[0           ]] = tp.texCoords[0];
            buffer.posCoords   [indices[1           ]] = posCoords[3];
            buffer.colorCoords [indices[1           ]] = shadowColor;
            buffer.texCoords[0][indices[1           ]] = tp.texCoords[3];
            buffer.posCoords   [indices[numVerts - 1]] = posCoords[2];
            buffer.colorCoords [indices[numVerts - 1]] = shadowColor;
            buffer.texCoords[0][indices[numVerts - 1]] = tp.texCoords[2];
            for(dint i = 0; i < rightEdge.divisionCount(); ++i) {
                WorldEdge::Event const &icpt = rightEdge.at(rightEdge.lastDivision() - i);
                buffer.posCoords   [indices[2 + i]] = icpt.origin();
                buffer.colorCoords [indices[2 + i]] = shadowColor;
                buffer.texCoords[0][indices[2 + i]] = Vector2f(tp.texCoords[3].x, tp.texCoords[2].y + (tp.texCoords[3].y - tp.texCoords[2].y) * icpt.distance());
            }
            if(::rendFakeRadio != 2) {
                shadowList.write(buffer, indices.constData(), numVerts, gl::TriangleFan);
            }
        } {
            duint const numVerts = 3 + leftEdge .divisionCount();
            duint const base     = buffer.allocateVertices(numVerts);
            if (indices.size() < int(numVerts)) indices.resize(numVerts);
            for(duint i = 0; i < numVerts; ++i) {
                indices[i] = base + i;
            }
            buffer.posCoords   [indices[0           ]] = posCoords[3];
            buffer.colorCoords [indices[0           ]] = shadowColor;
            buffer.texCoords[0][indices[0           ]] = tp.texCoords[3];
            buffer.posCoords   [indices[1           ]] = posCoords[0];
            buffer.colorCoords [indices[1           ]] = shadowColor;
            buffer.texCoords[0][indices[1           ]] = tp.texCoords[0];
            buffer.posCoords   [indices[numVerts - 1]] = posCoords[1];
            buffer.colorCoords [indices[numVerts - 1]] = shadowColor;
            buffer.texCoords[0][indices[numVerts - 1]] = tp.texCoords[1];
            for(dint i = 0; i < leftEdge.divisionCount(); ++i) {
                WorldEdge::Event const &icpt = leftEdge.at(leftEdge.firstDivision() + i);
                buffer.posCoords   [indices[2 + i]] = icpt.origin();
                buffer.colorCoords [indices[2 + i]] = shadowColor;
                buffer.texCoords[0][indices[2 + i]] = Vector2f(tp.texCoords[0].x, tp.texCoords[0].y + (tp.texCoords[1].y - tp.texCoords[0].y) * icpt.distance());
            }
            if(::rendFakeRadio != 2) {
                shadowList.write(buffer, indices.constData(), numVerts, gl::TriangleFan);
            }
        }
    }
    else {
        Store &buffer = ClientApp::renderSystem().buffer();
        duint base = buffer.allocateVertices(4);
        for(duint i = 0; i < 4; ++i) {
            indices[i] = base + i;
        }
        for(duint i = 0; i < 4; ++i) {
            buffer.posCoords   [indices[i]] = posCoords[i];
            buffer.colorCoords [indices[i]] = shadowColor;
            buffer.texCoords[0][indices[i]] = tp.texCoords[i];
        }
        if(::rendFakeRadio != 2) {
            shadowList.write(buffer, indices.constData(), 4, gl::TriangleStrip);
        }
    }
}
void Rend_DrawWallRadio(WallEdge const &leftEdge, WallEdge const &rightEdge, dfloat ambientLight) {
    if(!::rendFakeRadio || ::levelFullBright || leftEdge.spec().flags.testFlag(WallSpec::NoFakeRadio))
        return;
    dfloat const shadowDark = calcShadowDarkness(ambientLight);
    if(shadowDark < MIN_SHADOW_DARKNESS)
        return;
    dfloat const shadowSize = calcShadowSize(ambientLight);
    if(shadowSize < MIN_SHADOW_SIZE)
        return;
    leftEdge.lineSide().updateRadioForFrame(R_FrameCount());
    Vector3f const posCoords[] = {
        leftEdge .bottom().origin(),
        leftEdge .top   ().origin(),
        rightEdge.bottom().origin(),
        rightEdge.top   ().origin()
    };
    ProjectedShadowData projected;
    if(projectWallShadow(leftEdge, rightEdge, TopShadow, shadowSize, projected)) {
        drawWallShadow(posCoords, leftEdge, rightEdge, shadowDark,
                       projected);
    }
    if(projectWallShadow(leftEdge, rightEdge, BottomShadow, shadowSize, projected)) {
        drawWallShadow(posCoords, leftEdge, rightEdge, shadowDark,
                       projected);
    }
    if(projectWallShadow(leftEdge, rightEdge, LeftShadow, shadowSize, projected)) {
        drawWallShadow(posCoords, leftEdge, rightEdge,
                       shadowDark * de::cubed(wallSideOpenness(leftEdge, rightEdge, false/*left edge*/) * .8f),
                       projected);
    }
    if(projectWallShadow(leftEdge, rightEdge, RightShadow, shadowSize, projected)) {
        drawWallShadow(posCoords, leftEdge, rightEdge,
                       shadowDark * de::cubed(wallSideOpenness(leftEdge, rightEdge, true/*right edge*/) * .8f),
                       projected);
    }
}
/**
 * Determines whether FakeRadio flat, shadow geometry should be drawn between the vertices of
 * the given half-edges @a hEdges and prepares the ShadowEdges @a edges accordingly.
 *
 * @param edges             ShadowEdge descriptors for both edges { left, right }.
 * @param hEdges            Half-edge accessors for both edges { left, right }.
 * @param sectorPlaneIndex  Logical index of the sector plane to consider a shadow for.
 * @param shadowDark        Shadow darkness factor.
 *
 * @return  @c true if one or both edges are partially in shadow.
 */
static bool prepareFlatShadowEdges(ShadowEdge edges[2], HEdge const *hEdges[2], dint sectorPlaneIndex,
    dfloat shadowDark) {
    DENG2_ASSERT(edges && hEdges && hEdges[0] && hEdges[1]);
    if(!hEdges[0]->hasFace() || !hEdges[0]->face().hasMapElement())
        return false;
    if(!hEdges[0]->face().mapElementAs<ConvexSubspace>().subsector().as<world::ClientSubsector>().hasWorldVolume())
        return false;
    for(dint i = 0; i < 2; ++i) {
        edges[i].init(*hEdges[i], i);
        edges[i].prepare(sectorPlaneIndex);
    }
    return (edges[0].shadowStrength(shadowDark) >= .0001 && edges[1].shadowStrength(shadowDark) >= .0001);
}
static uint makeFlatShadowGeometry(DrawList::Indices &indices, Store &verts, gl::Primitive &primitive,
    ShadowEdge const edges[2], dfloat shadowDark, bool haveFloor) {
    static duint const floorOrder[][4] = { { 0, 1, 2, 3 }, { 1, 2, 3, 0 } };
    static duint const ceilOrder [][4] = { { 0, 3, 2, 1 }, { 1, 0, 3, 2 } };
    static Vector4ub const white(255, 255, 255, 0);
    static Vector4ub const black(  0,   0,   0, 0);
    duint const winding = (edges[1].length() > edges[0].length()? 1 : 0);
    duint const *order  = (haveFloor ? floorOrder[winding] : ceilOrder[winding]);
    duint base = verts.allocateVertices(4);
    DrawList::reserveSpace(indices, 4);
    for(duint i = 0; i < 4; ++i) {
        indices[i] = base + i;
    }
    primitive = gl::TriangleFan;
    verts.posCoords[indices[order[0]]] = edges[0].outer();
    verts.posCoords[indices[order[1]]] = edges[1].outer();
    verts.posCoords[indices[order[2]]] = edges[1].inner();
    verts.posCoords[indices[order[3]]] = edges[0].inner();
#if defined (DENG_OPENGL)
    Vector4ub const &uniformColor = (::renderWireframe? white : black);  // White to assist visual debugging.
#else
    Vector4ub const &uniformColor = black;
#endif
    for(duint i = 0; i < 4; ++i) {
        verts.colorCoords[indices[i]] = uniformColor;
    }
    for(duint i = 0; i < 2; ++i) {
        verts.colorCoords[indices[order[i]]].w = dbyte( edges[i].shadowStrength(shadowDark) * 255 );
    }
    return 4;
}
void Rend_DrawFlatRadio(ConvexSubspace const &subspace) {
    if(!::rendFakeRadio || ::levelFullBright) return;
    if(!subspace.shadowLineCount()) return;
    auto const &subsec = subspace.subsector().as<world::ClientSubsector>();
    dfloat const shadowDark = calcShadowDarkness(subsec.lightSourceIntensity());
    if(shadowDark < MIN_SHADOW_DARKNESS)
        return;
    static DrawList::Indices indices;
    static ShadowEdge shadowEdges[2/*left, right*/];  // Keep these around (needed often).
    auto const eyeToSubspace = Vector2f(Rend_EyeOrigin().xz() - subspace.poly().center());
    DrawList &shadowList = ClientApp::renderSystem().drawLists().find(
#if defined (DENG_OPENGL)
        DrawListSpec(renderWireframe? UnlitGeom : ShadowGeom)
#else
        DrawListSpec(ShadowGeom)
#endif
    );
    subspace.forAllShadowLines([&subsec, &shadowDark, &eyeToSubspace, &shadowList] (LineSide &side) {
        DENG2_ASSERT(side.hasSections() && !side.line().definesPolyobj() && side.leftHEdge());
        if(side.shadowVisCount() != R_FrameCount()) {
            side.setShadowVisCount(R_FrameCount());  // Mark processed.
            for (dint pln = 0; pln < subsec.visPlaneCount(); ++pln) {
                Plane const &plane = subsec.visPlane(pln);
                if (!plane.receivesShadow()) continue;
                if (Vector3f(eyeToSubspace, Rend_EyeOrigin().y - plane.heightSmoothed())
                         .dot(plane.surface().normal()) >= 0) {
                    HEdge const *hEdges[2/*left, right*/] = { side.leftHEdge(), side.leftHEdge() };
                    if (prepareFlatShadowEdges(shadowEdges, hEdges, pln, shadowDark)) {
                        bool const haveFloor = plane.surface().normal()[2] > 0;
                        Store &buffer = ClientApp::renderSystem().buffer();
                        gl::Primitive primitive;
                        uint vertCount = makeFlatShadowGeometry(indices, buffer, primitive, shadowEdges, shadowDark, haveFloor);
                        if (::rendFakeRadio == 2) continue;
                        shadowList.write(buffer, indices.constData(), vertCount, primitive);
                    }
                }
            }
        }
        return LoopContinue;
    });
}
void Rend_RadioRegister() {
    C_VAR_INT  ("rend-fakeradio",               &::rendFakeRadio,       0, 0, 2);
    C_VAR_FLOAT("rend-fakeradio-darkness",      &::fakeRadioDarkness,   0, 0, 2);
    C_VAR_BYTE ("rend-dev-fakeradio-update",    &::devFakeRadioUpdate,  CVF_NO_ARCHIVE, 0, 1);
}
/**
 * @ingroup drawTextFlags
 * @{
 */
#define DTF_INTERNAL_MASK               0xff00
#define DTF_NO_CHARACTER                0x8000 /// Only draw text decorations.
/**@}*/
typedef struct fr_state_attributes_s {
    int tracking;
    float leading;
    float rgba[4];
    int shadowOffsetX, shadowOffsetY;
    float shadowStrength;
    float glitterStrength;
    dd_bool caseScale;
} fr_state_attributes_t;
static fr_state_attributes_t defaultAttribs = {
    FR_DEF_ATTRIB_TRACKING,
    FR_DEF_ATTRIB_LEADING,
    { FR_DEF_ATTRIB_COLOR_RED, FR_DEF_ATTRIB_COLOR_GREEN, FR_DEF_ATTRIB_COLOR_BLUE, FR_DEF_ATTRIB_ALPHA },
    FR_DEF_ATTRIB_SHADOW_XOFFSET,
    FR_DEF_ATTRIB_SHADOW_YOFFSET,
    FR_DEF_ATTRIB_SHADOW_STRENGTH,
    FR_DEF_ATTRIB_GLITTER_STRENGTH,
    FR_DEF_ATTRIB_CASE_SCALE
};
typedef struct {
    fontid_t fontNum;
    int attribStackDepth;
    fr_state_attributes_t attribStack[FR_MAX_ATTRIB_STACK_DEPTH];
} fr_state_t;
static fr_state_t fr;
fr_state_t* frState = &fr;
typedef struct {
    fontid_t fontNum;
    float scaleX, scaleY;
    float offX, offY;
    float angle;
    float rgba[4];
    float glitterStrength, shadowStrength;
    int shadowOffsetX, shadowOffsetY;
    int tracking;
    float leading;
    int lastLineHeight;
    dd_bool typeIn;
    dd_bool caseScale;
    struct {
        float scale, offset;
    } caseMod[2]; // 1=upper, 0=lower
} drawtextstate_t;
static void drawChar(uchar ch, float posX, float posY, const AbstractFont &font, int alignFlags); //, short textFlags);
static void drawFlash(Point2Raw const *origin, Size2Raw const *size, bool bright);
static int initedFont = false;
static char smallTextBuffer[FR_SMALL_TEXT_BUFFER_SIZE+1];
static char *largeTextBuffer = NULL;
static size_t largeTextBufferSize = 0;
static int typeInTime;
static void errorIfNotInited(const char* callerName) {
    if (initedFont) return;
    App_Error("%s: font renderer module is not presently initialized.", callerName);
    exit(1);
}
static int topToAscent(const AbstractFont &font) {
    int lineHeight = font.lineSpacing();
    if (lineHeight == 0)
        return 0;
    return lineHeight - font.ascent();
}
static inline fr_state_attributes_t *currentAttribs(void) {
    return fr.attribStack + fr.attribStackDepth;
}
void FR_Shutdown(void) {
    if (!initedFont) return;
    initedFont = false;
}
dd_bool FR_Available(void) {
    return initedFont;
}
void FR_Ticker(timespan_t /*ticLength*/)
{
    if (!initedFont)
        return;
    if (!DD_IsSharpTick())
        return; // It's too soon.
    ++typeInTime;
}
#undef FR_ResetTypeinTimer
void FR_ResetTypeinTimer(void) {
    errorIfNotInited("FR_ResetTypeinTimer");
    typeInTime = 0;
}
#undef FR_SetFont
void FR_SetFont(fontid_t num) {
    errorIfNotInited("FR_SetFont");
    if (num != NOFONTID) {
        try {
            App_Resources().toFontManifest(num);
            fr.fontNum = num;
            return;
        }
        catch (ClientResources::UnknownFontIdError const &)
        {}
    }
    else {
        fr.fontNum = num;
    }
}
void FR_SetNoFont(void) {
    errorIfNotInited("FR_SetNoFont");
    fr.fontNum = 0;
}
#undef FR_Font
fontid_t FR_Font(void) {
    errorIfNotInited("FR_Font");
    return fr.fontNum;
}
#undef FR_LoadDefaultAttrib
void FR_LoadDefaultAttrib(void) {
    errorIfNotInited("FR_LoadDefaultAttrib");
    memcpy(currentAttribs(), &defaultAttribs, sizeof(defaultAttribs));
}
#undef FR_PushAttrib
void FR_PushAttrib(void) {
    errorIfNotInited("FR_PushAttrib");
    if (fr.attribStackDepth+1 == FR_MAX_ATTRIB_STACK_DEPTH) {
        App_Error("FR_PushAttrib: STACK_OVERFLOW.");
        exit(1); // Unreachable.
    }
    memcpy(fr.attribStack + fr.attribStackDepth + 1, fr.attribStack + fr.attribStackDepth, sizeof(fr.attribStack[0]));
    ++fr.attribStackDepth;
}
#undef FR_PopAttrib
void FR_PopAttrib(void) {
    errorIfNotInited("FR_PopAttrib");
    if (fr.attribStackDepth == 0) {
        App_Error("FR_PopAttrib: STACK_UNDERFLOW.");
        exit(1); // Unreachable.
    }
    --fr.attribStackDepth;
}
#undef FR_Leading
float FR_Leading(void) {
    errorIfNotInited("FR_Leading");
    return currentAttribs()->leading;
}
#undef FR_SetLeading
void FR_SetLeading(float value) {
    errorIfNotInited("FR_SetLeading");
    currentAttribs()->leading = value;
}
#undef FR_Tracking
int FR_Tracking(void) {
    errorIfNotInited("FR_Tracking");
    return currentAttribs()->tracking;
}
#undef FR_SetTracking
void FR_SetTracking(int value) {
    errorIfNotInited("FR_SetTracking");
    currentAttribs()->tracking = value;
}
#undef FR_ColorAndAlpha
void FR_ColorAndAlpha(float rgba[4]) {
    errorIfNotInited("FR_ColorAndAlpha");
    memcpy(rgba, currentAttribs()->rgba, sizeof(rgba[0]) * 4);
}
#undef FR_SetColor
void FR_SetColor(float red, float green, float blue) {
    fr_state_attributes_t* sat = currentAttribs();
    errorIfNotInited("FR_SetColor");
    sat->rgba[CR] = red;
    sat->rgba[CG] = green;
    sat->rgba[CB] = blue;
}
#undef FR_SetColorv
void FR_SetColorv(const float rgba[3]) {
    fr_state_attributes_t* sat = currentAttribs();
    errorIfNotInited("FR_SetColorv");
    sat->rgba[CR] = rgba[CR];
    sat->rgba[CG] = rgba[CG];
    sat->rgba[CB] = rgba[CB];
}
#undef FR_SetColorAndAlpha
void FR_SetColorAndAlpha(float red, float green, float blue, float alpha) {
    fr_state_attributes_t* sat = currentAttribs();
    errorIfNotInited("FR_SetColorAndAlpha");
    sat->rgba[CR] = red;
    sat->rgba[CG] = green;
    sat->rgba[CB] = blue;
    sat->rgba[CA] = alpha;
}
#undef FR_SetColorAndAlphav
void FR_SetColorAndAlphav(const float rgba[4]) {
    fr_state_attributes_t* sat = currentAttribs();
    errorIfNotInited("FR_SetColorAndAlphav");
    sat->rgba[CR] = rgba[CR];
    sat->rgba[CG] = rgba[CG];
    sat->rgba[CB] = rgba[CB];
    sat->rgba[CA] = rgba[CA];
}
#undef FR_ColorRed
float FR_ColorRed(void) {
    errorIfNotInited("FR_ColorRed");
    return currentAttribs()->rgba[CR];
}
#undef FR_SetColorRed
void FR_SetColorRed(float value) {
    errorIfNotInited("FR_SetColorRed");
    currentAttribs()->rgba[CR] = value;
}
#undef FR_ColorGreen
float FR_ColorGreen(void) {
    errorIfNotInited("FR_ColorGreen");
    return currentAttribs()->rgba[CG];
}
#undef FR_SetColorGreen
void FR_SetColorGreen(float value) {
    errorIfNotInited("FR_SetColorGreen");
    currentAttribs()->rgba[CG] = value;
}
#undef FR_ColorBlue
float FR_ColorBlue(void) {
    errorIfNotInited("FR_ColorBlue");
    return currentAttribs()->rgba[CB];
}
#undef FR_SetColorBlue
void FR_SetColorBlue(float value) {
    errorIfNotInited("FR_SetColorBlue");
    currentAttribs()->rgba[CB] = value;
}
#undef FR_Alpha
float FR_Alpha(void) {
    errorIfNotInited("FR_Alpha");
    return currentAttribs()->rgba[CA];
}
#undef FR_SetAlpha
void FR_SetAlpha(float value) {
    errorIfNotInited("FR_SetAlpha");
    currentAttribs()->rgba[CA] = value;
}
#undef FR_ShadowOffset
void FR_ShadowOffset(int* offsetX, int* offsetY) {
    fr_state_attributes_t* sat = currentAttribs();
    errorIfNotInited("FR_ShadowOffset");
    if (NULL != offsetX) *offsetX = sat->shadowOffsetX;
    if (NULL != offsetY) *offsetY = sat->shadowOffsetY;
}
#undef FR_SetShadowOffset
void FR_SetShadowOffset(int offsetX, int offsetY) {
    fr_state_attributes_t* sat = currentAttribs();
    errorIfNotInited("FR_SetShadowOffset");
    sat->shadowOffsetX = offsetX;
    sat->shadowOffsetY = offsetY;
}
#undef FR_ShadowStrength
float FR_ShadowStrength(void) {
    errorIfNotInited("FR_ShadowStrength");
    return currentAttribs()->shadowStrength;
}
#undef FR_SetShadowStrength
void FR_SetShadowStrength(float value) {
    errorIfNotInited("FR_SetShadowStrength");
    currentAttribs()->shadowStrength = MINMAX_OF(0, value, 1);
}
#undef FR_GlitterStrength
float FR_GlitterStrength(void) {
    errorIfNotInited("FR_GlitterStrength");
    return currentAttribs()->glitterStrength;
}
#undef FR_SetGlitterStrength
void FR_SetGlitterStrength(float value) {
    errorIfNotInited("FR_SetGlitterStrength");
    currentAttribs()->glitterStrength = MINMAX_OF(0, value, 1);
}
#undef FR_CaseScale
dd_bool FR_CaseScale(void) {
    errorIfNotInited("FR_CaseScale");
    return currentAttribs()->caseScale;
}
#undef FR_SetCaseScale
void FR_SetCaseScale(dd_bool value) {
    errorIfNotInited("FR_SetCaseScale");
    currentAttribs()->caseScale = value;
}
#undef FR_CharSize
void FR_CharSize(Size2Raw *size, uchar ch) {
    errorIfNotInited("FR_CharSize");
    if (size) {
        Vector2ui dimensions = App_Resources().font(fr.fontNum).glyphPosCoords(ch).size();
        size->width  = dimensions.x;
        size->height = dimensions.y;
    }
}
#undef FR_CharWidth
int FR_CharWidth(uchar ch) {
    errorIfNotInited("FR_CharWidth");
    if (fr.fontNum != 0)
        return App_Resources().font(fr.fontNum).glyphPosCoords(ch).width();
    return 0;
}
#undef FR_CharHeight
int FR_CharHeight(uchar ch) {
    errorIfNotInited("FR_CharHeight");
    if (fr.fontNum != 0)
        return App_Resources().font(fr.fontNum).glyphPosCoords(ch).height();
    return 0;
}
int FR_SingleLineHeight(char const *text) {
    errorIfNotInited("FR_SingleLineHeight");
    if (fr.fontNum == 0 || !text)
        return 0;
    AbstractFont &font = App_Resources().font(fr.fontNum);
    int ascent = font.ascent();
    if (ascent != 0)
        return ascent;
    return font.glyphPosCoords((uchar)text[0]).height();
}
int FR_GlyphTopToAscent(char const *text) {
    errorIfNotInited("FR_GlyphTopToAscent");
    if (fr.fontNum == 0 || !text)
        return 0;
    AbstractFont &font = App_Resources().font(fr.fontNum);
    int lineHeight = font.lineSpacing();
    if (lineHeight == 0)
        return 0;
    return lineHeight - font.ascent();
}
enum class TextFragmentPass {
    Shadow,
    Character,
    Glitter,
};
struct TextFragment {
    TextFragmentPass pass;
    const AbstractFont &font;
    const char *text;
    int x;
    int y;
    int alignFlags;
    uint16_t textFlags;
    int initialCount;
    int length;
    int width;
    int height;
    void updateDimensions() {
        length = int(strlen(text)); {
            width = 0;
            int i = 0;
            char const *ch = text;
            uchar c;
            while (i++ < length && (c = *ch++) != 0 && c != '\n') {
                width += FR_CharWidth(c);
            }
            if (length > 0) {
                width += currentAttribs()->tracking * (length - 1);
            }
        } {
            height = 0;
            int i = 0;
            char const *ch = text;
            uchar c;
            while (i++ < length && (c = *ch++) != 0 && c != '\n') {
                height = de::max(height, FR_CharHeight(c));
            }
            height += topToAscent(font);
        }
    }
    void applyBlendMode() const {
        switch (pass) {
        case TextFragmentPass::Shadow:
            DGL_BlendFunc(DGL_ZERO, DGL_ONE_MINUS_SRC_ALPHA);
            break;
        case TextFragmentPass::Character:
            DGL_BlendFunc(DGL_SRC_ALPHA, DGL_ONE_MINUS_SRC_ALPHA);
            break;
        case TextFragmentPass::Glitter:
            DGL_BlendFunc(DGL_SRC_ALPHA, DGL_ONE);
            break;
        }
    }
};
static void drawTextFragment(const TextFragment &fragment) {
    fr_state_attributes_t* sat = currentAttribs();
    bool noTypein = (fragment.textFlags & DTF_NO_TYPEIN) != 0;
    float glitterMul;
    float shadowMul;
    float flashColor[3] = { 0, 0, 0 };
    int x = fragment.x;
    int y = fragment.y;
    int w, h, cx, cy, count, yoff;
    unsigned char c;
    const char* ch; {
        if (fragment.pass == TextFragmentPass::Shadow &&
            (sat->shadowStrength <= 0 || fragment.font.flags().testFlag(AbstractFont::Shadowed))) {
            return;
        }
        if (fragment.pass == TextFragmentPass::Glitter && sat->glitterStrength <= 0) {
            return;
        }
    }
    if (fragment.pass == TextFragmentPass::Glitter) {
        flashColor[CR] = (1 + 2 * sat->rgba[CR]) / 3;
        flashColor[CG] = (1 + 2 * sat->rgba[CG]) / 3;
        flashColor[CB] = (1 + 2 * sat->rgba[CB]) / 3;
    }
    if (fragment.alignFlags & ALIGN_RIGHT) {
        x -= fragment.width;
    }
    else if (!(fragment.alignFlags & ALIGN_LEFT)) {
        x -= fragment.width / 2;
    }
    if (fragment.alignFlags & ALIGN_BOTTOM) {
        y -= fragment.height;
    }
    else if (!(fragment.alignFlags & ALIGN_TOP)) {
        y -= fragment.height / 2;
    }
    if (const BitmapFont *bmapFont = maybeAs<BitmapFont>(fragment.font)) {
        if (bmapFont->textureGLName()) {
            GL_BindTextureUnmanaged(bmapFont->textureGLName(), gl::ClampToEdge,
                                    gl::ClampToEdge, filterUI ? gl::Linear : gl::Nearest);
            DGL_MatrixMode(DGL_TEXTURE);
            DGL_PushMatrix();
            DGL_LoadIdentity();
            DGL_Scalef(1.f / bmapFont->textureDimensions().x,
                       1.f / bmapFont->textureDimensions().y, 1.f);
        }
    }
    bool isBlendModeSet = false; {
        count = fragment.initialCount;
        ch = fragment.text;
        cx = x + (fragment.pass == TextFragmentPass::Shadow? sat->shadowOffsetX : 0);
        cy = y + (fragment.pass == TextFragmentPass::Shadow? sat->shadowOffsetY : 0);
        for (;;) {
            c = *ch++;
            yoff = 0;
            glitterMul = 0;
            shadowMul = sat->rgba[CA];
            if (!noTypein) // && (pass || (!noShadow && !pass)))
            {
                int maxCount = (typeInTime > 0? typeInTime * 2 : 0);
                if (fragment.pass == TextFragmentPass::Shadow) {
                    if (count == maxCount) {
                        shadowMul = 0;
                    }
                    else if (count + 1 == maxCount) {
                        shadowMul *= .25f;
                    }
                    else if (count + 2 == maxCount) {
                        shadowMul *= .5f;
                    }
                    else if (count + 3 == maxCount) {
                        shadowMul *= .75f;
                    }
                    else if (count > maxCount) {
                        break;
                    }
                }
                else {
                    if (fragment.pass == TextFragmentPass::Glitter) {
                        if (count == maxCount) {
                            glitterMul = 1;
                            flashColor[CR] = sat->rgba[CR];
                            flashColor[CG] = sat->rgba[CG];
                            flashColor[CB] = sat->rgba[CB];
                        }
                        else if (count + 1 == maxCount) {
                            glitterMul = 0.88f;
                            flashColor[CR] = (1 + sat->rgba[CR]) / 2;
                            flashColor[CG] = (1 + sat->rgba[CG]) / 2;
                            flashColor[CB] = (1 + sat->rgba[CB]) / 2;
                        }
                        else if (count + 2 == maxCount) {
                            glitterMul = 0.75f;
                            flashColor[CR] = sat->rgba[CR];
                            flashColor[CG] = sat->rgba[CG];
                            flashColor[CB] = sat->rgba[CB];
                        }
                        else if (count + 3 == maxCount) {
                            glitterMul = 0.5f;
                            flashColor[CR] = sat->rgba[CR];
                            flashColor[CG] = sat->rgba[CG];
                            flashColor[CB] = sat->rgba[CB];
                        }
                        else if (count > maxCount) {
                            break;
                        }
                    }
                    else if (count > maxCount) {
                        break;
                    }
                }
            }
            count++;
            if (!c || c == '\n')
                break;
            w = FR_CharWidth(c);
            h = FR_CharHeight(c);
            if (c != ' ') {
                switch (fragment.pass) {
                    case TextFragmentPass::Character:
                        if (sat->rgba[CA] > 0.001f) {
                            DGL_Color4fv(sat->rgba);
                            if (!isBlendModeSet) {
                                fragment.applyBlendMode();
                                isBlendModeSet = true;
                            }
                            drawChar(c, cx, cy + yoff, fragment.font, ALIGN_TOPLEFT);
                        }
                        break;
                    case TextFragmentPass::Glitter:
                        if (/*!noGlitter &&*/ sat->glitterStrength > 0)
                        {
                            Point2Raw origin;
                            Size2Raw  size;
                            origin.x    = cx;
                            origin.y    = cy + yoff;
                            size.width  = w;
                            size.height = h;
                            const float alpha = sat->glitterStrength * glitterMul;
                            if (alpha > 0) {
                                if (!isBlendModeSet) {
                                    fragment.applyBlendMode();
                                    isBlendModeSet = true;
                                }
                                DGL_Color4f(flashColor[CR], flashColor[CG], flashColor[CB], alpha);
                                drawFlash(&origin, &size, true);
                            }
                        }
                        break;
                    case TextFragmentPass::Shadow: {
                        Point2Raw origin;
                        Size2Raw  size;
                        origin.x    = cx;
                        origin.y    = cy + yoff;
                        size.width  = w;
                        size.height = h;
                        const float alpha = sat->shadowStrength * shadowMul;
                        if (alpha > 0) {
                            if (!isBlendModeSet) {
                                fragment.applyBlendMode();
                                isBlendModeSet = true;
                            }
                            DGL_Color4f(1, 1, 1, alpha);
                            drawFlash(&origin, &size, false);
                        }
                        break;
                    }
                }
            }
            cx += w + sat->tracking;
        }
    }
    if (const BitmapFont *bmapFont = maybeAs<BitmapFont>(fragment.font)) {
        if (bmapFont->textureGLName()) {
            DGL_MatrixMode(DGL_TEXTURE);
            DGL_PopMatrix();
        }
    }
}
static void drawChar(uchar ch, float x, float y, const AbstractFont &font, int alignFlags) {
    if (alignFlags & ALIGN_RIGHT) {
        x -= font.glyphPosCoords(ch).width();
    }
    else if (!(alignFlags & ALIGN_LEFT)) {
        x -= font.glyphPosCoords(ch).width() / 2;
    }
    int const ascent = font.ascent();
    int const lineHeight = ascent ? ascent : font.glyphPosCoords(ch).height();
    if (alignFlags & ALIGN_BOTTOM) {
        y -= topToAscent(font) + lineHeight;
    }
    else if (!(alignFlags & ALIGN_TOP)) {
        y -= (topToAscent(font) + lineHeight) / 2;
    }
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_Translatef(x, y, 0);
    Rectanglei geometry = font.glyphPosCoords(ch);
    if (const BitmapFont *bmapFont = maybeAs<BitmapFont>(font)) {
        GL_BindTextureUnmanaged(bmapFont->textureGLName(), gl::ClampToEdge,
                                gl::ClampToEdge, filterUI? gl::Linear : gl::Nearest);
        geometry = geometry.expanded(bmapFont->textureMargin().toVector2i());
    }
    else if (const CompositeBitmapFont *compFont = maybeAs<CompositeBitmapFont>(font)) {
        GL_BindTexture(compFont->glyphTexture(ch));
        geometry = geometry.expanded(compFont->glyphTextureBorder(ch));
    }
    Vector2i coords[4] = { font.glyphTexCoords(ch).topLeft,
                           font.glyphTexCoords(ch).topRight(),
                           font.glyphTexCoords(ch).bottomRight,
                           font.glyphTexCoords(ch).bottomLeft() };
    GL_DrawRectWithCoords(geometry, coords);
    if (is<CompositeBitmapFont>(font)) {
        GL_SetNoTexture();
    }
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_Translatef(-x, -y, 0);
}
static void drawFlash(Point2Raw const *origin, Size2Raw const *size, bool bright) {
    float fsize = 4.f + bright;
    float fw = fsize * size->width  / 2.0f;
    float fh = fsize * size->height / 2.0f;
    int x, y, w, h;
    if (size->height <= 4) return;
    x = origin->x + (int) (size->width  / 2.0f - fw / 2);
    y = origin->y + (int) (size->height / 2.0f - fh / 2);
    w = (int) fw;
    h = (int) fh;
    GL_BindTextureUnmanaged(GL_PrepareLSTexture(LST_DYNAMIC),
                            gl::ClampToEdge, gl::ClampToEdge);
    DGL_Begin(DGL_QUADS);
        DGL_TexCoord2f(0, 0, 0);
        DGL_Vertex2f(x, y);
        DGL_TexCoord2f(0, 1, 0);
        DGL_Vertex2f(x + w, y);
        DGL_TexCoord2f(0, 1, 1);
        DGL_Vertex2f(x + w, y + h);
        DGL_TexCoord2f(0, 0, 1);
        DGL_Vertex2f(x, y + h);
    DGL_End();
}
/**
 * Expected:
 * <pre>
 *     "<whitespace> = <whitespace> <float>"
 * </pre>
 */
static float parseFloat(const char **str) {
    float value;
    char *end;
    *str = M_SkipWhite(*str);
    if (**str != '=') return 0; // Now I'm confused!
    *str = M_SkipWhite(*str + 1);
    value = (float) strtod(*str, &end);
    *str = end;
    return value;
}
/**
 * Expected:
 * <pre>
 *      "<whitespace> = <whitespace> [|"]<string>[|"]"
 * </pre>
 */
 static dd_bool parseString(const char **str, char* buf, size_t bufLen) {
    size_t len;
    const char* end;
    if (!buf || bufLen == 0) return false;
    *str = M_SkipWhite(*str);
    if (**str != '=') return false; // Now I'm confused!
    *str = M_SkipWhite(*str + 1);
    if (**str == '"') (*str)++;
    end = *str;
    while (*end && *end != '}' && *end != ',' && *end !='"') { end++; }
    len = end - *str;
    if (len != 0) {
        dd_snprintf(buf, MIN_OF(len+1, bufLen), "%s", *str);
        *str = end;
    }
    if (**str == '"')
        (*str)++;
    return true;
}
static void parseParamaterBlock(const char **strPtr, drawtextstate_t* state, int *numBreaks) {
    LOG_AS("parseParamaterBlock");
    (*strPtr)++;
    while (*(*strPtr) && *(*strPtr) != '}') {
        (*strPtr) = M_SkipWhite((*strPtr));
        if (!strnicmp((*strPtr), "flash", 5)) {
            (*strPtr) += 5;
            state->typeIn = true;
        }
        else if (!strnicmp((*strPtr), "noflash", 7)) {
            (*strPtr) += 7;
            state->typeIn = false;
        }
        else if (!strnicmp((*strPtr), "case", 4)) {
            (*strPtr) += 4;
            state->caseScale = true;
        }
        else if (!strnicmp((*strPtr), "nocase", 6)) {
            (*strPtr) += 6;
            state->caseScale = false;
        }
        else if (!strnicmp((*strPtr), "ups", 3)) {
            (*strPtr) += 3;
            state->caseMod[1].scale = parseFloat(&(*strPtr));
        }
        else if (!strnicmp((*strPtr), "upo", 3)) {
            (*strPtr) += 3;
            state->caseMod[1].offset = parseFloat(&(*strPtr));
        }
        else if (!strnicmp((*strPtr), "los", 3)) {
            (*strPtr) += 3;
            state->caseMod[0].scale = parseFloat(&(*strPtr));
        }
        else if (!strnicmp((*strPtr), "loo", 3)) {
            (*strPtr) += 3;
            state->caseMod[0].offset = parseFloat(&(*strPtr));
        }
        else if (!strnicmp((*strPtr), "break", 5)) {
            (*strPtr) += 5;
            ++(*numBreaks);
        }
        else if (!strnicmp((*strPtr), "r", 1)) {
            (*strPtr)++;
            state->rgba[CR] = parseFloat(&(*strPtr));
        }
        else if (!strnicmp((*strPtr), "g", 1)) {
            (*strPtr)++;
            state->rgba[CG] = parseFloat(&(*strPtr));
        }
        else if (!strnicmp((*strPtr), "b", 1)) {
            (*strPtr)++;
            state->rgba[CB] = parseFloat(&(*strPtr));
        }
        else if (!strnicmp((*strPtr), "a", 1)) {
            (*strPtr)++;
            state->rgba[CA] = parseFloat(&(*strPtr));
        }
        else if (!strnicmp((*strPtr), "x", 1)) {
            (*strPtr)++;
            state->offX = parseFloat(&(*strPtr));
        }
        else if (!strnicmp((*strPtr), "y", 1)) {
            (*strPtr)++;
            state->offY = parseFloat(&(*strPtr));
        }
        else if (!strnicmp((*strPtr), "scalex", 6)) {
            (*strPtr) += 6;
            state->scaleX = parseFloat(&(*strPtr));
        }
        else if (!strnicmp((*strPtr), "scaley", 6)) {
            (*strPtr) += 6;
            state->scaleY = parseFloat(&(*strPtr));
        }
        else if (!strnicmp((*strPtr), "scale", 5)) {
            (*strPtr) += 5;
            state->scaleX = state->scaleY = parseFloat(&(*strPtr));
        }
        else if (!strnicmp((*strPtr), "angle", 5)) {
            (*strPtr) += 5;
            state->angle = parseFloat(&(*strPtr));
        }
        else if (!strnicmp((*strPtr), "glitter", 7)) {
            (*strPtr) += 7;
            state->glitterStrength = parseFloat(&(*strPtr));
        }
        else if (!strnicmp((*strPtr), "shadow", 6)) {
            (*strPtr) += 6;
            state->shadowStrength = parseFloat(&(*strPtr));
        }
        else if (!strnicmp((*strPtr), "tracking", 8)) {
            (*strPtr) += 8;
            state->tracking = parseFloat(&(*strPtr));
        }
        else {
            if (!strnicmp((*strPtr), "font", 4)) {
                char buf[80];
                (*strPtr) += 4;
                if (parseString(&(*strPtr), buf, 80)) {
                    try {
                        state->fontNum = App_Resources().fontManifest(de::makeUri(buf)).uniqueId();
                        continue;
                    }
                    catch (Resources::MissingResourceManifestError const &)
                    {}
                }
                LOG_GL_WARNING("Unknown font '%s'") << *strPtr;
                continue;
            }
            if (*(*strPtr) != '}') {
                (*strPtr)++;
            }
        }
    }
    if (*(*strPtr))
        (*strPtr)++;
}
static void initDrawTextState(drawtextstate_t* state, uint16_t textFlags) {
    fr_state_attributes_t* sat = currentAttribs();
    state->typeIn = (textFlags & DTF_NO_TYPEIN) == 0;
    state->fontNum = fr.fontNum;
    memcpy(state->rgba, sat->rgba, sizeof(state->rgba));
    state->tracking = sat->tracking;
    state->glitterStrength = sat->glitterStrength;
    state->shadowStrength = sat->shadowStrength;
    state->shadowOffsetX = sat->shadowOffsetX;
    state->shadowOffsetY = sat->shadowOffsetY;
    state->leading = sat->leading;
    state->caseScale = sat->caseScale;
    state->scaleX = state->scaleY = 1;
    state->offX = state->offY = 0;
    state->angle = 0;
    state->typeIn = true;
    state->caseMod[0].scale = 1;
    state->caseMod[0].offset = 3;
    state->caseMod[1].scale = 1.25f;
    state->caseMod[1].offset = 0;
    state->lastLineHeight = FR_CharHeight('A') * state->scaleY * (1+state->leading);
    FR_PushAttrib();
}
static char* enlargeTextBuffer(size_t lengthMinusTerminator) {
    if (lengthMinusTerminator <= FR_SMALL_TEXT_BUFFER_SIZE) {
        return smallTextBuffer;
    }
    if (largeTextBuffer == NULL || lengthMinusTerminator > largeTextBufferSize) {
        largeTextBufferSize = lengthMinusTerminator;
        largeTextBuffer = (char*)realloc(largeTextBuffer, largeTextBufferSize+1);
        if (largeTextBuffer == NULL)
            App_Error("FR_EnlargeTextBuffer: Failed on reallocation of %lu bytes.",
                (unsigned long)(lengthMinusTerminator+1));
    }
    return largeTextBuffer;
}
static void freeTextBuffer(void) {
    if (largeTextBuffer == NULL)
        return;
    free(largeTextBuffer); largeTextBuffer = 0;
    largeTextBufferSize = 0;
}
#undef FR_TextWidth
int FR_TextWidth(const char* string) {
    int w, maxWidth = -1;
    dd_bool skipping = false, escaped = false;
    const char* ch;
    size_t i, len;
    errorIfNotInited("FR_TextWidth");
    if (!string || !string[0])
        return 0;
    w = 0;
    len = strlen(string);
    ch = string;
    for (i = 0; i < len; ++i, ch++) {
        unsigned char c = *ch;
        if (c == FR_FORMAT_ESCAPE_CHAR) {
            escaped = true;
            continue;
        }
        if (!escaped && c == '{') {
            skipping = true;
        }
        else if (skipping && c == '}') {
            skipping = false;
            continue;
        }
        if (skipping)
            continue;
        escaped = false;
        if (c == '\n') {
            if (w > maxWidth)
                maxWidth = w;
            w = 0;
            continue;
        }
        w += FR_CharWidth(c);
        if (i != len - 1) {
            w += FR_Tracking();
        }
        else if (maxWidth == -1) {
            maxWidth = w;
        }
    }
    return maxWidth;
}
#undef FR_TextHeight
int FR_TextHeight(const char* string) {
    int h, currentLineHeight;
    dd_bool skip = false;
    const char* ch;
    size_t i, len;
    if (!string || !string[0])
        return 0;
    errorIfNotInited("FR_TextHeight");
    currentLineHeight = 0;
    len = strlen(string);
    h = 0;
    ch = string;
    for (i = 0; i < len; ++i, ch++) {
        unsigned char c = *ch;
        int charHeight;
        if (c == '{') {
            skip = true;
        }
        else if (c == '}') {
            skip = false;
            continue;
        }
        if (skip)
            continue;
        if (c == '\n') {
            h += currentLineHeight == 0? (FR_CharHeight('A') * (1+FR_Leading())) : currentLineHeight;
            currentLineHeight = 0;
            continue;
        }
        charHeight = FR_CharHeight(c) * (1+FR_Leading());
        if (charHeight > currentLineHeight)
            currentLineHeight = charHeight;
    }
    h += currentLineHeight;
    return h;
}
#undef FR_TextSize
void FR_TextSize(Size2Raw* size, const char* text) {
    if (!size) return;
    size->width  = FR_TextWidth(text);
    size->height = FR_TextHeight(text);
}
#undef FR_DrawText3
void FR_DrawText3(const char* text, const Point2Raw* _origin, int alignFlags, uint16_t textFlags) {
    fontid_t origFont = FR_Font();
    float cx, cy, extraScale;
    drawtextstate_t state;
    const char* fragment;
    int curCase;
    Point2Raw origin;
    Size2Raw textSize;
    int charCount;
    float origColor[4];
    const char *str, *end;
    dd_bool escaped = false;
    errorIfNotInited("FR_DrawText");
    if (!text || !text[0]) return;
    origin.x = _origin? _origin->x : 0;
    origin.y = _origin? _origin->y : 0;
    textFlags &= ~(DTF_INTERNAL_MASK);
    if (alignFlags & ALIGN_RIGHT) {
        FR_TextSize(&textSize, text);
    }
    DENG2_ASSERT_IN_RENDER_THREAD();
    DENG_ASSERT_GL_CONTEXT_ACTIVE();
    DGL_CurrentColor(origColor);
    for (TextFragmentPass pass :
         {TextFragmentPass::Shadow, TextFragmentPass::Character, TextFragmentPass::Glitter}) {
        switch (pass) {
            case TextFragmentPass::Shadow:
                if (textFlags & DTF_NO_SHADOW) continue;
                break;
            case TextFragmentPass::Character:
                if (textFlags & DTF_NO_CHARACTER) continue;
                break;
            case TextFragmentPass::Glitter:
                if (textFlags & DTF_NO_GLITTER) continue;
                break;
        }
        cx = (float) origin.x;
        cy = (float) origin.y;
        curCase = -1;
        charCount = 0;
        initDrawTextState(&state, textFlags);
        str = text;
        while (*str) {
            if (*str == FR_FORMAT_ESCAPE_CHAR) {
                escaped = true;
                ++str;
                continue;
            }
            if (!escaped && *str == '{') // Paramaters included?
            {
                fontid_t lastFont = state.fontNum;
                int lastTracking = state.tracking;
                float lastLeading = state.leading;
                float lastShadowStrength = state.shadowStrength;
                float lastGlitterStrength = state.glitterStrength;
                dd_bool lastCaseScale = state.caseScale;
                float lastRGBA[4];
                int numBreaks = 0;
                lastRGBA[CR] = state.rgba[CR];
                lastRGBA[CG] = state.rgba[CG];
                lastRGBA[CB] = state.rgba[CB];
                lastRGBA[CA] = state.rgba[CA];
                parseParamaterBlock(&str, &state, &numBreaks);
                if (numBreaks != 0) {
                    do {
                        cx = (float) origin.x;
                        cy += state.lastLineHeight * (1+lastLeading);
                    } while (--numBreaks > 0);
                }
                if (state.fontNum != lastFont)
                    FR_SetFont(state.fontNum);
                if (state.tracking != lastTracking)
                    FR_SetTracking(state.tracking);
                if (state.leading != lastLeading)
                    FR_SetLeading(state.leading);
                if (state.rgba[CR] != lastRGBA[CR] || state.rgba[CG] != lastRGBA[CG] || state.rgba[CB] != lastRGBA[CB] || state.rgba[CA] != lastRGBA[CA])
                    FR_SetColorAndAlphav(state.rgba);
                if (state.shadowStrength != lastShadowStrength)
                    FR_SetShadowStrength(state.shadowStrength);
                if (state.glitterStrength != lastGlitterStrength)
                    FR_SetGlitterStrength(state.glitterStrength);
                if (state.caseScale != lastCaseScale)
                    FR_SetCaseScale(state.caseScale);
            }
            for (end = str; *end && *end != FR_FORMAT_ESCAPE_CHAR && (escaped || *end != '{');) {
                int newlines = 0, fragmentAlignFlags;
                float alignx = 0;
                if (FR_CaseScale()) {
                    curCase = -1;
                    for (; *end && *end != FR_FORMAT_ESCAPE_CHAR && (escaped || *end != '{') &&
                        *end != '\n'; end++) {
                        escaped = false;
                        if (isspace(*end))
                            continue;
                        if (curCase < 0)
                            curCase = (isupper(*end) != 0);
                        else if (curCase != (isupper(*end) != 0))
                            break;
                    }
                }
                else {
                    curCase = 0;
                    for (; *end && *end != FR_FORMAT_ESCAPE_CHAR && (escaped || *end != '{') &&
                        *end != '\n'; end++) { escaped = false; }
                }
                escaped = false;
                { char* buffer = enlargeTextBuffer(end - str);
                memcpy(buffer, str, end - str);
                buffer[end - str] = '\0';
                fragment = buffer;
                }
                while (*end == '\n') {
                    newlines++;
                    end++;
                }
                str = end;
                if (!(alignFlags & (ALIGN_LEFT|ALIGN_RIGHT))) {
                    fragmentAlignFlags = alignFlags;
                }
                else {
                    fragmentAlignFlags = (alignFlags & ~(ALIGN_RIGHT)) | ALIGN_LEFT;
                    if (alignFlags & ALIGN_RIGHT)
                        alignx = -textSize.width * state.scaleX;
                }
                DGL_MatrixMode(DGL_MODELVIEW);
                DGL_PushMatrix();
                if (!fequal(state.angle, 0)) {
                    DGL_Translatef((float)origin.x, (float)origin.y, 0);
                    DGL_Scalef(1, 200.0f / 240.0f, 1);
                    DGL_Rotatef(state.angle, 0, 0, 1);
                    DGL_Scalef(1, 240.0f / 200.0f, 1);
                    DGL_Translatef(-(float)origin.x, -(float)origin.y, 0);
                }
                DGL_Translatef(cx + state.offX + alignx, cy + state.offY + (FR_CaseScale() ? state.caseMod[curCase].offset : 0), 0);
                extraScale = (FR_CaseScale() ? state.caseMod[curCase].scale : 1);
                DGL_Scalef(state.scaleX, state.scaleY * extraScale, 1);
                DENG2_ASSERT(fr.fontNum);
                TextFragment frag{pass,
                                  App_Resources().font(fr.fontNum),
                                  fragment,
                                  0,
                                  0,
                                  fragmentAlignFlags,
                                  textFlags,
                                  state.typeIn ? charCount : DEFAULT_INITIALCOUNT,
                                  0,
                                  0,
                                  0};
                frag.updateDimensions();
                drawTextFragment(frag);
                charCount += frag.length;
                if (newlines == 0) {
                    cx += (float(frag.width) + currentAttribs()->tracking) * state.scaleX;
                }
                else {
                    if (frag.length > 0) {
                        state.lastLineHeight = frag.height;
                    }
                    cx = (float) origin.x;
                    cy += newlines * state.lastLineHeight * (1.f + FR_Leading());
                }
                DGL_MatrixMode(DGL_MODELVIEW);
                DGL_PopMatrix();
            }
        }
        FR_PopAttrib();
    }
    freeTextBuffer();
    FR_SetFont(origFont);
    DGL_Color4fv(origColor);
    DGL_BlendFunc(DGL_SRC_ALPHA, DGL_ONE_MINUS_SRC_ALPHA);
}
#undef FR_DrawText2
void FR_DrawText2(const char* text, const Point2Raw* origin, int alignFlags) {
    FR_DrawText3(text, origin, alignFlags, DEFAULT_DRAWFLAGS);
}
#undef FR_DrawText
void FR_DrawText(const char* text, const Point2Raw* origin) {
    FR_DrawText2(text, origin, DEFAULT_ALIGNFLAGS);
}
#undef FR_DrawTextXY3
void FR_DrawTextXY3(const char* text, int x, int y, int alignFlags, short flags) {
    Point2Raw origin;
    origin.x = x;
    origin.y = y;
    FR_DrawText3(text, &origin, alignFlags, flags);
}
#undef FR_DrawTextXY2
void FR_DrawTextXY2(const char* text, int x, int y, int alignFlags) {
    FR_DrawTextXY3(text, x, y, alignFlags, DEFAULT_DRAWFLAGS);
}
#undef FR_DrawTextXY
void FR_DrawTextXY(const char* text, int x, int y) {
    FR_DrawTextXY2(text, x, y, DEFAULT_ALIGNFLAGS);
}
#undef FR_DrawChar3
void FR_DrawChar3(unsigned char ch, const Point2Raw* origin, int alignFlags, short textFlags) {
    char str[2];
    str[0] = ch;
    str[1] = '\0';
    FR_DrawText3(str, origin, alignFlags, textFlags);
}
#undef FR_DrawChar2
void FR_DrawChar2(unsigned char ch, const Point2Raw* origin, int alignFlags) {
    FR_DrawChar3(ch, origin, alignFlags, DEFAULT_DRAWFLAGS);
}
#undef FR_DrawChar
void FR_DrawChar(unsigned char ch, const Point2Raw* origin) {
    FR_DrawChar2(ch, origin, DEFAULT_ALIGNFLAGS);
}
#undef FR_DrawCharXY3
void FR_DrawCharXY3(unsigned char ch, int x, int y, int alignFlags, short textFlags) {
    Point2Raw origin;
    origin.x = x;
    origin.y = y;
    FR_DrawChar3(ch, &origin, alignFlags, textFlags);
}
#undef FR_DrawCharXY2
void FR_DrawCharXY2(unsigned char ch, int x, int y, int alignFlags) {
    FR_DrawCharXY3(ch, x, y, alignFlags, DEFAULT_DRAWFLAGS);
}
#undef FR_DrawCharXY
void FR_DrawCharXY(unsigned char ch, int x, int y) {
    FR_DrawCharXY2(ch, x, y, DEFAULT_ALIGNFLAGS);
}
void FR_Init(void) {
    if (initedFont || novideo) return;
    initedFont = true;
    fr.fontNum = 0;
    FR_LoadDefaultAttrib();
    typeInTime = 0;
}
#undef Fonts_ResolveUri
DENG_EXTERN_C fontid_t Fonts_ResolveUri(uri_s const *uri) {
    if (!uri) return NOFONTID;
    try {
        return App_Resources().fontManifest(*reinterpret_cast<de::Uri const *>(uri)).uniqueId();
    }
    catch (Resources::MissingResourceManifestError const &)
    {}
    return NOFONTID;
}
DENG_DECLARE_API(FR) = {
    { DE_API_FONT_RENDER },
    Fonts_ResolveUri,
    FR_Font,
    FR_SetFont,
    FR_PushAttrib,
    FR_PopAttrib,
    FR_LoadDefaultAttrib,
    FR_Leading,
    FR_SetLeading,
    FR_Tracking,
    FR_SetTracking,
    FR_ColorAndAlpha,
    FR_SetColor,
    FR_SetColorv,
    FR_SetColorAndAlpha,
    FR_SetColorAndAlphav,
    FR_ColorRed,
    FR_SetColorRed,
    FR_ColorGreen,
    FR_SetColorGreen,
    FR_ColorBlue,
    FR_SetColorBlue,
    FR_Alpha,
    FR_SetAlpha,
    FR_ShadowOffset,
    FR_SetShadowOffset,
    FR_ShadowStrength,
    FR_SetShadowStrength,
    FR_GlitterStrength,
    FR_SetGlitterStrength,
    FR_CaseScale,
    FR_SetCaseScale,
    FR_DrawText,
    FR_DrawText2,
    FR_DrawText3,
    FR_DrawTextXY3,
    FR_DrawTextXY2,
    FR_DrawTextXY,
    FR_TextSize,
    FR_TextWidth,
    FR_TextHeight,
    FR_DrawChar3,
    FR_DrawChar2,
    FR_DrawChar,
    FR_DrawCharXY3,
    FR_DrawCharXY2,
    FR_DrawCharXY,
    FR_CharSize,
    FR_CharWidth,
    FR_CharHeight,
    FR_ResetTypeinTimer
};
struct flare_t {
    float offset;
    float size;
    float alpha;
    int   texture; // 0=round, 1=flare, 2=brflare, 3=bigflare
};
D_CMD(FlareConfig);
int     haloMode = 5, haloBright = 45, haloSize = 80;
int     haloRealistic = true;
int     haloOccludeSpeed = 48;
float   haloZMagDiv = 62, haloMinRadius = 20;
float   haloDimStart = 10, haloDimEnd = 100;
float   haloFadeMax = 0, haloFadeMin = 0, minHaloSize = 1;
flare_t flares[NUM_FLARES] = {
    {0, 1, 1, 0},               // Primary flare.
    {1, .41f, .5f, 0},          // Main secondary flare.
    {1.5f, .29f, .333f, 1},
    {-.6f, .24f, .333f, 0},
    {.4f, .29f, .25f, 0}
};
void H_Register(void) {
    cvartemplate_t cvars[] = {
        {"rend-halo", 0, CVT_INT, &haloMode, 0, 5},
        {"rend-halo-realistic", 0, CVT_INT, &haloRealistic, 0, 1},
        {"rend-halo-bright", 0, CVT_INT, &haloBright, 0, 100},
        {"rend-halo-occlusion", CVF_NO_MAX, CVT_INT, &haloOccludeSpeed, 0, 0},
        {"rend-halo-size", 0, CVT_INT, &haloSize, 0, 100},
        {"rend-halo-secondary-limit", CVF_NO_MAX, CVT_FLOAT, &minHaloSize, 0, 0},
        {"rend-halo-fade-far", CVF_NO_MAX, CVT_FLOAT, &haloFadeMax, 0, 0},
        {"rend-halo-fade-near", CVF_NO_MAX, CVT_FLOAT, &haloFadeMin, 0, 0},
        {"rend-halo-zmag-div", CVF_NO_MAX, CVT_FLOAT, &haloZMagDiv, 1, 1},
        {"rend-halo-radius-min", CVF_NO_MAX, CVT_FLOAT, &haloMinRadius, 0, 0},
        {"rend-halo-dim-near", CVF_NO_MAX, CVT_FLOAT, &haloDimStart, 0, 0},
        {"rend-halo-dim-far", CVF_NO_MAX, CVT_FLOAT, &haloDimEnd, 0, 0},
        {NULL}
    };
    Con_AddVariableList(cvars);
    C_CMD_FLAGS("flareconfig", NULL, FlareConfig, CMDF_NO_DEDICATED);
}
TextureVariantSpec const &Rend_HaloTextureSpec() {
    return App_Resources().textureSpec(TC_HALO_LUMINANCE,
        TSF_NO_COMPRESSION, 0, 0, 0, GL_CLAMP_TO_EDGE, GL_CLAMP_TO_EDGE, 1, 1, 0,
        false, false, false, true);
}
void H_SetupState(bool dosetup) {
    DENG2_ASSERT_IN_RENDER_THREAD();
    DENG_ASSERT_GL_CONTEXT_ACTIVE();
    if(dosetup) {
        DGL_Disable(DGL_DEPTH_WRITE);
        DGL_Disable(DGL_DEPTH_TEST);
        GL_BlendMode(BM_ADD);
    }
    else {
        DGL_Enable(DGL_DEPTH_WRITE);
        DGL_Enable(DGL_DEPTH_TEST);
        GL_BlendMode(BM_NORMAL);
    }
}
static inline float distanceDimFactorAt(coord_t distToViewer, float size) {
    if(haloDimStart && haloDimStart < haloDimEnd &&
       distToViewer / size > haloDimStart) {
        return 1 - (distToViewer / size - haloDimStart) / (haloDimEnd - haloDimStart);
    }
    return 1;
}
static inline float fadeFactorAt(coord_t distToViewer) {
    if(haloFadeMax && haloFadeMax != haloFadeMin &&
       distToViewer < haloFadeMax && distToViewer >= haloFadeMin) {
        return (distToViewer - haloFadeMin) / (haloFadeMax - haloFadeMin);
    }
    return 1;
}
bool H_RenderHalo(Vector3d const &origin, float size, DGLuint tex,
    Vector3f const &color, coord_t distanceToViewer,
    float occlusionFactor, float brightnessFactor, float viewXOffset,
    bool doPrimary, bool viewRelativeRotate) {
    if(!doPrimary && haloRealistic)
        return false;
    if(distanceToViewer <= 0 || occlusionFactor == 0 ||
       (haloFadeMax && distanceToViewer > haloFadeMax))
        return false;
    occlusionFactor = (1 + occlusionFactor) / 2;
    viewdata_t const *viewData = &viewPlayer->viewport();
    Vector3f const leftOff  = viewData->upVec + viewData->sideVec;
    Vector3f const rightOff = viewData->upVec - viewData->sideVec;
    Vector3f const center = Vector3f(origin.x, origin.z, origin.y)
                          - viewData->sideVec * viewXOffset;
    Vector3f const viewToCenter = center - Rend_EyeOrigin();
    float const scale = viewToCenter.dot(viewData->frontVec)
                        / viewData->frontVec.dot(viewData->frontVec);
    Vector3f const mirror =
        (viewData->frontVec * scale - viewToCenter) * 2;
    float const fadeFactor    = fadeFactorAt(distanceToViewer);
    float const secFadeFactor = viewToCenter.normalize().dot(viewData->frontVec);
    float turnAngle = 0;
    if(viewRelativeRotate) {
        Vector3f haloPos = mirror.normalize();
        if(haloPos.length()) {
            turnAngle = de::clamp<float>(-1, haloPos.dot(viewData->upVec), 1);
            if(turnAngle >= 1)
                turnAngle = 0;
            else if(turnAngle <= -1)
                turnAngle = float(de::PI);
            else
                turnAngle = acos(turnAngle);
            if(haloPos.dot(viewData->sideVec) < 0)
                turnAngle = -turnAngle;
        }
    }
    float const luminosity = (color.x + color.y + color.z) / 3;
    float const distanceDim = distanceDimFactorAt(distanceToViewer, size);
    if(doPrimary)
        H_SetupState(true);
    DENG2_ASSERT_IN_RENDER_THREAD();
    DENG_ASSERT_GL_CONTEXT_ACTIVE();
    DGL_MatrixMode(DGL_TEXTURE);
    DGL_PushMatrix();
    DGL_LoadIdentity();
    DGL_Translatef(0.5f, 0.5f, 0);
    DGL_Rotatef(turnAngle / float(de::PI) * 180, 0, 0, 1);
    DGL_Translatef(-0.5f, -0.5f, 0);
    flare_t *fl = flares;
    for(int i = 0; i < haloMode && i < NUM_FLARES; ++i, fl++) {
        bool const secondary = i != 0;
        if(doPrimary && secondary)
            break;
        if(!doPrimary && !secondary)
            continue;
        float alpha = fl->alpha * occlusionFactor * fadeFactor
                    + luminosity * luminosity / 5;
        alpha *= (!secondary? 1 : de::min<float>(minHaloSize * size / distanceToViewer, 1))
                 * distanceDim * brightnessFactor;
        alpha *= .8f * haloBright / 100.0f;
        if(fx::Bloom::isEnabled()) {
            alpha *= clamp(0.1f, (1.f - fx::Bloom::intensity() * .8f), 1.f);
        }
        if(secondary) {
            alpha *= luminosity - 8 * (1 - secFadeFactor);
        }
        if(haloRealistic) {
            alpha *= .6f;
        }
        if(alpha <= 0)
            break;
        float radius = size * (1 - luminosity / 3)
                     + distanceToViewer / haloZMagDiv;
        if(radius < haloMinRadius)
            radius = haloMinRadius;
        radius *= occlusionFactor;
        if(haloRealistic) {
            radius *= 0.8f;
        }
        if(haloRealistic) {
            if(!tex)
                tex = GL_PrepareSysFlaremap(FXT_ROUND);
        }
        else {
            if(!(doPrimary && tex)) {
                if(size > 45 || (luminosity > .90 && size > 20)) {
                    radius *= .65f;
                    if(!secondary)
                        tex = GL_PrepareSysFlaremap(FXT_BIGFLARE);
                    else
                        tex = GL_PrepareSysFlaremap(flaretexid_t(fl->texture));
                }
                else {
                    if(!secondary)
                        tex = GL_PrepareSysFlaremap(FXT_ROUND);
                    else
                        tex = GL_PrepareSysFlaremap(flaretexid_t(fl->texture));
                }
            }
        }
        Vector3f pos = center;
        if(secondary) {
            pos += mirror * fl->offset;
        }
        GL_BindTextureUnmanaged(renderTextures? tex : 0, gl::ClampToEdge, gl::ClampToEdge);
        DGL_Enable(DGL_TEXTURE_2D);
        float const radX = radius * fl->size;
        float const radY = radX / 1.2f; // Aspect correction.
        DGL_Color4f(color.x, color.y, color.z, alpha);
        DGL_Begin(DGL_QUADS);
            DGL_TexCoord2f(0, 0, 0);
            DGL_Vertex3f(pos.x + radX * leftOff.x,
                       pos.y + radY * leftOff.y,
                       pos.z + radX * leftOff.z);
            DGL_TexCoord2f(0, 1, 0);
            DGL_Vertex3f(pos.x + radX * rightOff.x,
                       pos.y + radY * rightOff.y,
                       pos.z + radX * rightOff.z);
            DGL_TexCoord2f(0, 1, 1);
            DGL_Vertex3f(pos.x - radX * leftOff.x,
                       pos.y - radY * leftOff.y,
                       pos.z - radX * leftOff.z);
            DGL_TexCoord2f(0, 0, 1);
            DGL_Vertex3f(pos.x - radX * rightOff.x,
                       pos.y - radY * rightOff.y,
                       pos.z - radX * rightOff.z);
        DGL_End();
        DGL_Disable(DGL_TEXTURE_2D);
    }
    DGL_MatrixMode(DGL_TEXTURE);
    DGL_PopMatrix();
    if(doPrimary)
        H_SetupState(false);
    return true;
}
/**
 * flareconfig list
 * flareconfig (num) pos/size/alpha/tex (val)
 */
D_CMD(FlareConfig) {
    DENG2_UNUSED(src);
    int             i;
    float           val;
    if(argc == 2) {
        if(!stricmp(argv[1], "list")) {
            for(i = 0; i < NUM_FLARES; ++i) {
                LOG_MSG("%i: pos:%f s:%.2f a:%.2f tex:%i")
                        << i << flares[i].offset << flares[i].size << flares[i].alpha
                        << flares[i].texture;
            }
        }
    }
    else if(argc == 4) {
        i = atoi(argv[1]);
        val = strtod(argv[3], NULL);
        if(i < 0 || i >= NUM_FLARES)
            return false;
        if(!stricmp(argv[2], "pos")) {
            flares[i].offset = val;
        }
        else if(!stricmp(argv[2], "size")) {
            flares[i].size = val;
        }
        else if(!stricmp(argv[2], "alpha")) {
            flares[i].alpha = val;
        }
        else if(!stricmp(argv[2], "tex")) {
            flares[i].texture = (int) val;
        }
    }
    else {
        LOG_SCR_NOTE("Usage:\n"
                     "  %s list\n"
                     "  %s (num) pos/size/alpha/tex (val)") << argv[0] << argv[0];
    }
    return true;
}
/**
 * POD structure used to transport vertex data refs as a logical unit.
 * @note The geometric data itself is not owned!
 */
struct Geometry {
    Vector3f *pos;
    Vector4f *color;
    Vector2f *tex;
    Vector2f *tex2;
};
/**
 * POD structure used to describe texture modulation data.
 */
struct TexModulationData {
    DGLuint texture = 0;
    Vector3f color;
    Vector2f topLeft;
    Vector2f bottomRight;
};
#define SVF_TANGENT             0x01
#define SVF_BITANGENT           0x02
#define SVF_NORMAL              0x04
/**
 * @defgroup soundOriginFlags  Sound Origin Flags
 * Flags for use with the sound origin debug display.
 * @ingroup flags
 */
#define SOF_SECTOR              0x01
#define SOF_PLANE               0x02
#define SOF_SIDE                0x04
void Rend_DrawBBox(Vector3d const &pos, coord_t w, coord_t l, coord_t h, dfloat a,
    dfloat const color[3], dfloat alpha, dfloat br, bool alignToBase = true);
void Rend_DrawArrow(Vector3d const &pos, dfloat a, dfloat s, dfloat const color3f[3], dfloat alpha);
D_CMD(OpenRendererAppearanceEditor);
D_CMD(LowRes);
D_CMD(MipMap);
D_CMD(TexReset);
D_CMD(CubeShot);
#if 0
dint useBias;  ///< Shadow Bias enabled? cvar
#endif
FogParams fogParams;
dfloat fieldOfView = 95.0f;
dbyte smoothTexAnim = true;
dint renderTextures = true;
#if defined (DENG_OPENGL)
dint renderWireframe;
#endif
dint dynlightBlend;
Vector3f torchColor(1, 1, 1);
dint useShinySurfaces = true;
dint useDynLights = true;
dfloat dynlightFactor = .7f;
dfloat dynlightFogBright = .15f;
dint useGlowOnWalls = true;
dfloat glowFactor = .8f;
dfloat glowHeightFactor = 3;  ///< Glow height as a multiplier.
dint glowHeightMax = 100;     ///< 100 is the default (0-1024).
dint useShadows = true;
dfloat shadowFactor = 1.2f;
dint shadowMaxRadius = 80;
dint shadowMaxDistance = 1000;
dbyte useLightDecorations = true;  ///< cvar
dfloat detailFactor = .5f;
dfloat detailScale = 4;
dint mipmapping = 5;
dint filterUI   = 1;
dint texQuality = TEXQ_BEST;
dint ratioLimit;      ///< Zero if none.
dd_bool fillOutlines = true;
dint useSmartFilter;  ///< Smart filter mode (cvar: 1=hq2x)
dint filterSprites = true;
dint texMagMode = 1;  ///< Linear.
dint texAniso = -1;   ///< Use best.
dbyte loadExtAlways;  ///< Always check for extres (cvar)
dfloat texGamma;
dint glmode[6] = // Indexed by 'mipmapping'.
{
    GL_NEAREST,
    GL_LINEAR,
    GL_NEAREST_MIPMAP_NEAREST,
    GL_LINEAR_MIPMAP_NEAREST,
    GL_NEAREST_MIPMAP_LINEAR,
    GL_LINEAR_MIPMAP_LINEAR
};
Vector3d vOrigin;
dfloat vang, vpitch;
dfloat viewsidex, viewsidey;
struct FixedView {
    Matrix4f projectionMatrix;
    Matrix4f modelViewMatrix;
    float yaw;
    float pitch;
    float horizontalFov;
};
static std::unique_ptr<FixedView> fixedView;
dbyte freezeRLs;
dint devNoCulling;  ///< @c 1= disabled (cvar).
dint devRendSkyMode;
dbyte devRendSkyAlways;
dint rAmbient, ambientLight;
dfloat yfov;
dint gameDrawHUD = 1;  ///< Set to zero when we advise that the HUD should not be drawn
/**
 * Implements a pre-calculated LUT for light level limiting and range
 * compression offsets, arranged such that it may be indexed with a
 * light level value. Return value is an appropriate delta (considering
 * all applicable renderer properties) which has been pre-clamped such
 * that when summed with the original light value the result remains in
 * the normalized range [0..1].
 */
dfloat lightRangeCompression;
dfloat lightModRange[255];
byte devLightModRange;
dfloat rendLightDistanceAttenuation = 924;
dint rendLightAttenuateFixedColormap = 1;
dfloat rendLightWallAngle = 1.2f; ///< Intensity of angle-based wall lighting.
dbyte rendLightWallAngleSmooth = true;
dfloat rendSkyLight = .273f;     ///< Intensity factor.
dbyte rendSkyLightAuto = true;
dint rendMaxLumobjs;             ///< Max lumobjs per viewer, per frame. @c 0= no maximum.
dint extraLight;  ///< Bumped light from gun blasts.
dfloat extraLightDelta;
/*
 * Debug/Development cvars:
 */
dbyte devMobjVLights;            ///< @c 1= Draw mobj vertex lighting vector.
dint devMobjBBox;                ///< @c 1= Draw mobj bounding boxes.
dint devPolyobjBBox;             ///< @c 1= Draw polyobj bounding boxes.
dbyte devVertexIndices;          ///< @c 1= Draw vertex indices.
dbyte devVertexBars;             ///< @c 1= Draw vertex position bars.
dbyte devDrawGenerators;         ///< @c 1= Draw active generators.
dbyte devSoundEmitters;          ///< @c 1= Draw sound emitters.
dbyte devSurfaceVectors;         ///< @c 1= Draw tangent space vectors for surfaces.
dbyte devNoTexFix;               ///< @c 1= Draw "missing" rather than fix materials.
dbyte devSectorIndices;          ///< @c 1= Draw sector indicies.
dbyte devThinkerIds;             ///< @c 1= Draw (mobj) thinker indicies.
dbyte rendInfoLums;              ///< @c 1= Print lumobj debug info to the console.
dbyte devDrawLums;               ///< @c 1= Draw lumobjs origins.
#if 0
dbyte devLightGrid;              ///< @c 1= Draw lightgrid debug visual.
dfloat devLightGridSize = 1.5f;  ///< Lightgrid debug visual size factor.
#endif
static void drawGenerators(Map &map);
static void drawLumobjs(Map &map);
static void drawMobjBoundingBoxes(Map &map);
static void drawSectors(Map &map);
static void drawSoundEmitters(Map &map);
static void drawSurfaceTangentVectors(Map &map);
static void drawThinkers(Map &map);
static void drawVertexes(Map &map);
static Vector3d eyeOrigin;            ///< Viewer origin.
static ConvexSubspace *curSubspace;   ///< Subspace currently being drawn.
static Vector3f curSectorLightColor;
static dfloat curSectorLightLevel;
static bool firstSubspace;            ///< No range checking for the first one.
static MaterialVariantSpec const *lookupMapSurfaceMaterialSpec = nullptr;
static QHash<Record const *, MaterialAnimator *> lookupSpriteMaterialAnimators;
void Rend_ResetLookups() {
    lookupMapSurfaceMaterialSpec = nullptr;
    lookupSpriteMaterialAnimators.clear();
    if (ClientApp::world().hasMap()) {
        auto &map = ClientApp::world().map();
        map.forAllSectors([] (Sector &sector) {
            sector.forAllPlanes([] (Plane &plane) {
                plane.surface().resetLookups();
                return LoopContinue;
            });
            return LoopContinue;
        });
        map.forAllLines([] (Line &line) {
            auto resetFunc = [] (Surface &surface) {
                surface.resetLookups();
                return LoopContinue;
            };
            line.front().forAllSurfaces(resetFunc);
            line.back() .forAllSurfaces(resetFunc);
            return LoopContinue;
        });
    }
}
static void reportWallDrawn(Line &line) {
    dint playerNum = DoomsdayApp::players().indexOf(viewPlayer);
    if (line.isMappedByPlayer(playerNum)) return;
    line.setMappedByPlayer(playerNum);
    if (gx.HandleMapObjectStatusReport) {
        gx.HandleMapObjectStatusReport(DMUSC_LINE_FIRSTRENDERED, line.indexInMap(),
                                       DMU_LINE, &playerNum);
    }
}
#if 0
static void scheduleFullLightGridUpdate() {
    if (!ClientApp::world().hasMap()) return;
    if (ClientApp::world().map().hasLightGrid()) {
        ClientApp::world().map().lightGrid().scheduleFullUpdate();
    }
}
#endif
void Rend_Reset() {
    R_ClearViewData();
    if (App_World().hasMap()) {
        App_World().map().removeAllLumobjs();
    }
}
bool Rend_IsMTexLights() {
    return IS_MTEX_LIGHTS;
}
bool Rend_IsMTexDetails() {
    return IS_MTEX_DETAILS;
}
dfloat Rend_FieldOfView() {
    if (fixedView) {
        return fixedView->horizontalFov;
    }
    if (vrCfg().mode() == VRConfig::OculusRift) {
        return vrCfg().oculusRift().fovX();
    }
    else {
        return de::clamp(1.f, fieldOfView, 179.f);
    }
}
static Vector3d vEyeOrigin;
Vector3d Rend_EyeOrigin() {
    return vEyeOrigin;
}
void Rend_SetFixedView(int consoleNum, float yaw, float pitch, float fov, Vector2f viewportSize) {
    viewdata_t const *viewData = &DD_Player(consoleNum)->viewport();
    fixedView.reset(new FixedView);
    fixedView->yaw = yaw;
    fixedView->pitch = pitch;
    fixedView->horizontalFov = fov;
    fixedView->modelViewMatrix =
            Matrix4f::rotate(pitch, Vector3f(1, 0, 0)) *
            Matrix4f::rotate(yaw,   Vector3f(0, 1, 0)) *
            Matrix4f::scale(Vector3f(1.0f, 1.2f, 1.0f)) * // This is the aspect correction.
            Matrix4f::translate(-viewData->current.origin.xzy());
    Rangef const clip = GL_DepthClipRange();
    fixedView->projectionMatrix = BaseGuiApp::app().vr()
            .projectionMatrix(fov, viewportSize, clip.start, clip.end) *
            Matrix4f::scale(Vector3f(1, 1, -1));
}
void Rend_UnsetFixedView() {
    fixedView.reset();
}
Matrix4f Rend_GetModelViewMatrix(dint consoleNum, bool inWorldSpace) {
    viewdata_t const *viewData = &DD_Player(consoleNum)->viewport();
    vOrigin = viewData->current.origin.xzy();
    vEyeOrigin = vOrigin;
    if (fixedView) {
        vang   = fixedView->yaw;
        vpitch = fixedView->pitch;
        return fixedView->modelViewMatrix;
    }
    vang   = viewData->current.angle() / (dfloat) ANGLE_MAX * 360 - 90;  // head tracking included
    vpitch = viewData->current.pitch * 85.0 / 110.0;
    dfloat bodyAngle = viewData->current.angleWithoutHeadTracking() / (dfloat) ANGLE_MAX * 360 - 90;
    OculusRift &ovr = vrCfg().oculusRift();
    bool const applyHead = (vrCfg().mode() == VRConfig::OculusRift && ovr.isReady());
    Matrix4f modelView;
    Matrix4f headOrientation;
    Matrix4f headOffset;
    if (applyHead) {
        Vector3f headPos = swizzle(Matrix4f::rotate(bodyAngle, Vector3f(0, 1, 0))
                                       * ovr.headPosition() * vrCfg().mapUnitsPerMeter()
                                   , AxisNegX, AxisNegY, AxisZ);
        headOffset = Matrix4f::translate(headPos);
        vEyeOrigin -= headPos;
    }
    if (inWorldSpace) {
        dfloat yaw   = vang;
        dfloat pitch = vpitch;
        dfloat roll  = 0;
        if (applyHead) {
            Vector3f const pry = ovr.headOrientation();
            roll  = -radianToDegree(pry[1]);
            pitch =  radianToDegree(pry[0]);
        }
        headOrientation = Matrix4f::rotate(roll,  Vector3f(0, 0, 1))
                        * Matrix4f::rotate(pitch, Vector3f(1, 0, 0))
                        * Matrix4f::rotate(yaw,   Vector3f(0, 1, 0));
        modelView = headOrientation * headOffset;
    }
    if (applyHead) {
        vEyeOrigin -= headOrientation.inverse() * (ovr.eyeOffset() * vrCfg().mapUnitsPerMeter());
    }
    return (modelView
            * Matrix4f::scale(Vector3f(1.0f, 1.2f, 1.0f)) // This is the aspect correction.
            * Matrix4f::translate(-vOrigin));
}
void Rend_ModelViewMatrix(bool inWorldSpace) {
    DENG2_ASSERT_IN_RENDER_THREAD();
    DENG_ASSERT_GL_CONTEXT_ACTIVE();
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_LoadMatrix(
        Rend_GetModelViewMatrix(DoomsdayApp::players().indexOf(viewPlayer), inWorldSpace).values());
}
Matrix4f Rend_GetProjectionMatrix(float fixedFov, float clipRangeScale) {
    if (fixedView) {
        return fixedView->projectionMatrix;
    }
    const Vector2f size = R_Console3DViewRect(displayPlayer).size();
    if (fixedFov > 0) {
        const float ASPECT_DEFAULT = 16.0f / 9.0f;
        const float ASPECT_MIN     = 4.0f / 3.0f;
        const float aspect         = de::max(size.x / size.y, ASPECT_MIN);
        if (aspect < ASPECT_DEFAULT) {
            fixedFov *= lerp(0.825f, 1.0f, (aspect - ASPECT_MIN) / (ASPECT_DEFAULT - ASPECT_MIN));
        }
    }
    const dfloat   fov  = (fixedFov > 0 ? fixedFov : Rend_FieldOfView());
    yfov                = vrCfg().verticalFieldOfView(fov, size);
    const Rangef clip   = GL_DepthClipRange();
    return vrCfg().projectionMatrix(
               fov, size, clip.start * clipRangeScale, clip.end * clipRangeScale) *
           Matrix4f::scale(Vector3f(1, 1, -1));
}
static inline ddouble viewFacingDot(Vector2d const &v1, Vector2d const &v2) {
    return (v1.y - v2.y) * (v1.x - Rend_EyeOrigin().x) + (v2.x - v1.x) * (v1.y - Rend_EyeOrigin().z);
}
dfloat Rend_ExtraLightDelta() {
    return extraLightDelta;
}
void Rend_ApplyTorchLight(Vector4f &color, dfloat distance) {
    ddplayer_t *ddpl = &viewPlayer->publicData();
    if (!ddpl->fixedColorMap) return;
    if (!rendLightAttenuateFixedColormap || distance < 1024) {
        dfloat d = (16 - ddpl->fixedColorMap) / 15.0f;
        if (rendLightAttenuateFixedColormap) {
            d *= (1024 - distance) / 1024.0f;
        }
        color = color.max(torchColor * d);
    }
}
void Rend_ApplyTorchLight(dfloat *color3, dfloat distance) {
    Vector4f tmp(color3, 0);
    Rend_ApplyTorchLight(tmp, distance);
    for (dint i = 0; i < 3; ++i) {
        color3[i] = tmp[i];
    }
}
dfloat Rend_AttenuateLightLevel(dfloat distToViewer, dfloat lightLevel) {
    if (distToViewer > 0 && rendLightDistanceAttenuation > 0) {
        dfloat real = lightLevel -
            (distToViewer - 32) / rendLightDistanceAttenuation *
                (1 - lightLevel);
        dfloat minimum = de::max(0.f, de::squared(lightLevel) + (lightLevel - .63f) * .5f);
        if (real < minimum)
            real = minimum; // Clamp it.
        return de::min(real, 1.f);
    }
    return lightLevel;
}
dfloat Rend_ShadowAttenuationFactor(coord_t distance) {
    if (shadowMaxDistance > 0 && distance > 3 * shadowMaxDistance / 4) {
        return (shadowMaxDistance - distance) / (shadowMaxDistance / 4);
    }
    return 1;
}
static Vector3f skyLightColor;
static Vector3f oldSkyAmbientColor(-1.f, -1.f, -1.f);
static dfloat oldRendSkyLight = -1;
bool Rend_SkyLightIsEnabled() {
    return rendSkyLight > .001f;
}
Vector3f Rend_SkyLightColor() {
    if (Rend_SkyLightIsEnabled() && ClientApp::world().hasMap()) {
        Sky &sky = ClientApp::world().map().sky();
        Vector3f const &ambientColor = sky.ambientColor();
        if (rendSkyLight != oldRendSkyLight
            || !INRANGE_OF(ambientColor.x, oldSkyAmbientColor.x, .001f)
            || !INRANGE_OF(ambientColor.y, oldSkyAmbientColor.y, .001f)
            || !INRANGE_OF(ambientColor.z, oldSkyAmbientColor.z, .001f)) {
            skyLightColor = ambientColor;
            R_AmplifyColor(skyLightColor);
            for (dint i = 0; i < 3; ++i) {
                skyLightColor[i] = skyLightColor[i] + (1 - rendSkyLight) * (1.f - skyLightColor[i]);
            }
#if 0
            scheduleFullLightGridUpdate();
#endif
            oldSkyAmbientColor = ambientColor;
        }
        oldRendSkyLight = rendSkyLight;
        return skyLightColor;
    }
    return Vector3f(1, 1, 1);
}
/**
 * Determine the effective ambient light color for the given @a sector. Usually
 * one would obtain this info from Subsector, however in some situations the
 * correct light color is *not* that of the subsector (e.g., where map hacks use
 * mapped planes to reference another sector).
 */
static Vector3f Rend_AmbientLightColor(Sector const &sector) {
    if (Rend_SkyLightIsEnabled() && sector.hasSkyMaskPlane()) {
        return Rend_SkyLightColor();
    }
    return sector.lightColor();
}
Vector3f Rend_LuminousColor(Vector3f const &color, dfloat light) {
    light = de::clamp(0.f, light, 1.f) * dynlightFactor;
    if (fogParams.usingFog) light *= dynlightFogBright;
    return color * light;
}
coord_t Rend_PlaneGlowHeight(dfloat intensity) {
    return de::clamp<ddouble>(0, GLOW_HEIGHT_MAX * intensity * glowHeightFactor, glowHeightMax);
}
ClientMaterial *Rend_ChooseMapSurfaceMaterial(Surface const &surface) {
    switch (renderTextures) {
    case 0:  // No texture mode.
    case 1:  // Normal mode.
        if (!(devNoTexFix && surface.hasFixMaterial())) {
            if (!surface.hasMaterial() && surface.parent().type() == DMU_SIDE) {
                const Line::Side &side = surface.parent().as<Line::Side>();
                if (side.hasSector()) {
                    if (auto *midAnim = side.middle().materialAnimator()) {
                        DENG2_ASSERT(&surface != &side.middle());
                        if (!midAnim->isOpaque()) {
                            return &midAnim->material();
                        }
                    }
                    if (&surface != &side.top() && !side.hasAtLeastOneMaterial()) {
                        return nullptr;
                    }
                    if (&surface == &side.bottom()) {
                        return static_cast<ClientMaterial *>(side.sector().floor().surface().materialPtr());
                    }
                    else if (&surface == &side.top()) {
                        /*
                        if (side.back().hasSector()) {
                            const auto &backSector = side.back().sector();
                            if (fequal(backSector.ceiling().height(), backSector.floor().height())) {
                                return nullptr;
                            }
                        }*/
                        return static_cast<ClientMaterial *>(side.sector().ceiling().surface().materialPtr());
                    }
                }
            }
            if (surface.hasMaterial() || surface.parent().type() != DMU_PLANE) {
                return static_cast<ClientMaterial *>(surface.materialPtr());
            }
        }
        return &ClientMaterial::find(de::Uri("System", Path("missing")));
    case 2:  // Lighting debug mode.
        if (surface.hasMaterial() && !(!devNoTexFix && surface.hasFixMaterial())) {
            if (!surface.hasSkyMaskedMaterial() || devRendSkyMode) {
                return &ClientMaterial::find(de::Uri("System", Path("gray")));
            }
        }
        break;
    default: break;
    }
    return nullptr;
}
/**
 * This doesn't create a rendering primitive but a vissprite! The vissprite
 * represents the masked poly and will be rendered during the rendering
 * of sprites. This is necessary because all masked polygons must be
 * rendered back-to-front, or there will be alpha artifacts along edges.
 */
void Rend_AddMaskedPoly(Vector3f const *rvertices, Vector4f const *rcolors,
    coord_t wallLength, MaterialAnimator *matAnimator, Vector2f const &origin,
    blendmode_t blendMode, duint lightListIdx, dfloat glow) {
    vissprite_t *vis = R_NewVisSprite(VSPR_MASKED_WALL);
    vis->pose.origin   = (rvertices[0] + rvertices[3]) / 2;
    vis->pose.distance = Rend_PointDist2D(vis->pose.origin);
    VS_WALL(vis)->texOffset[0] = origin[0];
    VS_WALL(vis)->texOffset[1] = origin[1];
    if (renderTextures) {
        matAnimator->prepare();
        Vector2ui const &matDimensions = matAnimator->dimensions();
        VS_WALL(vis)->texCoord[0][0] = VS_WALL(vis)->texOffset[0] / matDimensions.x;
        VS_WALL(vis)->texCoord[1][0] = VS_WALL(vis)->texCoord[0][0] + wallLength / matDimensions.x;
        VS_WALL(vis)->texCoord[0][1] = VS_WALL(vis)->texOffset[1] / matDimensions.y;
        VS_WALL(vis)->texCoord[1][1] = VS_WALL(vis)->texCoord[0][1] +
                (rvertices[3].z - rvertices[0].z) / matDimensions.y;
        dint wrapS = GL_REPEAT, wrapT = GL_REPEAT;
        if (!matAnimator->isOpaque()) {
            if (!(VS_WALL(vis)->texCoord[0][0] < 0 || VS_WALL(vis)->texCoord[0][0] > 1 ||
                 VS_WALL(vis)->texCoord[1][0] < 0 || VS_WALL(vis)->texCoord[1][0] > 1)) {
                wrapS = GL_CLAMP_TO_EDGE;
            }
            if (!(VS_WALL(vis)->texCoord[0][1] < 0 || VS_WALL(vis)->texCoord[0][1] > 1 ||
                 VS_WALL(vis)->texCoord[1][1] < 0 || VS_WALL(vis)->texCoord[1][1] > 1)) {
                wrapT = GL_CLAMP_TO_EDGE;
            }
        }
        matAnimator = &matAnimator->material().getAnimator(Rend_MapSurfaceMaterialSpec(wrapS, wrapT));
    }
    VS_WALL(vis)->animator  = matAnimator;
    VS_WALL(vis)->blendMode = blendMode;
    for (dint i = 0; i < 4; ++i) {
        VS_WALL(vis)->vertices[i].pos[0] = rvertices[i].x;
        VS_WALL(vis)->vertices[i].pos[1] = rvertices[i].y;
        VS_WALL(vis)->vertices[i].pos[2] = rvertices[i].z;
        for (dint c = 0; c < 4; ++c) {
            VS_WALL(vis)->vertices[i].color[c] = de::clamp(0.f, rcolors[i][c], 1.f);
        }
    }
    if (glow < 1 && lightListIdx && !(rcolors[0].w < 1)) {
        ClientApp::renderSystem().forAllSurfaceProjections(lightListIdx, [&vis] (ProjectedTextureData const &tp) {
            VS_WALL(vis)->modTex = tp.texture;
            VS_WALL(vis)->modTexCoord[0][0] = tp.topLeft.x;
            VS_WALL(vis)->modTexCoord[0][1] = tp.topLeft.y;
            VS_WALL(vis)->modTexCoord[1][0] = tp.bottomRight.x;
            VS_WALL(vis)->modTexCoord[1][1] = tp.bottomRight.y;
            for (dint c = 0; c < 4; ++c) {
                VS_WALL(vis)->modColor[c] = tp.color[c];
            }
            return LoopAbort;
        });
    }
    else {
        VS_WALL(vis)->modTex = 0;
    }
}
static void quadTexCoords(Vector2f *tc, Vector3f const *rverts, coord_t wallLength, Vector3d const &topLeft) {
    DENG2_ASSERT(tc && rverts);
    tc[0].x = tc[1].x = rverts[0].x - topLeft.x;
    tc[3].y = tc[1].y = rverts[0].y - topLeft.y;
    tc[3].x = tc[2].x = tc[0].x + wallLength;
    tc[2].y = tc[3].y + (rverts[1].z - rverts[0].z);
    tc[0].y = tc[3].y + (rverts[3].z - rverts[2].z);
}
static void lightVertex(Vector4f &color, Vector3f const &vtx, dfloat lightLevel, Vector3f const &ambientColor) {
    dfloat const dist = Rend_PointDist2D(vtx);
    lightLevel = Rend_AttenuateLightLevel(dist, lightLevel);
    lightLevel = de::clamp(0.f, lightLevel + Rend_ExtraLightDelta(), 1.f);
    Rend_ApplyLightAdaptation(lightLevel);
    color.x = lightLevel * ambientColor.x;
    color.y = lightLevel * ambientColor.y;
    color.z = lightLevel * ambientColor.z;
}
/**
 * Apply map-space lighting to the given geometry. All vertex lighting contributions affecting
 * map-space geometry are applied here.
 *
 * @param verts             Geometry to be illuminated.
 *
 * Surface geometry:
 * @param numVertices       Total number of map-space surface geometry vertices.
 * @param posCoords         Position coordinates for the map-space surface geometry.
 * @param mapElement        Source MapElement for the map-space surface geometry.
 * @param geomGroup         Source MapElement geometry-group selector, for the map-space surface geometry.
 * @param surfaceTangents   Tangent-space vectors for the map-space surface geometry.
 *
 * Surface lighting characteristics:
 * @param color             Tint color.
 * @param color2            Secondary tint color, for walls (if any).
 * @param glowing           Self-luminosity factor (normalized [0..1]).
 * @param luminosityDeltas  Edge luminosity deltas (for walls [left edge, right edge]).
 */
static void lightWallOrFlatGeometry(Geometry &verts, duint numVertices, Vector3f const *posCoords,
    MapElement &mapElement, dint /*geomGroup*/, Matrix3f const &/*surfaceTangents*/,
    Vector3f const &color, Vector3f const *color2, dfloat glowing, dfloat const luminosityDeltas[2]) {
    bool const haveWall = is<LineSideSegment>(mapElement);
    if (::levelFullBright || !(glowing < 1)) {
        dfloat const lum = de::clamp(0.f, ::curSectorLightLevel + (::levelFullBright? 1 : glowing), 1.f);
        Vector4f const uniformColor(lum, lum, lum, 0);
        for (duint i = 0; i < numVertices; ++i) {
            verts.color[i] = uniformColor;
        }
        return;
    }
#if 0
    if (::useBias)  // Bias lighting model.
    {
        Map &map     = subsec.sector().map();
        Shard &shard = subsec.shard(mapElement, geomGroup);
        if (map.hasLightGrid()) {
            for (duint i = 0; i < numVertices; ++i) {
                verts.color[i] = map.lightGrid().evaluate(posCoords[i]);
            }
        }
        shard.lightWithBiasSources(posCoords, verts.color, surfaceTangents, map.biasCurrentTime());
        if (glowing > 0) {
            Vector4f const glow(glowing, glowing, glowing, 0);
            for (duint i = 0; i < numVertices; ++i) {
                verts.color[i] += glow;
            }
        }
        for (duint i = 0; i < numVertices; ++i) {
            Vector4f &color = verts.color[i];
            for (dint k = 0; k < 3; ++k) {
                color[k] = de::clamp(0.f, color[k] + Rend_LightAdaptationDelta(color[k]), 1.f);
            }
        }
    }
    else  // Doom lighting model.
#endif
    {
        Vector3f const colorBlended = ::curSectorLightColor * color;
        dfloat const lumLeft  = de::clamp(0.f, ::curSectorLightLevel + luminosityDeltas[0] + glowing, 1.f);
        dfloat const lumRight = de::clamp(0.f, ::curSectorLightLevel + luminosityDeltas[1] + glowing, 1.f);
        if (haveWall && !de::fequal(lumLeft, lumRight)) {
            lightVertex(verts.color[0], posCoords[0], lumLeft,  colorBlended);
            lightVertex(verts.color[1], posCoords[1], lumLeft,  colorBlended);
            lightVertex(verts.color[2], posCoords[2], lumRight, colorBlended);
            lightVertex(verts.color[3], posCoords[3], lumRight, colorBlended);
        }
        else {
            for (duint i = 0; i < numVertices; ++i) {
                lightVertex(verts.color[i], posCoords[i], lumLeft, colorBlended);
            }
        }
        if (haveWall && color2) {
            Vector3f const color2Blended = ::curSectorLightColor * (*color2);
            lightVertex(verts.color[0], posCoords[0], lumLeft,  color2Blended);
            lightVertex(verts.color[2], posCoords[2], lumRight, color2Blended);
        }
    }
    DENG2_ASSERT(::viewPlayer);
    if (::viewPlayer->publicData().fixedColorMap) {
        for (duint i = 0; i < numVertices; ++i) {
            Rend_ApplyTorchLight(verts.color[i], Rend_PointDist2D(posCoords[i]));
        }
    }
}
static void makeFlatGeometry(Geometry &verts, duint numVertices, Vector3f const *posCoords,
    Vector3d const &topLeft, Vector3d const & /*bottomRight*/, MapElement &mapElement, dint geomGroup,
    Matrix3f const &surfaceTangents, dfloat uniformOpacity, Vector3f const &color, Vector3f const *color2,
    dfloat glowing, dfloat const luminosityDeltas[2], bool useVertexLighting = true) {
    DENG2_ASSERT(posCoords);
    for (duint i = 0; i < numVertices; ++i) {
        verts.pos[i] = posCoords[i];
        Vector3f const delta(posCoords[i] - topLeft);
        if (verts.tex)  // Primary.
        {
            verts.tex[i] = Vector2f(delta.x, -delta.y);
        }
        if (verts.tex2)  // Inter.
        {
            verts.tex2[i] = Vector2f(delta.x, -delta.y);
        }
    }
    if (useVertexLighting) {
        lightWallOrFlatGeometry(verts, numVertices, posCoords, mapElement, geomGroup, surfaceTangents,
                                color, color2, glowing, luminosityDeltas);
        for (duint i = 0; i < numVertices; ++i) {
            verts.color[i].w = uniformOpacity;
        }
    }
}
static void makeWallGeometry(Geometry &verts, duint numVertices, Vector3f const *posCoords,
    Vector3d const &topLeft, Vector3d const & /*bottomRight*/, coord_t sectionWidth,
    MapElement &mapElement, dint geomGroup, Matrix3f const &surfaceTangents, dfloat uniformOpacity,
    Vector3f const &color, Vector3f const *color2, dfloat glowing, dfloat const luminosityDeltas[2],
    bool useVertexLighting = true) {
    DENG2_ASSERT(posCoords);
    for (duint i = 0; i < numVertices; ++i) {
        verts.pos[i] = posCoords[i];
    }
    if (verts.tex)  // Primary.
    {
        quadTexCoords(verts.tex, posCoords, sectionWidth, topLeft);
    }
    if (verts.tex2)  // Inter.
    {
        quadTexCoords(verts.tex2, posCoords, sectionWidth, topLeft);
    }
    if (useVertexLighting) {
        lightWallOrFlatGeometry(verts, numVertices, posCoords, mapElement, geomGroup, surfaceTangents,
                                color, color2, glowing, luminosityDeltas);
        for (duint i = 0; i < numVertices; ++i) {
            verts.color[i].w = uniformOpacity;
        }
    }
}
static inline dfloat shineVertical(dfloat dy, dfloat dx) {
    return ((std::atan(dy/dx) / (PI/2)) + 1) / 2;
}
static void makeFlatShineGeometry(Geometry &verts, duint numVertices, Vector3f const *posCoords,
    Geometry const &mainVerts, Vector3f const &shineColor, dfloat shineOpacity) {
    DENG2_ASSERT(posCoords);
    for (duint i = 0; i < numVertices; ++i) {
        Vector3f const eye = Rend_EyeOrigin();
        dfloat distToEye = (eye.xz() - Vector2f(posCoords[i])).normalize().length();
        if (distToEye < 10) distToEye = 10;
        dfloat offset = ((eye.y - posCoords[i].y) * sin(.4f)/*viewFrontVec[0]*/ -
                         (eye.x - posCoords[i].x) * cos(.4f)/*viewFrontVec[2]*/);
        verts.tex[i] = Vector2f(((shineVertical(offset, distToEye) - .5f) * 2) + .5f,
                                shineVertical(eye.y - posCoords[i].z, distToEye));
    }
    for (duint i = 0; i < numVertices; ++i) {
        verts.color[i] = Vector4f(Vector3f(mainVerts.color[i]).max(shineColor), shineOpacity);
    }
}
static void makeWallShineGeometry(Geometry &verts, duint numVertices, Vector3f const *posCoords,
    Geometry const &mainVerts, coord_t sectionWidth, Vector3f const &shineColor, dfloat shineOpactiy) {
    DENG2_ASSERT(posCoords);
    Vector3f const &topLeft     = posCoords[1];
    Vector3f const &bottomRight = posCoords[2];
    Vector2f const normal = Vector2f( (bottomRight.y - topLeft.y) / sectionWidth,
                                     -(bottomRight.x - topLeft.x) / sectionWidth);
    Vector3f const eye = Rend_EyeOrigin();
    dfloat prevAngle = 0;
    for (duint i = 0; i < 2; ++i) {
        Vector2f const eyeToVert = eye.xz() - (i == 0? Vector2f(topLeft) : Vector2f(bottomRight));
        dfloat const distToEye   = eyeToVert.length();
        Vector2f const view      = eyeToVert.normalize();
        Vector2f projected;
        dfloat const div = normal.lengthSquared();
        if (div != 0) {
            projected = normal * view.dot(normal) / div;
        }
        Vector2f const reflected = view + (projected - view) * 2;
        dfloat angle = std::acos(reflected.y) / PI;
        if (reflected.x < 0) {
            angle = 1 - angle;
        }
        if (i == 0) {
            prevAngle = angle;
        }
        else if (angle > prevAngle) {
            angle -= 1;
        }
        verts.tex[ (i == 0 ? 1 : 2) ].x = angle + .3f /*std::acos(-dot)/PI*/;
        verts.tex[ (i == 0 ? 0 : 3) ].x = angle + .3f /*std::acos(-dot)/PI*/;
        verts.tex[ (i == 0 ? 0 : 2) ].y = shineVertical(eye.y - bottomRight.z, distToEye);
        verts.tex[ (i == 0 ? 1 : 3) ].y = shineVertical(eye.y - topLeft.z,     distToEye);
    }
    for (duint i = 0; i < numVertices; ++i) {
        verts.color[i] = Vector4f(Vector3f(mainVerts.color[i]).max(shineColor), shineOpactiy);
    }
}
static void makeFlatShadowGeometry(Geometry &verts, Vector3d const &topLeft, Vector3d const &bottomRight,
    duint numVertices, Vector3f const *posCoords,
    ProjectedTextureData const &tp) {
    DENG2_ASSERT(posCoords);
    Vector4f const colorClamped = tp.color.min(Vector4f(1, 1, 1, 1)).max(Vector4f(0, 0, 0, 0));
    for (duint i = 0; i < numVertices; ++i) {
        verts.color[i] = colorClamped;
    }
    dfloat const width  = bottomRight.x - topLeft.x;
    dfloat const height = bottomRight.y - topLeft.y;
    for (duint i = 0; i < numVertices; ++i) {
        verts.tex[i].x = ((bottomRight.x - posCoords[i].x) / width * tp.topLeft.x) +
            ((posCoords[i].x - topLeft.x) / width * tp.bottomRight.x);
        verts.tex[i].y = ((bottomRight.y - posCoords[i].y) / height * tp.topLeft.y) +
            ((posCoords[i].y - topLeft.y) / height * tp.bottomRight.y);
    }
    std::memcpy(verts.pos, posCoords, sizeof(Vector3f) * numVertices);
}
static void makeWallShadowGeometry(Geometry &verts, Vector3d const &/*topLeft*/, Vector3d const &/*bottomRight*/,
    duint numVertices, Vector3f const *posCoords, WallEdge const &leftEdge, WallEdge const &rightEdge,
    ProjectedTextureData const &tp) {
    DENG2_ASSERT(posCoords);
    Vector4f const colorClamped = tp.color.min(Vector4f(1, 1, 1, 1)).max(Vector4f(0, 0, 0, 0));
    for (duint i = 0; i < numVertices; ++i) {
        verts.color[i] = colorClamped;
    }
    verts.tex[1].x = verts.tex[0].x = tp.topLeft.x;
    verts.tex[1].y = verts.tex[3].y = tp.topLeft.y;
    verts.tex[3].x = verts.tex[2].x = tp.bottomRight.x;
    verts.tex[2].y = verts.tex[0].y = tp.bottomRight.y;
    if (leftEdge.divisionCount() || rightEdge.divisionCount()) {
        Vector3f origPosCoords[4]; std::memcpy(origPosCoords, posCoords,   sizeof(Vector3f) * 4);
        Vector2f origTexCoords[4]; std::memcpy(origTexCoords, verts.tex,   sizeof(Vector2f) * 4);
        Vector4f origColors[4];    std::memcpy(origColors,    verts.color, sizeof(Vector4f) * 4);
        R_DivVerts     (verts.pos,   origPosCoords, leftEdge, rightEdge);
        R_DivTexCoords (verts.tex,   origTexCoords, leftEdge, rightEdge);
        R_DivVertColors(verts.color, origColors,    leftEdge, rightEdge);
    }
    else {
        std::memcpy(verts.pos, posCoords, sizeof(Vector3f) * numVertices);
    }
}
static void makeFlatLightGeometry(Geometry &verts, Vector3d const &topLeft, Vector3d const &bottomRight,
    duint numVertices, Vector3f const *posCoords,
    ProjectedTextureData const &tp) {
    DENG2_ASSERT(posCoords);
    Vector4f const colorClamped = tp.color.min(Vector4f(1, 1, 1, 1)).max(Vector4f(0, 0, 0, 0));
    for (duint i = 0; i < numVertices; ++i) {
        verts.color[i] = colorClamped;
    }
    dfloat const width  = bottomRight.x - topLeft.x;
    dfloat const height = bottomRight.y - topLeft.y;
    for (duint i = 0; i < numVertices; ++i) {
        verts.tex[i].x = ((bottomRight.x - posCoords[i].x) / width * tp.topLeft.x) +
            ((posCoords[i].x - topLeft.x) / width * tp.bottomRight.x);
        verts.tex[i].y = ((bottomRight.y - posCoords[i].y) / height * tp.topLeft.y) +
            ((posCoords[i].y - topLeft.y) / height * tp.bottomRight.y);
    }
    std::memcpy(verts.pos, posCoords, sizeof(Vector3f) * numVertices);
}
static void makeWallLightGeometry(Geometry &verts, Vector3d const &/*topLeft*/, Vector3d const &/*bottomRight*/,
    duint numVertices, Vector3f const *posCoords, WallEdge const &leftEdge, WallEdge const &rightEdge,
    ProjectedTextureData const &tp) {
    DENG2_ASSERT(posCoords);
    Vector4f const colorClamped = tp.color.min(Vector4f(1, 1, 1, 1)).max(Vector4f(0, 0, 0, 0));
    for (duint i = 0; i < numVertices; ++i) {
        verts.color[i] = colorClamped;
    }
    verts.tex[1].x = verts.tex[0].x = tp.topLeft.x;
    verts.tex[1].y = verts.tex[3].y = tp.topLeft.y;
    verts.tex[3].x = verts.tex[2].x = tp.bottomRight.x;
    verts.tex[2].y = verts.tex[0].y = tp.bottomRight.y;
    if (leftEdge.divisionCount() || rightEdge.divisionCount()) {
        Vector3f origPosCoords[4]; std::memcpy(origPosCoords, posCoords,   sizeof(Vector3f) * 4);
        Vector2f origTexCoords[4]; std::memcpy(origTexCoords, verts.tex,   sizeof(Vector2f) * 4);
        Vector4f origColors[4];    std::memcpy(origColors,    verts.color, sizeof(Vector4f) * 4);
        R_DivVerts     (verts.pos,   origPosCoords, leftEdge, rightEdge);
        R_DivTexCoords (verts.tex,   origTexCoords, leftEdge, rightEdge);
        R_DivVertColors(verts.color, origColors,    leftEdge, rightEdge);
    }
    else {
        std::memcpy(verts.pos, posCoords, sizeof(Vector3f) * numVertices);
    }
}
static dfloat averageLuminosity(Vector4f const *rgbaValues, duint count) {
    DENG2_ASSERT(rgbaValues);
    dfloat avg = 0;
    for (duint i = 0; i < count; ++i) {
        Vector4f const &color = rgbaValues[i];
        avg += color.x + color.y + color.z;
    }
    return avg / (count * 3);
}
struct rendworldpoly_params_t {
    bool            skyMasked;
    blendmode_t     blendMode;
    Vector3d const *topLeft;
    Vector3d const *bottomRight;
    Vector2f const *materialOrigin;
    Vector2f const *materialScale;
    dfloat          alpha;
    dfloat          surfaceLuminosityDeltas[2];
    Vector3f const *surfaceColor;
    Matrix3f const *surfaceTangentMatrix;
    duint           lightListIdx;   ///< List of lights that affect this poly.
    duint           shadowListIdx;  ///< List of shadows that affect this poly.
    dfloat          glowing;
    bool            forceOpaque;
    MapElement     *mapElement;
    dint            geomGroup;
    bool            isWall;
    struct {
        coord_t width;
        Vector3f const *surfaceColor2;  ///< Secondary color.
        WallEdge const *leftEdge;
        WallEdge const *rightEdge;
    } wall;
};
static bool renderWorldPoly(Vector3f const *rvertices, duint numVertices,
    rendworldpoly_params_t const &p, MaterialAnimator &matAnimator) {
    using Parm = DrawList::PrimitiveParams;
    DENG2_ASSERT(rvertices);
    static DrawList::Indices indices;
    matAnimator.prepare();
    bool const skyMaskedMaterial        = (p.skyMasked || (matAnimator.material().isSkyMasked()));
    bool const drawAsVisSprite          = (!p.forceOpaque && !p.skyMasked && (!matAnimator.isOpaque() || p.alpha < 1 || p.blendMode > 0));
    GLTextureUnit const *layer0RTU      = (!p.skyMasked)? &matAnimator.texUnit(MaterialAnimator::TU_LAYER0) : nullptr;
    GLTextureUnit const *layer0InterRTU = (!p.skyMasked && !drawAsVisSprite && matAnimator.texUnit(MaterialAnimator::TU_LAYER0_INTER).hasTexture())? &matAnimator.texUnit(MaterialAnimator::TU_LAYER0_INTER) : nullptr;
    GLTextureUnit const *detailRTU      = (r_detail && !p.skyMasked && matAnimator.texUnit(MaterialAnimator::TU_DETAIL).hasTexture())? &matAnimator.texUnit(MaterialAnimator::TU_DETAIL) : nullptr;
    GLTextureUnit const *detailInterRTU = (r_detail && !p.skyMasked && matAnimator.texUnit(MaterialAnimator::TU_DETAIL_INTER).hasTexture())? &matAnimator.texUnit(MaterialAnimator::TU_DETAIL_INTER) : nullptr;
    GLTextureUnit const *shineRTU       = (::useShinySurfaces && !skyMaskedMaterial && !drawAsVisSprite && matAnimator.texUnit(MaterialAnimator::TU_SHINE).hasTexture())? &matAnimator.texUnit(MaterialAnimator::TU_SHINE) : nullptr;
    GLTextureUnit const *shineMaskRTU   = (::useShinySurfaces && !skyMaskedMaterial && !drawAsVisSprite && matAnimator.texUnit(MaterialAnimator::TU_SHINE).hasTexture() && matAnimator.texUnit(MaterialAnimator::TU_SHINE_MASK).hasTexture())? &matAnimator.texUnit(MaterialAnimator::TU_SHINE_MASK) : nullptr;
    Geometry verts;
    bool const mustSubdivide = (p.isWall && (p.wall.leftEdge->divisionCount() || p.wall.rightEdge->divisionCount()));
    duint const numVerts     = (mustSubdivide && !drawAsVisSprite?   3 + p.wall.leftEdge ->divisionCount()
                                                                   + 3 + p.wall.rightEdge->divisionCount()
                                                                 : numVertices);
    verts.pos   = R_AllocRendVertices(numVerts);
    verts.color = !skyMaskedMaterial? R_AllocRendColors   (numVerts) : nullptr;
    verts.tex   = layer0RTU         ? R_AllocRendTexCoords(numVerts) : nullptr;
    verts.tex2  = layer0InterRTU    ? R_AllocRendTexCoords(numVerts) : nullptr;
    if (p.isWall) {
        makeWallGeometry(verts, numVertices, rvertices, *p.topLeft, *p.bottomRight, p.wall.width,
                         *p.mapElement, p.geomGroup, *p.surfaceTangentMatrix,
                         p.alpha, *p.surfaceColor, p.wall.surfaceColor2, p.glowing, p.surfaceLuminosityDeltas,
                         !skyMaskedMaterial);
    }
    else {
        makeFlatGeometry(verts, numVertices, rvertices, *p.topLeft, *p.bottomRight,
                         *p.mapElement, p.geomGroup, *p.surfaceTangentMatrix,
                         p.alpha, *p.surfaceColor, p.wall.surfaceColor2, p.glowing, p.surfaceLuminosityDeltas,
                         !skyMaskedMaterial);
    }
    if (drawAsVisSprite) {
        DENG2_ASSERT(p.isWall);
        Rend_AddMaskedPoly(verts.pos, verts.color, p.wall.width, &matAnimator,
                           *p.materialOrigin, p.blendMode, p.lightListIdx, p.glowing);
        R_FreeRendVertices (verts.pos);
        R_FreeRendColors   (verts.color);
        R_FreeRendTexCoords(verts.tex);
        R_FreeRendTexCoords(verts.tex2);
        return false;  // We HAD to use a vissprite, so it MUST not be opaque.
    }
    bool useLights  = (p.lightListIdx  && !skyMaskedMaterial && !drawAsVisSprite && p.glowing < 1? true : false);
    bool useShadows = (p.shadowListIdx && !skyMaskedMaterial && !drawAsVisSprite && p.glowing < 1? true : false);
    if (useLights || useShadows) {
        dfloat const avg = averageLuminosity(verts.color, numVertices);
        if (avg > 0.98f)  // Fully bright.
        {
            useLights = false;  // Skip lights.
        }
        if (avg < 0.02f)  // Fully dark.
        {
            useShadows = false;  // Skip shadows.
        }
    }
    TexModulationData mod;
    Vector2f *modTexCoords = nullptr;
    if (useLights && Rend_IsMTexLights()) {
        ClientApp::renderSystem().forAllSurfaceProjections(p.lightListIdx, [&mod] (ProjectedTextureData const &dyn) {
            mod.texture     = dyn.texture;
            mod.color       = dyn.color;
            mod.topLeft     = dyn.topLeft;
            mod.bottomRight = dyn.bottomRight;
            return LoopAbort;
        });
        if (mod.texture) {
            modTexCoords = R_AllocRendTexCoords(numVerts);
            if (p.isWall) {
                modTexCoords[0] = Vector2f(mod.topLeft.x, mod.bottomRight.y);
                modTexCoords[1] = mod.topLeft;
                modTexCoords[2] = mod.bottomRight;
                modTexCoords[3] = Vector2f(mod.bottomRight.x, mod.topLeft.y);
            }
            else {
                for (duint i = 0; i < numVertices; ++i) {
                    modTexCoords[i] = (( Vector2f(*p.bottomRight) - Vector2f(rvertices[i]) ) / ( Vector2f(*p.bottomRight) - Vector2f(*p.topLeft) ) * mod.topLeft    )
                                    + (( Vector2f(rvertices[i]  ) - Vector2f(*p.topLeft  ) ) / ( Vector2f(*p.bottomRight) - Vector2f(*p.topLeft) ) * mod.bottomRight);
                }
            }
        }
    }
    bool hasDynlights = false;
    if (useLights) {
        bool const skipFirst = Rend_IsMTexLights();
        duint numProcessed = 0;
        ClientApp::renderSystem().forAllSurfaceProjections(p.lightListIdx,
                                           [&p, &mustSubdivide, &rvertices, &numVertices
                                           , &skipFirst, &numProcessed] (ProjectedTextureData const &tp) {
            if (!(skipFirst && numProcessed == 0)) {
                DrawListSpec listSpec;
                listSpec.group = LightGeom;
                listSpec.texunits[TU_PRIMARY] = GLTextureUnit(tp.texture, gl::ClampToEdge, gl::ClampToEdge);
                DrawList &lightList = ClientApp::renderSystem().drawLists().find(listSpec);
                Geometry verts;
                duint const numVerts = mustSubdivide ?   3 + p.wall.leftEdge ->divisionCount()
                                                       + 3 + p.wall.rightEdge->divisionCount()
                                                     : numVertices;
                verts.pos   = R_AllocRendVertices (numVerts);
                verts.color = R_AllocRendColors   (numVerts);
                verts.tex   = R_AllocRendTexCoords(numVerts);
                if (p.isWall) {
                    makeWallLightGeometry(verts, *p.topLeft, *p.bottomRight,
                                          numVertices, rvertices, *p.wall.leftEdge, *p.wall.rightEdge, tp);
                }
                else {
                    makeFlatLightGeometry(verts, *p.topLeft, *p.bottomRight,
                                          numVertices, rvertices, tp);
                }
                if (mustSubdivide) {
                    DENG2_ASSERT(p.isWall);
                    duint const numLeftVerts  = 3 + p.wall.leftEdge ->divisionCount();
                    duint const numRightVerts = 3 + p.wall.rightEdge->divisionCount();
                    Store &buffer = ClientApp::renderSystem().buffer(); {
                        duint base = buffer.allocateVertices(numRightVerts);
                        DrawList::reserveSpace(indices, numRightVerts);
                        for (duint i = 0; i < numRightVerts; ++i) {
                            indices[i] = base + i;
                            buffer.posCoords   [indices[i]] = verts.pos[numLeftVerts + i];
                            buffer.colorCoords [indices[i]] = (verts.color[numLeftVerts + i] * 255).toVector4ub();
                            buffer.texCoords[0][indices[i]] = verts.tex[numLeftVerts + i];
                        }
                        lightList.write(buffer, indices.constData(), numRightVerts, gl::TriangleFan);
                    } {
                        duint base = buffer.allocateVertices(numLeftVerts);
                        DrawList::reserveSpace(indices, numLeftVerts);
                        for (duint i = 0; i < numLeftVerts; ++i) {
                            indices[i] = base + i;
                            buffer.posCoords   [indices[i]] = verts.pos[i];
                            buffer.colorCoords [indices[i]] = (verts.color[i] * 255).toVector4ub();
                            buffer.texCoords[0][indices[i]] = verts.tex[i];
                        }
                        lightList.write(buffer, indices.constData(), numLeftVerts, gl::TriangleFan);
                    }
                }
                else {
                    Store &buffer = ClientApp::renderSystem().buffer();
                    duint base = buffer.allocateVertices(numVertices);
                    DrawList::reserveSpace(indices, numVertices);
                    for (duint i = 0; i < numVertices; ++i) {
                        indices[i] = base + i;
                        buffer.posCoords   [indices[i]] = verts.pos[i];
                        buffer.colorCoords [indices[i]] = (verts.color[i] * 255).toVector4ub();
                        buffer.texCoords[0][indices[i]] = verts.tex[i];
                    }
                    lightList.write(buffer, indices.constData(), numVertices, p.isWall? gl::TriangleStrip : gl::TriangleFan);
                }
                R_FreeRendVertices (verts.pos);
                R_FreeRendColors   (verts.color);
                R_FreeRendTexCoords(verts.tex);
            }
            numProcessed += 1;
            return LoopContinue;
        });
        hasDynlights = (numProcessed > duint(skipFirst? 1 : 0));
    }
    if (useShadows) {
        DrawListSpec listSpec;
        listSpec.group = ShadowGeom;
        listSpec.texunits[TU_PRIMARY] = GLTextureUnit(GL_PrepareLSTexture(LST_DYNAMIC), gl::ClampToEdge, gl::ClampToEdge);
        DrawList &shadowList = ClientApp::renderSystem().drawLists().find(listSpec);
        ClientApp::renderSystem().forAllSurfaceProjections(p.shadowListIdx,
                                           [&p, &mustSubdivide, &rvertices, &numVertices, &shadowList]
                                           (ProjectedTextureData const &tp) {
            Geometry verts;
            duint const numVerts = mustSubdivide ?   3 + p.wall.leftEdge ->divisionCount()
                                                   + 3 + p.wall.rightEdge->divisionCount()
                                                 : numVertices;
            verts.pos   = R_AllocRendVertices (numVerts);
            verts.color = R_AllocRendColors   (numVerts);
            verts.tex   = R_AllocRendTexCoords(numVerts);
            if (p.isWall) {
                makeWallShadowGeometry(verts, *p.topLeft, *p.bottomRight,
                                       numVertices, rvertices, *p.wall.leftEdge, *p.wall.rightEdge, tp);
            }
            else {
                makeFlatShadowGeometry(verts, *p.topLeft, *p.bottomRight,
                                       numVertices, rvertices, tp);
            }
            if (mustSubdivide) {
                DENG2_ASSERT(p.isWall);
                duint const numLeftVerts  = 3 + p.wall.leftEdge ->divisionCount();
                duint const numRightVerts = 3 + p.wall.rightEdge->divisionCount();
                Store &buffer = ClientApp::renderSystem().buffer(); {
                    duint base = buffer.allocateVertices(numRightVerts);
                    DrawList::reserveSpace(indices, numRightVerts);
                    for (duint i = 0; i < numRightVerts; ++i) {
                        indices[i] = base + i;
                        buffer.posCoords   [indices[i]] = verts.pos[numLeftVerts + i];
                        buffer.colorCoords [indices[i]] = (verts.color[numLeftVerts + i] * 255).toVector4ub();
                        buffer.texCoords[0][indices[i]] = verts.tex[numLeftVerts + i];
                    }
                    shadowList.write(buffer, indices.constData(), numRightVerts, gl::TriangleFan);
                } {
                    duint base = buffer.allocateVertices(numLeftVerts);
                    DrawList::reserveSpace(indices, numLeftVerts);
                    for (duint i = 0; i < numLeftVerts; ++i) {
                        indices[i] = base + i;
                        buffer.posCoords   [indices[i]] = verts.pos[i];
                        buffer.colorCoords [indices[i]] = (verts.color[i] * 255).toVector4ub();
                        buffer.texCoords[0][indices[i]] = verts.tex[i];
                    }
                    shadowList.write(buffer, indices.constData(), numLeftVerts, gl::TriangleFan);
                }
            }
            else {
                Store &buffer = ClientApp::renderSystem().buffer();
                duint base = buffer.allocateVertices(numVerts);
                DrawList::reserveSpace(indices, numVerts);
                for (duint i = 0; i < numVerts; ++i) {
                    indices[i] = base + i;
                    buffer.posCoords   [indices[i]] = verts.pos[i];
                    buffer.colorCoords [indices[i]] = (verts.color[i] * 255).toVector4ub();
                    buffer.texCoords[0][indices[i]] = verts.tex[i];
                }
                shadowList.write(buffer, indices.constData(), numVerts, p.isWall ? gl::TriangleStrip : gl::TriangleFan);
            }
            R_FreeRendVertices (verts.pos);
            R_FreeRendColors   (verts.color);
            R_FreeRendTexCoords(verts.tex);
            return LoopContinue;
        });
    }
    if (mustSubdivide) {
        DENG2_ASSERT(p.isWall);
        duint const numLeftVerts  = 3 + p.wall.leftEdge ->divisionCount();
        duint const numRightVerts = 3 + p.wall.rightEdge->divisionCount();
        Vector3f origPos[4]; std::memcpy(origPos, verts.pos, sizeof(origPos));
        R_DivVerts(verts.pos, origPos, *p.wall.leftEdge, *p.wall.rightEdge);
        if (verts.color) {
            Vector4f orig[4]; std::memcpy(orig, verts.color, sizeof(orig));
            R_DivVertColors(verts.color, orig, *p.wall.leftEdge, *p.wall.rightEdge);
        }
        if (verts.tex) {
            Vector2f orig[4]; std::memcpy(orig, verts.tex, sizeof(orig));
            R_DivTexCoords(verts.tex, orig, *p.wall.leftEdge, *p.wall.rightEdge);
        }
        if (verts.tex2) {
            Vector2f orig[4]; std::memcpy(orig, verts.tex2, sizeof(orig));
            R_DivTexCoords(verts.tex2, orig, *p.wall.leftEdge, *p.wall.rightEdge);
        }
        if (modTexCoords) {
            Vector2f orig[4]; std::memcpy(orig, modTexCoords, sizeof(orig));
            R_DivTexCoords(modTexCoords, orig, *p.wall.leftEdge, *p.wall.rightEdge);
        }
        if (p.skyMasked) {
            DrawList &skyMaskList = ClientApp::renderSystem().drawLists().find(DrawListSpec(SkyMaskGeom));
            Store &buffer = ClientApp::renderSystem().buffer(); {
                duint base = buffer.allocateVertices(numRightVerts);
                DrawList::reserveSpace(indices, numRightVerts);
                for (duint i = 0; i < numRightVerts; ++i) {
                    indices[i] = base + i;
                    buffer.posCoords[indices[i]] = verts.pos[numLeftVerts + i];
                }
                skyMaskList.write(buffer, indices.constData(), numRightVerts, gl::TriangleFan);
            } {
                duint base = buffer.allocateVertices(numLeftVerts);
                DrawList::reserveSpace(indices, numLeftVerts);
                for (duint i = 0; i < numLeftVerts; ++i) {
                    indices[i] = base + i;
                    buffer.posCoords[indices[i]] = verts.pos[i];
                }
                skyMaskList.write(buffer, indices.constData(), numLeftVerts, gl::TriangleFan);
            }
        }
        else {
            DrawListSpec listSpec((mod.texture || hasDynlights)? LitGeom : UnlitGeom);
            if (layer0RTU) {
                listSpec.texunits[TU_PRIMARY] = *layer0RTU;
                if (p.materialOrigin) {
                    listSpec.texunits[TU_PRIMARY].offset += *p.materialOrigin;
                }
                if (p.materialScale) {
                    listSpec.texunits[TU_PRIMARY].scale  *= *p.materialScale;
                    listSpec.texunits[TU_PRIMARY].offset *= *p.materialScale;
                }
            }
            if (layer0InterRTU) {
                listSpec.texunits[TU_INTER] = *layer0InterRTU;
                if (p.materialOrigin) {
                    listSpec.texunits[TU_INTER].offset += *p.materialOrigin;
                }
                if (p.materialScale) {
                    listSpec.texunits[TU_INTER].scale  *= *p.materialScale;
                    listSpec.texunits[TU_INTER].offset *= *p.materialScale;
                }
            }
            if (detailRTU) {
                listSpec.texunits[TU_PRIMARY_DETAIL] = *detailRTU;
                if (p.materialOrigin) {
                    listSpec.texunits[TU_PRIMARY_DETAIL].offset += *p.materialOrigin;
                }
            }
            if (detailInterRTU) {
                listSpec.texunits[TU_INTER_DETAIL] = *detailInterRTU;
                if (p.materialOrigin) {
                    listSpec.texunits[TU_INTER_DETAIL].offset += *p.materialOrigin;
                }
            }
            DrawList &drawList = ClientApp::renderSystem().drawLists().find(listSpec);
            Parm::Flags primFlags;
            if (mod.texture && !hasDynlights) {
                primFlags |= Parm::OneLight; //oneLight = true;  // Using modulation.
            }
            else if (mod.texture || hasDynlights) {
                primFlags |= Parm::ManyLights;
            }
            Store &buffer = ClientApp::renderSystem().buffer(); {
                duint base = buffer.allocateVertices(numRightVerts);
                DrawList::reserveSpace(indices, numRightVerts);
                static Vector4ub const white(255, 255, 255, 255);
                for (duint i = 0; i < numRightVerts; ++i) {
                    indices[i] = base + i;
                    buffer.posCoords[indices[i]] = verts.pos[numLeftVerts + i];
                    if (verts.color) {
                        buffer.colorCoords[indices[i]] = (verts.color[numLeftVerts + i] * 255).toVector4ub();
                    }
                    else {
                        buffer.colorCoords[indices[i]] = white;
                    }
                    if (verts.tex) {
                        buffer.texCoords[0][indices[i]] = verts.tex[numLeftVerts + i];
                    }
                    if (verts.tex2) {
                        buffer.texCoords[1][indices[i]] = verts.tex2[numLeftVerts + i];
                    }
                    if ((primFlags & (Parm::OneLight | Parm::ManyLights)) && Rend_IsMTexLights()) {
                        DENG2_ASSERT(modTexCoords);
                        buffer.modCoords[indices[i]] = modTexCoords[numLeftVerts + i];
                    }
                }
                drawList.write(buffer, indices.constData(), numRightVerts,
                               Parm(gl::TriangleFan,
                                    listSpec.unit(TU_PRIMARY       ).scale,
                                    listSpec.unit(TU_PRIMARY       ).offset,
                                    listSpec.unit(TU_PRIMARY_DETAIL).scale,
                                    listSpec.unit(TU_PRIMARY_DETAIL).offset,
                                    primFlags, BM_NORMAL,
                                    mod.texture, mod.color));
            } {
                duint base = buffer.allocateVertices(numLeftVerts);
                DrawList::reserveSpace(indices, numLeftVerts);
                static Vector4ub const white(255, 255, 255, 255);
                for (duint i = 0; i < numLeftVerts; ++i) {
                    indices[i] = base + i;
                    buffer.posCoords[indices[i]] = verts.pos[i];
                    if (verts.color) {
                        buffer.colorCoords[indices[i]] = (verts.color[i] * 255).toVector4ub();
                    }
                    else {
                        buffer.colorCoords[indices[i]] = white;
                    }
                    if (verts.tex) {
                        buffer.texCoords[0][indices[i]] = verts.tex[i];
                    }
                    if (verts.tex2) {
                        buffer.texCoords[1][indices[i]] = verts.tex2[i];
                    }
                    if ((primFlags & (Parm::OneLight | Parm::ManyLights)) && Rend_IsMTexLights()) {
                        DENG2_ASSERT(modTexCoords);
                        buffer.modCoords[indices[i]] = modTexCoords[i];
                    }
                }
                drawList.write(buffer, indices.constData(), numLeftVerts,
                               Parm(gl::TriangleFan,
                                    listSpec.unit(TU_PRIMARY       ).scale,
                                    listSpec.unit(TU_PRIMARY       ).offset,
                                    listSpec.unit(TU_PRIMARY_DETAIL).scale,
                                    listSpec.unit(TU_PRIMARY_DETAIL).offset,
                                    primFlags, BM_NORMAL,
                                    mod.texture, mod.color));
            }
        }
    }
    else {
        if (p.skyMasked) {
            Store &buffer = ClientApp::renderSystem().buffer();
            duint base = buffer.allocateVertices(numVerts);
            DrawList::reserveSpace(indices, numVerts);
            for (duint i = 0; i < numVerts; ++i) {
                indices[i] = base + i;
                buffer.posCoords[indices[i]] = verts.pos[i];
            }
            ClientApp::renderSystem()
                .drawLists().find(DrawListSpec(SkyMaskGeom))
                    .write(buffer, indices.constData(), numVerts, p.isWall? gl::TriangleStrip : gl::TriangleFan);
        }
        else {
            DrawListSpec listSpec((mod.texture || hasDynlights) ? LitGeom : UnlitGeom);
            if (layer0RTU) {
                listSpec.texunits[TU_PRIMARY] = *layer0RTU;
                if (p.materialOrigin) {
                    listSpec.texunits[TU_PRIMARY].offset += *p.materialOrigin;
                }
                if (p.materialScale) {
                    listSpec.texunits[TU_PRIMARY].scale  *= *p.materialScale;
                    listSpec.texunits[TU_PRIMARY].offset *= *p.materialScale;
                }
            }
            if (layer0InterRTU) {
                listSpec.texunits[TU_INTER] = *layer0InterRTU;
                if (p.materialOrigin) {
                    listSpec.texunits[TU_INTER].offset += *p.materialOrigin;
                }
                if (p.materialScale) {
                    listSpec.texunits[TU_INTER].scale  *= *p.materialScale;
                    listSpec.texunits[TU_INTER].offset *= *p.materialScale;
                }
            }
            if (detailRTU) {
                listSpec.texunits[TU_PRIMARY_DETAIL] = *detailRTU;
                if (p.materialOrigin) {
                    listSpec.texunits[TU_PRIMARY_DETAIL].offset += *p.materialOrigin;
                }
            }
            if (detailInterRTU) {
                listSpec.texunits[TU_INTER_DETAIL] = *detailInterRTU;
                if (p.materialOrigin) {
                    listSpec.texunits[TU_INTER_DETAIL].offset += *p.materialOrigin;
                }
            }
            Parm::Flags primFlags;
            if (mod.texture && !hasDynlights) {
                primFlags |= Parm::OneLight; // Using modulation.
            }
            else if (mod.texture || hasDynlights) {
                primFlags |= Parm::ManyLights; //manyLights = true;
            }
            Store &buffer = ClientApp::renderSystem().buffer();
            duint base = buffer.allocateVertices(numVertices);
            DrawList::reserveSpace(indices, numVertices);
            static Vector4ub const white(255, 255, 255, 255);
            for (duint i = 0; i < numVertices; ++i) {
                indices[i] = base + i;
                buffer.posCoords[indices[i]] = verts.pos[i];
                if (verts.color) {
                    buffer.colorCoords[indices[i]] = (verts.color[i] * 255).toVector4ub();
                }
                else {
                    buffer.colorCoords[indices[i]] = white;
                }
                if (verts.tex) {
                    buffer.texCoords[0][indices[i]] = verts.tex[i];
                }
                if (verts.tex2) {
                    buffer.texCoords[1][indices[i]] = verts.tex2[i];
                }
                if ((primFlags & (Parm::OneLight | Parm::ManyLights)) && Rend_IsMTexLights()) {
                    DENG2_ASSERT(modTexCoords);
                    buffer.modCoords[indices[i]] = modTexCoords[i];
                }
            }
            ClientApp::renderSystem()
                .drawLists().find(listSpec)
                    .write(buffer, indices.constData(), numVertices,
                           Parm(p.isWall? gl::TriangleStrip  : gl::TriangleFan,
                                listSpec.unit(TU_PRIMARY       ).scale,
                                listSpec.unit(TU_PRIMARY       ).offset,
                                listSpec.unit(TU_PRIMARY_DETAIL).scale,
                                listSpec.unit(TU_PRIMARY_DETAIL).offset,
                                primFlags, BM_NORMAL,
                                mod.texture, mod.color));
        }
    }
    if (shineRTU) {
        Geometry shineVerts = Geometry(); {
            Vector3f const &shineColor = matAnimator.shineMinColor();  // Shine strength.
            dfloat const shineOpacity  = shineRTU->opacity;
            shineVerts.color = R_AllocRendColors(numVerts);
            shineVerts.tex   = R_AllocRendTexCoords(numVerts);
            if (p.isWall) {
                makeWallShineGeometry(shineVerts, numVertices, rvertices, verts, p.wall.width,
                                      shineColor, shineOpacity);
            }
            else {
                makeFlatShineGeometry(shineVerts, numVertices, rvertices, verts,
                                      shineColor, shineOpacity);
            }
        }
        DrawListSpec listSpec(ShineGeom);
        listSpec.texunits[TU_PRIMARY] = *shineRTU;
        if (shineMaskRTU) {
            listSpec.texunits[TU_INTER] = *shineMaskRTU;
            if (p.materialOrigin) {
                listSpec.texunits[TU_INTER].offset += *p.materialOrigin;
            }
            if (p.materialScale) {
                listSpec.texunits[TU_INTER].scale  *= *p.materialScale;
                listSpec.texunits[TU_INTER].offset *= *p.materialScale;
            }
        }
        DrawList &shineList = ClientApp::renderSystem().drawLists().find(listSpec);
        Parm shineParams(gl::TriangleFan,
                         listSpec.unit(TU_INTER).scale,
                         listSpec.unit(TU_INTER).offset,
                         Vector2f(1, 1),
                         Vector2f(0, 0),
                         Parm::Unlit,
                         matAnimator.shineBlendMode());
        if (mustSubdivide) {
            DENG2_ASSERT(p.isWall);
            duint const numLeftVerts  = 3 + p.wall.leftEdge ->divisionCount();
            duint const numRightVerts = 3 + p.wall.rightEdge->divisionCount(); {
                Vector2f orig[4]; std::memcpy(orig, shineVerts.tex, sizeof(orig));
                R_DivTexCoords(shineVerts.tex, orig, *p.wall.leftEdge, *p.wall.rightEdge);
            } {
                Vector4f orig[4]; std::memcpy(orig, shineVerts.color, sizeof(orig));
                R_DivVertColors(shineVerts.color, orig, *p.wall.leftEdge, *p.wall.rightEdge);
            }
            Store &buffer = ClientApp::renderSystem().buffer(); {
                duint base = buffer.allocateVertices(numRightVerts);
                DrawList::reserveSpace(indices, numRightVerts);
                for (duint i = 0; i < numRightVerts; ++i) {
                    indices[i] = base + i;
                    buffer.posCoords   [indices[i]] = verts.pos[numLeftVerts + i];
                    buffer.colorCoords [indices[i]] = (shineVerts.color[numLeftVerts + i] * 255).toVector4ub();
                    buffer.texCoords[0][indices[i]] = shineVerts.tex[numLeftVerts + i];
                    if (shineMaskRTU) {
                        buffer.texCoords[1][indices[i]] = verts.tex[numLeftVerts + i];
                    }
                }
                shineList.write(buffer, indices.constData(), numRightVerts, shineParams);
            } {
                duint base = buffer.allocateVertices(numLeftVerts);
                DrawList::reserveSpace(indices, numLeftVerts);
                for (duint i = 0; i < numLeftVerts; ++i) {
                    indices[i] = base + i;
                    buffer.posCoords   [indices[i]] = verts.pos[i];
                    buffer.colorCoords [indices[i]] = (shineVerts.color[i] * 255).toVector4ub();
                    buffer.texCoords[0][indices[i]] = shineVerts.tex[i];
                    if (shineMaskRTU) {
                        buffer.texCoords[1][indices[i]] = verts.tex[i];
                    }
                }
                shineList.write(buffer, indices.constData(), numLeftVerts, shineParams);
            }
        }
        else {
            Store &buffer = ClientApp::renderSystem().buffer();
            duint base = buffer.allocateVertices(numVertices);
            DrawList::reserveSpace(indices, numVertices);
            for (duint i = 0; i < numVertices; ++i) {
                indices[i] = base + i;
                buffer.posCoords   [indices[i]] = verts.pos[i];
                buffer.colorCoords [indices[i]] = (shineVerts.color[i] * 255).toVector4ub();
                buffer.texCoords[0][indices[i]] = shineVerts.tex[i];
                if (shineMaskRTU) {
                    buffer.texCoords[1][indices[i]] = verts.tex[i];
                }
            }
            shineParams.type = p.isWall? gl::TriangleStrip : gl::TriangleFan;
            shineList.write(buffer, indices.constData(), numVertices, shineParams);
        }
        R_FreeRendColors   (shineVerts.color);
        R_FreeRendTexCoords(shineVerts.tex);
    }
    R_FreeRendTexCoords(modTexCoords);
    R_FreeRendVertices (verts.pos);
    R_FreeRendColors   (verts.color);
    R_FreeRendTexCoords(verts.tex);
    R_FreeRendTexCoords(verts.tex2);
    return (p.forceOpaque || skyMaskedMaterial
            || !(p.alpha < 1 || !matAnimator.isOpaque() || p.blendMode > 0));
}
static Lumobj::LightmapSemantic lightmapForSurface(Surface const &surface) {
    if (surface.parent().type() == DMU_SIDE) return Lumobj::Side;
    auto const &plane = surface.parent().as<Plane>();
    return plane.isSectorFloor() ? Lumobj::Down : Lumobj::Up;
}
static DGLuint prepareLightmap(ClientTexture *tex = nullptr) {
    if (tex) {
        if (TextureVariant *variant = tex->prepareVariant(Rend_MapSurfaceLightmapTextureSpec())) {
            return variant->glName();
        }
    }
    return GL_PrepareLSTexture(LST_DYNAMIC);
}
static bool projectDynlight(Vector3d const &topLeft, Vector3d const &bottomRight,
    Lumobj const &lum, Surface const &surface, dfloat blendFactor,
    ProjectedTextureData &projected) {
    if (blendFactor < OMNILIGHT_SURFACE_LUMINOSITY_ATTRIBUTION_MIN)
        return false;
    if (R_ViewerLumobjIsHidden(lum.indexInMap()))
        return false;
    DGLuint tex = prepareLightmap(lum.lightmap(lightmapForSurface(surface)));
    if (!tex) return false;
    Vector3d lumCenter = lum.origin();
    lumCenter.z += lum.zOffset();
    Vector3d topLeftToLum = topLeft - lumCenter;
    if (topLeftToLum.dot(surface.tangentMatrix().column(2)) > 0.f)
        return false;
    Vector3d pointOnPlane = R_ClosestPointOnPlane(surface.tangentMatrix().column(2)/*normal*/,
                                                  topLeft, lumCenter);
    coord_t distToLum = (lumCenter - pointOnPlane).length();
    if (distToLum <= 0 || distToLum > lum.radius())
        return false;
    dfloat luma = 1.5f - 1.5f * distToLum / lum.radius();
    luma *= lum.attenuation(R_ViewerLumobjDistance(lum.indexInMap()));
    if (luma * blendFactor < OMNILIGHT_SURFACE_LUMINOSITY_ATTRIBUTION_MIN)
        return false;
    Vector2f s, t;
    dfloat const scale = 1.0f / ((2.f * lum.radius()) - distToLum);
    if (!R_GenerateTexCoords(s, t, pointOnPlane, scale, scale * 1.08f,
                            topLeft, bottomRight, surface.tangentMatrix()))
        return false;
    projected = {};
    projected.texture     = tex;
    projected.topLeft     = Vector2f(s[0], t[0]);
    projected.bottomRight = Vector2f(s[1], t[1]);
    projected.color       = Vector4f(Rend_LuminousColor(lum.color(), luma), blendFactor);
    return true;
}
static bool projectPlaneGlow(Vector3d const &topLeft, Vector3d const &bottomRight,
    Plane const &plane, Vector3d const &pointOnPlane, dfloat blendFactor,
    ProjectedTextureData &projected) {
    if (blendFactor < OMNILIGHT_SURFACE_LUMINOSITY_ATTRIBUTION_MIN)
        return false;
    Vector3f color;
    dfloat intensity = plane.surface().glow(color);
    if (intensity < .05f)
        return false;
    coord_t glowHeight = Rend_PlaneGlowHeight(intensity);
    if (glowHeight < 2) return false;  // Not too small!
    dfloat bottom, top;
    if (plane.surface().normal().z < 0) {
              bottom = (pointOnPlane.z - topLeft.z) / glowHeight;
        top = bottom + (topLeft.z - bottomRight.z) / glowHeight;
    }
    else {
                 top = (bottomRight.z - pointOnPlane.z) / glowHeight;
        bottom = top + (topLeft.z - bottomRight.z) / glowHeight;
    }
    if (!(bottom <= 1 || top >= 0)) return false;
    de::zap(projected);
    projected.texture     = GL_PrepareLSTexture(LST_GRADIENT);
    projected.topLeft     = Vector2f(0, bottom);
    projected.bottomRight = Vector2f(1, top);
    projected.color       = Vector4f(Rend_LuminousColor(color, intensity), blendFactor);
    return true;
}
static bool projectShadow(Vector3d const &topLeft, Vector3d const &bottomRight,
    mobj_t const &mob, Surface const &surface, dfloat blendFactor,
    ProjectedTextureData &projected) {
    static Vector3f const black;  // shadows are black
    coord_t mobOrigin[3];
    Mobj_OriginSmoothed(const_cast<mobj_t *>(&mob), mobOrigin);
    coord_t distanceFromViewer = 0;
    if (shadowMaxDistance > 0) {
        distanceFromViewer = Rend_PointDist2D(mobOrigin);
        if (distanceFromViewer > shadowMaxDistance)
            return false;
    }
    dfloat shadowStrength = Mobj_ShadowStrength(mob) * ::shadowFactor;
    if (fogParams.usingFog) shadowStrength /= 2;
    if (shadowStrength <= 0) return false;
    coord_t shadowRadius = Mobj_ShadowRadius(mob);
    if (shadowRadius > ::shadowMaxRadius)
        shadowRadius = ::shadowMaxRadius;
    if (shadowRadius <= 0) return false;
    mobOrigin[2] -= mob.floorClip;
    if (mob.ddFlags & DDMF_BOB)
        mobOrigin[2] -= Mobj_BobOffset(mob);
    coord_t mobHeight = mob.height;
    if (!mobHeight) mobHeight = 1;
    Vector3d point = R_ClosestPointOnPlane(surface.tangentMatrix().column(2)/*normal*/,
                                           topLeft, mobOrigin);
    coord_t distFromSurface = (Vector3d(mobOrigin) - Vector3d(point)).length();
    if (distFromSurface > mob.height)
        return false;
    if (mobOrigin[2] + mob.height < point.z)
        return false;
    if (distFromSurface > shadowRadius)
        return false;
    shadowStrength *= 1.5f - 1.5f * distFromSurface / shadowRadius;
    coord_t halfMobjHeight = mobHeight / 2;
    if (distFromSurface > halfMobjHeight) {
        shadowStrength *= 1 - (distFromSurface - halfMobjHeight) / (mobHeight - halfMobjHeight);
    }
    shadowStrength *= Rend_ShadowAttenuationFactor(distanceFromViewer);
    shadowStrength *= blendFactor;
    if (shadowStrength < SHADOW_SURFACE_LUMINOSITY_ATTRIBUTION_MIN)
        return false;
    Vector2f s, t;
    dfloat const scale = 1.0f / ((2.f * shadowRadius) - distFromSurface);
    if (!R_GenerateTexCoords(s, t, point, scale, scale * 1.08f,
                            topLeft, bottomRight, surface.tangentMatrix()))
        return false;
    de::zap(projected);
    projected.texture     = GL_PrepareLSTexture(LST_DYNAMIC);
    projected.topLeft     = Vector2f(s[0], t[0]);
    projected.bottomRight = Vector2f(s[1], t[1]);
    projected.color       = Vector4f(black, shadowStrength);
    return true;
}
/**
 * @pre The coordinates of the given quad must be contained wholly within the subspoce
 * specified. This is due to an optimization within the lumobj management which separates
 * them by subspace.
 */
static void projectDynamics(Surface const &surface, dfloat glowStrength,
    Vector3d const &topLeft, Vector3d const &bottomRight,
    bool noLights, bool noShadows, bool sortLights,
    duint &lightListIdx, duint &shadowListIdx) {
    DENG2_ASSERT(curSubspace);
    if (levelFullBright) return;
    if (glowStrength >= 1) return;
    if (!noLights) {
        dfloat const blendFactor = 1;
        if (::useDynLights) {
            R_ForAllSubspaceLumContacts(*curSubspace, [&topLeft, &bottomRight, &surface
                                                      , &blendFactor, &sortLights
                                                      , &lightListIdx] (Lumobj &lum) {
                ProjectedTextureData projected;
                if (projectDynlight(topLeft, bottomRight, lum, surface, blendFactor,
                                   projected)) {
                    ClientApp::renderSystem().findSurfaceProjectionList(&lightListIdx, sortLights)
                                << projected;  // a copy is made.
                }
                return LoopContinue;
            });
        }
        if (::useGlowOnWalls && surface.parent().type() == DMU_SIDE && bottomRight.z < topLeft.z) {
            auto const &subsec = curSubspace->subsector().as<world::ClientSubsector>();
            for (dint i = 0; i < subsec.visPlaneCount(); ++i) {
                Plane const &plane = subsec.visPlane(i);
                Vector3d const pointOnPlane(subsec.center(), plane.heightSmoothed());
                ProjectedTextureData projected;
                if (projectPlaneGlow(topLeft, bottomRight, plane, pointOnPlane, blendFactor,
                                    projected)) {
                    ClientApp::renderSystem().findSurfaceProjectionList(&lightListIdx, sortLights)
                                << projected;  // a copy is made.
                }
            }
        }
    }
    if (!noShadows && ::useShadows) {
        dfloat blendFactor = 1 - glowStrength;
        if (blendFactor >= SHADOW_SURFACE_LUMINOSITY_ATTRIBUTION_MIN) {
            blendFactor = de::clamp(0.f, blendFactor, 1.f);
            R_ForAllSubspaceMobContacts(*curSubspace, [&topLeft, &bottomRight, &surface
                                                      , &blendFactor, &shadowListIdx] (mobj_t &mob) {
                ProjectedTextureData projected;
                if (projectShadow(topLeft, bottomRight, mob, surface, blendFactor,
                                 projected)) {
                    ClientApp::renderSystem().findSurfaceProjectionList(&shadowListIdx)
                                << projected;  // a copy is made.
                }
                return LoopContinue;
            });
        }
    }
}
/**
 * World light can both light and shade. Normal objects get more shade than light
 * (preventing them from ending up too bright compared to the ambient light).
 */
static bool lightWithWorldLight(Vector3d const & /*point*/, Vector3f const &ambientColor,
    bool starkLight, VectorLightData &vlight) {
    static Vector3f const worldLight(-.400891f, -.200445f, .601336f);
    vlight = {};
    vlight.direction         = worldLight;
    vlight.color             = ambientColor;
    vlight.affectedByAmbient = false;
    vlight.approxDist        = 0;
    if (starkLight) {
        vlight.lightSide = .35f;
        vlight.darkSide  = .5f;
        vlight.offset    = 0;
    }
    else {
        vlight.lightSide = .2f;
        vlight.darkSide  = .8f;
        vlight.offset    = .3f;
    }
    vlight.sourceMobj = nullptr;
    return true;
}
static bool lightWithLumobj(Vector3d const &point, Lumobj const &lum, VectorLightData &vlight) {
    Vector3d const lumCenter(lum.x(), lum.y(), lum.z() + lum.zOffset());
    ddouble const dist = M_ApproxDistance(M_ApproxDistance(lumCenter.x - point.x,
                                                           lumCenter.y - point.y),
                                          point.z - lumCenter.z);
    dfloat intensity = 0;
    if (dist < Lumobj::radiusMax()) {
        intensity = de::clamp(0.f, dfloat(1 - dist / lum.radius()) * 2, 1.f);
    }
    if (intensity < .05f) return false;
    vlight = {};
    vlight.direction         = (lumCenter - point) / dist;
    vlight.color             = lum.color() * intensity;
    vlight.affectedByAmbient = true;
    vlight.approxDist        = dist;
    vlight.lightSide         = 1;
    vlight.darkSide          = 0;
    vlight.offset            = 0;
    vlight.sourceMobj        = lum.sourceMobj();
    return true;
}
static bool lightWithPlaneGlow(Vector3d const &point, Subsector const &subsec,
    dint visPlaneIndex, VectorLightData &vlight) {
    Plane const &plane     = subsec.as<world::ClientSubsector>().visPlane(visPlaneIndex);
    Surface const &surface = plane.surface();
    Vector3f glowColor;
    dfloat intensity = surface.glow(glowColor);
    if (intensity < .05f) return false;
    coord_t const glowHeight = Rend_PlaneGlowHeight(intensity);
    if (glowHeight < 2) return false;  // Not too small!
    Vector3d const pointOnPlane(subsec.center(), plane.heightSmoothed());
    ddouble const dist = (point - pointOnPlane).dot(surface.normal());
    if (dist < 0) return false;
    intensity *= 1 - dist / glowHeight;
    if (intensity < .05f) return false;
    Vector3f const color = Rend_LuminousColor(glowColor, intensity);
    if (color == Vector3f()) return false;
    vlight = {};
    vlight.direction         = Vector3f(surface.normal().x, surface.normal().y, -surface.normal().z);
    vlight.color             = color;
    vlight.affectedByAmbient = true;
    vlight.approxDist        = dist;
    vlight.lightSide         = 1;
    vlight.darkSide          = 0;
    vlight.offset            = 0.3f;
    vlight.sourceMobj        = nullptr;
    return true;
}
duint Rend_CollectAffectingLights(Vector3d const &point, Vector3f const &ambientColor,
    ConvexSubspace *subspace, bool starkLight) {
    duint lightListIdx = 0; {
        VectorLightData vlight;
        if (lightWithWorldLight(point, ambientColor, starkLight, vlight)) {
            ClientApp::renderSystem().findVectorLightList(&lightListIdx)
                    << vlight;  // a copy is made.
        }
    }
    if (subspace) {
        R_ForAllSubspaceLumContacts(*subspace, [&point, &lightListIdx] (Lumobj &lum) {
            VectorLightData vlight;
            if (lightWithLumobj(point, lum, vlight)) {
                ClientApp::renderSystem().findVectorLightList(&lightListIdx)
                        << vlight;  // a copy is made.
            }
            return LoopContinue;
        });
        auto const &subsec = subspace->subsector().as<world::ClientSubsector>();
        for (dint i = 0; i < subsec.sector().planeCount(); ++i) {
            VectorLightData vlight;
            if (lightWithPlaneGlow(point, subsec, i, vlight)) {
                ClientApp::renderSystem().findVectorLightList(&lightListIdx)
                        << vlight;  // a copy is made.
            }
        }
    }
    return lightListIdx;
}
/**
 * Fade the specified @a opacity value to fully transparent the closer the view
 * player is to the geometry.
 *
 * @note When the viewer is close enough we should NOT try to occlude with this
 * section in the angle clipper, otherwise HOM would occur when directly on top
 * of the wall (e.g., passing through an opaque waterfall).
 *
 * @return  @c true= fading was applied (see above note), otherwise @c false.
 */
static bool applyNearFadeOpacity(WallEdge const &leftEdge, WallEdge const &rightEdge,
    dfloat &opacity) {
    if (!leftEdge.spec().flags.testFlag(WallSpec::NearFade))
        return false;
    if (Rend_EyeOrigin().y < leftEdge.bottom().z() || Rend_EyeOrigin().y > rightEdge.top().z())
        return false;
    mobj_t const *mo         = viewPlayer->publicData().mo;
    Line const &line         = leftEdge.lineSide().line();
    coord_t linePoint[2]     = { line.from().x(), line.from().y() };
    coord_t lineDirection[2] = {  line.direction().x,  line.direction().y };
    vec2d_t result;
    ddouble pos = V2d_ProjectOnLine(result, mo->origin, linePoint, lineDirection);
    if (!(pos > 0 && pos < 1))
        return false;
    coord_t const maxDistance = Mobj_Radius(*mo) * .8f;
    auto delta       = Vector2d(result) - Vector2d(mo->origin);
    coord_t distance = delta.length();
    if (de::abs(distance) > maxDistance)
        return false;
    if (distance > 0) {
        opacity = (opacity / maxDistance) * distance;
        opacity = de::clamp(0.f, opacity, 1.f);
    }
    return true;
}
/**
 * The DOOM lighting model applies a sector light level delta when drawing
 * walls based on their 2D world angle.
 *
 * @todo WallEdge should encapsulate.
 */
static dfloat wallLuminosityDeltaFromNormal(Vector3f const &normal) {
    return (1.0f / 255) * (normal.x * 18) * ::rendLightWallAngle;
}
static void wallLuminosityDeltas(WallEdge const &leftEdge, WallEdge const &rightEdge,
    dfloat luminosityDeltas[2]) {
    dfloat &leftDelta  = luminosityDeltas[0];
    dfloat &rightDelta = luminosityDeltas[1];
    if (leftEdge.spec().flags.testFlag(WallSpec::NoLightDeltas)) {
        leftDelta = rightDelta = 0;
        return;
    }
    leftDelta = wallLuminosityDeltaFromNormal(leftEdge.normal());
    if (leftEdge.normal() == rightEdge.normal()) {
        rightDelta = leftDelta;
    }
    else {
        rightDelta = wallLuminosityDeltaFromNormal(rightEdge.normal());
        coord_t const lineLength    = leftEdge.lineSide().line().length();
        coord_t const sectionOffset = leftEdge.lineSideOffset();
        coord_t const sectionWidth  = de::abs(Vector2d(rightEdge.origin() - leftEdge.origin()).length());
        dfloat deltaDiff = rightDelta - leftDelta;
        rightDelta = leftDelta + ((sectionOffset + sectionWidth) / lineLength) * deltaDiff;
        leftDelta += (sectionOffset / lineLength) * deltaDiff;
    }
}
static void writeWall(WallEdge const &leftEdge, WallEdge const &rightEdge,
    bool *retWroteOpaque = nullptr, coord_t *retBottomZ = nullptr, coord_t *retTopZ = nullptr) {
    DENG2_ASSERT(leftEdge.lineSideSegment().isFrontFacing() && leftEdge.lineSide().hasSections());
    if (retWroteOpaque) *retWroteOpaque = false;
    if (retBottomZ)     *retBottomZ     = 0;
    if (retTopZ)        *retTopZ        = 0;
    auto &subsec = curSubspace->subsector().as<world::ClientSubsector>();
    Surface &surface = leftEdge.lineSide().surface(leftEdge.spec().section);
    dfloat opacity = surface.opacity();
    if (opacity < .001f)
        return;
    ClientMaterial *material = Rend_ChooseMapSurfaceMaterial(surface);
    if (!material || !material->isDrawable())
        return;
    if (!leftEdge.isValid() || !rightEdge.isValid()
        || de::fequal(leftEdge.bottom().z(), rightEdge.top().z()))
        return;
    WallSpec const &wallSpec      = leftEdge.spec();
    bool const didNearFade        = applyNearFadeOpacity(leftEdge, rightEdge, opacity);
    bool const skyMasked          = material->isSkyMasked() && !::devRendSkyMode;
    bool const twoSidedMiddle     = (wallSpec.section == LineSide::Middle && !leftEdge.lineSide().considerOneSided());
    MaterialAnimator &matAnimator = material->getAnimator(Rend_MapSurfaceMaterialSpec());
    Vector2f const materialScale  = surface.materialScale();
    Vector3f const materialOrigin = leftEdge.materialOrigin();
    Vector3d const topLeft        = leftEdge .top   ().origin();
    Vector3d const bottomRight    = rightEdge.bottom().origin();
    rendworldpoly_params_t parm; de::zap(parm);
    parm.skyMasked            = skyMasked;
    parm.mapElement           = &leftEdge.lineSideSegment();
    parm.geomGroup            = wallSpec.section;
    parm.topLeft              = &topLeft;
    parm.bottomRight          = &bottomRight;
    parm.forceOpaque          = wallSpec.flags.testFlag(WallSpec::ForceOpaque);
    parm.alpha                = parm.forceOpaque? 1 : opacity;
    parm.surfaceTangentMatrix = &surface.tangentMatrix();
    parm.blendMode            = BM_NORMAL;
    parm.materialOrigin       = &materialOrigin;
    parm.materialScale        = &materialScale;
    parm.isWall               = true;
    parm.wall.width           = de::abs(Vector2d(rightEdge.origin() - leftEdge.origin()).length());
    parm.wall.leftEdge        = &leftEdge;
    parm.wall.rightEdge       = &rightEdge;
    wallLuminosityDeltas(leftEdge, rightEdge, parm.surfaceLuminosityDeltas);
    LineSide &side = leftEdge.lineSide();
    if (!parm.skyMasked) {
        if (glowFactor > .0001f) {
            parm.glowing = matAnimator.glowStrength();
            parm.glowing *= glowFactor;
        }
        projectDynamics(surface, parm.glowing, *parm.topLeft, *parm.bottomRight,
                        wallSpec.flags.testFlag(WallSpec::NoDynLights),
                        wallSpec.flags.testFlag(WallSpec::NoDynShadows),
                        wallSpec.flags.testFlag(WallSpec::SortDynLights),
                        parm.lightListIdx, parm.shadowListIdx);
        if (twoSidedMiddle) {
            parm.blendMode = surface.blendMode();
            if (parm.blendMode == BM_NORMAL && noSpriteTrans)
                parm.blendMode = BM_ZEROALPHA;  // "no translucency" mode
        }
        side.chooseSurfaceColors(wallSpec.section, &parm.surfaceColor, &parm.wall.surfaceColor2);
    }
    if (twoSidedMiddle && side.sectorPtr() != &subsec.sector()) {
        curSectorLightColor = Rend_AmbientLightColor(side.sector());
        curSectorLightLevel = side.sector().lightLevel();
    }
    Vector3f const posCoords[] = {
        leftEdge .bottom().origin(),
        leftEdge .top   ().origin(),
        rightEdge.bottom().origin(),
        rightEdge.top   ().origin()
    };
    bool const wroteOpaque = renderWorldPoly(posCoords, 4, parm, matAnimator);
    if (wroteOpaque && !skyMasked && !(parm.glowing > 0)) {
        Rend_DrawWallRadio(leftEdge, rightEdge, ::curSectorLightLevel);
    }
    if (twoSidedMiddle && side.sectorPtr() != &subsec.sector()) {
        Vector4f const color = subsec.lightSourceColorfIntensity();
        curSectorLightColor = color.toVector3f();
        curSectorLightLevel = color.w;
    }
    if (retWroteOpaque) *retWroteOpaque = wroteOpaque && !didNearFade;
    if (retBottomZ)     *retBottomZ     = leftEdge .bottom().z();
    if (retTopZ)        *retTopZ        = rightEdge.top   ().z();
}
/**
 * Prepare a trifan geometry according to the edges of the current subspace.
 * If a fan base HEdge has been chosen it will be used as the center of the
 * trifan, else the mid point of this leaf will be used instead.
 *
 * @param direction  Vertex winding direction.
 * @param height     Z map space height coordinate to be set for each vertex.
 * @param verts      Built position coordinates are written here. It is the
 *                   responsibility of the caller to release this storage with
 *                   @ref R_FreeRendVertices() when done.
 *
 * @return  Number of built vertices.
 */
static duint buildSubspacePlaneGeometry(ClockDirection direction, coord_t height,
    Vector3f **verts) {
    DENG2_ASSERT(verts);
    Face const &poly       = curSubspace->poly();
    HEdge *fanBase         = curSubspace->fanBase();
    duint const totalVerts = poly.hedgeCount() + (!fanBase? 2 : 0);
    *verts = R_AllocRendVertices(totalVerts);
    duint n = 0;
    if (!fanBase) {
        (*verts)[n] = Vector3f(poly.center(), height);
        n++;
    }
    HEdge *baseNode = fanBase? fanBase : poly.hedge();
    HEdge *node = baseNode;
    do {
        (*verts)[n] = Vector3f(node->origin(), height);
        n++;
    } while ((node = &node->neighbor(direction)) != baseNode);
    if (!fanBase) {
        (*verts)[n] = Vector3f(poly.hedge()->origin(), height);
    }
    return totalVerts;
}
static void writeSubspacePlane(Plane &plane) {
    Face const &poly       = curSubspace->poly();
    Surface const &surface = plane.surface();
    dfloat const opacity = surface.opacity();
    if (opacity < .001f) return;
    ClientMaterial *material = Rend_ChooseMapSurfaceMaterial(surface);
    if (!material || !material->isDrawable())
        return;
    if (!::devRendSkyMode && surface.hasSkyMaskedMaterial() && plane.indexInSector() <= Sector::Ceiling)
        return;
    MaterialAnimator &matAnimator = material->getAnimator(Rend_MapSurfaceMaterialSpec());
    Vec2f materialOrigin =
        Vec2f(fmod(curSubspace->poly().bounds().minX, material->width()),
              fmod(curSubspace->poly().bounds().maxY, material->height())) +
        surface.originSmoothed();
    if (plane.isSectorCeiling()) {
        materialOrigin.y -= poly.bounds().maxY - poly.bounds().minY;
    }
    materialOrigin.y = -materialOrigin.y;
    Vector2f const materialScale = surface.materialScale();
    Vector3d topLeft(poly.bounds().minX,
                     poly.bounds().arvec2[plane.isSectorFloor()? 1 : 0][1],
                     plane.heightSmoothed());
    Vector3d bottomRight(poly.bounds().maxX,
                         poly.bounds().arvec2[plane.isSectorFloor()? 0 : 1][1],
                         plane.heightSmoothed());
    rendworldpoly_params_t parm; de::zap(parm);
    parm.mapElement           = curSubspace;
    parm.geomGroup            = plane.indexInSector();
    parm.topLeft              = &topLeft;
    parm.bottomRight          = &bottomRight;
    parm.materialOrigin       = &materialOrigin;
    parm.materialScale        = &materialScale;
    parm.surfaceColor         = &surface.color();
    parm.surfaceTangentMatrix = &surface.tangentMatrix();
    if (material->isSkyMasked()) {
        if (devRendSkyMode) {
            parm.blendMode   = BM_NORMAL;
            parm.forceOpaque = true;
        }
        else {
            parm.skyMasked = true;
        }
    }
    else if (plane.indexInSector() <= Sector::Ceiling) {
        parm.blendMode   = BM_NORMAL;
        parm.forceOpaque = true;
    }
    else {
        parm.blendMode = surface.blendMode();
        if (parm.blendMode == BM_NORMAL && noSpriteTrans) {
            parm.blendMode = BM_ZEROALPHA;  // "no translucency" mode
        }
        parm.alpha = surface.opacity();
    }
    if (!parm.skyMasked) {
        if (glowFactor > .0001f) {
            if (material == surface.materialPtr()) {
                parm.glowing = matAnimator.glowStrength();
            }
            else {
                world::Material *actualMaterial =
                    surface.hasMaterial()? surface.materialPtr()
                                         : &world::Materials::get().material(de::Uri("System", Path("missing")));
                parm.glowing = actualMaterial->as<ClientMaterial>().getAnimator(Rend_MapSurfaceMaterialSpec()).glowStrength();
            }
            parm.glowing *= ::glowFactor;
        }
        projectDynamics(surface, parm.glowing, *parm.topLeft, *parm.bottomRight,
                        false /*do light*/, false /*do shadow*/, false /*don't sort*/,
                        parm.lightListIdx, parm.shadowListIdx);
    }
    if (&plane.sector() != &curSubspace->subsector().sector()) {
        curSectorLightColor = Rend_AmbientLightColor(plane.sector());
        curSectorLightLevel = plane.sector().lightLevel();
    }
    Vector3f *posCoords;
    duint vertCount = buildSubspacePlaneGeometry((plane.isSectorCeiling())? Anticlockwise : Clockwise,
                                                 plane.heightSmoothed(), &posCoords);
    renderWorldPoly(posCoords, vertCount, parm, matAnimator);
    if (&plane.sector() != &curSubspace->subsector().sector()) {
        Vector4f const color = curSubspace->subsector().as<world::ClientSubsector>().lightSourceColorfIntensity();
        curSectorLightColor = color.toVector3f();
        curSectorLightLevel = color.w;
    }
    R_FreeRendVertices(posCoords);
}
static void writeSkyMaskStrip(dint vertCount, Vector3f const *posCoords, Vector2f const *texCoords,
    Material *material) {
    DENG2_ASSERT(posCoords);
    static DrawList::Indices indices;
    if (!devRendSkyMode) {
        Store &buffer = ClientApp::renderSystem().buffer();
        duint base = buffer.allocateVertices(vertCount);
        DrawList::reserveSpace(indices, vertCount);
        for (dint i = 0; i < vertCount; ++i) {
            indices[i] = base + i;
            buffer.posCoords[indices[i]] = posCoords[i];
        }
        ClientApp::renderSystem().drawLists().find(DrawListSpec(SkyMaskGeom))
                      .write(buffer, indices.constData(), vertCount, gl::TriangleStrip);
    }
    else {
        DENG2_ASSERT(texCoords);
        DrawListSpec listSpec;
        listSpec.group = UnlitGeom;
        if (renderTextures != 2) {
            DENG2_ASSERT(material);
            MaterialAnimator &matAnimator = material->as<ClientMaterial>().getAnimator(Rend_MapSurfaceMaterialSpec());
            matAnimator.prepare();
            listSpec.texunits[TU_PRIMARY]        = matAnimator.texUnit(MaterialAnimator::TU_LAYER0);
            listSpec.texunits[TU_PRIMARY_DETAIL] = matAnimator.texUnit(MaterialAnimator::TU_DETAIL);
            listSpec.texunits[TU_INTER]          = matAnimator.texUnit(MaterialAnimator::TU_LAYER0_INTER);
            listSpec.texunits[TU_INTER_DETAIL]   = matAnimator.texUnit(MaterialAnimator::TU_DETAIL_INTER);
        }
        Store &buffer = ClientApp::renderSystem().buffer();
        duint base = buffer.allocateVertices(vertCount);
        DrawList::reserveSpace(indices, vertCount);
        for (dint i = 0; i < vertCount; ++i) {
            indices[i] = base + i;
            buffer.posCoords   [indices[i]] = posCoords[i];
            buffer.texCoords[0][indices[i]] = texCoords[i];
            buffer.colorCoords [indices[i]] = Vector4ub(255, 255, 255, 255);
        }
        ClientApp::renderSystem().drawLists().find(listSpec)
                      .write(buffer, indices.constData(), vertCount,
                             DrawList::PrimitiveParams(gl::TriangleStrip,
                                                       listSpec.unit(TU_PRIMARY       ).scale,
                                                       listSpec.unit(TU_PRIMARY       ).offset,
                                                       listSpec.unit(TU_PRIMARY_DETAIL).scale,
                                                       listSpec.unit(TU_PRIMARY_DETAIL).offset));
    }
}
static void writeSubspaceSkyMaskStrips(SkyFixEdge::FixType fixType) {
    ClockDirection const direction   = Clockwise;
    bool const splitOnMaterialChange = (::devRendSkyMode && ::renderTextures != 2);
    dint const relPlane = (fixType == SkyFixEdge::Upper ? Sector::Ceiling : Sector::Floor);
    TriangleStripBuilder stripBuilder(CPP_BOOL(::devRendSkyMode));
    Material *scanMaterial    = nullptr;
    dfloat scanMaterialOffset = 0;
    HEdge *scanNode           = nullptr;
    coord_t scanZBottom       = 0;
    coord_t scanZTop          = 0;
    HEdge *base  = curSubspace->poly().hedge();
    HEdge *hedge = base;
    forever {
        Material *skyMaterial = nullptr;
        if (splitOnMaterialChange) {
            skyMaterial = hedge->face().mapElementAs<ConvexSubspace>()
                              .subsector().as<world::ClientSubsector>()
                                   .visPlane(relPlane).surface().materialPtr();
        }
        if (!scanNode && hedge->hasMapElement()) {
            scanMaterialOffset = hedge->mapElementAs<LineSideSegment>().lineSideOffset();
            SkyFixEdge skyEdge(*hedge, fixType, (direction == Anticlockwise ? Line::To : Line::From),
                               scanMaterialOffset);
            if (skyEdge.isValid() && skyEdge.bottom().z() < skyEdge.top().z()) {
                stripBuilder.begin(direction);
                stripBuilder << skyEdge;
                scanNode     = hedge;
                scanZBottom  = skyEdge.bottom().z();
                scanZTop     = skyEdge.top   ().z();
                scanMaterial = skyMaterial;
            }
        }
        bool beginNewStrip = false;
        if (scanNode) {
            bool endStrip = false;
            if (hedge->hasMapElement()) {
                scanMaterialOffset += hedge->mapElementAs<LineSideSegment>().length() *
                                      (direction == Anticlockwise ? -1 : 1);
                SkyFixEdge skyEdge(*hedge, fixType, (direction == Anticlockwise)? Line::From : Line::To,
                                   scanMaterialOffset);
                if (!(skyEdge.isValid() && skyEdge.bottom().z() < skyEdge.top().z())) {
                    endStrip = true;
                }
                else if (hedge != scanNode &&
                        (   !de::fequal(skyEdge.bottom().z(), scanZBottom)
                         || !de::fequal(skyEdge.top   ().z(), scanZTop)
                         || (splitOnMaterialChange && skyMaterial != scanMaterial))) {
                    endStrip      = true;
                    beginNewStrip = true;  // We'll continue from here.
                }
                else {
                    stripBuilder << skyEdge;
                }
            }
            else {
                endStrip = true;
            }
            if (endStrip || &hedge->neighbor(direction) == base) {
                scanNode = nullptr;
                PositionBuffer *positions = nullptr;
                TexCoordBuffer *texcoords = nullptr;
                dint const numVerts = stripBuilder.take(&positions, &texcoords);
                writeSkyMaskStrip(numVerts, positions->constData(),
                                  (texcoords ? texcoords->constData() : nullptr),
                                  scanMaterial);
                delete positions;
                delete texcoords;
            }
        }
        if (beginNewStrip) continue;
        hedge = &hedge->neighbor(direction);
        if (hedge == base) break;
    }
}
/**
 * @defgroup skyCapFlags  Sky Cap Flags
 * @ingroup flags
 */
#define SKYCAP_LOWER        0x1
#define SKYCAP_UPPER        0x2
static uint makeFlatSkyMaskGeometry(DrawList::Indices &indices, Store &verts, gl::Primitive &primitive,
    ConvexSubspace const &subspace, coord_t worldZPosition = 0, ClockDirection direction = Clockwise) {
    Face const &poly = subspace.poly();
    HEdge *fanBase   = subspace.fanBase();
    duint const vertCount = poly.hedgeCount() + (!fanBase? 2 : 0);
    duint const base      = verts.allocateVertices(vertCount);
    DrawList::reserveSpace(indices, vertCount);
    for (duint i = 0; i < vertCount; ++i) {
        indices[i] = base + i;
    }
    primitive = gl::TriangleFan;
    duint n = 0;
    if (!fanBase) {
        verts.posCoords[indices[n++]] = Vector3f(poly.center(), worldZPosition);
    }
    HEdge *baseNode = fanBase? fanBase : poly.hedge();
    HEdge *node = baseNode;
    do {
        verts.posCoords[indices[n++]] = Vector3f(node->origin(), worldZPosition);
    } while ((node = &node->neighbor(direction)) != baseNode);
    if (!fanBase) {
        verts.posCoords[indices[n  ]] = Vector3f(node->origin(), worldZPosition);
    }
    return vertCount;
}
static void writeSubspaceSkyMask(dint skyCap = SKYCAP_LOWER | SKYCAP_UPPER) {
    DENG2_ASSERT(::curSubspace);
    if (!skyCap) return;
    auto &subsec = curSubspace->subsector().as<world::ClientSubsector>();
    world::Map &map = subsec.sector().map();
    DrawList &dlist = ClientApp::renderSystem().drawLists().find(DrawListSpec(SkyMaskGeom));
    static DrawList::Indices indices;
    if ((skyCap & SKYCAP_LOWER) && subsec.hasSkyFloor()) {
        world::ClSkyPlane &skyFloor = map.skyFloor();
        writeSubspaceSkyMaskStrips(SkyFixEdge::Lower);
        if (!::devRendSkyMode) {
            ddouble const height =
                P_IsInVoid(::viewPlayer) ? subsec.visFloor().heightSmoothed() : skyFloor.height();
            Store &verts = ClientApp::renderSystem().buffer();
            gl::Primitive primitive;
            uint vertCount = makeFlatSkyMaskGeometry(indices, verts, primitive, *curSubspace, height, Clockwise);
            dlist.write(verts, indices.constData(), vertCount, primitive);
        }
    }
    if ((skyCap & SKYCAP_UPPER) && subsec.hasSkyCeiling()) {
        world::ClSkyPlane &skyCeiling = map.skyCeiling();
        writeSubspaceSkyMaskStrips(SkyFixEdge::Upper);
        if (!::devRendSkyMode) {
            ddouble const height =
                P_IsInVoid(::viewPlayer) ? subsec.visCeiling().heightSmoothed() : skyCeiling.height();
            Store &verts = ClientApp::renderSystem().buffer();
            gl::Primitive primitive;
            uint vertCount = makeFlatSkyMaskGeometry(indices, verts, primitive, *curSubspace, height, Anticlockwise);
            dlist.write(verts, indices.constData(), vertCount, primitive);
        }
    }
}
static bool coveredOpenRange(HEdge &hedge, coord_t middleBottomZ, coord_t middleTopZ,
    bool wroteOpaqueMiddle) {
    LineSide const &front = hedge.mapElementAs<LineSideSegment>().lineSide();
    if (!front.hasAtLeastOneMaterial()) {
        return false;
    }
    if (auto *anim = front.middle().materialAnimator()) {
        if (!anim->isOpaque()) {
            return false;
        }
    }
    if (front.considerOneSided()) {
        return wroteOpaqueMiddle;
    }
    if (!hedge.twin().hasFace()) {
        return wroteOpaqueMiddle;
    }
    auto const &subsec     = hedge.face().mapElementAs<ConvexSubspace>().subsector().as<world::ClientSubsector>();
    auto const &backSubsec = hedge.twin().face().mapElementAs<ConvexSubspace>().subsector().as<world::ClientSubsector>();
    ddouble const ffloor   = subsec.visFloor().heightSmoothed();
    ddouble const fceil    = subsec.visCeiling().heightSmoothed();
    ddouble const bfloor   = backSubsec.visFloor().heightSmoothed();
    ddouble const bceil    = backSubsec.visCeiling().heightSmoothed();
    bool middleCoversOpening = false;
    if (wroteOpaqueMiddle) {
        ddouble xbottom = de::max(bfloor, ffloor);
        ddouble xtop    = de::min(bceil,  fceil);
        xbottom += front.middle().originSmoothed().y;
        xtop    += front.middle().originSmoothed().y;
        middleCoversOpening = (middleTopZ >= xtop && middleBottomZ <= xbottom);
    }
    if (wroteOpaqueMiddle && middleCoversOpening)
        return true;
    if (  (bceil  <= ffloor && (front.top   ().hasMaterial() || front.middle().hasMaterial()))
       || (bfloor >= fceil  && (front.bottom().hasMaterial() || front.middle().hasMaterial()))) {
        Surface const &ffloorSurface = subsec.visFloor  ().surface();
        Surface const &fceilSurface  = subsec.visCeiling().surface();
        Surface const &bfloorSurface = backSubsec.visFloor  ().surface();
        Surface const &bceilSurface  = backSubsec.visCeiling().surface();
        if (de::fequal(fceil, bfloor)) {
            return (bceil <= bfloor)
                   || !(   fceilSurface.hasSkyMaskedMaterial()
                        && bceilSurface.hasSkyMaskedMaterial());
        }
        if (de::fequal(ffloor, bceil)) {
            return (bfloor >= bceil)
                   || !(   ffloorSurface.hasSkyMaskedMaterial()
                        && bfloorSurface.hasSkyMaskedMaterial());
        }
        return true;
    }
    if (bceil <= bfloor
        || (!(bceil - bfloor > 0) && bfloor > ffloor && bceil < fceil
            && front.top().hasMaterial() && front.bottom().hasMaterial())) {
        return true;
    }
    return false;
}
static void writeAllWalls(HEdge &hedge) {
    if (!hedge.hasMapElement())
        return;
    auto &seg = hedge.mapElementAs<LineSideSegment>();
    if (!seg.isFrontFacing() || !seg.lineSide().hasSections())
        return;
    reportWallDrawn(seg.line());
    bool wroteOpaqueMiddle = false;
    coord_t middleBottomZ  = 0;
    coord_t middleTopZ     = 0;
    writeWall(WallEdge(WallSpec::fromMapSide(seg.lineSide(), LineSide::Bottom), hedge, Line::From),
              WallEdge(WallSpec::fromMapSide(seg.lineSide(), LineSide::Bottom), hedge, Line::To  ));
    writeWall(WallEdge(WallSpec::fromMapSide(seg.lineSide(), LineSide::Top),    hedge, Line::From),
              WallEdge(WallSpec::fromMapSide(seg.lineSide(), LineSide::Top),    hedge, Line::To  ));
    writeWall(WallEdge(WallSpec::fromMapSide(seg.lineSide(), LineSide::Middle), hedge, Line::From),
              WallEdge(WallSpec::fromMapSide(seg.lineSide(), LineSide::Middle), hedge, Line::To  ),
              &wroteOpaqueMiddle, &middleBottomZ, &middleTopZ);
    if (!P_IsInVoid(viewPlayer) && coveredOpenRange(hedge, middleBottomZ, middleTopZ, wroteOpaqueMiddle)) {
        if (hedge.hasMapElement()) {
            if (hedge.mapElementAs<LineSideSegment>().line().definesPolyobj()) {
                const Polyobj &poly = hedge.mapElementAs<LineSideSegment>().line().polyobj();
                if (poly.sector().ceiling().surface().hasSkyMaskedMaterial()) {
                    return;
                }
            }
        }
        ClientApp::renderSystem().angleClipper()
            .addRangeFromViewRelPoints(hedge.origin(), hedge.twin().origin());
    }
}
static void writeSubspaceWalls() {
    DENG2_ASSERT(::curSubspace);
    HEdge *base  = ::curSubspace->poly().hedge();
    DENG2_ASSERT(base);
    HEdge *hedge = base;
    do {
        writeAllWalls(*hedge);
    } while ((hedge = &hedge->next()) != base);
    ::curSubspace->forAllExtraMeshes([] (Mesh &mesh) {
        for (HEdge *hedge : mesh.hedges()) {
            writeAllWalls(*hedge);
        }
        return LoopContinue;
    });
    ::curSubspace->forAllPolyobjs([] (Polyobj &pob) {
        for (HEdge *hedge : pob.mesh().hedges()) {
            writeAllWalls(*hedge);
        }
        return LoopContinue;
    });
}
static void writeSubspaceFlats() {
    DENG2_ASSERT(::curSubspace);
    auto &subsec = ::curSubspace->subsector().as<world::ClientSubsector>();
    for (dint i = 0; i < subsec.visPlaneCount(); ++i) {
        Plane &plane = subsec.visPlane(i);
        Vector3d const pointOnPlane(subsec.center(), plane.heightSmoothed());
        if ((eyeOrigin - pointOnPlane).dot(plane.surface().normal()) < 0)
            continue;
        writeSubspacePlane(plane);
    }
}
static void markFrontFacingWalls(HEdge &hedge) {
    if (!hedge.hasMapElement()) return;
    hedge.mapElementAs<LineSideSegment>()
              .setFrontFacing(viewFacingDot(hedge.origin(), hedge.twin().origin()) >= 0);
}
static void markSubspaceFrontFacingWalls() {
    DENG2_ASSERT(::curSubspace);
    HEdge *base  = ::curSubspace->poly().hedge();
    DENG2_ASSERT(base);
    HEdge *hedge = base;
    do {
        markFrontFacingWalls(*hedge);
    } while ((hedge = &hedge->next()) != base);
    ::curSubspace->forAllExtraMeshes([] (Mesh &mesh) {
        for (HEdge *hedge : mesh.hedges()) {
            markFrontFacingWalls(*hedge);
        }
        return LoopContinue;
    });
    ::curSubspace->forAllPolyobjs([] (Polyobj &pob) {
        for (HEdge *hedge : pob.mesh().hedges()) {
            markFrontFacingWalls(*hedge);
        }
        return LoopContinue;
    });
}
static inline bool canOccludeEdgeBetweenPlanes(Plane const &frontPlane, Plane const &backPlane) {
    return !(frontPlane.surface().hasSkyMaskedMaterial() && backPlane .surface().hasSkyMaskedMaterial());
}
/**
 * Add angle clipper occlusion ranges for the edges of the current subspace.
 */
static void occludeSubspace(bool frontFacing) {
    if (devNoCulling) return;
    if (P_IsInVoid(viewPlayer)) return;
    AngleClipper &clipper = ClientApp::renderSystem().angleClipper();
    auto const &subsec = ::curSubspace->subsector().as<world::ClientSubsector>();
    HEdge const *base  = ::curSubspace->poly().hedge();
    DENG2_ASSERT(base);
    HEdge const *hedge = base;
    do {
        if (!hedge->hasMapElement())
            continue;
        auto &seg = hedge->mapElementAs<LineSideSegment>();
        if (frontFacing != seg.isFrontFacing())
            continue;
        if (!seg.lineSide().hasSections())
            continue;
        if (!hedge->hasTwin() || !hedge->twin().hasFace()
            || !hedge->twin().face().hasMapElement())
            continue;
        auto const &backSubspace = hedge->twin().face().mapElementAs<ConvexSubspace>();
        auto const &backSubsec   = backSubspace.subsector().as<world::ClientSubsector>();
        ddouble openBottom = de::max(backSubsec.visFloor().heightSmoothed(),
                                     subsec    .visFloor().heightSmoothed());
        ddouble openTop = de::min(backSubsec.visCeiling().heightSmoothed(),
                                  subsec    .visCeiling().heightSmoothed());
        Vertex const &from = frontFacing ? hedge->vertex() : hedge->twin().vertex();
        Vertex const &to   = frontFacing ? hedge->twin().vertex() : hedge->vertex();
        if ( (   (openBottom > subsec    .visFloor().heightSmoothed() && Rend_EyeOrigin().y <= openBottom)
              || (openBottom > backSubsec.visFloor().heightSmoothed() && Rend_EyeOrigin().y >= openBottom))
            && canOccludeEdgeBetweenPlanes(subsec.visFloor(), backSubsec.visFloor())) {
            clipper.addViewRelOcclusion(from.origin(), to.origin(), openBottom, false);
        }
        if ( (   (openTop < subsec    .visCeiling().heightSmoothed() && Rend_EyeOrigin().y >= openTop)
              || (openTop < backSubsec.visCeiling().heightSmoothed() && Rend_EyeOrigin().y <= openTop))
            && canOccludeEdgeBetweenPlanes(subsec.visCeiling(), backSubsec.visCeiling())) {
            clipper.addViewRelOcclusion(from.origin(), to.origin(), openTop, true);
        }
    } while ((hedge = &hedge->next()) != base);
}
static void clipSubspaceLumobjs() {
    DENG2_ASSERT(::curSubspace);
    ::curSubspace->forAllLumobjs([] (Lumobj &lob) {
        R_ViewerClipLumobj(&lob);
        return LoopContinue;
    });
}
/**
 * In the situation where a subspace contains both lumobjs and a polyobj, lumobjs
 * must be clipped more carefully. Here we check if the line of sight intersects
 * any of the polyobj half-edges facing the viewer.
 */
static void clipSubspaceLumobjsBySight() {
    DENG2_ASSERT(::curSubspace);
    if (!::curSubspace->polyobjCount())
        return;
    ::curSubspace->forAllLumobjs([] (Lumobj &lob) {
        R_ViewerClipLumobjBySight(&lob, ::curSubspace);
        return LoopContinue;
    });
}
static void clipFrontFacingWalls(HEdge &hedge) {
    if (!hedge.hasMapElement())
        return;
    auto &seg = hedge.mapElementAs<LineSideSegment>();
    if (seg.isFrontFacing()) {
        if (!ClientApp::renderSystem().angleClipper().checkRangeFromViewRelPoints(hedge.origin(), hedge.twin().origin())) {
            seg.setFrontFacing(false);
        }
    }
}
static void clipSubspaceFrontFacingWalls() {
    DENG2_ASSERT(::curSubspace);
    HEdge *base  = ::curSubspace->poly().hedge();
    DENG2_ASSERT(base);
    HEdge *hedge = base;
    do {
        clipFrontFacingWalls(*hedge);
    } while ((hedge = &hedge->next()) != base);
    ::curSubspace->forAllExtraMeshes([] (Mesh &mesh) {
        for (HEdge *hedge : mesh.hedges()) {
            clipFrontFacingWalls(*hedge);
        }
        return LoopContinue;
    });
    ::curSubspace->forAllPolyobjs([] (Polyobj &pob) {
        for (HEdge *hedge : pob.mesh().hedges()) {
            clipFrontFacingWalls(*hedge);
        }
        return LoopContinue;
    });
}
static void projectSubspaceSprites() {
    DENG2_ASSERT(::curSubspace);
    if (::curSubspace->lastSpriteProjectFrame() == R_FrameCount())
        return;  // Already added.
    R_ForAllSubspaceMobContacts(*::curSubspace, [] (mobj_t &mob) {
        auto const &subsec = ::curSubspace->subsector().as<world::ClientSubsector>();
        if (mob.addFrameCount != R_FrameCount()) {
            mob.addFrameCount = R_FrameCount();
            R_ProjectSprite(mob);
            if (subsec.visCeiling().surface().hasSkyMaskedMaterial()) {
                if (Record const *spriteRec = Mobj_SpritePtr(mob)) {
                    defn::Sprite const sprite(*spriteRec);
                    de::Uri const &viewMaterial = sprite.viewMaterial(0);
                    if (!viewMaterial.isEmpty()) {
                        if (world::Material *material = world::Materials::get().materialPtr(viewMaterial)) {
                            if (!(mob.dPlayer && (mob.dPlayer->flags & DDPF_CAMERA))
                               && mob.origin[2] <= subsec.visCeiling().heightSmoothed()
                               && mob.origin[2] >= subsec.visFloor  ().heightSmoothed()) {
                                world::ClSkyPlane &skyCeiling = subsec.sector().map().skyCeiling();
                                ddouble visibleTop = mob.origin[2] + material->height();
                                if (visibleTop > skyCeiling.height()) {
                                    skyCeiling.setHeight(visibleTop + 16/*leeway*/);
                                }
                            }
                        }
                    }
                }
            }
        }
        return LoopContinue;
    });
    ::curSubspace->setLastSpriteProjectFrame(R_FrameCount());
}
/**
 * @pre Assumes the subspace is at least partially visible.
 */
static void drawCurrentSubspace() {
    DENG2_ASSERT(curSubspace);
    Sector &sector = ::curSubspace->sector();
    R_ViewerSubspaceMarkVisible(*::curSubspace);
    markSubspaceFrontFacingWalls();
    sector.map().spreadAllContacts(::curSubspace->poly().bounds());
    Rend_DrawFlatRadio(*::curSubspace);
    occludeSubspace(false /* back facing */);
    clipSubspaceLumobjs();
    occludeSubspace(true /* front facing */);
    clipSubspaceFrontFacingWalls();
    clipSubspaceLumobjsBySight();
    if (::useParticles) {
        sector.map().forAllGeneratorsInSector(sector, [] (Generator &gen) {
            R_ViewerGeneratorMarkVisible(gen);
            return LoopContinue;
        });
    }
    projectSubspaceSprites();
    writeSubspaceSkyMask();
    writeSubspaceWalls();
    writeSubspaceFlats();
}
/**
 * Change the current subspace (updating any relevant draw state properties
 * accordingly).
 *
 * @param subspace  The new subspace to make current.
 */
static void makeCurrent(ConvexSubspace &subspace) {
    bool const subsecChanged = (!::curSubspace || ::curSubspace->subsectorPtr() != subspace.subsectorPtr());
    ::curSubspace = &subspace;
    if (subsecChanged) {
        Vector4f const color = subspace.subsector().as<world::ClientSubsector>().lightSourceColorfIntensity();
        ::curSectorLightColor = color.toVector3f();
        ::curSectorLightLevel = color.w;
    }
}
static void traverseBspTreeAndDrawSubspaces(BspTree const *bspTree) {
    DENG2_ASSERT(bspTree);
    AngleClipper const &clipper = ClientApp::renderSystem().angleClipper();
    while (!bspTree->isLeaf()) {
        auto const &bspNode = bspTree->userData()->as<BspNode>();
        dint const eyeSide  = bspNode.pointOnSide(eyeOrigin) < 0;
        traverseBspTreeAndDrawSubspaces(bspTree->childPtr(BspTree::ChildId(eyeSide)));
        if (!::firstSubspace && clipper.isFull())
            return;
        bspTree = bspTree->childPtr(BspTree::ChildId(eyeSide ^ 1));
    }
    if (ConvexSubspace *subspace = bspTree->userData()->as<BspLeaf>().subspacePtr()) {
        DENG2_ASSERT(subspace->hasSubsector());
        if (!subspace->subsector().as<world::ClientSubsector>().hasWorldVolume())
            return;
        if (!::firstSubspace && !clipper.isPolyVisible(subspace->poly()))
            return;
        makeCurrent(*subspace);
        drawCurrentSubspace();
        ::firstSubspace = false;
    }
}
/**
 * Project all the non-clipped decorations. They become regular vissprites.
 */
static void generateDecorationFlares(Map &map) {
    Vector3d const viewPos = Rend_EyeOrigin().xzy();
    map.forAllLumobjs([&viewPos] (Lumobj &lob) {
        lob.generateFlare(viewPos, R_ViewerLumobjDistance(lob.indexInMap()));
        return LoopContinue;
    });
}
MaterialAnimator *Rend_SpriteMaterialAnimator(Record const &spriteDef) {
    MaterialAnimator *matAnimator = nullptr;
    auto found = lookupSpriteMaterialAnimators.constFind(&spriteDef);
    if (found != lookupSpriteMaterialAnimators.constEnd()) {
        matAnimator = found.value();
    }
    else {
        defn::Sprite const sprite(spriteDef);
        de::Uri const &viewMaterial = sprite.viewMaterial(0);
        if (!viewMaterial.isEmpty()) {
            if (world::Material *mat = world::Materials::get().materialPtr(viewMaterial)) {
                matAnimator = &mat->as<ClientMaterial>().getAnimator(Rend_SpriteMaterialSpec());
            }
        }
        lookupSpriteMaterialAnimators.insert(&spriteDef, matAnimator);
    }
    return matAnimator;
}
ddouble Rend_VisualRadius(Record const &spriteDef) {
    if (auto *anim = Rend_SpriteMaterialAnimator(spriteDef)) {
        anim->prepare();  // Ensure we've up to date info.
        return anim->dimensions().x / 2;
    }
    return 0;
}
Lumobj *Rend_MakeLumobj(Record const &spriteDef) {
    LOG_AS("Rend_MakeLumobj");
    MaterialAnimator *matAnimator = Rend_SpriteMaterialAnimator(spriteDef); //mat->as<ClientMaterial>().getAnimator(Rend_SpriteMaterialSpec());
    if (!matAnimator) return nullptr;
    matAnimator->prepare();  // Ensure we have up-to-date info.
    TextureVariant *texture = matAnimator->texUnit(MaterialAnimator::TU_LAYER0).texture;
    if (!texture) return nullptr;  // Unloadable texture?
    auto const *pl = (pointlight_analysis_t const *)
            texture->base().analysisDataPointer(res::Texture::BrightPointAnalysis);
    if (!pl) {
        LOGDEV_RES_WARNING("Texture \"%s\" has no BrightPointAnalysis")
                << texture->base().manifest().composeUri();
        return nullptr;
    }
    return &(new Lumobj(Vector3d(), pl->brightMul, pl->color.rgb))
            ->setZOffset(-texture->base().origin().y - pl->originY * matAnimator->dimensions().y);
}
/**
 * Setup GL state for an entire rendering pass (compassing multiple lists).
 */
static void pushGLStateForPass(DrawMode mode, TexUnitMap &texUnitMap) {
    static dfloat const black[] = { 0, 0, 0, 0 };
    texUnitMap.fill(-1);
    switch (mode) {
    case DM_SKYMASK:
        GL_SelectTexUnits(0);
        DGL_Disable(DGL_ALPHA_TEST);
        DGL_Enable(DGL_DEPTH_WRITE);
        DGL_Enable(DGL_DEPTH_TEST);
        DGL_DepthFunc(DGL_LESS);
        break;
    case DM_BLENDED:
        GL_SelectTexUnits(2);
    case DM_ALL:
        texUnitMap[0] = AttributeSpec::TexCoord0;
        texUnitMap[1] = AttributeSpec::TexCoord1;
        DGL_Disable(DGL_ALPHA_TEST);
        DGL_Enable(DGL_DEPTH_WRITE);
        DGL_Enable(DGL_DEPTH_TEST);
        DGL_DepthFunc(DGL_LESS);
        if (fogParams.usingFog) {
            DGL_Enable(DGL_FOG);
        }
        DGL_Disable(DGL_BLEND);
        break;
    case DM_LIGHT_MOD_TEXTURE:
    case DM_TEXTURE_PLUS_LIGHT:
        GL_SelectTexUnits(2);
        if (mode == DM_LIGHT_MOD_TEXTURE) {
            texUnitMap[0] = AttributeSpec::ModTexCoord;
            texUnitMap[1] = AttributeSpec::TexCoord0;
            DGL_ModulateTexture(4);  // Light * texture.
        }
        else {
            texUnitMap[0] = AttributeSpec::TexCoord0;
            texUnitMap[1] = AttributeSpec::ModTexCoord;
            DGL_ModulateTexture(5);  // Texture + light.
        }
        DGL_Disable(DGL_ALPHA_TEST);
        DGL_Enable(DGL_DEPTH_WRITE);
        DGL_Enable(DGL_DEPTH_TEST);
        DGL_DepthFunc(DGL_LESS);
        if (fogParams.usingFog) {
            DGL_Enable(DGL_FOG);
        }
        DGL_Disable(DGL_BLEND);
        break;
    case DM_FIRST_LIGHT:
        GL_SelectTexUnits(1);
        texUnitMap[0] = AttributeSpec::ModTexCoord;
        DGL_ModulateTexture(6);
        DGL_Disable(DGL_ALPHA_TEST);
        DGL_Enable(DGL_DEPTH_WRITE);
        DGL_Enable(DGL_DEPTH_TEST);
        DGL_DepthFunc(DGL_LESS);
        DGL_Disable(DGL_BLEND);
        break;
    case DM_BLENDED_FIRST_LIGHT:
        GL_SelectTexUnits(1);
        texUnitMap[0] = AttributeSpec::ModTexCoord;
        DGL_ModulateTexture(7);  // Add light, no color.
        DGL_Enable(DGL_ALPHA_TEST);
        DGL_SetFloat(DGL_ALPHA_LIMIT, 1 / 255.0f);
        DGL_Disable(DGL_DEPTH_WRITE);
        DGL_Enable(DGL_DEPTH_TEST);
        DGL_DepthFunc(DGL_LEQUAL);
        DGL_Enable(DGL_BLEND);
        DGL_BlendFunc(DGL_ONE, DGL_ONE);
        break;
    case DM_WITHOUT_TEXTURE:
        GL_SelectTexUnits(0);
        DGL_ModulateTexture(1);
        DGL_Disable(DGL_ALPHA_TEST);
        DGL_Enable(DGL_DEPTH_WRITE);
        DGL_Enable(DGL_DEPTH_TEST);
        DGL_DepthFunc(DGL_LESS);
        DGL_Disable(DGL_BLEND);
        break;
    case DM_LIGHTS:
        GL_SelectTexUnits(1);
        texUnitMap[0] = AttributeSpec::TexCoord0;
        DGL_ModulateTexture(1);
        DGL_Enable(DGL_ALPHA_TEST);
        DGL_SetFloat(DGL_ALPHA_LIMIT, 1 / 255.0f);
        DGL_Disable(DGL_DEPTH_WRITE);
        DGL_Enable(DGL_DEPTH_TEST);
        DGL_DepthFunc(DGL_LEQUAL);
        if (fogParams.usingFog) {
            DGL_Enable(DGL_FOG);
            DGL_Fogfv(DGL_FOG_COLOR, black);
        }
        DGL_Enable(DGL_BLEND);
        GL_BlendMode(BM_ADD);
        break;
    case DM_MOD_TEXTURE:
    case DM_MOD_TEXTURE_MANY_LIGHTS:
    case DM_BLENDED_MOD_TEXTURE:
        texUnitMap[0] = AttributeSpec::TexCoord0;
        texUnitMap[1] = AttributeSpec::TexCoord1;
        DGL_Disable(DGL_ALPHA_TEST);
        DGL_Disable(DGL_DEPTH_WRITE);
        DGL_Enable(DGL_DEPTH_TEST);
        DGL_DepthFunc(DGL_LEQUAL);
        DGL_Enable(DGL_BLEND);
        DGL_BlendFunc(DGL_DST_COLOR, DGL_ZERO);
        break;
    case DM_UNBLENDED_TEXTURE_AND_DETAIL:
        texUnitMap[0] = AttributeSpec::TexCoord0;
        texUnitMap[1] = AttributeSpec::TexCoord0;
        DGL_Disable(DGL_ALPHA_TEST);
        DGL_Enable(DGL_DEPTH_WRITE);
        DGL_Enable(DGL_DEPTH_TEST);
        DGL_DepthFunc(DGL_LESS);
        DGL_Disable(DGL_BLEND);
        if (fogParams.usingFog) {
            DGL_Enable(DGL_FOG);
        }
        break;
    case DM_UNBLENDED_MOD_TEXTURE_AND_DETAIL:
        texUnitMap[0] = AttributeSpec::TexCoord0;
        texUnitMap[1] = AttributeSpec::TexCoord0;
        DGL_Disable(DGL_ALPHA_TEST);
        DGL_Disable(DGL_DEPTH_WRITE);
        DGL_Enable(DGL_DEPTH_TEST);
        DGL_DepthFunc(DGL_LEQUAL);
        DGL_Enable(DGL_BLEND);
        DGL_BlendFunc(DGL_DST_COLOR, DGL_ZERO);
        break;
    case DM_ALL_DETAILS:
        GL_SelectTexUnits(1);
        texUnitMap[0] = AttributeSpec::TexCoord0;
        DGL_ModulateTexture(0);
        DGL_Disable(DGL_ALPHA_TEST);
        DGL_Disable(DGL_DEPTH_WRITE);
        DGL_Enable(DGL_DEPTH_TEST);
        DGL_DepthFunc(DGL_LEQUAL);
        DGL_Enable(DGL_BLEND);
        DGL_BlendFunc(DGL_DST_COLOR, DGL_SRC_COLOR);
        if (fogParams.usingFog) {
            DGL_Enable(DGL_FOG);
            dfloat const midGray[] = { .5f, .5f, .5f, fogParams.fogColor[3] };  // The alpha is probably meaningless?
            DGL_Fogfv(DGL_FOG_COLOR, midGray);
        }
        break;
    case DM_BLENDED_DETAILS:
        GL_SelectTexUnits(2);
        texUnitMap[0] = AttributeSpec::TexCoord0;
        texUnitMap[1] = AttributeSpec::TexCoord1;
        DGL_ModulateTexture(3);
        DGL_Disable(DGL_ALPHA_TEST);
        DGL_Disable(DGL_DEPTH_WRITE);
        DGL_Enable(DGL_DEPTH_TEST);
        DGL_DepthFunc(DGL_LEQUAL);
        DGL_Enable(DGL_BLEND);
        DGL_BlendFunc(DGL_DST_COLOR, DGL_SRC_COLOR);
        if (fogParams.usingFog) {
            DGL_Enable(DGL_FOG);
            dfloat const midGray[] = { .5f, .5f, .5f, fogParams.fogColor[3] };  // The alpha is probably meaningless?
            DGL_Fogfv(DGL_FOG_COLOR, midGray);
        }
        break;
    case DM_SHADOW:
        GL_SelectTexUnits(1);
        texUnitMap[0] = AttributeSpec::TexCoord0;
        DGL_ModulateTexture(1);
        DGL_Enable(DGL_ALPHA_TEST);
        DGL_SetFloat(DGL_ALPHA_LIMIT, 1 / 255.0f);
        DGL_Disable(DGL_DEPTH_WRITE);
        DGL_Enable(DGL_DEPTH_TEST);
        DGL_DepthFunc(DGL_LEQUAL);
        if (fogParams.usingFog) {
            DGL_Enable(DGL_FOG);
            DGL_Fogfv(DGL_FOG_COLOR, fogParams.fogColor);
        }
        DGL_Enable(DGL_BLEND);
        GL_BlendMode(BM_NORMAL);
        break;
    case DM_SHINY:
        GL_SelectTexUnits(1);
        texUnitMap[0] = AttributeSpec::TexCoord0;
        DGL_ModulateTexture(1);  // 8 for multitexture
        DGL_Disable(DGL_ALPHA_TEST);
        DGL_Disable(DGL_DEPTH_WRITE);
        DGL_Enable(DGL_DEPTH_TEST);
        DGL_DepthFunc(DGL_LEQUAL);
        if (fogParams.usingFog) {
            DGL_Enable(DGL_FOG);
            DGL_Fogfv(DGL_FOG_COLOR, black);
        }
        DGL_Enable(DGL_BLEND);
        GL_BlendMode(BM_ADD);  // Purely additive.
        break;
    case DM_MASKED_SHINY:
        GL_SelectTexUnits(2);
        texUnitMap[0] = AttributeSpec::TexCoord0;
        texUnitMap[1] = AttributeSpec::TexCoord1;  // the mask
        DGL_ModulateTexture(8);  // same as with details
        DGL_Disable(DGL_ALPHA_TEST);
        DGL_Disable(DGL_DEPTH_WRITE);
        DGL_Enable(DGL_DEPTH_TEST);
        DGL_DepthFunc(DGL_LEQUAL);
        if (fogParams.usingFog) {
            DGL_Enable(DGL_FOG);
            DGL_Fogfv(DGL_FOG_COLOR, black);
        }
        DGL_Enable(DGL_BLEND);
        GL_BlendMode(BM_ADD);  // Purely additive.
        break;
    default: break;
    }
}
static void popGLStateForPass(DrawMode mode) {
    switch (mode) {
    default: break;
    case DM_SKYMASK:
        GL_SelectTexUnits(1);
        DGL_Enable(DGL_ALPHA_TEST);
        DGL_Disable(DGL_DEPTH_TEST);
        break;
    case DM_BLENDED:
        GL_SelectTexUnits(1);
    case DM_ALL:
        DGL_Enable(DGL_ALPHA_TEST);
        DGL_Disable(DGL_DEPTH_TEST);
        if (fogParams.usingFog) {
            DGL_Disable(DGL_FOG);
        }
        DGL_Enable(DGL_BLEND);
        break;
    case DM_LIGHT_MOD_TEXTURE:
    case DM_TEXTURE_PLUS_LIGHT:
        GL_SelectTexUnits(1);
        DGL_ModulateTexture(1);
        DGL_Enable(DGL_ALPHA_TEST);
        DGL_Disable(DGL_DEPTH_TEST);
        if (fogParams.usingFog) {
            DGL_Disable(DGL_FOG);
        }
        DGL_Enable(DGL_BLEND);
        break;
    case DM_FIRST_LIGHT:
        DGL_ModulateTexture(1);
        DGL_Enable(DGL_ALPHA_TEST);
        DGL_Disable(DGL_DEPTH_TEST);
        DGL_Enable(DGL_BLEND);
        break;
    case DM_BLENDED_FIRST_LIGHT:
        DGL_ModulateTexture(1);
        DGL_Disable(DGL_DEPTH_TEST);
        DGL_BlendFunc(DGL_SRC_ALPHA, DGL_ONE_MINUS_SRC_ALPHA);
        break;
    case DM_WITHOUT_TEXTURE:
        GL_SelectTexUnits(1);
        DGL_ModulateTexture(1);
        DGL_Enable(DGL_ALPHA_TEST);
        DGL_Disable(DGL_DEPTH_TEST);
        DGL_Enable(DGL_BLEND);
        break;
    case DM_LIGHTS:
        DGL_Disable(DGL_DEPTH_TEST);
        if (fogParams.usingFog) {
            DGL_Disable(DGL_FOG);
        }
        GL_BlendMode(BM_NORMAL);
        break;
    case DM_MOD_TEXTURE:
    case DM_MOD_TEXTURE_MANY_LIGHTS:
    case DM_BLENDED_MOD_TEXTURE:
        DGL_Enable(DGL_ALPHA_TEST);
        DGL_Disable(DGL_DEPTH_TEST);
        DGL_BlendFunc(DGL_SRC_ALPHA, DGL_ONE_MINUS_SRC_ALPHA);
        break;
    case DM_UNBLENDED_TEXTURE_AND_DETAIL:
        DGL_Enable(DGL_ALPHA_TEST);
        DGL_Disable(DGL_DEPTH_TEST);
        DGL_Enable(DGL_BLEND);
        if (fogParams.usingFog) {
            DGL_Disable(DGL_FOG);
        }
        break;
    case DM_UNBLENDED_MOD_TEXTURE_AND_DETAIL:
        DGL_Enable(DGL_ALPHA_TEST);
        DGL_Disable(DGL_DEPTH_TEST);
        DGL_BlendFunc(DGL_SRC_ALPHA, DGL_ONE_MINUS_SRC_ALPHA);
        break;
    case DM_ALL_DETAILS:
        DGL_ModulateTexture(1);
        DGL_Enable(DGL_ALPHA_TEST);
        DGL_Disable(DGL_DEPTH_TEST);
        DGL_BlendFunc(DGL_SRC_ALPHA, DGL_ONE_MINUS_SRC_ALPHA);
        if (fogParams.usingFog) {
            DGL_Disable(DGL_FOG);
        }
        break;
    case DM_BLENDED_DETAILS:
        GL_SelectTexUnits(1);
        DGL_ModulateTexture(1);
        DGL_Enable(DGL_ALPHA_TEST);
        DGL_Disable(DGL_DEPTH_TEST);
        DGL_BlendFunc(DGL_SRC_ALPHA, DGL_ONE_MINUS_SRC_ALPHA);
        if (fogParams.usingFog) {
            DGL_Disable(DGL_FOG);
        }
        break;
    case DM_SHADOW:
        DGL_Disable(DGL_DEPTH_TEST);
        if (fogParams.usingFog) {
            DGL_Disable(DGL_FOG);
        }
        break;
    case DM_SHINY:
        DGL_Enable(DGL_ALPHA_TEST);
        DGL_Disable(DGL_DEPTH_TEST);
        if (fogParams.usingFog) {
            DGL_Disable(DGL_FOG);
        }
        GL_BlendMode(BM_NORMAL);
        break;
    case DM_MASKED_SHINY:
        GL_SelectTexUnits(1);
        DGL_ModulateTexture(1);
        DGL_Enable(DGL_ALPHA_TEST);
        DGL_Disable(DGL_DEPTH_TEST);
        if (fogParams.usingFog) {
            DGL_Disable(DGL_FOG);
        }
        GL_BlendMode(BM_NORMAL);
        break;
    }
}
static void drawLists(DrawLists::FoundLists const &lists, DrawMode mode) {
    if (lists.isEmpty()) return;
    if (lists.at(0)->isEmpty()) return;
    TexUnitMap texUnitMap;
    pushGLStateForPass(mode, texUnitMap);
    for (dint i = 0; i < lists.count(); ++i) {
        lists.at(i)->draw(mode, texUnitMap);
    }
    popGLStateForPass(mode);
}
static void drawSky() {
    DrawLists::FoundLists lists;
    ClientApp::renderSystem().drawLists().findAll(SkyMaskGeom, lists);
    if (!devRendSkyAlways && lists.isEmpty()) {
        return;
    }
    DGL_PushState();
    DGL_Disable(DGL_DEPTH_TEST);
    DGL_Disable(DGL_DEPTH_WRITE);
    GLState::current().setColorMask(gl::WriteNone);
    LIBGUI_GL.glEnable(GL_STENCIL_TEST);
    LIBGUI_GL.glStencilOp(GL_REPLACE, GL_REPLACE, GL_REPLACE);
    LIBGUI_GL.glStencilFunc(GL_ALWAYS, 1, 0xffffffff);
    if (!devRendSkyAlways) {
        drawLists(lists, DM_SKYMASK);
    }
    else {
        LIBGUI_GL.glClearStencil(1);
        LIBGUI_GL.glClear(GL_STENCIL_BUFFER_BIT);
    }
    DGL_PopState();
    LIBGUI_GL.glDisable(GL_STENCIL_TEST);
    LIBGUI_GL.glEnable(GL_STENCIL_TEST);
    LIBGUI_GL.glStencilFunc(GL_EQUAL, 1, 0xffffffff);
    LIBGUI_GL.glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
    ClientApp::renderSystem().sky().draw(&ClientApp::world().map().skyAnimator());
    if (!devRendSkyAlways) {
        LIBGUI_GL.glClearStencil(0);
    }
    LIBGUI_GL.glDisable(GL_STENCIL_TEST);
}
static bool generateHaloForVisSprite(vissprite_t const *spr, bool primary = false) {
    DENG2_ASSERT(spr);
    if (primary && (spr->data.flare.flags & RFF_NO_PRIMARY)) {
        return false;
    }
    dfloat occlusionFactor;
    if (spr->data.flare.isDecoration) {
        occlusionFactor = R_ViewerLumobjIsClipped(spr->data.flare.lumIdx)? 0 : 1;
    }
    else {
        occlusionFactor = (spr->data.flare.factor & 0x7f) / 127.0f;
    }
    return H_RenderHalo(spr->pose.origin,
                        spr->data.flare.size,
                        spr->data.flare.tex,
                        spr->data.flare.color,
                        spr->pose.distance,
                        occlusionFactor, spr->data.flare.mul,
                        spr->data.flare.xOff, primary,
                        (spr->data.flare.flags & RFF_NO_TURN) == 0);
}
/**
 * Render sprites, 3D models, masked wall segments and halos, ordered back to
 * front. Halos are rendered with Z-buffer tests and writes disabled, so they
 * don't go into walls or interfere with real objects. It means that halos can
 * be partly occluded by objects that are closer to the viewpoint, but that's
 * the price to pay for not having access to the actual Z-buffer per-pixel depth
 * information. The other option would be for halos to shine through masked walls,
 * sprites and models, which looks even worse. (Plus, they are *halos*, not real
 * lens flares...)
 */
static void drawMasked() {
    if (::devNoSprites) return;
    R_SortVisSprites();
    if (::visSpriteP && ::visSpriteP > ::visSprites) {
        bool primaryHaloDrawn = false;
        for (vissprite_t *spr = ::visSprSortedHead.next; spr != &::visSprSortedHead; spr = spr->next) {
            switch (spr->type) {
            default: break;
            case VSPR_MASKED_WALL:
                Rend_DrawMaskedWall(spr->data.wall);
                break;
            case VSPR_SPRITE:
                Rend_DrawSprite(*spr);
                break;
            case VSPR_MODEL:
                Rend_DrawModel(*spr);
                break;
            case VSPR_MODELDRAWABLE:
                DGL_Flush();
                ClientApp::renderSystem().modelRenderer().render(*spr);
                break;
            case VSPR_FLARE:
                if (generateHaloForVisSprite(spr, true)) {
                    primaryHaloDrawn = true;
                }
                break;
            }
        }
        if (primaryHaloDrawn && ::haloMode > 1) {
            H_SetupState(true);
            for (vissprite_t *spr = ::visSprSortedHead.next; spr != &::visSprSortedHead; spr = spr->next) {
                if (spr->type == VSPR_FLARE) {
                    generateHaloForVisSprite(spr);
                }
            }
            H_SetupState(false);
        }
    }
}
/**
 * We have several different paths to accommodate both multitextured details and
 * dynamic lights. Details take precedence (they always cover entire primitives
 * and usually *all* of the surfaces in a scene).
 */
static void drawAllLists(Map &map) {
    DENG2_ASSERT(!Sys_GLCheckError());
    DENG2_ASSERT_IN_RENDER_THREAD();
    DENG_ASSERT_GL_CONTEXT_ACTIVE();
    drawSky();
    DrawLists::FoundLists lists;
    ClientApp::renderSystem().drawLists().findAll(UnlitGeom, lists);
    if (IS_MTEX_DETAILS) {
        drawLists(lists, DM_UNBLENDED_TEXTURE_AND_DETAIL);
        drawLists(lists, DM_BLENDED);
    }
    else {
        drawLists(lists, DM_ALL);
    }
    ClientApp::renderSystem().drawLists().findAll(LitGeom, lists);
    if (IS_MTEX_LIGHTS && dynlightBlend != 2) {
        if (IS_MUL) {
            drawLists(lists, DM_LIGHT_MOD_TEXTURE);
            drawLists(lists, DM_FIRST_LIGHT);
        }
        else // Additive ('foggy') lights.
        {
            drawLists(lists, DM_TEXTURE_PLUS_LIGHT);
            drawLists(lists, DM_BLENDED);
            drawLists(lists, DM_BLENDED_FIRST_LIGHT);
        }
    }
    else // Multitexturing is not available for lights.
    {
        if (IS_MUL) {
            drawLists(lists, DM_WITHOUT_TEXTURE);
        }
        else {
            if (IS_MTEX_DETAILS) // Draw detail textures using multitexturing.
            {
                drawLists(lists, DM_UNBLENDED_TEXTURE_AND_DETAIL);
                drawLists(lists, DM_BLENDED);
                drawLists(lists, DM_BLENDED_DETAILS);
            }
            else {
                drawLists(lists, DM_ALL);
            }
        }
    }
    if (dynlightBlend != 2) {
        ClientApp::renderSystem().drawLists().findAll(LightGeom, lists);
        drawLists(lists, DM_LIGHTS);
    }
    if (IS_MUL) {
        ClientApp::renderSystem().drawLists().findAll(LitGeom, lists);
        if (IS_MTEX_DETAILS) {
            drawLists(lists, DM_UNBLENDED_MOD_TEXTURE_AND_DETAIL);
            drawLists(lists, DM_BLENDED_MOD_TEXTURE);
            drawLists(lists, DM_BLENDED_DETAILS);
        }
        else {
            if (IS_MTEX_LIGHTS && dynlightBlend != 2) {
                drawLists(lists, DM_MOD_TEXTURE_MANY_LIGHTS);
            }
            else {
                drawLists(lists, DM_MOD_TEXTURE);
            }
        }
    }
    if (r_detail) {
        ClientApp::renderSystem().drawLists().findAll(UnlitGeom, lists);
        if (IS_MTEX_DETAILS) {
            drawLists(lists, DM_BLENDED_DETAILS);
        }
        else {
            drawLists(lists, DM_ALL_DETAILS);
            ClientApp::renderSystem().drawLists().findAll(LitGeom, lists);
            drawLists(lists, DM_ALL_DETAILS);
        }
    }
    ClientApp::renderSystem().drawLists().findAll(ShineGeom, lists);
    drawLists(lists, DM_SHINY);
    drawLists(lists, DM_MASKED_SHINY);
    dint const oldRenderTextures = renderTextures;
    renderTextures = true;
    ClientApp::renderSystem().drawLists().findAll(ShadowGeom, lists);
    drawLists(lists, DM_SHADOW);
    renderTextures = oldRenderTextures;
    DGL_Disable(DGL_TEXTURE_2D);
    DGL_Enable(DGL_BLEND);
    DGL_Enable(DGL_DEPTH_WRITE);
    DGL_Enable(DGL_DEPTH_TEST);
    DGL_DepthFunc(DGL_LESS);
    DGL_Enable(DGL_ALPHA_TEST);
    DGL_SetFloat(DGL_ALPHA_LIMIT, 0);
    if (fogParams.usingFog) {
        DGL_Enable(DGL_FOG);
        DGL_Fogfv(DGL_FOG_COLOR, fogParams.fogColor);
    }
    drawMasked();
    Rend_RenderParticles(map);
    if (fogParams.usingFog) {
        DGL_Disable(DGL_FOG);
    }
    DENG2_ASSERT(!Sys_GLCheckError());
}
void Rend_RenderMap(Map &map) {
    Rend_ModelViewMatrix();
    if (!freezeRLs) {
        ClientApp::renderSystem().beginFrame();
        generateDecorationFlares(map);
        viewdata_t const *viewData = &viewPlayer->viewport();
        eyeOrigin = viewData->current.origin;
        if (vpitch <= 90 - yfov / 2 && vpitch >= -90 + yfov / 2) {
            AngleClipper &clipper = ClientApp::renderSystem().angleClipper();
            dfloat a = de::abs(vpitch) / (90 - yfov / 2);
            binangle_t startAngle = binangle_t(BANG_45 * Rend_FieldOfView() / 90) * (1 + a);
            binangle_t angLen = BANG_180 - startAngle;
            binangle_t viewside = (viewData->current.angle() >> (32 - BAMS_BITS)) + startAngle;
            clipper.safeAddRange(viewside, viewside + angLen);
            clipper.safeAddRange(viewside + angLen, viewside + 2 * angLen);
        }
        viewsidex = -viewData->viewSin;
        viewsidey = viewData->viewCos;
        firstSubspace = true;
        curSubspace = nullptr;
        traverseBspTreeAndDrawSubspaces(&map.bspTree());
    }
    drawAllLists(map);
    drawSurfaceTangentVectors(map);
    drawLumobjs(map);
    drawMobjBoundingBoxes(map);
    drawSectors(map);
    drawVertexes(map);
    drawThinkers(map);
    drawSoundEmitters(map);
    drawGenerators(map);
    DGL_Flush();
}
#if 0
static void drawStar(Vector3d const &origin, dfloat size, Vector4f const &color) {
    static dfloat const black[] = { 0, 0, 0, 0 };
    DGL_Begin(DGL_LINES);
        DGL_Color4fv(black);
        DGL_Vertex3f(origin.x - size, origin.z, origin.y);
        DGL_Color4f(color.x, color.y, color.z, color.w);
        DGL_Vertex3f(origin.x, origin.z, origin.y);
        DGL_Vertex3f(origin.x, origin.z, origin.y);
        DGL_Color4fv(black);
        DGL_Vertex3f(origin.x + size, origin.z, origin.y);
        DGL_Vertex3f(origin.x, origin.z - size, origin.y);
        DGL_Color4f(color.x, color.y, color.z, color.w);
        DGL_Vertex3f(origin.x, origin.z, origin.y);
        DGL_Vertex3f(origin.x, origin.z, origin.y);
        DGL_Color4fv(black);
        DGL_Vertex3f(origin.x, origin.z + size, origin.y);
        DGL_Vertex3f(origin.x, origin.z, origin.y - size);
        DGL_Color4f(color.x, color.y, color.z, color.w);
        DGL_Vertex3f(origin.x, origin.z, origin.y);
        DGL_Vertex3f(origin.x, origin.z, origin.y);
        DGL_Color4fv(black);
        DGL_Vertex3f(origin.x, origin.z, origin.y + size);
    DGL_End();
}
#endif
static void drawLabel(String const &label, Vector3d const &origin, dfloat scale, dfloat opacity) {
    if (label.isEmpty()) return;
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    DGL_Translatef(origin.x, origin.z, origin.y);
    DGL_Rotatef(-vang + 180, 0, 1, 0);
    DGL_Rotatef(vpitch, 1, 0, 0);
    DGL_Scalef(-scale, -scale, 1);
    Point2Raw offset = {{{2, 2}}};
    UI_TextOutEx(label.toUtf8().constData(), &offset, UI_Color(UIC_TITLE), opacity);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
}
static void drawLabel(String const &label, Vector3d const &origin, ddouble maxDistance = 2000) {
    ddouble const distToEye = (Rend_EyeOrigin().xzy() - origin).length();
    if (distToEye < maxDistance) {
        drawLabel(label, origin, distToEye / (DENG_GAMEVIEW_WIDTH / 2), 1 - distToEye / maxDistance);
    }
}
void Rend_UpdateLightModMatrix() {
    if (novideo) return;
    de::zap(lightModRange);
    if (!App_World().hasMap()) {
        rAmbient = 0;
        return;
    }
    dint mapAmbient = App_World().map().ambientLightLevel();
    if (mapAmbient > ambientLight) {
        rAmbient = mapAmbient;
    }
    else {
        rAmbient = ambientLight;
    }
    for (dint i = 0; i < 255; ++i) {
        dfloat lightlevel = 0;
        if (lightRangeCompression != 0) {
            if (lightRangeCompression >= 0) {
                lightlevel = dfloat(255 - i) * lightRangeCompression;
            }
            else {
                lightlevel = dfloat(-i) * -lightRangeCompression;
            }
        }
        if (rAmbient != 0 && i+lightlevel <= rAmbient) {
            lightlevel = rAmbient - i;
        }
        if ((i + lightlevel) >= 255) {
            lightlevel = 255 - i;
        }
        else if ((i + lightlevel) <= 0) {
            lightlevel = -i;
        }
        lightModRange[i] = lightlevel / 255.0f;
        DENG2_ASSERT(INRANGE_OF(i / 255.0f + lightModRange[i], 0.f, 1.f));
    }
}
dfloat Rend_LightAdaptationDelta(dfloat val) {
    dint clampedVal = de::clamp(0, de::roundi(255.0f * val), 254);
    return lightModRange[clampedVal];
}
void Rend_ApplyLightAdaptation(dfloat &val) {
    val += Rend_LightAdaptationDelta(val);
}
void Rend_DrawLightModMatrix() {
#define BLOCK_WIDTH             ( 1.0f )
#define BLOCK_HEIGHT            ( BLOCK_WIDTH * 255.0f )
#define BORDER                  ( 20 )
    if (!devLightModRange) return;
    DGL_MatrixMode(DGL_PROJECTION);
    DGL_PushMatrix();
    DGL_LoadIdentity();
    DGL_Ortho(0, 0, DENG_GAMEVIEW_WIDTH, DENG_GAMEVIEW_HEIGHT, -1, 1);
    DGL_Translatef(BORDER, BORDER, 0);
    DGL_Color4f(1, 1, 0, 1);
    DGL_Begin(DGL_LINES);
        DGL_Vertex2f(-1, -1);
        DGL_Vertex2f(255 + 1, -1);
        DGL_Vertex2f(255 + 1, -1);
        DGL_Vertex2f(255 + 1, BLOCK_HEIGHT + 1);
        DGL_Vertex2f(255 + 1, BLOCK_HEIGHT + 1);
        DGL_Vertex2f(-1, BLOCK_HEIGHT + 1);
        DGL_Vertex2f(-1, BLOCK_HEIGHT + 1);
        DGL_Vertex2f(-1, -1);
    DGL_End();
    DGL_Begin(DGL_QUADS);
    dfloat c = 0;
    for (dint i = 0; i < 255; ++i, c += (1.0f/255.0f)) {
        dfloat off = lightModRange[i];
        DGL_Color4f(c + off, c + off, c + off, 1);
        DGL_Vertex2f(i * BLOCK_WIDTH, 0);
        DGL_Vertex2f(i * BLOCK_WIDTH + BLOCK_WIDTH, 0);
        DGL_Vertex2f(i * BLOCK_WIDTH + BLOCK_WIDTH, BLOCK_HEIGHT);
        DGL_Vertex2f(i * BLOCK_WIDTH, BLOCK_HEIGHT);
    }
    DGL_End();
    DGL_MatrixMode(DGL_PROJECTION);
    DGL_PopMatrix();
#undef BORDER
#undef BLOCK_HEIGHT
#undef BLOCK_WIDTH
}
static void drawBBox(dfloat br) {
    DGL_Begin(DGL_QUADS); {
        DGL_TexCoord2f(0, 1.0f, 1.0f); DGL_Vertex3f( 1.0f+br, 1.0f,-1.0f-br);  // TR
        DGL_TexCoord2f(0, 0.0f, 1.0f); DGL_Vertex3f(-1.0f-br, 1.0f,-1.0f-br);  // TL
        DGL_TexCoord2f(0, 0.0f, 0.0f); DGL_Vertex3f(-1.0f-br, 1.0f, 1.0f+br);  // BL
        DGL_TexCoord2f(0, 1.0f, 0.0f); DGL_Vertex3f( 1.0f+br, 1.0f, 1.0f+br);  // BR
        DGL_TexCoord2f(0, 1.0f, 1.0f); DGL_Vertex3f( 1.0f+br,-1.0f, 1.0f+br);  // TR
        DGL_TexCoord2f(0, 0.0f, 1.0f); DGL_Vertex3f(-1.0f-br,-1.0f, 1.0f+br);  // TL
        DGL_TexCoord2f(0, 0.0f, 0.0f); DGL_Vertex3f(-1.0f-br,-1.0f,-1.0f-br);  // BL
        DGL_TexCoord2f(0, 1.0f, 0.0f); DGL_Vertex3f( 1.0f+br,-1.0f,-1.0f-br);  // BR
        DGL_TexCoord2f(0, 1.0f, 1.0f); DGL_Vertex3f( 1.0f+br, 1.0f+br, 1.0f);  // TR
        DGL_TexCoord2f(0, 0.0f, 1.0f); DGL_Vertex3f(-1.0f-br, 1.0f+br, 1.0f);  // TL
        DGL_TexCoord2f(0, 0.0f, 0.0f); DGL_Vertex3f(-1.0f-br,-1.0f-br, 1.0f);  // BL
        DGL_TexCoord2f(0, 1.0f, 0.0f); DGL_Vertex3f( 1.0f+br,-1.0f-br, 1.0f);  // BR
        DGL_TexCoord2f(0, 1.0f, 1.0f); DGL_Vertex3f( 1.0f+br,-1.0f-br,-1.0f);  // TR
        DGL_TexCoord2f(0, 0.0f, 1.0f); DGL_Vertex3f(-1.0f-br,-1.0f-br,-1.0f);  // TL
        DGL_TexCoord2f(0, 0.0f, 0.0f); DGL_Vertex3f(-1.0f-br, 1.0f+br,-1.0f);  // BL
        DGL_TexCoord2f(0, 1.0f, 0.0f); DGL_Vertex3f( 1.0f+br, 1.0f+br,-1.0f);  // BR
        DGL_TexCoord2f(0, 1.0f, 1.0f); DGL_Vertex3f(-1.0f, 1.0f+br, 1.0f+br);  // TR
        DGL_TexCoord2f(0, 0.0f, 1.0f); DGL_Vertex3f(-1.0f, 1.0f+br,-1.0f-br);  // TL
        DGL_TexCoord2f(0, 0.0f, 0.0f); DGL_Vertex3f(-1.0f,-1.0f-br,-1.0f-br);  // BL
        DGL_TexCoord2f(0, 1.0f, 0.0f); DGL_Vertex3f(-1.0f,-1.0f-br, 1.0f+br);  // BR
        DGL_TexCoord2f(0, 1.0f, 1.0f); DGL_Vertex3f( 1.0f, 1.0f+br,-1.0f-br);  // TR
        DGL_TexCoord2f(0, 0.0f, 1.0f); DGL_Vertex3f( 1.0f, 1.0f+br, 1.0f+br);  // TL
        DGL_TexCoord2f(0, 0.0f, 0.0f); DGL_Vertex3f( 1.0f,-1.0f-br, 1.0f+br);  // BL
        DGL_TexCoord2f(0, 1.0f, 0.0f); DGL_Vertex3f( 1.0f,-1.0f-br,-1.0f-br);  // BR
    }
    DGL_End();
}
/**
 * Draws a textured cube using the currently bound gl texture.
 * Used to draw mobj bounding boxes.
 *
 * @param pos          Coordinates of the center of the box (in map space units).
 * @param w            Width of the box.
 * @param l            Length of the box.
 * @param h            Height of the box.
 * @param a            Angle of the box.
 * @param color        Color to make the box (uniform vertex color).
 * @param alpha        Alpha to make the box (uniform vertex color).
 * @param br           Border amount to overlap box faces.
 * @param alignToBase  @c true= align the base of the box to the Z coordinate.
 */
void Rend_DrawBBox(Vector3d const &pos, coord_t w, coord_t l, coord_t h,
    dfloat a, dfloat const color[3], dfloat alpha, dfloat br, bool alignToBase) {
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    if (alignToBase)
        DGL_Translatef(pos.x, pos.z + h, pos.y);
    else
        DGL_Translatef(pos.x, pos.z, pos.y);
    DGL_Rotatef(0, 0, 0, 1);
    DGL_Rotatef(0, 1, 0, 0);
    DGL_Rotatef(a, 0, 1, 0);
    DGL_Scalef(w - br - br, h - br - br, l - br - br);
    DGL_Color4f(color[0], color[1], color[2], alpha);
    drawBBox(.08f);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
}
/**
 * Draws a textured triangle using the currently bound gl texture.
 * Used to draw mobj angle direction arrow.
 *
 * @param pos    World space coordinates of the center of the base of the triangle.
 * @param a      Angle to point the triangle in.
 * @param s      Scale of the triangle.
 * @param color  Color to make the box (uniform vertex color).
 * @param alpha  Alpha to make the box (uniform vertex color).
 */
void Rend_DrawArrow(Vector3d const &pos, dfloat a, dfloat s, dfloat const color[3],
                    dfloat alpha) {
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    DGL_Translatef(pos.x, pos.z, pos.y);
    DGL_Rotatef(0, 0, 0, 1);
    DGL_Rotatef(0, 1, 0, 0);
    DGL_Rotatef(a, 0, 1, 0);
    DGL_Scalef(s, 0, s);
    DGL_Begin(DGL_TRIANGLES); {
        DGL_Color4f(0.0f, 0.0f, 0.0f, 0.5f);
        DGL_TexCoord2f(0, 1.0f, 1.0f);
        DGL_Vertex3f( 1.0f, 1.0f,-1.0f);  // L
        DGL_Color4f(color[0], color[1], color[2], alpha);
        DGL_TexCoord2f(0, 0.0f, 1.0f);
        DGL_Vertex3f(-1.0f, 1.0f,-1.0f);  // Point
        DGL_Color4f(0.0f, 0.0f, 0.0f, 0.5f);
        DGL_TexCoord2f(0, 0.0f, 0.0f);
        DGL_Vertex3f(-1.0f, 1.0f, 1.0f);  // R
    }
    DGL_End();
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
}
static void drawMobjBBox(mobj_t &mob) {
    static dfloat const red   [] = { 1,    0.2f, 0.2f };  // non-solid objects
    static dfloat const green [] = { 0.2f, 1,    0.2f };  // solid objects
    static dfloat const yellow[] = { 0.7f, 0.7f, 0.2f };  // missiles
    if (&mob == DD_Player(consolePlayer)->publicData().mo)
        return;
    if (!Mobj_IsLinked(mob)) return;
    BspLeaf const &bspLeaf = Mobj_BspLeafAtOrigin(mob);
    if (!bspLeaf.hasSubspace() || !R_ViewerSubspaceIsVisible(bspLeaf.subspace()))
        return;
    ddouble const distToEye = (eyeOrigin - Mobj_Origin(mob)).length();
    dfloat alpha = 1 - ((distToEye / (DENG_GAMEVIEW_WIDTH/2)) / 4);
    if (alpha < .25f)
        alpha = .25f; // Don't make them totally invisible.
    coord_t size = Mobj_Radius(mob);
    Rend_DrawBBox(mob.origin, size, size, mob.height/2, 0,
                  (mob.ddFlags & DDMF_MISSILE)? yellow :
                  (mob.ddFlags & DDMF_SOLID)? green : red,
                  alpha, .08f);
    Rend_DrawArrow(mob.origin, ((mob.angle + ANG45 + ANG90) / (dfloat) ANGLE_MAX *-360), size*1.25,
                   (mob.ddFlags & DDMF_MISSILE)? yellow :
                   (mob.ddFlags & DDMF_SOLID)? green : red, alpha);
}
/**
 * Renders bounding boxes for all mobj's (linked in sec->mobjList, except
 * the console player) in all sectors that are currently marked as vissible.
 *
 * Depth test is disabled to show all mobjs that are being rendered, regardless
 * if they are actually vissible (hidden by previously drawn map geometry).
 */
static void drawMobjBoundingBoxes(Map &map) {
    static dfloat const green [] = { 0.2f, 1,    0.2f };  // solid objects
    static dfloat const yellow[] = { 0.7f, 0.7f, 0.2f };  // missiles
    if (!devMobjBBox && !devPolyobjBBox) return;
#ifndef _DEBUG
    if (netGame) return;
#endif
    DGL_Disable(DGL_DEPTH_TEST);
    DGL_Enable(DGL_TEXTURE_2D);
    DGL_CullFace(DGL_NONE);
    MaterialAnimator &matAnimator = ClientMaterial::find(de::Uri("System", Path("bbox")))
            .getAnimator(Rend_SpriteMaterialSpec());
    matAnimator.prepare();
    GL_BindTexture(matAnimator.texUnit(MaterialAnimator::TU_LAYER0).texture);
    GL_BlendMode(BM_ADD);
    if (devMobjBBox) {
        map.thinkers().forAll(reinterpret_cast<thinkfunc_t>(gx.MobjThinker), 0x1, [] (thinker_t *th) {
            drawMobjBBox(*reinterpret_cast<mobj_t *>(th));
            return LoopContinue;
        });
    }
    if (devPolyobjBBox) {
        map.forAllPolyobjs([] (Polyobj &pob) {
            Sector const &sec = pob.sector();
            coord_t width  = (pob.bounds.maxX - pob.bounds.minX)/2;
            coord_t length = (pob.bounds.maxY - pob.bounds.minY)/2;
            coord_t height = (sec.ceiling().height() - sec.floor().height())/2;
            Vector3d pos(pob.bounds.minX + width,
                         pob.bounds.minY + length,
                         sec.floor().height());
            ddouble const distToEye = (eyeOrigin - pos).length();
            dfloat alpha = 1 - ((distToEye / (DENG_GAMEVIEW_WIDTH/2)) / 4);
            if (alpha < .25f)
                alpha = .25f; // Don't make them totally invisible.
            Rend_DrawBBox(pos, width, length, height, 0, yellow, alpha, .08f);
            for (Line *line : pob.lines()) {
                Vector3d pos(line->center(), sec.floor().height());
                Rend_DrawBBox(pos, 0, line->length() / 2, height,
                              BANG2DEG(BANG_90 - line->angle()),
                              green, alpha, 0);
            }
            return LoopContinue;
        });
    }
    GL_BlendMode(BM_NORMAL);
    DGL_PopState();
    DGL_Disable(DGL_TEXTURE_2D);
    DGL_Enable(DGL_DEPTH_TEST);
}
static void drawPoint(Vector3d const &origin, Vector4f const &color = Vector4f(1, 1, 1, 1)) {
    DGL_Begin(DGL_POINTS);
        DGL_Color4f(color.x, color.y, color.z, color.w);
        DGL_Vertex3f(origin.x, origin.z, origin.y);
    DGL_End();
}
static void drawVector(Vector3f const &vector, dfloat scalar, Vector4f const &color = Vector4f(1, 1, 1, 1)) {
    static dfloat const black[] = { 0, 0, 0, 0 };
    DGL_Begin(DGL_LINES);
        DGL_Color4fv(black);
        DGL_Vertex3f(scalar * vector.x, scalar * vector.z, scalar * vector.y);
        DGL_Color4f(color.x, color.y, color.z, color.w);
        DGL_Vertex3f(0, 0, 0);
    DGL_End();
}
static void drawTangentVectorsForSurface(Surface const &suf, Vector3d const &origin) {
    static dint const VISUAL_LENGTH = 20;
    static Vector4f const red  ( 1, 0, 0, 1);
    static Vector4f const green( 0, 1, 0, 1);
    static Vector4f const blue ( 0, 0, 1, 1);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    DGL_Translatef(origin.x, origin.z, origin.y);
    if (::devSurfaceVectors & SVF_TANGENT)   drawVector(suf.tangent(),   VISUAL_LENGTH, red);
    if (::devSurfaceVectors & SVF_BITANGENT) drawVector(suf.bitangent(), VISUAL_LENGTH, green);
    if (::devSurfaceVectors & SVF_NORMAL)    drawVector(suf.normal(),    VISUAL_LENGTH, blue);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
}
/**
 * @todo Determine Z-axis origin from a WallEdge.
 */
static void drawTangentVectorsForWalls(HEdge const *hedge) {
    if (!hedge || !hedge->hasMapElement())
        return;
    LineSideSegment const &seg = hedge->mapElementAs<LineSideSegment>();
    LineSide const &lineSide   = seg.lineSide();
    Line const &line           = lineSide.line();
    Vector2d const center      = (hedge->twin().origin() + hedge->origin()) / 2;
    if (lineSide.considerOneSided()) {
        auto &subsec =
            (line.definesPolyobj() ? line.polyobj().bspLeaf().subspace()
                                   : hedge->face().mapElementAs<ConvexSubspace>())
                .subsector().as<world::ClientSubsector>();
        ddouble const bottom = subsec.  visFloor().heightSmoothed();
        ddouble const top    = subsec.visCeiling().heightSmoothed();
        drawTangentVectorsForSurface(lineSide.middle(),
                                     Vector3d(center, bottom + (top - bottom) / 2));
    }
    else {
        auto &subsec =
            (line.definesPolyobj() ? line.polyobj().bspLeaf().subspace()
                                   : hedge->face().mapElementAs<ConvexSubspace>())
                .subsector().as<world::ClientSubsector>();
        auto &backSubsec =
            (line.definesPolyobj() ? line.polyobj().bspLeaf().subspace()
                                   : hedge->twin().face().mapElementAs<ConvexSubspace>())
                .subsector().as<world::ClientSubsector>();
        if (lineSide.middle().hasMaterial()) {
            ddouble const bottom = subsec.  visFloor().heightSmoothed();
            ddouble const top    = subsec.visCeiling().heightSmoothed();
            drawTangentVectorsForSurface(lineSide.middle(),
                                         Vector3d(center, bottom + (top - bottom) / 2));
        }
        if (backSubsec.visCeiling().heightSmoothed() < subsec.visCeiling().heightSmoothed() &&
           !(subsec.    visCeiling().surface().hasSkyMaskedMaterial() &&
             backSubsec.visCeiling().surface().hasSkyMaskedMaterial())) {
            coord_t const bottom = backSubsec.visCeiling().heightSmoothed();
            coord_t const top    = subsec.    visCeiling().heightSmoothed();
            drawTangentVectorsForSurface(lineSide.top(),
                                         Vector3d(center, bottom + (top - bottom) / 2));
        }
        if (backSubsec.visFloor().heightSmoothed() > subsec.visFloor().heightSmoothed() &&
           !(subsec.    visFloor().surface().hasSkyMaskedMaterial() &&
             backSubsec.visFloor().surface().hasSkyMaskedMaterial())) {
            coord_t const bottom = subsec.    visFloor().heightSmoothed();
            coord_t const top    = backSubsec.visFloor().heightSmoothed();
            drawTangentVectorsForSurface(lineSide.bottom(),
                                         Vector3d(center, bottom + (top - bottom) / 2));
        }
    }
}
/**
 * @todo Use drawTangentVectorsForWalls() for polyobjs too.
 */
static void drawSurfaceTangentVectors(Subsector &subsec) {
    subsec.forAllSubspaces([] (ConvexSubspace &space) {
        HEdge const *base  = space.poly().hedge();
        HEdge const *hedge = base;
        do {
            drawTangentVectorsForWalls(hedge);
        } while ((hedge = &hedge->next()) != base);
        space.forAllExtraMeshes([] (Mesh &mesh) {
            for (HEdge *hedge : mesh.hedges()) {
                drawTangentVectorsForWalls(hedge);
            }
            return LoopContinue;
        });
        space.forAllPolyobjs([] (Polyobj &pob) {
            for (HEdge *hedge : pob.mesh().hedges()) {
                drawTangentVectorsForWalls(hedge);
            }
            return LoopContinue;
        });
        return LoopContinue;
    });
    dint const planeCount = subsec.sector().planeCount();
    for (dint i = 0; i < planeCount; ++i) {
        Plane const &plane = subsec.as<world::ClientSubsector>().visPlane(i);
        ddouble height     = 0;
        if (plane.surface().hasSkyMaskedMaterial()
            && (plane.isSectorFloor() || plane.isSectorCeiling())) {
            height = plane.map().skyPlane(plane.isSectorCeiling()).height();
        }
        else {
            height = plane.heightSmoothed();
        }
        drawTangentVectorsForSurface(plane.surface(), Vector3d(subsec.center(), height));
    }
}
/**
 * Draw the surface tangent space vectors, primarily for debug.
 */
static void drawSurfaceTangentVectors(Map &map) {
    if (!::devSurfaceVectors) return;
    DGL_CullFace(DGL_NONE);
    map.forAllSectors([] (Sector &sec) {
        return sec.forAllSubsectors([] (Subsector &subsec) {
            drawSurfaceTangentVectors(subsec);
            return LoopContinue;
        });
    });
    DGL_PopState();
}
static void drawLumobjs(Map &map) {
    static dfloat const black[] = { 0, 0, 0, 0 };
    if (!devDrawLums) return;
    DGL_PushState();
    DGL_Disable(DGL_DEPTH_TEST);
    DGL_CullFace(DGL_NONE);
    map.forAllLumobjs([] (Lumobj &lob) {
        if (rendMaxLumobjs > 0 && R_ViewerLumobjIsHidden(lob.indexInMap()))
            return LoopContinue;
        DGL_MatrixMode(DGL_MODELVIEW);
        DGL_PushMatrix();
        DGL_Translatef(lob.origin().x, lob.origin().z + lob.zOffset(), lob.origin().y);
        DGL_Begin(DGL_LINES); {
            DGL_Color4fv(black);
            DGL_Vertex3f(-lob.radius(), 0, 0);
            DGL_Color4f(lob.color().x, lob.color().y, lob.color().z, 1);
            DGL_Vertex3f(0, 0, 0);
            DGL_Vertex3f(0, 0, 0);
            DGL_Color4fv(black);
            DGL_Vertex3f(lob.radius(), 0, 0);
            DGL_Vertex3f(0, -lob.radius(), 0);
            DGL_Color4f(lob.color().x, lob.color().y, lob.color().z, 1);
            DGL_Vertex3f(0, 0, 0);
            DGL_Vertex3f(0, 0, 0);
            DGL_Color4fv(black);
            DGL_Vertex3f(0, lob.radius(), 0);
            DGL_Vertex3f(0, 0, -lob.radius());
            DGL_Color4f(lob.color().x, lob.color().y, lob.color().z, 1);
            DGL_Vertex3f(0, 0, 0);
            DGL_Vertex3f(0, 0, 0);
            DGL_Color4fv(black);
            DGL_Vertex3f(0, 0, lob.radius());
        }
        DGL_End();
        DGL_MatrixMode(DGL_MODELVIEW);
        DGL_PopMatrix();
        return LoopContinue;
    });
    DGL_PopState();
}
static String labelForLineSideSection(LineSide &side, dint sectionId) {
    return String("Line #%1 (%2, %3)")
               .arg(side.line().indexInMap())
               .arg(Line::sideIdAsText(side.sideId()).upperFirstChar())
               .arg(LineSide::sectionIdAsText(sectionId));
}
static String labelForSector(Sector &sector) {
    return String("Sector #%1").arg(sector.indexInMap());
}
static String labelForSectorPlane(Plane &plane) {
    return String("Sector #%1 (%2)")
               .arg(plane.sector().indexInMap())
               .arg(Sector::planeIdAsText(plane.indexInSector()).upperFirstChar());
}
/**
 * Debugging aid for visualizing sound origins.
 */
static void drawSoundEmitters(Map &map) {
    static ddouble const MAX_DISTANCE = 384;
    if (!devSoundEmitters) return;
    FR_SetFont(fontFixed);
    FR_LoadDefaultAttrib();
    FR_SetShadowOffset(UI_SHADOW_OFFSET, UI_SHADOW_OFFSET);
    FR_SetShadowStrength(UI_SHADOW_STRENGTH);
    DGL_Disable(DGL_DEPTH_TEST);
    DGL_Enable(DGL_TEXTURE_2D);
    if (devSoundEmitters & SOF_SIDE) {
        map.forAllLines([] (Line &line) {
            for (dint i = 0; i < 2; ++i) {
                LineSide &side = line.side(i);
                if (!side.hasSections()) continue;
                drawLabel(labelForLineSideSection(side, LineSide::Middle)
                          , Vector3d(side.middleSoundEmitter().origin), MAX_DISTANCE);
                drawLabel(labelForLineSideSection(side, LineSide::Bottom)
                          , Vector3d(side.bottomSoundEmitter().origin), MAX_DISTANCE);
                drawLabel(labelForLineSideSection(side, LineSide::Top)
                          , Vector3d(side.topSoundEmitter   ().origin), MAX_DISTANCE);
            }
            return LoopContinue;
        });
    }
    if (devSoundEmitters & (SOF_SECTOR | SOF_PLANE)) {
        map.forAllSectors([] (Sector &sector) {
            if (devSoundEmitters & SOF_PLANE) {
                sector.forAllPlanes([] (Plane &plane) {
                    drawLabel(labelForSectorPlane(plane)
                              , Vector3d(plane.soundEmitter().origin), MAX_DISTANCE);
                    return LoopContinue;
                });
            }
            if (devSoundEmitters & SOF_SECTOR) {
                drawLabel(labelForSector(sector)
                          , Vector3d(sector.soundEmitter().origin), MAX_DISTANCE);
            }
            return LoopContinue;
        });
    }
    DGL_Enable(DGL_DEPTH_TEST);
    DGL_Disable(DGL_TEXTURE_2D);
}
void Rend_DrawVectorLight(VectorLightData const &vlight, dfloat alpha) {
    if (alpha < .0001f) return;
    dfloat const unitLength = 100;
    DGL_Begin(DGL_LINES);
        DGL_Color4f(vlight.color.x, vlight.color.y, vlight.color.z, alpha);
        DGL_Vertex3f(unitLength * vlight.direction.x, unitLength * vlight.direction.z, unitLength * vlight.direction.y);
        DGL_Color4f(vlight.color.x, vlight.color.y, vlight.color.z, 0);
        DGL_Vertex3f(0, 0, 0);
    DGL_End();
}
static String labelForGenerator(Generator const &gen) {
    return String("%1").arg(gen.id());
}
static void drawGenerator(Generator const &gen) {
    static dint const MAX_GENERATOR_DIST = 2048;
    if (gen.source || gen.isUntriggered()) {
        Vector3d const origin   = gen.origin();
        ddouble const distToEye = (eyeOrigin - origin).length();
        if (distToEye < MAX_GENERATOR_DIST) {
            drawLabel(labelForGenerator(gen), origin, distToEye / (DENG_GAMEVIEW_WIDTH / 2)
                      , 1 - distToEye / MAX_GENERATOR_DIST);
        }
    }
}
/**
 * Debugging aid; Draw all active generators.
 */
static void drawGenerators(Map &map) {
    if (!devDrawGenerators) return;
    FR_SetFont(fontFixed);
    FR_LoadDefaultAttrib();
    FR_SetShadowOffset(UI_SHADOW_OFFSET, UI_SHADOW_OFFSET);
    FR_SetShadowStrength(UI_SHADOW_STRENGTH);
    DGL_Disable(DGL_DEPTH_TEST);
    DGL_Enable(DGL_TEXTURE_2D);
    map.forAllGenerators([] (Generator &gen) {
        drawGenerator(gen);
        return LoopContinue;
    });
    DGL_Enable(DGL_DEPTH_TEST);
    DGL_Disable(DGL_TEXTURE_2D);
}
static void drawBar(Vector3d const &origin, coord_t height, dfloat opacity) {
    static dint const EXTEND_DIST = 64;
    static dfloat const black[] = { 0, 0, 0, 0 };
    DGL_Begin(DGL_LINES);
        DGL_Color4fv(black);
        DGL_Vertex3f(origin.x, origin.z - EXTEND_DIST, origin.y);
        DGL_Color4f(1, 1, 1, opacity);
        DGL_Vertex3f(origin.x, origin.z, origin.y);
        DGL_Vertex3f(origin.x, origin.z, origin.y);
        DGL_Vertex3f(origin.x, origin.z + height, origin.y);
        DGL_Vertex3f(origin.x, origin.z + height, origin.y);
        DGL_Color4fv(black);
        DGL_Vertex3f(origin.x, origin.z + height + EXTEND_DIST, origin.y);
    DGL_End();
}
static String labelForVertex(Vertex const *vtx) {
    DENG2_ASSERT(vtx);
    return String("%1").arg(vtx->indexInMap());
}
struct drawvertexvisual_parameters_t {
    dint maxDistance;
    bool drawOrigin;
    bool drawBar;
    bool drawLabel;
    QBitArray *drawnVerts;
};
static void drawVertexVisual(Vertex const &vertex, ddouble minHeight, ddouble maxHeight,
    drawvertexvisual_parameters_t &parms) {
    if (!parms.drawOrigin && !parms.drawBar && !parms.drawLabel)
        return;
    if (vertex.indexInArchive() == MapElement::NoIndex)
        return;
    if (parms.drawnVerts) {
        if (parms.drawnVerts->testBit(vertex.indexInArchive()))
            return;
        parms.drawnVerts->setBit(vertex.indexInArchive());
    }
    ddouble distToEye = (Vector2d(eyeOrigin.x, eyeOrigin.y) - vertex.origin()).length();
    if (distToEye >= parms.maxDistance)
        return;
    Vector3d const origin(vertex.origin(), minHeight);
    dfloat const opacity = 1 - distToEye / parms.maxDistance;
    if (parms.drawBar) {
        drawBar(origin, maxHeight - minHeight, opacity);
    }
    if (parms.drawOrigin) {
        drawPoint(origin, Vector4f(.7f, .7f, .2f, opacity * 4));
    }
    if (parms.drawLabel) {
        DGL_Disable(DGL_DEPTH_TEST);
        DGL_Enable(DGL_TEXTURE_2D);
        drawLabel(labelForVertex(&vertex), origin, distToEye / (DENG_GAMEVIEW_WIDTH / 2), opacity);
        DGL_Enable(DGL_DEPTH_TEST);
        DGL_Disable(DGL_TEXTURE_2D);
    }
}
/**
 * Find the relative next minmal and/or maximal visual height(s) of all sector
 * planes which "interface" at the half-edge, edge vertex.
 *
 * @param base  Base half-edge to find heights for.
 * @param edge  Edge of the half-edge.
 * @param min   Current minimal height to use as a base (will be overwritten).
 *              Use DDMAXFLOAT if the base is unknown.
 * @param min   Current maximal height to use as a base (will be overwritten).
 *              Use DDMINFLOAT if the base is unknown.
 *
 * @todo Don't stop when a zero-volume back neighbor is found; process all of
 * the neighbors at the specified vertex (the half-edge geometry will need to
 * be linked such that "outside" edges are neighbor-linked similarly to those
 * with a face).
 */
static void findMinMaxPlaneHeightsAtVertex(HEdge *base, dint edge,
    ddouble &min, ddouble &max) {
    if (!base || !base->hasFace())
        return;
    if (!base->face().hasMapElement() || !base->face().mapElementAs<ConvexSubspace>().hasSubsector())
        return;
    if (!Subsector::isInternalEdge(base)) {
        ClockDirection const direction = edge? Clockwise : Anticlockwise;
        HEdge *hedge = base;
        while ((hedge = &SubsectorCirculator::findBackNeighbor(*hedge, direction)) != base) {
            auto const *backSpace = hedge->hasFace() ? &hedge->face().mapElementAs<ConvexSubspace>() : nullptr;
            if (!backSpace) break;
            auto const &subsec = backSpace->subsector().as<world::ClientSubsector>();
            if (subsec.visFloor().heightSmoothed() < min) {
                min = subsec.visFloor().heightSmoothed();
            }
            if (subsec.visCeiling().heightSmoothed() > max) {
                max = subsec.visCeiling().heightSmoothed();
            }
        }
    }
}
static void drawSubspaceVertexs(ConvexSubspace &sub, drawvertexvisual_parameters_t &parms) {
    auto &subsec = sub.subsector().as<world::ClientSubsector>();
    ddouble const min = subsec.  visFloor().heightSmoothed();
    ddouble const max = subsec.visCeiling().heightSmoothed();
    HEdge *base  = sub.poly().hedge();
    HEdge *hedge = base;
    do {
        ddouble edgeMin = min;
        ddouble edgeMax = max;
        findMinMaxPlaneHeightsAtVertex(hedge, 0 /*left edge*/, edgeMin, edgeMax);
        drawVertexVisual(hedge->vertex(), min, max, parms);
    } while ((hedge = &hedge->next()) != base);
    sub.forAllExtraMeshes([&min, &max, &parms] (Mesh &mesh) {
        for (HEdge *hedge : mesh.hedges()) {
            drawVertexVisual(hedge->vertex(), min, max, parms);
            drawVertexVisual(hedge->twin().vertex(), min, max, parms);
        }
        return LoopContinue;
    });
    sub.forAllPolyobjs([&min, &max, &parms] (Polyobj &pob) {
        for (Line *line : pob.lines()) {
            drawVertexVisual(line->from(), min, max, parms);
            drawVertexVisual(line->to(), min, max, parms);
        }
        return LoopContinue;
    });
}
/**
 * Draw the various vertex debug aids.
 */
static void drawVertexes(Map &map) {
#define MAX_DISTANCE            1280  ///< From the viewer.
    dfloat oldLineWidth = -1;
    if (!devVertexBars && !devVertexIndices)
        return;
    AABoxd box(eyeOrigin.x - MAX_DISTANCE, eyeOrigin.y - MAX_DISTANCE,
               eyeOrigin.x + MAX_DISTANCE, eyeOrigin.y + MAX_DISTANCE);
    QBitArray drawnVerts(map.vertexCount());
    drawvertexvisual_parameters_t parms;
    parms.maxDistance = MAX_DISTANCE;
    parms.drawnVerts  = &drawnVerts;
    FR_SetFont(fontFixed);
    FR_LoadDefaultAttrib();
    FR_SetShadowOffset(UI_SHADOW_OFFSET, UI_SHADOW_OFFSET);
    FR_SetShadowStrength(UI_SHADOW_STRENGTH);
    if (devVertexBars) {
        DGL_Disable(DGL_DEPTH_TEST);
#if defined (DENG_OPENGL)
        LIBGUI_GL.glEnable(GL_LINE_SMOOTH);
#endif
        oldLineWidth = DGL_GetFloat(DGL_LINE_WIDTH);
        DGL_SetFloat(DGL_LINE_WIDTH, 2);
        parms.drawBar   = true;
        parms.drawLabel = parms.drawOrigin = false;
        map.subspaceBlockmap().forAllInBox(box, [&box, &parms] (void *object) {
            auto &subspace = *(ConvexSubspace *)object;
            AABoxd const &polyBounds = subspace.poly().bounds();
            if (!(   polyBounds.maxX < box.minX
                 || polyBounds.minX > box.maxX
                 || polyBounds.minY > box.maxY
                 || polyBounds.maxY < box.minY)) {
                drawSubspaceVertexs(subspace, parms);
            }
            return LoopContinue;
        });
        DGL_Enable(DGL_DEPTH_TEST);
    }
    dfloat const oldPointSize = DGL_GetFloat(DGL_POINT_SIZE);
#if defined (DENG_OPENGL)
    LIBGUI_GL.glEnable(GL_POINT_SMOOTH);
#endif
    DGL_SetFloat(DGL_POINT_SIZE, 6);
    DGL_Disable(DGL_DEPTH_TEST);
    parms.drawnVerts->fill(false);  // Process all again.
    parms.drawOrigin = true;
    parms.drawBar    = parms.drawLabel = false;
    map.subspaceBlockmap().forAllInBox(box, [&box, &parms] (void *object) {
        auto &subspace = *(ConvexSubspace *)object;
        AABoxd const &polyBounds = subspace.poly().bounds();
        if (!(   polyBounds.maxX < box.minX
              || polyBounds.minX > box.maxX
              || polyBounds.minY > box.maxY
              || polyBounds.maxY < box.minY)) {
            drawSubspaceVertexs(subspace, parms);
        }
        return LoopContinue;
    });
    DGL_Enable(DGL_DEPTH_TEST);
    if (devVertexIndices) {
        parms.drawnVerts->fill(false);  // Process all again.
        parms.drawLabel = true;
        parms.drawBar   = parms.drawOrigin = false;
        map.subspaceBlockmap().forAllInBox(box, [&box, &parms] (void *object) {
            auto &subspace = *(ConvexSubspace *)object;
            AABoxd const &polyBounds = subspace.poly().bounds();
            if (!(   polyBounds.maxX < box.minX
                  || polyBounds.minX > box.maxX
                  || polyBounds.minY > box.maxY
                  || polyBounds.maxY < box.minY)) {
                drawSubspaceVertexs(subspace, parms);
            }
            return LoopContinue;
        });
    }
    if (devVertexBars) {
        DGL_SetFloat(DGL_LINE_WIDTH, oldLineWidth);
#if defined (DENG_OPENGL)
        LIBGUI_GL.glDisable(GL_LINE_SMOOTH);
#endif
    }
    DGL_SetFloat(DGL_POINT_SIZE, oldPointSize);
#if defined (DENG_OPENGL)
    LIBGUI_GL.glDisable(GL_POINT_SMOOTH);
#endif
#undef MAX_VERTEX_POINT_DIST
}
static String labelForSubsector(Subsector const &subsec) {
    return String::number(subsec.sector().indexInMap());
}
/**
 * Draw the sector debugging aids.
 */
static void drawSectors(Map &map) {
    static ddouble const MAX_LABEL_DIST = 1280;
    if (!devSectorIndices) return;
    FR_SetFont(fontFixed);
    FR_LoadDefaultAttrib();
    FR_SetShadowOffset(UI_SHADOW_OFFSET, UI_SHADOW_OFFSET);
    FR_SetShadowStrength(UI_SHADOW_STRENGTH);
    DGL_Disable(DGL_DEPTH_TEST);
    DGL_Enable(DGL_TEXTURE_2D);
    map.forAllSectors([] (Sector &sec) {
        return sec.forAllSubsectors([] (Subsector &subsec) {
            Vector3d const origin(subsec.center(), subsec.as<world::ClientSubsector>().visFloor().heightSmoothed());
            ddouble const distToEye = (eyeOrigin - origin).length();
            if (distToEye < MAX_LABEL_DIST) {
                drawLabel(labelForSubsector(subsec), origin, distToEye / (DENG_GAMEVIEW_WIDTH / 2)
                          , 1 - distToEye / MAX_LABEL_DIST);
            }
            return LoopContinue;
        });
    });
    DGL_Enable(DGL_DEPTH_TEST);
    DGL_Disable(DGL_TEXTURE_2D);
}
static String labelForThinker(thinker_t *thinker) {
    DENG2_ASSERT(thinker);
    return String("%1").arg(thinker->id);
}
/**
 * Debugging aid for visualizing thinker IDs.
 */
static void drawThinkers(Map &map) {
    static ddouble const MAX_THINKER_DIST = 2048;
    if (!devThinkerIds) return;
    FR_SetFont(fontFixed);
    FR_LoadDefaultAttrib();
    FR_SetShadowOffset(UI_SHADOW_OFFSET, UI_SHADOW_OFFSET);
    FR_SetShadowStrength(UI_SHADOW_STRENGTH);
    DGL_Disable(DGL_DEPTH_TEST);
    DGL_Enable(DGL_TEXTURE_2D);
    map.thinkers().forAll(0x1 | 0x2, [] (thinker_t *th) {
        if (Thinker_IsMobj(th)) {
            Vector3d const origin   = Mobj_Center(*(mobj_t *)th);
            ddouble const distToEye = (eyeOrigin - origin).length();
            if (distToEye < MAX_THINKER_DIST) {
                drawLabel(labelForThinker(th), origin,  distToEye / (DENG_GAMEVIEW_WIDTH / 2)
                          , 1 - distToEye / MAX_THINKER_DIST);
            }
        }
        return LoopContinue;
    });
    DGL_Enable(DGL_DEPTH_TEST);
    DGL_Disable(DGL_TEXTURE_2D);
}
#if 0
void Rend_LightGridVisual(LightGrid &lg) {
    static Vector3f const red(1, 0, 0);
    static dint blink = 0;
    if (!devLightGrid) return;
    DENG_ASSERT_IN_MAIN_THREAD();
    DENG_ASSERT_GL_CONTEXT_ACTIVE();
    LightGrid::Index viewerGridIndex = 0;
    if (viewPlayer) {
        blink++;
        viewerGridIndex = lg.toIndex(lg.toRef(viewPlayer->publicData().mo->origin));
    }
    DGL_MatrixMode(DGL_PROJECTION);
    DGL_PushMatrix();
    DGL_LoadIdentity();
    DGL_Ortho(0, 0, DENG_GAMEVIEW_WIDTH, DENG_GAMEVIEW_HEIGHT, -1, 1);
    for (dint y = 0; y < lg.dimensions().y; ++y) {
        DGL_Begin(DGL_QUADS);
        for (dint x = 0; x < lg.dimensions().x; ++x) {
            LightGrid::Index gridIndex = lg.toIndex(x, lg.dimensions().y - 1 - y);
            bool const isViewerIndex   = (viewPlayer && viewerGridIndex == gridIndex);
            Vector3f const *color = 0;
            if (isViewerIndex && (blink & 16)) {
                color = &red;
            }
            else if (lg.primarySource(gridIndex)) {
                color = &lg.rawColorRef(gridIndex);
            }
            if (!color) continue;
            LIBGUI_GL.glColor3f(color->x, color->y, color->z);
            DGL_Vertex2f(x * devLightGridSize, y * devLightGridSize);
            DGL_Vertex2f(x * devLightGridSize + devLightGridSize, y * devLightGridSize);
            DGL_Vertex2f(x * devLightGridSize + devLightGridSize,
                       y * devLightGridSize + devLightGridSize);
            DGL_Vertex2f(x * devLightGridSize, y * devLightGridSize + devLightGridSize);
        }
        DGL_End();
    }
    DGL_MatrixMode(DGL_PROJECTION);
    DGL_PopMatrix();
}
#endif
MaterialVariantSpec const &Rend_MapSurfaceMaterialSpec(dint wrapS, dint wrapT) {
    return ClientApp::resources().materialSpec(MapSurfaceContext, 0, 0, 0, 0, wrapS, wrapT,
                                               -1, -1, -1, true, true, false, false);
}
MaterialVariantSpec const &Rend_MapSurfaceMaterialSpec() {
    if (!lookupMapSurfaceMaterialSpec) {
        lookupMapSurfaceMaterialSpec = &Rend_MapSurfaceMaterialSpec(GL_REPEAT, GL_REPEAT);
    }
    return *lookupMapSurfaceMaterialSpec;
}
TextureVariantSpec const &Rend_MapSurfaceLightmapTextureSpec() {
    return ClientApp::resources().textureSpec(TC_MAPSURFACE_LIGHTMAP, 0, 0, 0, 0, GL_CLAMP_TO_EDGE, GL_CLAMP_TO_EDGE,
                                1, -1, -1, false, false, false, true);
}
TextureVariantSpec const &Rend_MapSurfaceShinyTextureSpec() {
    return ClientApp::resources().textureSpec(TC_MAPSURFACE_REFLECTION, TSF_NO_COMPRESSION, 0, 0, 0,
                                GL_REPEAT, GL_REPEAT, 1, 1, -1, false, false, false, false);
}
TextureVariantSpec const &Rend_MapSurfaceShinyMaskTextureSpec() {
    return ClientApp::resources().textureSpec(TC_MAPSURFACE_REFLECTIONMASK, 0, 0, 0, 0, GL_REPEAT, GL_REPEAT,
                                -1, -1, -1, true, false, false, false);
}
D_CMD(OpenRendererAppearanceEditor) {
    DENG2_UNUSED3(src, argc, argv);
    if (!App_GameLoaded()) {
        LOG_ERROR("A game must be loaded before the Renderer Appearance editor can be opened");
        return false;
    }
    if (!ClientWindow::main().hasSidebar()) {
        RendererAppearanceEditor *editor = new RendererAppearanceEditor;
        editor->open();
    }
    return true;
}
D_CMD(OpenModelAssetEditor) {
    DENG2_UNUSED3(src, argc, argv);
    if (!App_GameLoaded()) {
        LOG_ERROR("A game must be loaded before the Model Asset editor can be opened");
        return false;
    }
    if (!ClientWindow::main().hasSidebar()) {
        ModelAssetEditor *editor = new ModelAssetEditor;
        editor->open();
    }
    return true;
}
D_CMD(LowRes) {
    DENG2_UNUSED3(src, argv, argc);
    filterSprites = 0;
    filterUI      = 0;
    texMagMode    = 0;
    GL_SetRawTexturesMinFilter(GL_NEAREST);
    GL_TexReset();
    return true;
}
D_CMD(MipMap) {
    DENG2_UNUSED2(src, argc);
    dint newMipMode = String(argv[1]).toInt();
    if (newMipMode < 0 || newMipMode > 5) {
        LOG_SCR_ERROR("Invalid mipmapping mode %i; the valid range is 0...5") << newMipMode;
        return false;
    }
    mipmapping = newMipMode;
    return true;
}
D_CMD(TexReset) {
    DENG2_UNUSED(src);
    if (argc == 2 && !String(argv[1]).compareWithoutCase("raw")) {
        GL_ReleaseTexturesForRawImages();
    }
    else {
        GL_TexReset();
    }
    if (auto *map = App_World().mapPtr()) {
        map->redecorate();
    }
    return true;
}
static void detailFactorChanged() {
    App_Resources().releaseGLTexturesByScheme("Details");
}
/*
static void fieldOfViewChanged() {
    if (vrCfg().mode() == VRConfig::OculusRift) {
        if (Con_GetFloat("rend-vr-rift-fovx") != fieldOfView)
            Con_SetFloat("rend-vr-rift-fovx", fieldOfView);
    }
    else {
        if (Con_GetFloat("rend-vr-nonrift-fovx") != fieldOfView)
            Con_SetFloat("rend-vr-nonrift-fovx", fieldOfView);
    }
}*/
static void loadExtAlwaysChanged() {
    GL_TexReset();
}
static void mipmappingChanged() {
    GL_TexReset();
}
static void texGammaChanged() {
    R_BuildTexGammaLut(texGamma);
    GL_TexReset();
    LOG_GL_MSG("Texture gamma correction set to %f") << texGamma;
}
static void texQualityChanged() {
    GL_TexReset();
}
static void useDynlightsChanged() {
    if (!ClientApp::world().hasMap()) return;
    ClientApp::world().map().thinkers()
        .forAll(reinterpret_cast<thinkfunc_t>(gx.MobjThinker), 0x1, [](thinker_t *th) {
        Mobj_UnlinkLumobjs(reinterpret_cast<mobj_t *>(th));
        return LoopContinue;
    });
}
#if 0
static void useSkylightChanged() {
    scheduleFullLightGridUpdate();
}
#endif
static void useSmartFilterChanged() {
    GL_TexReset();
}
void Rend_Register() {
    C_VAR_FLOAT("rend-camera-fov", &fieldOfView, 0, 1, 179);
    C_VAR_FLOAT("rend-glow", &glowFactor, 0, 0, 2);
    C_VAR_INT("rend-glow-height", &glowHeightMax, 0, 0, 1024);
    C_VAR_FLOAT("rend-glow-scale", &glowHeightFactor, 0, 0.1f, 10);
    C_VAR_INT("rend-glow-wall", &useGlowOnWalls, 0, 0, 1);
    C_VAR_BYTE("rend-info-lums", &rendInfoLums, 0, 0, 1);
    C_VAR_INT2("rend-light", &useDynLights, 0, 0, 1, useDynlightsChanged);
    C_VAR_INT2("rend-light-ambient", &ambientLight, 0, 0, 255, Rend_UpdateLightModMatrix);
    C_VAR_FLOAT("rend-light-attenuation", &rendLightDistanceAttenuation, CVF_NO_MAX, 0, 0);
    C_VAR_INT("rend-light-blend", &dynlightBlend, 0, 0, 2);
    C_VAR_FLOAT("rend-light-bright", &dynlightFactor, 0, 0, 1);
    C_VAR_FLOAT2("rend-light-compression", &lightRangeCompression, 0, -1, 1, Rend_UpdateLightModMatrix);
    C_VAR_BYTE("rend-light-decor", &useLightDecorations, 0, 0, 1);
    C_VAR_FLOAT("rend-light-fog-bright", &dynlightFogBright, 0, 0, 1);
    C_VAR_INT("rend-light-num", &rendMaxLumobjs, CVF_NO_MAX, 0, 0);
    C_VAR_FLOAT("rend-light-sky", &rendSkyLight, 0, 0, 1/*, useSkylightChanged*/);
    C_VAR_BYTE("rend-light-sky-auto", &rendSkyLightAuto, 0, 0, 1/*, useSkylightChanged*/);
    C_VAR_FLOAT("rend-light-wall-angle", &rendLightWallAngle, CVF_NO_MAX, 0, 0);
    C_VAR_BYTE("rend-light-wall-angle-smooth", &rendLightWallAngleSmooth, 0, 0, 1);
    C_VAR_BYTE("rend-map-material-precache", &precacheMapMaterials, 0, 0, 1);
    C_VAR_INT("rend-shadow", &useShadows, 0, 0, 1);
    C_VAR_FLOAT("rend-shadow-darkness", &shadowFactor, 0, 0, 2);
    C_VAR_INT("rend-shadow-far", &shadowMaxDistance, CVF_NO_MAX, 0, 0);
    C_VAR_INT("rend-shadow-radius-max", &shadowMaxRadius, CVF_NO_MAX, 0, 0);
    C_VAR_INT("rend-tex", &renderTextures, CVF_NO_ARCHIVE, 0, 2);
    C_VAR_BYTE("rend-tex-anim-smooth", &smoothTexAnim, 0, 0, 1);
    C_VAR_INT("rend-tex-detail", &r_detail, 0, 0, 1);
    C_VAR_FLOAT("rend-tex-detail-scale", &detailScale, CVF_NO_MIN | CVF_NO_MAX, 0, 0);
    C_VAR_FLOAT2("rend-tex-detail-strength", &detailFactor, 0, 0, 5, detailFactorChanged);
    C_VAR_BYTE2("rend-tex-external-always", &loadExtAlways, 0, 0, 1, loadExtAlwaysChanged);
    C_VAR_INT("rend-tex-filter-anisotropic", &texAniso, 0, -1, 4);
    C_VAR_INT("rend-tex-filter-mag", &texMagMode, 0, 0, 1);
    C_VAR_INT2("rend-tex-filter-smart", &useSmartFilter, 0, 0, 1, useSmartFilterChanged);
    C_VAR_INT("rend-tex-filter-sprite", &filterSprites, 0, 0, 1);
    C_VAR_INT("rend-tex-filter-ui", &filterUI, 0, 0, 1);
    C_VAR_FLOAT2("rend-tex-gamma", &texGamma, 0, 0, 1, texGammaChanged);
    C_VAR_INT2("rend-tex-mipmap", &mipmapping, CVF_PROTECTED, 0, 5, mipmappingChanged);
    C_VAR_INT2("rend-tex-quality", &texQuality, 0, 0, 8, texQualityChanged);
    C_VAR_INT("rend-tex-shiny", &useShinySurfaces, 0, 0, 1);
    C_VAR_BYTE("rend-dev-blockmap-debug", &bmapShowDebug, CVF_NO_ARCHIVE, 0, 4);
    C_VAR_FLOAT("rend-dev-blockmap-debug-size", &bmapDebugSize, CVF_NO_ARCHIVE, .1f, 100);
    C_VAR_INT("rend-dev-cull-leafs", &devNoCulling, CVF_NO_ARCHIVE, 0, 1);
    C_VAR_BYTE("rend-dev-freeze", &freezeRLs, CVF_NO_ARCHIVE, 0, 1);
    C_VAR_BYTE("rend-dev-generator-show-indices", &devDrawGenerators, CVF_NO_ARCHIVE, 0, 1);
    C_VAR_BYTE("rend-dev-light-mod", &devLightModRange, CVF_NO_ARCHIVE, 0, 1);
    C_VAR_BYTE("rend-dev-lums", &devDrawLums, CVF_NO_ARCHIVE, 0, 1);
    C_VAR_INT("rend-dev-mobj-bbox", &devMobjBBox, CVF_NO_ARCHIVE, 0, 1);
    C_VAR_BYTE("rend-dev-mobj-show-vlights", &devMobjVLights, CVF_NO_ARCHIVE, 0, 1);
    C_VAR_INT("rend-dev-polyobj-bbox", &devPolyobjBBox, CVF_NO_ARCHIVE, 0, 1);
    C_VAR_BYTE("rend-dev-sector-show-indices", &devSectorIndices, CVF_NO_ARCHIVE, 0, 1);
    C_VAR_INT("rend-dev-sky", &devRendSkyMode, CVF_NO_ARCHIVE, 0, 1);
    C_VAR_BYTE("rend-dev-sky-always", &devRendSkyAlways, CVF_NO_ARCHIVE, 0, 1);
    C_VAR_BYTE("rend-dev-soundorigins", &devSoundEmitters, CVF_NO_ARCHIVE, 0, 7);
    C_VAR_BYTE("rend-dev-surface-show-vectors", &devSurfaceVectors, CVF_NO_ARCHIVE, 0, 7);
    C_VAR_BYTE("rend-dev-thinker-ids", &devThinkerIds, CVF_NO_ARCHIVE, 0, 1);
    C_VAR_BYTE("rend-dev-tex-showfix", &devNoTexFix, CVF_NO_ARCHIVE, 0, 1);
    C_VAR_BYTE("rend-dev-vertex-show-bars", &devVertexBars, CVF_NO_ARCHIVE, 0, 1);
    C_VAR_BYTE("rend-dev-vertex-show-indices", &devVertexIndices, CVF_NO_ARCHIVE, 0, 1);
    C_CMD("rendedit", "", OpenRendererAppearanceEditor);
    C_CMD("modeledit", "", OpenModelAssetEditor);
    C_CMD("cubeshot", "i", CubeShot);
    C_CMD_FLAGS("lowres", "", LowRes, CMDF_NO_DEDICATED);
    C_CMD_FLAGS("mipmap", "i", MipMap, CMDF_NO_DEDICATED);
    C_CMD_FLAGS("texreset", "", TexReset, CMDF_NO_DEDICATED);
    C_CMD_FLAGS("texreset", "s", TexReset, CMDF_NO_DEDICATED);
    LightDecoration::consoleRegister();
#if 0
    BiasIllum::consoleRegister();
    LightGrid::consoleRegister();
#endif
    Lumobj::consoleRegister();
    SkyDrawable::consoleRegister();
    Rend_ModelRegister();
    Rend_ParticleRegister();
    world::Generator::consoleRegister();
    Rend_RadioRegister();
    Rend_SpriteRegister();
    PostProcessing::consoleRegister();
    fx::Bloom::consoleRegister();
    fx::Vignette::consoleRegister();
    fx::LensFlares::consoleRegister();
#if 0
    Shard::consoleRegister();
#endif
    VR_ConsoleRegister();
}
ResourceConfigVars &R_Config() {
    static ResourceConfigVars vars { nullptr, nullptr, nullptr};
    if (!vars.noHighResTex) {
        vars.noHighResTex     = &App::config("resource.noHighResTex");
        vars.noHighResPatches = &App::config("resource.noHighResPatches");
        vars.highResWithPWAD  = &App::config("resource.highResWithPWAD");
    }
    return vars;
}
#define QATAN2(y,x)         qatan2(y,x)
#define QASIN(x)            asin(x) // @todo Precalculate arcsin.
static inline float qatan2(float y, float x) {
    float ang = BANG2RAD(bamsAtan2(y * 512, x * 512));
    if (ang > PI) ang -= 2 * (float) PI;
    return ang;
}
enum rendcmd_t {
    RC_COMMAND_COORDS,
    RC_OTHER_COORDS,
    RC_BOTH_COORDS
};
byte useModels         = true;
int modelLight         = 4;
int frameInter         = true;
float modelAspectMod   = 1 / 1.2f; //.833334f;
int mirrorHudModels;
float modelShinyFactor = 1.0f;
float modelSpinSpeed   = 1;
int maxModelDistance   = 1500;
float rend_model_lod   = 256;
byte precacheSkins     = true;
static bool inited;
struct array_t {
    bool enabled;
    void *data;
};
#define MAX_ARRAYS (2 + MAX_TEX_UNITS)
static array_t arrays[MAX_ARRAYS];
static Vector3f *modelPosCoords;
static Vector3f *modelNormCoords;
static Vector4ub *modelColorCoords;
static Vector2f *modelTexCoords;
static Vector3f modelCenter;
static FrameModelLOD *activeLod;
static uint vertexBufferMax; ///< Maximum number of vertices we'll be required to render per submodel.
static uint vertexBufferSize; ///< Current number of vertices supported by the render buffer.
#ifdef DENG_DEBUG
static bool announcedVertexBufferMaxBreach; ///< @c true if an attempt has been made to expand beyond our capability.
#endif
/*static void modelAspectModChanged()
{
}*/
void Rend_ModelRegister() {
    C_VAR_BYTE ("rend-model",                &useModels,            0, 0, 1);
    C_VAR_INT  ("rend-model-lights",         &modelLight,           0, 0, 10);
    C_VAR_INT  ("rend-model-inter",          &frameInter,           0, 0, 1);
    C_VAR_FLOAT("rend-model-aspect",         &modelAspectMod,       CVF_NO_MAX | CVF_NO_MIN, 0, 0);
    C_VAR_INT  ("rend-model-distance",       &maxModelDistance,     CVF_NO_MAX, 0, 0);
    C_VAR_BYTE ("rend-model-precache",       &precacheSkins,        0, 0, 1);
    C_VAR_FLOAT("rend-model-lod",            &rend_model_lod,       CVF_NO_MAX, 0, 0);
    C_VAR_INT  ("rend-model-mirror-hud",     &mirrorHudModels,      0, 0, 1);
    C_VAR_FLOAT("rend-model-spin-speed",     &modelSpinSpeed,       CVF_NO_MAX | CVF_NO_MIN, 0, 0);
    C_VAR_FLOAT("rend-model-shiny-strength", &modelShinyFactor,     0, 0, 10);
    C_VAR_FLOAT("rend-model-fov",            &weaponFixedFOV,       0, 0, 180);
}
void Rend_ModelInit() {
    if (inited) return; // Already been here.
    modelPosCoords   = 0;
    modelNormCoords  = 0;
    modelColorCoords = 0;
    modelTexCoords   = 0;
    vertexBufferMax = vertexBufferSize = 0;
#ifdef DENG_DEBUG
    announcedVertexBufferMaxBreach = false;
#endif
    inited = true;
}
void Rend_ModelShutdown() {
    if (!inited) return;
    M_Free(modelPosCoords); modelPosCoords = 0;
    M_Free(modelNormCoords); modelNormCoords = 0;
    M_Free(modelColorCoords); modelColorCoords = 0;
    M_Free(modelTexCoords); modelTexCoords = 0;
    vertexBufferMax = vertexBufferSize = 0;
#ifdef DENG_DEBUG
    announcedVertexBufferMaxBreach = false;
#endif
    inited = false;
}
bool Rend_ModelExpandVertexBuffers(uint numVertices) {
    DENG2_ASSERT(inited);
    LOG_AS("Rend_ModelExpandVertexBuffers");
    if (numVertices <= vertexBufferMax) return true;
    if (numVertices >= RENDER_MAX_MODEL_VERTS) {
#ifdef DENG_DEBUG
        if (!announcedVertexBufferMaxBreach) {
            LOGDEV_GL_WARNING("Attempted to expand to %u vertices (max %u)")
                << numVertices << RENDER_MAX_MODEL_VERTS;
            announcedVertexBufferMaxBreach = true;
        }
#endif
        return false;
    }
    vertexBufferMax = numVertices;
    return true;
}
static bool resizeVertexBuffer(uint numVertices) {
    Rend_ModelExpandVertexBuffers(numVertices);
    if (vertexBufferMax != vertexBufferSize) {
        modelPosCoords   =  (Vector3f *) M_Realloc(modelPosCoords,   sizeof(*modelPosCoords)   * vertexBufferMax);
        modelNormCoords  =  (Vector3f *) M_Realloc(modelNormCoords,  sizeof(*modelNormCoords)  * vertexBufferMax);
        modelColorCoords = (Vector4ub *) M_Realloc(modelColorCoords, sizeof(*modelColorCoords) * vertexBufferMax);
        modelTexCoords   =  (Vector2f *) M_Realloc(modelTexCoords,   sizeof(*modelTexCoords)   * vertexBufferMax);
        vertexBufferSize = vertexBufferMax;
    }
    return vertexBufferSize >= numVertices;
}
static void disableArrays(int vertices, int colors, int coords) {
    DENG_ASSERT_IN_MAIN_THREAD();
    DENG_ASSERT_GL_CONTEXT_ACTIVE();
    if (vertices) {
        arrays[AR_VERTEX].enabled = false;
    }
    if (colors) {
        arrays[AR_COLOR].enabled = false;
    }
    for (int i = 0; i < MAX_TEX_UNITS; ++i) {
        if (coords & (1 << i)) {
            arrays[AR_TEXCOORD0 + i].enabled = false;
        }
    }
    DENG_ASSERT(!Sys_GLCheckError());
}
static inline void enableTexUnit(int id) {
    DENG_ASSERT_IN_MAIN_THREAD();
    DENG_ASSERT_GL_CONTEXT_ACTIVE();
    DGL_Enable(DGL_TEXTURE0 + id);
}
static inline void disableTexUnit(int id) {
    DENG_ASSERT_IN_MAIN_THREAD();
    DENG_ASSERT_GL_CONTEXT_ACTIVE();
    DGL_Disable(DGL_TEXTURE0 + id);
    disableArrays(0, 0, 1 << id);
}
/**
 * The first selected unit is active after this call.
 */
static void selectTexUnits(int count) {
    if (count < MAX_TEX_UNITS) {
        for (int i = MAX_TEX_UNITS - 1; i >= count; i--) {
            disableTexUnit(i);
        }
    }
    for (int i = count - 1; i >= 0; i--) {
        if (i >= MAX_TEX_UNITS) continue;
        enableTexUnit(i);
    }
}
/**
 * Enable, set and optionally lock all enabled arrays.
 */
static void configureArrays(void *vertices, void *colors, int numCoords = 0,
                            void **coords = 0) {
    DENG_ASSERT_IN_MAIN_THREAD();
    DENG_ASSERT_GL_CONTEXT_ACTIVE();
    if (vertices) {
        arrays[AR_VERTEX].enabled = true;
        arrays[AR_VERTEX].data = vertices;
    }
    if (colors) {
        arrays[AR_COLOR].enabled = true;
        arrays[AR_COLOR].data = colors;
    }
    for (int i = 0; i < numCoords && i < MAX_TEX_UNITS; ++i) {
        if (coords[i]) {
            arrays[AR_TEXCOORD0 + i].enabled = true;
            arrays[AR_TEXCOORD0 + i].data = coords[i];
        }
    }
    DENG_ASSERT(!Sys_GLCheckError());
}
static void drawArrayElement(int index) {
    DENG_ASSERT_IN_MAIN_THREAD();
    DENG_ASSERT_GL_CONTEXT_ACTIVE();
    for (int i = 0; i < MAX_TEX_UNITS; ++i) {
        if (arrays[AR_TEXCOORD0 + i].enabled) {
            Vector2f const &texCoord = reinterpret_cast<Vector2f const *>(arrays[AR_TEXCOORD0 + i].data)[index];
            DGL_TexCoord2fv(byte(i), texCoord.constPtr());
        }
    }
    if (arrays[AR_COLOR].enabled) {
        Vector4ub const &colorCoord = reinterpret_cast<Vector4ub const *>(arrays[AR_COLOR].data)[index];
        DGL_Color4ubv(colorCoord.constPtr());
    }
    if (arrays[AR_VERTEX].enabled) {
        Vector3f const &posCoord = reinterpret_cast<Vector3f const *>(arrays[AR_VERTEX].data)[index];
        DGL_Vertex3fv(posCoord.constPtr());
    }
}
/**
 * Return a pointer to the visible model frame.
 */
static FrameModelFrame &visibleModelFrame(FrameModelDef &modef, int subnumber, int mobjId) {
    if (subnumber >= int(modef.subCount())) {
        throw Error("Rend_DrawModel.visibleFrame",
                    QString("Model has %1 submodels, but submodel #%2 was requested")
                        .arg(modef.subCount()).arg(subnumber));
    }
    SubmodelDef const &sub = modef.subModelDef(subnumber);
    int curFrame = sub.frame;
    if (modef.flags & MFF_IDFRAME) {
        curFrame += mobjId % sub.frameRange;
    }
    return App_Resources().model(sub.modelId).frame(curFrame);
}
/**
 * Render a set of 3D model primitives using the given data.
 */
static void drawPrimitives(rendcmd_t mode,
                           FrameModel::Primitives const &primitives,
                           Vector3f *posCoords,
                           Vector4ub *colorCoords,
                           Vector2f *texCoords = 0) {
    DENG_ASSERT_IN_MAIN_THREAD();
    DENG_ASSERT_GL_CONTEXT_ACTIVE();
    disableArrays(true, true, DDMAXINT);
    void *coords[2];
    switch (mode) {
    case RC_OTHER_COORDS:
        coords[0] = texCoords;
        configureArrays(posCoords, colorCoords, 1, coords);
        break;
    case RC_BOTH_COORDS:
        coords[0] = NULL;
        coords[1] = texCoords;
        configureArrays(posCoords, colorCoords, 2, coords);
        break;
    default:
        configureArrays(posCoords, colorCoords);
        break;
    }
    FrameModel::Primitive::Element const *firstElem = nullptr;
    bool joining = false;
    auto submitElement = [mode, &firstElem] (FrameModel::Primitive::Element const &elem) {
        if (!firstElem) {
            firstElem = &elem;
        }
        if (mode != RC_OTHER_COORDS) {
            DGL_TexCoord2fv(0, elem.texCoord.constPtr());
        }
        drawArrayElement(elem.index);
    };
    int lastLength = 0;
    DGL_Begin(DGL_TRIANGLE_STRIP);
    foreach (FrameModel::Primitive const &prim, primitives) {
        DGLenum const primType = (prim.triFan? DGL_TRIANGLE_FAN : DGL_TRIANGLE_STRIP);
        joining = false;
        if (lastLength > 0) {
            DGL_Vertex3fv(nullptr);
            if (lastLength & 1) DGL_Vertex3fv(nullptr); // Retain the winding.
            joining = true;
        }
        firstElem = nullptr;
        if (primType == DGL_TRIANGLE_STRIP) {
            lastLength = prim.elements.size();
            foreach (FrameModel::Primitive::Element const &elem, prim.elements) {
                submitElement(elem);
                if (joining) {
                    DGL_Vertex3fv(nullptr);
                    joining = false;
                }
            }
        }
        else {
            lastLength = 2; // just make it even, so it doesn't affect winding (see above)
            for (int i = 1; i < prim.elements.size(); ++i) {
                submitElement(prim.elements.at(0));
                if (joining) {
                    DGL_Vertex3fv(nullptr);
                    joining = false;
                }
                submitElement(prim.elements.at(i));
            }
        }
    }
    DGL_End();
}
/**
 * Interpolate linearly between two sets of vertices.
 */
static void Mod_LerpVertices(float inter, int count, FrameModelFrame const &from,
    FrameModelFrame const &to, Vector3f *posOut, Vector3f *normOut) {
    DENG2_ASSERT(&from.model == &to.model); // sanity check.
    DENG2_ASSERT(!activeLod || &activeLod->model == &from.model); // sanity check.
    DENG2_ASSERT(from.vertices.count() == to.vertices.count()); // sanity check.
    FrameModelFrame::VertexBuf::const_iterator startIt = from.vertices.begin();
    FrameModelFrame::VertexBuf::const_iterator endIt   = to.vertices.begin();
    if (&from == &to || de::fequal(inter, 0)) {
        for (int i = 0; i < count; ++i, startIt++, posOut++, normOut++) {
            if (!activeLod || activeLod->hasVertex(i)) {
                *posOut  = startIt->pos;
                *normOut = startIt->norm;
            }
        }
    }
    else {
        for (int i = 0; i < count; ++i, startIt++, endIt++, posOut++, normOut++) {
            if (!activeLod || activeLod->hasVertex(i)) {
                *posOut  = de::lerp(startIt->pos,  endIt->pos,  inter);
                *normOut = de::lerp(startIt->norm, endIt->norm, inter);
            }
        }
    }
}
static void Mod_MirrorCoords(dint count, Vector3f *coords, dint axis) {
    DENG2_ASSERT(coords);
    for (; count-- > 0; coords++) {
        (*coords)[axis] = -(*coords)[axis];
    }
}
/**
 * Rotate a VectorLight direction vector from world space to model space.
 *
 * @param vlight  Light to process.
 * @param yaw     Yaw rotation angle.
 * @param pitch   Pitch rotation angle.
 * @param invert  @c true= flip light normal (for use with inverted models).
 *
 * @todo Construct a rotation matrix once and use it for all lights.
 */
static Vector3f rotateLightVector(VectorLightData const &vlight, dfloat yaw, dfloat pitch,
    bool invert = false) {
    dfloat rotated[3]; vlight.direction.decompose(rotated);
    M_RotateVector(rotated, yaw, pitch);
    if (invert) {
        rotated[0] = -rotated[0];
        rotated[1] = -rotated[1];
    }
    return Vector3f(rotated);
}
/**
 * Calculate vertex lighting.
 */
static void Mod_VertexColors(Vector4ub *out, dint count, Vector3f const *normCoords,
    duint lightListIdx, duint maxLights, Vector4f const &ambient, bool invert,
    dfloat rotateYaw, dfloat rotatePitch) {
    Vector4f const saturated(1, 1, 1, 1);
    for (dint i = 0; i < count; ++i, out++, normCoords++) {
        if (activeLod && !activeLod->hasVertex(i))
            continue;
        Vector3f const &normal = *normCoords;
        dint numProcessed = 0;
        Vector3f accum[2];  // Begin with total darkness [color, extra].
        ClientApp::renderSystem().forAllVectorLights(lightListIdx, [&maxLights, &invert, &rotateYaw
                                                      , &rotatePitch, &normal
                                                      , &accum, &numProcessed] (VectorLightData const &vlight) {
            numProcessed += 1;
            Vector3f const lightDirection
                    = rotateLightVector(vlight, rotateYaw, rotatePitch, invert);
            dfloat strength = lightDirection.dot(normal)
                            + vlight.offset;  // Shift a bit towards the light.
            if (strength > 0) strength *= vlight.lightSide;
            else             strength *= vlight.darkSide;
            accum[vlight.affectedByAmbient? 0 : 1]
                += vlight.color * de::clamp(-1.f, strength, 1.f);
            return (maxLights && duint(numProcessed) == maxLights);
        });
        Vector4f color(accum[0].max(ambient) + accum[1], ambient[3]);
        *out = (color.min(saturated) * 255).toVector4ub();
    }
}
/**
 * Set all the colors in the array to bright white.
 */
static void Mod_FullBrightVertexColors(dint count, Vector4ub *colorCoords, dfloat alpha) {
    DENG2_ASSERT(colorCoords);
    for (; count-- > 0; colorCoords++) {
        *colorCoords = Vector4ub(255, 255, 255, 255 * alpha);
    }
}
/**
 * Set all the colors into the array to the same values.
 */
static void Mod_FixedVertexColors(dint count, Vector4ub *colorCoords, Vector4ub const &color) {
    DENG2_ASSERT(colorCoords);
    for (; count-- > 0; colorCoords++) {
        *colorCoords = color;
    }
}
/**
 * Calculate cylindrically mapped, shiny texture coordinates.
 */
static void Mod_ShinyCoords(Vector2f *out, int count, Vector3f const *normCoords,
    float normYaw, float normPitch, float shinyAng, float shinyPnt, float reactSpeed) {
    for (int i = 0; i < count; ++i, out++, normCoords++) {
        if (activeLod && !activeLod->hasVertex(i))
            continue;
        float rotatedNormal[3] = { normCoords->x, normCoords->y, normCoords->z };
        M_RotateVector(rotatedNormal,
                       (shinyPnt + normYaw) * 360 * reactSpeed,
                       (shinyAng + normPitch - .5f) * 180 * reactSpeed);
        *out = Vector2f(rotatedNormal[0] + 1, rotatedNormal[2]);
    }
}
static int chooseSelSkin(FrameModelDef &mf, int submodel, int selector) {
    if (mf.def.hasSub(submodel)) {
        Record &subDef = mf.def.sub(submodel);
        int i = (selector >> DDMOBJ_SELECTOR_SHIFT) & subDef.geti("selSkinMask");
        int c = subDef.geti("selSkinShift");
        if (c > 0) i >>= c;
        else       i <<= -c;
        if (i > 7) i = 7; // Maximum number of skins for selskin.
        if (i < 0) i = 0; // Improbable (impossible?), but doesn't hurt.
        return subDef.geta("selSkins")[i].asInt();
    }
    return 0;
}
static int chooseSkin(FrameModelDef &mf, int submodel, int id, int selector, int tmap) {
    if (submodel >= int(mf.subCount())) {
        return 0;
    }
    SubmodelDef &smf = mf.subModelDef(submodel);
    FrameModel &mdl = App_Resources().model(smf.modelId);
    int skin = smf.skin;
    if (smf.testFlag(MFF_SELSKIN)) {
        skin = chooseSelSkin(mf, submodel, selector);
    }
    if (smf.skinRange > 1) {
        int offset;
        if (smf.testFlag(MFF_IDSKIN)) {
            offset = id;
        }
        else {
            offset = SECONDS_TO_TICKS(App_World().time()) / mf.skinTics;
        }
        skin += offset % smf.skinRange;
    }
    if (smf.testFlag(MFF_SKINTRANS)) {
        skin = tmap;
    }
    if (skin < 0 || skin >= mdl.skinCount()) {
        skin = 0;
    }
    return skin;
}
static inline MaterialVariantSpec const &modelSkinMaterialSpec() {
    return ClientApp::resources().materialSpec(ModelSkinContext, 0, 0, 0, 0, GL_REPEAT, GL_REPEAT,
                                 1, -2, -1, true, true, false, false);
}
static void drawSubmodel(uint number, vissprite_t const &spr) {
    drawmodelparams_t const &parm = *VS_MODEL(&spr);
    int const zSign = (spr.pose.mirrored? -1 : 1);
    FrameModelDef *mf = parm.mf, *mfNext = parm.nextMF;
    SubmodelDef const &smf = mf->subModelDef(number);
    FrameModel &mdl = App_Resources().model(smf.modelId);
    if (mf->scale == Vector3f(0, 0, 0))
        return;
    float alpha = spr.light.ambientColor[CA];
    if (smf.testFlag(MFF_BRIGHTSHADOW2) ||
       !(parm.flags & (DDMF_BRIGHTSHADOW|DDMF_SHADOW|DDMF_ALTSHADOW))) {
        alpha *= smf.alpha / 255.f;
    }
    if (alpha <= 0) return;
    blendmode_t blending = smf.blendMode;
    if (parm.flags & DDMF_BRIGHTSHADOW) {
        blending = BM_ADD;
    }
    int useSkin = chooseSkin(*mf, number, parm.id, parm.selector, parm.tmap);
    float endPos = 0;
    float inter = parm.inter;
    if ((mf->interRange[0] == 0 && mf->interRange[1] == 1) || smf.testFlag(MFF_WORLD_TIME_ANIM)) {
        endPos = (mf->interNext ? mf->interNext->interMark : 1);
        inter = (parm.inter - mf->interMark) / (endPos - mf->interMark);
    }
    FrameModelFrame *frame = &visibleModelFrame(*mf, number, parm.id);
    FrameModelFrame *nextFrame = 0;
    if (parm.alwaysInterpolate) {
        nextFrame = &mdl.frame((smf.frame + 1) % mdl.frameCount());
        mfNext = mf;
    }
    else {
        if (frameInter && mfNext && !smf.testFlag(MFF_DONT_INTERPOLATE)) {
            if (mfNext->hasSub(number) && mfNext->subModelId(number) == smf.modelId) {
                nextFrame = &visibleModelFrame(*mfNext, number, parm.id);
            }
        }
    }
    inter = de::clamp(0.f, inter, 1.f);
    if (!nextFrame) {
        nextFrame = frame;
        mfNext = mf;
    }
    int numVerts = mdl.vertexCount();
    if (!resizeVertexBuffer(numVerts)) {
        return;
    }
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    DGL_Translatef(spr.pose.origin[VX] + spr.pose.srvo[VX] +
                   de::lerp(mf->offset.x, mfNext->offset.x, inter),
                   spr.pose.origin[VZ] + spr.pose.srvo[VZ] +
                   de::lerp(mf->offset.y, mfNext->offset.y, inter),
                   spr.pose.origin[VY] + spr.pose.srvo[VY] + zSign *
                   de::lerp(mf->offset.z, mfNext->offset.z, inter));
    if (spr.pose.extraYawAngle || spr.pose.extraPitchAngle) {
        DGL_Scalef(1, 200 / 240.0f, 1);
        DGL_Rotatef(spr.pose.extraYawAngle, 1, 0, 0);
        DGL_Rotatef(spr.pose.extraPitchAngle, 0, 0, 1);
        DGL_Scalef(1, 240 / 200.0f, 1);
    }
    DGL_Rotatef(spr.pose.viewAligned? spr.pose.yawAngleOffset   : spr.pose.yaw,   0, 1, 0);
    DGL_Rotatef(spr.pose.viewAligned? spr.pose.pitchAngleOffset : spr.pose.pitch, 0, 0, 1);
    DGL_Scalef(de::lerp(mf->scale.x, mfNext->scale.x, inter),
             de::lerp(mf->scale.y, mfNext->scale.y, inter),
             de::lerp(mf->scale.z, mfNext->scale.z, inter));
    if (spr.pose.extraScale) {
        DGL_Scalef(spr.pose.extraScale, spr.pose.extraScale, spr.pose.extraScale);
    }
    DGL_Translatef(smf.offset.x, smf.offset.y, smf.offset.z);
    if (mdl.lodCount() > 1 && rend_model_lod != 0) {
        float lodFactor = rend_model_lod * DENG_GAMEVIEW_WIDTH / 640.0f / (Rend_FieldOfView() / 90.0f);
        if (!de::fequal(lodFactor, 0)) {
            lodFactor = 1 / lodFactor;
        }
        activeLod = &mdl.lod(de::clamp<int>(0, lodFactor * spr.pose.distance, mdl.lodCount() - 1));
    }
    else {
        activeLod = 0;
    }
    Mod_LerpVertices(inter, numVerts, *frame, *nextFrame,
                     modelPosCoords, modelNormCoords);
    if (zSign < 0) {
        Mod_MirrorCoords(numVerts, modelPosCoords, 2);
        Mod_MirrorCoords(numVerts, modelNormCoords, 1);
    }
    modelCenter = Vector3f(spr.pose.origin[VX], spr.pose.origin[VY], spr.pose.midZ())
            + Vector3d(spr.pose.srvo) + Vector3f(mf->offset.x, mf->offset.z, mf->offset.y);
    Vector4f ambient;
    if (smf.testFlag(MFF_FULLBRIGHT) && !smf.testFlag(MFF_DIM)) {
        ambient = Vector4f(1, 1, 1, 1);
        Mod_FullBrightVertexColors(numVerts, modelColorCoords, alpha);
    }
    else if (!spr.light.vLightListIdx) {
        ambient = Vector4f(spr.light.ambientColor, alpha);
        Mod_FixedVertexColors(numVerts, modelColorCoords,
                              (ambient * 255).toVector4ub());
    }
    else {
        ambient = Vector4f(spr.light.ambientColor, alpha);
        Mod_VertexColors(modelColorCoords, numVerts,
                         modelNormCoords, spr.light.vLightListIdx, modelLight + 1,
                         ambient, (mf->scale[VY] < 0), -spr.pose.yaw, -spr.pose.pitch);
    }
    TextureVariant *shinyTexture = 0;
    float shininess = 0;
    if (mf->def.hasSub(number)) {
        shininess = float(de::clamp(0.0, mf->def.sub(number).getd("shiny") * modelShinyFactor, 1.0));
        if (shininess > 0) {
            if (ClientTexture *tex = static_cast<ClientTexture *>(mf->subModelDef(number).shinySkin)) {
                shinyTexture = tex->prepareVariant(Rend_ModelShinyTextureSpec());
            }
            else {
                shininess = 0;
            }
        }
    }
    Vector4f color;
    if (shininess > 0) {
        Vector3f shinyColor = mf->def.sub(number).get("shinyColor");
        float offset = parm.shineYawOffset;
        float normYaw = M_CycleIntoRange(((spr.pose.viewAligned? spr.pose.yawAngleOffset
                                                               : spr.pose.yaw) + offset) / 360, 1);
        offset = parm.shinePitchOffset;
        float normPitch = M_CycleIntoRange(((spr.pose.viewAligned? spr.pose.pitchAngleOffset
                                                                 : spr.pose.pitch) + offset) / 360, 1);
        float shinyAng = 0;
        float shinyPnt = 0;
        if (parm.shinepspriteCoordSpace) {
            shinyPnt = 0.5;
        }
        else {
            Vector3f delta = modelCenter;
            if (!parm.shineTranslateWithViewerPos) {
                delta -= Rend_EyeOrigin().xzy();
            }
            shinyAng = QATAN2(delta.z, M_ApproxDistancef(delta.x, delta.y)) / PI + 0.5f; // shinyAng is [0,1]
            shinyPnt = QATAN2(delta.y, delta.x) / (2 * PI);
        }
        Mod_ShinyCoords(modelTexCoords, numVerts,
                        modelNormCoords, normYaw, normPitch, shinyAng, shinyPnt,
                        mf->def.sub(number).getf("shinyReact"));
        if (smf.testFlag(MFF_SHINY_LIT)) {
            color = Vector4f(ambient * shinyColor, shininess);
        }
        else {
            color = Vector4f(shinyColor, shininess);
        }
    }
    TextureVariant *skinTexture = 0;
    if (renderTextures == 2) {
        MaterialAnimator &matAnimator = ClientMaterial::find(de::Uri("System", Path("gray")))
                .getAnimator(modelSkinMaterialSpec());
        matAnimator.prepare();
        skinTexture = matAnimator.texUnit(MaterialAnimator::TU_LAYER0).texture;
    }
    else {
        skinTexture = 0;
        if (ClientTexture *tex = static_cast<ClientTexture *>(mdl.skin(useSkin).texture)) {
            skinTexture = tex->prepareVariant(Rend_ModelDiffuseTextureSpec(mdl.flags().testFlag(FrameModel::NoTextureCompression)));
        }
    }
    if (zSign < 0) {
        LIBGUI_GL.glFrontFace(GL_CCW);
    }
    if (smf.testFlag(MFF_TWO_SIDED)) {
        DGL_CullFace(DGL_NONE);
    }
    DGL_Enable(DGL_TEXTURE_2D);
    FrameModel::Primitives const &primitives =
        activeLod? activeLod->primitives : mdl.primitives();
    if (shininess <= 0 || alpha < 1 ||
        blending != BM_NORMAL || !smf.testFlag(MFF_SHINY_SPECULAR)) {
        if (shininess < 1 || smf.testFlag(MFF_SHINY_SPECULAR)) {
            selectTexUnits(1);
            GL_BlendMode(blending);
            GL_BindTexture(renderTextures? skinTexture : 0);
            drawPrimitives(RC_COMMAND_COORDS, primitives,
                           modelPosCoords, modelColorCoords);
        }
        if (shininess > 0) {
            DGL_DepthFunc(DGL_LEQUAL);
            if (smf.testFlag(MFF_SHINY_SPECULAR))
                GL_BlendMode(BM_ADD);
            else
                GL_BlendMode(BM_NORMAL);
            Mod_FixedVertexColors(numVerts, modelColorCoords,
                                  (color * 255).toVector4ub());
            selectTexUnits(2);
            DGL_ModulateTexture(11);
            DGL_SetInteger(DGL_ACTIVE_TEXTURE, 1);
            GL_BindTexture(renderTextures? shinyTexture : 0);
            DGL_SetInteger(DGL_ACTIVE_TEXTURE, 0);
            GL_BindTexture(renderTextures? skinTexture : 0);
            drawPrimitives(RC_BOTH_COORDS, primitives,
                           modelPosCoords, modelColorCoords, modelTexCoords);
            selectTexUnits(1);
            DGL_ModulateTexture(1);
        }
    }
    else {
        GL_BlendMode(blending);
        selectTexUnits(2);
        DGL_ModulateTexture(10);
        DGL_SetInteger(DGL_ACTIVE_TEXTURE, 1);
        GL_BindTexture(renderTextures? shinyTexture : 0);
        float colorv1[] = { color.x * color.w, color.y * color.w, color.z * color.w, color.w };
        DGL_SetModulationColor(colorv1);
        DGL_SetInteger(DGL_ACTIVE_TEXTURE, 0);
        GL_BindTexture(renderTextures? skinTexture : 0);
        drawPrimitives(RC_BOTH_COORDS, primitives,
                       modelPosCoords, modelColorCoords, modelTexCoords);
        selectTexUnits(1);
        DGL_ModulateTexture(1);
    }
    DGL_Disable(DGL_TEXTURE_2D);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
    if (smf.testFlag(MFF_TWO_SIDED)) {
        DGL_CullFace(DGL_BACK);
    }
    if (zSign < 0) {
        LIBGUI_GL.glFrontFace(GL_CW);
    }
    DGL_DepthFunc(DGL_LESS);
    GL_BlendMode(BM_NORMAL);
}
void Rend_DrawModel(vissprite_t const &spr) {
    drawmodelparams_t const &parm = *VS_MODEL(&spr);
    DENG2_ASSERT(inited);
    DENG_ASSERT_IN_MAIN_THREAD();
    DENG_ASSERT_GL_CONTEXT_ACTIVE();
    if (!parm.mf) return;
    DENG2_ASSERT(parm.mf->select == (parm.selector & DDMOBJ_SELECTOR_MASK))
    for (uint i = 0; i < parm.mf->subCount(); ++i) {
        if (parm.mf->subModelId(i)) {
            bool disableZ = (parm.mf->flags & MFF_DISABLE_Z_WRITE ||
                             parm.mf->testSubFlag(i, MFF_DISABLE_Z_WRITE));
            if (disableZ) {
                DGL_Disable(DGL_DEPTH_WRITE);
            }
            drawSubmodel(i, spr);
            if (disableZ) {
                DGL_Enable(DGL_DEPTH_WRITE);
            }
        }
    }
    if (devMobjVLights && spr.light.vLightListIdx) {
        DGL_PushState();
        DGL_Disable(DGL_DEPTH_TEST);
        DGL_CullFace(DGL_NONE);
        DGL_MatrixMode(DGL_MODELVIEW);
        DGL_PushMatrix();
        DGL_Translatef(spr.pose.origin[0], spr.pose.origin[2], spr.pose.origin[1]);
        coord_t const distFromViewer = de::abs(spr.pose.distance);
        ClientApp::renderSystem().forAllVectorLights(spr.light.vLightListIdx, [&distFromViewer] (VectorLightData const &vlight) {
            if (distFromViewer < 1600 - 8) {
                Rend_DrawVectorLight(vlight, 1 - distFromViewer / 1600);
            }
            return LoopContinue;
        });
        DGL_MatrixMode(DGL_MODELVIEW);
        DGL_PopMatrix();
        DGL_PopState();
    }
}
TextureVariantSpec const &Rend_ModelDiffuseTextureSpec(bool noCompression) {
    return ClientApp::resources().textureSpec(TC_MODELSKIN_DIFFUSE,
        (noCompression? TSF_NO_COMPRESSION : 0), 0, 0, 0, GL_REPEAT, GL_REPEAT,
        1, -2, -1, true, true, false, false);
}
TextureVariantSpec const &Rend_ModelShinyTextureSpec() {
    return ClientApp::resources().textureSpec(TC_MODELSKIN_REFLECTION,
        TSF_NO_COMPRESSION, 0, 0, 0, GL_REPEAT, GL_REPEAT, 1, -2, -1, false,
        false, false, false);
}
#define NUM_TEX_NAMES (MAX_PTC_TEXTURES)
static DGLuint pointTex, ptctexname[MAX_PTC_TEXTURES];
static bool hasPoints, hasLines, hasModels, hasNoBlend, hasAdditive;
static bool hasPointTexs[NUM_TEX_NAMES];
struct OrderedParticle {
    Generator const *generator;
    dint particleId;
    dfloat distance;
};
static OrderedParticle *order;
static size_t orderSize;
static size_t numParts;
/*
 * Console variables:
 */
dbyte useParticles = true;
static dint maxParticles;           ///< @c 0= Unlimited.
static dint particleNearLimit;
static dfloat particleDiffuse = 4;
static dfloat pointDist(fixed_t const c[3]) {
    viewdata_t const *viewData = &viewPlayer->viewport();
    dfloat dist = ((viewData->current.origin.y - FIX2FLT(c[1])) * -viewData->viewSin)
                - ((viewData->current.origin.x - FIX2FLT(c[0])) * viewData->viewCos);
    return de::abs(dist);  // Always return positive.
}
static Path tryFindImage(String name) {
    try {
        String foundPath = App_FileSystem().findPath(de::Uri("Textures", name + "-ck"),
                                                     RLF_DEFAULT, App_ResourceClass(RC_GRAPHIC));
        return App_BasePath() / foundPath;
    }
    catch(FS1::NotFoundError const&)
    {}  // Ignore this error.
    try {
        String foundPath = App_FileSystem().findPath(de::Uri("Textures", name),
                                                     RLF_DEFAULT, App_ResourceClass(RC_GRAPHIC));
        return App_BasePath() / foundPath;
    }
    catch(FS1::NotFoundError const&)
    {}  // Ignore this error.
    return Path();  // Not found.
}
static dbyte loadParticleTexture(duint particleTex) {
    DENG2_ASSERT(particleTex < MAX_PTC_TEXTURES);
    image_t image;
    try {
        String const assetId = QStringLiteral("texture.particle.%1").arg(particleTex, 2, 10, QChar('0'));
        if (App::assetExists(assetId)) {
            auto asset = App::asset(assetId);
            ImageFile const &img = App::rootFolder().locate<ImageFile const>
                    (asset.absolutePath(QStringLiteral("path")));
            Image_InitFromImage(image, img.image());
        }
        else {
            auto particleImageName = String("Particle%1").arg(particleTex, 2, 10, QChar('0'));
            Path foundPath = tryFindImage(particleImageName);
            if (foundPath.isEmpty())
                return 0;
            if (!GL_LoadImage(image, foundPath.toUtf8().constData())) {
                LOG_RES_WARNING("Failed to load \"%s\"") << NativePath(foundPath).pretty();
                return 0;
            }
        }
    }
    catch (Error const &er) {
        LOG_RES_ERROR("Failed to load texture for particle %i: %s")
                << particleTex << er.asText();
        return 0;
    }
    if (image.pixelSize == 1)
        Image_ConvertToAlpha(image, true);
    ptctexname[particleTex] = GL_NewTextureWithParams(
        image.pixelSize == 4 ? DGL_RGBA :
        image.pixelSize == 2 ? DGL_LUMINANCE_PLUS_A8 : DGL_RGB,
        image.size.x, image.size.y, image.pixels,
        TXCF_NO_COMPRESSION);
    Image_ClearPixelData(image);
    return 2; // External
}
void Rend_ParticleLoadSystemTextures() {
    if(novideo) return;
    if(!pointTex) {
        image_t image;
        if(GL_LoadExtImage(image, "Zeroth", LGM_WHITE_ALPHA)) {
            pointTex = GL_NewTextureWithParams(
                ( image.pixelSize == 2 ? DGL_LUMINANCE_PLUS_A8 :
                  image.pixelSize == 3 ? DGL_RGB :
                  image.pixelSize == 4 ? DGL_RGBA : DGL_LUMINANCE ),
                image.size.x, image.size.y, image.pixels,
                ( TXCF_MIPMAP | TXCF_NO_COMPRESSION ),
                0, glmode[mipmapping], GL_LINEAR, 0 /*no anisotropy*/, GL_CLAMP_TO_EDGE, GL_CLAMP_TO_EDGE);
            DENG2_ASSERT(pointTex != 0);
        }
        Image_ClearPixelData(image);
    }
}
void Rend_ParticleLoadExtraTextures() {
    if(novideo) return;
    Rend_ParticleReleaseExtraTextures();
    if(!App_GameLoaded()) return;
    QList<dint> loaded;
    for(dint i = 0; i < MAX_PTC_TEXTURES; ++i) {
        if(loadParticleTexture(i)) {
            loaded.append(i);
        }
    }
    if(!loaded.isEmpty()) {
        LOG_RES_NOTE("Loaded textures for particle IDs: %s") << Rangei::contiguousRangesAsText(loaded);
    }
}
void Rend_ParticleReleaseSystemTextures() {
    if(novideo) return;
    Deferred_glDeleteTextures(1, (GLuint const *) &pointTex);
    pointTex = 0;
}
void Rend_ParticleReleaseExtraTextures() {
    if(novideo) return;
    Deferred_glDeleteTextures(NUM_TEX_NAMES, (GLuint const *) ptctexname);
    de::zap(ptctexname);
}
/**
 * Sorts in descending order.
 */
static dint comparePOrder(void const *a, void const *b) {
    auto const &ptA = *(OrderedParticle const *) a;
    auto const &ptB = *(OrderedParticle const *) b;
    if(ptA.distance > ptB.distance) return -1;
    if(ptA.distance < ptB.distance) return 1;
    return 0;  // Highly unlikely (but possible).
}
/**
 * Allocate more memory for the particle ordering buffer, if necessary.
 */
static void expandOrderBuffer(size_t max) {
    size_t currentSize = orderSize;
    if(!orderSize) {
        orderSize = MAX_OF(max, 256);
    }
    else {
        while(max > orderSize) {
            orderSize *= 2;
        }
    }
    if(orderSize > currentSize) {
        order = (OrderedParticle *) Z_Realloc(order, sizeof(OrderedParticle) * orderSize, PU_APPSTATIC);
    }
}
/**
 * Determines whether the given particle is potentially visible for the current viewer.
 */
static bool particlePVisible(ParticleInfo const &pinfo) {
    if(pinfo.stage < 0) return false;
    if(!pinfo.bspLeaf || !pinfo.bspLeaf->hasSubspace())
        return false;
    return R_ViewerSubspaceIsVisible(pinfo.bspLeaf->subspace());
}
/**
 * @return  @c true if there are particles to be drawn.
 */
static dint listVisibleParticles(world::Map &map) {
    ::hasPoints = ::hasModels = ::hasLines = false;
    ::hasAdditive = ::hasNoBlend = false;
    de::zap(::hasPointTexs);
    ::numParts = 0;
    map.forAllGenerators([] (Generator &gen) {
        if(R_ViewerGeneratorIsVisible(gen)) {
            ::numParts += gen.activeParticleCount();
        }
        return LoopContinue;
    });
    if(!::numParts) return false;
    expandOrderBuffer(::numParts);
    size_t numVisibleParts = 0;
    map.forAllGenerators([&numVisibleParts] (Generator &gen) {
        if(!R_ViewerGeneratorIsVisible(gen)) return LoopContinue;  // Skip.
        for(dint i = 0; i < gen.count; ++i) {
            ParticleInfo const &pinfo = gen.particleInfo()[i];
            if(!particlePVisible(pinfo)) continue;  // Skip.
            dfloat const dist = de::max(pointDist(pinfo.origin), 1.f);
            if(gen.def->maxDist != 0 && dist > gen.def->maxDist) continue;
            if(dist < dfloat( ::particleNearLimit )) continue;
            OrderedParticle *slot = &::order[numVisibleParts++];
            slot->generator  = &gen;
            slot->particleId = i;
            slot->distance   = dist;
            dint const psType = gen.stages[pinfo.stage].type;
            if(psType == PTC_POINT) {
                ::hasPoints = true;
            }
            else if(psType == PTC_LINE) {
                ::hasLines = true;
            }
            else if(psType >= PTC_TEXTURE && psType < PTC_TEXTURE + MAX_PTC_TEXTURES) {
                if(::ptctexname[psType - PTC_TEXTURE]) {
                    ::hasPointTexs[psType - PTC_TEXTURE] = true;
                }
                else {
                    ::hasPoints = true;
                }
            }
            else if(psType >= PTC_MODEL && psType < PTC_MODEL + MAX_PTC_MODELS) {
                ::hasModels = true;
            }
            if(gen.blendmode() == BM_ADD) {
                ::hasAdditive = true;
            }
            else {
                ::hasNoBlend = true;
            }
        }
        return LoopContinue;
    });
    if(!numVisibleParts) return false;
    ::numParts = numVisibleParts;
    qsort(::order, ::numParts, sizeof(OrderedParticle), comparePOrder);
    return true;
}
static void setupModelParamsForParticle(vissprite_t &spr, ParticleInfo const *pinfo,
    GeneratorParticleStage const *st, ded_ptcstage_t const *dst, Vector3f const &origin,
    dfloat dist, dfloat size, dfloat mark, dfloat alpha) {
    drawmodelparams_t &parm = *VS_MODEL(&spr);
    spr.pose.origin     = Vector3d(origin.xz(), spr.pose.topZ = origin.y);
    spr.pose.distance   = dist;
    spr.pose.extraScale = size;  // Extra scaling factor.
    parm.mf = &ClientApp::resources().modelDef(dst->model);
    parm.alwaysInterpolate = true;
    dint frame;
    if(dst->endFrame < 0) {
        frame = dst->frame;
        parm.inter = 0;
    }
    else {
        frame = dst->frame + (dst->endFrame - dst->frame) * mark;
        parm.inter = M_CycleIntoRange(mark * (dst->endFrame - dst->frame), 1);
    }
    ClientApp::resources().setModelDefFrame(*parm.mf, frame);
    if(parm.mf->testSubFlag(0, MFF_MOVEMENT_YAW)) {
        spr.pose.yaw = R_MovementXYYaw(FIX2FLT(pinfo->mov[0]), FIX2FLT(pinfo->mov[1]));
    }
    else {
        spr.pose.yaw = pinfo->yaw / 32768.0f * 180;
    }
    if(parm.mf->testSubFlag(0, MFF_MOVEMENT_PITCH)) {
        spr.pose.pitch = R_MovementXYZPitch(FIX2FLT(pinfo->mov[0]), FIX2FLT(pinfo->mov[1]), FIX2FLT(pinfo->mov[2]));
    }
    else {
        spr.pose.pitch = pinfo->pitch / 32768.0f * 180;
    }
    spr.light.ambientColor.w = alpha;
    if(st->flags.testFlag(GeneratorParticleStage::Bright) || levelFullBright) {
        spr.light.ambientColor.x = spr.light.ambientColor.y = spr.light.ambientColor.z = 1;
        spr.light.vLightListIdx = 0;
    }
    else {
        world::Map &map = pinfo->bspLeaf->subspace().sector().map();
#if 0
        if(useBias && map.hasLightGrid()) {
            Vector4f color = map.lightGrid().evaluate(spr.pose.origin);
            for(dint i = 0; i < 3; ++i) {
                color[i] += Rend_LightAdaptationDelta(color[i]);
            }
            spr.light.ambientColor.x = color.x;
            spr.light.ambientColor.y = color.y;
            spr.light.ambientColor.z = color.z;
        }
        else
#endif
        {
            Vector4f const color = pinfo->bspLeaf->subspace().subsector().as<world::ClientSubsector>()
                                       .lightSourceColorfIntensity();
            dfloat lightLevel = color.w;
            lightLevel = Rend_AttenuateLightLevel(spr.pose.distance, lightLevel);
            lightLevel += Rend_ExtraLightDelta();
            Rend_ApplyLightAdaptation(lightLevel);
            for(dint i = 0; i < 3; ++i) {
                spr.light.ambientColor[i] = lightLevel * color[i];
            }
        }
        Rend_ApplyTorchLight(spr.light.ambientColor, spr.pose.distance);
        spr.light.vLightListIdx =
                Rend_CollectAffectingLights(spr.pose.origin, spr.light.ambientColor,
                                            map.bspLeafAt(spr.pose.origin).subspacePtr());
    }
}
/**
 * Calculate a unit vector parallel to @a line.
 *
 * @todo No longer needed (Surface has tangent space vectors).
 *
 * @param unitVect  Unit vector is written here.
 */
static Vector2f lineUnitVector(Line const &line) {
    ddouble len = M_ApproxDistance(line.direction().x, line.direction().y);
    if(len) {
        return line.direction() / len;
    }
    return Vector2f();
}
static void drawParticles(dint rtype, bool withBlend) {
    DENG2_ASSERT_IN_RENDER_THREAD();
    DENG_ASSERT_GL_CONTEXT_ACTIVE();
    viewdata_t const *viewData = &viewPlayer->viewport();
    Vector3f const leftoff     = viewData->upVec + viewData->sideVec;
    Vector3f const rightoff    = viewData->upVec - viewData->sideVec;
    DGLuint tex = 0;
    if (rtype == PTC_POINT
       || (rtype >= PTC_TEXTURE && rtype < PTC_TEXTURE + MAX_PTC_TEXTURES)) {
        if (renderTextures) {
            if (rtype == PTC_POINT || 0 == ptctexname[rtype - PTC_TEXTURE])
                tex = pointTex;
            else
                tex = ptctexname[rtype - PTC_TEXTURE];
        }
    }
    dglprimtype_t primType = DGL_QUADS;
    if (rtype == PTC_MODEL) {
        DGL_Enable(DGL_DEPTH_WRITE);
        DGL_Enable(DGL_DEPTH_TEST);
    }
    else if (tex != 0) {
        DGL_Disable(DGL_DEPTH_WRITE);
        DGL_DepthFunc(DGL_LEQUAL);
        DGL_CullFace(DGL_NONE);
        GL_BindTextureUnmanaged(tex, gl::ClampToEdge, gl::ClampToEdge);
        DGL_Enable(DGL_TEXTURE_2D);
        DGL_Begin(primType = DGL_QUADS);
    }
    else {
        DGL_Begin(primType = DGL_LINES);
    }
    size_t i = 0;
    if (maxParticles) {
        i = numParts - (unsigned) maxParticles;
    }
    blendmode_t mode = BM_NORMAL, newMode;
    for (; i < numParts; ++i) {
        OrderedParticle const *slot = &order[i];
        Generator const *gen        = slot->generator;
        ParticleInfo const &pinfo   = gen->particleInfo()[slot->particleId];
        GeneratorParticleStage const *st = &gen->stages[pinfo.stage];
        ded_ptcstage_t const *stDef      = &gen->def->stages[pinfo.stage];
        dshort stageType = st->type;
        if (stageType >= PTC_TEXTURE && stageType < PTC_TEXTURE + MAX_PTC_TEXTURES &&
            0 == ptctexname[stageType - PTC_TEXTURE]) {
            stageType = PTC_POINT;
        }
        if ((rtype == PTC_MODEL && stDef->model < 0) ||
            (rtype != PTC_MODEL && stageType != rtype)) {
            continue;
        }
        if (rtype >= PTC_TEXTURE && rtype < PTC_TEXTURE + MAX_PTC_TEXTURES &&
            0 == ptctexname[rtype - PTC_TEXTURE])
            continue;
        if ((gen->blendmode() != BM_ADD) == withBlend)
            continue;
        if (rtype != PTC_MODEL && !withBlend) {
            newMode = gen->blendmode();
            if(newMode != mode) {
                DGL_End();
                GL_BlendMode(mode = newMode);
                DGL_Begin(primType);
            }
        }
        ded_ptcstage_t const *nextStDef;
        if (pinfo.stage >= gen->def->stages.size() - 1 ||
            !gen->stages[pinfo.stage + 1].type) {
            nextStDef = &gen->def->stages[pinfo.stage];
        }
        else {
            nextStDef = &gen->def->stages[pinfo.stage + 1];
        }
        dfloat const inter = 1 - dfloat( pinfo.tics ) / stDef->tics;
        dfloat size = de::lerp(    stDef->particleRadius(slot->particleId),
                               nextStDef->particleRadius(slot->particleId), inter);
        if(!size) continue;
        Vector4f color = de::lerp(Vector4f(stDef->color), Vector4f(nextStDef->color), inter);
        if (!st->flags.testFlag(GeneratorParticleStage::Bright) && !levelFullBright) {
            if (world::ConvexSubspace *subspace = pinfo.bspLeaf->subspacePtr()) {
                dfloat const intensity = subspace->subsector().as<world::ClientSubsector>()
                                            .lightSourceIntensity();
                color *= Vector4f(intensity, intensity, intensity, 1);
            }
        }
        dfloat const maxDist = gen->def->maxDist;
        dfloat const dist    = order[i].distance;
        if(maxDist) {
            if(dist > maxDist * .75f) {
                color.w *= 1 - (dist - maxDist * .75f) / (maxDist * .25f);
            }
        }
        if(particleDiffuse > 0) {
            if(dist < particleDiffuse * size) {
                color.w -= 1 - dist / (particleDiffuse * size);
            }
        }
        if(color.w <= 0)
            continue;
        DGL_Color4f(color.x, color.y, color.z, color.w);
        bool const nearWall = (pinfo.contact && !pinfo.mov[0] && !pinfo.mov[1]);
        bool nearPlane = false;
        if (world::ConvexSubspace *space = pinfo.bspLeaf->subspacePtr()) {
            auto &subsec = space->subsector().as<world::ClientSubsector>();
            if (   FLT2FIX(subsec.  visFloor().heightSmoothed()) + 2 * FRACUNIT >= pinfo.origin[2]
                || FLT2FIX(subsec.visCeiling().heightSmoothed()) - 2 * FRACUNIT <= pinfo.origin[2]) {
                nearPlane = true;
            }
        }
        bool flatOnPlane = false, flatOnWall = false;
        if(stageType == PTC_POINT ||
           (stageType >= PTC_TEXTURE && stageType < PTC_TEXTURE + MAX_PTC_TEXTURES)) {
            if(st->flags.testFlag(GeneratorParticleStage::PlaneFlat) && nearPlane)
                flatOnPlane = true;
            if(st->flags.testFlag(GeneratorParticleStage::WallFlat) && nearWall)
                flatOnWall = true;
        }
        Vector3f center = gen->particleOrigin(pinfo).xzy();
        if(!flatOnPlane && !flatOnWall) {
            Vector3f offset(frameTimePos, nearPlane ? 0 : frameTimePos, frameTimePos);
            center += offset * gen->particleMomentum(pinfo).xzy();
        }
        if(rtype == PTC_MODEL && stDef->model >= 0) {
            vissprite_t temp;
            setupModelParamsForParticle(temp, &pinfo, st, stDef, center, dist, size, inter, color.w);
            Rend_DrawModel(temp);
            continue;
        }
        if(tex != 0) {
            if(flatOnPlane) {
                DGL_TexCoord2f(0, 0, 0);
                DGL_Vertex3f(center.x - size, center.y, center.z - size);
                DGL_TexCoord2f(0, 1, 0);
                DGL_Vertex3f(center.x + size, center.y, center.z - size);
                DGL_TexCoord2f(0, 1, 1);
                DGL_Vertex3f(center.x + size, center.y, center.z + size);
                DGL_TexCoord2f(0, 0, 1);
                DGL_Vertex3f(center.x - size, center.y, center.z + size);
            }
            else if(flatOnWall) {
                DENG2_ASSERT(pinfo.contact);
                Line const &contact = *pinfo.contact;
                vec2d_t origin;
                V2d_Set(origin, FIX2FLT(pinfo.origin[0]), FIX2FLT(pinfo.origin[1]));
                vec2d_t projected;
                V2d_ProjectOnLine(projected, origin,
                                  contact.from().origin().data().baseAs<ddouble>(),
                                  contact.direction().data().baseAs<ddouble>());
                ddouble const gap = -1;  // 1 map unit.
                ddouble diff[2], dist;
                V2d_Subtract(diff, projected, origin);
                if((dist = V2d_Length(diff)) != 0) {
                    projected[0] += diff[0] / dist * gap;
                    projected[1] += diff[1] / dist * gap;
                }
                Vector2f unitVec = lineUnitVector(*pinfo.contact);
                DGL_TexCoord2f(0, 0, 0);
                DGL_Vertex3f(projected[0] - size * unitVec.x, center.y - size,
                           projected[1] - size * unitVec.y);
                DGL_TexCoord2f(0, 1, 0);
                DGL_Vertex3f(projected[0] - size * unitVec.x, center.y + size,
                           projected[1] - size * unitVec.y);
                DGL_TexCoord2f(0, 1, 1);
                DGL_Vertex3f(projected[0] + size * unitVec.x, center.y + size,
                           projected[1] + size * unitVec.y);
                DGL_TexCoord2f(0, 0, 1);
                DGL_Vertex3f(projected[0] + size * unitVec.x, center.y - size,
                           projected[1] + size * unitVec.y);
            }
            else {
                DGL_TexCoord2f(0, 0, 0);
                DGL_Vertex3f(center.x + size * leftoff.x,
                           center.y + size * leftoff.y / 1.2f,
                           center.z + size * leftoff.z);
                DGL_TexCoord2f(0, 1, 0);
                DGL_Vertex3f(center.x + size * rightoff.x,
                           center.y + size * rightoff.y / 1.2f,
                           center.z + size * rightoff.z);
                DGL_TexCoord2f(0, 1, 1);
                DGL_Vertex3f(center.x - size * leftoff.x,
                           center.y - size * leftoff.y / 1.2f,
                           center.z - size * leftoff.z);
                DGL_TexCoord2f(0, 0, 1);
                DGL_Vertex3f(center.x - size * rightoff.x,
                           center.y - size * rightoff.y / 1.2f,
                           center.z - size * rightoff.z);
            }
        }
        else  // It's a line.
        {
            DGL_Vertex3f(center.x, center.y, center.z);
            DGL_Vertex3f(center.x - FIX2FLT(pinfo.mov[0]),
                         center.y - FIX2FLT(pinfo.mov[2]),
                         center.z - FIX2FLT(pinfo.mov[1]));
        }
    }
    if(rtype != PTC_MODEL) {
        DGL_End();
        if(tex != 0) {
            DGL_CullFace(DGL_BACK);
            DGL_Enable(DGL_DEPTH_WRITE);
            DGL_DepthFunc(DGL_LESS);
            DGL_Disable(DGL_TEXTURE_2D);
        }
    }
    if(!withBlend) {
        GL_BlendMode(BM_NORMAL);
    }
}
static void renderPass(bool useBlending) {
    DENG2_ASSERT(!Sys_GLCheckError());
    if(useBlending) {
        GL_BlendMode(BM_ADD);
    }
    if(hasModels) {
        drawParticles(PTC_MODEL, useBlending);
    }
    if(hasLines) {
        drawParticles(PTC_LINE, useBlending);
    }
    if(hasPoints) {
        drawParticles(PTC_POINT, useBlending);
    }
    for(dint i = 0; i < NUM_TEX_NAMES; ++i) {
        if(hasPointTexs[i]) {
            drawParticles(PTC_TEXTURE + i, useBlending);
        }
    }
    if(useBlending) {
        GL_BlendMode(BM_NORMAL);
    }
    DENG2_ASSERT(!Sys_GLCheckError());
}
void Rend_RenderParticles(world::Map &map) {
    if(!useParticles) return;
    if(!listVisibleParticles(map)) return;
    if(hasNoBlend) {
        renderPass(false);
    }
    if(hasAdditive) {
        renderPass(true);
    }
}
void Rend_ParticleRegister() {
    C_VAR_BYTE ("rend-particle",                   &useParticles,      0,              0, 1);
    C_VAR_INT  ("rend-particle-max",               &maxParticles,      CVF_NO_MAX,     0, 0);
    C_VAR_FLOAT("rend-particle-diffuse",           &particleDiffuse,   CVF_NO_MAX,     0, 0);
    C_VAR_INT  ("rend-particle-visible-near",      &particleNearLimit, CVF_NO_MAX,     0, 0);
}
DENG2_PIMPL(RenderSystem)
, DENG2_OBSERVES(PackageLoader, Load)
, DENG2_OBSERVES(PackageLoader, Unload) {
    Binder binder;
    Record renderModule;
    render::Environment environment;
    ModelRenderer models;
    SkyDrawable sky;
    ConfigProfiles settings;
    ConfigProfiles appearanceSettings;
    ImageBank images;
    AngleClipper clipper;
    Store buffer;
    DrawLists drawLists;
    GLUniform uMapTime          { "uMapTime",          GLUniform::Float };
    GLUniform uViewMatrix       { "uViewMatrix",       GLUniform::Mat4  };
    GLUniform uProjectionMatrix { "uProjectionMatrix", GLUniform::Mat4  };
    struct ProjectionLists {
        duint listCount = 0;
        duint cursorList = 0;
        ProjectionList *lists = nullptr;
        void init() {
            ProjectionList::init();
            lists = nullptr;
            listCount = 0;
            cursorList = 0;
        }
        void reset() {
            ProjectionList::rewind();  // start reusing list nodes.
            cursorList = 0;
            if(listCount) {
                std::memset(lists, 0, listCount * sizeof *lists);
            }
        }
        ProjectionList *tryFindList(duint listIdx) const {
            if(listIdx > 0 && listIdx <= listCount) {
                return &lists[listIdx - 1];
            }
            return nullptr;  // not found.
        }
        ProjectionList &findList(duint listIdx) const {
            if(ProjectionList *found = tryFindList(listIdx)) return *found;
            throw Error("RenderSystem::projector::findList", "Invalid index #" + String::number(listIdx));
        }
        ProjectionList &findOrCreateList(duint *listIdx, bool sortByLuma) {
            DENG2_ASSERT(listIdx);
            if(!(*listIdx)) {
                if(++cursorList >= listCount) {
                    listCount *= 2;
                    if(!listCount) listCount = 2;
                    lists = (ProjectionList *) Z_Realloc(lists, listCount * sizeof(*lists), PU_MAP);
                }
                ProjectionList *list = &lists[cursorList - 1];
                list->head       = nullptr;
                list->sortByLuma = sortByLuma;
                *listIdx = cursorList;
            }
            return lists[(*listIdx) - 1];  // 1-based index.
        }
    } projector;
    struct VectorLights {
        duint listCount = 0;
        duint cursorList = 0;
        VectorLightList *lists = nullptr;
        void init() {
            VectorLightList::init();
            lists = nullptr;
            listCount = 0;
            cursorList = 0;
        }
        void reset() {
            VectorLightList::rewind();  // start reusing list nodes.
            cursorList = 0;
            if(listCount) {
                std::memset(lists, 0, listCount * sizeof *lists);
            }
        }
        VectorLightList *tryFindList(duint listIdx) const {
            if(listIdx > 0 && listIdx <= listCount) {
                return &lists[listIdx - 1];
            }
            return nullptr;  // not found.
        }
        VectorLightList &findList(duint listIdx) const {
            if(VectorLightList *found = tryFindList(listIdx)) return *found;
            throw Error("RenderSystem::vlights::findList", "Invalid index #" + String::number(listIdx));
        }
        VectorLightList &findOrCreateList(duint *listIdx) {
            DENG2_ASSERT(listIdx);
            if(!(*listIdx)) {
                if(++cursorList >= listCount) {
                    listCount *= 2;
                    if(!listCount) listCount = 2;
                    lists = (VectorLightList *) Z_Realloc(lists, listCount * sizeof(*lists), PU_MAP);
                }
                VectorLightList *list = &lists[cursorList - 1];
                list->head = nullptr;
                *listIdx = cursorList;
            }
            return lists[(*listIdx) - 1];  // 1-based index.
        }
    } vlights;
    Impl(Public *i) : Base(i) {
        LOG_AS("RenderSystem");
        ModelRenderer::initBindings(binder, renderModule);
        ClientApp::scriptSystem().addNativeModule("Render", renderModule);
        App::packageLoader().audienceForLoad() += this;
        App::packageLoader().audienceForUnload() += this;
        App::packageLoader().load("net.dengine.client.renderer");
        App::packageLoader().load("net.dengine.client.renderer.lensflares");
        loadImages();
        typedef ConfigProfiles SReg;
        settings.define(SReg::FloatCVar, "rend-camera-fov", 95.f)
                .define(SReg::ConfigVariable, "render.pixelDensity")
                .define(SReg::IntCVar,   "rend-model-mirror-hud", 0)
                .define(SReg::IntCVar,   "rend-model-precache", 1)
                .define(SReg::IntCVar,   "rend-sprite-precache", 1)
                .define(SReg::IntCVar,   "rend-tex", 1)
                .define(SReg::IntCVar,   "rend-dev-wireframe", 0)
                .define(SReg::IntCVar,   "rend-dev-thinker-ids", 0)
                .define(SReg::IntCVar,   "rend-dev-mobj-bbox", 0)
                .define(SReg::IntCVar,   "rend-dev-polyobj-bbox", 0)
                .define(SReg::IntCVar,   "rend-dev-sector-show-indices", 0)
                .define(SReg::IntCVar,   "rend-dev-vertex-show-indices", 0)
                .define(SReg::IntCVar,   "rend-dev-generator-show-indices", 0);
        appearanceSettings.setPersistentName("renderer");
        appearanceSettings
                .define(SReg::IntCVar,   "rend-light", 1)
                .define(SReg::IntCVar,   "rend-light-decor", 1)
                .define(SReg::IntCVar,   "rend-light-blend", 0)
                .define(SReg::IntCVar,   "rend-light-num", 0)
                .define(SReg::FloatCVar, "rend-light-bright", .7f)
                .define(SReg::FloatCVar, "rend-light-fog-bright", .15f)
                .define(SReg::FloatCVar, "rend-light-radius-scale", 5.2f)
                .define(SReg::IntCVar,   "rend-light-radius-max", 320)
                .define(SReg::IntCVar,   "rend-light-ambient", 0)
                .define(SReg::FloatCVar, "rend-light-compression", 0)
                .define(SReg::IntCVar,   "rend-light-attenuation", 924)
                .define(SReg::IntCVar,   "rend-light-sky-auto", 1)
                .define(SReg::FloatCVar, "rend-light-sky", .273f)
                .define(SReg::IntCVar,   "rend-light-wall-angle-smooth", 1)
                .define(SReg::FloatCVar, "rend-light-wall-angle", 1.2f)
                .define(SReg::IntCVar,   "rend-vignette", 1)
                .define(SReg::FloatCVar, "rend-vignette-darkness", 1)
                .define(SReg::FloatCVar, "rend-vignette-width", 1)
                .define(SReg::IntCVar,   "rend-halo-realistic", 1)
                .define(SReg::IntCVar,   "rend-halo", 5)
                .define(SReg::IntCVar,   "rend-halo-bright", 45)
                .define(SReg::IntCVar,   "rend-halo-size", 80)
                .define(SReg::IntCVar,   "rend-halo-occlusion", 48)
                .define(SReg::FloatCVar, "rend-halo-radius-min", 20)
                .define(SReg::FloatCVar, "rend-halo-secondary-limit", 1)
                .define(SReg::FloatCVar, "rend-halo-dim-near", 10)
                .define(SReg::FloatCVar, "rend-halo-dim-far", 100)
                .define(SReg::FloatCVar, "rend-halo-zmag-div", 62)
                .define(SReg::FloatCVar, "rend-glow", .8f)
                .define(SReg::IntCVar,   "rend-glow-height", 100)
                .define(SReg::FloatCVar, "rend-glow-scale", 3)
                .define(SReg::IntCVar,   "rend-glow-wall", 1)
                .define(SReg::ConfigVariable, "render.fx.resize.factor")
                .define(SReg::IntCVar,   "rend-bloom", 1)
                .define(SReg::FloatCVar, "rend-bloom-intensity", .65f)
                .define(SReg::FloatCVar, "rend-bloom-threshold", .35f)
                .define(SReg::FloatCVar, "rend-bloom-dispersion", 1)
                .define(SReg::IntCVar,   "rend-fakeradio", 1)
                .define(SReg::FloatCVar, "rend-fakeradio-darkness", 1.2f)
                .define(SReg::IntCVar,   "rend-shadow", 1)
                .define(SReg::FloatCVar, "rend-shadow-darkness", 1.2f)
                .define(SReg::IntCVar,   "rend-shadow-far", 1000)
                .define(SReg::IntCVar,   "rend-shadow-radius-max", 80)
                .define(SReg::IntCVar,   "rend-tex-shiny", 1)
                .define(SReg::IntCVar,   "rend-tex-mipmap", 5)
                .define(SReg::IntCVar,   "rend-tex-quality", TEXQ_BEST)
                .define(SReg::IntCVar,   "rend-tex-anim-smooth", 1)
                .define(SReg::IntCVar,   "rend-tex-filter-smart", 0)
                .define(SReg::IntCVar,   "rend-tex-filter-sprite", 1)
                .define(SReg::IntCVar,   "rend-tex-filter-mag", 1)
                .define(SReg::IntCVar,   "rend-tex-filter-ui", 1)
                .define(SReg::IntCVar,   "rend-tex-filter-anisotropic", -1)
                .define(SReg::IntCVar,   "rend-tex-detail", 1)
                .define(SReg::FloatCVar, "rend-tex-detail-scale", 4)
                .define(SReg::FloatCVar, "rend-tex-detail-strength", .5f)
                .define(SReg::IntCVar,   "rend-mobj-smooth-move", 2)
                .define(SReg::IntCVar,   "rend-mobj-smooth-turn", 1)
                .define(SReg::IntCVar,   "rend-model", 1)
                .define(SReg::IntCVar,   "rend-model-inter", 1)
                .define(SReg::IntCVar,   "rend-model-distance", 1500)
                .define(SReg::FloatCVar, "rend-model-lod", 256)
                .define(SReg::FloatCVar, "rend-model-lights", 4)
                .define(SReg::IntCVar,   "rend-sprite-mode", 0)
                .define(SReg::IntCVar,   "rend-sprite-blend", 1)
                .define(SReg::IntCVar,   "rend-sprite-lights", 4)
                .define(SReg::IntCVar,   "rend-sprite-align", 0)
                .define(SReg::IntCVar,   "rend-sprite-noz", 0)
                .define(SReg::IntCVar,   "rend-particle", 1)
                .define(SReg::IntCVar,   "rend-particle-max", 0)
                .define(SReg::FloatCVar, "rend-particle-rate", 1)
                .define(SReg::FloatCVar, "rend-particle-diffuse", 4)
                .define(SReg::IntCVar,   "rend-particle-visible-near", 0)
                .define(SReg::FloatCVar, "rend-sky-distance", 1600);
    }
    void packageLoaded(String const &packageId) {
        FS::FoundFiles found;
        App::packageLoader().package(packageId).findPartialPath("shaders.dei", found);
        DENG2_FOR_EACH(FS::FoundFiles, i, found) {
            loadShaders(**i);
        }
    }
    void aboutToUnloadPackage(String const &packageId) {
        ClientApp::shaders().removeAllFromPackage(packageId);
    }
    /*
     * Reads all shader definitions and sets up a Bank where the actual
     * compiled shaders are stored once they're needed.
     *
     * @todo This should be reworked to support unloading packages, and
     * loading of new shaders from any newly loaded packages. -jk
     */
    void loadShaders(File const &defs) {
        LOG_MSG("Loading shader definitions from %s") << defs.description();
        ClientApp::shaders().addFromInfo(defs);
    }
    /**
     * Reads the renderer's image definitions and sets up a Bank for caching them
     * when they're needed.
     */
    void loadImages() {
    }
};
RenderSystem::RenderSystem() : d(new Impl(this))
{}
void RenderSystem::glInit() { {
        DictionaryValue defines;
        defines.add(new TextValue("DGL_BATCH_MAX"),
                    new NumberValue(DGL_BatchMaxSize()));
        shaders().setPreprocessorDefines(defines);
    }
    d->models.glInit();
}
void RenderSystem::glDeinit() {
    d->models.glDeinit();
    d->environment.glDeinit();
}
GLShaderBank &RenderSystem::shaders() {
    return BaseGuiApp::shaders();
}
ImageBank &RenderSystem::images() {
    return d->images;
}
GLUniform const &RenderSystem::uMapTime() const {
    return d->uMapTime;
}
GLUniform &RenderSystem::uProjectionMatrix() const {
    return d->uProjectionMatrix;
}
GLUniform &RenderSystem::uViewMatrix() const {
    return d->uViewMatrix;
}
render::Environment &RenderSystem::environment() {
    return d->environment;
}
ModelRenderer &RenderSystem::modelRenderer() {
    return d->models;
}
SkyDrawable &RenderSystem::sky() {
    return d->sky;
}
void RenderSystem::timeChanged(Clock const &) {
    d->uMapTime = ClientApp::world().time();
}
ConfigProfiles &RenderSystem::settings() {
    return d->settings;
}
ConfigProfiles &RenderSystem::appearanceSettings() {
    return d->appearanceSettings;
}
AngleClipper &RenderSystem::angleClipper() const {
    return d->clipper;
}
Store &RenderSystem::buffer() {
    return d->buffer;
}
void RenderSystem::clearDrawLists() {
    d->drawLists.clear();
    d->buffer.clear();
}
DrawLists &RenderSystem::drawLists() {
    return d->drawLists;
}
void RenderSystem::worldSystemMapChanged(world::Map &) {
    d->projector.init();
    d->vlights.init();
}
void RenderSystem::beginFrame() {
    d->drawLists.reset();
    d->buffer.rewind();  // Start reallocating storage from the global vertex buffer.
    d->clipper.clearRanges();
    d->projector.reset();
    d->vlights.reset();
    R_BeginFrame();
}
ProjectionList &RenderSystem::findSurfaceProjectionList(duint *listIdx, bool sortByLuma) {
    return d->projector.findOrCreateList(listIdx, sortByLuma);
}
LoopResult RenderSystem::forAllSurfaceProjections(duint listIdx, std::function<LoopResult (ProjectedTextureData const &)> func) const {
    if(ProjectionList *list = d->projector.tryFindList(listIdx)) {
        for(ProjectionList::Node *node = list->head; node; node = node->next) {
            if(auto result = func(node->projection))
                return result;
        }
    }
    return LoopContinue;
}
VectorLightList &RenderSystem::findVectorLightList(duint *listIdx) {
    return d->vlights.findOrCreateList(listIdx);
}
LoopResult RenderSystem::forAllVectorLights(duint listIdx, std::function<LoopResult (VectorLightData const &)> func) {
    if(VectorLightList *list = d->vlights.tryFindList(listIdx)) {
        for(VectorLightList::Node *node = list->head; node; node = node->next) {
            if(auto result = func(node->vlight))
                return result;
        }
    }
    return LoopContinue;
}
void RenderSystem::consoleRegister() {
    Viewports_Register();
    Rend_Register();
    H_Register();
}
enum RPolyDataType {
    RPT_VERT,
    RPT_COLOR,
    RPT_TEXCOORD
};
struct RPolyData {
    dd_bool inUse;
    RPolyDataType type;
    uint num;
    void* data;
};
byte rendInfoRPolys = 0;
static unsigned int numrendpolys = 0;
static unsigned int maxrendpolys = 0;
static RPolyData** rendPolys;
void R_PrintRendPoolInfo() {
    if(!rendInfoRPolys) return;
    LOGDEV_GL_MSG("RP Count: %-4i") << numrendpolys;
    for(uint i = 0; i < numrendpolys; ++i) {
        RPolyData* rp = rendPolys[i];
        LOGDEV_GL_VERBOSE(_E(m) "RP: %-4u %c (vtxs=%i t=%c)")
                << i << (rp->inUse? 'Y':'N') << rp->num
                << (rp->type == RPT_VERT? 'v' : rp->type == RPT_COLOR? 'c' : 't');
    }
}
void R_InitRendPolyPools() {
    numrendpolys = maxrendpolys = 0;
    rendPolys = 0;
    Vector3f *rvertices  = R_AllocRendVertices(24);
    Vector4f *rcolors    = R_AllocRendColors(24);
    Vector2f *rtexcoords = R_AllocRendTexCoords(24);
    R_FreeRendVertices(rvertices);
    R_FreeRendColors(rcolors);
    R_FreeRendTexCoords(rtexcoords);
}
Vector3f *R_AllocRendVertices(uint num) {
    uint idx;
    dd_bool found = false;
    for(idx = 0; idx < maxrendpolys; ++idx) {
        if(rendPolys[idx]->inUse) continue;
        if(rendPolys[idx]->type != RPT_VERT) continue;
        if(rendPolys[idx]->num >= num) {
            rendPolys[idx]->inUse = true;
            return (Vector3f *) rendPolys[idx]->data;
        }
        else if(rendPolys[idx]->num == 0) {
            numrendpolys++;
            found = true;
            break;
        }
    }
    if(!found) {
        if(++numrendpolys > maxrendpolys) {
            uint i, newCount;
            RPolyData *newPolyData, *ptr;
            maxrendpolys = (maxrendpolys > 0? maxrendpolys * 2 : 8);
            rendPolys = (RPolyData **) Z_Realloc(rendPolys, sizeof(RPolyData *) * maxrendpolys, PU_MAP);
            newCount = maxrendpolys - numrendpolys + 1;
            newPolyData = (RPolyData *) Z_Malloc(sizeof(RPolyData) * newCount, PU_MAP, 0);
            ptr = newPolyData;
            for(i = numrendpolys-1; i < maxrendpolys; ++i, ptr++) {
                ptr->inUse = false;
                ptr->num = 0;
                ptr->data = NULL;
                ptr->type = RPT_VERT;
                rendPolys[i] = ptr;
            }
        }
        idx = numrendpolys - 1;
    }
    rendPolys[idx]->inUse = true;
    rendPolys[idx]->type  = RPT_VERT;
    rendPolys[idx]->num   = num;
    rendPolys[idx]->data  = Z_Malloc(sizeof(Vector3f) * num, PU_MAP, 0);
    return (Vector3f *) rendPolys[idx]->data;
}
Vector4f *R_AllocRendColors(uint num) {
    uint idx;
    dd_bool found = false;
    for(idx = 0; idx < maxrendpolys; ++idx) {
        if(rendPolys[idx]->inUse) continue;
        if(rendPolys[idx]->type != RPT_COLOR) continue;
        if(rendPolys[idx]->num >= num) {
            rendPolys[idx]->inUse = true;
            return (Vector4f *) rendPolys[idx]->data;
        }
        else if(rendPolys[idx]->num == 0) {
            numrendpolys++;
            found = true;
            break;
        }
    }
    if(!found) {
        if(++numrendpolys > maxrendpolys) {
            uint i, newCount;
            RPolyData *newPolyData, *ptr;
            maxrendpolys = (maxrendpolys > 0? maxrendpolys * 2 : 8);
            rendPolys = (RPolyData **) Z_Realloc(rendPolys, sizeof(RPolyData *) * maxrendpolys, PU_MAP);
            newCount = maxrendpolys - numrendpolys + 1;
            newPolyData = (RPolyData *) Z_Malloc(sizeof(RPolyData) * newCount, PU_MAP, 0);
            ptr = newPolyData;
            for(i = numrendpolys-1; i < maxrendpolys; ++i, ptr++) {
                ptr->inUse = false;
                ptr->num = 0;
                ptr->data = NULL;
                ptr->type = RPT_COLOR;
                rendPolys[i] = ptr;
            }
        }
        idx = numrendpolys - 1;
    }
    rendPolys[idx]->inUse = true;
    rendPolys[idx]->type  = RPT_COLOR;
    rendPolys[idx]->num   = num;
    rendPolys[idx]->data  = Z_Malloc(sizeof(Vector4f) * num, PU_MAP, 0);
    return (Vector4f *) rendPolys[idx]->data;
}
Vector2f *R_AllocRendTexCoords(uint num) {
    uint idx;
    dd_bool found = false;
    for(idx = 0; idx < maxrendpolys; ++idx) {
        if(rendPolys[idx]->inUse) continue;
        if(rendPolys[idx]->type != RPT_TEXCOORD) continue;
        if(rendPolys[idx]->num >= num) {
            rendPolys[idx]->inUse = true;
            return (Vector2f *) rendPolys[idx]->data;
        }
        else if(rendPolys[idx]->num == 0) {
            numrendpolys++;
            found = true;
            break;
        }
    }
    if(!found) {
        if(++numrendpolys > maxrendpolys) {
            uint i, newCount;
            RPolyData *newPolyData, *ptr;
            maxrendpolys = (maxrendpolys > 0? maxrendpolys * 2 : 8);
            rendPolys = (RPolyData **) Z_Realloc(rendPolys, sizeof(RPolyData *) * maxrendpolys, PU_MAP);
            newCount = maxrendpolys - numrendpolys + 1;
            newPolyData = (RPolyData *) Z_Malloc(sizeof(RPolyData) * newCount, PU_MAP, 0);
            ptr = newPolyData;
            for(i = numrendpolys-1; i < maxrendpolys; ++i, ptr++) {
                ptr->inUse = false;
                ptr->num   = 0;
                ptr->data  = NULL;
                ptr->type  = RPT_TEXCOORD;
                rendPolys[i] = ptr;
            }
        }
        idx = numrendpolys - 1;
    }
    rendPolys[idx]->inUse = true;
    rendPolys[idx]->type  = RPT_TEXCOORD;
    rendPolys[idx]->num   = num;
    rendPolys[idx]->data  = Z_Malloc(sizeof(Vector2f) * num, PU_MAP, 0);
    return (Vector2f *) rendPolys[idx]->data;
}
void R_FreeRendVertices(Vector3f *rvertices) {
    if(!rvertices) return;
    for(uint i = 0; i < numrendpolys; ++i) {
        if(rendPolys[i]->data == rvertices) {
            rendPolys[i]->inUse = false;
            return;
        }
    }
    LOGDEV_GL_WARNING("R_FreeRendPoly: Dangling poly ptr!");
}
void R_FreeRendColors(Vector4f *rcolors) {
    if(!rcolors) return;
    for(uint i = 0; i < numrendpolys; ++i) {
        if(rendPolys[i]->data == rcolors) {
            rendPolys[i]->inUse = false;
            return;
        }
    }
    LOGDEV_GL_WARNING("R_FreeRendPoly: Dangling poly ptr!");
}
void R_FreeRendTexCoords(Vector2f *rtexcoords) {
    if(!rtexcoords) return;
    for(uint i = 0; i < numrendpolys; ++i) {
        if(rendPolys[i]->data == rtexcoords) {
            rendPolys[i]->inUse = false;
            return;
        }
    }
    LOGDEV_GL_WARNING("R_FreeRendPoly: Dangling poly ptr!");
}
void R_DivVerts(Vector3f *dst, Vector3f const *src,
    WorldEdge const &leftEdge, WorldEdge const &rightEdge) {
    int const numR = 3 + rightEdge.divisionCount();
    int const numL = 3 + leftEdge.divisionCount();
    if(numR + numL == 6) return; // Nothing to do.
    dst[numL + 0] = src[0];
    dst[numL + 1] = src[3];
    dst[numL + numR - 1] = src[2];
    for(int n = 0; n < rightEdge.divisionCount(); ++n) {
        WorldEdge::Event const &icpt = rightEdge.at(rightEdge.lastDivision() - n);
        dst[numL + 2 + n] = icpt.origin();
    }
    dst[0] = src[3];
    dst[1] = src[0];
    dst[numL - 1] = src[1];
    for(int n = 0; n < leftEdge.divisionCount(); ++n) {
        WorldEdge::Event const &icpt = leftEdge.at(leftEdge.firstDivision() + n);
        dst[2 + n] = icpt.origin();
    }
}
void R_DivTexCoords(Vector2f *dst, Vector2f const *src,
    WorldEdge const &leftEdge, WorldEdge const &rightEdge) {
    int const numR = 3 + rightEdge.divisionCount();
    int const numL = 3 + leftEdge.divisionCount();
    if(numR + numL == 6) return; // Nothing to do.
    dst[numL + 0] = src[0];
    dst[numL + 1] = src[3];
    dst[numL + numR-1] = src[2];
    for(int n = 0; n < rightEdge.divisionCount(); ++n) {
        WorldEdge::Event const &icpt = rightEdge.at(rightEdge.lastDivision() - n);
        dst[numL + 2 + n].x = src[3].x;
        dst[numL + 2 + n].y = src[2].y + (src[3].y - src[2].y) * icpt.distance();
    }
    dst[0] = src[3];
    dst[1] = src[0];
    dst[numL - 1] = src[1];
    for(int n = 0; n < leftEdge.divisionCount(); ++n) {
        WorldEdge::Event const &icpt = leftEdge.at(leftEdge.firstDivision() + n);
        dst[2 + n].x = src[0].x;
        dst[2 + n].y = src[0].y + (src[1].y - src[0].y) * icpt.distance();
    }
}
void R_DivVertColors(Vector4f *dst, Vector4f const *src,
    WorldEdge const &leftEdge, WorldEdge const &rightEdge) {
    int const numR = 3 + rightEdge.divisionCount();
    int const numL = 3 + leftEdge.divisionCount();
    if(numR + numL == 6) return; // Nothing to do.
    dst[numL + 0] = src[0];
    dst[numL + 1] = src[3];
    dst[numL + numR-1] = src[2];
    for(int n = 0; n < rightEdge.divisionCount(); ++n) {
        WorldEdge::Event const &icpt = rightEdge.at(rightEdge.lastDivision() - n);
        dst[numL + 2 + n] = src[2] + (src[3] - src[2]) * icpt.distance();
    }
    dst[0] = src[3];
    dst[1] = src[0];
    dst[numL - 1] = src[1];
    for(int n = 0; n < leftEdge.divisionCount(); ++n) {
        WorldEdge::Event const &icpt = leftEdge.at(leftEdge.firstDivision() + n);
        dst[2 + n] = src[0] + (src[1] - src[0]) * icpt.distance();
    }
}
static String const DEF_WRAP("wrap");
/**
 * Animatable variable bound to a GL uniform. The value can have 1...4 float
 * components.
 */
void ShaderVar::init(float value) {
    values.clear();
    values.append(new AnimationValue(Animation(value, Animation::Linear)));
}
ShaderVar::~ShaderVar() {
    delete uniform;
}
float ShaderVar::currentValue(int index) const {
    auto const &val = values.at(index);
    float v = val.anim->animation();
    if (val.wrap.isEmpty()) {
        return v;
    }
    return val.wrap.wrap(v);
}
void ShaderVar::updateUniform() {
    switch (values.size()) {
    case 1:
        *uniform = currentValue(0);
        break;
    case 2:
        *uniform = Vector2f(currentValue(0),
                            currentValue(1));
        break;
    case 3:
        *uniform = Vector3f(currentValue(0),
                            currentValue(1),
                            currentValue(2));
        break;
    case 4:
        *uniform = Vector4f(currentValue(0),
                            currentValue(1),
                            currentValue(2),
                            currentValue(3));
        break;
    }
}
static char const *componentNames[] = { "x", "y", "z", "w" };
void ShaderVar::updateValuePointers(Record &names, String const &varName) {
    if (values.size() == 1) {
        values[0].anim = &names[varName].value<AnimationValue>();
    }
    else {
        for (int i = 0; i < values.size(); ++i) {
            values[i].anim = &names[varName.concatenateMember(componentNames[i])]
                    .value<AnimationValue>();
        }
    }
}
void ShaderVars::initVariableFromDefinition(String const &variableName,
                                            Record const &valueDef,
                                            Record &bindingNames) {
    GLUniform::Type uniformType = GLUniform::Float;
    std::unique_ptr<ShaderVar> var(new ShaderVar);
    Value const &initialValue = valueDef.get("value");
    if (auto const *array = maybeAs<ArrayValue>(initialValue)) {
        switch (array->size()) {
        default:
            throw DefinitionError("StateAnimator::initVariables",
                                  QString("%1: Invalid initial value size (%2) for render.variable")
                                  .arg(ScriptedInfo::sourceLocation(valueDef))
                                  .arg(array->size()));
        case 2:
            var->init(vectorFromValue<Vector2f>(*array));
            uniformType = GLUniform::Vec2;
            break;
        case 3:
            var->init(vectorFromValue<Vector3f>(*array));
            uniformType = GLUniform::Vec3;
            break;
        case 4:
            var->init(vectorFromValue<Vector4f>(*array));
            uniformType = GLUniform::Vec4;
            break;
        }
        for (int k = 0; k < var->values.size(); ++k) {
            addBinding(bindingNames,
                       variableName.concatenateMember(componentNames[k]),
                       var->values[k].anim);
        }
    }
    else {
        var->init(float(initialValue.asNumber()));
        addBinding(bindingNames, variableName, var->values[0].anim);
    }
    if (valueDef.hasSubrecord(DEF_WRAP)) {
        for (int k = 0; k < 4; ++k) {
            String const varName = QString("%1.%2").arg(DEF_WRAP).arg(componentNames[k]);
            if (valueDef.has(varName)) {
                var->values[k].wrap = rangeFromValue<Rangef>(valueDef.geta(varName));
            }
        }
    }
    else if (valueDef.has(DEF_WRAP)) {
        var->values[0].wrap = rangeFromValue<Rangef>(valueDef.geta(DEF_WRAP));
    }
    var->uniform = new GLUniform(variableName.toLatin1(), uniformType);
    members[variableName] = var.release();
}
ShaderVars::~ShaderVars() {
    qDeleteAll(members.values());
}
void ShaderVars::addBinding(Record &names, String const &varName, AnimationValue *anim) {
    names.add(varName).set(anim).setReadOnly();
}
namespace de {
DENG2_PIMPL_NOREF(ShadowEdge) {
    HEdge const *leftMostHEdge = nullptr;
    dint edge = 0;
    Vector3d inner;
    Vector3d outer;
    dfloat sectorOpenness = 0;
    dfloat openness = 0;
};
ShadowEdge::ShadowEdge() : d(new Impl)
{}
void ShadowEdge::init(HEdge const &leftMostHEdge, dint edge) {
    d->leftMostHEdge = &leftMostHEdge;
    d->edge          = edge;
    d->inner = d->outer = Vector3d();
    d->sectorOpenness = d->openness = 0;
}
/**
 * Returns a value in the range of 0..2, representing how 'open' the edge is.
 *
 * @c =0 Completely closed, it is facing a wall or is relatively distant from the edge on
 *       the other side.
 * @c >0 && <1 How near the 'other' edge is.
 * @c =1 At the same height as "this" one.
 * @c >1 The 'other' edge is past our height (clearly 'open').
 */
static dfloat opennessFactor(dfloat fz, dfloat bz, dfloat bhz) {
    if(fz <= bz - SHADOWEDGE_OPEN_THRESHOLD || fz >= bhz)
        return 0;  // Fully closed.
    if(fz >= bhz - SHADOWEDGE_OPEN_THRESHOLD)
        return (bhz - fz) / SHADOWEDGE_OPEN_THRESHOLD;
    if(fz <= bz)
        return 1 - (bz - fz) / SHADOWEDGE_OPEN_THRESHOLD;
    if(fz <= bz + SHADOWEDGE_OPEN_THRESHOLD)
        return 1 + (fz - bz) / SHADOWEDGE_OPEN_THRESHOLD;
    return 2;  // Fully open!
}
static bool middleMaterialCoversOpening(LineSide const &side) {
    if (!side.hasSector()) return false;  // Never.
    if (!side.hasSections()) return false;
    MaterialAnimator *matAnimator = side.middle().materialAnimator();
    /* material().as<ClientMaterial>()
            .getAnimator(Rend_MapSurfaceMaterialSpec());*/
    if (!matAnimator) return false;
    matAnimator->prepare();
    if (matAnimator->isOpaque() && !side.middle().blendMode() && side.middle().opacity() >= 1) {
        if (side.isFlagged(SDF_MIDDLE_STRETCH))
            return true;
        Sector const &frontSec = side.sector();
        Sector const *backSec  = side.back().sectorPtr();
        coord_t openBottom;
        if (backSec && backSec->floor().heightSmoothed() > frontSec.floor().heightSmoothed()) {
            openBottom = backSec->floor().heightSmoothed();
        }
        else {
            openBottom = frontSec.floor().heightSmoothed();
        }
        coord_t openTop;
        if (backSec && backSec->ceiling().heightSmoothed() < frontSec.ceiling().heightSmoothed()) {
            openTop = backSec->ceiling().heightSmoothed();
        }
        else {
            openTop = frontSec.ceiling().heightSmoothed();
        }
        if (matAnimator->dimensions().y >= openTop - openBottom) {
            if(side.leftHEdge()) // possibility of degenerate BSP leaf
            {
                WallEdge edge(WallSpec::fromMapSide(side, LineSide::Middle),
                              *side.leftHEdge(), Line::From);
                return (edge.isValid() && edge.top().z() > edge.bottom().z()
                        && edge.top().z() >= openTop && edge.bottom().z() <= openBottom);
            }
        }
    }
    return false;
}
void ShadowEdge::prepare(dint planeIndex) {
    dint const otherPlaneIndex = planeIndex == Sector::Floor? Sector::Ceiling : Sector::Floor;
    HEdge const &hedge = *d->leftMostHEdge;
    auto const &subsec = hedge.face().mapElementAs<ConvexSubspace>()
                            .subsector().as<world::ClientSubsector>();
    Plane const &plane = subsec.visPlane(planeIndex);
    LineSide const &lineSide = hedge.mapElementAs<LineSideSegment>().lineSide();
    d->sectorOpenness = d->openness = 0; // Default is fully closed.
    if (hedge.twin().hasFace() &&
        hedge.twin().face().mapElementAs<ConvexSubspace>().hasSubsector()) {
        auto const &backSubsec = hedge.twin().face().mapElementAs<ConvexSubspace>()
                                    .subsector().as<world::ClientSubsector>();
        Plane const &backPlane = backSubsec.visPlane(planeIndex);
        Surface const &wallEdgeSurface =
            lineSide.back().hasSector() ? lineSide.surface(planeIndex == Sector::Ceiling ? LineSide::Top : LineSide::Bottom)
                                        : lineSide.middle();
        coord_t fz = plane.heightSmoothed();
        if (planeIndex == Sector::Ceiling)
            fz = -fz;
        coord_t bz = backPlane.heightSmoothed();
        if (planeIndex == Sector::Ceiling)
            bz = -bz;
        coord_t bhz = backSubsec.sector().plane(otherPlaneIndex).heightSmoothed();
        if (planeIndex == Sector::Ceiling)
            bhz = -bhz;
        if (fz < bz && !wallEdgeSurface.hasMaterial()) {
            d->sectorOpenness = 2; // Consider it fully open.
        }
        else if (subsec.visFloor().heightSmoothed() >= backSubsec.visCeiling().heightSmoothed() &&
                 subsec    .visPlane(otherPlaneIndex).surface().hasSkyMaskedMaterial() &&
                 backSubsec.visPlane(otherPlaneIndex).surface().hasSkyMaskedMaterial()) {
            d->sectorOpenness = 2; // Consider it fully open.
        }
        else {
            if (!middleMaterialCoversOpening(lineSide)) {
                d->sectorOpenness = opennessFactor(fz, bz, bhz);
            }
        }
    }
    if (d->sectorOpenness >= 1)
        return;
    dint const edge = lineSide.sideId() ^ d->edge;
    LineOwner const *vo = lineSide.line().vertexOwner(edge)->navigate(ClockDirection(d->edge ^ 1));
    Line const &neighborLine = vo->line();
    if (&neighborLine == &lineSide.line()) {
        d->openness = 1; // Fully open.
    }
    else if (neighborLine.isSelfReferencing()) /// @todo Skip over these? -ds
    {
        d->openness = 1;
    }
    else {
        LineSide const &neighborLineSide = neighborLine.side(&lineSide.line().vertex(edge) == &neighborLine.from()? d->edge ^ 1 : d->edge);
        if (!neighborLineSide.hasSections() && neighborLineSide.back().hasSector()) {
            d->openness = 1;
        }
        else if (!neighborLineSide.hasSector() ||
                 (neighborLineSide.back().hasSector() && middleMaterialCoversOpening(neighborLineSide))) {
            d->openness = 0;
        }
        else if (neighborLineSide.back().hasSector()) {
            Sector const *backSec  = neighborLineSide.back().sectorPtr();
            if (backSec != &subsec.sector() &&
                !((plane.isSectorFloor  () && backSec->ceiling().heightSmoothed() <= plane.heightSmoothed()) ||
                  (plane.isSectorCeiling() && backSec->floor  ().heightSmoothed() >= plane.heightSmoothed()))) {
                coord_t fz = plane.heightSmoothed();
                if (planeIndex == Sector::Ceiling)
                    fz = -fz;
                coord_t bz = backSec->plane(planeIndex).heightSmoothed();
                if (planeIndex == Sector::Ceiling)
                    bz = -bz;
                coord_t bhz = backSec->plane(otherPlaneIndex).heightSmoothed();
                if (planeIndex == Sector::Ceiling)
                    bhz = -bhz;
                d->openness = opennessFactor(fz, bz, bhz);
            }
        }
    }
    if (d->openness < 1) {
        LineOwner *vo = lineSide.line().vertexOwner(lineSide.sideId() ^ d->edge);
        if (d->edge) vo = vo->prev();
        d->inner = Vector3d(lineSide.vertex(d->edge).origin() + vo->innerShadowOffset(),
                            plane.heightSmoothed());
    }
    else {
        d->inner = Vector3d(lineSide.vertex(d->edge).origin() + vo->extendedShadowOffset(),
                            plane.heightSmoothed());
    }
    d->outer = Vector3d(lineSide.vertex(d->edge).origin(), plane.heightSmoothed());
}
Vector3d const &ShadowEdge::inner() const {
    return d->inner;
}
Vector3d const &ShadowEdge::outer() const {
    return d->outer;
}
dfloat ShadowEdge::openness() const {
    return d->openness;
}
dfloat ShadowEdge::sectorOpenness() const {
    return d->sectorOpenness;
}
dfloat ShadowEdge::shadowStrength(dfloat darkness) const {
    if(d->sectorOpenness < 1) {
        dfloat strength = de::min(darkness * (1 - d->sectorOpenness), 1.f);
        if(d->openness < 1) strength *= 1 - d->openness;
        return strength;
    }
    return 0;
}
}  // namespace de
namespace internal {
static int sphereDetail     = 6;
static float sphereDistance = 1600; ///< Map units.
static int sphereRows       = 3;    ///< Per hemisphere.
/**
 * Effective layer configuration used with both sphere and model drawing.
 */
struct LayerData {
    bool active;   ///< @c true= the layer is active/enabled (will be drawn).
    float offset;  ///< Layer material offset (along the horizon).
};
/**
 * Geometry used with the sky sphere. The crest of the hemisphere is up (i.e., y+).
 *
 * @todo Should be a subcomponent of SkyDrawable (cleaner interface). -ds
 */
struct Hemisphere {
    enum SphereComponent {
        UpperHemisphere,
        LowerHemisphere
    };
    int rows            = 3;
    int columns         = 4 * 6;
    float height        = 0.49f;
    float horizonOffset = -0.105f;
    typedef QVector<Vector3f> VBuf;
    VBuf verts;
    bool needRebuild = true;
    inline Vector3f const &vertex(int r, int c) const {
        return verts[r * columns + c % columns];
    }
    /**
     * Determine the material to use for the given sky @a layer.
     */
    static world::Material *chooseMaterialForSkyLayer(SkyLayer const &layer) {
        if(renderTextures == 0) {
            return nullptr;
        }
        if(renderTextures == 2) {
            return world::Materials::get().materialPtr(de::Uri("System", Path("gray")));
        }
        if(world::Material *mat = layer.material()) {
            return mat;
        }
        return world::Materials::get().materialPtr(de::Uri("System", Path("missing")));
    }
    /**
     * Determine the cap/fadeout color to use for the given sky @a layer.
     */
    static Vector3f chooseCapColor(SphereComponent hemisphere, SkyLayer const &layer,
                                   bool *needFadeOut = nullptr) {
        if (world::Material *mat = chooseMaterialForSkyLayer(layer)) {
            MaterialAnimator &matAnimator =
                mat->as<ClientMaterial>()
                    .getAnimator(SkyDrawable::layerMaterialSpec(layer.isMasked()));
            matAnimator.prepare();
            ClientTexture &pTex = matAnimator.texUnit(MaterialAnimator::TU_LAYER0).texture->base();
            averagecolor_analysis_t const *avgColor = reinterpret_cast<averagecolor_analysis_t const *>
                    (pTex.analysisDataPointer((hemisphere == UpperHemisphere ? res::Texture::AverageTopColorAnalysis
                                                                             : res::Texture::AverageBottomColorAnalysis)));
            if (!avgColor) {
                de::Uri const pTexUri = pTex.manifest().composeUri();
                throw Error("Hemisphere::capColor", String("Texture \"%1\" has no Average%2ColorAnalysis")
                                                        .arg(pTexUri)
                                                        .arg(hemisphere == UpperHemisphere? "Top" : "Bottom"));
            }
            Vector3f color(avgColor->color.rgb);
            dfloat const fadeOutLimit = layer.fadeOutLimit();
            if (color.x >= fadeOutLimit || color.y >= fadeOutLimit || color.z >= fadeOutLimit) {
                if (needFadeOut) *needFadeOut = true;
                return color;
            }
        }
        return Vector3f(); // Default color is black.
    }
    void drawCap(Vector3f const &color, bool drawFadeOut) const {
        GL_SetNoTexture();
        DGL_Color3f(color.x, color.y, color.z);
        DGL_Begin(DGL_TRIANGLE_FAN);
        for(int c = 0; c < columns; ++c) {
            Vector3f const &vtx = vertex(0, c);
            DGL_Vertex3f(vtx.x, vtx.y, vtx.z);
        }
        DGL_End();
        if(!drawFadeOut) return;
        DGL_Begin(DGL_TRIANGLE_STRIP);
        Vector3f const *vtx = &vertex(0, 0);
        DGL_Vertex3f(vtx->x, vtx->y, vtx->z);
        int c = 0;
        for(; c < columns; ++c) {
            vtx = &vertex(1, c);
            DGL_Vertex3f(vtx->x, vtx->y, vtx->z);
            vtx = &vertex(0, c + 1);
            DGL_Vertex3f(vtx->x, vtx->y, vtx->z);
        }
        vtx = &vertex(1, c);
        DGL_Vertex3f(vtx->x, vtx->y, vtx->z);
        DGL_End();
    }
    void draw(SphereComponent hemisphere, Sky const &sky, dint firstActiveLayer,
              LayerData const *layerData) const {
        DENG2_ASSERT(layerData);
        if (verts.isEmpty()) return;
        if (firstActiveLayer < 0) return;
        bool const yflip = (hemisphere == LowerHemisphere);
        if (yflip) {
            DGL_MatrixMode(DGL_MODELVIEW);
            DGL_PushMatrix();
            DGL_Scalef(1.0f, -1.0f, 1.0f);
        }
        bool drawFadeOut = true;
        drawCap(chooseCapColor(hemisphere, sky.layer(firstActiveLayer), &drawFadeOut), drawFadeOut);
        for (dint i = firstActiveLayer; i < MAX_LAYERS; ++i) {
            SkyLayer const &skyLayer = sky.layer(i);
            LayerData const &ldata   = layerData[i];
            if (!ldata.active) continue;
            TextureVariant *layerTex = nullptr;
            if (world::Material *mat = chooseMaterialForSkyLayer(skyLayer)) {
                MaterialAnimator &matAnimator = mat->as<ClientMaterial>()
                        .getAnimator(SkyDrawable::layerMaterialSpec(skyLayer.isMasked()));
                matAnimator.prepare();
                layerTex = matAnimator.texUnit(MaterialAnimator::TU_LAYER0).texture;
                GL_BindTexture(layerTex);
                DGL_Enable(DGL_TEXTURE_2D);
                DGL_MatrixMode(DGL_TEXTURE);
                DGL_PushMatrix();
                DGL_LoadIdentity();
                Vector2ui const &texSize = layerTex->base().dimensions();
                if (texSize.x > 0) {
                    DGL_Translatef(ldata.offset / texSize.x, 0, 0);
                    DGL_Scalef(1024.f / texSize.x, 1, 1);
                }
                if (yflip) {
                    DGL_Scalef(1, -1, 1);
                    DGL_Translatef(0, -1, 0);
                }
            }
            else {
                GL_SetNoTexture();
            }
#define WRITESKYVERTEX(r_, c_) { \
    svtx = &vertex(r_, c_); \
    if (layerTex) \
    { \
       DGL_TexCoord2f(0, (c_) / float(columns), (r_) / float(rows)); \
    } \
    if (drawFadeOut) \
    { \
        if ((r_) == 0) DGL_Color4f(1, 1, 1, 0); \
        else           DGL_Color3f(1, 1, 1); \
    } \
    else \
    { \
        if ((r_) == 0) DGL_Color3f(0, 0, 0); \
        else           DGL_Color3f(1, 1, 1); \
    } \
    DGL_Vertex3f(svtx->x, svtx->y, svtx->z); \
}
            Vector3f const *svtx;
            for (dint r = 0; r < rows; ++r) {
                DGL_Begin(DGL_TRIANGLE_STRIP);
                WRITESKYVERTEX(r, 0);
                WRITESKYVERTEX(r + 1, 0);
                for (dint c = 1; c <= columns; ++c) {
                    WRITESKYVERTEX(r, c);
                    WRITESKYVERTEX(r + 1, c);
                }
                DGL_End();
            }
            if (layerTex) {
                DGL_MatrixMode(DGL_TEXTURE);
                DGL_PopMatrix();
                DGL_Disable(DGL_TEXTURE_2D);
            }
#undef WRITESKYVERTEX
        }
        if (yflip) {
            DGL_MatrixMode(DGL_MODELVIEW);
            DGL_PopMatrix();
        }
    }
    /**
     * The top row (row 0) is the one that's faded out.
     * There must be at least 4 columns. The preferable number is 4n, where
     * n is 1, 2, 3... There should be at least two rows because the first
     * one is always faded.
     *
     * The total number of triangles per hemisphere can be calculated thus:
     *
     * Sum: rows * columns * 2 + (hemisphere)
     *      rows * 2 + (fadeout)
     *      rows - 2 (cap)
     */
    void makeVertices(float height, float horizonOffset) {
        DENG2_ASSERT(height > 0);
        rows    = de::max(sphereRows,   1);
        columns = de::max(sphereDetail, 1) * 4;
        verts.resize(columns * (rows + 1));
        float const maxSideAngle = float(de::PI / 2 * height);
        float const sideOffset   = float(de::PI / 2 * horizonOffset);
        for(int r = 0; r < rows + 1; ++r)
        for(int c = 0; c < columns; ++c) {
            Vector3f &svtx = verts[r * columns + c % columns];
            float const topAngle  = ((c / float(columns)) * 2) * PI;
            float const sideAngle = sideOffset + maxSideAngle * (rows - r) / float(rows);
            float const radius    = cos(sideAngle);
            svtx = Vector3f(radius * cos(topAngle),
                            sin(sideAngle), // The height.
                            radius * sin(topAngle));
        }
    }
};
static Hemisphere hemisphere;
} // namespace internal
using namespace ::internal;
DENG2_PIMPL(SkyDrawable)
, DENG2_OBSERVES(Sky, Deletion)
, DENG2_OBSERVES(Sky, HeightChange)
, DENG2_OBSERVES(Sky, HorizonOffsetChange) {
    Sky const *sky = nullptr;
    dfloat height = 1;
    bool needHeightUpdate = false;
    dfloat horizonOffset = 0;
    bool needHorizonOffsetUpdate = false;
    LayerData layers[MAX_LAYERS];
    dint firstActiveLayer = -1;  ///< @c -1= no active layers.
    bool needBuildHemisphere = true;
    struct ModelData {
        FrameModelDef *modef = nullptr;
    };
    ModelData models[MAX_MODELS];
    bool haveModels       = false;
    bool alwaysDrawSphere = false;
    Impl(Public *i) : Base(i) {
        de::zap(models);
    }
    ~Impl() {
        self().configure();
    }
    /**
     * Prepare for drawing; determine layer configuration, sphere dimensions, etc..
     */
    void prepare(Animator const *animator) {
        firstActiveLayer = -1;
        for (dint i = 0; i < MAX_LAYERS; ++i) {
            SkyLayer const &skyLayer = sky->layer(i);
            layers[i].active = skyLayer.isActive();
            layers[i].offset = skyLayer.offset() + animator->layer(i).offset;
            if(firstActiveLayer == -1 && layers[i].active) {
                firstActiveLayer = i;
            }
        }
        updateHeightIfNeeded();
        updateHorizonOffsetIfNeeded();
        if(needBuildHemisphere || hemisphere.needRebuild) {
            needBuildHemisphere = false;
            hemisphere.makeVertices(height, horizonOffset);
        }
    }
    void drawSphere() const {
        if(haveModels && !alwaysDrawSphere) return;
        DGL_PushState();
        DGL_CullFace(DGL_NONE);
        DGL_Disable(DGL_DEPTH_TEST);
        DGL_Disable(DGL_DEPTH_WRITE);
        DGL_MatrixMode(DGL_MODELVIEW);
        DGL_PushMatrix();
        DGL_Translatef(vOrigin.x, vOrigin.y, vOrigin.z);
        DGL_Scalef(sphereDistance, sphereDistance, sphereDistance);
        hemisphere.draw(Hemisphere::LowerHemisphere, *sky, firstActiveLayer, layers);
        hemisphere.draw(Hemisphere::UpperHemisphere, *sky, firstActiveLayer, layers);
        DGL_MatrixMode(DGL_MODELVIEW);
        DGL_PopMatrix();
        DGL_PopState();
    }
    void drawModels(Animator const *animator) const {
        if(!haveModels) return;
        DGL_Flush();
        DGL_Enable(DGL_DEPTH_TEST);
        DGL_Enable(DGL_DEPTH_WRITE);
        LIBGUI_GL.glClear(GL_DEPTH_BUFFER_BIT);
        DGL_MatrixMode(DGL_MODELVIEW);
        DGL_PushMatrix();
        DGL_Translatef(vOrigin.x, vOrigin.y, vOrigin.z);
        for(int i = 0; i < MAX_MODELS; ++i) {
            ModelData const &mdata = models[i];
            FrameModelDef *modef = mdata.modef;
            if(!modef) continue;
            Record const &skyModelDef = defn::Sky(*sky->def()).model(i);
            int const layerNum        = skyModelDef.geti("layer");
            if(layerNum > 0 && layerNum <= MAX_LAYERS) {
                if(!layers[layerNum - 1].active)
                    continue;
            }
            Animator::ModelState const &mstate = animator->model(i);
            vissprite_t vis{};
            vis.pose.origin          = vOrigin.xzy() * -Vector3f(skyModelDef.get("originOffset")).xzy();
            vis.pose.topZ            = vis.pose.origin.z;
            vis.pose.distance        = 1;
            Vector2f rotate(skyModelDef.get("rotate"));
            vis.pose.yaw             = mstate.yaw;
            vis.pose.extraYawAngle   = vis.pose.yawAngleOffset   = rotate.x;
            vis.pose.extraPitchAngle = vis.pose.pitchAngleOffset = rotate.y;
            drawmodelparams_t &visModel = *VS_MODEL(&vis);
            visModel.inter                       = (mstate.maxTimer > 0 ? mstate.timer / float(mstate.maxTimer) : 0);
            visModel.mf                          = modef;
            visModel.alwaysInterpolate           = true;
            visModel.shineTranslateWithViewerPos = true;
            ClientApp::resources().setModelDefFrame(*modef, mstate.frame);
            vis.light.ambientColor = Vector4f(skyModelDef.get("color"), 1);
            Rend_DrawModel(vis);
        }
        DGL_Flush();
        DGL_MatrixMode(DGL_MODELVIEW);
        DGL_PopMatrix();
        LIBGUI_GL.glClear(GL_DEPTH_BUFFER_BIT);
    }
    void setupModels(Record const *def) {
        alwaysDrawSphere = (def && (def->geti("flags") & SIF_DRAW_SPHERE) != 0);
        haveModels = false;
        de::zap(models);
        if(!def) return;
        defn::Sky const skyDef(*def);
        for(int i = 0; i < skyDef.modelCount(); ++i) {
            ModelData &mdata          = models[i];
            Record const &skyModelDef = skyDef.model(i);
            try {
                if(FrameModelDef *modef = &ClientApp::resources().modelDef(skyModelDef.gets("id"))) {
                    if(modef->subCount()) {
                        mdata.modef = modef;
                        haveModels = true; // There is at least one model here.
                    }
                }
            }
            catch(ClientResources::MissingModelDefError const &)
            {} // Ignore this error.
        }
    }
    void updateHeightIfNeeded() {
        if(!needHeightUpdate) return;
        needHeightUpdate = false;
        float newHeight = (sky? sky->height() : 1);
        if(!de::fequal(height, newHeight)) {
            height = newHeight;
            needBuildHemisphere = true;
        }
    }
    void updateHorizonOffsetIfNeeded() {
        if(!needHorizonOffsetUpdate) return;
        needHorizonOffsetUpdate = false;
        float newHorizonOffset = (sky? sky->horizonOffset() : 0);
        if(!de::fequal(horizonOffset, newHorizonOffset)) {
            horizonOffset = newHorizonOffset;
            needBuildHemisphere = true;
        }
    }
    void skyBeingDeleted(Sky const &) {
        self().configure();
    }
    void skyHeightChanged(Sky &) {
        needHeightUpdate = true;
    }
    void skyHorizonOffsetChanged(Sky &) {
        needHorizonOffsetUpdate = true;
    }
};
SkyDrawable::SkyDrawable(Sky const *sky) : d(new Impl(this)) {
    configure(sky);
}
SkyDrawable &SkyDrawable::configure(Sky const *sky) {
    if(d->sky) {
        d->sky->audienceForDeletion()            -= d;
        d->sky->audienceForHeightChange()        -= d;
        d->sky->audienceForHorizonOffsetChange() -= d;
    }
    d->sky = sky;
    d->needHeightUpdate        = true;
    d->needHorizonOffsetUpdate = true;
    if(d->sky) {
        d->sky->audienceForHorizonOffsetChange() += d;
        d->sky->audienceForHeightChange()        += d;
        d->sky->audienceForDeletion()            += d;
    }
    d->setupModels(d->sky? d->sky->def() : 0);
    return *this;
}
Sky const *SkyDrawable::sky() const {
    return d->sky;
}
void SkyDrawable::cacheAssets() {
    if (!d->sky) return;
    d->sky->forAllLayers([] (SkyLayer const &layer) {
        if (world::Material *mat = layer.material()) {
            ClientApp::resources().cache(mat->as<ClientMaterial>(), layerMaterialSpec(layer.isMasked()));
        }
        return LoopContinue;
    });
    if(!d->haveModels) return;
    for (dint i = 0; i < MAX_MODELS; ++i) {
        if (FrameModelDef *modef = d->models[i].modef) {
            ClientApp::resources().cache(modef);
        }
    }
}
FrameModelDef *SkyDrawable::modelDef(int modelIndex) const {
    if(modelIndex >= 0 && modelIndex < MAX_MODELS) {
        return d->models[modelIndex].modef;
    }
    return nullptr;
}
void SkyDrawable::draw(Animator const *animator) const {
    DENG2_ASSERT(animator);
    DENG2_ASSERT(&animator->sky() == this && d->sky == animator->sky().sky());
    DENG2_ASSERT_IN_RENDER_THREAD();
    DENG_ASSERT_GL_CONTEXT_ACTIVE();
    d->prepare(animator);
    if(d->firstActiveLayer < 0) return;
    if(fogParams.usingFog) DGL_Enable(DGL_FOG);
    d->drawSphere();
    d->drawModels(animator);
    if(fogParams.usingFog) DGL_Disable(DGL_FOG);
}
MaterialVariantSpec const &SkyDrawable::layerMaterialSpec(bool masked) // static
{
    return ClientApp::resources().materialSpec(SkySphereContext, TSF_NO_COMPRESSION | (masked? TSF_ZEROMASK : 0),
                                 0, 0, 0, GL_REPEAT, GL_CLAMP_TO_EDGE,
                                 0, -1, -1, false, true, false, false);
}
static void markSphereForRebuild() {
    hemisphere.needRebuild = true;
}
void SkyDrawable::consoleRegister() // static
{
    C_VAR_INT2 ("rend-sky-detail",   &sphereDetail,     0,          3, 7, markSphereForRebuild);
    C_VAR_INT2 ("rend-sky-rows",     &sphereRows,       0,          1, 8, markSphereForRebuild);
    C_VAR_FLOAT("rend-sky-distance", &sphereDistance,   CVF_NO_MAX, 1, 0);
}
DENG2_PIMPL_NOREF(SkyDrawable::Animator) {
    SkyDrawable *sky = nullptr;
    LayerState layers[MAX_LAYERS];
    ModelState models[MAX_MODELS];
    Impl() {
        de::zap(layers);
        de::zap(models);
    }
};
SkyDrawable::Animator::Animator() : d(new Impl)
{}
SkyDrawable::Animator::Animator(SkyDrawable &sky) : d(new Impl) {
    d->sky = &sky;
}
SkyDrawable::Animator::~Animator()
{}
void SkyDrawable::Animator::setSky(SkyDrawable *sky) {
    d->sky = sky;
    de::zap(d->layers);
    de::zap(d->models);
    if(!d->sky) return;
    defn::Sky const skyDef(*d->sky->sky()->def());
    for(int i = 0; i < MAX_MODELS; ++i) {
        ModelState &mstate = model(i);
        if(FrameModelDef const *modef = d->sky->modelDef(i)) {
            Record const &skyModelDef = skyDef.model(i);
            mstate.frame    = modef->subModelDef(0).frame;
            mstate.maxTimer = int(TICSPERSEC * skyModelDef.getf("frameInterval"));
            mstate.yaw      = skyModelDef.getf("yaw");
        }
    }
}
SkyDrawable &SkyDrawable::Animator::sky() const {
    DENG2_ASSERT(d->sky);
    return *d->sky;
}
bool SkyDrawable::Animator::hasLayer(int index) const {
    return (index >= 0 && index < MAX_LAYERS);
}
SkyDrawable::Animator::LayerState &SkyDrawable::Animator::layer(int index) {
    if(hasLayer(index)) return d->layers[index];
    throw MissingLayerStateError("SkyDrawable::Animator::layer", "Invalid layer state index #" + String::number(index) + ".");
}
SkyDrawable::Animator::LayerState const &SkyDrawable::Animator::layer(int index) const {
    return const_cast<SkyDrawable::Animator *>(this)->layer(index);
}
bool SkyDrawable::Animator::hasModel(int index) const {
    return (index >= 0 && index < MAX_MODELS);
}
SkyDrawable::Animator::ModelState &SkyDrawable::Animator::model(int index) {
    if(hasModel(index)) return d->models[index];
    throw MissingModelStateError("SkyDrawable::Animator::model", "Invalid model state index #" + String::number(index) + ".");
}
SkyDrawable::Animator::ModelState const &SkyDrawable::Animator::model(int index) const {
    return const_cast<SkyDrawable::Animator *>(this)->model(index);
}
void SkyDrawable::Animator::advanceTime(timespan_t /*elapsed*/)
{
    LOG_AS("SkyDrawable::Animator");
    if(!d->sky) return;
    if(!sky().sky()) return;
    if(clientPaused) return;
    if(!DD_IsSharpTick()) return;
    defn::Sky const skyDef(*sky().sky()->def());
    for(int i = 0; i < MAX_LAYERS; ++i) {
        Record const &skyLayerDef = skyDef.layer(i);
        LayerState &lstate        = layer(i);
        lstate.offset += skyLayerDef.getf("offsetSpeed") / TICSPERSEC;
    }
    for(int i = 0; i < MAX_MODELS; ++i) {
        FrameModelDef const *modef = sky().modelDef(i);
        if(!modef) continue;
        Record const &skyModelDef = skyDef.model(i);
        ModelState &mstate        = model(i);
        mstate.yaw += skyModelDef.getf("yawSpeed") / TICSPERSEC;
        if(mstate.maxTimer > 0 && ++mstate.timer >= mstate.maxTimer) {
            mstate.frame++;
            mstate.timer = 0;
            String const execute = skyModelDef.gets("execute");
            if(!execute.isEmpty()) {
                Con_Execute(CMDS_SCRIPT, execute.toUtf8().constData(), true, false);
            }
        }
    }
}
namespace de {
static ddouble skyFixFloorZ(Plane const *frontFloor, Plane const *backFloor) {
    DENG_UNUSED(backFloor);
    if(devRendSkyMode || P_IsInVoid(viewPlayer))
        return frontFloor->heightSmoothed();
    return frontFloor->map().skyFloor().height();
}
static ddouble skyFixCeilZ(Plane const *frontCeil, Plane const *backCeil) {
    DENG_UNUSED(backCeil);
    if(devRendSkyMode || P_IsInVoid(viewPlayer))
        return frontCeil->heightSmoothed();
    return frontCeil->map().skyCeiling().height();
}
DENG2_PIMPL_NOREF(SkyFixEdge::Event) {
    SkyFixEdge &owner;
    ddouble distance;
    Impl(SkyFixEdge &owner, ddouble distance)
        : owner(owner), distance(distance)
    {}
};
SkyFixEdge::Event::Event(SkyFixEdge &owner, ddouble distance)
    : WorldEdge::Event()
    , d(new Impl(owner, distance))
{}
bool SkyFixEdge::Event::operator < (Event const &other) const {
    return d->distance < other.distance();
}
ddouble SkyFixEdge::Event::distance() const {
    return d->distance;
}
Vector3d SkyFixEdge::Event::origin() const {
    return d->owner.pOrigin() + d->owner.pDirection() * distance();
}
DENG2_PIMPL(SkyFixEdge) {
    HEdge *hedge;
    FixType fixType;
    dint edge;
    Vector3d pOrigin;
    Vector3d pDirection;
    coord_t lo, hi;
    Event bottom;
    Event top;
    bool isValid;
    Vector2f materialOrigin;
    Impl(Public *i, HEdge &hedge, FixType fixType, int edge,
             Vector2f const &materialOrigin)
        : Base(i),
          hedge(&hedge),
          fixType(fixType),
          edge(edge),
          bottom(*i, 0),
          top(*i, 1),
          isValid(false),
          materialOrigin(materialOrigin)
    {}
    /**
     * Determines whether a sky fix is actually necessary.
     */
    bool wallSectionNeedsSkyFix() const {
        DENG_ASSERT(hedge->hasFace());
        bool const lower = fixType == SkyFixEdge::Lower;
        if (!hedge->hasMapElement()) return false;
        const auto &lineSeg  = hedge->mapElementAs<LineSideSegment>();
        const auto &lineSide = lineSeg.lineSide();
        auto const *space     = &hedge->face().mapElementAs<ConvexSubspace>();
        auto const *backSpace = hedge->twin().hasFace() ? &hedge->twin().face().mapElementAs<ConvexSubspace>()
                                                        : nullptr;
        auto const *subsec     = &space->subsector().as<ClientSubsector>();
        auto const *backSubsec = backSpace && backSpace->hasSubsector() ? &backSpace->subsector().as<ClientSubsector>()
                                                                        : nullptr;
        if (backSubsec && &backSubsec->sector() == &subsec->sector())
            return false;
        if (lineSide.middle().hasMaterial() &&
            !lineSide.middle().materialAnimator()->isOpaque() &&
            ((!lower && !lineSide.top().hasMaterial()) ||
             (lower  && !lineSide.bottom().hasMaterial()))) {
            return false;
        }
        dint relPlane = lower ? Sector::Floor : Sector::Ceiling;
        Plane const *front   = &subsec->visPlane(relPlane);
        Plane const *back    = backSubsec ? &backSubsec->visPlane(relPlane) : nullptr;
        if (!front->surface().hasSkyMaskedMaterial())
            return false;
        bool const hasClosedBack = R_SideBackClosed(lineSide);
        if (!devRendSkyMode) {
            if (!P_IsInVoid(viewPlayer)
                && !(hasClosedBack || !(back && back->surface().hasSkyMaskedMaterial())))
                return false;
        }
        else {
            dint relSection = lower ? LineSide::Bottom : LineSide::Top;
            if (lineSide.surface(relSection).hasMaterial()
                || !(hasClosedBack || (back && back->surface().hasSkyMaskedMaterial())))
                return false;
        }
        coord_t fz = front->heightSmoothed();
        if (relPlane == Sector::Ceiling)
            fz = -fz;
        coord_t bz = 0;
        if (back) {
            bz = back->heightSmoothed();
            if(relPlane == Sector::Ceiling)
                bz = -bz;
        }
        coord_t planeZ = (back && back->surface().hasSkyMaskedMaterial() &&
                          fz < bz? bz : fz);
        coord_t skyZ = lower ? skyFixFloorZ(front, back)
                             : -skyFixCeilZ(front, back);
        return (planeZ > skyZ);
    }
    void prepare() {
        if (!wallSectionNeedsSkyFix()) {
            isValid = false;
            return;
        }
        auto const *subspace     = &hedge->face().mapElementAs<ConvexSubspace>();
        auto const *backSubspace = hedge->twin().hasFace() ? &hedge->twin().face().mapElementAs<ConvexSubspace>() : nullptr;
        auto const *subsec       = &subspace->subsector().as<world::ClientSubsector>();
        auto const *backSubsec   = backSubspace && backSubspace->hasSubsector() ? &backSubspace->subsector().as<world::ClientSubsector>()
                                                                                : nullptr;
        Plane const *ffloor = &subsec->visFloor();
        Plane const *fceil  = &subsec->visCeiling();
        Plane const *bceil  = backSubsec ? &backSubsec->visCeiling() : nullptr;
        Plane const *bfloor = backSubsec ? &backSubsec->visFloor()   : nullptr;
        if (fixType == Upper) {
            hi = skyFixCeilZ(fceil, bceil);
            lo = de::max((backSubsec && bceil->surface().hasSkyMaskedMaterial()) ? bceil->heightSmoothed()
                                                                                 : fceil->heightSmoothed()
                         , ffloor->heightSmoothed());
        }
        else {
            hi = de::min((backSubsec && bfloor->surface().hasSkyMaskedMaterial()) ? bfloor->heightSmoothed()
                                                                                  : ffloor->heightSmoothed()
                         , fceil->heightSmoothed());
            lo = skyFixFloorZ(ffloor, bfloor);
        }
        isValid = hi > lo;
        if (!isValid) return;
        pOrigin = Vector3d(self().origin(), lo);
        pDirection = Vector3d(0, 0, hi - lo);
    }
};
SkyFixEdge::SkyFixEdge(HEdge &hedge, FixType fixType, int edge, float materialOffsetS)
    : WorldEdge((edge? hedge.twin() : hedge).origin()),
      d(new Impl(this, hedge, fixType, edge, Vector2f(materialOffsetS, 0))) {
    d->prepare();
}
Vector3d const &SkyFixEdge::pOrigin() const {
    return d->pOrigin;
}
Vector3d const &SkyFixEdge::pDirection() const {
    return d->pDirection;
}
Vector2f SkyFixEdge::materialOrigin() const {
    return d->materialOrigin;
}
bool SkyFixEdge::isValid() const {
    return d->isValid;
}
SkyFixEdge::Event const &SkyFixEdge::first() const {
    return d->bottom;
}
SkyFixEdge::Event const &SkyFixEdge::last() const {
    return d->top;
}
SkyFixEdge::Event const &SkyFixEdge::at(EventIndex index) const {
    if(index >= 0 && index < 2) {
        return index == 0? d->bottom : d->top;
    }
    throw Error("SkyFixEdge::at", QString("Index '%1' does not map to a known event (count: 2)").arg(index));
}
} // namespace de
namespace render {
static String const DEF_ALWAYS_TRIGGER("alwaysTrigger");
static String const DEF_MUST_FINISH   ("mustFinish");
static String const DEF_ANGLE         ("angle");
static String const DEF_AXIS          ("axis");
static String const DEF_DURATION      ("duration");
static String const DEF_ENABLED       ("enabled");
static String const DEF_LOOPING       ("looping");
static String const DEF_NODE          ("node");
static String const DEF_PRIORITY      ("priority");
static String const DEF_PROBABILITY   ("prob");
static String const DEF_SPEED         ("speed");
static String const DEF_VARIABLE      ("variable");
static String const VAR_ID            ("ID");           // model asset ID
static String const VAR_ASSET         ("__asset__");    // runtime reference to asset metadata
static String const VAR_ENABLED       ("enabled");
static String const VAR_MATERIAL      ("material");
static String const VAR_NOTIFIED_STATES("notifiedStates");
static String const PASS_GLOBAL       ("");
static String const DEFAULT_MATERIAL  ("default");
static int const ANIM_DEFAULT_PRIORITY = 1;
DENG2_PIMPL(StateAnimator)
, DENG2_OBSERVES(Variable, Change) {
    enum BindOperation { Bind, Unbind };
    /**
     * Specialized animation sequence state for a running animation.
     */
    struct Sequence : public ModelDrawable::Animator::OngoingSequence {
        enum LoopMode {
            NotLooping = 0,
            Looping = 1
        };
        LoopMode looping = NotLooping;
        bool mustFinish = false; // Sequence will play to end unless higher priority started.
        int priority = ANIM_DEFAULT_PRIORITY;
        Timeline const *timeline = nullptr; // owned by ModelRenderer::AnimSequence
        std::unique_ptr<Timeline::Clock> clock;
        TimeSpan overrideDuration = -1.0;
        TimeSpan actualRuntime = 0.0;
        Sequence() {}
        Sequence(int animationId, String const &rootNode, LoopMode looping, bool mustFinish,
                 int priority, Timeline const *timeline = nullptr)
            : looping(looping)
            , mustFinish(mustFinish)
            , priority(priority)
            , timeline(timeline) {
            animId = animationId;
            node   = rootNode;
        }
        Sequence(Sequence const &other) : OngoingSequence(other) {
            apply(other);
        }
        Sequence &operator=(Sequence const &other) {
            apply(other);
            return *this;
        }
        void initialize() override {
            ModelDrawable::Animator::OngoingSequence::initialize();
        }
        void apply(Sequence const &other) {
            animId     = other.animId;
            node       = other.node;
            looping    = other.looping;
            mustFinish = other.mustFinish;
            priority   = other.priority;
            timeline   = other.timeline;
            if (other.overrideDuration >= 0) {
                duration = other.overrideDuration;
            }
            clock.reset();
        }
        bool isRunning() const {
            if (looping == Looping) {
                return true;
            }
            return !atEnd();
        }
        int loopCount() const {
            if (duration > 0) {
                return int(actualRuntime / duration);
            }
            return 0;
        }
        void operator>>(Writer &to) const override {
            ModelDrawable::Animator::OngoingSequence::operator>>(to);
            to << duint8(looping) << priority;
        }
        void operator<<(Reader &from) override {
            ModelDrawable::Animator::OngoingSequence::operator<<(from);
            from.readAs<duint8>(looping);
            from >> priority;
        }
        static Sequence *make() { return new Sequence; }
    };
    QHash<String, Sequence> pendingAnimForNode;
    String currentStateName;
    Record names; ///< Local context for scripts, i.e., per-object model state.
    String ownerNamespaceVarName;
    ModelDrawable::Appearance appearance;
    QHash<String, int> indexForPassName;
    QHash<Variable *, int> passForMaterialVariable;
    QHash<String, ShaderVars *> passVars;
    struct AnimVar {
        String variableName;
        SafePtr<AnimationValue> angle; // not owned
        SafePtr<AnimationValue> speed; // not owned
        Vector3f axis;
    };
    typedef QHash<String, AnimVar *> AnimVars;
    AnimVars animVars;
    struct StateCallback : public ConditionalTrigger {
        Record &names;
        StateCallback(Record &names) : names(names) {
            setCondition(names[VAR_NOTIFIED_STATES]);
        }
        void handleTriggered(String const &trigger) override {
            Record ns;
            ns.add(QStringLiteral("self")).set(new RecordValue(names));
            Process::scriptCall(Process::IgnoreResult, ns,
                                QStringLiteral("self.__asset__.onStateChange"),
                                "$self",    // StateAnimator instance
                                trigger);   // new state
        }
    };
    std::unique_ptr<StateCallback> stateCallback;
    std::unique_ptr<Scheduler> scheduler; // only created if needed for additional timelines
    Impl(Public *i, DotPath const &assetId) : Base(i) {
        if (!assetId.isEmpty()) {
            initVariables(assetId);
            if (!self().model().passes.isEmpty()) {
                appearance.drawPasses = &self().model().passes;
            }
            appearance.programCallback = [this] (GLProgram &program, ModelDrawable::ProgramBinding binding) {
                bindUniforms(program,
                    binding == ModelDrawable::AboutToBind? Bind : Unbind);
            };
            appearance.passCallback = [this] (ModelDrawable::Pass const &pass, ModelDrawable::PassState state) {
                bindPassUniforms(*self().model().currentProgram(),
                    pass.name,
                    state == ModelDrawable::PassBegun? Bind : Unbind);
            };
        }
    }
    ~Impl() {
        deinitVariables();
    }
    void initVariables(DotPath const &assetId) {
        names.add(Record::VAR_NATIVE_SELF).set(new NativePointerValue(thisPublic)).setReadOnly();
        names.addSuperRecord(ScriptSystem::builtInClass(QStringLiteral("Render"),
                                                        QStringLiteral("StateAnimator")));
        names.addText(VAR_ID, assetId).setReadOnly();
        Record const &assetDef = App::asset(assetId).accessedRecord();
        names.add(VAR_ASSET).set(new RecordValue(assetDef)).setReadOnly();
        if (assetDef.has(VAR_NOTIFIED_STATES)) {
            names.add(VAR_NOTIFIED_STATES).set(assetDef.get(VAR_NOTIFIED_STATES));
        }
        else {
            names.addArray(VAR_NOTIFIED_STATES);
        }
        if (assetDef.has(QStringLiteral("onStateChange"))) {
            stateCallback.reset(new StateCallback(names));
        }
        int const passCount = self().model().passes.size();
        indexForPassName.clear();
        appearance.passMaterial.clear();
        if (!passCount) {
            appearance.passMaterial << 0;
        }
        else {
            for (int i = 0; i < passCount; ++i) {
                appearance.passMaterial << 0;
            }
        }
        appearance.passMask.resize(passCount);
        names.addText(VAR_MATERIAL, DEFAULT_MATERIAL);
        int passIndex = 0;
        auto const &def = names[VAR_ASSET].valueAsRecord();
        if (def.has(ModelLoader::DEF_RENDER)) {
            Record const &renderBlock = def.subrecord(ModelLoader::DEF_RENDER);
            initVariablesForPass(renderBlock);
            auto passes = ScriptedInfo::subrecordsOfType(ModelLoader::DEF_PASS, renderBlock);
            for (String passName : ScriptedInfo::sortRecordsBySource(passes)) {
                Record const &passDef = *passes[passName];
                indexForPassName[passName] = passIndex++;
                Record &passRec = names.addSubrecord(passName);
                passRec.addBoolean(VAR_ENABLED,
                                   ScriptedInfo::isTrue(passDef, DEF_ENABLED, true))
                       .audienceForChange() += this;
                initVariablesForPass(passDef, passName);
            }
        }
        if (def.has(ModelLoader::DEF_ANIMATION)) {
            auto varDefs = ScriptedInfo::subrecordsOfType(DEF_VARIABLE, def.subrecord(ModelLoader::DEF_ANIMATION));
            for (String varName : varDefs.keys()) {
                initAnimationVariable(varName, *varDefs[varName]);
            }
        }
        DENG2_ASSERT(passIndex == passCount);
        DENG2_ASSERT(indexForPassName.size() == passCount);
        updatePassMask();
        updatePassMaterials();
    }
    void initVariablesForPass(Record const &block,
                              String const &passName = PASS_GLOBAL) {
        Variable &passMaterialVar = names.addText(passName.concatenateMember(VAR_MATERIAL),
                                                  block.gets(ModelLoader::DEF_MATERIAL, DEFAULT_MATERIAL));
        passMaterialVar.audienceForChange() += this;
        passForMaterialVariable.insert(&passMaterialVar, self().model().passes.findName(passName));
        ShaderVars *vars = passVars[passName];
        if (!vars) {
            vars = new ShaderVars;
            passVars.insert(passName, vars);
        }
        auto varDefs = ScriptedInfo::subrecordsOfType(DEF_VARIABLE, block);
        for (auto i = varDefs.constBegin(); i != varDefs.constEnd(); ++i) {
            vars->initVariableFromDefinition(
                        i.key(), *i.value(),
                        names.addSubrecord(passName, Record::KeepExisting));
        }
    }
    void initAnimationVariable(String const &variableName,
                               Record const &variableDef) {
        try {
            std::unique_ptr<AnimVar> var(new AnimVar);
            var->variableName = variableName;
            var->angle.reset(new AnimationValue(Animation(variableDef.getf(DEF_ANGLE, 0.f), Animation::Linear)));
            var->speed.reset(new AnimationValue(Animation(variableDef.getf(DEF_SPEED, 0.f), Animation::Linear)));
            var->axis = vectorFromValue<Vector3f>(variableDef.get(DEF_AXIS));
            addBinding(variableName.concatenateMember(DEF_ANGLE), var->angle);
            addBinding(variableName.concatenateMember(DEF_SPEED), var->speed);
            animVars.insert(variableDef.gets(DEF_NODE), var.get());
            var.release();
            self().setFlags(AlwaysTransformNodes);
        }
        catch (Error const &er) {
            LOG_GL_WARNING("%s: %s") << ScriptedInfo::sourceLocation(variableDef)
                                     << er.asText();
        }
    }
    void addBinding(String const &varName, AnimationValue *anim) {
        names.add(varName).set(anim).setReadOnly(); // ownership of anim taken
    }
    void deinitVariables() {
        appearance.passMaterial.clear();
        qDeleteAll(passVars.values());
        passVars.clear();
        qDeleteAll(animVars.values());
        animVars.clear();
    }
    Variable const &materialVariableForPass(duint passIndex) const {
        auto const &model = self().model();
        if (!model.passes.isEmpty()) {
            String const varName = model.passes.at(passIndex).name.concatenateMember(VAR_MATERIAL);
            if (names.has(varName)) {
                return names[varName];
            }
        }
        return names[VAR_MATERIAL];
    }
    void updatePassMaterials() {
        for (int i = 0; i < appearance.passMaterial.size(); ++i) {
            appearance.passMaterial[i] = materialForUserProvidedName(
                        materialVariableForPass(i).value().asText());
        }
    }
    void variableValueChanged(Variable &var, Value const &newValue) {
        if (var.name() == VAR_MATERIAL) {
            int passIndex = passForMaterialVariable[&var];
            if (passIndex < 0) {
                updatePassMaterials();
            }
            else {
                appearance.passMaterial[passIndex] = materialForUserProvidedName(newValue.asText());
            }
        }
        else {
            DENG2_ASSERT(var.name() == VAR_ENABLED);
            updatePassMask();
        }
    }
    /**
     * Determines the material to use during a rendering pass. These are
     * determined by the "material" variables in the object's namespace.
     *
     * @param materialName  Name of the material to use.
     *
     * @return Material index.
     */
    duint materialForUserProvidedName(String const &materialName) const {
       auto const &model = self().model();
       auto const matIndex = model.materialIndexForName.constFind(materialName);
       if (matIndex != model.materialIndexForName.constEnd()) {
           return matIndex.value();
       }
       LOG_RES_WARNING("Asset \"%s\" does not have a material called '%s'")
               << names.gets(VAR_ID) << materialName;
       return 0; // default material
    }
    void updatePassMask() {
        Record::Subrecords enabledPasses = names.subrecords([] (Record const &sub) {
            return sub.getb(DEF_ENABLED, false);
        });
        appearance.passMask.fill(false);
        for (String name : enabledPasses.keys()) {
            DENG2_ASSERT(indexForPassName.contains(name));
            appearance.passMask.setBit(indexForPassName[name], true);
        }
    }
    void updateAnimationValuePointers() {
        foreach (String passName, passVars.keys()) {
            ShaderVars *vars = passVars[passName];
            foreach (String name, vars->members.keys()) {
                vars->members[name]->updateValuePointers(names, passName.concatenateMember(name));
            }
        }
        foreach (String name, animVars.keys()) {
            AnimVar &animVar = *animVars[name];
            animVar.angle = &names[animVar.variableName.concatenateMember(DEF_ANGLE)].value<AnimationValue>();
            animVar.speed = &names[animVar.variableName.concatenateMember(DEF_SPEED)].value<AnimationValue>();
        }
    }
    /**
     * Checks if a shader definition has a declaration for a variable.
     *
     * @param program  Shader definition.
     * @param uniform  Uniform.
     * @return @c true, if a variable exists matching @a uniform.
     */
    static bool hasDeclaredVariable(Record const &shaderDef, GLUniform const &uniform) {
        return shaderDef.hasMember(String::fromUtf8(uniform.name()));
    }
    /**
     * Binds or unbinds uniforms that apply to all rendering passes.
     *
     * @param program    Program where bindings are made.
     * @param operation  Bind or unbind.
     */
    void bindUniforms(de::GLProgram &program, BindOperation operation) const {
        bindPassUniforms(program, PASS_GLOBAL, operation);
    }
    /**
     * Binds or unbinds uniforms that apply to a single rendering pass.
     *
     * @param program    Program where bindings are made.
     * @param passName   Name of the rendering pass. The render variables are
     *                   named, e.g., "render.(passName).uName".
     * @param operation  Bind or unbind.
     */
    void bindPassUniforms(de::GLProgram &program,
                          de::String const &passName,
                          BindOperation operation) const {
        auto const &modelLoader = ClientApp::renderSystem().modelRenderer().loader();
        auto const vars = passVars.constFind(passName);
        if (vars != passVars.constEnd()) {
            for (auto i : vars.value()->members) {
                if (!hasDeclaredVariable(modelLoader.shaderDefinition(program),
                                         *i->uniform))
                    continue;
                if (operation == Bind) {
                    i->updateUniform();
                    program.bind(*i->uniform);
                }
                else {
                    program.unbind(*i->uniform);
                }
            }
        }
    }
    int animationId(String const &name) const {
        return ModelLoader::identifierFromText(name, [this] (String const &name) {
            return self().model().animationIdForName(name);
        });
    }
    Sequence &start(Sequence const &spec) {
        Sequence &anim = self().start(spec.animId, spec.node).as<Sequence>();
        anim.apply(spec);
        if (anim.timeline) {
            anim.clock.reset(new Timeline::Clock(*anim.timeline, &names));
        }
        applyFlagOperation(anim.flags, Sequence::ClampToDuration,
                           anim.looping == Sequence::Looping? UnsetFlags : SetFlags);
        return anim;
    }
};
StateAnimator::StateAnimator()
    : ModelDrawable::Animator(Impl::Sequence::make)
    , d(new Impl(this, {}))
{}
StateAnimator::StateAnimator(DotPath const &id, Model const &model)
    : ModelDrawable::Animator(model, Impl::Sequence::make)
    , d(new Impl(this, id))
{}
Model const &StateAnimator::model() const {
    return static_cast<Model const &>(ModelDrawable::Animator::model());
}
Scheduler &StateAnimator::scheduler() {
    if (!d->scheduler) {
        d->scheduler.reset(new Scheduler);
    }
    return *d->scheduler;
}
void StateAnimator::setOwnerNamespace(Record &names, String const &varName) {
    d->ownerNamespaceVarName = varName;
    d->names.add(d->ownerNamespaceVarName).set(new RecordValue(names));
    if (d->names.has(QStringLiteral("__asset__.onInit"))) {
        Record ns;
        ns.add(QStringLiteral("self")).set(new RecordValue(d->names));
        Process::scriptCall(Process::IgnoreResult, ns,
                            "self.__asset__.onInit",
                            "$self");
    }
}
String StateAnimator::ownerNamespaceName() const {
    return d->ownerNamespaceVarName;
}
void StateAnimator::triggerByState(String const &stateName) {
    using Sequence = Impl::Sequence;
    if (d->stateCallback) {
        d->stateCallback->tryTrigger(stateName);
    }
    auto const *stateAnims = &model().animations;
    if (!stateAnims) return;
    auto found = stateAnims->constFind(stateName);
    if (found == stateAnims->constEnd()) return;
    LOG_AS("StateAnimator");
    d->currentStateName = stateName;
    foreach (Model::AnimSequence const &seq, found.value()) {
        try {
            float chance = seq.def->getf(DEF_PROBABILITY, 1.f);
            if (randf() > chance) continue;
            String const node = seq.def->gets(DEF_NODE, "");
            int animId = d->animationId(seq.name);
            if (animId < 0) {
                LOG_GL_ERROR("%s: animation sequence \"%s\" not found")
                    << ScriptedInfo::sourceLocation(*seq.def) << seq.name;
                break;
            }
            const bool alwaysTrigger = ScriptedInfo::isTrue(*seq.def, DEF_ALWAYS_TRIGGER, false);
            const bool mustFinish    = ScriptedInfo::isTrue(*seq.def, DEF_MUST_FINISH, false);
            if (!alwaysTrigger) {
                if (isRunning(animId, node)) continue;
            }
            int const priority = seq.def->geti(DEF_PRIORITY, ANIM_DEFAULT_PRIORITY);
            Timeline *timeline = seq.timeline;
            if (!seq.sharedTimeline.isEmpty()) {
                auto tl = model().timelines.constFind(seq.sharedTimeline);
                if (tl != model().timelines.constEnd()) {
                    timeline = tl.value();
                }
            }
            Sequence anim(animId, node,
                          ScriptedInfo::isTrue(*seq.def, DEF_LOOPING)? Sequence::Looping :
                                                                       Sequence::NotLooping,
                          mustFinish,
                          priority,
                          timeline);
            if (seq.def->has(DEF_DURATION)) {
                anim.overrideDuration = seq.def->getd(DEF_DURATION);
            }
            if (auto *existing = maybeAs<Sequence>(find(node))) {
                if (priority < existing->priority ||
                    (priority == existing->priority && existing->mustFinish &&
                     !existing->atEnd() && existing->loopCount() == 0)) {
                    d->pendingAnimForNode[node] = anim;
                    continue;
                }
            }
            d->pendingAnimForNode.remove(node);
            d->start(anim);
        }
        catch (ModelDrawable::Animator::InvalidError const &er) {
            LOG_GL_WARNING("Failed to start animation \"%s\": %s")
                    << seq.name << er.asText();
            continue;
        }
        break;
    }
}
void StateAnimator::triggerDamage(int points, struct mobj_s const *inflictor) {
    /*
     * Here we check for the onDamage() function in the asset. The __asset__
     * variable holds a direct pointer to the asset definition, where the
     * function is defined.
     */
    if (d->names.has(QStringLiteral("__asset__.onDamage"))) {
        /*
         * We need to provide the StateAnimator instance to the script as an
         * argument, because onDamage() is a member of the asset -- when the
         * method is executed, "self" refers to the asset.
         */
        Record ns;
        ns.add(QStringLiteral("self")).set(new RecordValue(d->names));
        Process::scriptCall(Process::IgnoreResult, ns,
                            QStringLiteral("self.__asset__.onDamage"),
                            "$self", points,
                            inflictor? &THINKER_DATA(inflictor->thinker, ThinkerData) :
                                       nullptr);
    }
}
void StateAnimator::startAnimation(int animationId, int priority, bool looping, String const &node) {
    LOG_AS("StateAnimator::startAnimation");
    try {
        using Seq = Impl::Sequence;
        d->start(Seq(animationId,
                     node,
                     looping ? Seq::Looping : Seq::NotLooping,
                     false, // same priority override allowed
                     priority));
    }
    catch (const Error &er) {
        LOG_GL_ERROR("%s: %s") << d->names.gets(VAR_ID) << er.asText();
    }
}
int StateAnimator::animationId(String const &name) const {
    return d->animationId(name);
}
void StateAnimator::advanceTime(TimeSpan const &elapsed) {
    ModelDrawable::Animator::advanceTime(elapsed);
    using Sequence = Impl::Sequence;
    bool retrigger = false;
    for (auto *var : d->animVars.values()) {
        if (!var->angle || !var->speed) continue;
        var->angle->animation().shift(var->speed->animation() * elapsed);
        float varAngle = var->angle->animation();
        if (varAngle > 360)    var->angle->animation().shift(-360);
        else if (varAngle < 0) var->angle->animation().shift(+360);
    }
    for (int i = 0; i < count(); ++i) {
        auto &anim = at(i).as<Sequence>();
        ddouble factor = 1.0;
        TimeSpan animElapsed = factor * elapsed;
        anim.time += animElapsed;
        anim.actualRuntime += animElapsed;
        if (anim.looping == Sequence::NotLooping) {
            anim.time = min(anim.time, anim.duration);
        }
        if (anim.looping == Sequence::Looping) {
            if (anim.atEnd()) {
                retrigger = true;
                anim.time -= anim.duration; // Trigger only once per loop.
                if (anim.clock) {
                    anim.clock->rewind(anim.time);
                }
            }
        }
        if (anim.clock) {
            anim.clock->advanceTime(animElapsed);
        }
        if (!anim.isRunning()) {
            String const node = anim.node;
            if (count() > 1) {
                stop(i--); // `anim` gets deleted
            }
            auto &pending = d->pendingAnimForNode;
            if (pending.contains(node)) {
                LOG_GL_VERBOSE("Starting pending animation %i") << pending[node].animId;
                d->start(pending[node]);
                pending.remove(node);
            }
        }
    }
    if (d->scheduler) {
        d->scheduler->advanceTime(elapsed);
    }
    if (retrigger && !d->currentStateName.isEmpty()) {
        triggerByState(d->currentStateName);
    }
}
ddouble StateAnimator::currentTime(int index) const {
    return ModelDrawable::Animator::currentTime(index); // + frameTimePos;
}
Vector4f StateAnimator::extraRotationForNode(String const &nodeName) const {
    auto found = d->animVars.constFind(nodeName);
    if (found != d->animVars.constEnd()) {
        Impl::AnimVar const &var = *found.value();
        if (var.angle) {
            return Vector4f(var.axis, var.angle->animation());
        }
    }
    return Vector4f();
}
ModelDrawable::Appearance const &StateAnimator::appearance() const {
    return d->appearance;
}
Record &StateAnimator::objectNamespace() {
    return d->names;
}
Record const &StateAnimator::objectNamespace() const {
    return d->names;
}
void StateAnimator::operator >> (Writer &to) const {
    ModelDrawable::Animator::operator >> (to);
    to << d->currentStateName
       << Record(d->names, Record::IgnoreDoubleUnderscoreMembers);
}
void StateAnimator::operator << (Reader &from) {
    d->pendingAnimForNode.clear();
    ModelDrawable::Animator::operator << (from);
    from >> d->currentStateName;
    Record storedNames;
    from >> storedNames;
    Record const oldNames = d->names;
    try {
        d->names.assignPreservingVariables(storedNames, Record::IgnoreDoubleUnderscoreMembers);
        d->updateAnimationValuePointers();
    }
    catch (...) {
        d->names = oldNames;
        d->updateAnimationValuePointers();
        throw;
    }
}
} // namespace render
Store::Store() {
    de::zap(texCoords);
}
Store::~Store() {
    clear();
}
void Store::rewind() {
    _vertCount = 0;
}
void Store::clear() {
    _vertCount = _vertMax = 0;
    M_Free(posCoords); posCoords = nullptr;
    M_Free(colorCoords); colorCoords = nullptr;
    for(dint i = 0; i < 2; ++i) {
        M_Free(texCoords[i]); texCoords[i] = nullptr;
    }
    M_Free(modCoords); modCoords = nullptr;
}
duint Store::allocateVertices(duint count) {
    duint const base = _vertCount;
    _vertCount += count;
    while(_vertCount > _vertMax) {
        if(_vertMax == 0) {
            _vertMax = 16;
        }
        else {
            _vertMax *= 2;
        }
        posCoords   = (Vector3f *)  M_Realloc(posCoords,   sizeof(*posCoords) * _vertMax);
        colorCoords = (Vector4ub *) M_Realloc(colorCoords, sizeof(*colorCoords) * _vertMax);
        for(dint i = 0; i < 2; ++i) {
            texCoords[i] = (Vector2f *) M_Realloc(texCoords[i], sizeof(Vector2f) * _vertMax);
        }
        modCoords   = (Vector2f *) M_Realloc(modCoords,  sizeof(*modCoords) * _vertMax);
    }
    return base;
}
DENG2_PIMPL(TriangleStripBuilder) {
    ClockDirection direction;
    bool buildTexCoords;
    int initialReserveElements;
    QScopedPointer<PositionBuffer> positions;
    QScopedPointer<TexCoordBuffer> texcoords;
    Impl(Public *i, bool buildTexCoords)
        : Base(i),
          direction(Clockwise),
          buildTexCoords(buildTexCoords),
          initialReserveElements(0),
          positions(0),
          texcoords(0)
    {}
    void reserveElements(int num) {
        if(num < 0) return; // Huh?
        if(positions.isNull()) {
            positions.reset(new PositionBuffer());
            if(buildTexCoords)
                texcoords.reset(new TexCoordBuffer());
            num += initialReserveElements;
        }
        positions->reserve(positions->size() + num);
        if(buildTexCoords)
            texcoords->reserve(texcoords->size() + num);
    }
};
TriangleStripBuilder::TriangleStripBuilder(bool buildTexCoords)
    : d(new Impl(this, buildTexCoords))
{}
void TriangleStripBuilder::begin(ClockDirection direction, int reserveElements) {
    d->direction = direction;
    d->initialReserveElements = de::max(reserveElements, 0);
    d->positions.reset();
    d->texcoords.reset();
}
void TriangleStripBuilder::extend(AbstractEdge &edge) {
    if(!edge.isValid()) return;
    AbstractEdge::Event const &from = edge.first();
    AbstractEdge::Event const &to   = edge.last();
    d->reserveElements(2);
    d->positions->append((d->direction == Anticlockwise? to : from).origin());
    d->positions->append((d->direction == Anticlockwise? from : to).origin());
    if(d->buildTexCoords) {
        double edgeLength = to.origin().z - from.origin().z;
        d->texcoords->append(edge.materialOrigin() + Vector2f(0, (d->direction == Anticlockwise? 0 : edgeLength)));
        d->texcoords->append(edge.materialOrigin() + Vector2f(0, (d->direction == Anticlockwise? edgeLength : 0)));
    }
}
int TriangleStripBuilder::numElements() const {
    return d->positions.isNull()? 0 : d->positions->size();
}
int TriangleStripBuilder::take(PositionBuffer **positions, TexCoordBuffer **texcoords) {
    int retNumElements = numElements();
    *positions = d->positions.take();
    if(texcoords) {
        *texcoords = d->texcoords.take();
    }
    return retNumElements;
}
VectorLightList::Node *VectorLightList::firstNode  = nullptr;
VectorLightList::Node *VectorLightList::cursorNode = nullptr;
void VectorLightList::init()  // static
{
    static bool firstTime = true;
    if(firstTime) {
        firstNode  = 0;
        cursorNode = 0;
        firstTime  = false;
    }
}
void VectorLightList::rewind()  // static
{
    cursorNode = firstNode;
}
VectorLightList &VectorLightList::add(VectorLightData &vlight) {
    Node *node = newNode();
    node->vlight = vlight;
    if(head) {
        Node *iter = head;
        Node *last = iter;
        do {
            VectorLightData *vlight = &node->vlight;
            if(node->vlight.approxDist > vlight->approxDist) {
                last = iter;
                iter = iter->next;
            }
            else {
                node->next = last->next;
                last->next = node;
                return *this;
            }
        } while(iter);
    }
    node->next = head;
    head = node;
    return *this;
}
VectorLightList::Node *VectorLightList::newNode()  // static
{
    Node *node;
    if(!cursorNode) {
        node = (Node *) Z_Malloc(sizeof(*node), PU_APPSTATIC, nullptr);
        node->nextUsed = firstNode;
        firstNode = node;
    }
    else {
        node = cursorNode;
        cursorNode = cursorNode->nextUsed;
    }
    node->next = nullptr;
    return node;
}
dd_bool firstFrameAfterLoad;
static dint loadInStartupMode;
static dint rendCameraSmooth = true;  ///< Smoothed by default.
static dbyte showFrameTimePos;
static dbyte showViewAngleDeltas;
static dbyte showViewPosDeltas;
dint rendInfoTris;
static viewport_t *currentViewport;
struct FrameLuminous {
    coord_t distance;
    duint isClipped;
};
static QVector<FrameLuminous> frameLuminous;
static QBitArray subspacesVisible;
static QBitArray generatorsVisible(Map::MAX_GENERATORS);
static dint frameCount;
static dint gridCols, gridRows;
static viewport_t viewportOfLocalPlayer[DDMAXPLAYERS];
static dint resetNextViewer = true;
dint R_FrameCount() {
    return frameCount;
}
void R_ResetFrameCount() {
    frameCount = 0;
}
#undef R_SetViewOrigin
DENG_EXTERN_C void R_SetViewOrigin(dint consoleNum, coord_t const origin[3]) {
    if(consoleNum < 0 || consoleNum >= DDMAXPLAYERS) return;
    DD_Player(consoleNum)->viewport().latest.origin = Vector3d(origin);
}
#undef R_SetViewAngle
DENG_EXTERN_C void R_SetViewAngle(dint consoleNum, angle_t angle) {
    if(consoleNum < 0 || consoleNum >= DDMAXPLAYERS) return;
    DD_Player(consoleNum)->viewport().latest.setAngle(angle);
}
#undef R_SetViewPitch
DENG_EXTERN_C void R_SetViewPitch(dint consoleNum, dfloat pitch) {
    if(consoleNum < 0 || consoleNum >= DDMAXPLAYERS) return;
    DD_Player(consoleNum)->viewport().latest.pitch = pitch;
}
void R_SetupDefaultViewWindow(dint consoleNum) {
    viewdata_t *vd = &DD_Player(consoleNum)->viewport();
    if(consoleNum < 0 || consoleNum >= DDMAXPLAYERS) return;
    vd->window =
        vd->windowOld =
            vd->windowTarget = Rectanglei::fromSize(Vector2i(0, 0), Vector2ui(DENG_GAMEVIEW_WIDTH, DENG_GAMEVIEW_HEIGHT));
    vd->windowInter = 1;
}
void R_ViewWindowTicker(dint consoleNum, timespan_t ticLength) {
    viewdata_t *vd = &DD_Player(consoleNum)->viewport();
    if(consoleNum < 0 || consoleNum >= DDMAXPLAYERS) {
        return;
    }
    vd->windowInter += dfloat(.4 * ticLength * TICRATE);
    if(vd->windowInter >= 1) {
        vd->window = vd->windowTarget;
    }
    else {
        vd->window.moveTopLeft(Vector2i(de::roundf(de::lerp<dfloat>(vd->windowOld.topLeft.x, vd->windowTarget.topLeft.x, vd->windowInter)),
                                        de::roundf(de::lerp<dfloat>(vd->windowOld.topLeft.y, vd->windowTarget.topLeft.y, vd->windowInter))));
        vd->window.setSize(Vector2ui(de::roundf(de::lerp<dfloat>(vd->windowOld.width(),  vd->windowTarget.width(),  vd->windowInter)),
                                     de::roundf(de::lerp<dfloat>(vd->windowOld.height(), vd->windowTarget.height(), vd->windowInter))));
    }
}
#undef R_ViewWindowGeometry
DENG_EXTERN_C dint R_ViewWindowGeometry(dint player, RectRaw *geometry) {
    if(!geometry) return false;
    if(player < 0 || player >= DDMAXPLAYERS) return false;
    viewdata_t const &vd = DD_Player(player)->viewport();
    geometry->origin.x    = vd.window.topLeft.x;
    geometry->origin.y    = vd.window.topLeft.y;
    geometry->size.width  = vd.window.width();
    geometry->size.height = vd.window.height();
    return true;
}
#undef R_ViewWindowOrigin
DENG_EXTERN_C dint R_ViewWindowOrigin(dint player, Point2Raw *origin) {
    if(!origin) return false;
    if(player < 0 || player >= DDMAXPLAYERS) return false;
    viewdata_t const &vd = DD_Player(player)->viewport();
    origin->x = vd.window.topLeft.x;
    origin->y = vd.window.topLeft.y;
    return true;
}
#undef R_ViewWindowSize
DENG_EXTERN_C dint R_ViewWindowSize(dint player, Size2Raw *size) {
    if(!size) return false;
    if(player < 0 || player >= DDMAXPLAYERS) return false;
    viewdata_t const &vd = DD_Player(player)->viewport();
    size->width  = vd.window.width();
    size->height = vd.window.height();
    return true;
}
/**
 * @note Do not change values used during refresh here because we might be
 * partway through rendering a frame. Changes should take effect on next
 * refresh only.
 */
#undef R_SetViewWindowGeometry
DENG_EXTERN_C void R_SetViewWindowGeometry(dint player, RectRaw const *geometry, dd_bool interpolate) {
    dint p = P_ConsoleToLocal(player);
    if(p < 0) return;
    viewport_t const *vp = &viewportOfLocalPlayer[p];
    viewdata_t *vd = &DD_Player(player)->viewport();
    Rectanglei newGeom = Rectanglei::fromSize(Vector2i(de::clamp<dint>(0, geometry->origin.x, vp->geometry.width()),
                                                       de::clamp<dint>(0, geometry->origin.y, vp->geometry.height())),
                                              Vector2ui(de::abs(geometry->size.width),
                                                        de::abs(geometry->size.height)));
    if((unsigned) newGeom.bottomRight.x > vp->geometry.width()) {
        newGeom.setWidth(vp->geometry.width() - newGeom.topLeft.x);
    }
    if((unsigned) newGeom.bottomRight.y > vp->geometry.height()) {
        newGeom.setHeight(vp->geometry.height() - newGeom.topLeft.y);
    }
    if(vd->window == newGeom) {
        return;
    }
    vd->windowTarget = newGeom;
    if(interpolate && vd->window.size() != Vector2ui(0, 0)) {
        vd->windowOld   = vd->window;
        vd->windowInter = 0;
    }
    else {
        vd->windowOld   = vd->windowTarget;
        vd->windowInter = 1;  // Update on next frame.
    }
}
#undef R_ViewPortGeometry
DENG_EXTERN_C dint R_ViewPortGeometry(dint player, RectRaw *geometry) {
    if(!geometry) return false;
    dint p = P_ConsoleToLocal(player);
    if(p == -1) return false;
    viewport_t const &vp = viewportOfLocalPlayer[p];
    geometry->origin.x    = vp.geometry.topLeft.x;
    geometry->origin.y    = vp.geometry.topLeft.y;
    geometry->size.width  = vp.geometry.width();
    geometry->size.height = vp.geometry.height();
    return true;
}
#undef R_ViewPortOrigin
DENG_EXTERN_C dint R_ViewPortOrigin(dint player, Point2Raw *origin) {
    if(!origin) return false;
    dint p = P_ConsoleToLocal(player);
    if(p == -1) return false;
    viewport_t const &vp = viewportOfLocalPlayer[p];
    origin->x = vp.geometry.topLeft.x;
    origin->y = vp.geometry.topLeft.y;
    return true;
}
#undef R_ViewPortSize
DENG_EXTERN_C dint R_ViewPortSize(dint player, Size2Raw *size) {
    if(!size) return false;
    dint p = P_ConsoleToLocal(player);
    if(p == -1) return false;
    viewport_t const &vp = viewportOfLocalPlayer[p];
    size->width  = vp.geometry.width();
    size->height = vp.geometry.height();
    return true;
}
#undef R_SetViewPortPlayer
DENG_EXTERN_C void R_SetViewPortPlayer(dint consoleNum, dint viewPlayer) {
    dint p = P_ConsoleToLocal(consoleNum);
    if(p != -1) {
        viewportOfLocalPlayer[p].console = viewPlayer;
    }
}
/**
 * Calculate the placement and dimensions of a specific viewport.
 * Assumes that the grid has already been configured.
 */
void R_UpdateViewPortGeometry(viewport_t *port, dint col, dint row) {
    DENG2_ASSERT(port);
    Rectanglei newGeom = Rectanglei(Vector2i(DENG_GAMEVIEW_X + col * DENG_GAMEVIEW_WIDTH  / gridCols,
                                             DENG_GAMEVIEW_Y + row * DENG_GAMEVIEW_HEIGHT / gridRows),
                                    Vector2i(DENG_GAMEVIEW_X + (col+1) * DENG_GAMEVIEW_WIDTH  / gridCols,
                                             DENG_GAMEVIEW_Y + (row+1) * DENG_GAMEVIEW_HEIGHT / gridRows));
    ddhook_viewport_reshape_t p;
    if(port->geometry == newGeom) return;
    bool doReshape = false;
    if(port->console != -1 && Plug_CheckForHook(HOOK_VIEWPORT_RESHAPE)) {
        p.oldGeometry.origin.x    = port->geometry.topLeft.x;
        p.oldGeometry.origin.y    = port->geometry.topLeft.y;
        p.oldGeometry.size.width  = port->geometry.width();
        p.oldGeometry.size.height = port->geometry.height();
        doReshape = true;
    }
    port->geometry = newGeom;
    if(doReshape) {
        p.geometry.origin.x    = port->geometry.topLeft.x;
        p.geometry.origin.y    = port->geometry.topLeft.y;
        p.geometry.size.width  = port->geometry.width();
        p.geometry.size.height = port->geometry.height();
        DoomsdayApp::plugins().callAllHooks(HOOK_VIEWPORT_RESHAPE, port->console, (void *)&p);
    }
}
bool R_SetViewGrid(dint numCols, dint numRows) {
    if(numCols > 0 && numRows > 0) {
        if(numCols * numRows > DDMAXPLAYERS) {
            return false;
        }
        if(numCols != gridCols || numRows != gridRows) {
            LensFx_GLRelease();
        }
        if(numCols > DDMAXPLAYERS)
            numCols = DDMAXPLAYERS;
        if(numRows > DDMAXPLAYERS)
            numRows = DDMAXPLAYERS;
        gridCols = numCols;
        gridRows = numRows;
    }
    dint p = 0;
    for(dint y = 0; y < gridRows; ++y)
    for(dint x = 0; x < gridCols; ++x) {
        viewport_t *vp = &viewportOfLocalPlayer[p];
        dint const console = P_LocalToConsole(p);
        if(console != -1) {
            vp->console = DD_Player(console)->viewConsole;
        }
        else {
            vp->console = -1;
        }
        R_UpdateViewPortGeometry(vp, x, y);
        ++p;
    }
    return true;
}
void R_ResetViewer() {
    resetNextViewer = 1;
}
dint R_NextViewer() {
    return resetNextViewer;
}
/**
 * The components whose difference is too large for interpolation will be
 * snapped to the sharp values.
 */
void R_CheckViewerLimits(viewer_t *src, viewer_t *dst) {
    dint const MAXMOVE = 32;
    if(fabs(dst->origin.x - src->origin.x) > MAXMOVE ||
       fabs(dst->origin.y - src->origin.y) > MAXMOVE) {
        src->origin = dst->origin;
    }
    /*
    if(abs(dint(dst->angle) - dint(src->angle)) >= ANGLE_45) {
        LOG_DEBUG("R_CheckViewerLimits: Snap camera angle to %08x.") << dst->angle;
        src->angle = dst->angle;
    }
    */
}
/**
 * Retrieve the current sharp camera position.
 */
viewer_t R_SharpViewer(ClientPlayer &player) {
    DENG2_ASSERT(player.publicData().mo);
    ddplayer_t const &ddpl = player.publicData();
    viewer_t view(player.viewport().latest);
    if((ddpl.flags & DDPF_CHASECAM) && !(ddpl.flags & DDPF_CAMERA)) {
        dfloat const distance = 90;
        duint angle = view.angle() >> ANGLETOFINESHIFT;
        duint pitch = angle_t(LOOKDIR2DEG(view.pitch) / 360 * ANGLE_MAX) >> ANGLETOFINESHIFT;
        view.origin -= Vector3d(FIX2FLT(fineCosine[angle]),
                                FIX2FLT(finesine[angle]),
                                FIX2FLT(finesine[pitch])) * distance;
    }
    if(!(ddpl.flags & DDPF_CAMERA)) {
        if(view.origin.z > ddpl.mo->ceilingZ - 4) {
            view.origin.z = ddpl.mo->ceilingZ - 4;
        }
        if(view.origin.z < ddpl.mo->floorZ + 4) {
            view.origin.z = ddpl.mo->floorZ + 4;
        }
    }
    return view;
}
void R_NewSharpWorld() {
    if(resetNextViewer) {
        resetNextViewer = 2;
    }
    for(dint i = 0; i < DDMAXPLAYERS; ++i) {
        player_t *plr  = DD_Player(i);
        viewdata_t *vd = &plr->viewport();
        if(!plr->isInGame()) {
            continue;
        }
        viewer_t sharpView = R_SharpViewer(*plr);
        vd->lastSharp[0] = vd->lastSharp[1];
        vd->lastSharp[1] = sharpView;
        R_CheckViewerLimits(vd->lastSharp, &sharpView);
    }
    if(ClientApp::world().hasMap()) {
        Map &map = ClientApp::world().map();
        map.updateTrackedPlanes();
        map.updateScrollingSurfaces();
    }
}
void R_UpdateViewer(dint consoleNum) {
    DENG2_ASSERT(consoleNum >= 0 && consoleNum < DDMAXPLAYERS);
    dint const VIEWPOS_MAX_SMOOTHDISTANCE = 172;
    player_t *player = DD_Player(consoleNum);
    viewdata_t *vd   = &player->viewport();
    if(!player->isInGame()) return;
    viewer_t sharpView = R_SharpViewer(*player);
    if(resetNextViewer ||
       (sharpView.origin - vd->current.origin).length() > VIEWPOS_MAX_SMOOTHDISTANCE) {
        if(resetNextViewer > 1) {
            resetNextViewer = 0;
        }
        vd->current = sharpView;
        vd->lastSharp[0] = vd->lastSharp[1] = sharpView;
    }
    else  //if(!clientPaused)
    {
        viewer_t smoothView = vd->lastSharp[0].lerp(vd->lastSharp[1], frameTimePos);
        /*if(!(player->shared.flags & DDPF_INTERYAW))
            smoothView.angle = sharpView.angle;*/
        /*if(!(player->shared.flags & DDPF_INTERPITCH))
            smoothView.pitch = sharpView.pitch;*/
        vd->current = smoothView;
        if(showViewAngleDeltas) {
            struct OldAngle {
                ddouble time;
                dfloat yaw;
                dfloat pitch;
            };
            static OldAngle oldAngle[DDMAXPLAYERS];
            OldAngle *old = &oldAngle[DoomsdayApp::players().indexOf(viewPlayer)];
            dfloat yaw    = (ddouble)smoothView.angle() / ANGLE_MAX * 360;
            LOGDEV_MSG("(%i) F=%.3f dt=%-10.3f dx=%-10.3f dy=%-10.3f "
                       "Rdx=%-10.3f Rdy=%-10.3f")
                    << SECONDS_TO_TICKS(gameTime)
                    << frameTimePos
                    << sysTime - old->time
                    << yaw - old->yaw
                    << smoothView.pitch - old->pitch
                    << (yaw - old->yaw) / (sysTime - old->time)
                    << (smoothView.pitch - old->pitch) / (sysTime - old->time);
            old->yaw   = yaw;
            old->pitch = smoothView.pitch;
            old->time  = sysTime;
        }
        if(showViewPosDeltas) {
            struct OldPos {
                ddouble time;
                Vector3f pos;
            };
            static OldPos oldPos[DDMAXPLAYERS];
            OldPos *old = &oldPos[DoomsdayApp::players().indexOf(viewPlayer)];
            LOGDEV_MSG("(%i) F=%.3f dt=%-10.3f dx=%-10.3f dy=%-10.3f dz=%-10.3f dx/dt=%-10.3f dy/dt=%-10.3f")
                    << SECONDS_TO_TICKS(gameTime)
                    << frameTimePos
                    << sysTime - old->time
                    << smoothView.origin.x - old->pos.x
                    << smoothView.origin.y - old->pos.y
                    << smoothView.origin.z - old->pos.z
                    << (smoothView.origin.x - old->pos.x) / (sysTime - old->time)
                    << (smoothView.origin.y - old->pos.y) / (sysTime - old->time);
            old->pos  = smoothView.origin;
            old->time = sysTime;
        }
    }
    angle_t const viewYaw = vd->current.angle();
    duint const an = viewYaw >> ANGLETOFINESHIFT;
    vd->viewSin = FIX2FLT(finesine[an]);
    vd->viewCos = FIX2FLT(fineCosine[an]);
    dfloat const yawRad   = ((viewYaw / (dfloat) ANGLE_MAX) *2) * PI;
    dfloat const pitchRad = vd->current.pitch * 85 / 110.f / 180 * PI;
    vd->frontVec.x = cos(yawRad) * cos(pitchRad);
    vd->frontVec.z = sin(yawRad) * cos(pitchRad);
    vd->frontVec.y = sin(pitchRad);
    vd->upVec.x = -cos(yawRad) * sin(pitchRad);
    vd->upVec.z = -sin(yawRad) * sin(pitchRad);
    vd->upVec.y = cos(pitchRad);
    vd->sideVec = vd->frontVec.cross(vd->upVec);
}
/**
 * Prepare rendering the view of the given player.
 */
void R_SetupFrame(player_t *player) {
#define MINEXTRALIGHTFRAMES         2
    viewPlayer = player;
    if(showFrameTimePos) {
        LOGDEV_VERBOSE("frametime = %f") << frameTimePos;
    }
    if(player->targetExtraLight != player->publicData().extraLight) {
        player->targetExtraLight = player->publicData().extraLight;
        player->extraLightCounter = MINEXTRALIGHTFRAMES;
    }
    if(player->extraLightCounter > 0) {
        player->extraLightCounter--;
        if(player->extraLightCounter == 0)
            player->extraLight = player->targetExtraLight;
    }
    validCount++;
    extraLight      = player->extraLight;
    extraLightDelta = extraLight / 16.0f;
    if(!freezeRLs) {
        R_ClearVisSprites();
    }
#undef MINEXTRALIGHTFRAMES
}
void R_RenderPlayerViewBorder() {
    R_DrawViewBorder();
}
void R_UseViewPort(viewport_t const *vp) {
    DENG2_ASSERT_IN_RENDER_THREAD();
    DENG_ASSERT_GL_CONTEXT_ACTIVE();
    if (!vp) {
        currentViewport = nullptr;
        /*ClientWindow::main().game().glApplyViewport(
                Rectanglei::fromSize(Vector2i(DENG_GAMEVIEW_X, DENG_GAMEVIEW_Y),
                                     Vector2ui(DENG_GAMEVIEW_WIDTH, DENG_GAMEVIEW_HEIGHT)));*/
    }
    else {
        currentViewport = const_cast<viewport_t *>(vp);
    }
}
void R_UseViewPort(int consoleNum) {
    int local = P_ConsoleToLocal(consoleNum);
    if (local >= 0) {
        R_UseViewPort(&viewportOfLocalPlayer[local]);
    }
    else {
        R_UseViewPort(nullptr);
    }
}
Rectanglei R_ConsoleRect(int console) {
    int local = P_ConsoleToLocal(console);
    if (local < 0) return Rectanglei();
    auto const &port = viewportOfLocalPlayer[local];
    return Rectanglei(port.geometry.topLeft.x,
                      port.geometry.topLeft.y,
                      port.geometry.width(),
                      port.geometry.height());
}
Rectanglei R_Console3DViewRect(int console) {
    Rectanglei rect = R_ConsoleRect(console);
    auto const &pv = DD_Player(console)->viewport();
    return Rectanglei(rect.left() + pv.window.topLeft.x,
                      rect.top()  + pv.window.topLeft.y,
                      de::min(rect.width(),  pv.window.width()),
                      de::min(rect.height(), pv.window.height()));
}
viewport_t const *R_CurrentViewPort() {
    return currentViewport;
}
void R_RenderBlankView() {
    UI_DrawDDBackground(Point2Raw{0, 0}, Size2Raw{{{320, 200}}}, 1);
}
static void setupPlayerSprites() {
    DENG2_ASSERT(viewPlayer);
    ::psp3d = false;
    ddplayer_t *ddpl = &viewPlayer->publicData();
    if((ddpl->flags & DDPF_CAMERA) || (ddpl->flags & DDPF_CHASECAM))
        return;
    if(!ddpl->mo) return;
    mobj_t *mob = ddpl->mo;
    if(!Mobj_HasSubsector(*mob)) return;
    auto &subsec = Mobj_Subsector(*mob).as<world::ClientSubsector>();
    bool fullBright = CPP_BOOL(::levelFullBright);
    if(!fullBright) {
        for(ddpsprite_t const &psp : ddpl->pSprites) {
            if(!psp.statePtr) continue;
            if(psp.statePtr->flags & STF_FULLBRIGHT) {
                fullBright = true;
            }
        }
    }
    viewdata_t const *viewData = &viewPlayer->viewport();
    for(dint i = 0; i < DDMAXPSPRITES; ++i) {
        vispsprite_t *spr = &visPSprites[i];
        spr->type    = VPSPR_SPRITE;
        spr->psp     = &ddpl->pSprites[i];
        spr->origin  = viewData->current.origin;
        spr->bspLeaf = &Mobj_BspLeafAtOrigin(*mob);
        spr->alpha   = spr->psp->alpha;
        if(!spr->psp->statePtr) continue;
        spr->light.isFullBright = (spr->psp->flags & DDPSPF_FULLBRIGHT) != 0;
        FrameModelDef *mf = nullptr, *nextmf = nullptr;
        dfloat inter = 0;
        if(useModels) {
            if(viewPlayer->playerWeaponAnimator().hasModel()) {
                viewPlayer->playerWeaponAnimator().setupVisPSprite(*spr);
                ::psp3d = true;
                continue;
            }
            else {
                MobjThinker dummy;
                dummy->state = spr->psp->statePtr;
                dummy->tics  = spr->psp->tics;
                mf = Mobj_ModelDef(dummy, &nextmf, &inter);
            }
        }
        if(mf) {
            ::psp3d = true;
            spr->type = VPSPR_MODEL;
            spr->data.model.flags       = 0;
            spr->data.model.topZ        = viewData->current.origin.z;
            spr->data.model.secFloor    = subsec.visFloor().heightSmoothed();
            spr->data.model.secCeil     = subsec.visCeiling().heightSmoothed();
            spr->data.model.pClass      = 0;
            spr->data.model.floorClip   = 0;
            spr->data.model.mf          = mf;
            spr->data.model.nextMF      = nextmf;
            spr->data.model.inter       = inter;
            spr->data.model.viewAligned = true;
            spr->data.model.yawAngleOffset   = spr->psp->pos[0] * weaponOffsetScale - 90;
            spr->data.model.pitchAngleOffset =
                (32 - spr->psp->pos[1]) * weaponOffsetScale * weaponOffsetScaleY / 1000.0f;
            if (weaponFOVShift > 0 && weaponFixedFOV > 90) {
                spr->data.model.pitchAngleOffset -= weaponFOVShift * (weaponFixedFOV - 90) / 90;
            }
            spr->data.model.yaw =
                viewData->current.angle() / (dfloat) ANGLE_MAX *-360 + spr->data.model.yawAngleOffset + 90;
            spr->data.model.pitch = viewData->current.pitch * 85 / 110 + spr->data.model.yawAngleOffset;
            std::memset(spr->data.model.visOff, 0, sizeof(spr->data.model.visOff));
        }
        else {
            spr->type = VPSPR_SPRITE;
        }
    }
}
static Matrix4f frameViewerMatrix;
static void setupViewMatrix() {
    auto &rend = ClientApp::renderSystem();
    rend.uProjectionMatrix() = Rend_GetProjectionMatrix();
    rend.uViewMatrix()       = Rend_GetModelViewMatrix(DoomsdayApp::players().indexOf(viewPlayer));
    frameViewerMatrix        = rend.uProjectionMatrix().toMatrix4f() * rend.uViewMatrix().toMatrix4f();
}
Matrix4f const &Viewer_Matrix() {
    return frameViewerMatrix;
}
enum ViewState { Default2D, PlayerView3D, PlayerSprite2D };
static void changeViewState(ViewState viewState) //, viewport_t const *port, viewdata_t const *viewData)
{
    DENG2_ASSERT_IN_RENDER_THREAD();
    DENG_ASSERT_GL_CONTEXT_ACTIVE();
    switch (viewState) {
    case PlayerView3D:
        DGL_CullFace(DGL_BACK);
        DGL_Enable(DGL_DEPTH_TEST);
        GL_ProjectionMatrix(); // The 3D projection matrix.
        break;
    case PlayerSprite2D: {
        auto const conRect  = R_ConsoleRect(displayPlayer);
        auto const viewRect = R_Console3DViewRect(displayPlayer);
        dint const height = dint(SCREENHEIGHT
                * ( float(conRect.width()) * float(viewRect.height())
                                           / float(viewRect.width()) )
                / float(conRect.height()));
        scalemode_t sm = R_ChooseScaleMode(SCREENWIDTH, SCREENHEIGHT,
                                           conRect.width(), conRect.height(),
                                           scalemode_t(weaponScaleMode));
        DGL_MatrixMode(DGL_PROJECTION);
        DGL_LoadIdentity();
        if(sm == SCALEMODE_STRETCH) {
            DGL_Ortho(0, 0, SCREENWIDTH, height, -1, 1);
        }
        else {
            DGL_Ortho(0, 0, conRect.width(), conRect.height(), -1, 1);
            DGL_Translatef(conRect.width()/2, conRect.height(), 0);
            if(conRect.width() >= conRect.height()) {
                DGL_Scalef(dfloat( conRect.height() ) / SCREENHEIGHT,
                                   dfloat( conRect.height() ) / SCREENHEIGHT, 1);
            }
            else {
                DGL_Scalef(dfloat( conRect.width() ) / SCREENWIDTH,
                                   dfloat( conRect.width() ) / SCREENWIDTH, 1);
            }
            if(conRect.height() > conRect.width()) {
                dfloat extraScale = (dfloat(conRect.height() * 2) / conRect.width()) / 2;
                DGL_Scalef(extraScale, extraScale, 1);
            }
            DGL_Translatef(-(SCREENWIDTH / 2), -SCREENHEIGHT, 0);
            DGL_Scalef(1, dfloat( SCREENHEIGHT ) / height, 1);
        }
        DGL_MatrixMode(DGL_MODELVIEW);
        DGL_LoadIdentity();
        DGL_Disable(DGL_DEPTH_TEST);
        break;
    }
    case Default2D:
        DGL_CullFace(DGL_NONE);
        DGL_Disable(DGL_DEPTH_TEST);
        break;
    }
    /*auto const viewRect = R_Console3DViewRect(displayPlayer);
    viewpx = 0;
    viewpy = 0;
    viewpw = int(viewRect.width());
    viewph = int(viewRect.height());*/
    /*ClientWindow::main().game().glApplyViewport(Rectanglei::fromSize(Vector2i(viewpx, viewpy),
                                                                     Vector2ui(viewpw, viewph)));*/
}
#undef R_RenderPlayerView
DENG_EXTERN_C void R_RenderPlayerView(dint num) {
    if (num < 0 || num >= DDMAXPLAYERS) return; // Huh?
    player_t *player = DD_Player(num);
    if (!player->publicData().inGame) return;
    if (!player->publicData().mo) return;
    if (!ClientApp::world().hasMap()) return;
    if (firstFrameAfterLoad) {
        firstFrameAfterLoad = false;
        DD_ResetTimer();
    }
    viewdata_t *vd = &player->viewport();
    if (vd->window.isNull()) return;
    R_SetupFrame(player);
    vrCfg().setEyeHeightInMapUnits(Con_GetInteger("player-eyeheight"));
    setupViewMatrix();
    setupPlayerSprites();
    if (ClientApp::vr().mode() == VRConfig::OculusRift &&
        ClientApp::world().map().isPointInVoid(Rend_EyeOrigin().xzy())) {
        GLState::current().target().clear(GLFramebuffer::Color);
        return;
    }
    dint oldFlags = 0;
    if (!(player->publicData().flags & DDPF_CHASECAM)) {
        oldFlags = player->publicData().mo->ddFlags;
        player->publicData().mo->ddFlags |= DDMF_DONTDRAW;
    }
#if defined (DENG_OPENGL)
    if (renderWireframe) {
        LIBGUI_GL.glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    }
#endif
    DGL_MatrixMode(DGL_PROJECTION);
    DGL_PushMatrix();
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    changeViewState(PlayerView3D);
    Rend_RenderMap(ClientApp::world().map());
    changeViewState(PlayerSprite2D);
#if defined (DENG_OPENGL)
    if (renderWireframe) {
        LIBGUI_GL.glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    }
#endif
    Rend_Draw2DPlayerSprites();  // If the 2D versions are needed.
#if defined (DENG_OPENGL)
    if (renderWireframe) {
        LIBGUI_GL.glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    }
#endif
    if (psp3d) {
        changeViewState(PlayerView3D);
        Rend_Draw3DPlayerSprites();
    }
    changeViewState(Default2D);
    DGL_MatrixMode(DGL_PROJECTION);
    DGL_PopMatrix();
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
    DGL_Flush();
#if defined (DENG_OPENGL)
    if (renderWireframe) {
        LIBGUI_GL.glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    }
#endif
    if (!(player->publicData().flags & DDPF_CHASECAM)) {
        player->publicData().mo->ddFlags = oldFlags;
    }
    R_PrintRendPoolInfo();
}
/**
 * Should be called when returning from a game-side drawing method to ensure
 * that our assumptions of the GL state are valid. This is necessary because
 * DGL affords the user the posibility of modifiying the GL state.
 *
 * @todo: A cleaner approach would be a DGL state stack which could simply pop.
 */
static void restoreDefaultGLState() {
    DGL_Disable(DGL_FOG);
    DGL_Disable(DGL_SCISSOR_TEST);
    DGL_Disable(DGL_TEXTURE_2D);
    DGL_Enable(DGL_LINE_SMOOTH);
    DGL_Enable(DGL_POINT_SMOOTH);
}
void R_RenderViewPort(int playerNum) {
    int localNum = P_ConsoleToLocal(playerNum);
    if (localNum < 0) return;
    viewport_t const *vp = &viewportOfLocalPlayer[localNum];
    DENG2_ASSERT(vp->console == playerNum);
    dint const oldDisplay = displayPlayer;
    displayPlayer = vp->console;
    if(displayPlayer < 0 || (DD_Player(displayPlayer)->publicData().flags & DDPF_UNDEFINED_ORIGIN)) {
        R_RenderBlankView();
        displayPlayer = oldDisplay;
        return;
    }
    DGL_MatrixMode(DGL_PROJECTION);
    DGL_PushMatrix();
    DGL_LoadIdentity();
    Rectanglei viewRect = R_Console3DViewRect(playerNum);
    DGL_Ortho(0, 0, viewRect.width(), viewRect.height(), -1, 1);
    viewdata_t const *vd = &DD_Player(vp->console)->viewport();
    RectRaw vpGeometry = {{vp->geometry.topLeft.x, vp->geometry.topLeft.y},
                          {int(vp->geometry.width()), int(vp->geometry.height())}};
    RectRaw vdWindow = {{vd->window.topLeft.x, vd->window.topLeft.y},
                        {int(vd->window.width()), int(vd->window.height())}};
    R_UpdateViewer(vp->console);
    gx.DrawViewPort(localNum, &vpGeometry, &vdWindow, displayPlayer, /* layer: */ 0);
    LensFx_Draw(vp->console);
    restoreDefaultGLState();
    DGL_MatrixMode(DGL_PROJECTION);
    DGL_PopMatrix();
    frameCount++;
    if(resetNextViewer > 1) resetNextViewer = 0;
    displayPlayer = oldDisplay;
}
void R_ClearViewData() {
    frameLuminous.clear();
}
/**
 * Viewer specific override controlling whether a given sky layer is enabled.
 *
 * @todo The override should be applied at SkyDrawable level. We have Raven to
 * thank for this nonsense (Hexen's sector special 200)... -ds
 */
#undef R_SkyParams
DENG_EXTERN_C void R_SkyParams(dint layerIndex, dint param, void * /*data*/)
{
    LOG_AS("R_SkyParams");
    if (!ClientApp::world().hasMap()) {
        LOG_GL_WARNING("No map currently loaded, ignoring");
        return;
    }
    Sky &sky = ClientApp::world().map().sky();
    if (layerIndex >= 0 && layerIndex < sky.layerCount()) {
        SkyLayer &layer = sky.layer(layerIndex);
        switch (param) {
        case DD_ENABLE:  layer.enable();  break;
        case DD_DISABLE: layer.disable(); break;
        default: // Log but otherwise ignore this error.
            LOG_GL_WARNING("Failed configuring layer #%i: bad parameter %i")
                << layerIndex << param;
        }
        return;
    }
    LOG_GL_WARNING("Invalid layer #%i") << + layerIndex;
}
bool R_ViewerSubspaceIsVisible(ConvexSubspace const &subspace) {
    DENG2_ASSERT(subspace.indexInMap() != MapElement::NoIndex);
    return subspacesVisible.testBit(subspace.indexInMap());
}
void R_ViewerSubspaceMarkVisible(ConvexSubspace const &subspace, bool yes) {
    DENG2_ASSERT(subspace.indexInMap() != MapElement::NoIndex);
    subspacesVisible.setBit(subspace.indexInMap(), yes);
}
bool R_ViewerGeneratorIsVisible(Generator const &generator) {
    return generatorsVisible.testBit(generator.id() - 1 /* id is 1-based index */);
}
void R_ViewerGeneratorMarkVisible(Generator const &generator, bool yes) {
    generatorsVisible.setBit(generator.id() - 1 /* id is 1-based index */, yes);
}
ddouble R_ViewerLumobjDistance(dint idx) {
    if(idx >= 0 && idx < ClientApp::world().map().lumobjCount()) {
        return frameLuminous.at(idx).distance;
    }
    return 0;
}
bool R_ViewerLumobjIsClipped(dint idx) {
    if (idx >= 0 && idx < frameLuminous.size()) {
        return CPP_BOOL(frameLuminous.at(idx).isClipped);
    }
    return false;
}
bool R_ViewerLumobjIsHidden(dint idx) {
    if (idx >= 0 && idx < frameLuminous.size()) {
        return frameLuminous.at(idx).isClipped == 2;
    }
    return false;
}
static void markLumobjClipped(Lumobj const &lob, bool yes = true) {
    dint const index = lob.indexInMap();
    DENG_ASSERT(index >= 0 && index < lob.map().lumobjCount());
    DENG_ASSERT(index < frameLuminous.size());
    frameLuminous[index].isClipped = yes? 1 : 0;
}
void R_BeginFrame() {
    static QVector<int> frameLuminousOrder;
    Map &map = ClientApp::world().map();
    subspacesVisible.resize(map.subspaceCount());
    subspacesVisible.fill(false);
    generatorsVisible.fill(false);
    int numLuminous = map.lumobjCount();
    if (!numLuminous) return;
    if (frameLuminous.size() < numLuminous) {
        frameLuminous.resize(numLuminous);
        frameLuminousOrder.resize(numLuminous);
    }
    viewdata_t const *viewData = &viewPlayer->viewport();
    map.forAllLumobjs([&viewData] (Lumobj &lob) {
        Vector3d delta = lob.origin() - viewData->current.origin;
        frameLuminous[lob.indexInMap()].distance = M_ApproxDistance3(delta.x, delta.y, delta.z * 1.2 /*correct aspect*/);
        return LoopContinue;
    });
    if (rendMaxLumobjs > 0 && numLuminous > rendMaxLumobjs) {
        for (int i = 0; i < numLuminous; ++i) {
            frameLuminousOrder[i] = i;
            frameLuminous[i].isClipped = 2;
        }
        qSort(frameLuminousOrder.begin(),
              frameLuminousOrder.begin() + numLuminous,
              [] (int a, int b) {
            return frameLuminous[a].distance < frameLuminous[b].distance;
        });
        for (int i = 0, n = 0; i < numLuminous; ++i) {
            if (n++ > rendMaxLumobjs)
                break;
            frameLuminous[frameLuminousOrder[i]].isClipped = 1;
        }
    }
    else {
        for (auto &lum : frameLuminous) {
            lum.isClipped = 1;
        }
    }
}
void R_ViewerClipLumobj(Lumobj *lum) {
    if (!lum) return;
    dint lumIdx = lum->indexInMap();
    if (frameLuminous.at(lumIdx).isClipped > 1) return;
    markLumobjClipped(*lum, false);
    Vector3d const origin(lum->x(), lum->y(), lum->z() + lum->zOffset());
    if (!P_IsInVoid(DD_Player(displayPlayer)) && !devNoCulling) {
        if (!ClientApp::renderSystem().angleClipper().isPointVisible(origin)) {
            markLumobjClipped(*lum); // Won't have a halo.
        }
    }
    else {
        markLumobjClipped(*lum);
        Vector3d const eye = Rend_EyeOrigin().xzy();
        if (LineSightTest(eye, origin, -1, 1, LS_PASSLEFT | LS_PASSOVER | LS_PASSUNDER)
                .trace(lum->map().bspTree())) {
            markLumobjClipped(*lum, false); // Will have a halo.
        }
    }
}
void R_ViewerClipLumobjBySight(Lumobj *lob, ConvexSubspace *subspace) {
    if(!lob || !subspace) return;
    if (frameLuminous.at(lob->indexInMap()).isClipped)
        return;
    Vector3d const eye = Rend_EyeOrigin().xzy();
    subspace->forAllPolyobjs([&lob, &eye] (Polyobj &pob) {
        for(HEdge *hedge : pob.mesh().hedges()) {
            if(!hedge->hasMapElement())
                continue;
            if(hedge->mapElementAs<LineSideSegment>().isFrontFacing()) {
                coord_t eyeV1[2]       = { eye.x, eye.y };
                coord_t lumOriginV1[2] = { lob->origin().x, lob->origin().y };
                coord_t fromV1[2]      = { hedge->origin().x, hedge->origin().y };
                coord_t toV1[2]        = { hedge->twin().origin().x, hedge->twin().origin().y };
                if(V2d_Intercept2(lumOriginV1, eyeV1, fromV1, toV1, 0, 0, 0)) {
                    markLumobjClipped(*lob);
                    break;
                }
            }
        }
        return LoopContinue;
    });
}
angle_t viewer_t::angle() const {
    angle_t a = _angle;
    if(DD_GetInteger(DD_USING_HEAD_TRACKING)) {
        a += fixed_t(radianToDegree(vrCfg().oculusRift().headOrientation().z) / 180 * ANGLE_180);
    }
    return a;
}
D_CMD(ViewGrid) {
    DENG2_UNUSED2(src, argc);
    return R_SetViewGrid(String(argv[1]).toInt(), String(argv[2]).toInt());
}
void Viewports_Register() {
    C_VAR_INT ("con-show-during-setup",     &loadInStartupMode,     0, 0, 1);
    C_VAR_INT ("rend-camera-smooth",        &rendCameraSmooth,      CVF_HIDE, 0, 1);
    C_VAR_BYTE("rend-info-deltas-angles",   &showViewAngleDeltas,   0, 0, 1);
    C_VAR_BYTE("rend-info-deltas-pos",      &showViewPosDeltas,     0, 0, 1);
    C_VAR_BYTE("rend-info-frametime",       &showFrameTimePos,      0, 0, 1);
    C_VAR_BYTE("rend-info-rendpolys",       &rendInfoRPolys,        CVF_NO_ARCHIVE, 0, 1);
    C_CMD("viewgrid", "ii", ViewGrid);
}
vissprite_t visSprites[MAXVISSPRITES], *visSpriteP;
vispsprite_t visPSprites[DDMAXPSPRITES];
vissprite_t visSprSortedHead;
static vissprite_t overflowVisSprite;
void R_ClearVisSprites() {
    visSpriteP = visSprites;
}
vissprite_t *R_NewVisSprite(visspritetype_t type) {
    vissprite_t *spr;
    if (visSpriteP == &visSprites[MAXVISSPRITES]) {
        spr = &overflowVisSprite;
    }
    else {
        visSpriteP++;
        spr = visSpriteP - 1;
    }
    *spr = {};
    spr->type = type;
    return spr;
}
void VisSprite_SetupSprite(vissprite_t *spr, VisEntityPose const &pose, VisEntityLighting const &light,
    dfloat /*secFloor*/, dfloat /*secCeil*/, dfloat /*floorClip*/, dfloat /*top*/,
    ClientMaterial &material, bool matFlipS, bool matFlipT, blendmode_t blendMode,
    dint tClass, dint tMap, world::BspLeaf *bspLeafAtOrigin,
    bool /*floorAdjust*/, bool /*fitTop*/, bool /*fitBottom*/)
{
    drawspriteparams_t &p = *VS_SPRITE(spr);
    MaterialVariantSpec const &spec = Rend_SpriteMaterialSpec(tClass, tMap);
    MaterialAnimator *matAnimator   = &material.getAnimator(spec);
    DENG2_ASSERT((tClass == 0 && tMap == 0) ||
                 (spec.primarySpec->variant.flags & TSF_HAS_COLORPALETTE_XLAT));
    spr->pose = pose;
    p.bspLeaf     = bspLeafAtOrigin;
    p.noZWrite    = noSpriteZWrite;
    p.matAnimator = matAnimator;
    p.matFlip[0]  = matFlipS;
    p.matFlip[1]  = matFlipT;
    p.blendMode   = (useSpriteBlend? blendMode : BM_NORMAL);
    spr->light = light;
    spr->light.ambientColor[3] = (useSpriteAlpha? light.ambientColor.w : 1);
}
void VisSprite_SetupModel(vissprite_t *spr, VisEntityPose const &pose, VisEntityLighting const &light,
    FrameModelDef *mf, FrameModelDef *nextMF, dfloat inter,
    dint id, dint selector, world::BspLeaf * /*bspLeafAtOrigin*/, dint mobjDDFlags, dint tmap,
    bool /*fullBright*/, bool alwaysInterpolate)
{
    drawmodelparams_t &p = *VS_MODEL(spr);
    p.mf                    = mf;
    p.nextMF                = nextMF;
    p.inter                 = inter;
    p.alwaysInterpolate     = alwaysInterpolate;
    p.id                    = id;
    p.selector              = selector;
    p.flags                 = mobjDDFlags;
    p.tmap                  = tmap;
    spr->pose               = pose;
    spr->light              = light;
    p.shineYawOffset        = 0;
    p.shinePitchOffset      = 0;
    p.shineTranslateWithViewerPos = p.shinepspriteCoordSpace = false;
}
void R_SortVisSprites() {
    if(!visSpriteP) return;
    dint const count = visSpriteP - visSprites;
    if(count <= 0) return;
    vissprite_t unsorted;
    unsorted.next = unsorted.prev = &unsorted;
    for(vissprite_t *ds = visSprites; ds < visSpriteP; ds++) {
        ds->next = ds + 1;
        ds->prev = ds - 1;
    }
    visSprites[0].prev = &unsorted;
    unsorted.next = &visSprites[0];
    (visSpriteP - 1)->next = &unsorted;
    unsorted.prev = visSpriteP - 1;
    visSprSortedHead.next = visSprSortedHead.prev = &visSprSortedHead;
    vissprite_t *best = nullptr;
    for(dint i = 0; i < count; ++i) {
        ddouble bestdist = 0;
        for(vissprite_t *ds = unsorted.next; ds != &unsorted; ds = ds->next) {
            if(ds->pose.distance >= bestdist) {
                bestdist = ds->pose.distance;
                best = ds;
            }
        }
        best->next->prev = best->prev;
        best->prev->next = best->next;
        best->next = &visSprSortedHead;
        best->prev = visSprSortedHead.prev;
        visSprSortedHead.prev->next = best;
        visSprSortedHead.prev = best;
    }
}
void VisEntityLighting::setupLighting(Vector3d const &origin, ddouble distance,
                                      world::BspLeaf const &bspLeaf) {
#if 0
    world::Map &map = ClientApp::world().map();
    if(useBias && map.hasLightGrid()) {
        Vector4f color = map.lightGrid().evaluate(origin);
        for(dint i = 0; i < 3; ++i) {
            color[i] += Rend_LightAdaptationDelta(color[i]);
        }
        ambientColor.x = color.x;
        ambientColor.y = color.y;
        ambientColor.z = color.z;
    }
    else
#endif
    {
        auto const &subsec   = bspLeaf.subspace().subsector().as<world::ClientSubsector>();
        Vector4f const color = subsec.lightSourceColorfIntensity();
        dfloat lightLevel = color.w;
        lightLevel += Rend_ExtraLightDelta();
        Rend_ApplyLightAdaptation(lightLevel);
        for(dint i = 0; i < 3; ++i) {
            ambientColor[i] = lightLevel * color[i];
        }
    }
    Rend_ApplyTorchLight(ambientColor, distance);
    vLightListIdx =
            Rend_CollectAffectingLights(origin, ambientColor, bspLeaf.subspacePtr(),
                                        true /*stark world light*/);
}
namespace VR {
    float weaponDistance = 10.f; // global
}
static int   vrMode             = VRConfig::Mono;
static int   vrRiftFBSamples;
static float vrHudDistance;
static float vrPlayerHeight;
static float vrIpd;
static byte  vrSwapEyes;
static float vrDominantEye;
VRConfig &vrCfg() {
    DENG2_ASSERT(DENG2_BASE_GUI_APP != 0);
    return DENG2_BASE_GUI_APP->vr();
}
/*
float VR_RiftFovX() {
    return vrRiftFovX;
}
*/
static void vrConfigVariableChanged() {
    vrCfg().setDominantEye(vrDominantEye);
    vrCfg().setScreenDistance(vrHudDistance);
    vrCfg().setInterpupillaryDistance(vrIpd);
    vrCfg().setPhysicalPlayerHeight(vrPlayerHeight);
    vrCfg().setRiftFramebufferSampleCount(vrRiftFBSamples);
    vrCfg().setSwapEyes(vrSwapEyes);
}
static void vrModeChanged() {
    /*
    if(vrMode == VRConfig::OculusRift && !vrCfg().oculusRift().isReady()) {
        vrMode = VRConfig::Mono;
        LOG_WARNING("Oculus Rift not connected, reverting to normal 3D mode");
    }*/
    vrCfg().setMode(VRConfig::StereoMode(vrMode));
    if (ClientWindow::mainExists()) {
        ClientWindow &win = ClientWindow::main();
        win.updateRootSize();
    }
    /*
    if(vrMode == VRConfig::OculusRift) {
        if(Con_GetFloat("rend-camera-fov") != vrRiftFovX)
            Con_SetFloat("rend-camera-fov", vrRiftFovX);
        vrCfg().oculusRift().setPredictionLatency(vrRiftLatency);
    }
    else {
        if(Con_GetFloat("rend-camera-fov") != vrNonRiftFovX)
            Con_SetFloat("rend-camera-fov", vrNonRiftFovX);
    }*/
}
/*
static void vrRiftFovXChanged() {
    if(vrCfg().mode() == VRConfig::OculusRift) {
        if(Con_GetFloat("rend-camera-fov") != vrRiftFovX)
            Con_SetFloat("rend-camera-fov", vrRiftFovX);
    }
}
static void vrNonRiftFovXChanged() {
    if(vrCfg().mode() != VRConfig::OculusRift) {
        if(Con_GetFloat("rend-camera-fov") != vrNonRiftFovX)
            Con_SetFloat("rend-camera-fov", vrNonRiftFovX);
    }
}
D_CMD(LoadRiftParams) {
    DENG2_UNUSED3(src, argc, argv);
    return VR_LoadRiftParameters();
}*/
D_CMD(ResetRiftPose) {
    DENG2_UNUSED3(src, argc, argv);
    vrCfg().oculusRift().resetTracking();
    LOG_INPUT_MSG("Reset Oculus Rift position tracking");
    return true;
}
void VR_ConsoleRegister() {
    vrDominantEye   = vrCfg().dominantEye();
    vrHudDistance   = vrCfg().screenDistance();
    vrIpd           = vrCfg().interpupillaryDistance();
    vrPlayerHeight  = vrCfg().physicalPlayerHeight();
    vrRiftFBSamples = vrCfg().riftFramebufferSampleCount();
    vrSwapEyes      = vrCfg().swapEyes();
    /**
     * @todo When old-style console variables become obsolete, VRConfig should expose
     * these settings as a Record that can be attached under Config.
     */
    C_VAR_INT2  ("rend-vr-mode",                 &vrMode,             0, 0, VRConfig::NUM_STEREO_MODES - 1, vrModeChanged);
    C_VAR_FLOAT2("rend-vr-dominant-eye",  &vrDominantEye,   0, -1.0f, 1.0f,   vrConfigVariableChanged);
    C_VAR_FLOAT2("rend-vr-hud-distance",  &vrHudDistance,   0, 0.01f, 40.0f,  vrConfigVariableChanged);
    C_VAR_FLOAT2("rend-vr-ipd",           &vrIpd,           0, 0.02f, 0.1f,   vrConfigVariableChanged);
    C_VAR_FLOAT2("rend-vr-player-height", &vrPlayerHeight,  0, 1.0f,  2.4f,   vrConfigVariableChanged);
    C_VAR_INT2  ("rend-vr-rift-samples",  &vrRiftFBSamples, 0, 1, 4,          vrConfigVariableChanged);
    C_VAR_BYTE2 ("rend-vr-swap-eyes",     &vrSwapEyes,      0, 0, 1,          vrConfigVariableChanged);
    C_CMD("resetriftpose", NULL, ResetRiftPose);
}
#if 0
bool VR_LoadRiftParameters() {
    de::OculusRift &ovr = vrCfg().oculusRift();
    if(ovr.isReady()) {
        Con_SetFloat("rend-vr-ipd", ovr.interpupillaryDistance()); // from Oculus SDK
        Con_SetFloat("rend-vr-rift-fovx", ovr.fovX());
        /*
        float fov = 180.0f / de::PI * 2.0f * (atan2(
                  info.EyeToScreenDistance,
                  0.5f * (info.HScreenSize - info.InterpupillaryDistance)));
                */
        return true;
    }
    return false;
}
#endif
/**
 * Determines whether normal smoothing should be performed for the given pair of
 * map surfaces (which are assumed to share an edge).
 *
 * Yes if the angle between the two surfaces is less than 45 degrees.
 * @todo Should be user customizable with a Material property. -ds
 *
 * @param sufA       The "left"  map surface which shares an edge with @a sufB.
 * @param sufB       The "right" map surface which shares an edge with @a sufA.
 * @param angleDiff  Angle difference (i.e., normal delta) between the two surfaces.
 */
static bool shouldSmoothNormals(Surface &sufA, Surface &sufB, binangle_t angleDiff) {
    DENG2_UNUSED2(sufA, sufB);
    return INRANGE_OF(angleDiff, BANG_180, BANG_45);
}
WallEdge::Event::Event()
    : IHPlane::IIntercept(0)
    , _owner(nullptr)
{}
WallEdge::Event::Event(WallEdge &owner, ddouble distance)
    : WorldEdge::Event()
    , IHPlane::IIntercept(distance)
    , _owner(&owner)
{}
WallEdge::Event &WallEdge::Event::operator = (Event const &other) {
    _owner    = other._owner;
    _distance = other._distance;
    return *this;
}
bool WallEdge::Event::operator < (Event const &other) const {
    return distance() < other.distance();
}
ddouble WallEdge::Event::distance() const {
    return IHPlane::IIntercept::distance();
}
Vector3d WallEdge::Event::origin() const {
    return _owner->pOrigin() + _owner->pDirection() * distance();
}
static inline coord_t lineSideOffset(LineSideSegment &seg, dint edge) {
    return seg.lineSideOffset() + (edge? seg.length() : 0);
}
QList<WallEdge::Impl *> WallEdge::recycledImpls;
struct WallEdge::Impl : public IHPlane {
    WallEdge *self = nullptr;
    WallSpec spec;
    dint edge = 0;
    HEdge *wallHEdge = nullptr;
    Partition hplane;
    Vector3d pOrigin;
    Vector3d pDirection;
    coord_t lo = 0, hi = 0;
    Event bottom;
    Event top;
    /**
     * Special-purpose array whose memory is never freed while the array exists.
     * `WallEdge::Impl`s are recycled, so it would be a waste of time to keep
     * allocating and freeing the event arrays.
     */
    struct EventArray : private QVector<Event> {
        using Base = QVector<Event>;
        int size = 0;
    public:
        EventArray() {}
        using Base::at;
        using Base::begin;
        using Base::iterator;
        using Base::const_iterator;
        inline bool isEmpty() const {
            return size == 0;
        }
        inline void clear() {
            size = 0;
        }
        void append(Event const &event) {
            if (size < Base::size()) {
                (*this)[size++] = event;
            }
            else {
                Base::append(event);
                ++size;
            }
        }
        inline Event &last() {
            return (*this)[size - 1];
        }
        inline Base::iterator end() {
            return Base::begin() + size;
        }
        inline Base::const_iterator end() const {
            return Base::begin() + size;
        }
    };
    EventArray events;
    bool needSortEvents = false;
    Vector2f materialOrigin;
    Vector3f normal;
    bool needUpdateNormal = true;
    Impl() {}
    void deinit() {
        self = nullptr;
        edge = 0;
        wallHEdge = nullptr;
        lo = hi = 0;
        events.clear();
        needSortEvents = false;
        needUpdateNormal = true;
    }
    void init(WallEdge *i, WallSpec const &wallSpec, HEdge &hedge, dint edge) {
        self = i;
        spec       = wallSpec;
        this->edge = edge;
        wallHEdge  = &hedge;
        bottom     = Event(*self, 0);
        top        = Event(*self, 1);
        LineSideSegment &seg   = lineSideSegment();
        Line const &line       = seg.line();
        bool const unpegBottom = (line.flags() & DDLF_DONTPEGBOTTOM) != 0;
        bool const unpegTop    = (line.flags() & DDLF_DONTPEGTOP)    != 0;
        ConvexSubspace const &space = (line.definesPolyobj() ? line.polyobj().bspLeaf().subspace()
                                                             : wallHEdge->face().mapElementAs<world::ConvexSubspace>());
        auto const &subsec = space.subsector().as<world::ClientSubsector>();
        /*
         * For reference, see "Texture aligment" in Doomwiki.org:
         * https://doomwiki.org/wiki/Texture_alignment
         */
        if (seg.lineSide().considerOneSided()
            || // Mapping errors may result in a line segment missing a back face.
               (!line.definesPolyobj() && !wallHEdge->twin().hasFace())) {
            if (spec.section == LineSide::Middle) {
                lo = subsec.visFloor().heightSmoothed();
                hi = subsec.visCeiling().heightSmoothed();
            }
            else {
                lo = hi = subsec.visFloor().heightSmoothed();
            }
            materialOrigin = seg.lineSide().middle().originSmoothed();
            if (unpegBottom) {
                materialOrigin.y -= hi - lo;
            }
        }
        else {
            auto const &backSubsec =
                line.definesPolyobj() ? subsec
                                      : wallHEdge->twin().face().mapElementAs<world::ConvexSubspace>()
                                            .subsector().as<world::ClientSubsector>();
            Plane const *ffloor = &subsec.visFloor();
            Plane const *fceil  = &subsec.visCeiling();
            Plane const *bfloor = &backSubsec.visFloor();
            Plane const *bceil  = &backSubsec.visCeiling();
            switch (spec.section) {
            case LineSide::Top:
                if (!line.isSelfReferencing()) {
                    if (bceil->heightSmoothed() < ffloor->heightSmoothed())
                        lo = ffloor->heightSmoothed();
                    else
                        lo = bceil->heightSmoothed();
                    hi = fceil->heightSmoothed();
                    if (spec.flags.testFlag(WallSpec::SkyClip)
                        && fceil->surface().hasSkyMaskedMaterial()
                        && bceil->surface().hasSkyMaskedMaterial()) {
                        hi = lo;
                    }
                    materialOrigin = seg.lineSide().middle().originSmoothed();
                    if (!unpegTop) {
                        materialOrigin.y -= fceil->heightSmoothed() - bceil->heightSmoothed();
                    }
                }
                break;
            case LineSide::Bottom:
                if (!line.isSelfReferencing()) {
                    const double bceilZ  = bceil->heightSmoothed();
                    const double bfloorZ = bfloor->heightSmoothed();
                    const double fceilZ  = fceil->heightSmoothed();
                    const double ffloorZ = ffloor->heightSmoothed();
                    const bool raiseToBackFloor =
                        (   fceil->surface().hasSkyMaskedMaterial()
                         && bceil->surface().hasSkyMaskedMaterial()
                         && fceilZ < bceilZ
                         && bfloorZ > fceilZ);
                    coord_t t = bfloorZ;
                    lo = ffloorZ;
                    if (bfloorZ > bceilZ)
                        t = bceilZ;
                    if (t > fceilZ && !raiseToBackFloor)
                        t = fceilZ;
                    hi = t;
                    if (spec.flags.testFlag(WallSpec::SkyClip)
                        && ffloor->surface().hasSkyMaskedMaterial()
                        && bfloor->surface().hasSkyMaskedMaterial()) {
                        lo = hi;
                    }
                    materialOrigin = seg.lineSide().bottom().originSmoothed();
                    if (bfloor->heightSmoothed() > fceil->heightSmoothed()) {
                        materialOrigin.y -= (raiseToBackFloor? t : fceilZ)
                                          - bfloorZ;
                    }
                    if (unpegBottom) {
                        materialOrigin.y += (raiseToBackFloor ? t : de::max(fceilZ, bceilZ)) - bfloorZ;
                    }
                }
                break;
            case LineSide::Middle: {
                LineSide const &lineSide = seg.lineSide();
                Surface const &middle    = lineSide.middle();
                const bool isExtendedMasked = middle.hasMaterial() &&
                                              !middle.materialAnimator()->isOpaque() &&
                                              !lineSide.top().hasMaterial() &&
                                              lineSide.sector().ceiling().surface().material().isSkyMasked();
                if (!line.isSelfReferencing() && ffloor == &subsec.sector().floor()) {
                    lo = de::max(bfloor->heightSmoothed(), ffloor->heightSmoothed());
                }
                else {
                    lo = lineSide.sector().floor().heightSmoothed();
                }
                if (!line.isSelfReferencing() && fceil == &subsec.sector().ceiling()) {
                    hi = de::min(bceil->heightSmoothed(), fceil->heightSmoothed());
                }
                else {
                    hi = lineSide.back().sector().ceiling().heightSmoothed();
                }
                materialOrigin = Vector2f(middle.originSmoothed().x, 0);
                if (middle.hasMaterial()
                    && !seg.lineSide().isFlagged(SDF_MIDDLE_STRETCH)) {
                    coord_t openBottom, openTop;
                    if (!line.isSelfReferencing()) {
                        openBottom = lo;
                        openTop    = hi;
                    }
                    else {
                        openBottom = ffloor->heightSmoothed();
                        openTop    = fceil->heightSmoothed();
                    }
                    if (openTop > openBottom) {
                        if (unpegBottom) {
                            lo += middle.originSmoothed().y;
                            hi = lo + middle.material().height();
                        }
                        else {
                            hi += middle.originSmoothed().y;
                            lo = hi - middle.material().height();
                        }
                        if (hi > openTop) {
                            materialOrigin.y = hi - openTop;
                        }
                        bool const clipBottom = !(!(devRendSkyMode || P_IsInVoid(viewPlayer)) && ffloor->surface().hasSkyMaskedMaterial() && bfloor->surface().hasSkyMaskedMaterial());
                        bool const clipTop    = !(!(devRendSkyMode || P_IsInVoid(viewPlayer)) && fceil->surface().hasSkyMaskedMaterial()  && bceil->surface().hasSkyMaskedMaterial());
                        if (clipTop || clipBottom) {
                            if (clipBottom && lo < openBottom)
                                lo = openBottom;
                            if (clipTop && hi > openTop)
                                hi = openTop;
                        }
                        if (!clipTop) {
                            materialOrigin.y = 0;
                        }
                    }
                }
                if (isExtendedMasked) {
                    if (hi - lo < middle.material().height() + middle.originSmoothed().y) {
                        hi = lo + middle.material().height() + middle.originSmoothed().y;
                    }
                }
                break; }
            }
        }
        materialOrigin += Vector2f(::lineSideOffset(seg, edge), 0);
        pOrigin    = Vector3d(self->origin(), lo);
        pDirection = Vector3d(0, 0, hi - lo);
    }
    inline LineSideSegment &lineSideSegment() {
        return wallHEdge->mapElementAs<LineSideSegment>();
    }
    void verifyValid() const {
        if(!self->isValid()) {
            throw InvalidError("WallEdge::verifyValid", "Range geometry is not valid (top < bottom)");
        }
    }
    EventIndex toEventIndex(ddouble distance) const {
        for (EventIndex i = 0; i < events.size; ++i) {
            if (de::fequal(events.at(i).distance(), distance))
                return i;
        }
        return InvalidIndex;
    }
    inline bool haveEvent(ddouble distance) const {
        return toEventIndex(distance) != InvalidIndex;
    }
    Event &createEvent(ddouble distance) {
        return *intercept(distance);
    }
    void configure(Partition const &newPartition) {
        hplane = newPartition;
    }
    Partition const &partition() const {
        return hplane;
    }
    Event *intercept(ddouble distance) {
        events.append(Event(*self, distance));
        needSortEvents = true;
        return &events.last();
    }
    void sortAndMergeIntercepts() {
        if (needSortEvents) {
            qSort(events.begin(), events.end(), [] (WorldEdge::Event const &a, WorldEdge::Event const &b) {
                return a < b;
            });
            needSortEvents = false;
        }
    }
    void clearIntercepts() {
        events.clear();
        needSortEvents = false;
    }
    Event const &at(EventIndex index) const {
        if(index >= 0 && index < interceptCount()) {
            return events.at(index);
        }
        throw UnknownInterceptError("WallEdge::at", String("Index '%1' does not map to a known intercept (count: %2)")
                                                        .arg(index).arg(interceptCount()));
    }
    dint interceptCount() const {
        return events.size;
    }
#ifdef DENG2_DEBUG
    void printIntercepts() const {
        EventIndex index = 0;
        foreach (Event const &icpt, events) {
            LOGDEV_MAP_MSG(" %u: >%1.2f ") << (index++) << icpt.distance();
        }
    }
#endif
    /**
     * Ensure all intercepts do not exceed the specified closed range.
     */
    void assertInterceptsInRange(ddouble low, ddouble hi) const {
#ifdef DENG2_DEBUG
        foreach (Event const &icpt, events) {
            DENG2_ASSERT(icpt.distance() >= low && icpt.distance() <= hi);
        }
#else
        DENG2_UNUSED2(low, hi);
#endif
    }
    inline ddouble distanceTo(coord_t worldHeight) const {
        return (worldHeight - lo) / (hi - lo);
    }
    void addNeighborIntercepts(ddouble bottom, ddouble top) {
        ClockDirection const direction = edge ? Clockwise : Anticlockwise;
        HEdge const *hedge = wallHEdge;
        while ((hedge = &SubsectorCirculator::findBackNeighbor(*hedge, direction)) != wallHEdge) {
            if (!hedge->hasFace() || !hedge->hasMapElement())
                break;
            auto const &backSpace = hedge->face().mapElementAs<ConvexSubspace>();
            auto const &subsec    = backSpace.subsector().as<world::ClientSubsector>();
            if (subsec.hasWorldVolume()) {
                for (dint i = 0; i < subsec.visPlaneCount(); ++i) {
                    Plane const &plane = subsec.visPlane(i);
                    if (plane.heightSmoothed() > bottom && plane.heightSmoothed() < top) {
                        ddouble distance = distanceTo(plane.heightSmoothed());
                        if (!haveEvent(distance)) {
                            createEvent(distance);
                            if (interceptCount() == WALLEDGE_MAX_INTERCEPTS)
                                return;
                        }
                    }
                    if (plane.isSectorFloor() && plane.heightSmoothed() > bottom)
                        bottom = plane.heightSmoothed();
                    else if (plane.isSectorCeiling() && plane.heightSmoothed() < top)
                        top = plane.heightSmoothed();
                    if (bottom >= top)
                        return;
                }
            }
            else {
                ddouble z = subsec.visCeiling().heightSmoothed();
                if(z > bottom && z < top) {
                    ddouble distance = distanceTo(z);
                    if(!haveEvent(distance)) {
                        createEvent(distance);
                        return; // All clipped away.
                    }
                }
            }
        }
    }
    /**
     * Determines whether the wall edge should be intercepted with neighboring
     * planes from other subsectors.
     */
    bool shouldInterceptNeighbors() {
        if(spec.flags & WallSpec::NoEdgeDivisions)
            return false;
        if(de::fequal(hi, lo))
            return false;
        if(Subsector::isInternalEdge(wallHEdge))
            return false;
        return true;
    }
    void prepareEvents() {
        DENG2_ASSERT(self->isValid());
        needSortEvents = false;
        events.append(bottom);
        events.append(top);
        if(shouldInterceptNeighbors()) {
            configure(Partition(Vector2d(0, hi - lo)));
            addNeighborIntercepts(lo, hi);
            sortAndMergeIntercepts();
        }
#ifdef DENG2_DEBUG
        assertInterceptsInRange(0, 1);
#endif
    }
    /**
     * Find the neighbor surface for the edge which we will use to calculate the
     * "blend" properties (e.g., smoothed edge normal).
     *
     * @todo: Use the half-edge rings instead of LineOwners.
     */
    Surface *findBlendNeighbor(binangle_t &diff) {
        diff = 0;
        if(spec.flags.testFlag(WallSpec::NoEdgeNormalSmoothing))
            return nullptr;
        LineSide const &lineSide = lineSideSegment().lineSide();
        if(lineSide.line().definesPolyobj())
            return nullptr;
        ClockDirection const direction = (edge? Anticlockwise : Clockwise);
        LineOwner const &farVertOwner  = *lineSide.line().vertexOwner(lineSide.sideId() ^ edge);
        Line *neighbor;
        if(R_SideBackClosed(lineSide)) {
            neighbor = R_FindSolidLineNeighbor(lineSide.line(), farVertOwner, direction,
                                               lineSide.sectorPtr(), &diff);
        }
        else {
            neighbor = R_FindLineNeighbor(lineSide.line(), farVertOwner, direction,
                                          lineSide.sectorPtr(), &diff);
        }
        if(!neighbor) return nullptr;
        LineSide *otherSide;
        if(&neighbor->vertex(edge ^ 1) == &lineSide.vertex(edge))
            otherSide = &neighbor->front();
        else
            otherSide = &neighbor->back();
        if(!otherSide->hasSections()) return nullptr;
        return &otherSide->middle();
    }
    /**
     * Determine the (possibly smoothed) edge normal.
     * @todo Cache the smoothed normal value somewhere...
     */
    void updateNormal() {
        needUpdateNormal = false;
        LineSide &lineSide = lineSideSegment().lineSide();
        Surface &surface   = lineSide.surface(spec.section);
        binangle_t angleDiff;
        Surface *blendSurface = findBlendNeighbor(angleDiff);
        if(blendSurface && shouldSmoothNormals(surface, *blendSurface, angleDiff)) {
            normal = Vector3f(surface.normal() + blendSurface->normal()) / 2;
        }
        else {
            normal = surface.normal();
        }
    }
};
WallEdge::WallEdge(WallSpec const &spec, HEdge &hedge, int edge)
    : WorldEdge((edge? hedge.twin() : hedge).origin())
    , d(getRecycledImpl()) {
    d->init(this, spec, hedge, edge);
}
WallEdge::~WallEdge() {
    recycleImpl(d);
}
Vector3d const &WallEdge::pOrigin() const {
    return d->pOrigin;
}
Vector3d const &WallEdge::pDirection() const {
    return d->pDirection;
}
Vector2f WallEdge::materialOrigin() const {
    return d->materialOrigin;
}
Vector3f WallEdge::normal() const {
    if(d->needUpdateNormal) {
        d->updateNormal();
    }
    return d->normal;
}
WallSpec const &WallEdge::spec() const {
    return d->spec;
}
LineSideSegment &WallEdge::lineSideSegment() const {
    return d->lineSideSegment();
}
coord_t WallEdge::lineSideOffset() const {
    return ::lineSideOffset(d->lineSideSegment(), d->edge);
}
dint WallEdge::divisionCount() const {
    if (!isValid()) return 0;
    if (d->events.isEmpty()) {
        d->prepareEvents();
    }
    return d->interceptCount() - 2;
}
WallEdge::EventIndex WallEdge::firstDivision() const {
    return divisionCount()? 1 : InvalidIndex;
}
WallEdge::EventIndex WallEdge::lastDivision() const {
    return divisionCount()? (d->interceptCount() - 2) : InvalidIndex;
}
WallEdge::Event const &WallEdge::at(EventIndex index) const {
    return d->events.at(index);
}
bool WallEdge::isValid() const {
    return d->hi > d->lo;
}
WallEdge::Event const &WallEdge::first() const {
    return d->bottom;
}
WallEdge::Event const &WallEdge::last() const {
    return d->top;
}
WallEdge::Impl *WallEdge::getRecycledImpl() // static
{
    if (recycledImpls.isEmpty()) {
        return new Impl;
    }
    return recycledImpls.takeLast();
}
void WallEdge::recycleImpl(Impl *d) // static
{
    d->deinit();
    recycledImpls.append(d);
}
/**
 * Should angle based light level deltas be applied?
 */
static bool useWallSectionLightLevelDeltas(LineSide const &side, dint section) {
    if (rendLightWallAngle <= 0) return false;
    if (side.surface(section).hasFixMaterial() &&
        side.hasSector() && side.back().hasSector()) {
        Sector &backSector = side.back().sector();
        if (backSector.floor().height() < backSector.ceiling().height())
            return false;
    }
    return true;
}
WallSpec WallSpec::fromMapSide(LineSide const &side, dint section) // static
{
    bool const isTwoSidedMiddle = (section == LineSide::Middle && !side.considerOneSided());
    WallSpec spec(section);
    if (side.line().definesPolyobj() || isTwoSidedMiddle) {
        spec.flags &= ~WallSpec::ForceOpaque;
        spec.flags |= WallSpec::NoEdgeDivisions;
    }
    if (auto *midMan = side.middle().materialAnimator()) {
        if (section != LineSide::Middle && !midMan->isOpaque()) {
            spec.flags &= ~WallSpec::ForceOpaque;
        }
    }
    if (isTwoSidedMiddle) {
        if (viewPlayer &&
            ((viewPlayer->publicData().flags & (DDPF_NOCLIP | DDPF_CAMERA)) ||
             !side.line().isFlagged(DDLF_BLOCKING)))
            spec.flags |= WallSpec::NearFade;
        spec.flags |= WallSpec::SortDynLights;
    }
    if (devRendSkyMode)
        spec.flags &= ~WallSpec::SkyClip;
    if (side.line().definesPolyobj())
        spec.flags |= WallSpec::NoFakeRadio;
    bool useLightLevelDeltas = useWallSectionLightLevelDeltas(side, section);
    if (!useLightLevelDeltas)
        spec.flags |= WallSpec::NoLightDeltas;
    if (!useLightLevelDeltas || !rendLightWallAngleSmooth)
        spec.flags |= WallSpec::NoEdgeNormalSmoothing;
    return spec;
}
