﻿/** @file atlas.cpp  Image-based atlas.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Atlas"
#include <de/LogBuffer>
#include <QSet>
#include <QRect>
#include <QImage>
#include <QPainter>
/** @file atlastexture.h  Atlas stored on a GLTexture.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/AtlasTexture"
#include "de/RowAtlasAllocator"
#include "de/KdTreeAtlasAllocator"
/** @file colorbank.cpp  Bank of colors.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/ColorBank"
#include <de/math.h>
#include <de/ArrayValue>
/** @file drawable.cpp  Drawable object with buffers, programs and states.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Drawable"
#include <QMap>
/** @file glatlasbuffer.cpp  GLBuffer from where GLSubBuffers are allocated.
 *
 * @authors Copyright (c) 2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/GLAtlasBuffer"
#include "de/GLSubBuffer"
#include <QList>
/** @file glbuffer.cpp  GL vertex buffer.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/GLBuffer"
#include "de/GLState"
#include "de/GLFramebuffer"
#include "de/GLProgram"
#include "de/GLInfo"
/** @file gldrawqueue.cpp  Utility for managing and drawing semi-static GL buffers.
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/GLDrawQueue"
#include "de/GLInfo"
#include "de/GLProgram"
#include "de/GLShader"
#include "de/GLState"
#include "de/GLSubBuffer"
#include "de/GLUniform"
/** @file gltarget.cpp  GL render target.
 *
 * Implementation does not use QGLFrameBufferObject because it does not allow
 * attaching manually created textures.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/GLFramebuffer"
#include "de/GLTexture"
#include "de/GLState"
#include "de/GLInfo"
#include "de/GLWindow"
#include <de/Asset>
#include <de/LogBuffer>
/** @file gltarget_alternativebuffer.cpp  Alternative buffer attachment for GLFramebuffer.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#if 0
#include "de/GLFramebuffer"
#include "de/GLTexture"
#include "de/GLInfo"
namespace de {
DENG2_PIMPL_NOREF(GLFramebuffer::AlternativeBuffer)
{
    GLFramebuffer *target;
    GLTexture *texture;
    GLFramebuffer::Flags attachment;
    GLTexture *original;
    GLuint originalRendBuf;
    Impl()
        : target(nullptr)
        , texture(nullptr)
        , attachment(GLFramebuffer::NoAttachments)
        , original(nullptr)
        , originalRendBuf(0)
    {}
};
GLFramebuffer::AlternativeBuffer::AlternativeBuffer(GLFramebuffer &target, GLTexture &texture, Flags const &attachment)
    : d(new Impl)
{
    d->target = &target;
    d->texture = &texture;
    d->attachment = attachment;
}
GLFramebuffer::AlternativeBuffer::AlternativeBuffer(GLFramebuffer &target, Flags const &attachment)
    : d(new Impl)
{
    d->target = &target;
    d->attachment = attachment;
}
GLFramebuffer::AlternativeBuffer::~AlternativeBuffer()
{
    deinit();
}
bool GLFramebuffer::AlternativeBuffer::init()
{
    if (d->attachment != GLFramebuffer::DepthStencil)
    {
        DENG2_ASSERT(!"GLFramebuffer::AlternativeBuffer only supports DepthStencil attachments");
        return false;
    }
    if (d->original || d->originalRendBuf)
    {
        return false;
    }
    if (d->texture)
    {
        d->original = d->target->attachedTexture(d->attachment);
        DENG2_ASSERT(d->original != 0);
        if (d->texture->size() != d->target->size())
        {
            d->texture->setDepthStencilContent(d->target->size());
        }
        d->target->replaceAttachment(d->attachment, *d->texture);
    }
    else
    {
        d->originalRendBuf = d->target->attachedRenderBuffer(d->attachment);
        if (d->originalRendBuf == 0)
        {
            d->original = d->target->attachedTexture(d->attachment);
            if (!d->original) return false; // Not supported.
        }
        d->target->replaceWithNewRenderBuffer(d->attachment);
    }
    return true;
}
bool GLFramebuffer::AlternativeBuffer::deinit()
{
    if (!d->original && !d->originalRendBuf) return false; // Not inited.
    if (!d->texture)
    {
        d->target->releaseAttachment(d->attachment);
    }
    if (d->original)
    {
        d->target->replaceAttachment(d->attachment, *d->original);
    }
    else if (d->originalRendBuf)
    {
        d->target->replaceAttachment(d->attachment, d->originalRendBuf);
    }
    return true;
}
GLFramebuffer &GLFramebuffer::AlternativeBuffer::target() const
{
    return *d->target;
}
} // namespace de
#endif
/** @file glinfo.cpp  OpenGL information.
 *
 * @authors Copyright (c) 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright (c) 2007-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/GLInfo"
#include "de/GLWindow"
#include "de/graphics/opengl.h"
#include <cstring>
#include <de/String>
#include <de/Log>
#include <de/math.h>
#include <de/c_wrapper.h>
#ifdef DENG2_DEBUG
#  define DENG_ENABLE_OPENGL_DEBUG_LOGGER
#  include <QOpenGLDebugLogger>
#endif
#if defined (MACOSX)
#  include <OpenGL/OpenGL.h>
#endif
#if defined (DENG_X11)
#  include <QX11Info>
#  include <GL/glx.h>
#  include <GL/glxext.h>
#  undef None
#  undef Always
#  undef Status
#endif
/** @file glprogram.cpp  GL shader program.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/GLProgram"
#include "de/GLInfo"
#include "de/GLUniform"
#include "de/GLBuffer"
#include "de/GLShader"
#include "de/GLTexture"
#include "de/GuiApp"
#include <de/Block>
#include <de/Log>
#include <QList>
#include <QSet>
#include <QHash>
#include <QStack>
/** @file glshader.cpp  GL shader.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/GLShader"
#include "de/GLInfo"
#include "de/GuiApp"
#include "de/graphics/opengl.h"
#include <de/Block>
#include <de/String>
/** @file glshaderbank.cpp  Bank containing GL shaders.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/GLShaderBank"
#include "de/GLProgram"
#include "de/GLShader"
#include "de/GLUniform"
#include <de/App>
#include <de/ArrayValue>
#include <de/ByteArrayFile>
#include <de/DictionaryValue>
#include <de/FS>
#include <de/Folder>
#include <de/ScriptedInfo>
#include <de/math.h>
#include <QRegularExpression>
#include <QMap>
/** @file glstate.cpp  GL state.
 *
 * @todo This implementation assumes OpenGL drawing occurs only in one thread.
 * If multithreaded rendering is done at some point in the future, the GL state
 * stack must be part of the thread-local data.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/GLState"
#include "de/GLFramebuffer"
#include "de/GLWindow"
#include "de/graphics/opengl.h"
#include <de/GLInfo>
#include <de/BitField>
#include <de/Log>
/** @file glsubbuffer.cpp  Sub-range of a larger GLBuffer.
 *
 * @authors Copyright (c) 2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/GLSubBuffer"
/** @file gltexture.cpp  GL texture atlas.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/GLTexture"
#include "de/GLInfo"
#include "de/graphics/opengl.h"
/** @file glframebuffer.cpp  GL frame buffer.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/GLTextureFramebuffer"
#include "de/GuiApp"
#include <de/Log>
#include <de/Drawable>
#include <de/GLInfo>
#include <de/Property>
/** @file gluniform.cpp  GL uniform.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/GLUniform"
#include "de/GLTexture"
#include "de/GLProgram"
#include "de/GLInfo"
#include <de/Block>
#include <de/Log>
#include <cstring>
/** @file heightmap.h  Height map.
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/HeightMap"
/** @file image.h  Wrapper over QImage.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Image"
#include "de/graphics/opengl.h"
#include <de/Block>
#include <de/File>
#include <de/FixedByteArray>
#include <de/Log>
#include <de/Reader>
#include <de/Vector>
#include <de/Writer>
#include <de/Zeroed>
#include <QDataStream>
#include <QPainter>
#include <QColor>
/** @file imagebank.h  Bank containing Image instances loaded from files.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/ImageBank"
#include "de/App"
#include "de/FileSystem"
#include "de/ImageFile"
#include <de/ScriptedInfo>
/** @file imagefile.cpp  Image file.
 *
 * @authors Copyright (c) 2015-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/ImageFile"
#include "de/HeightMap"
#include <de/App>
#include <de/Folder>
#include <de/LogBuffer>
#include <QHash>
/** @file kdtreeatlasallocator.cpp  KD-tree based atlas allocator.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright (c) 2009-2010 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/KdTreeAtlasAllocator"
#include <de/BinaryTree>
#include <QList>
/** @file modelbank.cpp  Bank containing 3D models.
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/ModelBank"
#include <de/App>
#include <de/Folder>
/** @file modeldrawable.cpp  Drawable specialized for 3D models.
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/ModelDrawable"
#include "de/HeightMap"
#include "de/ImageFile"
#include <de/Animation>
#include <de/App>
#include <de/ByteArrayFile>
#include <de/Folder>
#include <de/GLBuffer>
#include <de/GLProgram>
#include <de/GLState>
#include <de/GLUniform>
#include <de/Matrix>
#include <de/NativePath>
#include <de/TextureBank>
#include <assimp/IOStream.hpp>
#include <assimp/IOSystem.hpp>
#include <assimp/Importer.hpp>
#include <assimp/LogStream.hpp>
#include <assimp/DefaultLogger.hpp>
#include <assimp/scene.h>
#include <assimp/postprocess.h>
#include <array>
/** @file multiatlas.cpp
 *
 * @authors Copyright (c) 2015-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/MultiAtlas"
#include <de/Deletable>
#include <QHash>
#include <QSet>
/** @file packageiconbank.cpp  Bank for package icons.
 *
 * @authors Copyright (c) 2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/PackageIconBank"
#include <de/FileSystem>
#include <de/ImageFile>
/** @file rowatlasallocator.cpp  Row-based atlas allocator.
 *
 * The row allocator works according to the following principles:
 *
 * - In the beginning, there is a single row that spans the height of the entire atlas.
 *   The row contains a single empty segment.
 * - If a row is completely empty, the empty space below will be split into a new empty
 *   row when the first allocation is made on the line. The first allocation also
 *   determines the initial height of the row.
 * - The height of a row may expand if there is empty space below.
 * - All the empty spaces are kept ordered from narrow to wide, so that when a new
 *   allocation is needed, the smallest suitable space can be picked.
 * - Each row is a doubly-linked list containing the used and free regions.
 * - If there are two adjacent free regions on a row, they will be merged into a larger
 *   empty space. Similarly empty rows are merged together.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/RowAtlasAllocator"
#include <QList>
#include <set>
/** @file texturebank.cpp
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/TextureBank"
#include <QHash>

namespace de {
DENG2_PIMPL(Atlas) {
    Flags flags;
    Size totalSize;
    int margin;
    int border;
    std::unique_ptr<IAllocator> allocator;
    QHash<Id::Type, Image *> deferred; ///< Only used with DeferredAllocations.
    Image backing;
    bool needCommit;
    bool needFullCommit;
    bool mayDefrag;
    QList<Rectanglei> changedAreas;
    Time fullReportedAt;
    Impl(Public *i, Flags const &flg, Size const &size)
        : Base(i)
        , flags(flg)
        , totalSize(size.max(Size(1, 1)))
        , margin(1)
        , border(0)
        , needCommit(false)
        , needFullCommit(true)
        , mayDefrag(false) {
        if (hasBacking()) {
            backing = QImage(QSize(totalSize.x, totalSize.y), QImage::Format_ARGB32);
        }
    }
    ~Impl() {
        allocator.reset();
        self().cancelDeferred();
    }
    inline bool usingDeferredMode() const {
        return flags.testFlag(DeferredAllocations);
    }
    inline bool hasBacking() const {
        return flags.testFlag(BackingStore);
    }
    void markAsChanged(Rectanglei const &changedRect) {
            needCommit = true;
        changedAreas.append(changedRect);
    }
    void markFullyChanged() {
        needCommit = true;
        needFullCommit = true;
        changedAreas.clear();
        changedAreas.append(backing.rect());
    }
    bool mustCommitFull() const {
        /*
         * Simple heuristic: if more than 95% of the pixels are included in the
         * changed area, simply copy the whole thing.
         */
        return (needFullCommit || changedPercentage() > .95f);
    }
    float changedPercentage() const {
        if (!needCommit || totalSize == Size(0, 0)) return 0.0f;
        duint const totalPx = totalSize.x * totalSize.y;
        duint changedPx = 0;
        for (Rectanglei const &rect : changedAreas) {
            changedPx += rect.width() * rect.height();
        }
        return float(changedPx) / float(totalPx);
    }
    float usedPercentage() const {
        if (!allocator) return 0;
        duint totalPx = totalSize.x * totalSize.y;
        duint usedPx = 0;
        foreach (Rectanglei const &alloc, allocator->allocs().values()) {
            usedPx += alloc.width() * alloc.height();
        }
        return float(usedPx) / float(totalPx);
    }
    /**
     * Submits the image to the backing store, or commits it if no backing
     * store is available.
     *
     * @param image  Image.
     * @param rect   Rectangle for the image determined by an IAllocator.
     */
    void submitImage(Image const &image, Rectanglei const &rect) {
        Rectanglei const noBorders  = rect.shrunk(border);
        Rectanglei const withMargin = rect.expanded(margin);
        if (hasBacking()) {
            backing.fill(withMargin, Image::Color(0, 0, 0, 0));
            if (border > 0) {
                if (flags.testFlag(WrapBordersInBackingStore)) {
                    backing.drawPartial(image, Rectanglei(0, 0, border, image.height()),
                                           rect.topRight() + Vector2i(-border, border));
                    backing.drawPartial(image, Rectanglei(image.width() - border, 0,
                                                             border, image.height()),
                                           rect.topLeft + Vector2i(0, border));
                    backing.drawPartial(image, Rectanglei(0, 0, image.width(), border),
                                           rect.bottomLeft() + Vector2i(border, -border));
                    backing.drawPartial(image, Rectanglei(0, image.height() - border,
                                                             image.width(), border),
                                           rect.topLeft + Vector2i(border, 0));
                }
            }
            backing.draw(image, noBorders.topLeft);
            markAsChanged(rect);
        }
        else {
            if (border > 0) {
                QImage const srcImg = image.toQImage();
                int const sw = srcImg.width();
                int const sh = srcImg.height();
                QImage bordered(QSize(rect.width(), rect.height()), srcImg.format());
                int const w = bordered.width();
                int const h = bordered.height();
                QPainter painter(&bordered);
                painter.setCompositionMode(QPainter::CompositionMode_Source);
                painter.fillRect(bordered.rect(), QColor(0, 0, 0, 0));
                painter.drawImage(border, border, srcImg);
                painter.drawImage(border, 0,     srcImg, 0, 0, sw, 1); // top
                painter.drawImage(border, h - 1, srcImg, 0, sh - 1, sw, 1); // bottom
                painter.drawImage(0, border,     srcImg, 0, 0, 1, sh); // left
                painter.drawImage(w - 1, border, srcImg, sw - 1, 0, 1, sh); // right
                painter.drawImage(0, 0,         srcImg, 0, 0, 1, 1);
                painter.drawImage(w - 1, 0,     srcImg, sw - 1, 0, 1, 1);
                painter.drawImage(0, h - 1,     srcImg, 0, sh - 1, 1, 1);
                painter.drawImage(w - 1, h - 1, srcImg, sw - 1, sh - 1, 1, 1);
                self().commit(bordered, rect.topLeft);
            }
            else {
                self().commit(image, noBorders.topLeft);
            }
        }
    }
    void submitDeferred() {
        for (auto i = deferred.begin(); i != deferred.end(); ++i) {
            try {
                Rectanglei rect;
                allocator->rect(i.key(), rect);
                submitImage(*i.value(), rect);
            }
            catch (Error const &er) {
                LOG_GL_ERROR("Allocation %s could not be submitted: %s")
                        << i.key() << er.asText();
            }
            delete i.value(); // we own the Image
        }
        deferred.clear();
    }
    /**
     * Compose a new backing store with an optimal layout.
     */
    void defragment() {
        DENG2_ASSERT(hasBacking());
        IAllocator::Allocations const oldLayout = allocator->allocs();
        if (!allocator->optimize()) {
            mayDefrag = false;
            return;
        }
        Image defragged(QImage(QSize(backing.size().x, backing.size().y), backing.qtFormat()));
        defragged.fill(Image::Color(0, 0, 0, 0));
        IAllocator::Allocations optimal = allocator->allocs();
        DENG2_FOR_EACH(IAllocator::Allocations, i, optimal) {
            defragged.draw(backing.subImage(oldLayout[i.key()]),
                           i.value().topLeft);
        }
        backing = defragged;
        markFullyChanged();
        mayDefrag = false;
        DENG2_FOR_PUBLIC_AUDIENCE2(Reposition, i) {
            i->atlasContentRepositioned(self());
        }
    }
    Image::Size sizeWithBorders(Image::Size const &size) {
        return size + Image::Size(2 * border, 2 * border);
    }
    Rectanglei rectWithoutBorder(Id const &id) const {
        Rectanglei rect;
        allocator->rect(id, rect);
        return rect.shrunk(border);
    }
    DENG2_PIMPL_AUDIENCE(Reposition)
    DENG2_PIMPL_AUDIENCE(OutOfSpace)
};
DENG2_AUDIENCE_METHOD(Atlas, Reposition)
DENG2_AUDIENCE_METHOD(Atlas, OutOfSpace)
Atlas::Atlas(Flags const &flags, Size const &totalSize)
    : d(new Impl(this, flags, totalSize))
{}
Atlas::Flags Atlas::flags() const {
    return d->flags;
}
void Atlas::setAllocator(IAllocator *allocator) {
    DENG2_GUARD(this);
    clear();
    d->allocator.reset(allocator);
    if (d->allocator.get()) {
        d->allocator->setMetrics(d->totalSize, d->margin);
        d->allocator->clear(); // using new metrics
    }
    d->markFullyChanged();
}
Atlas::IAllocator *Atlas::takeAllocator() {
    return d->allocator.release();
}
void Atlas::setMarginSize(dint marginPixels) {
    d->margin = marginPixels;
    if (d->allocator.get()) {
        d->allocator->setMetrics(d->totalSize, d->margin);
    }
}
void Atlas::setBorderSize(dint borderPixels) {
    d->border = borderPixels;
}
void Atlas::clear() {
    DENG2_GUARD(this);
    if (d->allocator) {
        d->allocator->clear();
    }
    if (d->hasBacking()) {
        d->backing.fill(Image::Color(0, 0, 0, 0));
        d->markFullyChanged();
    }
    d->mayDefrag = false;
}
void Atlas::setTotalSize(Size const &totalSize) {
    DENG2_GUARD(this);
    d->totalSize = totalSize;
    if (d->allocator) {
        d->allocator->setMetrics(totalSize, d->margin);
    }
    if (d->hasBacking()) {
        d->backing.resize(totalSize);
        d->markFullyChanged();
        d->defragment();
    }
}
Atlas::Size Atlas::totalSize() const {
    DENG2_GUARD(this);
    return d->totalSize;
}
Id Atlas::alloc(Image const &image, Id const &chosenId) {
    if (image.isNull()) {
        LOG_AS("Atlas");
        LOGDEV_GL_WARNING("Cannot allocate a zero-size image");
        return Id::None;
    }
    DENG2_GUARD(this);
    DENG2_ASSERT(d->allocator.get());
    Rectanglei rect;
    Id id = d->allocator->allocate(d->sizeWithBorders(image.size()), rect, chosenId);
    if (id.isNone() && d->flags.testFlag(AllowDefragment) && d->mayDefrag) {
        d->defragment();
        id = d->allocator->allocate(d->sizeWithBorders(image.size()), rect, chosenId);
    }
    if (!id.isNone()) {
        d->mayDefrag = true;
        if (!d->usingDeferredMode()) {
            d->submitImage(image, rect);
        }
        else {
            d->deferred.insert(id, new Image(image));
        }
    }
    else if (!d->usingDeferredMode()) {
        LOG_AS("Atlas");
        if (!d->fullReportedAt.isValid() || d->fullReportedAt.since() > 1.0) {
            LOGDEV_GL_XVERBOSE("Full with %.1f%% usage", d->usedPercentage() * 100);
            d->fullReportedAt = Time::currentHighPerformanceTime();
        }
        DENG2_FOR_AUDIENCE2(OutOfSpace, i) {
            i->atlasOutOfSpace(*this);
        }
    }
    return id;
}
void Atlas::release(Id const &id) {
    if (id.isNone()) return;
    DENG2_GUARD(this);
    DENG2_ASSERT(d->allocator.get());
    d->allocator->release(id);
    d->mayDefrag = true;
}
bool Atlas::contains(Id const &id) const {
    DENG2_GUARD(this);
    if (d->allocator.get()) {
        return d->allocator->ids().contains(id);
    }
    return false;
}
int Atlas::imageCount() const {
    DENG2_GUARD(this);
    DENG2_ASSERT(d->allocator.get());
    return d->allocator->count();
}
Atlas::Ids Atlas::allImages() const {
    DENG2_GUARD(this);
    DENG2_ASSERT(d->allocator.get());
    return d->allocator->ids();
}
Rectanglei Atlas::imageRect(Id const &id) const {
    DENG2_GUARD(this);
    DENG2_ASSERT(d->allocator.get());
    return d->rectWithoutBorder(id);
}
Rectanglef Atlas::imageRectf(Id const &id) const {
    DENG2_GUARD(this);
    DENG2_ASSERT(d->allocator.get());
    Rectanglei const rect = d->rectWithoutBorder(id);
    return Rectanglef(float(rect.topLeft.x) / float(d->totalSize.x),
                      float(rect.topLeft.y) / float(d->totalSize.y),
                      float(rect.width()  ) / float(d->totalSize.x),
                      float(rect.height() ) / float(d->totalSize.y));
}
Image Atlas::image(Id const &id) const {
    DENG2_GUARD(this);
    if (d->deferred.contains(id)) {
        return *d->deferred[id];
    }
    if (d->hasBacking() && d->allocator.get() && contains(id)) {
        return d->backing.subImage(imageRect(id));
    }
    return Image();
}
void Atlas::commit() const {
    DENG2_GUARD(this);
    LOG_AS("Atlas");
    d->submitDeferred();
    if (!d->needCommit || !d->hasBacking()) return;
    if (d->mustCommitFull()) {
        DENG2_ASSERT(d->backing.size() == d->totalSize);
        commitFull(d->backing);
    }
    else {
        for (Rectanglei const &rect : d->changedAreas) {
            commit(d->backing, rect);
        }
    }
    d->changedAreas.clear();
    d->needCommit = false;
    d->needFullCommit = false;
}
void Atlas::cancelDeferred() {
    for (auto i = d->deferred.constBegin(); i != d->deferred.constEnd(); ++i) {
        delete i.value(); // we own a copy of the Image
        if (d->allocator) {
            release(i.key());
        }
    }
    d->deferred.clear();
}
} // namespace de
namespace de {
AtlasTexture::AtlasTexture(Atlas::Flags const &flags, Atlas::Size const &totalSize)
    : Atlas(flags, totalSize) {
    setState(Ready);
}
AtlasTexture *AtlasTexture::newWithRowAllocator(Atlas::Flags const &flags, Atlas::Size const &totalSize) {
    AtlasTexture *atlas = new AtlasTexture(flags, totalSize);
    atlas->setAllocator(new RowAtlasAllocator);
    return atlas;
}
AtlasTexture *AtlasTexture::newWithKdTreeAllocator(Atlas::Flags const &flags, Atlas::Size const &totalSize) {
    AtlasTexture *atlas = new AtlasTexture(flags, totalSize);
    atlas->setAllocator(new KdTreeAtlasAllocator);
    return atlas;
}
void AtlasTexture::clear() {
    Atlas::clear();
    GLTexture::clear();
    setState(Ready);
}
void AtlasTexture::aboutToUse() const {
    Atlas::commit();
}
void AtlasTexture::commitFull(Image const &fullImage) const {
    DENG2_ASSERT(fullImage.size() == totalSize());
    const_cast<AtlasTexture *>(this)->setImage(fullImage);
}
void AtlasTexture::commit(Image const &image, Vector2i const &topLeft) const {
    GLTexture *tex = const_cast<AtlasTexture *>(this);
    if (size() == GLTexture::Size(0, 0)) {
        tex->setUndefinedImage(totalSize(), Image::RGBA_8888);
    }
    tex->setSubImage(image, topLeft);
}
void AtlasTexture::commit(Image const &fullImage, Rectanglei const &subregion) const {
    GLTexture *tex = const_cast<AtlasTexture *>(this);
    if (size() == GLTexture::Size(0, 0)) {
        tex->setUndefinedImage(totalSize(), Image::RGBA_8888);
    }
    tex->setSubImage(fullImage, subregion);
}
} // namespace de
namespace de {
DENG2_PIMPL(ColorBank) {
    struct ColorSource : public ISource {
        ColorBank &bank;
        String id;
        ColorSource(ColorBank &b, String const &colorId) : bank(b), id(colorId) {}
        Time modifiedAt() const { return bank.sourceModifiedAt(); }
        Vector4d load() const {
            Record const &def = bank[id];
            ArrayValue const *colorDef = 0;
            if (def.has("rgb")) {
                colorDef = &def.geta("rgb");
            }
            else {
                colorDef = &def.geta("rgba");
            }
            ddouble alpha = 1.0;
            if (colorDef->size() >= 4) {
                alpha = colorDef->at(3).asNumber();
            }
            return Vector4d(colorDef->at(0).asNumber(),
                            colorDef->at(1).asNumber(),
                            colorDef->at(2).asNumber(),
                            alpha);
        }
    };
    struct ColorData : public IData {
        Vector4d color;
        ColorData(Vector4d const &c = Vector4d()) : color(c) {}
    };
    Impl(Public *i) : Base(i)
    {}
};
ColorBank::ColorBank() : InfoBank("ColorBank", DisableHotStorage), d(new Impl(this))
{}
void ColorBank::addFromInfo(File const &file) {
    LOG_AS("ColorBank");
    parse(file);
    addFromInfoBlocks("color");
}
ColorBank::Color ColorBank::color(DotPath const &path) const {
    if (path.isEmpty()) return Color();
    Colorf col = colorf(path);
    return Color(round<dbyte>(col.x * 255),
                 round<dbyte>(col.y * 255),
                 round<dbyte>(col.z * 255),
                 round<dbyte>(col.w * 255));
}
ColorBank::Colorf ColorBank::colorf(DotPath const &path) const {
    if (path.isEmpty()) return Colorf();
    Vector4d clamped = data(path).as<Impl::ColorData>().color;
    clamped = clamped.max(Vector4d(0, 0, 0, 0)).min(Vector4d(1, 1, 1, 1));
    return Colorf(float(clamped.x), float(clamped.y), float(clamped.z), float(clamped.w));
}
Bank::ISource *ColorBank::newSourceFromInfo(String const &id) {
    return new Impl::ColorSource(*this, id);
}
Bank::IData *ColorBank::loadFromSource(ISource &source) {
    return new Impl::ColorData(source.as<Impl::ColorSource>().load());
}
} // namespace de
namespace de {
DENG2_PIMPL(Drawable) {
    typedef QMap<Id, GLBuffer *>  Buffers;
    typedef QMap<Id, GLProgram *> Programs;
    typedef QMap<Id, GLState *>   States;
    typedef QMap<String, Id>      Names;
    struct BufferConfig {
        GLProgram const *program;
        GLState const *state;
        BufferConfig(GLProgram const *p = 0, GLState const *s = 0)
            : program(p), state(s) {}
    };
    typedef QMap<Id, BufferConfig> BufferConfigs;
    Buffers       buffers;
    Programs      programs;
    States        states;
    Names         bufferNames;
    Names         programNames;
    Names         stateNames;
    BufferConfigs configs;
    GLProgram     defaultProgram;
    Impl(Public *i) : Base(i) {
        self() += defaultProgram;
    }
    ~Impl() {
        clear();
    }
    void clear() {
        qDeleteAll(buffers.values());
        qDeleteAll(programs.values());
        qDeleteAll(states.values());
        defaultProgram.clear();
        buffers.clear();
        programs.clear();
        states.clear();
        configs.clear();
        bufferNames.clear();
        programNames.clear();
        stateNames.clear();
    }
    Id nextBufferId() const {
        Id next = 1;
        if (!buffers.isEmpty()) next = buffers.keys().back() + 1;
        return next;
    }
    Id nextProgramId() const {
        Id next = 1;
        if (!programs.isEmpty()) next = programs.keys().back() + 1;
        return next;
    }
    Id nextStateId() const {
        Id next = 1;
        if (!states.isEmpty()) next = states.keys().back() + 1;
        return next;
    }
    void replaceProgram(GLProgram const *src, GLProgram const *dest) {
        DENG2_FOR_EACH(BufferConfigs, i, configs) {
            if (i.value().program == src) {
                i.value().program = dest;
            }
        }
    }
    void replaceState(GLState const *src, GLState const *dest) {
        DENG2_FOR_EACH(BufferConfigs, i, configs) {
            if (i.value().state == src) {
                i.value().state = dest;
            }
        }
    }
    void removeName(Names &names, Id id) {
        QMutableMapIterator<String, Id> iter(names);
        while (iter.hasNext()) {
            iter.next();
            if (iter.value() == id) {
                iter.remove();
            }
        }
    }
};
Drawable::Drawable() : d(new Impl(this))
{}
void Drawable::clear() {
    d->clear();
}
Drawable::Ids Drawable::allBuffers() const {
    return d->buffers.keys();
}
Drawable::Ids Drawable::allPrograms() const {
    Ids ids;
    ids << 0 // default program is always there
        << d->programs.keys();
    return ids;
}
Drawable::Ids Drawable::allStates() const {
    return d->states.keys();
}
bool Drawable::hasBuffer(Id id) const {
    return d->buffers.contains(id);
}
GLBuffer &Drawable::buffer(Id id) const {
    DENG2_ASSERT(d->buffers.contains(id));
    return *d->buffers[id];
}
GLBuffer &Drawable::buffer(Name const &bufferName) const {
    return buffer(bufferId(bufferName));
}
Drawable::Id Drawable::bufferId(Name const &bufferName) const {
    DENG2_ASSERT(d->bufferNames.contains(bufferName));
    return d->bufferNames[bufferName];
}
GLProgram &Drawable::program(Id id) const {
    if (!id) return d->defaultProgram;
    DENG2_ASSERT(d->programs.contains(id));
    return *d->programs[id];
}
GLProgram &Drawable::program(Name const &programName) const {
    return program(programId(programName));
}
Drawable::Id Drawable::programId(Name const &programName) const {
    if (programName.isEmpty()) return 0; // Default program.
    DENG2_ASSERT(d->programNames.contains(programName));
    return d->programNames[programName];
}
GLProgram const &Drawable::programForBuffer(Id bufferId) const {
    DENG2_ASSERT(d->configs.contains(bufferId));
    DENG2_ASSERT(d->configs[bufferId].program != 0);
    return *d->configs[bufferId].program;
}
GLProgram const &Drawable::programForBuffer(Name const &bufferName) const {
    return programForBuffer(bufferId(bufferName));
}
GLState const *Drawable::stateForBuffer(Id bufferId) const {
    return d->configs[bufferId].state;
}
GLState const *Drawable::stateForBuffer(Name const &bufferName) const {
    return stateForBuffer(bufferId(bufferName));
}
GLState &Drawable::state(Id id) const {
    DENG2_ASSERT(d->states.contains(id));
    return *d->states[id];
}
GLState &Drawable::state(Name const &stateName) const {
    return state(stateId(stateName));
}
Drawable::Id Drawable::stateId(Name const &stateName) const {
    DENG2_ASSERT(d->stateNames.contains(stateName));
    return d->stateNames[stateName];
}
void Drawable::addBuffer(Id id, GLBuffer *buffer) {
    removeBuffer(id);
    d->buffers[id] = buffer;
    setProgram(id, d->defaultProgram);
    insert(*buffer, Required);
}
Drawable::Id Drawable::addBuffer(Name const &bufferName, GLBuffer *buffer) {
    Id id = d->nextBufferId();
    d->bufferNames.insert(bufferName, id);
    addBuffer(id, buffer);
    return id;
}
Drawable::Id Drawable::addBuffer(GLBuffer *buffer) {
    Id const id = d->nextBufferId();
    addBuffer(id, buffer);
    return id;
}
Drawable::Id Drawable::addBufferWithNewProgram(GLBuffer *buffer, Name const &programName) {
    Id const bufId = d->nextBufferId();
    addBuffer(bufId, buffer);
    Id const progId = addProgram(programName);
    setProgram(bufId, program(progId));
    return bufId;
}
void Drawable::addBufferWithNewProgram(Id id, GLBuffer *buffer, Name const &programName) {
    addBuffer(id, buffer);
    addProgram(programName);
    setProgram(id, programName);
}
Drawable::Id Drawable::addBufferWithNewProgram(Name const &bufferName, GLBuffer *buffer,
                                               Name const &programName) {
    Id const progId = addProgram(programName);
    Id const bufId = addBuffer(bufferName, buffer);
    setProgram(bufId, program(progId));
    return bufId;
}
GLProgram &Drawable::addProgram(Id id) {
    DENG2_ASSERT(id != 0);
    removeProgram(id);
    GLProgram *p = new GLProgram;
    d->programs[id] = p;
    insert(*p, Required);
    return *p;
}
Drawable::Id Drawable::addProgram(Name const &programName) {
    Id const id = d->nextProgramId();
    addProgram(id);
    if (!programName.isEmpty()) {
        d->programNames.insert(programName, id);
    }
    return id;
}
GLState &Drawable::addState(Id id, GLState const &state) {
    removeState(id);
    GLState *s = new GLState(state);
    d->states[id] = s;
    return *s;
}
Drawable::Id Drawable::addState(Name const &stateName, GLState const &state) {
    Id const id = d->nextStateId();
    addState(id, state);
    d->stateNames.insert(stateName, id);
    return id;
}
void Drawable::removeBuffer(Id id) {
    if (d->buffers.contains(id)) {
        remove(*d->buffers[id]);
        delete d->buffers.take(id);
    }
    d->configs.remove(id);
}
void Drawable::removeProgram(Id id) {
    if (d->programs.contains(id)) {
        GLProgram *prog = d->programs[id];
        d->replaceProgram(prog, &d->defaultProgram);
        remove(*prog);
        delete d->programs.take(id);
    }
}
void Drawable::removeState(Id id) {
    if (d->programs.contains(id)) {
        GLState *st = d->states[id];
        d->replaceState(st, 0);
        delete d->states.take(id);
    }
}
void Drawable::removeBuffer(Name const &bufferName) {
    Id const id = bufferId(bufferName);
    removeBuffer(id);
    d->removeName(d->bufferNames, id);
}
void Drawable::removeProgram(Name const &programName) {
    Id const id = programId(programName);
    removeProgram(id);
    d->removeName(d->programNames, id);
}
void Drawable::removeState(Name const &stateName) {
    Id const id = stateId(stateName);
    removeState(id);
    d->removeName(d->stateNames, id);
}
void Drawable::setProgram(Id bufferId, GLProgram &program) {
    d->configs[bufferId].program = &program;
}
void Drawable::setProgram(Id bufferId, Name const &programName) {
    setProgram(bufferId, program(programName));
}
void Drawable::setProgram(Name const &bufferName, GLProgram &program) {
    setProgram(bufferId(bufferName), program);
}
void Drawable::setProgram(Name const &bufferName, Name const &programName) {
    setProgram(bufferId(bufferName), program(programName));
}
void Drawable::setProgram(GLProgram &program) {
    foreach (Id id, allBuffers()) {
        setProgram(id, program);
    }
}
void Drawable::setProgram(Id programId) {
    setProgram(program(programId));
}
void Drawable::setProgram(Name const &programName) {
    setProgram(program(programName));
}
void Drawable::setState(Id bufferId, GLState &state) {
    d->configs[bufferId].state = &state;
}
void Drawable::setState(Id bufferId, Name const &stateName) {
    setState(bufferId, state(stateName));
}
void Drawable::setState(Name const &bufferName, GLState &state) {
    setState(bufferId(bufferName), state);
}
void Drawable::setState(Name const &bufferName, Name const &stateName) {
    setState(bufferId(bufferName), state(stateName));
}
void Drawable::setState(GLState &state) {
    foreach (Id id, allBuffers()) {
        setState(id, state);
    }
}
void Drawable::setState(Name const &stateName) {
    setState(state(stateName));
}
void Drawable::unsetState(Id bufferId) {
    d->configs[bufferId].state = 0;
}
void Drawable::unsetState(Name const &bufferName) {
    unsetState(bufferId(bufferName));
}
void Drawable::unsetState() {
    foreach (Id id, allBuffers()) {
        unsetState(id);
    }
}
void Drawable::draw() const {
    if (!isReady()) return;
    GLProgram const *currentProgram = 0;
    GLState   const *currentState   = 0;
    GLState::current().apply();
    DENG2_FOR_EACH_CONST(Impl::Buffers, i, d->buffers) {
        Id const id = i.key();
        GLProgram const &bufProg = programForBuffer(id);
        if (currentProgram != &bufProg) {
            if (currentProgram) currentProgram->endUse();
            currentProgram = &bufProg;
            currentProgram->beginUse();
        }
        GLState const *bufState = stateForBuffer(id);
        if (bufState && currentState != bufState) {
            currentState = bufState;
            currentState->apply();
        }
        else if (!bufState && currentState != 0) {
            currentState = 0;
            GLState::current().apply();
        }
        i.value()->draw();
    }
    if (currentProgram) {
        currentProgram->endUse();
    }
    if (currentState) {
        GLState::current().apply();
    }
}
} // namespace de
namespace de {
static dsize const MIN_RANGE_SIZE = 16;
DENG2_PIMPL_NOREF(GLAtlasBuffer) {
    struct HostBuffer {
        GLBuffer buffer;
        QList<Rangeui16> availableRanges;
        void addAvailable(Rangeui16 const &range) {
            DENG2_ASSERT(range.size() >= MIN_RANGE_SIZE);
            availableRanges.append(range);
            qSort(availableRanges.begin(), availableRanges.end(),
                  [] (Rangeui16 const &a, Rangeui16 const &b) { return a.size() < b.size(); });
        }
        Rangeui16 findBestAvailable(dsize forSize) {
            for (auto i = availableRanges.begin(); i != availableRanges.end(); ++i) {
                if (i->size() >= forSize) {
                    Rangeui16 chosen = *i;
                    availableRanges.erase(i);
                    if (chosen.size() - forSize >= MIN_RANGE_SIZE) {
                        addAvailable(Rangeui16(chosen.start + forSize, chosen.end));
                    }
                    return chosen;
                }
            }
            return Rangeui16();
        }
    };
    QList<HostBuffer *> hostBuffers;
    gl::Usage usage = gl::Static;
    dsize elementSize = 0;
    duint16 maxElementCount = 0;
    internal::AttribSpecs format;
    ~Impl() {
        deinit();
    }
    HostBuffer &allocNewHostBuffer() {
        DENG2_ASSERT(maxElementCount > 0);
        DENG2_ASSERT(elementSize > 0);
        auto *host = new HostBuffer;
        host->availableRanges << Rangeui16(0, maxElementCount);
        host->buffer.setFormat(format);
        host->buffer.setUninitializedData(elementSize * maxElementCount, usage);
        hostBuffers.append(host);
        return *host;
    }
    void clear() {
        for (auto *host : hostBuffers) {
            host->availableRanges.clear();
            host->availableRanges << Rangeui16(0, maxElementCount);
        }
    }
    void deinit() {
        qDeleteAll(hostBuffers);
        hostBuffers.clear();
    }
    GLSubBuffer *allocSubBuffer(duint16 elementCount) {
        GLSubBuffer *sub = nullptr;
        for (auto *host : hostBuffers) {
            auto const bestRange = host->findBestAvailable(elementCount);
            if (!bestRange.isEmpty()) {
                sub = new GLSubBuffer(bestRange, host->buffer);
                break;
            }
        }
        if (!sub) {
            auto &empty = allocNewHostBuffer();
            sub = new GLSubBuffer(empty.findBestAvailable(elementCount), empty.buffer);
        }
        sub->setFormat(format);
        return sub;
    }
    void releaseSubBuffer(GLSubBuffer &) {
        DENG2_ASSERT(false);
    }
};
GLAtlasBuffer::GLAtlasBuffer(internal::AttribSpecs const &vertexFormat)
    : d(new Impl) {
    d->format = vertexFormat;
    d->elementSize = vertexFormat.first->stride;
}
void GLAtlasBuffer::setMaxElementCount(dsize maxElementCount) {
    DENG2_ASSERT(maxElementCount < 0x10000); // 16-bit indexing
    d->maxElementCount = duint16(maxElementCount);
}
void GLAtlasBuffer::setUsage(gl::Usage usage) {
    d->usage = usage;
}
void GLAtlasBuffer::clear() {
    d->clear();
}
GLSubBuffer *GLAtlasBuffer::alloc(dsize elementCount) {
    DENG2_ASSERT(elementCount > 0);
    DENG2_ASSERT(elementCount < 0x10000);
    return d->allocSubBuffer(duint16(elementCount));
}
void GLAtlasBuffer::release(GLSubBuffer &buf) {
    d->releaseSubBuffer(buf);
}
} // namespace de
namespace de {
#ifdef DENG2_DEBUG
extern int GLDrawQueue_queuedElems;
#endif
using namespace internal;
using namespace gl;
AttribSpec const Vertex2Tex::_spec[2] = {
    { AttribSpec::Position,  2, GL_FLOAT, false, sizeof(Vertex2Tex), 0 },
    { AttribSpec::TexCoord0, 2, GL_FLOAT, false, sizeof(Vertex2Tex), 2 * sizeof(float) }
};
LIBGUI_VERTEX_FORMAT_SPEC(Vertex2Tex, 4 * sizeof(float))
AttribSpec const Vertex2Rgba::_spec[2] = {
    { AttribSpec::Position, 2, GL_FLOAT, false, sizeof(Vertex2Rgba), 0 },
    { AttribSpec::Color,    4, GL_FLOAT, false, sizeof(Vertex2Rgba), 2 * sizeof(float) }
};
LIBGUI_VERTEX_FORMAT_SPEC(Vertex2Rgba, 6 * sizeof(float))
AttribSpec const Vertex2TexRgba::_spec[3] = {
    { AttribSpec::Position,  2, GL_FLOAT, false, sizeof(Vertex2TexRgba), 0 },
    { AttribSpec::TexCoord0, 2, GL_FLOAT, false, sizeof(Vertex2TexRgba), 2 * sizeof(float) },
    { AttribSpec::Color,     4, GL_FLOAT, false, sizeof(Vertex2TexRgba), 4 * sizeof(float) }
};
LIBGUI_VERTEX_FORMAT_SPEC(Vertex2TexRgba, 8 * sizeof(float))
AttribSpec const Vertex3::_spec[1] = {
    { AttribSpec::Position, 3, GL_FLOAT, false, sizeof(Vertex3), 0 },
};
LIBGUI_VERTEX_FORMAT_SPEC(Vertex3, 3 * sizeof(float))
AttribSpec const Vertex3Tex::_spec[2] = {
    { AttribSpec::Position,  3, GL_FLOAT, false, sizeof(Vertex3Tex), 0 },
    { AttribSpec::TexCoord0, 2, GL_FLOAT, false, sizeof(Vertex3Tex), 3 * sizeof(float) }
};
LIBGUI_VERTEX_FORMAT_SPEC(Vertex3Tex, 5 * sizeof(float))
AttribSpec const Vertex3TexRgba::_spec[3] = {
    { AttribSpec::Position,  3, GL_FLOAT, false, sizeof(Vertex3TexRgba), 0 },
    { AttribSpec::TexCoord0, 2, GL_FLOAT, false, sizeof(Vertex3TexRgba), 3 * sizeof(float) },
    { AttribSpec::Color,     4, GL_FLOAT, false, sizeof(Vertex3TexRgba), 5 * sizeof(float) }
};
LIBGUI_VERTEX_FORMAT_SPEC(Vertex3TexRgba, 9 * sizeof(float))
AttribSpec const Vertex3TexBoundsRgba::_spec[4] = {
    { AttribSpec::Position,   3, GL_FLOAT, false, sizeof(Vertex3TexBoundsRgba), 0 },
    { AttribSpec::TexCoord0,  2, GL_FLOAT, false, sizeof(Vertex3TexBoundsRgba), 3 * sizeof(float) },
    { AttribSpec::TexBounds0, 4, GL_FLOAT, false, sizeof(Vertex3TexBoundsRgba), 5 * sizeof(float) },
    { AttribSpec::Color,      4, GL_FLOAT, false, sizeof(Vertex3TexBoundsRgba), 9 * sizeof(float) }
};
LIBGUI_VERTEX_FORMAT_SPEC(Vertex3TexBoundsRgba, 13 * sizeof(float))
AttribSpec const Vertex3Tex2BoundsRgba::_spec[5] = {
    { AttribSpec::Position,   3, GL_FLOAT, false, sizeof(Vertex3Tex2BoundsRgba), 0 },
    { AttribSpec::TexCoord0,  2, GL_FLOAT, false, sizeof(Vertex3Tex2BoundsRgba), 3 * sizeof(float) },
    { AttribSpec::TexCoord1,  2, GL_FLOAT, false, sizeof(Vertex3Tex2BoundsRgba), 5 * sizeof(float) },
    { AttribSpec::TexBounds0, 4, GL_FLOAT, false, sizeof(Vertex3Tex2BoundsRgba), 7 * sizeof(float) },
    { AttribSpec::Color,      4, GL_FLOAT, false, sizeof(Vertex3Tex2BoundsRgba), 11 * sizeof(float) }
};
LIBGUI_VERTEX_FORMAT_SPEC(Vertex3Tex2BoundsRgba, 15 * sizeof(float))
AttribSpec const Vertex3Tex2Rgba::_spec[4] = {
    { AttribSpec::Position,  3, GL_FLOAT, false, sizeof(Vertex3Tex2Rgba), 0 },
    { AttribSpec::TexCoord0, 2, GL_FLOAT, false, sizeof(Vertex3Tex2Rgba), 3 * sizeof(float) },
    { AttribSpec::TexCoord1, 2, GL_FLOAT, false, sizeof(Vertex3Tex2Rgba), 5 * sizeof(float) },
    { AttribSpec::Color,     4, GL_FLOAT, false, sizeof(Vertex3Tex2Rgba), 7 * sizeof(float) }
};
LIBGUI_VERTEX_FORMAT_SPEC(Vertex3Tex2Rgba, 11 * sizeof(float))
AttribSpec const Vertex3Tex3Rgba::_spec[5] = {
    { AttribSpec::Position,  3, GL_FLOAT, false, sizeof(Vertex3Tex3Rgba), 0 },
    { AttribSpec::TexCoord0, 2, GL_FLOAT, false, sizeof(Vertex3Tex3Rgba), 3 * sizeof(float) },
    { AttribSpec::TexCoord1, 2, GL_FLOAT, false, sizeof(Vertex3Tex3Rgba), 5 * sizeof(float) },
    { AttribSpec::TexCoord2, 2, GL_FLOAT, false, sizeof(Vertex3Tex3Rgba), 7 * sizeof(float) },
    { AttribSpec::Color,     4, GL_FLOAT, false, sizeof(Vertex3Tex3Rgba), 9 * sizeof(float) }
};
LIBGUI_VERTEX_FORMAT_SPEC(Vertex3Tex3Rgba, 13 * sizeof(float))
AttribSpec const Vertex3NormalTexRgba::_spec[4] = {
    { AttribSpec::Position,  3, GL_FLOAT, false, sizeof(Vertex3NormalTexRgba), 0 },
    { AttribSpec::Normal,    3, GL_FLOAT, false, sizeof(Vertex3NormalTexRgba), 3 * sizeof(float) },
    { AttribSpec::TexCoord0, 2, GL_FLOAT, false, sizeof(Vertex3NormalTexRgba), 6 * sizeof(float) },
    { AttribSpec::Color,     4, GL_FLOAT, false, sizeof(Vertex3NormalTexRgba), 8 * sizeof(float) }
};
LIBGUI_VERTEX_FORMAT_SPEC(Vertex3NormalTexRgba, 12 * sizeof(float))
AttribSpec const Vertex3NormalTangentTex::_spec[5] = {
    { AttribSpec::Position,  3, GL_FLOAT, false, sizeof(Vertex3NormalTangentTex), 0 },
    { AttribSpec::Normal,    3, GL_FLOAT, false, sizeof(Vertex3NormalTangentTex), 3 * sizeof(float) },
    { AttribSpec::Tangent,   3, GL_FLOAT, false, sizeof(Vertex3NormalTangentTex), 6 * sizeof(float) },
    { AttribSpec::Bitangent, 3, GL_FLOAT, false, sizeof(Vertex3NormalTangentTex), 9 * sizeof(float) },
    { AttribSpec::TexCoord0, 2, GL_FLOAT, false, sizeof(Vertex3NormalTangentTex), 12 * sizeof(float) }
};
LIBGUI_VERTEX_FORMAT_SPEC(Vertex3NormalTangentTex, 14 * sizeof(float))
static duint drawCounter = 0;
DENG2_PIMPL(GLBuffer) {
    GLuint           vao             = 0;
    GLProgram const *vaoBoundProgram = nullptr;
    GLuint           name            = 0;
    GLuint           idxName         = 0;
    dsize            count           = 0;
    dsize            idxCount        = 0;
    DrawRanges       defaultRange; ///< All vertices.
    Primitive        prim  = Points;
    AttribSpecs      specs{nullptr, 0};
    Impl(Public *i) : Base(i)
    {}
    ~Impl() {
        release();
        releaseIndices();
        releaseArray();
    }
    void allocArray() {
#if defined (DENG_HAVE_VAOS)
        if (!vao) {
            LIBGUI_GL.glGenVertexArrays(1, &vao);
        }
#endif
    }
    void releaseArray() {
#if defined (DENG_HAVE_VAOS)
        if (vao) {
            LIBGUI_GL.glDeleteVertexArrays(1, &vao);
            vao = 0;
            vaoBoundProgram = nullptr;
        }
#endif
    }
    void alloc() {
        if (!name) {
            LIBGUI_GL.glGenBuffers(1, &name);
        }
    }
    void allocIndices() {
        if (!idxName) {
            LIBGUI_GL.glGenBuffers(1, &idxName);
        }
    }
    void release() {
        if (name) {
            LIBGUI_GL.glDeleteBuffers(1, &name);
            name = 0;
            count = 0;
            vaoBoundProgram = nullptr;
        }
    }
    void releaseIndices() {
        if (idxName) {
            LIBGUI_GL.glDeleteBuffers(1, &idxName);
            idxName = 0;
            idxCount = 0;
        }
    }
    static GLenum glUsage(Usage u) {
        switch (u) {
        case Static:  return GL_STATIC_DRAW;
        case Dynamic: return GL_DYNAMIC_DRAW;
        case Stream:  return GL_STREAM_DRAW;
        }
        DENG2_ASSERT(false);
        return GL_STATIC_DRAW;
    }
    static GLenum glPrimitive(Primitive p) {
        switch (p) {
        case Points:        return GL_POINTS;
        case LineStrip:     return GL_LINE_STRIP;
        case LineLoop:      return GL_LINE_LOOP;
        case Lines:         return GL_LINES;
        case TriangleStrip: return GL_TRIANGLE_STRIP;
        case TriangleFan:   return GL_TRIANGLE_FAN;
        case Triangles:     return GL_TRIANGLES;
        }
        DENG2_ASSERT(false);
        return GL_TRIANGLES;
    }
    void setAttribPointer(GLuint index, AttribSpec const &spec, int divisor, int part = 0) const {
        DENG2_ASSERT(!part || spec.type == GL_FLOAT);
        auto &GL = LIBGUI_GL;
        GL.glEnableVertexAttribArray(index + part);
        LIBGUI_ASSERT_GL_OK();
        GL.glVertexAttribPointer(index + part,
                                        min(4, spec.size),
                                        spec.type,
                                        spec.normalized,
                                        spec.stride,
                                        (void const *) dintptr(spec.startOffset + part * 4 * sizeof(float)));
        LIBGUI_ASSERT_GL_OK();
#if defined (DENG_HAVE_INSTANCES)
        GL.glVertexAttribDivisor(index + part, divisor);
        LIBGUI_ASSERT_GL_OK();
#else
        DENG2_UNUSED(divisor);
#endif
    }
    void enableArrays(bool enable, int divisor = 0, GLuint vaoName = 0) {
        auto &GL = LIBGUI_GL;
        if (!enable) {
#if defined (DENG_HAVE_VAOS)
            GL.glBindVertexArray(0);
#endif
            return;
        }
        DENG2_ASSERT(GLProgram::programInUse());
        DENG2_ASSERT(specs.first != 0); // must have a spec
#if defined (DENG_HAVE_VAOS)
        DENG2_ASSERT(vaoName || vao);
#else
        DENG2_UNUSED(vaoName);
#endif
#if defined (DENG_HAVE_VAOS)
        GL.glBindVertexArray(vaoName? vaoName : vao);
#endif
        GL.glBindBuffer(GL_ARRAY_BUFFER, name);
        vaoBoundProgram = GLProgram::programInUse();
        for (duint i = 0; i < specs.second; ++i) {
            AttribSpec const &spec = specs.first[i];
            int index = vaoBoundProgram->attributeLocation(spec.semantic);
            if (index < 0) continue; // Not used.
            if (spec.size == 16) {
                for (int part = 0; part < 4; ++part) {
                    if (enable)
                        setAttribPointer(index, spec, divisor, part);
                    else {
                        GL.glDisableVertexAttribArray(index + part);
                        LIBGUI_ASSERT_GL_OK();
                    }
                }
            }
            else {
                if (enable)
                    setAttribPointer(index, spec, divisor);
                else {
                    GL.glDisableVertexAttribArray(index);
                    LIBGUI_ASSERT_GL_OK();
                }
            }
        }
        GL.glBindBuffer(GL_ARRAY_BUFFER, 0);
    }
    void bindArray(bool doBind) {
#if defined (DENG_HAVE_VAOS)
        if (doBind) {
            DENG2_ASSERT(vao != 0);
            DENG2_ASSERT(GLProgram::programInUse());
            if (vaoBoundProgram != GLProgram::programInUse()) {
                enableArrays(true);
            }
            else {
                LIBGUI_GL.glBindVertexArray(vao);
            }
        }
        else {
            LIBGUI_GL.glBindVertexArray(0);
        }
#else
        enableArrays(doBind);
#endif
    }
};
GLBuffer::GLBuffer() : d(new Impl(this))
{}
void GLBuffer::clear() {
    setState(NotReady);
    d->release();
    d->releaseIndices();
    d->releaseArray();
}
void GLBuffer::setVertices(dsize count, void const *data, dsize dataSize, Usage usage) {
    setVertices(Points, count, data, dataSize, usage);
}
void GLBuffer::setVertices(Primitive primitive, dsize count, void const *data, dsize dataSize, Usage usage) {
    d->prim  = primitive;
    d->count = count;
    d->defaultRange.clear();
    d->defaultRange.append(Rangeui(0, count));
    if (data) {
        d->allocArray();
        d->alloc();
        if (dataSize && count) {
            auto &GL = LIBGUI_GL;
            GL.glBindBuffer(GL_ARRAY_BUFFER, d->name);
            GL.glBufferData(GL_ARRAY_BUFFER, dataSize, data, Impl::glUsage(usage));
            GL.glBindBuffer(GL_ARRAY_BUFFER, 0);
        }
        setState(Ready);
    }
    else {
        d->release();
        setState(NotReady);
    }
}
void GLBuffer::setIndices(Primitive primitive, dsize count, Index const *indices, Usage usage) {
    d->prim     = primitive;
    d->idxCount = count;
    d->defaultRange.clear();
    d->defaultRange.append(Rangeui(0, count));
    if (indices && count) {
        d->allocArray();
        d->allocIndices();
        auto &GL = LIBGUI_GL;
        GL.glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, d->idxName);
        GL.glBufferData(GL_ELEMENT_ARRAY_BUFFER, GLsizeiptr(count * sizeof(Index)),
                        indices, Impl::glUsage(usage));
        GL.glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
    }
    else {
        d->releaseIndices();
    }
}
void GLBuffer::setIndices(Primitive primitive, Indices const &indices, Usage usage) {
    setIndices(primitive, indices.size(), indices.constData(), usage);
}
void GLBuffer::setData(void const *data, dsize dataSize, gl::Usage usage) {
    if (data && dataSize) {
        d->alloc();
        auto &GL = LIBGUI_GL;
        GL.glBindBuffer(GL_ARRAY_BUFFER, d->name);
        GL.glBufferData(GL_ARRAY_BUFFER, GLsizeiptr(dataSize), data, Impl::glUsage(usage));
        GL.glBindBuffer(GL_ARRAY_BUFFER, 0);
    }
    else {
        d->release();
    }
}
void GLBuffer::setData(dsize startOffset, void const *data, dsize dataSize) {
    DENG2_ASSERT(isReady());
    if (data && dataSize) {
        auto &GL = LIBGUI_GL;
        GL.glBindBuffer   (GL_ARRAY_BUFFER, d->name);
        GL.glBufferSubData(GL_ARRAY_BUFFER, GLintptr(startOffset), GLsizeiptr(dataSize), data);
        GL.glBindBuffer   (GL_ARRAY_BUFFER, 0);
    }
}
void GLBuffer::setUninitializedData(dsize dataSize, gl::Usage usage) {
    d->count = 0;
    d->defaultRange.clear();
    d->allocArray();
    d->alloc();
    LIBGUI_GL.glBindBuffer(GL_ARRAY_BUFFER, d->name);
    LIBGUI_GL.glBufferData(GL_ARRAY_BUFFER, GLsizeiptr(dataSize), nullptr, Impl::glUsage(usage));
    LIBGUI_GL.glBindBuffer(GL_ARRAY_BUFFER, 0);
    setState(Ready);
}
void GLBuffer::draw(DrawRanges const *ranges) const {
    if (!isReady() || !GLProgram::programInUse()) return;
    GLState::current().target().markAsChanged();
    auto &GL = LIBGUI_GL;
    d->bindArray(true);
    if (d->idxName) {
        GL.glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, d->idxName);
        DENG2_ASSERT(GLProgram::programInUse()->validate());
        for (Rangeui const &range : (ranges? *ranges : d->defaultRange)) {
            GL.glDrawElements(Impl::glPrimitive(d->prim),
                              range.size(), GL_UNSIGNED_SHORT,
                              (void const *) dintptr(range.start * 2));
            LIBGUI_ASSERT_GL_OK();
        }
        GL.glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
    }
    else {
        DENG2_ASSERT(GLProgram::programInUse()->validate());
        for (Rangeui const &range : (ranges? *ranges : d->defaultRange)) {
            GL.glDrawArrays(Impl::glPrimitive(d->prim), range.start, range.size());
            LIBGUI_ASSERT_GL_OK();
        }
    }
    ++drawCounter;
#ifdef DENG2_DEBUG
    DENG2_ASSERT(GLDrawQueue_queuedElems == 0);
#endif
    d->bindArray(false);
}
void GLBuffer::drawWithIndices(GLBuffer const &indexBuffer) const {
    if (!isReady() || !indexBuffer.d->idxName || !GLProgram::programInUse()) return;
    GLState::current().target().markAsChanged();
    auto &GL = LIBGUI_GL;
    d->bindArray(true);
    DENG2_ASSERT(GLProgram::programInUse()->validate());
    GL.glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, indexBuffer.d->idxName);
    GL.glDrawElements(Impl::glPrimitive(indexBuffer.d->prim),
                      GLsizei(indexBuffer.d->idxCount),
                      GL_UNSIGNED_SHORT, nullptr);
    LIBGUI_ASSERT_GL_OK();
    GL.glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
    ++drawCounter;
    d->bindArray(false);
}
void GLBuffer::drawWithIndices(gl::Primitive primitive, Index const *indices, dsize count) const {
    if (!isReady() || !indices || !count || !GLProgram::programInUse()) return;
    GLState::current().target().markAsChanged();
    auto &GL = LIBGUI_GL;
    d->bindArray(true);
    DENG2_ASSERT(GLProgram::programInUse()->validate());
    GL.glDrawElements(Impl::glPrimitive(primitive), GLsizei(count), GL_UNSIGNED_SHORT, indices);
    LIBGUI_ASSERT_GL_OK();
    ++drawCounter;
    d->bindArray(false);
}
void GLBuffer::drawInstanced(GLBuffer const &instanceAttribs, duint first, dint count) const {
#if defined (DENG_HAVE_INSTANCES)
    if (!isReady() || !instanceAttribs.isReady() || !GLProgram::programInUse()) return;
    GLState::current().target().markAsChanged();
    auto &GL = LIBGUI_GL;
    d->enableArrays(true);
    instanceAttribs.d->enableArrays(true, 1 /* per instance */, d->vao);
    if (d->idxName) {
        if (count < 0) count = d->idxCount;
        if (first + count > d->idxCount) count = d->idxCount - first;
        DENG2_ASSERT(count >= 0);
        GL.glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, d->idxName);
        DENG2_ASSERT(GLProgram::programInUse()->validate());
        GL.glDrawElementsInstanced(Impl::glPrimitive(d->prim), count, GL_UNSIGNED_SHORT,
                                   reinterpret_cast<void const *>(dintptr(first * 2)),
                                   GLsizei(instanceAttribs.count()));
        LIBGUI_ASSERT_GL_OK();
        GL.glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
    }
    else {
        if (count < 0) count = d->count;
        if (first + count > d->count) count = d->count - first;
        DENG2_ASSERT(count >= 0);
        DENG2_ASSERT(GLProgram::programInUse()->validate());
        GL.glDrawArraysInstanced(Impl::glPrimitive(d->prim), first, count,
                                 GLsizei(instanceAttribs.count()));
        LIBGUI_ASSERT_GL_OK();
    }
    d->enableArrays(false);
    instanceAttribs.d->enableArrays(false);
#else
    DENG2_UNUSED(instanceAttribs);
    DENG2_UNUSED(first);
    DENG2_UNUSED(count);
#endif
}
dsize GLBuffer::count() const {
    return d->count;
}
void GLBuffer::setFormat(AttribSpecs const &format) {
    d->specs = format;
}
GLuint GLBuffer::glName() const {
    return d->name;
}
duint GLBuffer::drawCount() // static
{
    return drawCounter;
}
void GLBuffer::resetDrawCount() // static
{
    drawCounter = 0;
}
} // namespace de
namespace de {
#ifdef DENG2_DEBUG
extern int GLDrawQueue_queuedElems;
int GLDrawQueue_queuedElems = 0;
#endif
DENG2_PIMPL_NOREF(GLDrawQueue) {
    GLProgram *       currentProgram = nullptr;
    GLBuffer const *  currentBuffer  = nullptr;
    GLBuffer::Indices indices;
    QList<GLBuffer *> indexBuffers;
    int               indexBufferPos = 0;
    dsize batchIndex = 0;
    Vector4f defaultColor;
    std::unique_ptr<GLUniform> uBatchColors;
    Vector4f defaultScissor;
    GLUniform uBatchScissors { "uScissorRect", GLUniform::Vec4Array, GLShader::MAX_BATCH_UNIFORMS };
    float defaultSaturation = 1.f;
    GLUniform uBatchSaturation { "uSaturation", GLUniform::FloatArray, GLShader::MAX_BATCH_UNIFORMS };
    ~Impl() {
        qDeleteAll(indexBuffers);
    }
    GLBuffer &nextIndexBuffer() {
        if (indexBufferPos == indexBuffers.size()) {
            indexBuffers << new GLBuffer;
        }
        indexBufferPos++;
        return *indexBuffers.last();
    }
    void unsetProgram() {
        if (currentProgram) {
            if (uBatchColors) {
                currentProgram->unbind(*uBatchColors);
                uBatchColors.reset();
                currentProgram->unbind(uBatchScissors);
                currentProgram->unbind(uBatchSaturation);
            }
            currentProgram = nullptr;
        }
    }
    void restoreBatchValues() {
        if (uBatchColors) {
            uBatchColors   ->set(batchIndex, defaultColor);
            uBatchScissors  .set(batchIndex, defaultScissor);
            uBatchSaturation.set(batchIndex, defaultSaturation);
        }
    }
};
GLDrawQueue::GLDrawQueue()
    : d(new Impl)
{}
void GLDrawQueue::setProgram(GLProgram &program,
                             Block const &batchUniformName,
                             GLUniform::Type batchUniformType) {
    if (d->currentProgram && d->currentProgram != &program) {
        flush();
    }
    d->unsetProgram();
    d->currentProgram = &program;
    if (batchUniformName) {
        d->uBatchColors.reset(new GLUniform(batchUniformName, batchUniformType,
                                            GLShader::MAX_BATCH_UNIFORMS));
        program << *d->uBatchColors;
        program << d->uBatchScissors;
        program << d->uBatchSaturation;
    }
}
int GLDrawQueue::batchIndex() const {
    return int(d->batchIndex);
}
void GLDrawQueue::setBatchColor(Vector4f const &color) {
    if (d->uBatchColors) {
        d->uBatchColors->set(d->batchIndex, color);
    }
    d->defaultColor = color;
}
void GLDrawQueue::setBatchSaturation(float saturation) {
    d->uBatchSaturation.set(d->batchIndex, saturation);
    d->defaultSaturation = saturation;
}
void GLDrawQueue::setBatchScissorRect(Vector4f const &scissor) {
    d->uBatchScissors.set(d->batchIndex, scissor);
    d->defaultScissor = scissor;
}
void GLDrawQueue::setBuffer(GLBuffer const &buffer) {
    if (d->currentBuffer && &buffer != d->currentBuffer) {
        flush();
    }
    d->currentBuffer = &buffer;
}
void GLDrawQueue::enqueueDraw(GLSubBuffer const &buffer) {
    DENG2_ASSERT(&buffer.hostBuffer() == d->currentBuffer);
    DENG2_ASSERT(d->currentProgram);
    if (buffer.size() == 0) return;
    if (d->indices.size() > 0) {
        d->indices << d->indices.last();
        d->indices << buffer.hostRange().start;
    }
    for (duint16 i = buffer.hostRange().start;
                 i < buffer.hostRange().start + buffer.size(); ++i) {
        d->indices << i;
    }
    if (d->uBatchColors) {
        d->batchIndex++;
        if (d->batchIndex == GLShader::MAX_BATCH_UNIFORMS) {
            flush();
        }
        d->restoreBatchValues();
    }
#ifdef DENG2_DEBUG
    GLDrawQueue_queuedElems = d->indices.size();
#endif
}
void GLDrawQueue::flush() {
    DENG2_ASSERT(d->currentProgram);
    if (d->currentBuffer) {
#ifdef DENG2_DEBUG
        GLDrawQueue_queuedElems = 0;
#endif
        GLState::current().apply();
        dsize const batchCount = d->batchIndex;
        /*qDebug() << "[GLDrawQueue] Flushing" << d->indices.size() << "elements"
                 << "consisting of" << batchCount << "batches";*/
        GLBuffer &indexBuffer = d->nextIndexBuffer();
        indexBuffer.setIndices(gl::TriangleStrip, d->indices, gl::Stream);
        d->indices.clear();
        if (d->uBatchColors) {
            d->uBatchColors   ->setUsedElementCount(batchCount);
            d->uBatchScissors  .setUsedElementCount(batchCount);
            d->uBatchSaturation.setUsedElementCount(batchCount);
        }
        d->currentProgram->beginUse();
        d->currentBuffer ->drawWithIndices(indexBuffer);
        d->currentProgram->endUse();
        d->indices.clear();
    }
    d->currentBuffer = nullptr;
    d->batchIndex = 0;
    d->restoreBatchValues();
}
void GLDrawQueue::beginFrame() {
    d->indexBufferPos = 0;
}
} // namespace de
namespace de {
static Vector2ui const nullSize;
static GLuint defaultFramebuffer = 0;
DENG2_PIMPL(GLFramebuffer),
DENG2_OBSERVES(Asset, Deletion) {
    enum AttachmentId {
        ColorBuffer,
        DepthBuffer,
        StencilBuffer,
        DepthStencilBuffer,
        MAX_ATTACHMENTS
    };
    static AttachmentId attachmentToId(GLenum atc) {
        switch (atc) {
        case GL_COLOR_ATTACHMENT0:
            return ColorBuffer;
        case GL_DEPTH_ATTACHMENT:
            return DepthBuffer;
        case GL_STENCIL_ATTACHMENT:
            return StencilBuffer;
        case GL_DEPTH_STENCIL_ATTACHMENT:
            return DepthStencilBuffer;
        default:
            DENG2_ASSERT(false);
            break;
        }
        return ColorBuffer; // should not be reached
    }
    static GLenum flagsToGLAttachment(Flags const &flags) {
        DENG2_ASSERT(!flags.testFlag(ColorDepth));
        DENG2_ASSERT(!flags.testFlag(ColorDepthStencil));
        return flags == Color?   GL_COLOR_ATTACHMENT0  :
               flags == Depth?   GL_DEPTH_ATTACHMENT   :
               flags == Stencil? GL_STENCIL_ATTACHMENT :
                                 GL_DEPTH_STENCIL_ATTACHMENT;
    }
    GLuint      fbo;
    GLuint      renderBufs[MAX_ATTACHMENTS];
    GLTexture * bufTextures[MAX_ATTACHMENTS];
    Flags       flags;
    Flags       textureAttachment; ///< Where to attach @a texture.
    GLTexture * texture;
    Vector2ui   size;
    Vector4f    clearColor;
    Rectangleui activeRect; ///< Initially null.
    int         sampleCount;
    Impl(Public *i)
        : Base(i)
        , fbo(0)
        , flags(DefaultFlags)
        , textureAttachment(NoAttachments)
        , texture(0)
        , sampleCount(0) {
        zap(renderBufs);
        zap(bufTextures);
    }
    Impl(Public *i, Flags const &texAttachment, GLTexture &colorTexture, Flags const &otherAtm)
        : Base(i)
        , fbo(0)
        , flags(texAttachment | otherAtm)
        , textureAttachment(texAttachment)
        , texture(&colorTexture)
        , size(colorTexture.size())
        , sampleCount(0) {
        zap(renderBufs);
        zap(bufTextures);
    }
    Impl(Public *i, Vector2ui const &targetSize, Flags const &fboFlags)
        : Base(i)
        , fbo(0)
        , flags(fboFlags)
        , textureAttachment(NoAttachments)
        , texture(0)
        , size(targetSize)
        , sampleCount(0) {
        zap(renderBufs);
        zap(bufTextures);
    }
    ~Impl() {
        release();
    }
    bool isDefault() const {
        return !texture && size == nullSize;
    }
    static AttachmentId flagsToAttachmentId(Flags const &flags) {
        if (flags == Color) {
            return ColorBuffer;
        }
        if (flags == Depth) {
            return DepthBuffer;
        }
        if (flags == Stencil) {
            return StencilBuffer;
        }
        if (flags == DepthStencil) {
            return DepthStencilBuffer;
        }
        DENG2_ASSERT(0!="Invalid attachment flags");
        return MAX_ATTACHMENTS;
    }
    GLTexture *bufferTexture(Flags const &flags) const {
        auto attachId = flagsToAttachmentId(flags);
        if (attachId != MAX_ATTACHMENTS) {
            return bufTextures[attachId];
        }
        return nullptr;
    }
    GLuint renderBuffer(Flags const &flags) const {
        auto attachId = flagsToAttachmentId(flags);
        if (attachId != MAX_ATTACHMENTS) {
            return renderBufs[attachId];
        }
        return 0;
    }
    void allocFBO() {
        if (isDefault() || fbo) return;
        LIBGUI_GL.glGenFramebuffers(1, &fbo);
        LIBGUI_GL.glBindFramebuffer(GL_FRAMEBUFFER, fbo);
        LIBGUI_ASSERT_GL_OK();
        LOG_GL_XVERBOSE("Creating FBO %i", fbo);
    }
    void attachTexture(GLTexture &tex, GLenum attachment, int level = 0) {
        DENG2_ASSERT(tex.isReady());
        LOG_GL_XVERBOSE("FBO %i: glTex %i (level %i) => attachment %i",
                        fbo << tex.glName() << level << attachmentToId(attachment));
        LIBGUI_GL.glFramebufferTexture2D(GL_FRAMEBUFFER, attachment, GL_TEXTURE_2D, tex.glName(), level);
        LIBGUI_ASSERT_GL_OK();
        bufTextures[attachmentToId(attachment)] = &tex;
    }
    void attachRenderbuffer(AttachmentId id, GLenum type, GLenum attachment) {
        DENG2_ASSERT(size != Vector2ui(0, 0));
        LIBGUI_GL.glGenRenderbuffers(1, &renderBufs[id]);
        LIBGUI_GL.glBindRenderbuffer(GL_RENDERBUFFER, renderBufs[id]);
        LIBGUI_ASSERT_GL_OK();
#if !defined(DENG_OPENGL_ES)
        if (sampleCount > 1) {
            if (GLInfo::extensions().NV_framebuffer_multisample_coverage) {
                LOG_GL_VERBOSE("FBO %i: renderbuffer %ix%i is multisampled with %i CSAA samples => attachment %i")
                        << fbo << size.x << size.y << sampleCount
                        << attachmentToId(attachment);
                GLInfo::NV_framebuffer_multisample_coverage()->glRenderbufferStorageMultisampleCoverageNV(
                        GL_RENDERBUFFER, 8, sampleCount, type, size.x, size.y);
                LIBGUI_ASSERT_GL_OK();
            }
            else {
                LOG_GL_VERBOSE("FBO %i: renderbuffer %ix%i is multisampled with %i samples => attachment %i")
                        << fbo << size.x << size.y << sampleCount
                        << attachmentToId(attachment);
                LIBGUI_GL.glRenderbufferStorageMultisample(
                        GL_RENDERBUFFER, sampleCount, type, size.x, size.y);
                LIBGUI_ASSERT_GL_OK();
            }
        }
        else
#endif
        {
            LIBGUI_GL.glRenderbufferStorage(GL_RENDERBUFFER, type, size.x, size.y);
            LIBGUI_ASSERT_GL_OK();
        }
        LIBGUI_GL.glFramebufferRenderbuffer(
                    GL_FRAMEBUFFER, attachment, GL_RENDERBUFFER, renderBufs[id]);
        LIBGUI_ASSERT_GL_OK();
    }
    void alloc() {
        allocFBO();
        if (texture) {
            DENG2_ASSERT(textureAttachment == Color   ||
                         textureAttachment == Depth   ||
                         textureAttachment == Stencil ||
                         textureAttachment == DepthStencil);
            attachTexture(*texture,
                          textureAttachment == Color?   GL_COLOR_ATTACHMENT0  :
                          textureAttachment == Depth?   GL_DEPTH_ATTACHMENT   :
                          textureAttachment == Stencil? GL_STENCIL_ATTACHMENT :
                                                        GL_DEPTH_STENCIL_ATTACHMENT);
        }
        if (size != nullSize) // A non-default target: size must be specified.
        {
            allocRenderBuffers();
        }
        validate();
    }
    void allocRenderBuffers() {
        DENG2_ASSERT(size != nullSize);
        if (flags.testFlag(Color) && !textureAttachment.testFlag(Color)) {
            LOG_GL_VERBOSE("FBO %i: color renderbuffer %s") << fbo << size.asText();
            attachRenderbuffer(ColorBuffer, GL_RGBA8, GL_COLOR_ATTACHMENT0);
        }
        allocDepthStencilRenderBuffers();
        LIBGUI_GL.glBindRenderbuffer(GL_RENDERBUFFER, 0);
    }
    void allocDepthStencilRenderBuffers() {
        if (flags.testFlag(DepthStencil) && !flags.testFlag(SeparateDepthAndStencil) &&
            (!texture || textureAttachment == Color)) {
            LOG_GL_VERBOSE("FBO %i: depth+stencil renderbuffer %s") << fbo << size.asText();
            attachRenderbuffer(DepthStencilBuffer, GL_DEPTH24_STENCIL8, GL_DEPTH_STENCIL_ATTACHMENT);
        }
        else {
            if (flags.testFlag(Depth) && !textureAttachment.testFlag(Depth)) {
                LOG_GL_VERBOSE("FBO %i: depth renderbuffer %s") << fbo << size.asText();
                attachRenderbuffer(DepthBuffer, GL_DEPTH_COMPONENT, GL_DEPTH_ATTACHMENT);
            }
#if defined (DENG_OPENGL)
            if (flags.testFlag(Stencil) && !textureAttachment.testFlag(Stencil)) {
                LOG_GL_VERBOSE("FBO %i: stencil renderbuffer %s") << fbo << size.asText();
                attachRenderbuffer(StencilBuffer, GL_STENCIL_INDEX, GL_STENCIL_ATTACHMENT);
            }
#endif
        }
    }
    void releaseRenderBuffers() {
        LIBGUI_GL.glDeleteRenderbuffers(MAX_ATTACHMENTS, renderBufs);
        zap(renderBufs);
        zap(bufTextures);
    }
    void release() {
        self().setState(NotReady);
        if (fbo) {
            releaseRenderBuffers();
            LIBGUI_GL.glDeleteFramebuffers(1, &fbo);
            fbo = 0;
        }
        zap(bufTextures);
        texture = 0;
        size = nullSize;
    }
    void releaseAndReset() {
        release();
        textureAttachment = NoAttachments;
        flags = NoAttachments;
        sampleCount = 0;
    }
    void releaseRenderBuffer(AttachmentId id) {
        if (renderBufs[id]) {
            LIBGUI_GL.glDeleteRenderbuffers(1, &renderBufs[id]);
            renderBufs[id] = 0;
        }
    }
    void resizeRenderBuffers(Size const &newSize) {
        size = newSize;
        releaseRenderBuffers();
        allocRenderBuffers();
    }
    void replace(GLenum attachment, GLTexture &newTexture) {
        DENG2_ASSERT(self().isReady()); // must already be inited
        DENG2_ASSERT(bufTextures[attachmentToId(attachment)] != 0); // must have an attachment already
        LIBGUI_GL.glBindFramebuffer(GL_FRAMEBUFFER, fbo);
        attachTexture(newTexture, attachment);
        validate();
    }
    void replaceWithNewRenderBuffer(Flags const &attachment) {
        DENG2_ASSERT(self().isReady()); // must already be inited
        if (attachment == DepthStencil) // this supported only
        {
            LIBGUI_GL.glBindFramebuffer(GL_FRAMEBUFFER, fbo);
            allocDepthStencilRenderBuffers();
            validate();
        }
    }
    void replaceWithExistingRenderBuffer(Flags const &attachment, GLuint renderBufId) {
        DENG2_ASSERT(self().isReady());       // must already be inited
        auto id = flagsToAttachmentId(attachment);
        renderBufs[id] = renderBufId;
        LIBGUI_GL.glBindFramebuffer(GL_FRAMEBUFFER, fbo);
        LIBGUI_GL.glFramebufferRenderbuffer(
                    GL_FRAMEBUFFER, flagsToGLAttachment(attachment),
                    GL_RENDERBUFFER, renderBufs[id]);
        LIBGUI_ASSERT_GL_OK();
        GLState::current().target().glBind();
    }
    void validate() {
        if (isDefault()) {
            self().setState(Ready);
            return;
        }
        DENG2_ASSERT(fbo != 0);
        LIBGUI_GL.glBindFramebuffer(GL_FRAMEBUFFER, fbo);
        GLenum status = LIBGUI_GL.glCheckFramebufferStatus(GL_FRAMEBUFFER);
        if (status != GL_FRAMEBUFFER_COMPLETE) {
            releaseAndReset();
            throw ConfigError("GLFramebuffer::validate",
                status == GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT? "Incomplete attachments" :
                status == GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS? "Mismatch with dimensions" :
                status == GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT? "No images attached" :
                                                                        QString("Unsupported (0x%1)").arg(status, 0, 16));
        }
        self().setState(Ready);
        GLState::current().target().glBind();
        LIBGUI_ASSERT_GL_OK();
    }
    void assetBeingDeleted(Asset &asset) {
        if (texture == &asset) {
            release();
        }
    }
};
void GLFramebuffer::setDefaultFramebuffer(GLuint defaultFBO) {
    defaultFramebuffer = defaultFBO;
}
GLFramebuffer::GLFramebuffer() : d(new Impl(this)) {
    setState(Ready);
}
GLFramebuffer::GLFramebuffer(GLTexture &colorTarget, Flags const &otherAttachments)
    : d(new Impl(this, Color, colorTarget, otherAttachments)) {
    LOG_AS("GLFramebuffer");
    d->alloc();
}
GLFramebuffer::GLFramebuffer(Flags const &attachment, GLTexture &texture, Flags const &otherAttachments)
    : d(new Impl(this, attachment, texture, otherAttachments)) {
    LOG_AS("GLFramebuffer");
    d->alloc();
}
GLFramebuffer::GLFramebuffer(Vector2ui const &size, Flags const &flags)
    : d(new Impl(this, size, flags)) {
    LOG_AS("GLFramebuffer");
    d->alloc();
}
GLFramebuffer::Flags GLFramebuffer::flags() const {
    return d->flags;
}
void GLFramebuffer::markAsChanged() {
    d->flags |= Changed;
}
void GLFramebuffer::configure() {
    LOG_AS("GLFramebuffer");
    d->releaseAndReset();
    setState(Ready);
}
void GLFramebuffer::configure(Vector2ui const &size, Flags const &flags, int sampleCount) {
    LOG_AS("GLFramebuffer");
    d->releaseAndReset();
    d->flags = flags;
    d->size = size;
#if defined (DENG_OPENGL_ES)
    DENG2_UNUSED(sampleCount);
#else
    d->sampleCount = (sampleCount > 1? sampleCount : 0);
#endif
    d->allocFBO();
    d->allocRenderBuffers();
    d->validate();
    LIBGUI_ASSERT_GL_OK();
}
void GLFramebuffer::configure(GLTexture *colorTex, GLTexture *depthStencilTex) {
    DENG2_ASSERT(colorTex || depthStencilTex);
    LOG_AS("GLFramebuffer");
    d->releaseAndReset();
    d->flags = ColorDepthStencil;
    d->size = (colorTex? colorTex->size() : depthStencilTex->size());
    d->allocFBO();
    if (colorTex) {
        DENG2_ASSERT(colorTex->isReady());
        DENG2_ASSERT(d->size == colorTex->size());
        d->attachTexture(*colorTex, GL_COLOR_ATTACHMENT0);
    }
    else {
        d->attachRenderbuffer(Impl::ColorBuffer, GL_RGBA8, GL_COLOR_ATTACHMENT0);
    }
    if (depthStencilTex) {
        DENG2_ASSERT(depthStencilTex->isReady());
        DENG2_ASSERT(d->size == depthStencilTex->size());
        d->attachTexture(*depthStencilTex, GL_DEPTH_STENCIL_ATTACHMENT);
    }
    else {
        d->attachRenderbuffer(Impl::DepthStencilBuffer, GL_DEPTH24_STENCIL8, GL_DEPTH_STENCIL_ATTACHMENT);
    }
    LIBGUI_ASSERT_GL_OK();
    d->validate();
}
void GLFramebuffer::configure(Flags const &attachment, GLTexture &texture, Flags const &otherAttachments) {
    LOG_AS("GLFramebuffer");
    d->releaseAndReset();
    d->texture = &texture;
    d->textureAttachment = attachment;
    d->flags = attachment | otherAttachments;
    d->size = texture.size();
    d->alloc();
}
void GLFramebuffer::deinit() {
    LOG_AS("GLFramebuffer");
    d->releaseAndReset();
}
void GLFramebuffer::glBind() const {
    LIBGUI_ASSERT_GL_OK();
    DENG2_ASSERT(isReady());
    if (!isReady()) {
        return;
    }
    GLuint const fbo = (d->fbo? d->fbo : defaultFramebuffer);
    LIBGUI_GL.glBindFramebuffer(GL_FRAMEBUFFER, fbo);
    LIBGUI_ASSERT_GL_OK();
}
void GLFramebuffer::glRelease() const {
    LIBGUI_ASSERT_GL_OK();
    LIBGUI_GL.glBindFramebuffer(GL_FRAMEBUFFER, defaultFramebuffer); // both read and write FBOs
    LIBGUI_ASSERT_GL_OK();
}
QImage GLFramebuffer::toImage() const {
    if (!d->fbo) {
        return GLWindow::main().grabImage();
    }
    else if (d->flags & Color) {
        Size imgSize = size();
        QImage img(QSize(imgSize.x, imgSize.y), QImage::Format_ARGB32);
        LIBGUI_GL.glBindFramebuffer(GL_READ_FRAMEBUFFER, d->fbo);
        LIBGUI_GL.glPixelStorei(GL_PACK_ALIGNMENT, 4);
        LIBGUI_GL.glReadPixels(0, 0, imgSize.x, imgSize.y, GL_BGRA, GL_UNSIGNED_BYTE,
                               (GLvoid *) img.constBits());
        GLState::current().target().glBind();
        return img.mirrored(false, true);
    }
    return QImage();
}
void GLFramebuffer::setClearColor(Vector4f const &color) {
    d->clearColor = color;
}
void GLFramebuffer::clear(Flags const &attachments) {
    DENG2_ASSERT(isReady());
    markAsChanged();
    GLState::current().apply();
    glBind();
    Flags which = attachments & d->flags;
    LIBGUI_GL.glClearColor(d->clearColor.x, d->clearColor.y, d->clearColor.z, d->clearColor.w);
    LIBGUI_GL.glClear((which & Color?   GL_COLOR_BUFFER_BIT   : 0) |
                      (which & Depth?   GL_DEPTH_BUFFER_BIT   : 0) |
                      (which & Stencil? GL_STENCIL_BUFFER_BIT : 0));
    GLState::current().target().glBind();
}
void GLFramebuffer::resize(Size const &size) {
    if (d->size == size || d->isDefault()) return;
    LIBGUI_GL.glBindFramebuffer(GL_FRAMEBUFFER, d->fbo);
    if (d->texture) {
        d->texture->setUndefinedImage(size, d->texture->imageFormat());
    }
    d->resizeRenderBuffers(size);
    GLState::current().target().glBind();
}
GLTexture *GLFramebuffer::attachedTexture(Flags const &attachment) const {
    return d->bufferTexture(attachment);
}
GLuint GLFramebuffer::attachedRenderBuffer(Flags const &attachment) const {
    return d->renderBuffer(attachment);
}
void GLFramebuffer::replaceAttachment(Flags const &attachment, GLTexture &texture) {
    d->replace(d->flagsToGLAttachment(attachment), texture);
}
void GLFramebuffer::replaceAttachment(Flags const &attachment, GLuint renderBufferId) {
    d->replaceWithExistingRenderBuffer(attachment, renderBufferId);
}
void GLFramebuffer::replaceWithNewRenderBuffer(Flags const &attachment) {
    d->replaceWithNewRenderBuffer(attachment);
}
void GLFramebuffer::releaseAttachment(Flags const &attachment) {
    d->releaseRenderBuffer(d->flagsToAttachmentId(attachment));
}
void GLFramebuffer::blit(GLFramebuffer &dest, Flags const &attachments, gl::Filter filtering) const {
    LIBGUI_ASSERT_GL_OK();
    LIBGUI_GL.glBindFramebuffer(GL_DRAW_FRAMEBUFFER, dest.glName());
    LIBGUI_ASSERT_GL_OK();
#if defined (DENG_HAVE_BLIT_FRAMEBUFFER)
    LIBGUI_GL.glBindFramebuffer(GL_READ_FRAMEBUFFER, glName());
    LIBGUI_ASSERT_GL_OK();
    Flags common = d->flags & dest.flags() & attachments;
    LIBGUI_GL.glBlitFramebuffer(
                0, 0, size().x, size().y,
                0, 0, dest.size().x, dest.size().y,
                (common.testFlag(Color)?   GL_COLOR_BUFFER_BIT   : 0) |
                (common.testFlag(Depth)?   GL_DEPTH_BUFFER_BIT   : 0) |
                (common.testFlag(Stencil)? GL_STENCIL_BUFFER_BIT : 0),
                filtering == gl::Nearest? GL_NEAREST : GL_LINEAR);
    LIBGUI_ASSERT_GL_OK();
#else
    qDebug() << "need to implement glBlitFramebuffer:" << glName() << "->" << dest.glName();
    qDebug() << "\t- from texture:" << attachedTexture(Color);
    qDebug() << "\t- to texture:" << dest.attachedTexture(Color);
#endif
    dest.markAsChanged();
    GLState::current().target().glBind();
}
void GLFramebuffer::blit(gl::Filter filtering) const {
    LIBGUI_ASSERT_GL_OK();
#if defined (DENG_HAVE_BLIT_FRAMEBUFFER)
    LIBGUI_GL.glBindFramebuffer(GL_READ_FRAMEBUFFER, glName());
    LIBGUI_GL.glBindFramebuffer(GL_DRAW_FRAMEBUFFER, defaultFramebuffer);
    LIBGUI_GL.glBlitFramebuffer(
                0, 0, size().x, size().y,
                0, 0, size().x, size().y,
                GL_COLOR_BUFFER_BIT,
                filtering == gl::Nearest? GL_NEAREST : GL_LINEAR);
    LIBGUI_ASSERT_GL_OK();
#else
    qDebug() << "need to implement glBlitFramebuffer:" << glName() << "-> 0";
    qDebug() << "\t- texture:" << attachedTexture(Color);
#endif
    GLState::current().target().glBind();
}
GLuint GLFramebuffer::glName() const {
    return d->fbo? d->fbo : defaultFramebuffer;
}
GLFramebuffer::Size GLFramebuffer::size() const {
    if (d->texture) {
        return d->texture->size();
    }
    else if (d->size != nullSize) {
        return d->size;
    }
    return GLWindow::main().pixelSize();
}
void GLFramebuffer::setActiveRect(Rectangleui const &rect, bool applyGLState) {
    d->activeRect = rect;
    if (applyGLState) {
        GLState::considerNativeStateUndefined();
        GLState::current().apply();
    }
}
void GLFramebuffer::unsetActiveRect(bool applyGLState) {
    setActiveRect(Rectangleui(), applyGLState);
}
Vector2f GLFramebuffer::activeRectScale() const {
    if (!hasActiveRect()) {
        return Vector2f(1, 1);
    }
    return Vector2f(d->activeRect.size()) / size();
}
Vector2f GLFramebuffer::activeRectNormalizedOffset() const {
    if (!hasActiveRect()) {
        return Vector2f(0, 0);
    }
    return Vector2f(d->activeRect.topLeft) / size();
}
Rectangleui GLFramebuffer::scaleToActiveRect(Rectangleui const &rectInTarget) const {
    if (!hasActiveRect()) {
        return rectInTarget;
    }
    Vector2f const scaling = activeRectScale();
    return Rectangleui(d->activeRect.left()  + scaling.x * rectInTarget.left(),
                       d->activeRect.top()   + scaling.y * rectInTarget.top(),
                       rectInTarget.width()  * scaling.x,
                       rectInTarget.height() * scaling.y);
}
Rectangleui const &GLFramebuffer::activeRect() const {
    return d->activeRect;
}
bool GLFramebuffer::hasActiveRect() const {
    return !d->activeRect.isNull();
}
Rectangleui GLFramebuffer::rectInUse() const {
    if (hasActiveRect()) {
        return activeRect();
    }
    return Rectangleui::fromSize(size());
}
} // namespace de
namespace de {
static GLInfo info;
DENG2_PIMPL_NOREF(GLInfo), public QOpenGLFunctions_Doomsday {
    bool inited = false;
    Extensions ext;
    Limits lim;
#if defined (DENG_OPENGL)
    std::unique_ptr<QOpenGLExtension_NV_framebuffer_multisample_coverage> NV_framebuffer_multisample_coverage;
    std::unique_ptr<QOpenGLExtension_NV_texture_barrier>                  NV_texture_barrier;
#endif
#ifdef WIN32
    BOOL (APIENTRY *wglSwapIntervalEXT)(int interval) = nullptr;
#endif
#ifdef DENG_X11
    PFNGLXSWAPINTERVALEXTPROC  glXSwapIntervalEXT  = nullptr;
    PFNGLXSWAPINTERVALSGIPROC  glXSwapIntervalSGI  = nullptr;
    PFNGLXSWAPINTERVALMESAPROC glXSwapIntervalMESA = nullptr;
#endif
#ifdef DENG_ENABLE_OPENGL_DEBUG_LOGGER
    QOpenGLDebugLogger *logger = nullptr;
#endif
    Impl() {
        zap(ext);
    }
    /**
     * This routine is based on the method used by David Blythe and Tom
     * McReynolds in the book "Advanced Graphics Programming Using OpenGL"
     * ISBN: 1-55860-659-9.
     */
    bool checkExtensionString(char const *name, char const *extensions) {
        GLubyte const *start;
        GLubyte const *c, *terminator;
        c = (GLubyte const *) strchr(name, ' ');
        if (c || *name == '\0')
            return false;
        if (!extensions)
            return false;
        start = (GLubyte const *) extensions;
        for (;;) {
            c = (GLubyte const *) strstr((char const *) start, name);
            if (!c)
                break;
            terminator = c + strlen(name);
            if (c == start || *(c - 1) == ' ') {
                if (*terminator == ' ' || *terminator == '\0') {
                    return true;
                }
            }
            start = terminator;
        }
        return false;
    }
    bool doQuery(char const *ext) {
        DENG2_ASSERT(ext);
        DENG2_ASSERT(QOpenGLContext::currentContext() != nullptr);
#if 0
#ifdef WIN32
        if (wglGetExtensionsStringARB != nullptr &&
           checkExtensionString(ext, (GLubyte const *)wglGetExtensionsStringARB(wglGetCurrentDC())))
            return true;
#endif
#endif
#ifdef DENG_X11
        if (checkExtensionString(ext, glXQueryExtensionsString(QX11Info::display(),
                                                               QX11Info::appScreen())))
            return true;
#endif
        return QOpenGLContext::currentContext()->hasExtension(ext);
    }
    bool query(char const *ext) {
        bool found = doQuery(ext);
        LOGDEV_GL_VERBOSE("%s: %b") << ext << found;
        return found;
    }
    void init() {
        LOG_AS("GLInfo");
        if (inited) return;
        #if defined (DENG_OPENGL_ES)
        {
            initializeOpenGLFunctions();
            static OpenGLFeature const requiredFeatures[] = {
                Multitexture,
                Shaders,
                Buffers,
                Framebuffers,
                BlendEquationSeparate,
                BlendSubtract,
                NPOTTextures,
                NPOTTextureRepeat,
            };
            for (auto const &feature : requiredFeatures) {
                if (!openGLFeatures().testFlag(feature)) {
                    qDebug() << "OpenGL feature reported as not available:" << feature;
                }
            }
        }
        #else
        {
            if (!initializeOpenGLFunctions()) {
                throw InitError("GLInfo::init", "Failed to initialize OpenGL");
            }
        }
        #endif
        inited = true;
        ext.EXT_texture_compression_s3tc   = query("GL_EXT_texture_compression_s3tc");
        ext.EXT_texture_filter_anisotropic = query("GL_EXT_texture_filter_anisotropic");
        ext.NV_framebuffer_multisample_coverage
                                           = query("GL_NV_framebuffer_multisample_coverage");
        ext.NV_texture_barrier             = query("GL_NV_texture_barrier");
        ext.KHR_debug                      = query("GL_KHR_debug");
        #ifdef WIN32
        {
            ext.Windows_EXT_swap_control       = query("WGL_EXT_swap_control");
            if (ext.Windows_EXT_swap_control) {
                wglSwapIntervalEXT = de::function_cast<decltype(wglSwapIntervalEXT)>
                        (QOpenGLContext::currentContext()->getProcAddress("wglSwapIntervalEXT"));
            }
        }
        #endif
        #ifdef DENG_X11
        {
            ext.X11_EXT_swap_control           = query("GLX_EXT_swap_control");
            ext.X11_SGI_swap_control           = query("GLX_SGI_swap_control");
            ext.X11_MESA_swap_control          = query("GLX_MESA_swap_control");
            if (ext.X11_EXT_swap_control) {
                glXSwapIntervalEXT = de::function_cast<decltype(glXSwapIntervalEXT)>
                        (glXGetProcAddress(reinterpret_cast<GLubyte const *>("glXSwapIntervalEXT")));
            }
            if (ext.X11_SGI_swap_control) {
                glXSwapIntervalSGI = de::function_cast<decltype(glXSwapIntervalSGI)>
                        (glXGetProcAddress(reinterpret_cast<GLubyte const *>("glXSwapIntervalSGI")));
            }
            if (ext.X11_MESA_swap_control) {
                glXSwapIntervalMESA = de::function_cast<decltype(glXSwapIntervalMESA)>
                        (glXGetProcAddress(reinterpret_cast<GLubyte const *>("glXSwapIntervalMESA")));
            }
        }
        #endif
        #ifdef DENG_ENABLE_OPENGL_DEBUG_LOGGER
        {
            logger = new QOpenGLDebugLogger(&GLWindow::main());
            if (!ext.KHR_debug) {
                qDebug() << "[GLInfo] GL_KHR_debug is not available";
            }
            else if (!logger->initialize()) {
                qWarning() << "[GLInfo] Failed to initialize debug logger!";
            }
            else {
                QObject::connect(logger, &QOpenGLDebugLogger::messageLogged,
                                 [] (QOpenGLDebugMessage const &debugMessage) {
                    if (debugMessage.severity() == QOpenGLDebugMessage::NotificationSeverity) {
                        return;
                    }
                    char const *mType = "--";
                    char const *mSeverity = "--";
                    switch (debugMessage.type()) {
                    case QOpenGLDebugMessage::ErrorType:
                        mType = "ERROR";
                        break;
                    case QOpenGLDebugMessage::DeprecatedBehaviorType:
                        mType = "Deprecated";
                        break;
                    case QOpenGLDebugMessage::UndefinedBehaviorType:
                        mType = "Undefined";
                        break;
                    case QOpenGLDebugMessage::PortabilityType:
                        mType = "Portability";
                        break;
                    case QOpenGLDebugMessage::PerformanceType:
                        mType = "Performance";
                        break;
                    case QOpenGLDebugMessage::OtherType:
                        mType = "Other";
                        break;
                    case QOpenGLDebugMessage::MarkerType:
                        mType = "Marker";
                        break;
                    case QOpenGLDebugMessage::GroupPushType:
                        mType = "Group Push";
                        break;
                    case QOpenGLDebugMessage::GroupPopType:
                        mType = "Group Pop";
                        break;
                    default:
                        break;
                    }
                    switch (debugMessage.severity()) {
                    case QOpenGLDebugMessage::HighSeverity:
                        mType = " HIGH ";
                        break;
                    case QOpenGLDebugMessage::MediumSeverity:
                        mType = "MEDIUM";
                        break;
                    case QOpenGLDebugMessage::LowSeverity:
                        mType = " low  ";
                        break;
                    case QOpenGLDebugMessage::NotificationSeverity:
                        mType = " note ";
                        break;
                    default:
                        break;
                    }
                    qDebug("[OpenGL] %04x %s (%s): %s",
                           debugMessage.id(),
                           mType,
                           mSeverity,
                           debugMessage.message().toLatin1().constData());
                });
                logger->startLogging(QOpenGLDebugLogger::SynchronousLogging);
            }
        }
        #endif
        /*
        if (ext.ARB_draw_instanced) {
            ARB_draw_instanced.reset(new QOpenGLExtension_ARB_draw_instanced);
            ARB_draw_instanced->initializeOpenGLFunctions();
        }
        if (ext.ARB_instanced_arrays) {
            ARB_instanced_arrays.reset(new QOpenGLExtension_ARB_instanced_arrays);
            ARB_instanced_arrays->initializeOpenGLFunctions();
        }
        if (ext.EXT_framebuffer_blit) {
            EXT_framebuffer_blit.reset(new QOpenGLExtension_EXT_framebuffer_blit);
            EXT_framebuffer_blit->initializeOpenGLFunctions();
        }
        if (ext.EXT_framebuffer_multisample) {
            EXT_framebuffer_multisample.reset(new QOpenGLExtension_EXT_framebuffer_multisample);
            EXT_framebuffer_multisample->initializeOpenGLFunctions();
        }
        if (ext.EXT_framebuffer_object) {
            EXT_framebuffer_object.reset(new QOpenGLExtension_EXT_framebuffer_object);
            EXT_framebuffer_object->initializeOpenGLFunctions();
        }*/
        #if defined (DENG_OPENGL)
        {
            if (ext.NV_framebuffer_multisample_coverage) {
                NV_framebuffer_multisample_coverage.reset(new QOpenGLExtension_NV_framebuffer_multisample_coverage);
                NV_framebuffer_multisample_coverage->initializeOpenGLFunctions();
            }
            if (ext.NV_texture_barrier) {
                NV_texture_barrier.reset(new QOpenGLExtension_NV_texture_barrier);
                NV_texture_barrier->initializeOpenGLFunctions();
            }
        }
        #endif
        glGetIntegerv(GL_MAX_TEXTURE_SIZE,              reinterpret_cast<GLint *>(&lim.maxTexSize));
        glGetIntegerv(GL_MAX_TEXTURE_IMAGE_UNITS,       reinterpret_cast<GLint *>(&lim.maxTexUnits)); // at least 16
        #if defined (DENG_OPENGL)
        {
            glGetFloatv(GL_SMOOTH_LINE_WIDTH_RANGE,       &lim.smoothLineWidth.start);
            glGetFloatv(GL_SMOOTH_LINE_WIDTH_GRANULARITY, &lim.smoothLineWidthGranularity);
        }
        #endif
        LIBGUI_ASSERT_GL_OK();
        if (ext.EXT_texture_filter_anisotropic) {
            glGetIntegerv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, (GLint *) &lim.maxTexFilterAniso);
        }
        if (CommandLine_CheckWith("-maxtex", 1)) {
            lim.maxTexSize = min(ceilPow2(String(CommandLine_Next()).toInt()),
                                 lim.maxTexSize);
            LOG_GL_NOTE("Using requested maximum texture size of %i x %i") << lim.maxTexSize << lim.maxTexSize;
        }
        QOpenGLContext const *ctx = QOpenGLContext::currentContext();
        QSurfaceFormat form = ctx->format();
        LOGDEV_GL_MSG("Initial OpenGL format:");
        LOGDEV_GL_MSG(" - version: %i.%i") << form.majorVersion() << form.minorVersion();
        LOGDEV_GL_MSG(" - profile: %s") << (form.profile() == QSurfaceFormat::CompatibilityProfile? "Compatibility" : "Core");
        LOGDEV_GL_MSG(" - color: R%i G%i B%i A%i bits") << form.redBufferSize() << form.greenBufferSize() << form.blueBufferSize() << form.alphaBufferSize();
        LOGDEV_GL_MSG(" - depth: %i bits") << form.depthBufferSize();
        LOGDEV_GL_MSG(" - stencil: %i bits") << form.stencilBufferSize();
        LOGDEV_GL_MSG(" - samples: %i") << form.samples();
        LOGDEV_GL_MSG(" - swap behavior: %i") << form.swapBehavior();
        LIBGUI_ASSERT_GL_OK();
    }
};
GLInfo::GLInfo() : d(new Impl)
{}
void GLInfo::glInit() {
    LIBGUI_ASSERT_GL_CONTEXT_ACTIVE();
    info.d->init();
}
void GLInfo::glDeinit() {
    #ifdef DENG_ENABLE_OPENGL_DEBUG_LOGGER
    {
        if (info.d->logger) {
            info.d->logger->stopLogging();
        }
    }
    #endif
    info.d.reset();
}
QOpenGLFunctions_Doomsday &GLInfo::api() // static
{
    DENG2_ASSERT(QOpenGLContext::currentContext() != nullptr);
    DENG2_ASSERT(info.d->inited);
    return *info.d;
}
/*
QOpenGLExtension_ARB_draw_instanced *GLInfo::ARB_draw_instanced() {
    DENG2_ASSERT(info.d->inited);
    return info.d->ARB_draw_instanced.get();
}
QOpenGLExtension_ARB_instanced_arrays *GLInfo::ARB_instanced_arrays() {
    DENG2_ASSERT(info.d->inited);
    return info.d->ARB_instanced_arrays.get();
}
QOpenGLExtension_EXT_framebuffer_blit *GLInfo::EXT_framebuffer_blit() {
    DENG2_ASSERT(info.d->inited);
    return info.d->EXT_framebuffer_blit.get();
}
QOpenGLExtension_EXT_framebuffer_multisample *GLInfo::EXT_framebuffer_multisample() {
    DENG2_ASSERT(info.d->inited);
    return info.d->EXT_framebuffer_multisample.get();
}
QOpenGLExtension_EXT_framebuffer_object *GLInfo::EXT_framebuffer_object() {
    DENG2_ASSERT(info.d->inited);
    return info.d->EXT_framebuffer_object.get();
}
*/
#if defined (DENG_OPENGL)
QOpenGLExtension_NV_framebuffer_multisample_coverage *GLInfo::NV_framebuffer_multisample_coverage() {
    DENG2_ASSERT(info.d->inited);
    return info.d->NV_framebuffer_multisample_coverage.get();
}
QOpenGLExtension_NV_texture_barrier *GLInfo::NV_texture_barrier() {
    DENG2_ASSERT(info.d->inited);
    return info.d->NV_texture_barrier.get();
}
#endif
void GLInfo::setSwapInterval(int interval) {
    DENG2_ASSERT(info.d->inited);
    #if defined (WIN32)
    {
        if (extensions().Windows_EXT_swap_control) {
            info.d->wglSwapIntervalEXT(interval);
        }
    }
    #elif defined (MACOSX)
    {
        CGLContextObj context = CGLGetCurrentContext();
        DENG2_ASSERT(context != nullptr);
        if (context) {
            GLint params[1] = { interval };
            CGLSetParameter(context, kCGLCPSwapInterval, params);
        }
    }
    #elif defined (DENG_X11)
    {
        if (extensions().X11_SGI_swap_control) {
            info.d->glXSwapIntervalSGI(interval);
        }
        else if (extensions().X11_MESA_swap_control) {
            info.d->glXSwapIntervalMESA(uint(interval));
        }
        else if (extensions().X11_EXT_swap_control) {
            info.d->glXSwapIntervalEXT(QX11Info::display(),
                                       GLWindow::main().winId(),
                                       interval);
        }
    }
    #endif
}
#if 0
void GLInfo::setLineWidth(float lineWidth) {
    #if defined (DENG_OPENGL)
    {
        LIBGUI_ASSERT_GL_CONTEXT_ACTIVE();
        DENG2_ASSERT(info.d->inited);
        info.d->glLineWidth(info.d->lim.smoothLineWidth.clamp(lineWidth));
        LIBGUI_ASSERT_GL_OK();
    }
    #else
    {
        DENG2_UNUSED(lineWidth);
    }
    #endif
}
#endif
GLInfo::Extensions const &GLInfo::extensions() {
    DENG2_ASSERT(info.d->inited);
    return info.d->ext;
}
GLInfo::Limits const &GLInfo::limits() {
    DENG2_ASSERT(info.d->inited);
    return info.d->lim;
}
bool GLInfo::isFramebufferMultisamplingSupported() {
    return true;
}
void GLInfo::checkError(char const *file, int line) {
    GLuint error = GL_NO_ERROR;
    do {
        error = LIBGUI_GL.glGetError();
        if (error != GL_NO_ERROR) {
            LogBuffer_Flush();
            qWarning("%s:%i: OpenGL error: 0x%x (%s)", file, line, error,
                     LIBGUI_GL_ERROR_STR(error));
            LIBGUI_ASSERT_GL(0!="OpenGL operation failed");
        }
    }
    while (error != GL_NO_ERROR);
}
} // namespace de
namespace de {
static GLProgram const *currentProgram = 0;
using namespace internal;
DENG2_PIMPL(GLProgram)
, DENG2_OBSERVES(GLUniform, ValueChange)
, DENG2_OBSERVES(GLUniform, Deletion) {
    typedef QSet<GLUniform const *>  Uniforms;
    typedef QList<GLUniform const *> UniformList;
    typedef QSet<GLShader const *>   Shaders;
    QHash<Block, QStack<GLUniform const *>> stacks;
    Uniforms    allBound;
    Uniforms    active; ///< Currently active bindings.
    Uniforms    changed;
    UniformList textures;
    bool        texturesChanged;
    int         attribLocation[AttribSpec::NUM_SEMANTICS]; ///< Where each attribute is bound.
    GLuint  name;
    Shaders shaders;
    bool    inUse;
    bool    needRebuild;
    Impl(Public *i)
        : Base(i)
        , texturesChanged(false)
        , name(0)
        , inUse(false)
        , needRebuild(false)
    {}
    ~Impl() {
        release();
    }
    void alloc() {
        if (!name) {
            name = LIBGUI_GL.glCreateProgram();
            if (!name) {
                throw AllocError("GLProgram::alloc", "Failed to create program");
            }
        }
    }
    void releaseButRetainBindings() {
        self().setState(NotReady);
        detachAllShaders();
        if (name) {
            LIBGUI_GL.glDeleteProgram(name);
            name = 0;
        }
    }
    void release() {
        unbindAll();
        releaseButRetainBindings();
    }
    void attach(GLShader const *shader) {
        DENG2_ASSERT(shader->isReady());
        alloc();
        LIBGUI_GL.glAttachShader(name, shader->glName());
        LIBGUI_ASSERT_GL_OK();
        shaders.insert(holdRef(shader));
    }
    void detach(GLShader const *shader) {
        if (shader->isReady()) {
            LIBGUI_GL.glDetachShader(name, shader->glName());
        }
        shaders.remove(shader);
        shader->release();
    }
    void detachAllShaders() {
        foreach (GLShader const *shader, shaders) {
            detach(shader);
        }
        shaders.clear();
    }
    void unbindAll() {
        for (GLUniform const *u : allBound) {
            u->audienceForValueChange() -= this;
            u->audienceForDeletion() -= this;
        }
        allBound.clear();
        stacks.clear();
        active.clear();
        changed.clear();
        textures.clear();
        texturesChanged = false;
    }
    /**
     * Bind all known vertex attributes to the indices used by GLBuffer. The
     * program is automatically (re)linked after binding the vertex attributes,
     * if there are already shaders attached.
     */
    void bindVertexAttribs() {
        alloc();
        if (!shaders.isEmpty()) {
            link();
        }
        static struct {
            AttribSpec::Semantic semantic;
            char const *varName;
        }
        const names[] = {
            { AttribSpec::Position,       "aVertex"      },
            { AttribSpec::TexCoord0,      "aUV"          },
            { AttribSpec::TexCoord1,      "aUV2"         },
            { AttribSpec::TexCoord2,      "aUV3"         },
            { AttribSpec::TexCoord3,      "aUV4"         },
            { AttribSpec::TexBounds0,     "aBounds"      },
            { AttribSpec::TexBounds1,     "aBounds2"     },
            { AttribSpec::TexBounds2,     "aBounds3"     },
            { AttribSpec::TexBounds3,     "aBounds4"     },
            { AttribSpec::Color,          "aColor"       },
            { AttribSpec::Normal,         "aNormal"      },
            { AttribSpec::Tangent,        "aTangent"     },
            { AttribSpec::Bitangent,      "aBitangent"   },
            { AttribSpec::BoneIDs,        "aBoneIDs"     },
            { AttribSpec::BoneWeights,    "aBoneWeights" },
            { AttribSpec::Index,          "aIndex"       },
            { AttribSpec::InstanceMatrix, "aInstanceMatrix" }, // x4
            { AttribSpec::InstanceColor,  "aInstanceColor"  }
        };
        for (uint i = 0; i < AttribSpec::NUM_SEMANTICS; ++i) {
            attribLocation[i] = -1; // not in use
        }
        auto &GL = LIBGUI_GL;
        for (uint i = 0; i < sizeof(names)/sizeof(names[0]); ++i) {
            attribLocation[names[i].semantic] = GL.glGetAttribLocation(name, names[i].varName);
        }
    }
    Block getInfoLog() const {
        dint32 logSize;
        dint32 count;
        LIBGUI_GL.glGetProgramiv(name, GL_INFO_LOG_LENGTH, &logSize);
        Block log{Block::Size(logSize)};
        LIBGUI_GL.glGetProgramInfoLog(name, logSize, &count, reinterpret_cast<GLchar *>(log.data()));
        return log;
    }
    void link() {
        DENG2_ASSERT(name != 0);
        LIBGUI_GL.glLinkProgram(name);
        GLint ok;
        LIBGUI_GL.glGetProgramiv(name, GL_LINK_STATUS, &ok);
        if (!ok) {
            throw LinkerError("GLProgram::link", "Linking failed:\n" + getInfoLog());
        }
    }
    void markAllBoundUniformsChanged() {
        foreach (GLUniform const *u, active) {
            changed.insert(u);
        }
    }
    void updateUniforms() {
        if (changed.isEmpty()) return;
        foreach (GLUniform const *u, changed) {
            DENG2_ASSERT(active.contains(changed));
            if (!u->isSampler()) {
                u->applyInProgram(self());
            }
        }
        if (texturesChanged) {
            for (int unit = 0; unit < textures.size(); ++unit) {
                int loc = self().glUniformLocation(textures[unit]->name());
                if (loc >= 0) {
                    LIBGUI_GL.glUniform1i(loc, unit);
                    LIBGUI_ASSERT_GL_OK();
                }
            }
            texturesChanged = false;
        }
        changed.clear();
    }
    void bindTextures() {
        for (int unit = textures.size() - 1; unit >= 0; --unit) {
            GLTexture const *tex = *textures[unit];
            if (tex) {
                tex->glBindToUnit(unit);
            }
        }
    }
    void rebuild() {
        if (name) {
            LIBGUI_GL.glDeleteProgram(name);
            name = 0;
        }
        alloc();
        foreach (GLShader const *shader, shaders) {
            LIBGUI_GL.glAttachShader(name, shader->glName());
            LIBGUI_ASSERT_GL_OK();
        }
        bindVertexAttribs();
        markAllBoundUniformsChanged();
    }
    void uniformValueChanged(GLUniform &uniform) {
        if (active.contains(&uniform)) {
            changed.insert(&uniform);
        }
    }
    void uniformDeleted(GLUniform &uniform) {
        self().unbind(uniform);
    }
    void addBinding(GLUniform const *uniform) {
        allBound.insert(uniform);
        uniform->audienceForValueChange() += this;
        uniform->audienceForDeletion() += this;
        auto &stack = stacks[uniform->name()];
        if (!stack.isEmpty()) {
            active.remove(stack.top());
            changed.remove(stack.top());
        }
        stack.push(uniform);
        active.insert(uniform);
        changed.insert(uniform);
        if (uniform->isSampler()) {
            textures << uniform;
            texturesChanged = true;
        }
    }
    void removeBinding(GLUniform const *uniform) {
        allBound.remove(uniform);
        uniform->audienceForValueChange() -= this;
        uniform->audienceForDeletion() -= this;
        active.remove(uniform);
        changed.remove(uniform);
        auto &stack = stacks[uniform->name()];
        if (stack.top() == uniform) {
            stack.pop();
            if (!stack.isEmpty()) {
                active.insert(stack.top());
                changed.insert(stack.top());
            }
        }
        else {
            int found = stack.indexOf(uniform);
            if (found >= 0) {
                stack.remove(found);
            }
        }
        if (stack.isEmpty()) {
            stacks.remove(uniform->name());
        }
        if (uniform->isSampler()) {
            textures.removeAll(uniform);
            texturesChanged = true;
        }
    }
};
GLProgram::GLProgram() : d(new Impl(this))
{}
void GLProgram::clear() {
    d->release();
}
GLProgram &GLProgram::build(GLShader const *vertexShader, GLShader const *fragmentShader) {
    DENG2_ASSERT(vertexShader != 0);
    DENG2_ASSERT(vertexShader->isReady());
    DENG2_ASSERT(vertexShader->type() == GLShader::Vertex);
    DENG2_ASSERT(fragmentShader != 0);
    DENG2_ASSERT(fragmentShader->isReady());
    DENG2_ASSERT(fragmentShader->type() == GLShader::Fragment);
    d->releaseButRetainBindings();
    d->attach(vertexShader);
    d->attach(fragmentShader);
    d->bindVertexAttribs();
    d->markAllBoundUniformsChanged();
    setState(Ready);
    return *this;
}
GLProgram &GLProgram::build(IByteArray const &vertexShaderSource,
                            IByteArray const &fragmentShaderSource) {
    return build(refless(new GLShader(GLShader::Vertex,   vertexShaderSource)),
                 refless(new GLShader(GLShader::Fragment, fragmentShaderSource)));
}
void GLProgram::rebuildBeforeNextUse() {
    d->needRebuild = true;
}
void GLProgram::rebuild() {
    d->rebuild();
}
GLProgram &GLProgram::operator << (GLUniform const &uniform) {
    return bind(uniform);
}
GLProgram &GLProgram::bind(GLUniform const &uniform) {
    if (!d->allBound.contains(&uniform)) {
        if (!isReady() || glHasUniform(uniform.name())) {
            d->addBinding(&uniform);
        }
    }
    return *this;
}
GLProgram &GLProgram::unbind(GLUniform const &uniform) {
    if (d->allBound.contains(&uniform)) {
        d->removeBinding(&uniform);
    }
    return *this;
}
void GLProgram::beginUse() const {
    LIBGUI_ASSERT_GL_OK();
    DENG2_ASSERT_IN_RENDER_THREAD();
    DENG2_ASSERT(isReady());
    DENG2_ASSERT(!d->inUse);
    LIBGUI_ASSERT_GL_CONTEXT_ACTIVE();
    if (d->needRebuild) {
        d->needRebuild = false;
        const_cast<GLProgram *>(this)->rebuild();
    }
    DENG2_ASSERT(LIBGUI_GL.glIsProgram(d->name));
    d->inUse = true;
    currentProgram = this;
    LIBGUI_GL.glUseProgram(d->name);
    LIBGUI_ASSERT_GL_OK();
    d->updateUniforms();
    d->bindTextures();
    LIBGUI_ASSERT_GL_OK();
}
void GLProgram::endUse() const {
    DENG2_ASSERT_IN_RENDER_THREAD();
    DENG2_ASSERT(d->inUse);
    d->inUse = false;
    currentProgram = 0;
    LIBGUI_GL.glUseProgram(0);
}
GLProgram const *GLProgram::programInUse() // static
{
    return currentProgram;
}
GLuint GLProgram::glName() const {
    return d->name;
}
int GLProgram::glUniformLocation(char const *uniformName) const {
    return LIBGUI_GL.glGetUniformLocation(d->name, uniformName);
}
bool GLProgram::glHasUniform(char const *uniformName) const {
    return glUniformLocation(uniformName) >= 0;
}
int GLProgram::attributeLocation(AttribSpec::Semantic semantic) const {
    DENG2_ASSERT(semantic >= 0);
    DENG2_ASSERT(semantic < AttribSpec::NUM_SEMANTICS);
    return d->attribLocation[semantic];
}
bool GLProgram::validate() const {
    LIBGUI_GL.glValidateProgram(d->name);
    int valid;
    LIBGUI_GL.glGetProgramiv(d->name, GL_VALIDATE_STATUS, &valid);
    if (valid == GL_FALSE) {
        qDebug() << "GLProgram" << d->name << this << "is not validated:";
        qDebug() << d->getInfoLog().constData();
        return false;
    }
    return true;
}
} // namespace de
namespace de {
DENG2_PIMPL(GLShader) {
    GLuint name;
    Type   type;
    Impl(Public *i) : Base(i), name(0), type(Vertex)
    {}
    ~Impl() {
        release();
    }
    void alloc() {
        if (!name) {
            name = LIBGUI_GL.glCreateShader(type == Vertex? GL_VERTEX_SHADER : GL_FRAGMENT_SHADER);
            LIBGUI_ASSERT_GL_OK();
            if (!name) {
                throw AllocError("GLShader::alloc", "Failed to create shader");
            }
        }
    }
    void release() {
        if (name) {
            LIBGUI_GL.glDeleteShader(name);
            name = 0;
        }
        self().setState(Asset::NotReady);
    }
};
GLShader::GLShader() : d(new Impl(this))
{}
GLShader::GLShader(Type shaderType, IByteArray const &source) : d(new Impl(this)) {
    try {
        compile(shaderType, source);
    }
    catch (...) {
        addRef(-1);
        throw;
    }
}
GLShader::Type GLShader::type() const {
    return d->type;
}
GLuint GLShader::glName() const {
    return d->name;
}
void GLShader::clear() {
    d->release();
}
Block GLShader::prefixToSource(Block const &source, Block const &prefix) {
    Block src = source;
    int versionPos = src.indexOf("#version ");
    if (versionPos >= 0) {
        int pos = src.indexOf('\n', versionPos);
        src.insert(pos + 1, prefix);
    }
    else {
        src = prefix + src;
    }
    return src;
}
void GLShader::compile(Type shaderType, IByteArray const &shaderSource) {
#if defined (DENG_OPENGL)
    static const Block DEFAULT_VERSION("#version 330\n");
    static const Block PREFIX("#ifndef GL_ES\n"
                              "#  define lowp\n"
                              "#  define mediump\n"
                              "#  define highp\n"
                              "#endif\n");
#else
    int const glesVer = DENG_OPENGL_ES;
    static Block const DEFAULT_VERSION(glesVer == 30? "#version 300 es\n" : "#version 100\n");
    static Block const PREFIX("\n");
#endif
    DENG2_ASSERT(shaderType == Vertex || shaderType == Fragment);
    Block preamble;
    Block source = shaderSource;
    if (!source.contains("#version")) {
        preamble = DEFAULT_VERSION;
    }
    preamble += PREFIX;
    setState(NotReady);
    d->type = shaderType;
    d->alloc();
    if (shaderType == Vertex) {
        preamble += "#define DENG_VERTEX_SHADER\n";
#if defined (DENG_OPENGL) || (defined (DENG_OPENGL_ES) && DENG_OPENGL_ES == 30)
        preamble += "#define DENG_VAR out\n"
                    "#define DENG_ATTRIB in\n";
#else
        preamble += "#define DENG_VAR varying\n"
                    "#define DENG_ATTRIB attribute\n";
#endif
    }
    else {
        preamble += "#define DENG_FRAGMENT_SHADER\n";
#if defined (DENG_OPENGL_ES)
        preamble += "precision highp float;\n"
                    "precision highp int;\n";
#endif
#if defined (DENG_OPENGL) || (defined (DENG_OPENGL_ES) && DENG_OPENGL_ES == 30)
        preamble += "#define DENG_VAR in\n"
                    "out vec4 out_FragColor;\n";
#else
        preamble += "#define DENG_VAR varying\n"
                    "#define out_FragColor gl_FragColor\n";
#endif
    }
    preamble += String::format("#define DENG_MAX_BATCH_UNIFORMS %d\n", MAX_BATCH_UNIFORMS).toLatin1();
#if defined (DENG_OPENGL) || (defined (DENG_OPENGL_ES) && DENG_OPENGL_ES == 30)
    preamble += "#define DENG_LAYOUT_LOC(x) layout(location = x)\n";
#else
    preamble += "#define DENG_LAYOUT_LOC(x)\n";
#endif
    preamble += "#line 1\n";
    const char *srcPtr[] = {preamble.constData(), source.constData()};
    LIBGUI_GL.glShaderSource(d->name, 2, srcPtr, 0);
    LIBGUI_GL.glCompileShader(d->name);
    LIBGUI_ASSERT_GL_OK();
    GLint status;
    LIBGUI_GL.glGetShaderiv(d->name, GL_COMPILE_STATUS, &status);
    if (!status) {
        dint32 logSize = 0;
        dint32 count = 0;
        LIBGUI_GL.glGetShaderiv(d->name, GL_INFO_LOG_LENGTH, &logSize);
        Block log{Block::Size(logSize)};
        LIBGUI_GL.glGetShaderInfoLog(d->name, logSize, &count, reinterpret_cast<GLchar *>(log.data()));
        throw CompilerError("GLShader::compile",
                            "Compilation of " + String(d->type == Fragment? "fragment" : "vertex") +
                            " shader failed:\n" + log);
    }
    setState(Ready);
}
} // namespace de
namespace de {
static String processIncludes(String source, String const &sourceFolderPath) {
    QRegularExpression const re("#include\\s+['\"]([^\"']+)['\"]");
    forever {
        auto found = re.match(source);
        if (!found.hasMatch()) break; // No more includes.
        String incFilePath = sourceFolderPath / found.captured(1);
        String incSource = String::fromUtf8(Block(App::rootFolder().locate<File const>(incFilePath)));
        incSource = processIncludes(incSource, incFilePath.fileNamePath());
        Rangei const capRange(found.capturedStart(), found.capturedEnd());
        source = source.substr(0, capRange.start)
               + incSource
               + source.substr(capRange.end);
    }
    return source;
}
DENG2_PIMPL(GLShaderBank) {
    struct Source : public ISource {
        struct ShaderSource {
            String source;
            enum Type {
                FilePath,
                ShaderSourceText
            };
            Type type;
            ShaderSource(String const &str = "", Type t = ShaderSourceText)
                : source(str), type(t) {}
            void convertToSourceText() {
                if (type == FilePath) {
                    source = String::fromLatin1(Block(App::rootFolder().locate<File const>(source)));
                    type = ShaderSourceText;
                }
            }
            void insertFromFile(String const &path) {
                convertToSourceText();
                source += "\n";
                Block combo = GLShader::prefixToSource(source.toLatin1(),
                        Block(App::rootFolder().locate<File const>(path)));
                source = String::fromLatin1(combo);
            }
            void insertDefinition(String const &macroName, String const &content) {
                convertToSourceText();
                Block combo = GLShader::prefixToSource(
                    source.toLatin1(),
                    Block(String("#define %1 %2\n").arg(macroName).arg(content).toLatin1()));
                source = String::fromLatin1(combo);
            }
            void insertIncludes(GLShaderBank const &bank, Record const &def) {
                convertToSourceText();
                source = processIncludes(source, bank.absolutePathInContext(def, ".").fileNamePath());
            }
        };
        GLShaderBank &bank;
        String id;
        ShaderSource vertex;
        ShaderSource fragment;
        Source(GLShaderBank &b, String const &id, ShaderSource const &vtx, ShaderSource const &frag)
            : bank(b), id(id), vertex(vtx), fragment(frag)
        {}
        Time sourceModifiedAt(ShaderSource const &src) const {
            if (src.type == ShaderSource::FilePath && !src.source.isEmpty()) {
                return App::rootFolder().locate<File const>(src.source).status().modifiedAt;
            }
            return bank.sourceModifiedAt();
        }
        Time modifiedAt() const {
            Time vtxTime  = sourceModifiedAt(vertex);
            Time fragTime = sourceModifiedAt(fragment);
            return de::max(vtxTime, fragTime);
        }
        GLShader *load(GLShader::Type type) const {
            ShaderSource const &src = (type == GLShader::Vertex? vertex : fragment);
            if (src.type == ShaderSource::FilePath) {
                return bank.d->findShader(src.source, type);
            }
            return refless(
                new GLShader(type, bank.d->prependPredefines(Block(src.source.toLatin1()))));
        }
    };
    struct Data : public IData {
        GLShader *vertex;
        GLShader *fragment;
        QSet<GLUniform *> defaultUniforms;
        Data(GLShader *v, GLShader *f)
            : vertex(holdRef(v))
            , fragment(holdRef(f))
        {}
        ~Data() {
            qDeleteAll(defaultUniforms);
            releaseRef(vertex);
            releaseRef(fragment);
        }
    };
    typedef QMap<String, GLShader *> Shaders; // path -> shader
    Shaders shaders;
    std::unique_ptr<DictionaryValue> preDefines;
    Impl(Public *i) : Base(i)
    {}
    ~Impl() {
        clearShaders();
    }
    void clearShaders() {
        foreach (GLShader *shader, shaders.values()) {
            shader->release();
        }
        shaders.clear();
    }
    Block prependPredefines(const IByteArray &source) const {
        Block sourceText = source;
        if (preDefines) {
            Block predefs;
            for (auto i : preDefines->elements()) {
                predefs += String::format("#define %s %s\n",
                                          i.first.value->asText().toLatin1().constData(),
                                          i.second->asText().toLatin1().constData())
                               .toLatin1();
            }
            predefs += "#line 1\n";
            sourceText = predefs + sourceText;
        }
        return sourceText;
    }
    GLShader *findShader(String const &path, GLShader::Type type) {
        if (shaders.contains(path)) {
            return shaders[path];
        }
        GLShader *shader =
            new GLShader(type, prependPredefines(FS::locate<const ByteArrayFile>(path)));
        shaders.insert(path, shader);
        return shader;
    }
};
GLShaderBank::GLShaderBank() : InfoBank("GLShaderBank"), d(new Impl(this))
{}
void GLShaderBank::clear() {
    d->clearShaders();
    InfoBank::clear();
}
void GLShaderBank::addFromInfo(File const &file) {
    LOG_AS("GLShaderBank");
    parse(file);
    addFromInfoBlocks("shader");
}
GLShader &GLShaderBank::shader(DotPath const &path, GLShader::Type type) const {
    Impl::Data &i = data(path).as<Impl::Data>();
    if (type == GLShader::Vertex) {
        return *i.vertex;
    }
    else {
        return *i.fragment;
    }
}
GLProgram &GLShaderBank::build(GLProgram &program, DotPath const &path) const {
    Impl::Data &i = data(path).as<Impl::Data>();
    program.build(i.vertex, i.fragment);
    for (GLUniform *uniform : i.defaultUniforms) {
        program << *uniform;
    }
    return program;
}
void GLShaderBank::setPreprocessorDefines(const DictionaryValue &preDefines) {
    d->preDefines.reset(static_cast<DictionaryValue *>(preDefines.duplicate()));
}
Bank::ISource *GLShaderBank::newSourceFromInfo(String const &id) {
    typedef Impl::Source Source;
    typedef Impl::Source::ShaderSource ShaderSource;
    Record const &def = info()[id];
    ShaderSource vtx;
    ShaderSource frag;
    if (def.has("vertex")) {
        vtx = ShaderSource(def["vertex"], ShaderSource::ShaderSourceText);
    }
    else if (def.has("path.vertex")) {
        vtx = ShaderSource(absolutePathInContext(def, def["path.vertex"]), ShaderSource::FilePath);
    }
    else if (def.has("path")) {
        vtx = ShaderSource(absolutePathInContext(def, def.gets("path") + ".vsh"), ShaderSource::FilePath);
    }
    if (def.has("fragment")) {
        frag = ShaderSource(def["fragment"], ShaderSource::ShaderSourceText);
    }
    else if (def.has("path.fragment")) {
        frag = ShaderSource(absolutePathInContext(def, def["path.fragment"]), ShaderSource::FilePath);
    }
    else if (def.has("path")) {
        frag = ShaderSource(absolutePathInContext(def, def.gets("path") + ".fsh"), ShaderSource::FilePath);
    }
    if (def.has("include.vertex")) {
        auto const &incs = def["include.vertex"].value<ArrayValue>().elements();
        for (int i = incs.size() - 1; i >= 0; --i) {
            vtx.insertFromFile(absolutePathInContext(def, incs.at(i)->asText()));
        }
    }
    if (def.has("include.fragment")) {
        auto const &incs = def["include.fragment"].value<ArrayValue>().elements();
        for (int i = incs.size() - 1; i >= 0; --i) {
            frag.insertFromFile(absolutePathInContext(def, incs.at(i)->asText()));
        }
    }
    vtx .insertIncludes(*this, def);
    frag.insertIncludes(*this, def);
    if (def.has("defines")) {
        const DictionaryValue &dict = def.getdt("defines");
        for (auto i : dict.elements()) {
            String const macroName = i.first.value->asText();
            String const content   = i.second->asText();
            vtx .insertDefinition(macroName, content);
            frag.insertDefinition(macroName, content);
        }
    }
    return new Source(*this, id, vtx, frag);
}
Bank::IData *GLShaderBank::loadFromSource(ISource &source) {
    Impl::Source &src = source.as<Impl::Source>();
    std::unique_ptr<Impl::Data> data(new Impl::Data(src.load(GLShader::Vertex),
                                                            src.load(GLShader::Fragment)));
    Record const &def = info()[src.id];
    auto const vars = ScriptedInfo::subrecordsOfType(QStringLiteral("variable"), def);
    for (auto i = vars.begin(); i != vars.end(); ++i) {
        std::unique_ptr<GLUniform> uniform;
        Block const uName = i.key().toLatin1();
        if (!i.value()->has(QStringLiteral("value"))) continue;
        Value const &valueDef = i.value()->get(QStringLiteral("value"));
        if (auto const *array = maybeAs<ArrayValue>(valueDef)) {
            switch (array->size()) {
            default:
                throw DefinitionError("GLShaderBank::loadFromSource",
                                      QString("%1: Invalid initial value size (%2) for shader variable")
                                      .arg(ScriptedInfo::sourceLocation(*i.value()))
                                      .arg(array->size()));
            case 1:
                uniform.reset(new GLUniform(uName, GLUniform::Float));
                *uniform = array->element(0).asNumber();
                break;
            case 2:
                uniform.reset(new GLUniform(uName, GLUniform::Vec2));
                *uniform = vectorFromValue<Vector2f>(*array);
                break;
            case 3:
                uniform.reset(new GLUniform(uName, GLUniform::Vec3));
                *uniform = vectorFromValue<Vector3f>(*array);
                break;
            case 4:
                uniform.reset(new GLUniform(uName, GLUniform::Vec4));
                *uniform = vectorFromValue<Vector4f>(*array);
                break;
            }
        }
        else {
            uniform.reset(new GLUniform(uName, GLUniform::Float));
            *uniform = valueDef.asNumber();
        }
        data->defaultUniforms.insert(uniform.release());
    }
    return data.release();
}
} // namespace de
namespace de {
#ifdef DENG2_DEBUG
extern int GLDrawQueue_queuedElems;
#endif
namespace internal {
    enum Property {
        CullMode,
        DepthTest,
        DepthFunc,
        DepthWrite,
        AlphaTest,
        AlphaLimit,
        Blend,
        BlendFuncSrc,
        BlendFuncDest,
        BlendOp,
        ColorMask,
        Scissor,
        ScissorX,
        ScissorY,
        ScissorWidth,
        ScissorHeight,
        ViewportX,
        ViewportY,
        ViewportWidth,
        ViewportHeight,
        MAX_PROPERTIES
    };
    static BitField::Spec const propSpecs[MAX_PROPERTIES] = {
        { CullMode,       2  },
        { DepthTest,      1  },
        { DepthFunc,      3  },
        { DepthWrite,     1  },
        { AlphaTest,      1  },
        { AlphaLimit,     8  },
        { Blend,          1  },
        { BlendFuncSrc,   4  },
        { BlendFuncDest,  4  },
        { BlendOp,        2  },
        { ColorMask,      4  },
        { Scissor,        1  },
        { ScissorX,       13 }, // 13 bits == 8192 max
        { ScissorY,       13 },
        { ScissorWidth,   13 },
        { ScissorHeight,  13 },
        { ViewportX,      13 },
        { ViewportY,      13 },
        { ViewportWidth,  13 },
        { ViewportHeight, 13 }
    };
    static BitField::Elements const glStateProperties(propSpecs, MAX_PROPERTIES);
    struct GLStateStack : public QList<GLState *> {
        GLStateStack() {
            append(new GLState);
        }
        ~GLStateStack() { qDeleteAll(*this); }
    };
    static GLStateStack stack;
    static BitField currentProps;
    class CurrentTarget : DENG2_OBSERVES(Asset, Deletion) {
        GLFramebuffer *_target;
        void assetBeingDeleted(Asset &asset) {
            if (&asset == _target) {
                LOG_AS("GLState");
                LOGDEV_GL_NOTE("Current target destroyed, clearing pointer");
                _target = 0;
            }
        }
    public:
        CurrentTarget() : _target(0) {}
        ~CurrentTarget() {
            set(0);
        }
        void set(GLFramebuffer *trg) {
            if (_target) {
                _target->audienceForDeletion() -= this;
            }
            _target = trg;
            if (_target) {
                _target->audienceForDeletion() += this;
            }
        }
        CurrentTarget &operator = (GLFramebuffer *trg) {
            set(trg);
            return *this;
        }
        bool operator != (GLFramebuffer *trg) const {
            return _target != trg;
        }
        GLFramebuffer *get() const {
            return _target;
        }
        operator GLFramebuffer *() const {
            return _target;
        }
    };
    static CurrentTarget currentTarget;
}
DENG2_PIMPL(GLState) {
    BitField       props;
    GLFramebuffer *target;
    Impl(Public *i)
        : Base(i)
        , props(internal::glStateProperties)
        , target(0)
    {}
    Impl(Public *i, Impl const &other)
        : Base(i)
        , props(other.props)
        , target(other.target)
    {}
    static GLenum glComp(gl::Comparison comp) {
        switch (comp) {
        case gl::Never:          return GL_NEVER;
        case gl::Always:         return GL_ALWAYS;
        case gl::Equal:          return GL_EQUAL;
        case gl::NotEqual:       return GL_NOTEQUAL;
        case gl::Less:           return GL_LESS;
        case gl::Greater:        return GL_GREATER;
        case gl::LessOrEqual:    return GL_LEQUAL;
        case gl::GreaterOrEqual: return GL_GEQUAL;
        }
        return GL_NEVER;
    }
    static GLenum glBFunc(gl::Blend f) {
        switch (f) {
        case gl::Zero:              return GL_ZERO;
        case gl::One:               return GL_ONE;
        case gl::SrcColor:          return GL_SRC_COLOR;
        case gl::OneMinusSrcColor:  return GL_ONE_MINUS_SRC_COLOR;
        case gl::SrcAlpha:          return GL_SRC_ALPHA;
        case gl::OneMinusSrcAlpha:  return GL_ONE_MINUS_SRC_ALPHA;
        case gl::DestColor:         return GL_DST_COLOR;
        case gl::OneMinusDestColor: return GL_ONE_MINUS_DST_COLOR;
        case gl::DestAlpha:         return GL_DST_ALPHA;
        case gl::OneMinusDestAlpha: return GL_ONE_MINUS_DST_ALPHA;
        }
        return GL_ZERO;
    }
    static gl::Blend fromGlBFunc(GLenum e) {
        switch (e) {
        case GL_ZERO: return gl::Zero;
        case GL_ONE: return gl::One;
        case GL_SRC_COLOR: return gl::SrcColor;
        case GL_ONE_MINUS_SRC_COLOR: return gl::OneMinusSrcColor;
        case GL_SRC_ALPHA: return gl::SrcAlpha;
        case GL_ONE_MINUS_SRC_ALPHA: return gl::OneMinusSrcAlpha;
        case GL_DST_COLOR: return gl::DestColor;
        case GL_ONE_MINUS_DST_COLOR: return gl::OneMinusDestColor;
        case GL_DST_ALPHA: return gl::DestAlpha;
        case GL_ONE_MINUS_DST_ALPHA: return gl::OneMinusDestAlpha;
        }
        return gl::Zero;
    }
    void glApply(internal::Property prop) {
        switch (prop) {
        case internal::CullMode:
            switch (self().cull()) {
            case gl::None:
                LIBGUI_GL.glDisable(GL_CULL_FACE);
                break;
            case gl::Front:
                LIBGUI_GL.glEnable(GL_CULL_FACE);
                LIBGUI_GL.glCullFace(GL_FRONT);
                break;
            case gl::Back:
                LIBGUI_GL.glEnable(GL_CULL_FACE);
                LIBGUI_GL.glCullFace(GL_BACK);
                break;
            }
            break;
        case internal::DepthTest:
            if (self().depthTest())
                LIBGUI_GL.glEnable(GL_DEPTH_TEST);
            else
                LIBGUI_GL.glDisable(GL_DEPTH_TEST);
            break;
        case internal::DepthFunc:
            LIBGUI_GL.glDepthFunc(glComp(self().depthFunc()));
            break;
        case internal::DepthWrite:
            if (self().depthWrite())
                LIBGUI_GL.glDepthMask(GL_TRUE);
            else
                LIBGUI_GL.glDepthMask(GL_FALSE);
            break;
        case internal::AlphaTest:
            /*if (self().alphaTest())
                LIBGUI_GL.glEnable(GL_ALPHA_TEST);
            else
                LIBGUI_GL.glDisable(GL_ALPHA_TEST);*/
            break;
        case internal::AlphaLimit:
            break;
        case internal::Blend:
            if (self().blend())
                LIBGUI_GL.glEnable(GL_BLEND);
            else
                LIBGUI_GL.glDisable(GL_BLEND);
            break;
        case internal::BlendFuncSrc:
        case internal::BlendFuncDest:
            LIBGUI_GL.glBlendFuncSeparate(glBFunc(self().srcBlendFunc()), glBFunc(self().destBlendFunc()),
                                          GL_ONE, GL_ONE);
            break;
        case internal::BlendOp:
            switch (self().blendOp()) {
            case gl::Add:
                LIBGUI_GL.glBlendEquation(GL_FUNC_ADD);
                break;
            case gl::Subtract:
                LIBGUI_GL.glBlendEquation(GL_FUNC_SUBTRACT);
                break;
            case gl::ReverseSubtract:
                LIBGUI_GL.glBlendEquation(GL_FUNC_REVERSE_SUBTRACT);
                break;
            }
            break;
        case internal::ColorMask: {
            gl::ColorMask const mask = self().colorMask();
            LIBGUI_GL.glColorMask((mask & gl::WriteRed)   != 0,
                                  (mask & gl::WriteGreen) != 0,
                                  (mask & gl::WriteBlue)  != 0,
                                  (mask & gl::WriteAlpha) != 0);
            break;
        }
        case internal::Scissor:
        case internal::ScissorX:
        case internal::ScissorY:
        case internal::ScissorWidth:
        case internal::ScissorHeight: {
            if (self().scissor() || self().target().hasActiveRect()) {
                LIBGUI_GL.glEnable(GL_SCISSOR_TEST);
                Rectangleui origScr;
                if (self().scissor()) {
                    origScr = self().scissorRect();
                }
                else {
                    origScr = Rectangleui::fromSize(self().target().size());
                }
                Rectangleui const scr = self().target().scaleToActiveRect(origScr);
                LIBGUI_GL.glScissor(scr.left(), self().target().size().y - scr.bottom(),
                                    scr.width(), scr.height());
            }
            else {
                LIBGUI_GL.glDisable(GL_SCISSOR_TEST);
            }
            break;
        }
        case internal::ViewportX:
        case internal::ViewportY:
        case internal::ViewportWidth:
        case internal::ViewportHeight: {
            Rectangleui const vp = self().target().scaleToActiveRect(self().viewport());
            LIBGUI_GL.glViewport(vp.left(), self().target().size().y - vp.bottom(),
                                 vp.width(), vp.height());
            break;
        }
        default:
            break;
        }
        LIBGUI_ASSERT_GL_OK();
    }
    void removeRedundancies(BitField::Ids &changed) {
        if (changed.contains(internal::BlendFuncSrc) && changed.contains(internal::BlendFuncDest)) {
            changed.remove(internal::BlendFuncDest);
        }
        if (changed.contains(internal::ScissorX) || changed.contains(internal::ScissorY) ||
            changed.contains(internal::ScissorWidth) || changed.contains(internal::ScissorHeight)) {
            changed.insert(internal::ScissorX);
            changed.remove(internal::ScissorY);
            changed.remove(internal::ScissorWidth);
            changed.remove(internal::ScissorHeight);
        }
        if (changed.contains(internal::ViewportX) || changed.contains(internal::ViewportY) ||
            changed.contains(internal::ViewportWidth) || changed.contains(internal::ViewportHeight)) {
            changed.insert(internal::ViewportX);
            changed.remove(internal::ViewportY);
            changed.remove(internal::ViewportWidth);
            changed.remove(internal::ViewportHeight);
        }
    }
};
GLState::GLState() : d(new Impl(this)) {
    setCull      (gl::None);
    setDepthTest (false);
    setDepthFunc (gl::Less);
    setDepthWrite(true);
    setAlphaTest (true);
    setAlphaLimit(0);
    setBlend     (true);
    setBlendFunc (gl::One, gl::Zero);
    setBlendOp   (gl::Add);
    setColorMask (gl::WriteAll);
    setDefaultTarget();
}
GLState::GLState(GLState const &other) : d(new Impl(this, *other.d))
{}
GLState &GLState::operator=(GLState const &other) {
    d.reset(new Impl(this, *other.d));
    return *this;
}
bool GLState::operator==(const GLState &other) {
    return d->target == other.d->target && d->props == other.d->props;
}
GLState &GLState::setCull(gl::Cull mode) {
    d->props.set(internal::CullMode, duint(mode));
    return *this;
}
GLState &GLState::setDepthTest(bool enable) {
    d->props.set(internal::DepthTest, enable);
    return *this;
}
GLState &GLState::setDepthFunc(gl::Comparison func) {
    d->props.set(internal::DepthFunc, duint(func));
    return *this;
}
GLState &GLState::setDepthWrite(bool enable) {
    d->props.set(internal::DepthWrite, enable);
    return *this;
}
GLState &GLState::setAlphaTest(bool enable) {
    d->props.set(internal::AlphaTest, enable);
    return *this;
}
GLState &GLState::setAlphaLimit(float greaterThanValue) {
    d->props.set(internal::AlphaLimit, unsigned(clamp(0.f, greaterThanValue, 1.f) * 255));
    return *this;
}
GLState &GLState::setBlend(bool enable) {
    d->props.set(internal::Blend, enable);
    return *this;
}
GLState &GLState::setBlendFunc(gl::Blend src, gl::Blend dest) {
    d->props.set(internal::BlendFuncSrc,  duint(src));
    d->props.set(internal::BlendFuncDest, duint(dest));
    return *this;
}
GLState &GLState::setBlendFunc(gl::BlendFunc func) {
    d->props.set(internal::BlendFuncSrc,  duint(func.first));
    d->props.set(internal::BlendFuncDest, duint(func.second));
    return *this;
}
GLState &GLState::setBlendOp(gl::BlendOp op) {
    d->props.set(internal::BlendOp, duint(op));
    return *this;
}
GLState &GLState::setColorMask(gl::ColorMask mask) {
    d->props.set(internal::ColorMask, duint(mask));
    return *this;
}
GLState &GLState::setTarget(GLFramebuffer &target) {
    d->target = &target;
    return *this;
}
GLState &GLState::setDefaultTarget() {
    d->target = 0;
    return *this;
}
GLState &GLState::setViewport(Rectanglei const &viewportRect) {
    return setViewport(viewportRect.toRectangleui());
}
GLState &GLState::setViewport(Rectangleui const &viewportRect) {
    d->props.set(internal::ViewportX,      viewportRect.left());
    d->props.set(internal::ViewportY,      viewportRect.top());
    d->props.set(internal::ViewportWidth,  viewportRect.width());
    d->props.set(internal::ViewportHeight, viewportRect.height());
    return *this;
}
GLState &GLState::setNormalizedViewport(Rectanglef const &normViewportRect) {
    GLFramebuffer::Size const size = target().size();
    Rectangleui vp(Vector2ui(normViewportRect.left() * size.x,
                             normViewportRect.top()  * size.y),
                   Vector2ui(std::ceil(normViewportRect.right()  * size.x),
                             std::ceil(normViewportRect.bottom() * size.y)));
    return setViewport(vp);
}
GLState &GLState::setScissor(Rectanglei const &scissorRect) {
    return setScissor(scissorRect.toRectangleui());
}
GLState &GLState::setScissor(Rectangleui const &newScissorRect) {
    Rectangleui cumulative;
    if (scissor()) {
        cumulative = scissorRect() & newScissorRect;
    }
    else {
        cumulative = newScissorRect;
    }
    d->props.set(internal::Scissor,       true);
    d->props.set(internal::ScissorX,      cumulative.left());
    d->props.set(internal::ScissorY,      cumulative.top());
    d->props.set(internal::ScissorWidth,  cumulative.width());
    d->props.set(internal::ScissorHeight, cumulative.height());
    return *this;
}
GLState &GLState::setNormalizedScissor(Rectanglef const &normScissorRect) {
    Rectangleui vp = viewport();
    Rectanglei scis(Vector2i(normScissorRect.left()   * vp.width(),
                             normScissorRect.top()    * vp.height()),
                    Vector2i(std::ceil(normScissorRect.right()  * vp.width()),
                             std::ceil(normScissorRect.bottom() * vp.height())));
    return setScissor(scis.moved(vp.topLeft.toVector2i()));
}
GLState &GLState::clearScissor() {
    d->props.set(internal::Scissor,       false);
    d->props.set(internal::ScissorX,      0u);
    d->props.set(internal::ScissorY,      0u);
    d->props.set(internal::ScissorWidth,  0u);
    d->props.set(internal::ScissorHeight, 0u);
    return *this;
}
gl::Cull GLState::cull() const {
    return d->props.valueAs<gl::Cull>(internal::CullMode);
}
bool GLState::depthTest() const {
    return d->props.asBool(internal::DepthTest);
}
gl::Comparison GLState::depthFunc() const {
    return d->props.valueAs<gl::Comparison>(internal::DepthFunc);
}
bool GLState::depthWrite() const {
    return d->props.asBool(internal::DepthWrite);
}
bool GLState::alphaTest() const {
    return d->props.asBool(internal::AlphaTest);
}
float GLState::alphaLimit() const {
    return float(d->props.asUInt(internal::AlphaLimit)) / 255.f;
}
bool GLState::blend() const {
    return d->props.asBool(internal::Blend);
}
gl::Blend GLState::srcBlendFunc() const {
    return d->props.valueAs<gl::Blend>(internal::BlendFuncSrc);
}
gl::Blend GLState::destBlendFunc() const {
    return d->props.valueAs<gl::Blend>(internal::BlendFuncDest);
}
gl::BlendFunc GLState::blendFunc() const {
    return gl::BlendFunc(srcBlendFunc(), destBlendFunc());
}
gl::BlendOp GLState::blendOp() const {
    return d->props.valueAs<gl::BlendOp>(internal::BlendOp);
}
gl::ColorMask GLState::colorMask() const {
    return d->props.valueAs<gl::ColorMask>(internal::ColorMask);
}
GLFramebuffer &GLState::target() const {
    if (d->target) {
        return *d->target;
    }
    return GLWindow::main().framebuffer();
}
Rectangleui GLState::viewport() const {
    return Rectangleui(d->props[internal::ViewportX],
                       d->props[internal::ViewportY],
                       d->props[internal::ViewportWidth],
                       d->props[internal::ViewportHeight]);
}
Rectanglef GLState::normalizedViewport() const {
    GLFramebuffer::Size const size = target().size();
    Rectangleui const vp = viewport();
    return Rectanglef(float(vp.left())   / float(size.x),
                      float(vp.top())    / float(size.y),
                      float(vp.width())  / float(size.x),
                      float(vp.height()) / float(size.y));
}
bool GLState::scissor() const {
    return d->props.asBool(internal::Scissor);
}
Rectangleui GLState::scissorRect() const {
    return Rectangleui(d->props[internal::ScissorX],
                       d->props[internal::ScissorY],
                       d->props[internal::ScissorWidth],
                       d->props[internal::ScissorHeight]);
}
void GLState::apply() const {
    LIBGUI_ASSERT_GL_OK();
#ifdef DENG2_DEBUG
    DENG2_ASSERT(GLDrawQueue_queuedElems == 0);
#endif
    DENG2_ASSERT_IN_RENDER_THREAD();
    bool forceViewportAndScissor = false;
    GLFramebuffer *newTarget = &target();
    DENG2_ASSERT(newTarget != 0);
    if (internal::currentTarget != newTarget) {
        GLFramebuffer const *oldTarget = internal::currentTarget;
        if (oldTarget) {
            oldTarget->glRelease();
        }
        internal::currentTarget = newTarget;
        newTarget->glBind();
        if ((oldTarget && oldTarget->hasActiveRect()) || newTarget->hasActiveRect()) {
            forceViewportAndScissor = true;
        }
    }
    LIBGUI_ASSERT_GL_OK();
    BitField::Ids changed;
    if (internal::currentProps.isEmpty()) {
        changed = d->props.elements().ids();
    }
    else {
        changed = d->props.delta(internal::currentProps);
        if (forceViewportAndScissor) {
            changed.insert(internal::ViewportX);
            changed.insert(internal::ScissorX);
        }
    }
    if (!changed.isEmpty()) {
        d->removeRedundancies(changed);
        foreach (BitField::Id id, changed) {
            d->glApply(internal::Property(id));
        }
        internal::currentProps = d->props;
    }
#if 0
    for (int i = 0; i < d->props.elements().size(); ++i) {
        auto const &elem = d->props.elements().at(i);
        int val;
        switch (elem.id) {
        case internal::Blend:
            LIBGUI_GL.glGetIntegerv(GL_BLEND, &val);
            DENG2_ASSERT(!val == !d->props.asBool(elem.id));
            break;
        case internal::BlendFuncSrc:
            LIBGUI_GL.glGetIntegerv(GL_BLEND_SRC_RGB, &val);
            DENG2_ASSERT(d->fromGlBFunc(val) == d->props.asUInt(elem.id));
            break;
        case internal::BlendFuncDest:
            LIBGUI_GL.glGetIntegerv(GL_BLEND_DST_RGB, &val);
            DENG2_ASSERT(d->fromGlBFunc(val) == d->props.asUInt(elem.id));
            break;
        case internal::BlendOp:
            LIBGUI_GL.glGetIntegerv(GL_BLEND_EQUATION_RGB, &val);
            val = (val == GL_FUNC_ADD? gl::Add :
                   val == GL_FUNC_SUBTRACT? gl::Subtract :
                   val == GL_FUNC_REVERSE_SUBTRACT? gl::ReverseSubtract : 0);
            DENG2_ASSERT(val == d->props.asUInt(elem.id));
            break;
        }
    }
#endif
}
void GLState::considerNativeStateUndefined() {
    internal::currentProps.clear();
    internal::currentTarget = 0;
}
GLState &GLState::current() {
    DENG2_ASSERT(!internal::stack.isEmpty());
    return *internal::stack.last();
}
GLState &GLState::push() {
    push(new GLState(current()));
    return current();
}
GLState &GLState::pop() {
    delete take();
    return current();
}
void GLState::push(GLState *state) {
    internal::stack.append(state);
}
GLState *GLState::take() {
    DENG2_ASSERT(internal::stack.size() > 1);
    return internal::stack.takeLast();
}
dsize GLState::stackDepth() {
    return internal::stack.size();
}
} // namespace de
namespace de {
DENG2_PIMPL_NOREF(GLSubBuffer) {
    GLBuffer *            host = nullptr;
    Rangeui16             hostRange;
    dsize                 size = 0;
    internal::AttribSpecs format;
    int                   batchIndexOffset = -1;
};
GLSubBuffer::GLSubBuffer(Rangeui16 const &hostRange, GLBuffer &hostBuffer)
    : d(new Impl) {
    d->host      = &hostBuffer;
    d->hostRange = hostRange;
}
dsize GLSubBuffer::size() const {
    return d->size;
}
void GLSubBuffer::clear() {
    d->size = 0;
}
void GLSubBuffer::setVertices(dsize count, void const *data) {
    DENG2_ASSERT(count <= d->hostRange.size());
    dsize const elementSize = d->format.first->stride;
    d->host->setData(elementSize * d->hostRange.start, data,
                     elementSize * count);
    d->size = count;
}
void GLSubBuffer::setBatchVertices(int batchIndex, dsize elementCount, void *data) {
    DENG2_ASSERT(d->batchIndexOffset >= 0);
    dsize const elementSize = d->format.first->stride;
    duint8 *elems = reinterpret_cast<duint8 *>(data);
    for (dsize i = 0; i < elementCount; ++i) {
        *reinterpret_cast<float *>(elems + d->batchIndexOffset + i * elementSize)
                = float(batchIndex);
    }
    setVertices(elementCount, data);
}
GLBuffer &GLSubBuffer::hostBuffer() const {
    return *d->host;
}
void GLSubBuffer::setHostBuffer(GLBuffer &hostBuffer) {
    d->host = &hostBuffer;
}
Rangeui16 const &GLSubBuffer::hostRange() const {
    return d->hostRange;
}
void GLSubBuffer::setFormat(internal::AttribSpecs const &format) {
    d->format = format;
    d->host->setFormat(format);
    auto const *specs = format.first;
    for (dsize i = 0; i < format.second; ++i) {
        if (specs[i].semantic == internal::AttribSpec::Index) {
            d->batchIndexOffset = int(specs[i].startOffset);
        }
    }
}
void GLSubBuffer::setHostRange(Rangeui16 const &range) {
    d->hostRange = range;
}
} // namespace de
namespace de {
namespace internal {
    enum TextureFlag {
        AutoMips        = 0x1,
        MipmapAvailable = 0x2,
        ParamsChanged   = 0x4
    };
    Q_DECLARE_FLAGS(TextureFlags, TextureFlag)
}
Q_DECLARE_OPERATORS_FOR_FLAGS(internal::TextureFlags)
using namespace internal;
using namespace gl;
DENG2_PIMPL(GLTexture) {
    Size          size;
    Image::Format format;
    GLuint        name;
    GLenum        texTarget;
    Filter        minFilter;
    Filter        magFilter;
    MipFilter     mipFilter;
    Wraps         wrap;
    dfloat        maxAnisotropy;
    dfloat        maxLevel;
    TextureFlags  flags;
    Impl(Public *i)
        : Base(i)
        , format(Image::Unknown)
        , name(0)
        , texTarget(GL_TEXTURE_2D)
        , minFilter(Linear), magFilter(Linear), mipFilter(MipNone)
        , wrap(Wraps(Repeat, Repeat))
        , maxAnisotropy(1.0f)
        , maxLevel(1000.f)
        , flags(ParamsChanged)
    {}
    ~Impl() {
        release();
    }
    void alloc() {
        if (!name) {
            LIBGUI_GL.glGenTextures(1, &name);
        }
    }
    void release() {
        if (name) {
            LIBGUI_GL.glDeleteTextures(1, &name);
            name = 0;
        }
    }
    void clear() {
        release();
        size = Size();
        texTarget = GL_TEXTURE_2D;
        flags |= ParamsChanged;
        self().setState(NotReady);
    }
    bool isCube() const {
        return texTarget == GL_TEXTURE_CUBE_MAP;
    }
    static GLenum glWrap(gl::Wrapping w) {
        switch (w) {
        case Repeat:         return GL_REPEAT;
        case RepeatMirrored: return GL_MIRRORED_REPEAT;
        case ClampToEdge:    return GL_CLAMP_TO_EDGE;
        }
        return GL_REPEAT;
    }
    static GLenum glMinFilter(gl::Filter min, gl::MipFilter mip) {
        if (mip == MipNone) {
            if (min == Nearest) return GL_NEAREST;
            if (min == Linear)  return GL_LINEAR;
        }
        else if (mip == MipNearest) {
            if (min == Nearest) return GL_NEAREST_MIPMAP_NEAREST;
            if (min == Linear)  return GL_LINEAR_MIPMAP_NEAREST;
        }
        else // MipLinear
        {
            if (min == Nearest) return GL_NEAREST_MIPMAP_LINEAR;
            if (min == Linear)  return GL_LINEAR_MIPMAP_LINEAR;
        }
        return GL_NEAREST;
    }
    static GLenum glFace(gl::CubeFace face) {
        switch (face) {
        case PositiveX: return GL_TEXTURE_CUBE_MAP_POSITIVE_X;
        case PositiveY: return GL_TEXTURE_CUBE_MAP_POSITIVE_Y;
        case PositiveZ: return GL_TEXTURE_CUBE_MAP_POSITIVE_Z;
        case NegativeX: return GL_TEXTURE_CUBE_MAP_NEGATIVE_X;
        case NegativeY: return GL_TEXTURE_CUBE_MAP_NEGATIVE_Y;
        case NegativeZ: return GL_TEXTURE_CUBE_MAP_NEGATIVE_Z;
        }
        return GL_TEXTURE_CUBE_MAP_POSITIVE_X;
    }
    void glBind() const {
        LIBGUI_GL.glBindTexture(texTarget, name); LIBGUI_ASSERT_GL_OK();
    }
    void glUnbind() const {
        LIBGUI_GL.glBindTexture(texTarget, 0);
    }
    /**
     * Update the OpenGL texture parameters. You must bind the texture before
     * calling.
     */
    void glUpdateParamsOfBoundTexture() {
        LIBGUI_GL.glTexParameteri(texTarget, GL_TEXTURE_WRAP_S,     glWrap(wrap.x));
        LIBGUI_GL.glTexParameteri(texTarget, GL_TEXTURE_WRAP_T,     glWrap(wrap.y));
        LIBGUI_GL.glTexParameteri(texTarget, GL_TEXTURE_MAG_FILTER, magFilter == Nearest? GL_NEAREST : GL_LINEAR);
        LIBGUI_GL.glTexParameteri(texTarget, GL_TEXTURE_MIN_FILTER, glMinFilter(minFilter, mipFilter));
        LIBGUI_GL.glTexParameterf(texTarget, GL_TEXTURE_MAX_LEVEL,  maxLevel);
        if (GLInfo::extensions().EXT_texture_filter_anisotropic) {
            LIBGUI_GL.glTexParameterf(texTarget, GL_TEXTURE_MAX_ANISOTROPY_EXT, maxAnisotropy);
        }
        LIBGUI_ASSERT_GL_OK();
        flags &= ~ParamsChanged;
    }
    void glImage(int level, Size const &size, GLPixelFormat const &glFormat,
                 void const *data, CubeFace face = PositiveX) {
        GLenum const internalFormat =
                (glFormat.format == GL_BGRA?          GL_RGBA :
                 glFormat.format == GL_DEPTH_STENCIL? GL_DEPTH24_STENCIL8 :
                                                      glFormat.format);
        /*qDebug() << "glTexImage2D:" << name << (isCube()? glFace(face) : texTarget)
                << level << internalFormat << size.x << size.y << 0
                << glFormat.format << glFormat.type << data;*/
        if (data) LIBGUI_GL.glPixelStorei(GL_UNPACK_ALIGNMENT, GLint(glFormat.rowAlignment));
        LIBGUI_GL.glTexImage2D(isCube()? glFace(face) : texTarget,
                     level, internalFormat, size.x, size.y, 0,
                     glFormat.format, glFormat.type, data);
        LIBGUI_ASSERT_GL_OK();
    }
    void glSubImage(int level, Vector2i const &pos, Size const &size,
                    GLPixelFormat const &glFormat, void const *data, CubeFace face = PositiveX) {
        if (data) LIBGUI_GL.glPixelStorei(GL_UNPACK_ALIGNMENT, GLint(glFormat.rowAlignment));
        LIBGUI_GL.glTexSubImage2D(isCube()? glFace(face) : texTarget,
                        level, pos.x, pos.y, size.x, size.y,
                        glFormat.format, glFormat.type, data);
        LIBGUI_ASSERT_GL_OK();
    }
    void glSubImage(int level, Rectanglei const &rect, Image const &image,
                    CubeFace face = PositiveX) {
        auto const &glFormat = image.glFormat();
        LIBGUI_GL.glPixelStorei(GL_UNPACK_ALIGNMENT,  GLint(glFormat.rowAlignment));
        LIBGUI_GL.glPixelStorei(GL_UNPACK_ROW_LENGTH, GLint(image.width()));
        int const bytesPerPixel = image.depth() / 8;
        LIBGUI_GL.glTexSubImage2D(isCube()? glFace(face) : texTarget,
                                  level, rect.left(), rect.top(), rect.width(), rect.height(),
                                  glFormat.format, glFormat.type,
                                  static_cast<dbyte const *>(image.bits()) +
                                  bytesPerPixel * rect.left() + image.stride() * rect.top());
        LIBGUI_GL.glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
        LIBGUI_ASSERT_GL_OK();
    }
};
GLTexture::GLTexture() : d(new Impl(this))
{}
GLTexture::GLTexture(GLuint existingTexture, Size const &size) : d(new Impl(this)) {
    d->size = size;
    d->name = existingTexture;
    d->flags |= ParamsChanged;
}
void GLTexture::clear() {
    d->clear();
}
void GLTexture::setMagFilter(Filter magFilter) {
    d->magFilter = magFilter;
    d->flags |= ParamsChanged;
}
void GLTexture::setMinFilter(Filter minFilter, MipFilter mipFilter) {
    d->minFilter = minFilter;
    d->mipFilter = mipFilter;
    d->flags |= ParamsChanged;
}
void GLTexture::setWrapS(Wrapping mode) {
    d->wrap.x = mode;
    d->flags |= ParamsChanged;
}
void GLTexture::setWrapT(Wrapping mode) {
    d->wrap.y = mode;
    d->flags |= ParamsChanged;
}
void GLTexture::setMaxAnisotropy(dfloat maxAnisotropy) {
    d->maxAnisotropy = maxAnisotropy;
    d->flags |= ParamsChanged;
}
void GLTexture::setMaxLevel(dfloat maxLevel) {
    d->maxLevel = maxLevel;
    d->flags |= ParamsChanged;
}
Filter GLTexture::minFilter() const {
    return d->minFilter;
}
Filter GLTexture::magFilter() const {
    return d->magFilter;
}
MipFilter GLTexture::mipFilter() const {
    return d->mipFilter;
}
Wrapping GLTexture::wrapS() const {
    return d->wrap.x;
}
Wrapping GLTexture::wrapT() const {
    return d->wrap.y;
}
GLTexture::Wraps GLTexture::wrap() const {
    return d->wrap;
}
dfloat GLTexture::maxAnisotropy() const {
    return d->maxAnisotropy;
}
dfloat GLTexture::maxLevel() const {
    return d->maxLevel;
}
bool GLTexture::isCubeMap() const {
    return d->isCube();
}
void GLTexture::setAutoGenMips(bool genMips) {
    if (genMips)
        d->flags |= AutoMips;
    else
        d->flags &= ~AutoMips;
}
bool GLTexture::autoGenMips() const {
    return d->flags.testFlag(AutoMips);
}
void GLTexture::setUndefinedImage(GLTexture::Size const &size, Image::Format format, int level) {
    d->texTarget = GL_TEXTURE_2D;
    d->size = size;
    d->format = format;
    d->alloc();
    d->glBind();
    d->glImage(level, size, Image::glFormat(format), NULL);
    d->glUnbind();
    setState(Ready);
}
void GLTexture::setUndefinedImage(CubeFace face, GLTexture::Size const &size,
                                  Image::Format format, int level) {
    d->texTarget = GL_TEXTURE_CUBE_MAP;
    d->size = size;
    d->format = format;
    d->alloc();
    d->glBind();
    d->glImage(level, size, Image::glFormat(format), NULL, face);
    d->glUnbind();
    setState(Ready);
}
void GLTexture::setUndefinedContent(Size const &size, GLPixelFormat const &glFormat, int level) {
    d->texTarget = GL_TEXTURE_2D;
    d->size = size;
    d->format = Image::Unknown;
    d->alloc();
    d->glBind();
    d->glImage(level, size, glFormat, NULL);
    d->glUnbind();
    setState(Ready);
}
void GLTexture::setUndefinedContent(CubeFace face, Size const &size, GLPixelFormat const &glFormat, int level) {
    d->texTarget = GL_TEXTURE_CUBE_MAP;
    d->size = size;
    d->format = Image::Unknown;
    d->alloc();
    d->glBind();
    d->glImage(level, size, glFormat, NULL, face);
    d->glUnbind();
    setState(Ready);
}
void GLTexture::setDepthStencilContent(Size const &size) {
    setUndefinedContent(size, GLPixelFormat(GL_DEPTH_STENCIL, GL_UNSIGNED_INT_24_8));
}
void GLTexture::setImage(Image const &image, int level) {
    d->texTarget = GL_TEXTURE_2D;
    d->size = image.size();
    d->format = image.format();
    d->alloc();
    d->glBind();
    d->glImage(level, image.size(), image.glFormat(), image.bits());
    d->glUnbind();
    if (!level && d->flags.testFlag(AutoMips)) {
        generateMipmap();
    }
    setState(Ready);
}
void GLTexture::setImage(CubeFace face, Image const &image, int level) {
    d->texTarget = GL_TEXTURE_CUBE_MAP;
    d->size = image.size();
    d->format = image.format();
    d->alloc();
    d->glBind();
    d->glImage(level, image.size(), image.glFormat(), image.bits(), face);
    d->glUnbind();
    if (!level && d->flags.testFlag(AutoMips)) {
        generateMipmap();
    }
    setState(Ready);
}
void GLTexture::setSubImage(Image const &image, Vector2i const &pos, int level) {
    d->texTarget = GL_TEXTURE_2D;
    d->alloc();
    d->glBind();
    d->glSubImage(level, pos, image.size(), image.glFormat(), image.bits());
    d->glUnbind();
    if (!level && d->flags.testFlag(AutoMips)) {
        generateMipmap();
    }
}
void GLTexture::setSubImage(Image const &image, Rectanglei const &rect, int level) {
    d->texTarget = GL_TEXTURE_2D;
    d->alloc();
    d->glBind();
    d->glSubImage(level, rect, image);
    d->glUnbind();
    if (!level && d->flags.testFlag(AutoMips)) {
        generateMipmap();
    }
}
void GLTexture::setSubImage(CubeFace face, Image const &image, Vector2i const &pos, int level) {
    d->texTarget = GL_TEXTURE_CUBE_MAP;
    d->alloc();
    d->glBind();
    d->glSubImage(level, pos, image.size(), image.glFormat(), image.bits(), face);
    d->glUnbind();
    if (!level && d->flags.testFlag(AutoMips)) {
        generateMipmap();
    }
}
void GLTexture::setSubImage(CubeFace face, Image const &image, Rectanglei const &rect, int level) {
    d->texTarget = GL_TEXTURE_CUBE_MAP;
    d->alloc();
    d->glBind();
    d->glSubImage(level, rect, image, face);
    d->glUnbind();
    if (!level && d->flags.testFlag(AutoMips)) {
        generateMipmap();
    }
}
void GLTexture::generateMipmap() {
    if (d->name) {
        d->glBind();
        LIBGUI_GL.glGenerateMipmap(d->texTarget); LIBGUI_ASSERT_GL_OK();
        d->glUnbind();
        d->flags |= MipmapAvailable;
    }
}
GLTexture::Size GLTexture::size() const {
    return d->size;
}
int GLTexture::mipLevels() const {
    if (!isReady()) return 0;
    return d->flags.testFlag(MipmapAvailable)? levelsForSize(d->size) : 1;
}
GLTexture::Size GLTexture::levelSize(int level) const {
    if (level < 0) return Size();
    return levelSize(d->size, level);
}
GLuint GLTexture::glName() const {
    return d->name;
}
void GLTexture::glBindToUnit(int unit) const {
    LIBGUI_GL.glActiveTexture(GL_TEXTURE0 + unit);
    aboutToUse();
    d->glBind();
    if (d->flags.testFlag(ParamsChanged)) {
        d->glUpdateParamsOfBoundTexture();
    }
}
void GLTexture::glApplyParameters() {
    if (d->flags.testFlag(ParamsChanged)) {
        d->glBind();
        d->glUpdateParamsOfBoundTexture();
        d->glUnbind();
    }
}
Image::Format GLTexture::imageFormat() const {
    return d->format;
}
GLTexture::Size GLTexture::maximumSize() {
    int v = 0;
    LIBGUI_GL.glGetIntegerv(GL_MAX_TEXTURE_SIZE, &v); LIBGUI_ASSERT_GL_OK();
    return Size(v, v);
}
void GLTexture::aboutToUse() const {
}
int GLTexture::levelsForSize(GLTexture::Size const &size) {
    int mipLevels = 0;
    duint w = size.x;
    duint h = size.y;
    while (w > 1 || h > 1) {
        w = de::max(1u, w >> 1);
        h = de::max(1u, h >> 1);
        mipLevels++;
    }
    return mipLevels;
}
GLTexture::Size GLTexture::levelSize(GLTexture::Size const &size0, int level) {
    Size s = size0;
    for (int i = 0; i < level; ++i) {
        s.x = de::max(1u, s.x >> 1);
        s.y = de::max(1u, s.y >> 1);
    }
    return s;
}
} // namespace de
namespace de {
DENG2_STATIC_PROPERTY(DefaultSampleCount, int)
DENG2_PIMPL(GLTextureFramebuffer)
, DENG2_OBSERVES(DefaultSampleCount, Change) {
    Image::Format colorFormat;
    Size          size;
    int           _samples; ///< don't touch directly (0 == default)
    GLTexture     color;
    GLTexture     depthStencil;
    GLFramebuffer resolvedFbo;
    Asset         texFboState;
    Impl(Public *i)
        : Base(i)
        , colorFormat(Image::RGB_888)
        , _samples(0) {
        pDefaultSampleCount.audienceForChange() += this;
    }
    ~Impl() {
        release();
    }
    int sampleCount() const {
        if (_samples <= 0) return pDefaultSampleCount;
        return _samples;
    }
    bool isMultisampled() const {
        return sampleCount() > 1;
    }
    void valueOfDefaultSampleCountChanged() {
        reconfigure();
    }
    void release() {
        color.clear();
        depthStencil.clear();
        self().deinit();
        resolvedFbo.deinit();
        texFboState.setState(NotReady);
    }
#if 0
    void configureTexturesWithFallback(GLFramebuffer &fbo) {
        fbo.configure(&color, &depthStencil);
        for (int attempt = 0; ; ++attempt) {
            String failMsg;
            try {
                switch (attempt) {
                case 0:
                    failMsg = "Texture-based framebuffer failed: %s\n"
                              "Trying again without depth/stencil texture";
                    fbo.configure(&color, &depthStencil);
                    break;
                case 1:
                    failMsg = "Color texture with unified depth/stencil renderbuffer failed: %s\n"
                              "Trying again without stencil";
                    fbo.configure(GLFramebuffer::Color, color, GLFramebuffer::DepthStencil);
                    LOG_GL_WARNING("Renderer feature unavailable: lensflare depth");
                    break;
                case 2:
                    failMsg = "Color texture with depth renderbuffer failed: %s\n"
                              "Trying again without texture buffers";
                    fbo.configure(GLFramebuffer::Color, color, GLFramebuffer::Depth);
                    LOG_GL_WARNING("Renderer features unavailable: sky mask, lensflare depth");
                    break;
                case 3:
                    failMsg = "Renderbuffer-based framebuffer failed: %s\n"
                              "Trying again without stencil";
                    fbo.configure(size, GLFramebuffer::ColorDepthStencil);
                    LOG_GL_WARNING("Renderer features unavailable: postfx, lensflare depth");
                    break;
                case 4:
                    fbo.configure(size, GLFramebuffer::ColorDepth);
                    LOG_GL_WARNING("Renderer features unavailable: postfx, sky mask, lensflare depth");
                    break;
                default:
                    break;
                }
                break; // success!
            }
            catch (GLFramebuffer::ConfigError const &er) {
                if (failMsg.isEmpty()) throw er; // Can't handle it.
                LOG_GL_NOTE(failMsg) << er.asText();
            }
        }
    }
#endif
    void reconfigure() {
        if (!texFboState.isReady() || size == Size()) return;
        LOGDEV_GL_VERBOSE("Reconfiguring framebuffer: %s ms:%i")
                << size.asText() << sampleCount();
        color.setUndefinedImage(size, colorFormat);
        color.setWrap(gl::ClampToEdge, gl::ClampToEdge);
        color.setFilter(gl::Nearest, gl::Linear, gl::MipNone);
        DENG2_ASSERT(color.isReady());
        depthStencil.setDepthStencilContent(size);
        depthStencil.setWrap(gl::ClampToEdge, gl::ClampToEdge);
        depthStencil.setFilter(gl::Nearest, gl::Nearest, gl::MipNone);
        DENG2_ASSERT(depthStencil.isReady());
        if (isMultisampled()) {
            self().configure(size, ColorDepthStencil, sampleCount());
            resolvedFbo.configure(&color, &depthStencil);
        }
        else {
            try {
                self().configure(&color, &depthStencil);
                resolvedFbo.setState(NotReady);
            }
            catch (ConfigError const &er) {
                try {
                    LOG_GL_WARNING("Using framebuffer configuration fallback 1 "
                                   "(depth & stencil will be used for rendering but "
                                   "are inaccessible in shaders): ") << er.asText();
                    self().configure(size, ColorDepthStencil);
                    resolvedFbo.configure(Color, color);
                }
                catch (ConfigError const &er) {
                    try {
                        LOG_GL_WARNING("Using framebuffer configuration fallback 2 "
                                       "(only depth used for rendering, depth & stencil "
                                       "inaccessible in shaders): ") << er.asText();
                        self().configure(Color, color, Depth);
                        resolvedFbo.setState(NotReady);
                    }
                    catch (ConfigError const &er) {
                        LOG_GL_WARNING("Using final framebuffer configuration fallback 3 "
                                       "(only depth used for rendering, depth & stencil "
                                       "inaccessible in shaders): ") << er.asText();
                        self().configure(size, ColorDepth);
                        resolvedFbo.configure(GLFramebuffer::Color, color);
                    }
                }
            }
        }
        self().clear(ColorDepthStencil);
        if (resolvedFbo.isReady()) resolvedFbo.clear(ColorDepthStencil);
        LIBGUI_ASSERT_GL_OK();
    }
    void resize(Size const &newSize) {
        if (size != newSize) {
            size = newSize;
            reconfigure();
        }
    }
};
GLTextureFramebuffer::GLTextureFramebuffer(Image::Format const &colorFormat, Size const &initialSize, int sampleCount)
    : d(new Impl(this)) {
    d->colorFormat = colorFormat;
    d->size        = initialSize;
    d->_samples    = sampleCount;
}
bool GLTextureFramebuffer::areTexturesReady() const {
    return d->texFboState.isReady();
}
void GLTextureFramebuffer::glInit() {
    if (d->texFboState.isReady()) return;
    LOG_AS("GLFramebuffer");
    d->texFboState.setState(Ready);
    d->reconfigure();
}
void GLTextureFramebuffer::glDeinit() {
    d->release();
}
void GLTextureFramebuffer::setSampleCount(int sampleCount) {
    if (!GLInfo::isFramebufferMultisamplingSupported()) {
        sampleCount = 1;
    }
    if (d->_samples != sampleCount) {
        LOG_AS("GLFramebuffer");
        d->_samples = sampleCount;
        d->reconfigure();
    }
}
void GLTextureFramebuffer::setColorFormat(Image::Format const &colorFormat) {
    if (d->colorFormat != colorFormat) {
        d->colorFormat = colorFormat;
        d->reconfigure();
    }
}
void GLTextureFramebuffer::resize(Size const &newSize) {
    d->resize(newSize);
}
void GLTextureFramebuffer::resolveSamples() {
    if (d->resolvedFbo.isReady()) {
        blit(d->resolvedFbo, ColorDepthStencil);
    }
}
GLFramebuffer &GLTextureFramebuffer::resolvedFramebuffer() {
    if (d->resolvedFbo.isReady()) {
        return d->resolvedFbo;
    }
    return *this;
}
GLTextureFramebuffer::Size GLTextureFramebuffer::size() const {
    return d->size;
}
GLTexture &GLTextureFramebuffer::colorTexture() const {
    return d->color;
}
GLTexture &GLTextureFramebuffer::depthStencilTexture() const {
    return d->depthStencil;
}
int GLTextureFramebuffer::sampleCount() const {
    return d->sampleCount();
}
GLTexture *GLTextureFramebuffer::attachedTexture(Flags const &attachment) const {
    if (d->resolvedFbo.isReady()) {
        return d->resolvedFbo.attachedTexture(attachment);
    }
    return GLFramebuffer::attachedTexture(attachment);
}
/*void GLTextureFramebuffer::drawBuffer(float opacity)
{
    d->uColor = Vector4f(1, 1, 1, opacity);
    GLState::push()
            .setCull(gl::None)
            .setDepthTest(false)
            .setDepthWrite(false);
    d->drawSwap();
    GLState::pop();
    d->uColor = Vector4f(1, 1, 1, 1);
}*/
bool GLTextureFramebuffer::setDefaultMultisampling(int sampleCount) {
    LOG_AS("GLFramebuffer");
    int const newCount = max(1, sampleCount);
    if (pDefaultSampleCount != newCount) {
        pDefaultSampleCount = newCount;
        return true;
    }
    return false;
}
int GLTextureFramebuffer::defaultMultisampling() {
    return pDefaultSampleCount;
}
} // namespace de
namespace de {
DENG2_PIMPL(GLUniform)
, DENG2_OBSERVES(Asset, Deletion) {
    Block name;
    Type type;
    union Value {
        dint     int32;
        duint    uint32;
        dfloat   float32;
        dint     *ints;
        dfloat   *floats;
        Vector2f *vec2array;
        Vector3f *vec3array;
        Vector4f *vector;
        Matrix3f *mat3;
        Matrix4f *mat4;
        GLTexture const *tex;
    } value;
    duint16 usedElemCount;
    duint16 elemCount;
    Impl(Public *i, QLatin1String const &n, Type t, duint elems)
        : Base(i)
        , name(n.latin1())
        , type(t)
        , usedElemCount(duint16(elems))
        , elemCount(duint16(elems)) {
        name.append('\0');
        DENG2_ASSERT(elemCount == 1 || (elemCount > 1 && (type == IntArray   ||
                                                          type == FloatArray ||
                                                          type == Mat4Array  ||
                                                          type == Vec4Array  ||
                                                          type == Vec3Array  ||
                                                          type == Vec2Array)));
        zap(value);
        switch (type) {
        case Vec2:
        case Vec3:
        case Vec4:
            value.vector = new Vector4f;
            break;
        case IntArray:
            value.ints = new int[elemCount];
            break;
        case FloatArray:
            value.floats = new float[elemCount];
            break;
        case Vec2Array:
            value.vec2array = new Vector2f[elemCount];
            break;
        case Vec3Array:
            value.vec3array = new Vector3f[elemCount];
            break;
        case Vec4Array:
            value.vector = new Vector4f[elemCount];
            break;
        case Mat3:
            value.mat3 = new Matrix3f;
            break;
        case Mat4:
            value.mat4 = new Matrix4f;
            break;
        case Mat4Array:
            value.mat4 = new Matrix4f[elemCount];
            break;
        default:
            break;
        }
    }
    ~Impl() {
        DENG2_FOR_PUBLIC_AUDIENCE2(Deletion, i) i->uniformDeleted(self());
        switch (type) {
        case Vec2:
        case Vec3:
        case Vec4:
            delete value.vector;
            break;
        case IntArray:
            delete [] value.ints;
            break;
        case FloatArray:
            delete [] value.floats;
            break;
        case Vec2Array:
            delete [] value.vec2array;
            break;
        case Vec3Array:
            delete [] value.vec3array;
            break;
        case Vec4Array:
            delete [] value.vector;
            break;
        case Mat3:
            delete value.mat3;
            break;
        case Mat4:
            delete value.mat4;
            break;
        case Mat4Array:
            delete [] value.mat4;
            break;
        case Sampler2D:
        case SamplerCube:
            break;
        default:
            break;
        }
    }
    /**
     * @return Returns @c true if the value changed.
     */
    template <typename Type>
    bool set(Type numValue) {
        DENG2_ASSERT(type == Int || type == UInt || type == Float);
        switch (type) {
        case Int:
            if (value.int32 != dint(numValue)) {
                value.int32 = dint(numValue);
                return true;
            }
            break;
        case UInt:
            if (value.uint32 != duint(numValue)) {
                value.uint32 = duint(numValue);
                return true;
            }
            break;
        case Float:
            if (!fequal(value.float32, dfloat(numValue))) {
                value.float32 = dfloat(numValue);
                return true;
            }
            break;
        default:
            break;
        }
        return false;
    }
    void markAsChanged() {
        DENG2_FOR_PUBLIC_AUDIENCE2(ValueChange, i) {
            i->uniformValueChanged(self());
        }
    }
    void assetBeingDeleted(Asset &asset) {
        if (self().isSampler()) {
            if (value.tex == &asset) {
                value.tex = 0;
            }
        }
    }
    DENG2_PIMPL_AUDIENCE(Deletion)
    DENG2_PIMPL_AUDIENCE(ValueChange)
};
DENG2_AUDIENCE_METHOD(GLUniform, Deletion)
DENG2_AUDIENCE_METHOD(GLUniform, ValueChange)
GLUniform::GLUniform(char const *nameInShader, Type uniformType, duint elements)
    : d(new Impl(this, QLatin1String(nameInShader), uniformType, elements))
{}
void GLUniform::setName(char const *nameInShader) {
    d->name = Block(nameInShader);
    d->name.append('\0');
}
Block GLUniform::name() const {
    return d->name;
}
GLUniform::Type GLUniform::type() const {
    return d->type;
}
bool GLUniform::isSampler() const {
    return d->type == Sampler2D || d->type == SamplerCube;
}
GLUniform &GLUniform::operator = (dint value) {
    if (d->set(value)) {
        d->markAsChanged();
    }
    return *this;
}
GLUniform &GLUniform::operator = (duint value) {
    if (d->set(value)) {
        d->markAsChanged();
    }
    return *this;
}
GLUniform &GLUniform::operator = (dfloat value) {
    if (d->set(value)) {
        d->markAsChanged();
    }
    return *this;
}
GLUniform &GLUniform::operator = (ddouble value) {
    return *this = dfloat(value);
}
GLUniform &GLUniform::operator = (Vector2f const &vec) {
    DENG2_ASSERT(d->type == Vec2);
    if (Vector2f(*d->value.vector) != vec) {
        *d->value.vector = Vector4f(vec);
        d->markAsChanged();
    }
    return *this;
}
GLUniform &GLUniform::operator = (Vector3f const &vec) {
    DENG2_ASSERT(d->type == Vec3);
    if (Vector3f(*d->value.vector) != vec) {
        *d->value.vector = vec;
        d->markAsChanged();
    }
    return *this;
}
GLUniform &GLUniform::operator = (Vector4f const &vec) {
    DENG2_ASSERT(d->type == Vec4);
    if (*d->value.vector != vec) {
        *d->value.vector = vec;
        d->markAsChanged();
    }
    return *this;
}
GLUniform &GLUniform::operator = (Matrix3f const &mat) {
    DENG2_ASSERT(d->type == Mat3);
    *d->value.mat3 = mat;
    d->markAsChanged();
    return *this;
}
GLUniform &GLUniform::operator = (Matrix4f const &mat) {
    DENG2_ASSERT(d->type == Mat4);
    *d->value.mat4 = mat;
    d->markAsChanged();
    return *this;
}
GLUniform &GLUniform::operator = (GLTexture const &texture) {
    return *this = &texture;
}
GLUniform &GLUniform::operator = (GLTexture const *texture) {
    if (texture && texture->isReady()) {
        DENG2_ASSERT(d->type != Sampler2D   || !texture->isCubeMap());
        DENG2_ASSERT(d->type != SamplerCube ||  texture->isCubeMap());
    }
    if (d->value.tex != texture) {
        if (d->value.tex) d->value.tex->audienceForDeletion() -= d;
        d->value.tex = texture;
        d->markAsChanged();
        if (d->value.tex) d->value.tex->audienceForDeletion() += d;
    }
    return *this;
}
GLUniform &GLUniform::set(duint elementIndex, dfloat value) {
    DENG2_ASSERT(d->type == FloatArray);
    DENG2_ASSERT(elementIndex < d->elemCount);
    if (!fequal(d->value.floats[elementIndex], value)) {
        d->value.floats[elementIndex] = value;
        d->markAsChanged();
    }
    return *this;
}
GLUniform &GLUniform::set(duint elementIndex, Vector3f const &vec) {
    DENG2_ASSERT(d->type == Vec3Array);
    DENG2_ASSERT(elementIndex < d->elemCount);
    if (d->value.vec3array[elementIndex] != vec) {
        d->value.vec3array[elementIndex] = vec;
        d->usedElemCount = d->elemCount;
        d->markAsChanged();
    }
    return *this;
}
GLUniform &GLUniform::set(duint elementIndex, Vector4f const &vec) {
    DENG2_ASSERT(d->type == Vec4Array);
    DENG2_ASSERT(elementIndex < d->elemCount);
    if (d->value.vector[elementIndex] != vec) {
        d->value.vector[elementIndex] = vec;
        d->usedElemCount = d->elemCount;
        d->markAsChanged();
    }
    return *this;
}
GLUniform &GLUniform::set(duint elementIndex, Matrix4f const &mat) {
    DENG2_ASSERT(d->type == Mat4Array);
    DENG2_ASSERT(elementIndex < d->elemCount);
    d->value.mat4[elementIndex] = mat;
    d->usedElemCount = d->elemCount;
    d->markAsChanged();
    return *this;
}
GLUniform &GLUniform::set(const int *intArray, dsize count) {
    DENG2_ASSERT(d->type == IntArray);
    DENG2_ASSERT(count <= d->elemCount);
    memcpy(d->value.ints, intArray, sizeof(int) * count);
    d->usedElemCount = duint16(count);
    d->markAsChanged();
    return *this;
}
GLUniform &GLUniform::set(float const *floatArray, dsize count) {
    DENG2_ASSERT(d->type == FloatArray);
    DENG2_ASSERT(count <= d->elemCount);
    memcpy(d->value.floats, floatArray, sizeof(float) * count);
    d->usedElemCount = duint16(count);
    d->markAsChanged();
    return *this;
}
GLUniform &GLUniform::set(Vector4f const *vectorArray, dsize count) {
    DENG2_ASSERT(d->type == Vec4Array);
    DENG2_ASSERT(count <= d->elemCount);
    memcpy(d->value.vector, vectorArray, sizeof(Vector4f) * count);
    d->usedElemCount = duint16(count);
    d->markAsChanged();
    return *this;
}
GLUniform &GLUniform::set(const Matrix4f *mat4Array, dsize count) {
    DENG2_ASSERT(d->type == Mat4Array);
    DENG2_ASSERT(count <= d->elemCount);
    memcpy(d->value.mat4, mat4Array, sizeof(Matrix4f) * count);
    d->usedElemCount = duint16(count);
    d->markAsChanged();
    return *this;
}
GLUniform &GLUniform::setUsedElementCount(duint elementCount) {
    DENG2_ASSERT(elementCount <= d->elemCount);
    d->usedElemCount = elementCount;
    d->markAsChanged();
    return *this;
}
dint GLUniform::toInt() const {
    DENG2_ASSERT(d->type == Int || d->type == UInt || d->type == Float);
    switch (d->type) {
    case Int:
        return d->value.int32;
    case UInt:
        return dint(d->value.uint32);
    case Float:
        return dint(d->value.float32);
    default:
        return 0;
    }
}
duint de::GLUniform::toUInt() const {
    DENG2_ASSERT(d->type == Int || d->type == UInt || d->type == Float);
    switch (d->type) {
    case Int:
        return duint(d->value.int32);
    case UInt:
        return d->value.uint32;
    case Float:
        return duint(d->value.float32);
    default:
        return 0;
    }
}
dfloat GLUniform::toFloat() const {
    DENG2_ASSERT(d->type == Int || d->type == UInt || d->type == Float);
    switch (d->type) {
    case Int:
        return dfloat(d->value.int32);
    case UInt:
        return dfloat(d->value.uint32);
    case Float:
        return d->value.float32;
    default:
        return 0;
    }
}
Vector2f const &GLUniform::toVector2f() const {
    DENG2_ASSERT(d->type == Vec2 || d->type == Vec3 || d->type == Vec4);
    return *d->value.vector;
}
Vector3f const &GLUniform::toVector3f() const {
    DENG2_ASSERT(d->type == Vec2 || d->type == Vec3 || d->type == Vec4);
    return *d->value.vector;
}
Vector4f const &GLUniform::toVector4f() const {
    DENG2_ASSERT(d->type == Vec2 || d->type == Vec3 || d->type == Vec4);
    return *d->value.vector;
}
Matrix3f const &GLUniform::toMatrix3f() const {
    DENG2_ASSERT(d->type == Mat3);
    return *d->value.mat3;
}
Matrix4f const &GLUniform::toMatrix4f() const {
    DENG2_ASSERT(d->type == Mat4);
    return *d->value.mat4;
}
GLTexture const *GLUniform::texture() const {
    DENG2_ASSERT(isSampler());
    return d->value.tex;
}
void GLUniform::applyInProgram(GLProgram &program) const {
    LIBGUI_ASSERT_GL_OK();
    int loc = program.glUniformLocation(d->name.constData());
    if (loc < 0) {
        LOG_AS("applyInProgram");
        LOGDEV_GL_WARNING("'%s' not in the program") << d->name.constData();
        return;
    }
    switch (d->type) {
    case Int:
        LIBGUI_GL.glUniform1i(loc, d->value.int32);
        break;
    case IntArray:
        LIBGUI_GL.glUniform1iv(loc, d->usedElemCount, d->value.ints);
        break;
    case UInt:
        LIBGUI_GL.glUniform1ui(loc, d->value.uint32);
        break;
    case Float:
        LIBGUI_GL.glUniform1f(loc, d->value.float32);
        break;
    case FloatArray:
        LIBGUI_GL.glUniform1fv(loc, d->usedElemCount, d->value.floats);
        break;
    case Vec2:
        LIBGUI_GL.glUniform2f(loc, d->value.vector->x, d->value.vector->y);
        break;
    case Vec2Array:
        LIBGUI_GL.glUniform2fv(loc, d->usedElemCount, &d->value.vec2array->x);
        break;
    case Vec3:
        LIBGUI_GL.glUniform3f(loc, d->value.vector->x, d->value.vector->y, d->value.vector->z);
        break;
    case Vec3Array:
        LIBGUI_GL.glUniform3fv(loc, d->usedElemCount, &d->value.vec3array->x); // sequentially laid out
        break;
    case Vec4:
    case Vec4Array:
        LIBGUI_GL.glUniform4fv(loc, d->usedElemCount, &d->value.vector->x); // sequentially laid out
        break;
    case Mat3:
        LIBGUI_GL.glUniformMatrix3fv(loc, 1, GL_FALSE, d->value.mat3->values());
        break;
    case Mat4:
    case Mat4Array:
        LIBGUI_GL.glUniformMatrix4fv(loc, d->usedElemCount, GL_FALSE, d->value.mat4->values()); // sequentially laid out
        break;
    case Sampler2D:
    case SamplerCube:
        break;
    }
#if defined (DENG2_DEBUG)
    {
        GLenum err = LIBGUI_GL.glGetError();
        if (err != GL_NO_ERROR) {
            qDebug() << "[GLUniform] Failure with uniform:" << d->name << "loc:" << loc;
        }
    }
#endif
}
} // namespace de
namespace de {
static float const NORMAL_Z = .2f;
DENG2_PIMPL_NOREF(HeightMap) {
    QImage   heightImage;
    QImage   normalImage;
    Vector2f mapSize;
    float    heightRange = 1.f;
    Vector2f pixelCoordf(Vector2f const &worldPos) const {
        Vector2f normPos = worldPos / mapSize + Vector2f(.5f, .5f);
        return normPos * Vector2f(heightImage.width(), heightImage.height())
                - Vector2f(.5f, .5f);
    }
    Vector3f normalAtCoord(Vector2i const &pos) const {
        int const w = heightImage.width();
        int const h = heightImage.height();
        int x0 = max(pos.x - 1, 0);
        int y0 = max(pos.y - 1, 0);
        int x2 = min(pos.x + 1, w - 1);
        int y2 = min(pos.y + 1, h - 1);
        float base  = qRed(heightImage.pixel(pos.x, pos.y)) / 255.f;
        float left  = qRed(heightImage.pixel(x0,    pos.y)) / 255.f;
        float right = qRed(heightImage.pixel(x2,    pos.y)) / 255.f;
        float up    = qRed(heightImage.pixel(pos.x, y0))    / 255.f;
        float down  = qRed(heightImage.pixel(pos.x, y2))    / 255.f;
        return (Vector3f(base - right, base - down, NORMAL_Z) +
                Vector3f(left - base,  up - base,   NORMAL_Z)).normalize();
    }
};
HeightMap::HeightMap() : d(new Impl)
{}
void HeightMap::setMapSize(Vector2f const &worldSize, float heightRange) {
    d->mapSize     = worldSize;
    d->heightRange = heightRange;
}
void HeightMap::loadGrayscale(Image const &heightImage) {
    d->heightImage = heightImage.toQImage();
}
Image HeightMap::toImage() const {
    return d->heightImage;
}
Image HeightMap::makeNormalMap() const {
    QImage const &heightMap = d->heightImage;
    QImage img(heightMap.size(), QImage::Format_ARGB32);
    int const w = heightMap.width();
    int const h = heightMap.height();
    for (int y = 0; y < h; y++) {
        for (int x = 0; x < w; x++) {
            Vector3f norm = d->normalAtCoord(Vector2i(x, y));
            img.setPixel(x, y, qRgba(clamp(0.f, (norm.x + 1) * 128.f, 255.f),
                                     clamp(0.f, (norm.y + 1) * 128.f, 255.f),
                                     clamp(0.f, (norm.z + 1) * 128.f, 255.f),
                                     255));
        }
    }
    d->normalImage = img;
    return img;
}
float HeightMap::heightAtPosition(Vector2f const &worldPos) const {
    QImage const &img = d->heightImage;
    Vector2f coord = d->pixelCoordf(worldPos);
    Vector2i pixelCoord = coord.toVector2i();
    if (pixelCoord.x < 0 || pixelCoord.y < 0 ||
       pixelCoord.x >= img.width() - 1 || pixelCoord.y >= img.height() - 1) return 0;
    float A = qRed(img.pixel(pixelCoord.x,     pixelCoord.y))     / 255.f - 0.5f;
    float B = qRed(img.pixel(pixelCoord.x + 1, pixelCoord.y))     / 255.f - 0.5f;
    float C = qRed(img.pixel(pixelCoord.x + 1, pixelCoord.y + 1)) / 255.f - 0.5f;
    float D = qRed(img.pixel(pixelCoord.x,     pixelCoord.y + 1)) / 255.f - 0.5f;
    Vector2f i(fract(coord.x), fract(coord.y));
    float value = A + i.x * (B - A) + i.y * (D - A) + i.x * i.y * (A - B - D + C);
    return value * -d->heightRange;
}
Vector3f HeightMap::normalAtPosition(Vector2f const &worldPos) const {
    Vector2i const pos = d->pixelCoordf(worldPos).toVector2i();
    return d->normalAtCoord(pos); // * Vector3f(1, 1, NORMAL_Z)).normalize();
}
} // namespace de
namespace de {
#define IMAGE_ASSERT_EDITABLE(d) DENG2_ASSERT(d->format == UseQImageFormat)
namespace internal {
namespace pcx {
static dbyte const MAGIC        = 0x0a;
static dbyte const RLE_ENCODING = 1;
static dsize const HEADER_SIZE  = 128;
struct Header : public IReadable {
    dbyte   magic;
    dbyte   version;
    dbyte   encoding;
    dbyte   bitsPerPixel;
    duint16 xMin, yMin;
    duint16 xMax, yMax;
    duint16 hRes, vRes;
    dbyte   colorPlanes;
    duint16 bytesPerLine;
    duint16 paletteType;
    void operator << (Reader &from) {
        from >> magic
             >> version
             >> encoding
             >> bitsPerPixel
             >> xMin
             >> yMin
             >> xMax
             >> yMax
             >> hRes
             >> vRes;
        from.seek(48);  // skip EGA palette
        from.seek(1);   // skip reserved field
        from >> colorPlanes
             >> bytesPerLine
             >> paletteType;
    }
};
static bool recognize(Block const &data) {
    try {
        Header header;
        Reader(data) >> header;
        return (header.magic == MAGIC &&
                header.version == 5 /* latest */ &&
                header.encoding == RLE_ENCODING &&
                header.bitsPerPixel == 8);
    }
    catch (Error const &) {
        return false;
    }
}
/**
 * Loads a PCX image into a QImage using an RGB888 buffer. The PCX palette is used
 * to map color indices to RGB values.
 *
 * @param data  Source data containing a PCX image.
 *
 * @return QImage using the RGB888 (24-bit) format.
 */
static QImage load(Block const &data) {
    Header header;
    Reader reader(data);
    reader >> header;
    Image::Size const size(header.xMax + 1, header.yMax + 1);
    QImage image(size.x, size.y, QImage::Format_RGB888);
    DENG2_ASSERT(image.depth() == 24);
    dbyte const *palette = data.data() + data.size() - 768;
    dbyte const *pos = data.data() + HEADER_SIZE;
    dbyte *dst = image.bits();
    for (duint y = 0; y < size.y; ++y, dst += size.x * 3) {
        for (duint x = 0; x < size.x; ) {
            dbyte value = *pos++;
            int runLength = 1;
            if ((value & 0xc0) == 0xc0) {
                runLength = value & 0x3f;
                value = *pos++;
            }
            while (runLength-- > 0) {
                std::memcpy(&dst[3 * x++], &palette[3 * value], 3);
            }
        }
    }
    return image;
}
} // namespace pcx
namespace tga {
struct Header : public IReadable {
    enum Flag {
        NoFlags           = 0,
        ScreenOriginUpper = 0x1,
        InterleaveTwoWay  = 0x2,
        InterleaveFourWay = 0x4
    };
    Q_DECLARE_FLAGS(Flags, Flag)
    enum ColorMapType {
        ColorMapNone = 0,
        ColorMap256  = 1 // not supported
    };
    enum ImageType {
        ColorMapped = 1, // Uncompressed and color-mapped.
        RGB         = 2, // Uncompressed RGB.
        RleRGB      = 10 // Run-length encoded RGB.
    };
    Block identification;
    Zeroed<duint8> colorMapType;
    Zeroed<duint8> imageType;
    Zeroed<dint16> mapIndex;
    Zeroed<dint16> mapCount;        ///< Number of color map entries.
    Zeroed<duint8> mapEntrySize;    ///< Bits in a color map entry.
    Flags flags;
    Vector2<dint16> origin;
    Vector2<dint16> size;
    Zeroed<duint8> depth;
    Zeroed<duint8> attrib;
    void operator << (Reader &from) {
        duint8 identificationSize;
        from >> identificationSize
             >> colorMapType
             >> imageType
             >> mapIndex
             >> mapCount
             >> mapEntrySize
             >> origin.x >> origin.y
             >> size.x >> size.y
             >> depth;
        duint8 f;
        from >> f;
        /*
           Flags:
             0-3 : Number of attribute bits
               4 : reserved
               5 : Screen origin in upper left corner
             6-7 : Data storage interleave
                   00 - no interleave
                   01 - even/odd interleave
                   10 - four way interleave
                   11 - reserved
        */
        attrib = f & 0x0f;
        flags = (f & 0x20? ScreenOriginUpper : NoFlags);
        if ((f & 0xc0) == 0x40) flags |= InterleaveTwoWay;
        if ((f & 0xc0) == 0x80) flags |= InterleaveFourWay;
        from.readBytes(identificationSize, identification);
    }
};
Q_DECLARE_OPERATORS_FOR_FLAGS(Header::Flags)
static bool recognize(Block const &data) {
    try {
        Header header;
        Reader(data) >> header;
        if (header.size.x == 0 || header.size.y == 0) {
            return false;
        }
        if (header.imageType == Header::ColorMapped && header.colorMapType == Header::ColorMap256 &&
            header.depth == 8) {
            return true;
        }
        return (header.imageType == Header::RGB || header.imageType == Header::RleRGB) &&
               header.colorMapType == Header::ColorMapNone &&
               (header.depth == 24 || header.depth == 32);
    }
    catch (...) {
        return false;
    }
}
static QImage load(Block const &data) {
    Header header;
    Reader input(data);
    input >> header;
    int const pixelSize = header.depth / 8;
    QImage img(QSize(header.size.x, header.size.y),
               pixelSize == 4 || header.attrib > 0? QImage::Format_ARGB32 : QImage::Format_RGB888);
    dbyte *base = img.bits();
    bool const isUpperOrigin = header.flags.testFlag(Header::ScreenOriginUpper);
    if (header.imageType == Header::RGB) {
        for (int y = 0; y < header.size.y; y++) {
            int inY = (isUpperOrigin? y : (header.size.y - y - 1));
            ByteRefArray line(base + (inY * img.bytesPerLine()), header.size.x * pixelSize);
            input.readBytesFixedSize(line);
        }
    }
    else if (header.imageType == Header::RleRGB) {
        img.fill(0);
        int x = 0;
        int y = (isUpperOrigin? 0 : (header.size.y - 1));
        int endY = header.size.y - y - 1;
        int stepY = (isUpperOrigin? 1 : -1);
        while (y != endY && x < header.size.x) {
            dbyte rle;
            input >> rle;
            int count;
            bool repeat = false;
            if (rle & 0x80) // Repeat?
            {
                repeat = true;
                count = (rle & 0x7f) + 1;
            }
            else {
                count = rle + 1;
            }
            Block pixel;
            for (int i = 0; i < count; ++i) {
                if (i == 0 || !repeat) {
                    input.readBytes(pixelSize, pixel);
                }
                std::memcpy(base + (x + y * header.size.x) * pixelSize,
                            pixel.constData(), pixelSize);
                if (++x == header.size.x) {
                    x = 0;
                    y += stepY;
                }
            }
        }
    }
    else if (header.imageType == Header::ColorMapped) {
        DENG2_ASSERT(header.colorMapType == Header::ColorMap256);
        DENG2_ASSERT(header.depth == 8);
        QVector<QRgb> colorTable(256);
        for (int i = 0; i < header.mapCount; ++i) {
            uint8_t color[4] = {0, 0, 0, 255};
            ByteRefArray buf(color, 4);
            input.readBytes(header.mapEntrySize / 8, buf);
            DENG2_ASSERT(header.mapIndex + i < 256);
            colorTable[header.mapIndex + i] = qRgba(color[2], color[1], color[0], color[3]);
        }
        img = QImage(QSize(header.size.x, header.size.y), QImage::Format_Indexed8);
        img.setColorTable(colorTable);
        dbyte *base = img.bits();
        for (int y = 0; y < header.size.y; y++) {
            int inY = (isUpperOrigin? y : (header.size.y - y - 1));
            ByteRefArray line(base + inY * img.bytesPerLine(), header.size.x);
            input.readBytesFixedSize(line);
        }
    }
    if (pixelSize >= 3) {
        img = img.rgbSwapped();
    }
    return img;
}
} // namespace tga
} // namespace internal
using namespace internal;
DENG2_PIMPL(Image) {
    Format format;
    Size size;
    QImage image;
    Block pixels;
    ByteRefArray refPixels;
    float pointRatio = 1.f;
    Impl(Public *i, QImage const &img = QImage())
        : Base(i), format(UseQImageFormat), image(img) {
        size = Size(img.width(), img.height());
    }
    Impl(Public * i, Impl const &other)
        : Base(i)
        , format(other.format)
        , size(other.size)
        , image(other.image)
        , pixels(other.pixels)
        , refPixels(other.refPixels)
        , pointRatio(other.pointRatio)
    {}
    Impl(Public *i, Size const &imgSize, Format imgFormat, IByteArray const &imgPixels)
        : Base(i), format(imgFormat), size(imgSize), pixels(imgPixels)
    {}
    Impl(Public *i, Size const &imgSize, Format imgFormat, ByteRefArray const &imgRefPixels)
        : Base(i), format(imgFormat), size(imgSize), refPixels(imgRefPixels)
    {}
};
Image::Image() : d(new Impl(this))
{}
Image::Image(Image const &other)
    : d(new Impl(this, *other.d))
{}
Image::Image(QImage const &image) : d(new Impl(this, image))
{}
Image::Image(Size const &size, Format format, IByteArray const &pixels)
    : d(new Impl(this, size, format, pixels))
{}
Image::Image(Size const &size, Format format, ByteRefArray const &refPixels)
    : d(new Impl(this, size, format, refPixels))
{}
Image &Image::operator=(Image const &other) {
    d.reset(new Impl(this, *other.d));
    return *this;
}
Image &Image::operator = (QImage const &other) {
    d.reset(new Impl(this, other));
    return *this;
}
Image::Format Image::format() const {
    return d->format;
}
QImage::Format Image::qtFormat() const {
    if (d->format == UseQImageFormat) {
        return d->image.format();
    }
    return QImage::Format_Invalid;
}
Image::Size Image::size() const {
    return d->size;
}
Rectanglei Image::rect() const {
    return Rectanglei(0, 0, d->size.x, d->size.y);
}
int Image::depth() const {
    switch (d->format) {
    case UseQImageFormat:
        return d->image.depth();
    case Luminance_8:
    case Alpha_8:
        return 8;
    case LuminanceAlpha_88:
    case RGB_565:
    case RGBA_4444:
    case RGBA_5551:
        return 16;
    case RGB_888:
        return 24;
    case RGBA_8888:
    case RGBx_8888:
        return 32;
    default:
        return 0;
    }
}
int Image::stride() const {
    if (d->format == UseQImageFormat) {
        return d->image.bytesPerLine();
    }
    return depth() / 8 * d->size.x;
}
int Image::byteCount() const {
    if (d->format == UseQImageFormat) {
        return d->image.byteCount();
    }
    if (!d->pixels.isEmpty()) {
        return d->pixels.size();
    }
    return depth() / 8 * d->size.x * d->size.y;
}
void const *Image::bits() const {
    if (d->format == UseQImageFormat) {
        return d->image.constBits();
    }
    if (!d->pixels.isEmpty()) {
        return d->pixels.constData();
    }
    return d->refPixels.readBase();
}
void *Image::bits() {
    if (d->format == UseQImageFormat) {
        return d->image.bits();
    }
    if (!d->pixels.isEmpty()) {
        return d->pixels.data();
    }
    return d->refPixels.base();
}
bool Image::isNull() const {
    return size() == Size(0, 0);
}
bool Image::isGLCompatible() const {
    if (d->format == UseQImageFormat) {
        switch (qtFormat()) {
        case QImage::Format_ARGB32: // 8888
        case QImage::Format_RGB32:  // 8888
        case QImage::Format_RGB888: // 888
        case QImage::Format_RGB16:  // 565
        case QImage::Format_RGB555: // 555
        case QImage::Format_RGB444: // 444
            return true;
        default:
            return false;
        }
    }
    return d->format >= Luminance_8 && d->format <= RGBx_8888;
}
bool Image::canConvertToQImage() const {
    switch (d->format) {
    case RGB_444:
    case RGB_555:
    case RGB_565:
    case RGB_888:
    case RGBA_8888:
    case RGBx_8888:
    case UseQImageFormat:
        return true;
    default:
        return false;
    }
}
QImage Image::toQImage() const {
    if (d->format == UseQImageFormat) {
        return d->image;
    }
    QImage::Format form = QImage::Format_Invalid;
    switch (d->format) {
    case RGB_444:
        form = QImage::Format_RGB444;
        break;
    case RGB_555:
        form = QImage::Format_RGB555;
        break;
    case RGB_565:
        form = QImage::Format_RGB16;
        break;
    case RGB_888:
        form = QImage::Format_RGB888;
        break;
    case RGBA_8888:
        form = QImage::Format_ARGB32;
        break;
    case RGBx_8888:
        form = QImage::Format_RGB32;
        break;
    default:
        return QImage();
    }
    QImage img(QSize(d->size.x, d->size.y), form);
    std::memcpy(const_cast<uchar *>(img.constBits()), bits(), byteCount());
    return img;
}
GLPixelFormat Image::glFormat() const {
    if (d->format == UseQImageFormat) {
        return glFormat(d->image.format());
    }
    return glFormat(d->format);
}
float Image::pointRatio() const {
    return d->pointRatio;
}
void Image::setPointRatio(float pointsPerPixel) {
    d->pointRatio = pointsPerPixel;
}
Image Image::subImage(Rectanglei const &subArea) const {
    IMAGE_ASSERT_EDITABLE(d);
    return Image(d->image.copy(subArea.topLeft.x, subArea.topLeft.y,
                               subArea.width(), subArea.height()));
}
void Image::resize(Size const &size) {
    IMAGE_ASSERT_EDITABLE(d);
    DENG2_ASSERT(d->image.format() != QImage::Format_Invalid);
    QImage resized(QSize(size.x, size.y), d->image.format());
    resized.fill(0);
    QPainter painter(&resized);
    painter.drawImage(QRect(QPoint(0, 0), resized.size()), d->image);
    d->image = resized;
    d->size = size;
}
void Image::fill(Color const &color) {
    IMAGE_ASSERT_EDITABLE(d);
    d->image.fill(QColor(color.x, color.y, color.z, color.w).rgba());
}
void Image::fill(Rectanglei const &rect, Color const &color) {
    IMAGE_ASSERT_EDITABLE(d);
    QPainter painter(&d->image);
    painter.setCompositionMode(QPainter::CompositionMode_Source);
    painter.fillRect(QRect(rect.topLeft.x, rect.topLeft.y, rect.width(), rect.height()),
                     QColor(color.x, color.y, color.z, color.w));
}
void Image::draw(Image const &image, Vector2i const &topLeft) {
    IMAGE_ASSERT_EDITABLE(d);
    IMAGE_ASSERT_EDITABLE(image.d);
    QPainter painter(&d->image);
    painter.drawImage(QPoint(topLeft.x, topLeft.y), image.d->image);
}
void Image::drawPartial(Image const &image, Rectanglei const &part, Vector2i const &topLeft) {
    IMAGE_ASSERT_EDITABLE(d);
    IMAGE_ASSERT_EDITABLE(image.d);
    QPainter painter(&d->image);
    painter.drawImage(QPoint(topLeft.x, topLeft.y),
                      image.d->image,
                      QRect(part.left(), part.top(), part.width(), part.height()));
}
Image Image::multiplied(Image const &factorImage) const {
    QImage multiplied = toQImage();
    QPainter painter(&multiplied);
    painter.setCompositionMode(QPainter::CompositionMode_Multiply);
    painter.drawImage(0, 0, factorImage.toQImage());
    return multiplied;
}
Image Image::multiplied(Color const &color) const {
    if (color == Color(255, 255, 255, 255)) return *this; // No change.
    QImage copy = toQImage().convertToFormat(QImage::Format_ARGB32);
    for (duint y = 0; y < height(); ++y) {
        duint32 *ptr = reinterpret_cast<duint32 *>(copy.bits() + y * copy.bytesPerLine());
        for (duint x = 0; x < width(); ++x) {
            duint16 b =  *ptr & 0xff;
            duint16 g = (*ptr & 0xff00) >> 8;
            duint16 r = (*ptr & 0xff0000) >> 16;
            duint16 a = (*ptr & 0xff000000) >> 24;
            *ptr++ = qRgba((color.x + 1) * r >> 8,
                           (color.y + 1) * g >> 8,
                           (color.z + 1) * b >> 8,
                           (color.w + 1) * a >> 8);
        }
    }
    return copy;
}
Image Image::colorized(Color const &color) const {
    QImage copy = toQImage().convertToFormat(QImage::Format_ARGB32);
    QColor targetColor(color.x, color.y, color.z, 255);
    int targetHue = targetColor.hue();
    for (duint y = 0; y < height(); ++y) {
        duint32 *ptr = reinterpret_cast<duint32 *>(copy.bits() + y * copy.bytesPerLine());
        for (duint x = 0; x < width(); ++x) {
            duint16 b =  *ptr & 0xff;
            duint16 g = (*ptr & 0xff00) >> 8;
            duint16 r = (*ptr & 0xff0000) >> 16;
            duint16 a = (*ptr & 0xff000000) >> 24;
            QColor rgba(r, g, b, a);
            QColor colorized;
            colorized.setHsv(targetHue, rgba.saturation(), rgba.value(), color.w * a >> 8);
            *ptr++ = colorized.rgba();
        }
    }
    return copy;
}
void Image::operator >> (Writer &to) const {
    to << duint8(d->format);
    if (d->format == UseQImageFormat) {
        Block block;
        QDataStream os(&block, QIODevice::WriteOnly);
        os.setVersion(QDataStream::Qt_4_8);
        os << d->image;
        to << block;
    }
    else {
        to << d->size << ByteRefArray(bits(), byteCount());
    }
}
void Image::operator << (Reader &from) {
    d->pixels.clear();
    d->refPixels = ByteRefArray();
    from.readAs<duint8>(d->format);
    if (d->format == UseQImageFormat) {
        Block block;
        from >> block;
        QDataStream is(block);
        is.setVersion(QDataStream::Qt_4_8);
        is >> d->image;
        d->size.x = d->image.width();
        d->size.y = d->image.height();
    }
    else {
        from >> d->size >> d->pixels;
    }
}
GLPixelFormat Image::glFormat(Format imageFormat) {
    DENG2_ASSERT(imageFormat >= Luminance_8 && imageFormat <= RGBx_8888);
    switch (imageFormat) {
    case Luminance_8:
        return GLPixelFormat(GL_LUMINANCE, GL_UNSIGNED_BYTE, 1);
    case LuminanceAlpha_88:
        return GLPixelFormat(GL_LUMINANCE_ALPHA, GL_UNSIGNED_BYTE, 2);
    case Alpha_8:
        return GLPixelFormat(GL_ALPHA, GL_UNSIGNED_BYTE, 1);
    case RGB_555:
        return GLPixelFormat(GL_RGB, GL_UNSIGNED_SHORT_5_5_5_1, 2);
    case RGB_565:
        return GLPixelFormat(GL_RGB, GL_UNSIGNED_SHORT_5_6_5, 2);
    case RGB_444:
        return GLPixelFormat(GL_RGB, GL_UNSIGNED_SHORT_4_4_4_4, 2);
    case RGB_888:
        return GLPixelFormat(GL_RGB, GL_UNSIGNED_BYTE, 1);
    case RGBA_4444:
        return GLPixelFormat(GL_RGBA, GL_UNSIGNED_SHORT_4_4_4_4, 2);
    case RGBA_5551:
        return GLPixelFormat(GL_RGBA, GL_UNSIGNED_SHORT_5_5_5_1, 2);
    case RGBA_8888:
        return GLPixelFormat(GL_RGBA, GL_UNSIGNED_BYTE, 4);
    default:
    case RGBx_8888:
        return GLPixelFormat(GL_RGBA, GL_UNSIGNED_BYTE, 4);
    }
}
GLPixelFormat Image::glFormat(QImage::Format format) {
    switch (format) {
    case QImage::Format_Indexed8:
        return GLPixelFormat(GL_LUMINANCE, GL_UNSIGNED_BYTE, 1);
    case QImage::Format_RGB444:
        return GLPixelFormat(GL_RGB, GL_UNSIGNED_SHORT_4_4_4_4, 2);
    case QImage::Format_ARGB4444_Premultiplied:
        return GLPixelFormat(GL_RGBA, GL_UNSIGNED_SHORT_4_4_4_4, 2);
    case QImage::Format_RGB555:
        return GLPixelFormat(GL_RGB, GL_UNSIGNED_SHORT_5_5_5_1, 2);
    case QImage::Format_RGB16:
        return GLPixelFormat(GL_RGB, GL_UNSIGNED_SHORT_5_6_5, 2);
    case QImage::Format_RGB888:
        return GLPixelFormat(GL_RGB, GL_UNSIGNED_BYTE, 1);
    case QImage::Format_RGB32:
#if defined (DENG_OPENGL)
        return GLPixelFormat(GL_BGR, GL_UNSIGNED_BYTE, 4);
#else
        return GLPixelFormat(GL_BGRA8_EXT, GL_UNSIGNED_BYTE, 4);
#endif
    case QImage::Format_ARGB32:
        return GLPixelFormat(GL_BGRA, GL_UNSIGNED_BYTE, 4);
    default:
        break;
    }
    return GLPixelFormat(GL_RGBA, GL_UNSIGNED_BYTE, 4);
}
Image Image::solidColor(Color const &color, Size const &size) {
    QImage img(QSize(size.x, size.y), QImage::Format_ARGB32);
    img.fill(QColor(color.x, color.y, color.z, color.w).rgba());
    return img;
}
Image Image::fromData(IByteArray const &data, String const &formatHint) {
    return fromData(Block(data), formatHint);
}
Image Image::fromData(Block const &data, String const &formatHint) {
    if (!formatHint.compareWithoutCase(".tga") && tga::recognize(data)) {
        return tga::load(data);
    }
    if (pcx::recognize(data)) {
        return pcx::load(data);
    }
    return QImage::fromData(data).convertToFormat(QImage::Format_ARGB32);
}
Image Image::fromIndexedData(Size const &size, IByteArray const &image, IByteArray const &palette) {
    QImage rgba(size.x, size.y, QImage::Format_ARGB32);
    Block color(size.x);
    for (duint y = 0; y < size.y; ++y) {
        duint32 *dest = reinterpret_cast<duint32 *>(rgba.bits() + y * rgba.bytesPerLine());
        image.get(size.x * y, color.data(), color.size());
        auto const *srcColor = color.dataConst();
        for (duint x = 0; x < size.x; ++x) {
            duint8 paletteColor[3];
            palette.get(*srcColor++ * 3, paletteColor, 3);
            *dest++ = qRgba(paletteColor[0], paletteColor[1], paletteColor[2], 255);
        }
    }
    return rgba;
}
Image Image::fromMaskedIndexedData(Size const &size, IByteArray const &imageAndMask,
                                   IByteArray const &palette) {
    duint const layerSize = size.x * size.y;
    QImage rgba(size.x, size.y, QImage::Format_ARGB32);
    Block color(size.x);
    Block alpha(size.x);
    for (duint y = 0; y < size.y; ++y) {
        duint32 *dest = reinterpret_cast<duint32 *>(rgba.bits() + y * rgba.bytesPerLine());
        imageAndMask.get(size.x * y,             color.data(), color.size());
        imageAndMask.get(size.x * y + layerSize, alpha.data(), alpha.size());
        auto const *srcColor = color.dataConst();
        auto const *srcAlpha = alpha.dataConst();
        for (duint x = 0; x < size.x; ++x) {
            duint8 paletteColor[3];
            palette.get(*srcColor++ * 3, paletteColor, 3);
            *dest++ = qRgba(paletteColor[0], paletteColor[1], paletteColor[2], *srcAlpha++);
        }
    }
    return rgba;
}
bool Image::recognize(File const &file) {
    String const ext = file.extension().toLower();
    return (ext == ".tga"  || ext == ".pcx" || ext == ".png"  || ext == ".jpg" ||
            ext == ".jpeg" || ext == ".gif" || ext == ".tiff" || ext == ".ico");
}
} // namespace de
namespace de {
DENG2_PIMPL_NOREF(ImageBank) {
    struct ImageSource : public ISource {
        String filePath;
        float  pointRatio;
        ImageSource(String const &path, float pointRatio)
            : filePath(path)
            , pointRatio(pointRatio)
        {}
        Time modifiedAt() const {
            return App::rootFolder().locate<File>(filePath).status().modifiedAt;
        }
        Image load() const {
            Image img = FS::locate<const ImageFile>(filePath).image();
            if (pointRatio > 0) {
                img.setPointRatio(pointRatio);
            }
            return img;
        }
    };
    struct ImageData : public IData {
        Image image;
        ImageData() {}
        ImageData(Image const &img) : image(img) {}
        ISerializable *asSerializable() {
            return &image;
        }
        duint sizeInMemory() const {
            return duint(image.byteCount());
        }
    };
};
ImageBank::ImageBank(Flags const &flags) : InfoBank("ImageBank", flags), d(new Impl)
{}
void ImageBank::add(DotPath const &path, String const &imageFilePath) {
    Bank::add(path, new Impl::ImageSource(imageFilePath, 0.f));
}
void ImageBank::addFromInfo(File const &file) {
    LOG_AS("ImageBank");
    parse(file);
    addFromInfoBlocks("image");
}
Image const &ImageBank::image(DotPath const &path) const {
    return data(path).as<Impl::ImageData>().image;
}
Bank::ISource *ImageBank::newSourceFromInfo(String const &id) {
    Record const &def = info()[id];
    return new Impl::ImageSource(absolutePathInContext(def, def["path"]),
                                 def.getf("pointRatio", 0.f));
}
Bank::IData *ImageBank::loadFromSource(ISource &source) {
    return new Impl::ImageData(source.as<Impl::ImageSource>().load());
}
Bank::IData *ImageBank::newData() {
    return new Impl::ImageData();
}
} // namespace de
namespace de {
static String const MULTIPLY            ("Multiply:");
static String const HEIGHTMAP_TO_NORMALS("HeightMap.toNormals");
static String const COLOR_DESATURATE    ("Color.desaturate");
DENG2_PIMPL(ImageFile) {
    BuiltInFilter filter = NoFilter;
    QHash<BuiltInFilter, ImageFile *> filtered; // owned
    String filterParameter;
    Impl(Public *i) : Base(i) {}
    ~Impl() {
        qDeleteAll(filtered);
    }
    ImageFile *makeOrGetFiltered(BuiltInFilter filter) {
        auto found = filtered.constFind(filter);
        if (found != filtered.constEnd()) {
            return found.value();
        }
        if (filter != NoFilter) //filter == HeightMapToNormals || filter == Multiply)
        {
            ImageFile *sub = new ImageFile(filter, self());
            filtered.insert(filter, sub);
            return sub;
        }
        return nullptr;
    }
    ImageFile &filterSource() {
        return self().Node::parent()->as<ImageFile>();
    }
    static String filterTypeToText(BuiltInFilter filter) {
        switch (filter) {
        case HeightMapToNormals:
            return HEIGHTMAP_TO_NORMALS;
        case Multiply:
            return MULTIPLY;
        case ColorDesaturate:
            return COLOR_DESATURATE;
        default:
            break;
        }
        return "";
    }
};
ImageFile::ImageFile(File *source)
    : File(source->name())
    , d(new Impl(this)) {
    setSource(source);
}
ImageFile::ImageFile(BuiltInFilter filterType, ImageFile &filterSource)
    : File(Impl::filterTypeToText(filterType))
    , d(new Impl(this)) {
    d->filter = filterType;
    setParent(&filterSource);
}
ImageFile::~ImageFile() {
    DENG2_GUARD(this);
    DENG2_FOR_AUDIENCE2(Deletion, i) i->fileBeingDeleted(*this);
    audienceForDeletion().clear();
    deindex();
}
String ImageFile::describe() const {
    String desc = String("image \"%1\"").arg(d->filter == NoFilter?
                                             name() : d->filterSource().name());
    switch (d->filter) {
    case HeightMapToNormals:
        desc += " (filter: heightfield to normals)";
        break;
    case Multiply:
        desc += " (filter: multiplied with " + d->filterParameter + ")";
        break;
    case ColorDesaturate:
        desc += " (filter: desaturate)";
        break;
    default:
        break;
    }
    return desc;
}
Image ImageFile::image() const {
    if (d->filter != NoFilter) {
        Image img = d->filterSource().image();
        if (d->filter == HeightMapToNormals) {
            HeightMap heightMap;
            heightMap.loadGrayscale(img);
            img = heightMap.makeNormalMap();
        }
        else if (d->filter == Multiply) {
            String const refPath = d->filterSource().path().fileNamePath();
            Image factorImg = App::rootFolder().locate<ImageFile const>
                    (refPath / d->filterParameter).image();
            if (img.size() != factorImg.size()) {
                throw FilterError("ImageFile::image",
                                  String("Cannot multiply %1 and %2 due to different sizes")
                                  .arg(d->filterSource().path())
                                  .arg(refPath / d->filterParameter));
            }
            img = img.multiplied(factorImg);
        }
        else if (d->filter == ColorDesaturate) {
            img = img.colorized(Image::Color(255, 255, 255, 255));
        }
        return img;
    }
    else {
        Image img = Image::fromData(*source(), extension());
        if (source()->name().contains("@2x.", Qt::CaseInsensitive)) {
            img.setPointRatio(.5f);
        }
        return img;
    }
}
IIStream const &ImageFile::operator >> (IByteArray &bytes) const {
    *source() >> bytes;
    return *this;
}
filesys::Node const *ImageFile::tryGetChild(String const &name) const {
    if (!name.compareWithoutCase(HEIGHTMAP_TO_NORMALS)) {
        return d->makeOrGetFiltered(HeightMapToNormals);
    }
    else if (name.startsWith(MULTIPLY, Qt::CaseInsensitive)) {
        ImageFile *filtered = d->makeOrGetFiltered(Multiply);
        filtered->d->filterParameter = name.substr(MULTIPLY.size());
        return filtered;
    }
    else if (!name.compareWithoutCase(COLOR_DESATURATE)) {
        return d->makeOrGetFiltered(ColorDesaturate);
    }
    else if (d->filter == Multiply) {
        d->filterParameter = d->filterParameter / name;
        return this;
    }
    return nullptr;
}
File *ImageFile::Interpreter::interpretFile(File *sourceData) const {
    if (Image::recognize(*sourceData)) {
        LOG_RES_XVERBOSE("Interpreted %s as an image", sourceData->description());
        return new ImageFile(sourceData);
    }
    return nullptr;
}
} // namespace de
namespace de {
DENG2_PIMPL(KdTreeAtlasAllocator) {
    Atlas::Size size;
    int margin;
    Allocations allocs;
    struct Partition {
        Rectanglei area;
        Id alloc;   ///< Id of the allocation in this partition (or Id::None).
        Partition() : alloc(Id::None) {}
    };
    typedef BinaryTree<Partition> Node;
    Node root;
    Impl(Public *i) : Base(i), margin(0), root(Partition()) {}
    void initTree(Node &rootNode) {
        Partition full;
        full.area = Rectanglei(margin, margin, size.x - margin, size.y - margin);
        rootNode.setUserData(full);
    }
    Node *treeInsert(Node *parent, Atlas::Size const &allocSize) {
        if (!parent->isLeaf()) {
            Node *subTree;
            if ((subTree = treeInsert(parent->rightPtr(), allocSize)) != 0)
                return subTree;
            return treeInsert(parent->leftPtr(), allocSize);
        }
        /*
         * We have arrived at a leaf.
         */
        Partition pnode = parent->userData();
        if (!pnode.alloc.isNone())
            return 0;
        if (pnode.area.width() < allocSize.x || pnode.area.height() < allocSize.y)
            return 0; // No, too small.
        if (pnode.area.size() == allocSize)
            return parent;
        /*
         * The leaf will be split.
         */
        Partition rnode, lnode;
        if (pnode.area.width() - allocSize.x > pnode.area.height() - allocSize.y) {
            rnode.area = Rectanglei(pnode.area.left(), pnode.area.top(),
                                    allocSize.x, pnode.area.height());
            lnode.area = Rectanglei(pnode.area.left() + allocSize.x, pnode.area.top(),
                                    pnode.area.width() - allocSize.x, pnode.area.height());
        }
        else {
            rnode.area = Rectanglei(pnode.area.left(), pnode.area.top(),
                                    pnode.area.width(), allocSize.y);
            lnode.area = Rectanglei(pnode.area.left(), pnode.area.top() + allocSize.y,
                                    pnode.area.width(), pnode.area.height() - allocSize.y);
        }
        parent->setRight(new Node(rnode, parent));
        parent->setLeft(new Node(lnode, parent));
        return treeInsert(parent->rightPtr(), allocSize);
    }
    /**
     * Attempts to find a large enough free space for the requested size.
     * Room is left for margins.
     *
     * @param rootNode  Partitioning root node to use for allocation.
     * @param size      Size to allocate.
     * @param rect      Found rectangle is returned here.
     * @param preallocId  Id for the allocation, if one has already been determined.
     *
     * @return Allocated Id, if successful.
     */
    Id allocate(Node &rootNode, Atlas::Size const &size, Rectanglei &rect, Id const &preallocId = 0) {
        Atlas::Size const allocSize(size.x + margin, size.y + margin);
        if (Node *inserted = treeInsert(&rootNode, allocSize)) {
            Partition part = inserted->userData();
            part.alloc = (preallocId.isNone()? Id() : preallocId);
            inserted->setUserData(part);
            rect = part.area.adjusted(Vector2i(), Vector2i(-margin, -margin));
            return part.alloc;
        }
        return Id::None;
    }
    struct EraseArgs { Id id; };
    static int allocationEraser(Node &node, void *argsPtr) {
        EraseArgs *args = reinterpret_cast<EraseArgs *>(argsPtr);
        Partition part = node.userData();
        if (part.alloc == args->id) {
            part.alloc = Id::None;
            node.setUserData(part);
            return 1; // Can stop here.
        }
        return 0;
    }
    void releaseAlloc(Id const &id) {
        allocs.remove(id);
        EraseArgs args;
        args.id = id;
        root.traverseInOrder(allocationEraser, &args);
    }
    struct ContentSize {
        Id::Type id;
        duint64 area;
        ContentSize(Id const &allocId, Vector2ui const &size)
            : id(allocId), area(size.x * size.y) {}
        bool operator < (ContentSize const &other) const {
            return area > other.area;
        }
    };
    bool optimize() {
        QList<ContentSize> descending;
        DENG2_FOR_EACH(Allocations, i, allocs) {
            descending.append(ContentSize(i.key(), i.value().size()));
        }
        qSort(descending);
        Allocations optimal;
        Node optimalRoot;
        initTree(optimalRoot);
        /*
         * Attempt to optimize space usage by placing the largest allocations
         * first.
         */
        foreach (ContentSize const &iter, descending) {
            Rectanglei newRect;
            if (allocate(optimalRoot, allocs[iter.id].size(), newRect, iter.id).isNone()) {
                return false;
            }
            optimal.insert(iter.id, newRect);
        }
        root = optimalRoot;
        allocs = optimal;
        return true;
    }
};
KdTreeAtlasAllocator::KdTreeAtlasAllocator() : d(new Impl(this))
{}
void KdTreeAtlasAllocator::setMetrics(Atlas::Size const &totalSize, int margin) {
    DENG2_ASSERT(d->allocs.isEmpty());
    d->size   = totalSize;
    d->margin = margin;
    d->initTree(d->root);
}
void KdTreeAtlasAllocator::clear() {
    d->allocs.clear();
    d->root.clear();
}
Id KdTreeAtlasAllocator::allocate(Atlas::Size const &size, Rectanglei &rect,
                                  Id const &knownId) {
    Id newId = d->allocate(d->root, size, rect, knownId);
    if (newId.isNone()) {
        return 0;
    }
    d->allocs[newId] = rect;
    return newId;
}
void KdTreeAtlasAllocator::release(Id const &id) {
    DENG2_ASSERT(d->allocs.contains(id));
    d->releaseAlloc(id);
}
int KdTreeAtlasAllocator::count() const {
    return d->allocs.size();
}
Atlas::Ids KdTreeAtlasAllocator::ids() const {
    Atlas::Ids ids;
    foreach (Id const &id, d->allocs.keys()) {
        ids.insert(id);
    }
    return ids;
}
void KdTreeAtlasAllocator::rect(Id const &id, Rectanglei &rect) const {
    DENG2_ASSERT(d->allocs.contains(id));
    rect = d->allocs[id];
}
KdTreeAtlasAllocator::Allocations KdTreeAtlasAllocator::allocs() const {
    return d->allocs;
}
bool KdTreeAtlasAllocator::optimize() {
    return d->optimize();
}
} // namespace de
namespace de {
DENG2_PIMPL(ModelBank) {
    Constructor modelConstructor;
    struct Source : public ISource {
        String path; ///< Path to a model file.
        Source(String const &sourcePath) : path(sourcePath) {}
    };
    struct Data : public IData {
        std::unique_ptr<ModelDrawable> model;
        std::unique_ptr<IUserData> userData;
        Data(ModelDrawable *model, String const &path)
            : model(model) {
            model->load(App::rootFolder().locate<File>(path));
        }
    };
    Impl(Public *i, Constructor c)
        : Base(i)
        , modelConstructor(c? c : [] () { return new ModelDrawable; })
    {}
};
ModelBank::ModelBank(Constructor modelConstructor)
    : Bank("ModelBank", BackgroundThread)
    , d(new Impl(this, modelConstructor))
{}
void ModelBank::add(DotPath const &id, String const &sourcePath) {
    return Bank::add(id, new Impl::Source(sourcePath));
}
ModelDrawable &ModelBank::model(DotPath const &id) {
    return *data(id).as<Impl::Data>().model;
}
void ModelBank::setUserData(DotPath const &id, IUserData *anim) {
    data(id).as<Impl::Data>().userData.reset(anim);
}
ModelBank::IUserData const *ModelBank::userData(DotPath const &id) const {
    return data(id).as<Impl::Data>().userData.get();
}
ModelBank::ModelWithData ModelBank::modelAndData(DotPath const &id) const {
    auto &item = data(id).as<Impl::Data>();
    return ModelWithData(item.model.get(), item.userData.get());
}
Bank::IData *ModelBank::loadFromSource(ISource &source) {
    return new Impl::Data(d->modelConstructor(),
                              source.as<Impl::Source>().path);
}
} // namespace de
namespace de {
namespace internal {
class ImpIOStream : public Assimp::IOStream {
public:
    ImpIOStream(ByteArrayFile const &file) : _file(file), _pos(0)
    {}
    size_t Read(void *pvBuffer, size_t pSize, size_t pCount) {
        size_t const num = pSize * pCount;
        _file.get(_pos, reinterpret_cast<IByteArray::Byte *>(pvBuffer), num);
        _pos += num;
        return pCount;
    }
    size_t Write(void const *, size_t, size_t) {
        throw Error("ImpIOStream::Write", "Writing not allowed");
    }
    aiReturn Seek(size_t pOffset, aiOrigin pOrigin) {
        switch (pOrigin) {
        case aiOrigin_SET:
            _pos = pOffset;
            break;
        case aiOrigin_CUR:
            _pos += pOffset;
            break;
        case aiOrigin_END:
            _pos = _file.size() - pOffset;
            break;
        default:
            break;
        }
        return aiReturn_SUCCESS;
    }
    size_t Tell() const {
        return _pos;
    }
    size_t FileSize() const {
        return _file.size();
    }
    void Flush() {}
private:
    ByteArrayFile const &_file;
    size_t _pos;
};
/**
 * Adapter between FS2 and Assimp. Each ModelDrawable instance has its own
 * instance of this class.
 */
struct ImpIOSystem : public Assimp::IOSystem {
    String referencePath;
    ImpIOSystem() {}
    ~ImpIOSystem() {}
    char getOsSeparator() const { return '/'; }
    Path resolvePath(char const *fn) const {
        Path path(fn);
        if (path.isAbsolute()) return path;
        return referencePath / path;
    }
    bool Exists(char const *pFile) const {
        return App::rootFolder().has(resolvePath(pFile));
    }
    Assimp::IOStream *Open(char const *pFile, char const *) {
        Path const path = resolvePath(pFile);
        return new ImpIOStream(App::rootFolder().locate<ByteArrayFile const>(path));
    }
    void Close(Assimp::IOStream *pFile) {
        delete pFile;
    }
};
struct ImpLogger : public Assimp::LogStream {
    void write(char const *message) {
        LOG_GL_VERBOSE("[ai] %s") << message;
    }
    static void registerLogger() {
        if (registered) return;
        registered = true;
        Assimp::DefaultLogger::get()->attachStream(
                    new ImpLogger,
                    Assimp::Logger::Info |
                    Assimp::Logger::Warn |
                    Assimp::Logger::Err);
    }
    static bool registered;
};
bool ImpLogger::registered = false;
struct DefaultImageLoader : public ModelDrawable::IImageLoader {
    Image loadImage(String const &path) {
        Image img = App::rootFolder().locate<ImageFile>(path).image();
        if (img.depth() == 24) {
            DENG2_ASSERT(img.canConvertToQImage());
            return Image(img.toQImage().convertToFormat(QImage::Format_ARGB32));
        }
        return img;
    }
};
static DefaultImageLoader defaultImageLoader;
} // namespace internal
using namespace internal;
static int const MAX_BONES = 64;
static int const MAX_BONES_PER_VERTEX = 4;
static int const MAX_TEXTURES = 4;
struct ModelVertex {
    Vector3f pos;
    Vector4f color;
    Vector4f boneIds;
    Vector4f boneWeights;
    Vector3f normal;
    Vector3f tangent;
    Vector3f bitangent;
    Vector2f texCoord;
    Vector4f texBounds[4];
    LIBGUI_DECLARE_VERTEX_FORMAT(12)
};
AttribSpec const ModelVertex::_spec[12] = {
    { AttribSpec::Position,    3, GL_FLOAT, false, sizeof(ModelVertex),  0 },
    { AttribSpec::Color,       4, GL_FLOAT, false, sizeof(ModelVertex),  3 * sizeof(float) },
    { AttribSpec::BoneIDs,     4, GL_FLOAT, false, sizeof(ModelVertex),  7 * sizeof(float) },
    { AttribSpec::BoneWeights, 4, GL_FLOAT, false, sizeof(ModelVertex), 11 * sizeof(float) },
    { AttribSpec::Normal,      3, GL_FLOAT, false, sizeof(ModelVertex), 15 * sizeof(float) },
    { AttribSpec::Tangent,     3, GL_FLOAT, false, sizeof(ModelVertex), 18 * sizeof(float) },
    { AttribSpec::Bitangent,   3, GL_FLOAT, false, sizeof(ModelVertex), 21 * sizeof(float) },
    { AttribSpec::TexCoord0,   2, GL_FLOAT, false, sizeof(ModelVertex), 24 * sizeof(float) },
    { AttribSpec::TexBounds0,  4, GL_FLOAT, false, sizeof(ModelVertex), 26 * sizeof(float) },
    { AttribSpec::TexBounds1,  4, GL_FLOAT, false, sizeof(ModelVertex), 30 * sizeof(float) },
    { AttribSpec::TexBounds2,  4, GL_FLOAT, false, sizeof(ModelVertex), 34 * sizeof(float) },
    { AttribSpec::TexBounds3,  4, GL_FLOAT, false, sizeof(ModelVertex), 38 * sizeof(float) },
};
LIBGUI_VERTEX_FORMAT_SPEC(ModelVertex, 42 * sizeof(float))
static Matrix4f convertMatrix(aiMatrix4x4 const &aiMat) {
    return Matrix4f(&aiMat.a1).transpose();
}
static ddouble secondsToTicks(ddouble seconds, aiAnimation const &anim) {
    ddouble const ticksPerSec = anim.mTicksPerSecond? anim.mTicksPerSecond : 25.0;
    return seconds * ticksPerSec;
}
static ddouble ticksToSeconds(ddouble ticks, aiAnimation const &anim) {
    return ticks / secondsToTicks(1.0, anim);
}
static String const DUMMY_BONE_NAME = "__deng_dummy-bone__";
DENG2_PIMPL(ModelDrawable) {
    typedef GLBufferT<ModelVertex> VBuf;
    typedef QHash<String, int> AnimLookup;
    static TextureMap textureMapType(aiTextureType type) {
        switch (type) {
        case aiTextureType_DIFFUSE:  return Diffuse;
        case aiTextureType_NORMALS:  return Normals;
        case aiTextureType_HEIGHT:   return Height;
        case aiTextureType_SPECULAR: return Specular;
        case aiTextureType_EMISSIVE: return Emissive;
        default:
            DENG2_ASSERT(!"Unsupported texture type");
            return Diffuse;
        }
    }
    static aiTextureType impTextureType(TextureMap map) {
        switch (map) {
        case Diffuse:  return aiTextureType_DIFFUSE;
        case Normals:  return aiTextureType_NORMALS;
        case Height:   return aiTextureType_HEIGHT;
        case Specular: return aiTextureType_SPECULAR;
        case Emissive: return aiTextureType_EMISSIVE;
        default:
            break;
        }
        return aiTextureType_UNKNOWN;
    }
    struct VertexBone {
        duint16 ids[MAX_BONES_PER_VERTEX];
        dfloat weights[MAX_BONES_PER_VERTEX];
        VertexBone() {
            zap(ids);
            zap(weights);
        }
    };
    struct BoneData {
        Matrix4f offset;
    };
    Asset modelAsset;
    String sourcePath;
    ImpIOSystem *importerIoSystem; // not owned
    Assimp::Importer importer;
    aiScene const *scene { nullptr };
    Vector3f minPoint; ///< Bounds in default pose.
    Vector3f maxPoint;
    Matrix4f globalInverse;
    QVector<VertexBone> vertexBones; // indexed by vertex
    QHash<String, duint16> boneNameToIndex;
    QHash<String, aiNode const *> nodeNameToPtr;
    QVector<BoneData> bones; // indexed by bone index
    AnimLookup animNameToIndex;
    QVector<Rangeui> meshIndexRanges;
    /**
     * Management of texture maps.
     */
    struct GLData {
        /**
         * Each material has its own VBO with a particular, fixed set of texture
         * coordinates.
         */
        struct Material {
            struct MeshTextures {
                Id::Type texIds[MAX_TEXTURES];
                QHash<TextureMap, String> customPaths;
                MeshTextures() { zap(texIds); }
            };
            QVector<MeshTextures> meshTextures; // indexed by mesh index
            std::unique_ptr<VBuf> buffer;
        };
        /**
         * Source information for a texture used in one or more of the meshes.
         * Depends on the ModelDrawable's IImageLoader.
         */
        struct TextureSource : public TextureBank::ImageSource {
            GLData *d;
            TextureSource(String const &path, GLData *glData)
                : ImageSource(Path(path))
                , d(glData) {}
            Image load() const {
                return d->imageLoader->loadImage(sourcePath().toString());
            }
        };
        Id::Type defaultTexIds[MAX_TEXTURES];  ///< Used if no other texture is provided.
        TextureMap textureOrder[MAX_TEXTURES]; ///< Order of textures for vertex buffer texcoords.
        IImageLoader *imageLoader { &defaultImageLoader };
        TextureBank textureBank;
        QList<Material *> materials; // owned
        bool needMakeBuffer { false };
        String sourcePath; ///< Location of the model file (imported with Assimp).
        aiScene const *scene { nullptr };
        GLData() {
            textureBank.setSeparator('/');
            textureOrder[0] = Diffuse;
            textureOrder[1] = textureOrder[2] = textureOrder[3] = Unknown;
            zap(defaultTexIds);
        }
        ~GLData() {
            qDeleteAll(materials);
        }
        void initMaterials() {
            deinitMaterials();
            addMaterial(); // default is at index zero
        }
        void deinitMaterials() {
            qDeleteAll(materials);
            materials.clear();
        }
        duint addMaterial() {
            DENG2_ASSERT(scene != nullptr);
            needMakeBuffer = true;
            Material *material = new Material;
            for (duint i = 0; i < scene->mNumMeshes; ++i) {
                material->meshTextures << Material::MeshTextures();
            }
            materials << material;
            return materials.size() - 1;
        }
        void glInit(String modelSourcePath) {
            sourcePath = modelSourcePath;
            initTextures();
        }
        void glDeinit() {
            releaseTexturesFromAtlas();
        }
        void releaseTexture(Id const &id) {
            if (!id) return; // We don't own this, don't release.
            Path texPath = textureBank.sourcePathForAtlasId(id);
            DENG2_ASSERT(!texPath.isEmpty());
            LOGDEV_GL_VERBOSE("Releasing model texture '%s' path: \"%s\"")
                    << id.asText() << texPath;
            textureBank.unload(texPath);
        }
        void releaseTexturesFromAtlas() {
            textureBank.unloadAll(Bank::ImmediatelyInCurrentThread);
            for (Material *mat : materials) {
                for (auto &mesh : mat->meshTextures) {
                    zap(mesh.texIds);
                }
            }
            textureBank.clear();
        }
        void fallBackToDefaultTexture(Material::MeshTextures &mesh,
                                      TextureMap map) const {
            if (!mesh.texIds[map]) {
                mesh.texIds[map] = defaultTexIds[map];
            }
        }
        /**
         * Load all the textures of the model, for all materials. The textures
         * are allocated into the atlas provided to the model.
         *
         * Only a single copy of each texture image is kept in the atlas even
         * if the same image is beig used in many meshes.
         */
        void initTextures() {
            for (duint matIdx = 0; matIdx < duint(materials.size()); ++matIdx) {
                for (duint i = 0; i < scene->mNumMeshes; ++i) {
                    MeshId const mesh(i, matIdx);
                    auto &textures = materials[matIdx]->meshTextures[i];
                    loadTextureImage(mesh, aiTextureType_DIFFUSE);
                    fallBackToDefaultTexture(textures, Diffuse);
                    loadTextureImage(mesh, aiTextureType_NORMALS);
                    if (!textures.texIds[Normals]) {
                        loadTextureImage(mesh, aiTextureType_HEIGHT);
                    }
                    fallBackToDefaultTexture(textures, Normals);
                    loadTextureImage(mesh, aiTextureType_SPECULAR);
                    fallBackToDefaultTexture(textures, Specular);
                    loadTextureImage(mesh, aiTextureType_EMISSIVE);
                    fallBackToDefaultTexture(textures, Emissive);
                }
            }
            textureBank.atlas()->commit();
        }
        /**
         * Attempts to load a texture image specified in the material. Also checks if
         * an overridden custom path is provided, though.
         *
         * @param mesh  Identifies the mesh whose texture is being loaded.
         * @param type  AssImp texture type.
         */
        void loadTextureImage(MeshId const &mesh, aiTextureType type) {
            DENG2_ASSERT(imageLoader != 0);
            aiMesh     const &sceneMesh     = *scene->mMeshes[mesh.index];
            aiMaterial const &sceneMaterial = *scene->mMaterials[sceneMesh.mMaterialIndex];
            auto const &meshTextures = materials.at(mesh.material)->meshTextures[mesh.index];
            TextureMap const map = textureMapType(type);
            try {
                if (meshTextures.customPaths.contains(map)) {
                    LOG_GL_VERBOSE("Loading custom path \"%s\"") << meshTextures.customPaths[map];
                    return setTexture(mesh, map, meshTextures.customPaths[map]);
                }
            }
            catch (Error const &er) {
                LOG_GL_WARNING("Failed to load user-defined %s texture for "
                               "mesh %i (material %i): %s")
                        << textureMapToText(textureMapType(type))
                        << mesh.index << mesh.material
                        << er.asText();
            }
            aiString texPath;
            for (duint s = 0; s < sceneMaterial.GetTextureCount(type); ++s) {
                try {
                    if (sceneMaterial.GetTexture(type, s, &texPath) == AI_SUCCESS) {
                        setTexture(mesh, map, sourcePath.fileNamePath() / NativePath(texPath.C_Str()));
                        break;
                    }
                }
                catch (Error const &er) {
                    LOG_GL_WARNING("Failed to load %s texture for mesh %i "
                                   "(material %i) based on info from model file: %s")
                            << textureMapToText(textureMapType(type))
                            << mesh.index << mesh.material << er.asText();
                }
            }
        }
        void setTexture(MeshId const &mesh, TextureMap map, String contentPath) {
            if (!scene) return;
            if (map == Unknown) return; // Ignore unmapped textures.
            if (mesh.material >= duint(materials.size())) return;
            if (mesh.index >= scene->mNumMeshes) return;
            DENG2_ASSERT(textureBank.atlas());
            Material &material = *materials[mesh.material];
            auto &meshTextures = material.meshTextures[mesh.index];
            Id::Type &destId = (map == Height? meshTextures.texIds[Normals] :
                                               meshTextures.texIds[map]);
            if (map == Height) {
                contentPath = contentPath.concatenatePath("HeightMap.toNormals");
            }
            Path const path(contentPath);
            if (!textureBank.has(path)) {
                textureBank.add(path, new TextureSource(contentPath, this));
            }
            LOGDEV_GL_VERBOSE("material: %i mesh: %i file: \"%s\"")
                    << mesh.material
                    << mesh.index
                    << textureMapToText(map)
                    << contentPath;
            destId = textureBank.texture(path);
            needMakeBuffer = true;
        }
        void setTextureMapping(Mapping mapsToUse) {
            for (int i = 0; i < MAX_TEXTURES; ++i) {
                textureOrder[i] = (i < mapsToUse.size()? mapsToUse.at(i) : Unknown);
                if (textureOrder[i] == Height) textureOrder[i] = Normals;
            }
            needMakeBuffer = true;
        }
        /**
         * Sets a custom texture that will be loaded when the model is glInited.
         *
         * @param matId  Material.
         * @param tex    Texture map.
         * @param path   Image file path.
         */
        void setCustomTexturePath(MeshId const &mesh, TextureMap map, String const &path) {
            DENG2_ASSERT(!textureBank.atlas()); // in-use textures cannot be replaced on the fly
            DENG2_ASSERT(mesh.index < scene->mNumMeshes);
            DENG2_ASSERT(mesh.material < duint(materials.size()));
            materials[mesh.material]->meshTextures[mesh.index].customPaths.insert(map, path);
        }
    };
    GLData glData;
    Passes defaultPasses;
    GLProgram *program { nullptr }; ///< Default program (overridden by pass shaders).
    mutable GLUniform uBoneMatrices { "uBoneMatrices", GLUniform::Mat4Array, MAX_BONES };
    Impl(Public *i) : Base(i) {
        importer.SetIOHandler(importerIoSystem = new ImpIOSystem);
        ImpLogger::registerLogger();
    }
    ~Impl() {
        glDeinit();
    }
    void import(File const &file) {
        LOG_GL_MSG("Loading model from %s") << file.description();
#if defined (DENG_HAVE_CUSTOMIZED_ASSIMP)
        {
        /*
         * MD5: Multiple animation sequences are supported via multiple .md5anim files.
         * Autodetect if these exist and make a list of their names.
         */
        String anims;
        if (file.extension() == ".md5mesh") {
            String const baseName = file.name().fileNameWithoutExtension() + "_";
            file.parent()->forContents([&anims, &baseName] (String fileName, File &) {
                if (fileName.startsWith(baseName) &&
                    fileName.fileNameExtension() == ".md5anim") {
                    if (!anims.isEmpty()) anims += ";";
                    anims += fileName.substr(baseName.size()).fileNameWithoutExtension();
                }
                return LoopContinue;
            });
        }
        importer.SetPropertyString(AI_CONFIG_IMPORT_MD5_ANIM_SEQUENCE_NAMES, anims.toStdString());
        }
#endif
        scene = glData.scene = nullptr;
        sourcePath = file.path();
        importerIoSystem->referencePath = sourcePath.fileNamePath();
        if (!importer.ReadFile(sourcePath.toUtf8(),
                              aiProcess_CalcTangentSpace |
                              aiProcess_GenSmoothNormals |
                              aiProcess_JoinIdenticalVertices |
                              aiProcess_Triangulate |
                              aiProcess_GenUVCoords |
                              aiProcess_FlipUVs |
                              aiProcess_SortByPType)) {
            throw LoadError("ModelDrawable::import", String("Failed to load model from %1: %2")
                            .arg(file.description()).arg(importer.GetErrorString()));
        }
        scene = glData.scene = importer.GetScene();
        initBones();
        globalInverse = convertMatrix(scene->mRootNode->mTransformation).inverse();
        maxPoint      = Vector3f(1.0e-9f, 1.0e-9f, 1.0e-9f);
        minPoint      = Vector3f(1.0e9f,  1.0e9f,  1.0e9f);
        for (duint i = 0; i < scene->mNumMeshes; ++i) {
            aiMesh const &mesh = *scene->mMeshes[i];
            for (duint i = 0; i < mesh.mNumVertices; ++i) {
                addToBounds(Vector3f(&mesh.mVertices[i].x));
            }
        }
        LOG_GL_VERBOSE("Bone count: %i\n"
                       "Animation count: %i")
                << boneCount()
                << scene->mNumAnimations;
        animNameToIndex.clear();
        for (duint i = 0; i < scene->mNumAnimations; ++i) {
            LOG_GL_VERBOSE("Animation #%i name:%s tps:%f")
                    << i << scene->mAnimations[i]->mName.C_Str()
                    << scene->mAnimations[i]->mTicksPerSecond;
            String const name = scene->mAnimations[i]->mName.C_Str();
            if (!name.isEmpty()) {
                animNameToIndex.insert(name, i);
            }
        }
        nodeNameToPtr.clear();
        nodeNameToPtr.insert("", scene->mRootNode);
        buildNodeLookup(*scene->mRootNode);
        glData.initMaterials();
        Pass pass;
        pass.meshes.resize(scene->mNumMeshes);
        pass.meshes.fill(true);
        defaultPasses << pass;
    }
    void buildNodeLookup(aiNode const &node) {
        String const name = node.mName.C_Str();
#ifdef DENG2_DEBUG
        qDebug() << "Node:" << name;
#endif
        if (!name.isEmpty()) {
            nodeNameToPtr.insert(name, &node);
        }
        for (duint i = 0; i < node.mNumChildren; ++i) {
            buildNodeLookup(*node.mChildren[i]);
        }
    }
    void clear() {
        glDeinit();
        sourcePath.clear();
        defaultPasses.clear();
        importer.FreeScene();
        scene = glData.scene = nullptr;
    }
    void glInit() {
        DENG2_ASSERT_IN_MAIN_THREAD();
        if (!scene) return;
        if (modelAsset.isReady()) {
            return;
        }
        DENG2_FOR_PUBLIC_AUDIENCE2(AboutToGLInit, i) {
            i->modelAboutToGLInit(self());
        }
        glData.glInit(sourcePath);
        makeBuffer();
        modelAsset.setState(Ready);
    }
    void glDeinit() {
        glData.glDeinit();
        clearBones();
        modelAsset.setState(NotReady);
    }
    void addToBounds(Vector3f const &point) {
        minPoint = minPoint.min(point);
        maxPoint = maxPoint.max(point);
    }
    int findMaterial(String const &name) const {
        if (!scene) return -1;
        for (duint i = 0; i < scene->mNumMaterials; ++i) {
            aiMaterial const &material = *scene->mMaterials[i];
            aiString matName;
            if (material.Get(AI_MATKEY_NAME, matName) == AI_SUCCESS) {
                if (name == matName.C_Str()) return i;
            }
        }
        return -1;
    }
    void clearBones() {
        vertexBones.clear();
        bones.clear();
        boneNameToIndex.clear();
    }
    int boneCount() const {
        return bones.size();
    }
    int addBone(String const &name) {
        int idx = boneCount();
        bones << BoneData();
        boneNameToIndex[name] = idx;
        return idx;
    }
    int findBone(String const &name) const {
        if (boneNameToIndex.contains(name)) {
            return boneNameToIndex[name];
        }
        return -1;
    }
    int addOrFindBone(String const &name) {
        int i = findBone(name);
        if (i >= 0) {
            return i;
        }
        return addBone(name);
    }
    void addVertexWeight(duint vertexIndex, duint16 boneIndex, dfloat weight) {
        VertexBone &vb = vertexBones[vertexIndex];
        for (int i = 0; i < MAX_BONES_PER_VERTEX; ++i) {
            if (vb.weights[i] == 0.f) {
                vb.ids[i] = boneIndex;
                vb.weights[i] = weight;
                return;
            }
        }
        LOG_GL_WARNING("\"%s\": too many weights for vertex %i (only 4 supported), bone index: %i")
            << sourcePath << vertexIndex << boneIndex;
        DENG2_ASSERT(!"Too many bone weights for a vertex");
    }
    /**
     * Initializes the per-vertex bone weight information, and indexes the bones
     * of the mesh in a sequential order.
     *
     * @param mesh        Source mesh.
     * @param vertexBase  Index of the first vertex of the mesh.
     */
    void initMeshBones(aiMesh const &mesh, duint vertexBase) {
        vertexBones.resize(vertexBase + mesh.mNumVertices);
        if (mesh.HasBones()) {
            for (duint i = 0; i < mesh.mNumBones; ++i) {
                aiBone const &bone = *mesh.mBones[i];
                duint const boneIndex = addOrFindBone(bone.mName.C_Str());
                bones[boneIndex].offset = convertMatrix(bone.mOffsetMatrix);
                for (duint w = 0; w < bone.mNumWeights; ++w) {
                    addVertexWeight(vertexBase + bone.mWeights[w].mVertexId,
                                    boneIndex, bone.mWeights[w].mWeight);
                }
            }
        }
        else {
            duint const boneIndex = addOrFindBone(DUMMY_BONE_NAME);
            bones[boneIndex].offset = Matrix4f();
            for (duint i = 0; i < mesh.mNumVertices; ++i) {
                addVertexWeight(vertexBase + i, boneIndex, 1.f);
            }
        }
    }
    /**
     * Initializes all bones in the scene.
     */
    void initBones() {
        clearBones();
        int base = 0;
        for (duint i = 0; i < scene->mNumMeshes; ++i) {
            aiMesh const &mesh = *scene->mMeshes[i];
            LOGDEV_GL_VERBOSE("Initializing %i bones for mesh #%i %s")
                    << mesh.mNumBones << i << mesh.mName.C_Str();
            initMeshBones(mesh, base);
            base += mesh.mNumVertices;
        }
    }
    void makeBuffer() {
        glData.needMakeBuffer = false;
        for (GLData::Material *material : glData.materials) {
            makeBuffer(*material);
        }
    }
    /**
     * Allocates and fills in the GL buffer containing vertex information.
     *
     * @param material  Material whose vertex buffer to create.
     */
    void makeBuffer(GLData::Material &material) {
        VBuf::Vertices verts;
        VBuf::Indices indx;
        aiVector3D const zero(0, 0, 0);
        aiColor4D const white(1, 1, 1, 1);
        int base = 0;
        meshIndexRanges.clear();
        meshIndexRanges.resize(scene->mNumMeshes);
        for (duint m = 0; m < scene->mNumMeshes; ++m) {
            aiMesh const &mesh = *scene->mMeshes[m];
            for (duint i = 0; i < mesh.mNumVertices; ++i) {
                aiVector3D const *pos      = &mesh.mVertices[i];
                aiColor4D  const *color    = (mesh.HasVertexColors(0)? &mesh.mColors[0][i] : &white);
                aiVector3D const *normal   = (mesh.HasNormals()? &mesh.mNormals[i] : &zero);
                aiVector3D const *texCoord = (mesh.HasTextureCoords(0)? &mesh.mTextureCoords[0][i] : &zero);
                aiVector3D const *tangent  = (mesh.HasTangentsAndBitangents()? &mesh.mTangents[i] : &zero);
                aiVector3D const *bitang   = (mesh.HasTangentsAndBitangents()? &mesh.mBitangents[i] : &zero);
                VBuf::Type v;
                v.pos       = Vector3f(pos->x, pos->y, pos->z);
                v.color     = Vector4f(color->r, color->g, color->b, color->a);
                v.normal    = Vector3f(normal ->x, normal ->y, normal ->z);
                v.tangent   = Vector3f(tangent->x, tangent->y, tangent->z);
                v.bitangent = Vector3f(bitang ->x, bitang ->y, bitang ->z);
                v.texCoord     = Vector2f(texCoord->x, texCoord->y);
                v.texBounds[0] = Vector4f(0, 0, 1, 1);
                v.texBounds[1] = Vector4f(0, 0, 1, 1);
                v.texBounds[2] = Vector4f(0, 0, 1, 1);
                v.texBounds[3] = Vector4f(0, 0, 1, 1);
                auto const &meshTextures = material.meshTextures[m];
                for (int t = 0; t < MAX_TEXTURES; ++t) {
                    TextureMap map = glData.textureOrder[t];
                    if (map < 0 || map >= MAX_TEXTURES) continue;
                    if (meshTextures.texIds[map]) {
                        v.texBounds[t] = glData.textureBank.atlas()->imageRectf(meshTextures.texIds[map]).xywh();
                    }
                    else if (glData.defaultTexIds[map]) {
                        v.texBounds[t] = glData.textureBank.atlas()->imageRectf(glData.defaultTexIds[map]).xywh();
                    }
                    else {
                        v.texBounds[t] = Vector4f();
                    }
                }
                for (int b = 0; b < MAX_BONES_PER_VERTEX; ++b) {
                    v.boneIds[b]     = vertexBones[base + i].ids[b];
                    v.boneWeights[b] = vertexBones[base + i].weights[b];
                }
                verts << v;
            }
            duint firstFace = indx.size();
            for (duint i = 0; i < mesh.mNumFaces; ++i) {
                aiFace const &face = mesh.mFaces[i];
                DENG2_ASSERT(face.mNumIndices == 3); // expecting triangles
                indx << face.mIndices[0] + base
                     << face.mIndices[1] + base
                     << face.mIndices[2] + base;
            }
            meshIndexRanges[m] = Rangeui::fromSize(firstFace, mesh.mNumFaces * 3);
            base += mesh.mNumVertices;
        }
        std::unique_ptr<VBuf> buf(new VBuf);
        buf->setVertices(verts, gl::Static);
        buf->setIndices(gl::Triangles, indx, gl::Static);
        material.buffer = std::move(buf);
    }
    struct AccumData {
        Animator const &animator;
        ddouble time = 0.0;
        aiAnimation const *anim = nullptr;
        QVector<Matrix4f> finalTransforms;
        AccumData(Animator const &animator, int boneCount)
            : animator(animator)
            , finalTransforms(boneCount)
        {}
        aiNodeAnim const *findNodeAnim(aiNode const &node) const {
            if (!anim) return nullptr;
            for (duint i = 0; i < anim->mNumChannels; ++i) {
                aiNodeAnim const *na = anim->mChannels[i];
                if (na->mNodeName == node.mName) {
                    return na;
                }
            }
            return nullptr;
        }
    };
    void accumulateAnimationTransforms(Animator const &animator,
                                       ddouble time,
                                       aiAnimation const *animSeq,
                                       aiNode const &rootNode) const {
        AccumData data(animator, boneCount());
        data.anim = animSeq;
        data.time = animSeq? std::fmod(secondsToTicks(time, *animSeq), animSeq->mDuration) : time;
        accumulateTransforms(rootNode, data);
        for (int i = 0; i < boneCount(); ++i) {
            uBoneMatrices.set(i, data.finalTransforms.at(i));
        }
    }
    void accumulateTransforms(aiNode const &node, AccumData &data,
                              Matrix4f const &parentTransform = Matrix4f()) const {
        Matrix4f nodeTransform = convertMatrix(node.mTransformation);
        Vector4f const axisAngle = data.animator.extraRotationForNode(node.mName.C_Str());
        if (aiNodeAnim const *anim = data.findNodeAnim(node)) {
            Matrix4f const translation = Matrix4f::translate(interpolatePosition(data.time, *anim));
            Matrix4f const scaling     = Matrix4f::scale(interpolateScaling(data.time, *anim));
            Matrix4f       rotation    = convertMatrix(aiMatrix4x4(interpolateRotation(data.time, *anim).GetMatrix()));
            if (!fequal(axisAngle.w, 0)) {
                rotation = Matrix4f::rotate(axisAngle.w, axisAngle) * rotation;
            }
            nodeTransform = translation * rotation * scaling;
        }
        else {
            if (!fequal(axisAngle.w, 0)) {
                nodeTransform = Matrix4f::rotate(axisAngle.w, axisAngle) * nodeTransform;
            }
        }
        Matrix4f globalTransform = parentTransform * nodeTransform;
        int boneIndex = findBone(String(node.mName.C_Str()));
        if (boneIndex >= 0) {
            data.finalTransforms[boneIndex] =
                    globalInverse * globalTransform * bones.at(boneIndex).offset;
        }
        for (duint i = 0; i < node.mNumChildren; ++i) {
            accumulateTransforms(*node.mChildren[i], data, globalTransform);
        }
    }
    template <typename Type>
    static duint findAnimKey(ddouble time, Type const *keys, duint count) {
        DENG2_ASSERT(count > 0);
        for (duint i = 0; i < count - 1; ++i) {
            if (time < keys[i + 1].mTime) {
                return i;
            }
        }
        DENG2_ASSERT(!"Failed to find animation key (invalid time?)");
        return 0;
    }
    static Vector3f interpolateVectorKey(ddouble time, aiVectorKey const *keys, duint at) {
        Vector3f const start(&keys[at]    .mValue.x);
        Vector3f const end  (&keys[at + 1].mValue.x);
        return start + (end - start) *
               float((time - keys[at].mTime) / (keys[at + 1].mTime - keys[at].mTime));
    }
    static aiQuaternion interpolateRotation(ddouble time, aiNodeAnim const &anim) {
        if (anim.mNumRotationKeys == 1) {
            return anim.mRotationKeys[0].mValue;
        }
        aiQuatKey const *key = anim.mRotationKeys + findAnimKey(time, anim.mRotationKeys, anim.mNumRotationKeys);
        aiQuaternion interp;
        aiQuaternion::Interpolate(interp, key[0].mValue, key[1].mValue,
                                  (time - key[0].mTime) / (key[1].mTime - key[0].mTime));
        interp.Normalize();
        return interp;
    }
    static Vector3f interpolateScaling(ddouble time, aiNodeAnim const &anim) {
        if (anim.mNumScalingKeys == 1) {
            return Vector3f(&anim.mScalingKeys[0].mValue.x);
        }
        return interpolateVectorKey(time, anim.mScalingKeys,
                                    findAnimKey(time, anim.mScalingKeys,
                                                anim.mNumScalingKeys));
    }
    static Vector3f interpolatePosition(ddouble time, aiNodeAnim const &anim) {
        if (anim.mNumPositionKeys == 1) {
            return Vector3f(&anim.mPositionKeys[0].mValue.x);
        }
        return interpolateVectorKey(time, anim.mPositionKeys,
                                    findAnimKey(time, anim.mPositionKeys,
                                                anim.mNumPositionKeys));
    }
    void updateMatricesFromAnimation(Animator const *animator) const {
        if (!animator) return;
        if (!scene->HasAnimations() || !animator->count()) {
            if (animator->flags().testFlag(Animator::AlwaysTransformNodes)) {
                accumulateAnimationTransforms(*animator, 0, nullptr, *scene->mRootNode);
                return;
            }
        }
        for (int i = 0; i < animator->count(); ++i) {
            auto const &animSeq = animator->at(i);
            DENG2_ASSERT(duint(animSeq.animId) < scene->mNumAnimations);
            DENG2_ASSERT(nodeNameToPtr.contains(animSeq.node));
            accumulateAnimationTransforms(*animator,
                                          animator->currentTime(i),
                                          scene->mAnimations[animSeq.animId],
                                          *nodeNameToPtr[animSeq.node]);
        }
    }
    GLProgram *drawProgram = nullptr;
    Pass const *drawPass = nullptr;
    void preDraw(Animator const *animation) {
        if (glData.needMakeBuffer) makeBuffer();
        DENG2_ASSERT(drawProgram == nullptr);
        updateMatricesFromAnimation(animation);
        GLState::current().apply();
    }
    void setDrawProgram(GLProgram *prog, Appearance const *appearance = nullptr) {
        if (drawProgram) {
            drawProgram->unbind(uBoneMatrices);
            if (appearance && appearance->programCallback) {
                appearance->programCallback(*drawProgram, Unbound);
            }
        }
        drawProgram = prog;
        if (drawProgram) {
            if (appearance && appearance->programCallback) {
                appearance->programCallback(*drawProgram, AboutToBind);
            }
            drawProgram->bind(uBoneMatrices);
        }
    }
    void initRanges(GLBuffer::DrawRanges &ranges, QBitArray const &meshes) {
        Rangeui current;
        for (int i = 0; i < meshIndexRanges.size(); ++i) {
            if (!meshes.at(i)) continue;
            auto const &mesh = meshIndexRanges.at(i);
            if (current.isEmpty()) {
                current = mesh;
            }
            else if (current.end == mesh.start) {
                current.end = mesh.end;
            }
            else {
                ranges.append(current);
                current = mesh;
            }
        }
        if (!current.isEmpty()) {
            ranges.append(current);
        }
    }
    void draw(Appearance const *appearance, Animator const *animation) {
        Passes const *passes = appearance && appearance->drawPasses?
                    appearance->drawPasses : &defaultPasses;
        preDraw(animation);
        try {
            GLBuffer::DrawRanges ranges;
            for (int i = 0; i < passes->size(); ++i) {
                Pass const &pass = passes->at(i);
                if (appearance && !appearance->passMask.isEmpty() &&
                                  !appearance->passMask.testBit(i)) {
                    continue;
                }
                drawPass = &pass;
                setDrawProgram(pass.program? pass.program : program, appearance);
                if (!drawProgram) {
                    throw ProgramError("ModelDrawable::draw",
                                       QString("Rendering pass %1 (\"%2\") has no shader program")
                                        .arg(i).arg(pass.name));
                }
                if (appearance && appearance->passCallback) {
                    appearance->passCallback(pass, PassBegun);
                }
                duint material = 0;
                if (appearance && appearance->passMaterial.size() >= passes->size()) {
                    material = appearance->passMaterial.at(i);
                }
                ranges.clear();
                initRanges(ranges, pass.meshes);
                GLState::push()
                        .setBlendFunc (pass.blendFunc)
                        .setBlendOp   (pass.blendOp)
                        .setDepthTest (pass.depthFunc != gl::Always)
                        .setDepthFunc (pass.depthFunc)
                        .setDepthWrite(pass.depthWrite)
                        .apply(); {
                    drawProgram->beginUse();
                    glData.materials.at(material)->buffer->draw(&ranges);
                    drawProgram->endUse();
                }
                GLState::pop();
                if (appearance && appearance->passCallback) {
                    appearance->passCallback(pass, PassEnded);
                }
            }
        }
        catch (Error const &er) {
            LOG_GL_ERROR("Failed to draw model \"%s\": %s")
                    << sourcePath
                    << er.asText();
        }
        postDraw();
    }
    void drawInstanced(GLBuffer const &attribs, Animator const *animation) {
        preDraw(animation);
        setDrawProgram(program);
        glData.materials.at(0)->buffer->drawInstanced(attribs);
        postDraw();
    }
    void postDraw() {
        setDrawProgram(nullptr);
        drawPass = nullptr;
    }
    DENG2_PIMPL_AUDIENCE(AboutToGLInit)
};
DENG2_AUDIENCE_METHOD(ModelDrawable, AboutToGLInit)
namespace internal {
static struct {
    String const text;
    ModelDrawable::TextureMap map;
} const mappings[] {
    { "diffuse",  ModelDrawable::Diffuse  },
    { "normals",  ModelDrawable::Normals  },
    { "specular", ModelDrawable::Specular },
    { "emission", ModelDrawable::Emissive },
    { "height",   ModelDrawable::Height   },
    { "unknown",  ModelDrawable::Unknown  }
};
} // namespace internal
ModelDrawable::TextureMap ModelDrawable::textToTextureMap(String const &text) // static
{
    for (auto const &mapping : internal::mappings) {
        if (!text.compareWithoutCase(mapping.text))
            return mapping.map;
    }
    return Unknown;
}
String ModelDrawable::textureMapToText(TextureMap map) // static
{
    for (auto const &mapping : internal::mappings) {
        if (mapping.map == map)
            return mapping.text;
    }
    return QStringLiteral("unknown");
}
ModelDrawable::ModelDrawable() : d(new Impl(this)) {
    *this += d->modelAsset;
}
void ModelDrawable::setImageLoader(IImageLoader &loader) {
    d->glData.imageLoader = &loader;
}
void ModelDrawable::useDefaultImageLoader() {
    d->glData.imageLoader = &defaultImageLoader;
}
void ModelDrawable::load(File const &file) {
    LOG_AS("ModelDrawable");
    clear();
    d->import(file);
}
void ModelDrawable::clear() {
    glDeinit();
    d->clear();
}
int ModelDrawable::animationIdForName(String const &name) const {
    Impl::AnimLookup::const_iterator found = d->animNameToIndex.constFind(name);
    if (found != d->animNameToIndex.constEnd()) {
        return found.value();
    }
    return -1;
}
String ModelDrawable::animationName(int id) const {
    if (!d->scene || id < 0 || id >= int(d->scene->mNumAnimations)) {
        return "";
    }
    String const name = d->scene->mAnimations[id]->mName.C_Str();
    if (name.isEmpty()) {
        return QString("@%1").arg(id);
    }
    return name;
}
int ModelDrawable::animationCount() const {
    if (!d->scene) return 0;
    return d->scene->mNumAnimations;
}
int ModelDrawable::meshCount() const {
    if (!d->scene) return 0;
    return d->scene->mNumMeshes;
}
int ModelDrawable::meshId(String const &name) const {
    if (!d->scene) return -1;
    for (duint i = 0; i < d->scene->mNumMeshes; ++i) {
        if (name == d->scene->mMeshes[i]->mName.C_Str()) {
            return i;
        }
    }
    return -1;
}
String ModelDrawable::meshName(int id) const {
    if (!d->scene || id < 0 || id >= int(d->scene->mNumMeshes)) {
        return "";
    }
    String const name = d->scene->mMeshes[id]->mName.C_Str();
    if (name.isEmpty()) {
        return QString("@%1").arg(id);
    }
    return name;
}
bool ModelDrawable::nodeExists(String const &name) const {
    return d->nodeNameToPtr.contains(name);
}
void ModelDrawable::setAtlas(IAtlas &atlas) {
    d->glData.textureBank.setAtlas(&atlas);
}
void ModelDrawable::unsetAtlas() {
    d->glData.releaseTexturesFromAtlas();
    d->glData.textureBank.setAtlas(nullptr);
}
IAtlas *ModelDrawable::atlas() const {
    return d->glData.textureBank.atlas();
}
ModelDrawable::Mapping ModelDrawable::diffuseNormalsSpecularEmission() // static
{
    return Mapping() << Diffuse << Normals << Specular << Emissive;
}
duint ModelDrawable::addMaterial() {
    DENG2_ASSERT(!d->modelAsset.isReady());
    return d->glData.addMaterial();
}
void ModelDrawable::resetMaterials() {
    DENG2_ASSERT(!d->modelAsset.isReady());
    d->glData.deinitMaterials();
    d->glData.initMaterials();
}
void ModelDrawable::setTextureMapping(Mapping mapsToUse) {
    d->glData.setTextureMapping(mapsToUse);
}
void ModelDrawable::setDefaultTexture(TextureMap textureType, Id const &atlasId) {
    DENG2_ASSERT(textureType >= 0 && textureType < MAX_TEXTURES);
    if (textureType < 0 || textureType >= MAX_TEXTURES) return;
    d->glData.defaultTexIds[textureType] = atlasId;
}
void ModelDrawable::glInit() {
    d->glInit();
}
void ModelDrawable::glDeinit() {
    d->glDeinit();
}
int ModelDrawable::materialId(String const &name) const {
    return d->findMaterial(name);
}
void ModelDrawable::setTexturePath(MeshId const &mesh, TextureMap tex, String const &path) {
    if (d->glData.textureBank.atlas()) {
        d->glData.setTexture(mesh, tex, path);
    }
    else {
        d->glData.setCustomTexturePath(mesh, tex, path);
    }
}
void ModelDrawable::setProgram(GLProgram *program) {
    d->program = program;
}
GLProgram *ModelDrawable::program() const {
    return d->program;
}
void ModelDrawable::draw(Appearance const *appearance,
                           Animator const *animation) const {
    const_cast<ModelDrawable *>(this)->glInit();
    if (isReady() && d->glData.textureBank.atlas()) {
        d->draw(appearance, animation);
    }
}
void ModelDrawable::drawInstanced(GLBuffer const &instanceAttribs,
                                  Animator const *animation) const {
    const_cast<ModelDrawable *>(this)->glInit();
    if (isReady() && d->program && d->glData.textureBank.atlas()) {
        d->drawInstanced(instanceAttribs, animation);
    }
}
ModelDrawable::Pass const *ModelDrawable::currentPass() const {
    return d->drawPass;
}
GLProgram *ModelDrawable::currentProgram() const {
    return d->drawProgram;
}
Vector3f ModelDrawable::dimensions() const {
    return d->maxPoint - d->minPoint;
}
Vector3f ModelDrawable::midPoint() const {
    return (d->maxPoint + d->minPoint) / 2.f;
}
int ModelDrawable::Passes::findName(String const &name) const {
    for (int i = 0; i < size(); ++i) {
        if (at(i).name == name) // case sensitive
            return i;
    }
    return -1;
}
} // namespace de
namespace de {
DENG2_PIMPL_NOREF(ModelDrawable::Animator)
, DENG2_OBSERVES(Asset, Deletion) {
    Constructor constructor;
    ModelDrawable const *model = nullptr;
    QList<OngoingSequence *> anims;
    Flags flags = DefaultFlags;
    Impl(Constructor ctr, ModelDrawable const *mdl = 0)
        : constructor(ctr) {
        setModel(mdl);
    }
    ~Impl() {
        setModel(nullptr);
        qDeleteAll(anims);
    }
    void setModel(ModelDrawable const *mdl) {
        if (model) model->audienceForDeletion() -= this;
        model = mdl;
        if (model) model->audienceForDeletion() += this;
    }
    void assetBeingDeleted(Asset &a) {
        if (model == &a) model = nullptr;
    }
    OngoingSequence &add(OngoingSequence *seq) {
        DENG2_ASSERT(seq != nullptr);
        DENG2_ASSERT(model != nullptr);
        if (seq->animId < 0 || seq->animId >= model->animationCount()) {
            throw InvalidError("ModelDrawable::Animator::add",
                               "Specified animation does not exist");
        }
        if (!model->nodeExists(seq->node)) {
            throw InvalidError("ModelDrawable::Animator::add",
                               "Node '" + seq->node + "' does not exist");
        }
        anims.append(seq);
        return *anims.last();
    }
    void stopByNode(String const &node) {
        QMutableListIterator<OngoingSequence *> iter(anims);
        while (iter.hasNext()) {
            iter.next();
            if (iter.value()->node == node) {
                delete iter.value();
                iter.remove();
            }
        }
    }
    OngoingSequence const *findAny(String const &rootNode) const {
        foreach (OngoingSequence const *anim, anims) {
            if (anim->node == rootNode)
                return anim;
        }
        return nullptr;
    }
    OngoingSequence const *find(int animId, String const &rootNode) const {
        foreach (OngoingSequence const *anim, anims) {
            if (anim->animId == animId && anim->node == rootNode)
                return anim;
        }
        return nullptr;
    }
    bool isRunning(int animId, String const &rootNode) const {
        return find(animId, rootNode) != nullptr;
    }
};
ModelDrawable::Animator::Animator(Constructor constructor)
    : d(new Impl(constructor))
{}
ModelDrawable::Animator::Animator(ModelDrawable const &model, Constructor constructor)
    : d(new Impl(constructor, &model))
{}
void ModelDrawable::Animator::setModel(ModelDrawable const &model) {
    d->setModel(&model);
}
void ModelDrawable::Animator::setFlags(Flags const &flags, FlagOp op) {
    applyFlagOperation(d->flags, flags, op);
}
ModelDrawable::Animator::Flags ModelDrawable::Animator::flags() const {
    return d->flags;
}
ModelDrawable const &ModelDrawable::Animator::model() const {
    DENG2_ASSERT(d->model != 0);
    return *d->model;
}
int ModelDrawable::Animator::count() const {
    return d->anims.size();
}
ModelDrawable::Animator::OngoingSequence const &
ModelDrawable::Animator::at(int index) const {
    return *d->anims.at(index);
}
ModelDrawable::Animator::OngoingSequence &
ModelDrawable::Animator::at(int index) {
    return *d->anims[index];
}
bool ModelDrawable::Animator::isRunning(String const &animName, String const &rootNode) const {
    return d->isRunning(model().animationIdForName(animName), rootNode);
}
bool ModelDrawable::Animator::isRunning(int animId, String const &rootNode) const {
    return d->isRunning(animId, rootNode);
}
ModelDrawable::Animator::OngoingSequence *ModelDrawable::Animator::find(String const &rootNode) const {
    return const_cast<OngoingSequence *>(d->findAny(rootNode));
}
ModelDrawable::Animator::OngoingSequence *ModelDrawable::Animator::find(int animId, String const &rootNode) const {
    return const_cast<OngoingSequence *>(d->find(animId, rootNode));
}
ModelDrawable::Animator::OngoingSequence &
ModelDrawable::Animator::start(String const &animName, String const &rootNode) {
    return start(model().animationIdForName(animName), rootNode);
}
ModelDrawable::Animator::OngoingSequence &
ModelDrawable::Animator::start(int animId, String const &rootNode) {
    d->stopByNode(rootNode);
    aiScene const &scene = *model().d->scene;
    if (animId < 0 || animId >= int(scene.mNumAnimations)) {
        throw InvalidError("ModelDrawable::Animator::start",
                           QString("Invalid animation ID %1").arg(animId));
    }
    auto const &animData = *scene.mAnimations[animId];
    OngoingSequence *anim = d->constructor();
    anim->animId = animId;
    anim->node   = rootNode;
    anim->time   = 0.0;
    anim->duration = ticksToSeconds(animData.mDuration, animData);
    anim->initialize();
    return d->add(anim);
}
void ModelDrawable::Animator::stop(int index) {
    d->anims.removeAt(index);
}
void ModelDrawable::Animator::clear() {
    qDeleteAll(d->anims);
    d->anims.clear();
}
void ModelDrawable::Animator::advanceTime(TimeSpan const &) {
}
ddouble ModelDrawable::Animator::currentTime(int index) const {
    auto const &anim = at(index);
    ddouble t = anim.time;
    if (anim.flags.testFlag(OngoingSequence::ClampToDuration)) {
        t = min(t, anim.duration - de::FLOAT_EPSILON);
    }
    return t;
}
Vector4f ModelDrawable::Animator::extraRotationForNode(String const &) const {
    return Vector4f();
}
void ModelDrawable::Animator::operator >> (Writer &to) const {
    to.writeObjects(d->anims);
}
void ModelDrawable::Animator::operator << (Reader &from) {
    clear();
    from.readObjects<OngoingSequence>(d->anims, [this] () { return d->constructor(); });
}
void ModelDrawable::Animator::OngoingSequence::initialize()
{}
bool ModelDrawable::Animator::OngoingSequence::atEnd() const {
    return time >= duration;
}
void ModelDrawable::Animator::OngoingSequence::operator >> (Writer &to) const {
    to << animId
       << time
       << duration
       << node
       << duint32(flags);
}
void ModelDrawable::Animator::OngoingSequence::operator << (Reader &from) {
    from >> animId
         >> time
         >> duration
         >> node;
    from.readAs<duint32>(flags);
}
ModelDrawable::Animator::OngoingSequence *
ModelDrawable::Animator::OngoingSequence::make() // static
{
    return new OngoingSequence;
}
} // namespace de
uint qHash(de::ModelDrawable::Pass const &pass) {
    return qHash(pass.name);
}
namespace de {
typedef QHash<Id::Type, Image *> PendingImages;
DENG2_PIMPL(MultiAtlas)
, public Deletable {
    IAtlasFactory &factory;
    QList<Atlas *> atlases;
    Impl(Public *i, IAtlasFactory &factory)
        : Base(i)
        , factory(factory)
    {}
    ~Impl() {
        release();
    }
    void release() {
        qDeleteAll(atlases);
        atlases.clear();
    }
    Atlas *getEmptyAtlas() {
        foreach (Atlas *atlas, atlases) {
            if (atlas->isEmpty()) return atlas;
        }
        Atlas *blank = factory.makeAtlas(self());
        DENG2_ASSERT(blank->flags().testFlag(Atlas::DeferredAllocations));
        atlases.prepend(blank);
        return blank;
    }
    bool tryAllocatePending(Atlas &atlas, PendingImages const &pending) {
        DENG2_ASSERT(atlas.flags().testFlag(Atlas::DeferredAllocations));
        for (auto i = pending.begin(); i != pending.end(); ++i) {
            if (!atlas.alloc(*i.value(), i.key())) {
                atlas.cancelDeferred();
                return false;
            }
        }
        atlas.commit();
        return true;
    }
    /**
     * Selects a suitable atlas for the provided set of images. All the images
     * must fit on a single atlas. A new atlas is created if necessary.
     *
     * @param pendingAllocs  Set of images to be allocated.
     *
     * @return Atlas object. Caller must commit the deferred allocations.
     */
    Atlas &allocatePending(PendingImages const &pending) {
        foreach (Atlas *atlas, atlases) {
            if (tryAllocatePending(*atlas, pending)) {
                return *atlas;
            }
        }
        Atlas *blank = getEmptyAtlas();
        if (tryAllocatePending(*blank, pending)) {
            return *blank;
        }
        throw IAtlas::OutOfSpaceError("MultiAtlas::allocatePending",
                                      "Even an empty atlas cannot fit the pending allocations");
    }
};
MultiAtlas::MultiAtlas(IAtlasFactory &factory)
    : d(new Impl(this, factory))
{}
void MultiAtlas::clear() {
    d->release();
}
} // namespace de
namespace de {
DENG2_PIMPL_NOREF(MultiAtlas::AllocGroup)
, DENG2_OBSERVES(Deletable, Deletion) {
    AllocGroup *self;
    MultiAtlas *owner;
    PendingImages pending; // owned
    Atlas *atlas = nullptr; ///< Committed atlas.
    QSet<Id::Type> allocated; // committed to the atlas
    Impl(AllocGroup *i, MultiAtlas &owner)
        : self(i)
        , owner(&owner) {
        this->owner->d->audienceForDeletion += this;
    }
    ~Impl() {
        if (owner) {
            release();
        }
    }
    void objectWasDeleted(Deletable *deleted) {
        if (deleted == owner->d) {
            owner = nullptr;
            if (atlas) atlas->audienceForDeletion -= this;
            atlas = nullptr;
            cancelPending();
        }
        else {
            DENG2_ASSERT(deleted == atlas);
            atlas = nullptr;
        }
        allocated.clear();
        self->setState(Asset::NotReady);
    }
    void cancelPending() {
        qDeleteAll(pending.values());
        pending.clear();
    }
    void release() {
        cancelPending();
        if (atlas) {
            for (auto i : allocated) {
                atlas->release(i);
            }
        }
        allocated.clear();
    }
};
MultiAtlas::AllocGroup::AllocGroup(MultiAtlas &multiAtlas)
    : d(new Impl(this, multiAtlas))
{}
Id MultiAtlas::AllocGroup::alloc(Image const &image, Id const &knownId) {
    if (!d->atlas) {
        Id allocId { knownId.isNone()? Id() : knownId };
        d->pending.insert(allocId, new Image(image));
        return allocId;
    }
    else {
        Id allocId { d->atlas->alloc(image, knownId) };
        d->allocated.insert(allocId);
        return allocId;
    }
}
void MultiAtlas::AllocGroup::release(Id const &id) {
    auto foundPending = d->pending.constFind(id);
    if (foundPending != d->pending.constEnd()) {
        delete foundPending.value();
        d->pending.remove(id);
        return;
    }
    if (d->atlas && d->allocated.contains(id)) {
        d->allocated.remove(id);
        d->atlas->release(id);
    }
}
bool MultiAtlas::AllocGroup::contains(Id const &id) const {
    return d->pending.contains(id) || d->allocated.contains(id);
}
void MultiAtlas::AllocGroup::commit() const {
    if (!d->owner) {
        throw InvalidError("MultiAtlas::AllocGroup::commit",
                           "Allocation group has been invalidated");
    }
    if (!d->atlas) {
        d->atlas = &d->owner->d->allocatePending(d->pending);
        d->atlas->audienceForDeletion += d;
    }
    for (auto i = d->pending.begin(); i != d->pending.end(); ++i) {
        d->allocated.insert(i.key());
        delete i.value(); // free the Image
    }
    d->pending.clear();
    const_cast<MultiAtlas::AllocGroup *>(this)->setState(Ready);
}
Rectanglef MultiAtlas::AllocGroup::imageRectf(Id const &id) const {
    if (d->atlas) {
        return d->atlas->imageRectf(id);
    }
    throw InvalidError("MultiAtlas::AllocGroup::imageRectf",
                       "Allocation group has not yet been committed to an atlas");
}
Atlas const *MultiAtlas::AllocGroup::atlas() const {
    return d->atlas;
}
MultiAtlas &MultiAtlas::AllocGroup::multiAtlas() {
    DENG2_ASSERT(d->owner != nullptr);
    return *d->owner;
}
} // namespace de
namespace de {
DENG2_PIMPL_NOREF(PackageIconBank) {
    class PackageImageSource : public TextureBank::ImageSource {
    public:
        PackageImageSource(Path const &packagePath,
                           Size const &displaySize)
            : ImageSource(packagePath)
            , _displaySize(displaySize) {
            DENG2_ASSERT(packagePath.lastSegment().toStringRef().endsWith(".pack"));
        }
        Image load() const override {
            const String iconPath = sourcePath() / QStringLiteral("icon");
            Image img;
            if (ImageFile const *file = FS::tryLocate<ImageFile const>(iconPath + ".jpg")) {
                img = file->image();
            }
            else if (ImageFile const *file = FS::tryLocate<ImageFile const>(iconPath + ".png")) {
                img = file->image();
            }
            if (!img.isNull()) {
                if (img.height() > img.width()) {
                    duint const off = img.height() - img.width();
                    img = img.subImage(Rectanglei(0, off/2, img.width(), img.width()));
                }
                else if (img.width() > img.height()) {
                    duint const off = img.width() - img.height();
                    img = img.subImage(Rectanglei(off/2, 0, img.height(), img.height()));
                }
                if (img.width() > _displaySize.x) {
                    img.resize(_displaySize);
                }
            }
            img.setPointRatio(1.f);
            return img;
        }
    private:
        Size _displaySize;
    };
    Size displaySize;
    static bool doesPackageHaveIcon(const Path &packagePath) {
        return FS::exists(packagePath / "icon.jpg") ||
               FS::exists(packagePath / "icon.png");
    }
};
PackageIconBank::PackageIconBank()
    : TextureBank("PackageIconBank", BackgroundThread | DisableHotStorage)
    , d(new Impl) {
    setSeparator('/'); // expect package paths as keys
}
void PackageIconBank::setDisplaySize(Size const &displaySize) {
    d->displaySize = displaySize;
}
Id PackageIconBank::packageIcon(File const &packageFile) {
    const Path packagePath = packageFile.path();
    if (!has(packagePath)) {
        add(packagePath, new Impl::PackageImageSource(packagePath, d->displaySize));
    }
    if (isLoaded(packagePath)) {
        return TextureBank::texture(packagePath);
    }
    LOG_AS("PackageIconBank");
    LOG_RES_VERBOSE("Loading icon from \"%s\"") << packagePath.toString();
    load(packagePath, BeforeQueued);
    return Id::None;
}
bool PackageIconBank::packageContainsIcon(const de::File &packageFile) const {
    return Impl::doesPackageHaveIcon(packageFile.path());
}
} // namespace de
namespace de {
template <typename Type>
void linkAfter(Type *where, Type *object) {
    object->next = where->next;
    object->prev = where;
    if (where->next) where->next->prev = object;
    where->next = object;
}
template <typename Type>
void unlink(Type *object) {
    if (object->prev) object->prev->next = object->next;
    if (object->next) object->next->prev = object->prev;
    object->next = object->prev = nullptr;
}
static float const OPTIMIZATION_USAGE_THRESHOLD = .7f;
DENG2_PIMPL(RowAtlasAllocator) {
    struct Rows {
        struct Row;
        /**
         * Each row is composed of one or more used or empty slots.
         */
        struct Slot {
            Slot *next = nullptr;
            Slot *prev = nullptr;
            Row * row;
            Id    id{Id::None}; ///< Id of allocation here, or Id::None if free.
            int   x        = 0; ///< Left edge of the slot.
            duint width    = 0; ///< Width of the slot.
            dsize usedArea = 0;
            Slot(Row *owner) : row(owner) {}
            bool isEmpty() const {
                return id.isNone();
            }
            /**
             * Take an empty slot into use. The remaining empty space is split off
             * into a new slot.
             *
             * @param allocId  Allocation identifier.
             * @param widthWithMargin  Needed width.
             *
             * @return If an empty slot was created, it is returned. Otherwise @c nullptr.
             */
            Slot *allocateAndSplit(Id const &allocId, duint widthWithMargin) {
                DENG2_ASSERT(isEmpty());
                DENG2_ASSERT(width >= widthWithMargin);
                int const remainder = width - widthWithMargin;
                id    = allocId;
                width = widthWithMargin;
                if (remainder > 0) {
                    Slot *split = new Slot(row);
                    linkAfter(this, split);
                    split->x = x + width;
                    split->width = remainder;
                    return split;
                }
                return nullptr;
            }
            Slot *mergeWithNext() {
                DENG2_ASSERT(isEmpty());
                if (!next || !next->isEmpty()) return nullptr;
                Slot *merged = next;
                unlink(merged);
                width += merged->width;
                return merged; // Caller gets ownership.
            }
            Slot *mergeWithPrevious() {
                DENG2_ASSERT(isEmpty());
                if (!prev || !prev->isEmpty()) return nullptr;
                Slot *merged = prev;
                unlink(merged);
                if (row->first == merged) {
                    row->first = this;
                }
                x -= merged->width;
                width += merged->width;
                return merged; // Caller gets ownership.
            }
            struct SortByWidth {
                bool operator () (Slot const *a, Slot const *b) const {
                    if (a->width == b->width) return a < b;
                    return a->width > b->width;
                }
            };
        };
        struct Row {
            Row *next = nullptr;
            Row *prev = nullptr;
            int y = 0;       ///< Top edge of the row.
            duint height = 0;
            Slot *first;    ///< There's always at least one empty slot.
            Row() : first(new Slot(this)) {}
            ~Row() {
                Slot *next;
                for (Slot *s = first; s; s = next) {
                    next = s->next;
                    delete s;
                }
            }
            bool isEmpty() const {
                return first->isEmpty() && !first->next;
            }
            bool isTallEnough(duint heightWithMargin) const {
                if (height >= heightWithMargin) return true;
                if (next && next->isEmpty()) {
                    return (height + next->height) >= heightWithMargin;
                }
                return false;
            }
            Row *split(duint newHeight) {
                DENG2_ASSERT(isEmpty());
                DENG2_ASSERT(newHeight <= height);
                duint const remainder = height - newHeight;
                height = newHeight;
                if (remainder > 0) {
                    Row *below = new Row;
                    linkAfter(this, below);
                    below->y = y + height;
                    below->height = remainder;
                    return below;
                }
                return nullptr;
            }
            void grow(duint newHeight) {
                DENG2_ASSERT(newHeight > height);
                DENG2_ASSERT(next);
                DENG2_ASSERT(next->isEmpty());
                duint delta = newHeight - height;
                height += delta;
                next->y += delta;
                next->height -= delta;
            }
        };
        Row *top; ///< Always at least one row exists.
        std::set<Slot *, Slot::SortByWidth> vacant; // not owned
        QHash<Id, Slot *> slotsById; // not owned
        dsize usedArea = 0; ///< Total allocated pixels.
        Impl *d;
        Rows(Impl *inst) : d(inst) {
            top = new Row;
            /*
             * Set up one big row, excluding the margins. This is all the space that
             * we will be using; it will be chopped up and merged back together, but
             * space will not be added or removed. Margin is reserved on the top/left
             * edge; individual slots reserve it on the right, rows reserve it in the
             * bottom.
             */
            top->y            = d->margin;
            top->height       = d->size.y - d->margin;
            top->first->x     = d->margin;
            top->first->width = d->size.x - d->margin;
            addVacant(top->first);
        }
        ~Rows() {
            Row *next;
            for (Row *r = top; r; r = next) {
                next = r->next;
                delete r;
            }
        }
        void addVacant(Slot *slot) {
            DENG2_ASSERT(slot->isEmpty());
            vacant.insert(slot);
            DENG2_ASSERT(*vacant.find(slot) == slot);
        }
        void removeVacant(Slot *slot) {
            DENG2_ASSERT(vacant.find(slot) != vacant.end());
            vacant.erase(slot);
            DENG2_ASSERT(vacant.find(slot) == vacant.end());
        }
        Slot *findBestVacancy(Atlas::Size const &size) const {
            Slot *best = nullptr;
            for (Slot *s : vacant) {
                if (s->width >= size.x + d->margin) {
                    if (s->row->isTallEnough(size.y + d->margin)) {
                        best = s;
                    }
                }
                else {
                    break;
                }
            }
            return best;
        }
        /**
         * Allocate a slot for the specified size. The area used by the slot may be
         * larger than the requested size.
         *
         * @param size  Dimensions of area to allocate.
         * @param rect  Allocated rectangle is returned here.
         * @param id    Id for the new slot.
         *
         * @return Allocated slot, or @c nullptr.
         */
        Slot *alloc(Atlas::Size const &size, Rectanglei &rect, Id::Type id = Id::None) {
            Slot *slot = findBestVacancy(size);
            if (!slot) return nullptr;
            DENG2_ASSERT(slot->isEmpty());
            removeVacant(slot);
            Atlas::Size const needed = size + Atlas::Size(d->margin, d->margin);
            if (slot->row->isEmpty()) {
                if (Row *addedRow = slot->row->split(needed.y)) {
                    addedRow->first->x = d->margin;
                    addedRow->first->width = d->size.x - d->margin;
                    addVacant(addedRow->first);
                }
            }
            if (slot->row->height < needed.y) {
                slot->row->grow(needed.y);
            }
            if (Slot *addedSlot = slot->allocateAndSplit(id? Id(id) : Id(), needed.x)) {
                addVacant(addedSlot);
            }
            slotsById.insert(slot->id, slot);
            rect = Rectanglei::fromSize(Vector2i(slot->x, slot->row->y), size);
            slot->usedArea = size.x * size.y;
            usedArea += slot->usedArea;
            DENG2_ASSERT(usedArea <= d->size.x * d->size.y);
            DENG2_ASSERT(vacant.find(slot) == vacant.end());
            DENG2_ASSERT(!slot->isEmpty());
            return slot;
        }
        void mergeLeft(Slot *slot) {
            if (Slot *removed = slot->mergeWithPrevious()) {
                removeVacant(removed);
                delete removed;
            }
        }
        void mergeRight(Slot *slot) {
            if (Slot *removed = slot->mergeWithNext()) {
                removeVacant(removed);
                delete removed;
            }
        }
        void mergeAbove(Row *row) {
            DENG2_ASSERT(row->isEmpty());
            if (row->prev && row->prev->isEmpty()) {
                Row *merged = row->prev;
                unlink(merged);
                if (top == merged) {
                    top = row;
                }
                row->y -= merged->height;
                row->height += merged->height;
                removeVacant(merged->first);
                delete merged;
            }
        }
        void mergeBelow(Row *row) {
            DENG2_ASSERT(row->isEmpty());
            if (row->next && row->next->isEmpty()) {
                Row *merged = row->next;
                unlink(merged);
                row->height += merged->height;
                removeVacant(merged->first);
                delete merged;
            }
        }
        void release(Id const &id) {
            DENG2_ASSERT(slotsById.contains(id));
            Slot *slot = slotsById.take(id);
            slot->id = Id::None;
            DENG2_ASSERT(slot->usedArea > 0);
            DENG2_ASSERT(usedArea >= slot->usedArea);
            usedArea -= slot->usedArea;
            mergeLeft(slot);
            mergeRight(slot);
            addVacant(slot);
            if (slot->row->isEmpty()) {
                mergeAbove(slot->row);
                mergeBelow(slot->row);
            }
        }
    };
    Atlas::Size size;
    int margin { 0 };
    Allocations allocs;
    std::unique_ptr<Rows> rows;
    Impl(Public *i)
        : Base(i)
        , rows(new Rows(this))
    {}
    struct ContentSize {
        Id::Type id;
        Atlas::Size size;
        ContentSize(Id const &allocId, Vector2ui const &sz) : id(allocId), size(sz) {}
        bool operator < (ContentSize const &other) const {
            if (size.y == other.size.y) {
                return size.x > other.size.x;
            }
            return size.y > other.size.y;
        }
    };
    bool optimize() {
        QList<ContentSize> descending;
        DENG2_FOR_EACH(Allocations, i, allocs) {
            descending.append(ContentSize(i.key(), i.value().size()));
        }
        qSort(descending);
        Allocations optimal;
        std::unique_ptr<Rows> revised(new Rows(this));
        for (auto const &ct : descending) {
            Rectanglei optRect;
            if (!revised->alloc(ct.size, optRect, ct.id)) {
                return false; // Ugh, can't actually fit these.
            }
            optimal[ct.id] = optRect;
        }
        allocs = optimal;
        rows.reset(revised.release());
        return true;
    }
    float usage() const {
        return float(rows->usedArea) / float(size.x * size.y);
    }
};
RowAtlasAllocator::RowAtlasAllocator() : d(new Impl(this))
{}
void RowAtlasAllocator::setMetrics(Atlas::Size const &totalSize, int margin) {
    d->size   = totalSize;
    d->margin = margin;
    DENG2_ASSERT(d->allocs.isEmpty());
    d->rows.reset(new Impl::Rows(d));
}
void RowAtlasAllocator::clear() {
    d->rows.reset(new Impl::Rows(d));
    d->allocs.clear();
}
Id RowAtlasAllocator::allocate(Atlas::Size const &size, Rectanglei &rect,
                               Id const &knownId) {
    if (auto *slot = d->rows->alloc(size, rect, knownId)) {
        d->allocs[slot->id] = rect;
        return slot->id;
    }
    return 0;
}
void RowAtlasAllocator::release(Id const &id) {
    DENG2_ASSERT(d->allocs.contains(id));
    d->rows->release(id);
    d->allocs.remove(id);
}
int RowAtlasAllocator::count() const {
    return d->allocs.size();
}
Atlas::Ids RowAtlasAllocator::ids() const {
    Atlas::Ids ids;
    foreach (Id const &id, d->allocs.keys()) {
        ids.insert(id);
    }
    return ids;
}
void RowAtlasAllocator::rect(Id const &id, Rectanglei &rect) const {
    DENG2_ASSERT(d->allocs.contains(id));
    rect = d->allocs[id];
}
RowAtlasAllocator::Allocations RowAtlasAllocator::allocs() const {
    return d->allocs;
}
bool RowAtlasAllocator::optimize() {
    if (d->usage() >= OPTIMIZATION_USAGE_THRESHOLD) return false;
    return d->optimize();
}
} // namespace de
namespace de {
DENG2_PIMPL_NOREF(TextureBank::ImageSource) {
    DotPath sourcePath;
};
TextureBank::ImageSource::ImageSource(DotPath const &sourcePath) : d(new Impl) {
    d->sourcePath = sourcePath;
}
DotPath const &TextureBank::ImageSource::sourcePath() const {
    return d->sourcePath;
}
DENG2_PIMPL(TextureBank) {
    struct TextureData : public IData {
        Impl *d;
        Id _id { Id::None };
        std::unique_ptr<Image> pendingImage;
        TextureData(Image const &image, Impl *owner) : d(owner) {
            if (image) {
                if (d->atlas) {
                    _id = d->atlas->alloc(image);
                }
                else {
                    pendingImage.reset(new Image(image));
                }
            }
        }
        ~TextureData() {
            if (_id) {
                d->pathForAtlasId.remove(_id);
                d->atlas->release(_id);
            }
        }
        Id const &id() {
            if (pendingImage && d->atlas) {
                _id = d->atlas->alloc(*pendingImage);
                pendingImage.reset();
            }
            return _id;
        }
    };
    IAtlas *atlas { nullptr };
    QHash<Id::Type, String> pathForAtlasId; // reverse lookup
    Impl(Public *i) : Base(i) {}
    ~Impl() {
        self().clear();
    }
};
TextureBank::TextureBank(char const *nameForLog, Flags const &flags)
    : Bank(nameForLog, flags), d(new Impl(this))
{}
void TextureBank::setAtlas(IAtlas *atlas) {
    d->atlas = atlas;
}
IAtlas *TextureBank::atlas() {
    return d->atlas;
}
Id const &TextureBank::texture(DotPath const &id) {
    return data(id).as<Impl::TextureData>().id();
}
Path TextureBank::sourcePathForAtlasId(Id const &id) const {
    auto found = d->pathForAtlasId.constFind(id);
    if (found != d->pathForAtlasId.constEnd()) {
        return found.value();
    }
    return "";
}
Bank::IData *TextureBank::loadFromSource(ISource &source) {
    auto *data = new Impl::TextureData(source.as<ImageSource>().load(), d);
    if (auto const &texId = data->id()) {
        d->pathForAtlasId.insert(texId, source.as<ImageSource>().sourcePath());
    }
    return data;
}
} // namespace de
